// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: dir_cli_ap_udp.cc,v $ - the Directory Service 
 *            Client access point through UDP protocol.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/07/12 08:11:34 $
 * Version :  $Revision: 1.3 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

# include <netinet/in.h>
# include <stdio.h>
# include "sock_conn.h"
# include "dir_cli_ap_udp.h"

# include <string>

/**
 * Status translation.
 * @param status UDP client error code.
 * @return status Dir service error code.
 */
static int udp_to_dir_status(int status)
{
	switch (status) {
		case UDP_RES_OKAY : return DIR_RES_OKAY;
		case UDP_RES_TOUT : return DIR_RES_ETOUT;
		case UDP_RES_SYST : return DIR_RES_ENETWORK;
		case UDP_RES_PORT :
		case UDP_RES_ADDR :
		case UDP_RES_LENG : return DIR_RES_EDATA;
		default			  : return DIR_RES_ECLIENT;
	}
}

/** 
 * DIR_Cli_AP_UDP constructor. 
 * @param ip Server IP address (as a string).
 * @param port Server port number.
 */
DIR_Cli_AP_UDP::DIR_Cli_AP_UDP(const char *addr, int port) :
	udp_cli(NULL, SOCK_PORT_ANY, 1000),
	dir_srv_port(port)
{ 
	// convert to IP address.
	struct sockaddr_in sin;
	char buf[1024];
	if (sock_filladdr(addr, port, &sin)<0) {
		snprintf(buf, 1024, "Bad server address: %s:%d", addr, port);
		throw std::string(buf);
	}

	if (sock_dotaddr(buf, 1024, &(sin.sin_addr))<0) {
		snprintf(buf, 1024, "Bad server address: %s:%d", addr, port);
		throw std::string(buf);
	}

	dir_srv_addr=buf;

	snprintf(buf, 1024, "udp://%s:%d/", addr, port);
	protocol=buf;
}

/** DIR_Cli_AP_UDP destructor. */
DIR_Cli_AP_UDP::~DIR_Cli_AP_UDP() { }

/**
 * Allocate an ident request.
 * @param len An pointer to an interger, will be used 
 * 	to store the size of the request structure
 * @return The request structure.
 */
DIR_REQ *DIR_Cli_AP_UDP::new_req_ident(int *len)
{
	if (!len)
		return NULL;

	*len=DIR_REQ_HEADER_SZ+sizeof(DIR_REQ_IDENT);
	DIR_REQ *ret=(DIR_REQ *)calloc(1, *len);
	if (!ret)
		return NULL;

	ret->type=DIR_REQ_TYPE_IDENT;

	return ret;
}

/**
 * Allocate a dummy response. The dummy response is only used to 
 * report error conditions before we hit to the server.
 * @return The dummy response structure.
 */
DIR_RES *DIR_Cli_AP_UDP::new_res_dummy()
{
	DIR_RES *res=(DIR_RES *)calloc(1, sizeof(DIR_RES));

	return res;
}

/** 
 * sending a request and wait for the response. 
 * @param req The request structure.
 * @param req_len The length of the request structure (variable len).
 * @return response The status and the response structure.
 */ 
DIR_RES * DIR_Cli_AP_UDP::request(DIR_REQ *req, int req_len) 
{
	DIR_RES *res=NULL;

	Udp_res **ress=udp_cli.sync_req(dir_srv_addr.c_str(), dir_srv_port, req, req_len, 10);

	if ((!ress)||(!ress[0])) {
		res=new_res_dummy();
		if (res)
			res->status=DIR_RES_EMEM;
		goto _exit;
	}

	if (ress[0]->status!=UDP_RES_OKAY) {
		res=new_res_dummy();
		if (res)
			res->status=udp_to_dir_status(ress[0]->status);
		goto _exit;
	}

	res=(DIR_RES *)malloc(ress[0]->length());

	if (!res) {
		goto _exit;
	}

	memcpy(res, ress[0]->result(), ress[0]->length());;

_exit:
	if (ress) {
		udp_cli.free_results(ress);
	}

	return res;
}

/** 
 * get the server identity. 
 * @param id (output) The result will be filled in.
 * @return status (zero: success; non-zero: error).
 */
int DIR_Cli_AP_UDP::get_ident(const char * &id)
{
	/** check whether the ident has already been known. */
	if (ident.c_str()[0] != '\0') {
		id=ident.c_str();
		return DIR_RES_OKAY;
	}

	/** Get the ident from server. */
	int req_len;
	DIR_REQ *req=new_req_ident(&req_len); /**< construct a request. */
	if (!req) {
		id=NULL;
		return DIR_RES_EMEM;
	}

	int ret=DIR_RES_OKAY;
	DIR_RES *res=request(req, req_len);
	free(req);

	if (res->status!=DIR_RES_OKAY) {
		ret=res->status;
		id=NULL;
		goto _exit;
	}

	ident=res->u.d_ident.ident;
	id=ident.c_str();

_exit:
	if (res)
		free_response(res);

	return ret;
}

void DIR_Cli_AP_UDP::free_response(DIR_RES *res)
{
	free(res);
}
