// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: dir_cli_base.cc,v $ - directory server client, shared functionalities.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/08/01 11:21:20 $
 * Version :  $Revision: 1.15 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 
# include <stdlib.h>
# include <time.h>
# include "dir_cli_base.h"
# include "tmpi_debug.h"

# if defined(__DEBUG__)
# include "asmalloc.h"
# endif

const DIR_Base::node_t DIR_Base::NIL=(DIR_Base::node_t)DIRDB_NIL; /* 0 */
const DIR_Base::node_t DIR_Base::ROOT=(DIR_Base::node_t)DIRDB_ROOT; /* 1 */
const DIR_Base::node_t DIR_Base::ERR=(DIR_Base::node_t)DIRDB_ERR; /* 2 */
const int DIR_Base::MAX_NAME_LEN=DIR_MAX_NAME_LEN;
const int DIR_Base::MAX_PATH_LEN=DIR_MAX_PATH_LEN;

// cached attributes expires in 1 minute.
# define CACHE_ATTRIB_EXPIRE 60

DIR_Base::DIR_Base() :
	nid(ERR),
	expire(0),
	child(ERR),
	parent(ERR),
	next_node(NIL)
{
	pwd[0]='\0';
	cur_param.reset();
}

void DIR_Base::check_expire(bool force_expire)
{
	if (!force_expire) {
		time_t now;
		time(&now);

		if (now<=expire) { // not expired yet.
			return;
		}
	}

	/**
	 * expired or force_expire==true
	 */
	{
		if ((parent!=ROOT)||(nid!=parent))
			parent=ERR;
		child=ERR;
		cur_param.reset();
	}

	return;
}

void DIR_Base::set_expire()
{
	time_t now;
	time(&now);

	expire=now+CACHE_ATTRIB_EXPIRE; 
}

DIR_REQ *DIR_Base::new_req_query(char *path, int buf_len, unsigned int flags, int *len)
{
	if (!len)
		return NULL;

	if (path) {
		*len=DIR_REQ_HEADER_SZ+DIR_REQ_QUERY_HEADER_SZ+strlen(path)+1;
	}
	else {
		*len=DIR_REQ_HEADER_SZ+sizeof(DIR_REQ_QUERY);
	}
	DIR_REQ *ret=(DIR_REQ *)calloc(1, *len);
	if (!ret)
		return NULL;

	ret->type=DIR_REQ_TYPE_QUERY;

	DIR_REQ_QUERY *req=&(ret->u.d_query);

	req->buf_len=buf_len;
	req->flags=flags;

	if (flags & DIR_CREATE) 
		req->param.options=DIR_OPTION_DIR;

	check_expire();
	if (path) {
		strcpy(req->path, path);
		req->base=nid;
		req->b_child=(child==NIL)?ERR:child;
		req->b_parent=parent;
		req->next_list=ERR;
	}
	else {
		req->path[0]='\0';
		req->base=ERR;
		req->b_child=ERR;
		req->b_parent=ERR;
		req->next_list=next_node;
	}

	return ret;
}

DIR_REQ *DIR_Base::new_req_del(char *path, unsigned int flags, int *len)
{
	if ((!len)||(!path))
		return NULL;

	*len=DIR_REQ_HEADER_SZ+DIR_REQ_DEL_HEADER_SZ+strlen(path)+1;

	DIR_REQ *ret=(DIR_REQ *)calloc(1, *len);

	if (!ret)
		return NULL;

	ret->type=DIR_REQ_TYPE_DEL;

	DIR_REQ_DEL *req=&(ret->u.d_del);

	req->flags=flags;

	strcpy(req->path, path);
	check_expire();
	req->base=nid;
	req->b_child=(child==NIL)?ERR:child;
	req->b_parent=parent;

	return ret;
}

DIR_REQ *DIR_Base::new_req_setp(char *path, DIR_Param *param, int *len)
{
	if ((!len)||(!path))
		return NULL;

	*len=DIR_REQ_HEADER_SZ+DIR_REQ_SETP_HEADER_SZ+strlen(path)+1;

	DIR_REQ *ret=(DIR_REQ *)calloc(1, *len);

	if (!ret)
		return NULL;

	ret->type=DIR_REQ_TYPE_SETP;

	DIR_REQ_SETP *req=&(ret->u.d_setp);

	strcpy(req->path, path);
	check_expire();
	req->base=nid;
	req->b_child=(child==NIL)?ERR:child;
	req->b_parent=parent;
	memcpy(&(req->param), param, sizeof(DIR_Param));

	return ret;
}

int	DIR_Base::do_local_cd_dotdot()
{
	if (parent==ERR) {
		return -1;
	}
	if (nid==parent) {
		// root, do nothing
		return 0;
	}

	char *p=strrchr(pwd, '/');
	if (!p) {
		return -1;
	}

	if (*(p+1)=='\0') { // a trailing '/'
		*p='\0'; // remove it, do it again.
		p=strrchr(pwd, '/');
		if (!p) {
			return -1;
		}
	}

	if (p==pwd) { // parent is root.
		*(p+1)='\0';
	}
	else {
		*p='\0';
	}

	nid=parent;
	check_expire(true);

	return 0;
}

int DIR_Base::get_param(DIR_Param *p)
{
	if (!p) {
		return DIR_RES_EINVALID;
	}

	if ( (cur_param.t_create==(time_t)0) 
# if defined(__DEBUG__)
		 || (memcmp(cur_param.guid, (Guid::GUID_NIL), sizeof(guid_t))==0)
# endif
		 ) {
		QUERY_RES res;

		if (query_(pwd, &res, NULL, NULL, 0)<0) {
			return res.status;
		}

		nid=res.nid; // probably redundant.
		child=res.child;
		parent=res.parent;
		memcpy(&cur_param, &(res.param), sizeof(DIR_Param));
		set_expire();
	}

	memcpy(p, &cur_param, sizeof(DIR_Param));

	return DIR_RES_OKAY;
}

// path must have already been standardized
int DIR_Base::query_(char *path, DIR_Base::QUERY_RES *meta, char *buf, int *len, unsigned int flags)
{
	if ((!meta)) {
		return -1;
	}

	if (!good()) {
		meta->status=DIR_RES_ECLIENT;
		return -1;
	}

	if (((!buf)&&(len))||((buf)&&(!len))) { // must both be NULL or none NULL
		meta->status=DIR_RES_EINVALID;
		return -1;
	}

	int buf_len=(len)?*len:0;
	int req_len;

	if ((!path)&&(next_node==NIL)) {
		meta->status=DIR_RES_EEND;
		return -1;
	}

# if defined(__INVALID_CHILD__)
	bool invalidate_child=false;
	if ((path)&&(flags&DIR_CREATE)) {
		if (path[0]=='/') {
			if (pwd[0]=='\0') {
				invalidate_child=true;
			}
			else {
				int pwd_len=strlen(pwd);
				if ( (strncmp(path, pwd, pwd_len)==0) &&
					 (path[pwd_len]) ) {
					invalidate_child=true;
				}
			}
		}
		else {
			char *p=path;
			while ((*p)&&((*p=='.')||(*p=='/')))
				p++;
			if (*p)
				invalidate_child=true;
		}
	}
# endif

	DIR_REQ *req=new_req_query(path, buf_len, flags, &req_len);
	if (!req) {
		meta->status=DIR_RES_EMEM;
		return -1;
	}

	/* could be send to server, or directly call dir_db function */
	DIR_RES *res=hook_query(req, req_len); 

	if (!res) { /* all other errors will store the error code in the struct */
		meta->status=DIR_RES_ESERVER;
		return -1;
	}

	free(req);

	int ret=0;
	DIR_RES_QUERY *res_qry=NULL;

	if (res->status!=DIR_RES_OKAY) {
		ret=-1;
		meta->status=res->status;
		goto _exit;
	}

	res_qry=(DIR_RES_QUERY *)&(res->u.d_query);
	meta->nid=res_qry->nid;
	meta->child=res_qry->n_child;
	meta->parent=res_qry->n_parent;
	memcpy(&(meta->param), &(res_qry->param), sizeof(DIR_Param));
# if defined(__INVALID_CHILD__)
	if (invalidate_child) {
		child=ERR; /* child might have been changed */
	}
# endif
	if (buf_len>0) { // looks for child
		meta->next=res_qry->n_next;
		if (*len<res_qry->buf_len) {
			ret=-1;
			meta->status=DIR_RES_EBUFFER;
			goto _exit;
		}
		*len=res_qry->buf_len;
		if (*len>0)
			memcpy(buf, res_qry->buf, *len);
	}

_exit:
	free_res(res);

	return ret;
}

int DIR_Base::change(char *dir, unsigned int flags)
{
	if (!dir) {
		dir="/";
	}

	char n_host[MAX_NAME_LEN];
	int n_port;
	char n_lpath[MAX_PATH_LEN];
	int ret=dir_parse_path(dir, n_host, MAX_NAME_LEN, &n_port, n_lpath, MAX_PATH_LEN);

	if (!(ret&DIR_PATH_LPATH)) {
		return DIR_RES_EINVALID;
	}

	char *t_host=NULL;
	int t_port=0;
	if (ret&DIR_PATH_HOST) {
		t_host=n_host;
	}
	if (ret&DIR_PATH_PORT) {
		t_port=n_port;
	}

	ret=DIR_RES_OKAY;
	if ((ret=change_server(t_host, t_port))!=DIR_RES_OKAY) {
		return ret;
	}

	if (nid==ERR) {
		// first time, starts at ROOT. must invalidate all the cached stuff.		
		check_expire(true);
		strcpy(pwd,"/");
		nid=ROOT;
		parent=ROOT;
		next_node=NIL;
	}

	char n_std_lpath[MAX_PATH_LEN];
	if (dir_std_path(NULL, n_lpath, n_std_lpath, MAX_PATH_LEN)<0) {
		return DIR_RES_EINVALID;
	}

	if (strcmp(n_std_lpath, ".")==0) {
		return DIR_RES_OKAY;
	}

	if (strcmp(n_std_lpath, "..")==0) {
		if (do_local_cd_dotdot()==0) {
			return DIR_RES_OKAY;
		}
	}

	char n_opt_lpath[MAX_PATH_LEN];
	if (dir_opt_path(pwd, n_std_lpath, n_opt_lpath, MAX_PATH_LEN)<0) {
# if defined(__DEBUG__)
		tmpi_error(DBG_INTERNAL, "dir_opt_path failed <%s, %s>.\n", pwd, n_std_lpath);
# endif
		strcpy(n_opt_lpath, n_std_lpath);
	}

	QUERY_RES res;
	if (query_(n_opt_lpath, &res, NULL, NULL, flags)<0) {
		return res.status;
	}

	char n_pwd[MAX_PATH_LEN];
	if (dir_std_path(pwd, n_opt_lpath, n_pwd, MAX_PATH_LEN)<0) {
		return DIR_RES_EBUFFER;
	}

	nid=res.nid;
	child=res.child;
	parent=res.parent;
	memcpy(&cur_param, &(res.param), sizeof(DIR_Param));
	strcpy(pwd, n_pwd);
	set_expire();

	return DIR_RES_OKAY;
}

/**
 * More rigorous checking?
 */
bool DIR_Base::is_valid_local_path(char *path)
{
	if (!path)
		return false;

	if (strchr(path, ':'))
		return false;
	return true;
}

// not return -1 or 0 but a detailed error code
int DIR_Base::list(char *path, DIR_Param *pm, char *buf, int *len, unsigned int flags)
{
	QUERY_RES res;

	if (!path) {
		if (query_(NULL, &res, buf, len, flags)<0) {
			return res.status;
		}
		next_node=res.next;
	}
	else {
		if (!pm)
			return DIR_RES_EINVALID;

		if (!is_valid_local_path(path)){ 
			// list only works on the current server.
			return DIR_RES_EINVALID;
		}

		char n_std_lpath[MAX_PATH_LEN];
		if (dir_std_path(NULL, path, n_std_lpath, MAX_PATH_LEN)<0) { // bad path
			return DIR_RES_EINVALID;
		}

		char n_opt_lpath[MAX_PATH_LEN];
		if (dir_opt_path(pwd, n_std_lpath, n_opt_lpath, MAX_PATH_LEN)<0) {
# if defined(__DEBUG__)
			tmpi_error(DBG_INTERNAL, "dir_opt_path failed <%s, %s>.\n", pwd, n_std_lpath);
# endif
			strcpy(n_opt_lpath, n_std_lpath);
		}

		if (query_(n_opt_lpath, &res, buf, len, flags)<0) {
			return res.status;
		}
		memcpy(pm, &(res.param), sizeof(DIR_Param));
		next_node=res.next;
		if (res.nid==nid) {
			parent=res.parent;
			child=res.child;
			memcpy(&cur_param, &(res.param), sizeof(DIR_Param));
			set_expire();
		}
	}

	return DIR_RES_OKAY;
}

int DIR_Base::read_list(char *buf, int len, char **p, char **iname, DIR_Param *ipm)
{
	static const int min_item_sz=sizeof(DIR_Param)+sizeof(node_t)+2;

	if ((!buf)||(len<0)||(!p)||(!iname)||(!ipm)) {
		return DIR_RES_EINVALID;
	}

	if (*p==NULL) { /* start call */
		*p=buf;
	}
	
	char *buf_end=buf+len;
	if (*p+min_item_sz>buf_end) {
		return DIR_RES_EEND;
	}
	memcpy(ipm, *p+sizeof(node_t), sizeof(DIR_Param));
	*iname=*p+sizeof(node_t)+sizeof(DIR_Param);
	*p=*iname;
	while (*p<buf_end) {
		if (**p=='\0') {
			break;
		}
		(*p)++;
	}

	if (*p<buf_end) {
		(*p)++;
		return DIR_RES_OKAY;
	}

	return DIR_RES_EEND;
}

int DIR_Base::del(char *path, unsigned int flags)
{
	if (!good()) {
		return DIR_RES_ECLIENT;
	}

	if (!path) {
		return DIR_RES_EINVALID;
	}

	if (!is_valid_local_path(path)) { 
		return DIR_RES_EINVALID;
	}

	char n_std_lpath[MAX_PATH_LEN];
	if (dir_std_path(NULL, path, n_std_lpath, MAX_PATH_LEN)<0) { // bad path
		return DIR_RES_EINVALID;
	}

	char n_opt_lpath[MAX_PATH_LEN];
	if (dir_opt_path(pwd, n_std_lpath, n_opt_lpath, MAX_PATH_LEN)<0) {
# if defined(__DEBUG__)
		tmpi_error(DBG_INTERNAL, "dir_opt_path failed <%s, %s>.\n", pwd, n_std_lpath);
# endif
		strcpy(n_opt_lpath, n_std_lpath);
	}

# if defined(__INVALID_CHILD__)
	/**
	 * If the immediate child of the current node is gone,
	 * we need to invalidate the current child.
	 * For absolute path, we compare with cached pwd.
	 * Otherwise, it is only disabled if it is a single
	 * level normal path.
	 */
	bool invalidate_child=false;
	if (n_opt_lpath[0]=='/') {
		if (pwd[0]=='\0')
			invalidate_child=true;
		else {
			int pwd_len=strlen(pwd);
			if (strncmp(pwd, n_opt_lpath, pwd_len)==0) {
				if ( (n_opt_lpath[pwd_len]=='/')&& 
					 (strchr(n_opt_lpath+pwd_len+1, '/')==NULL) )
					invalidate_child=true;
			}
		}
	}
	else if (!strchr(n_opt_lpath, '/')) {
		if (strcmp(n_opt_lpath, ".")&& strcmp(n_opt_lpath, ".."))
			invalidate_child=true;
	}
# endif

	// construct a request.
	int req_len;
	DIR_REQ *req=new_req_del(n_opt_lpath, flags, &req_len);
	if (!req) {
		return DIR_RES_EMEM;
	}

	DIR_RES *res=hook_del(req, req_len);
	if (!res) {
		return DIR_RES_ESERVER;
	}

	free(req);

	int ret=DIR_RES_OKAY;

	if ((res->status!=DIR_RES_OKAY)&&(res->status!=DIR_RES_ESELF)) {
		ret=res->status;
		goto _exit;
	}

	// succeeded, but current directory no longer exists.
	if (res->status==DIR_RES_ESELF) {
		check_expire(true);
		nid=ROOT;
		strcpy(pwd,"/");
		parent=ROOT;
		next_node=NIL;

		ret=DIR_RES_ESELF;
	}
	else {
# if defined(__INVALID_CHILD__)
		if (invalidate_child)
			child=ERR;
# endif
		DIR_RES_DEL *res_del=&(res->u.d_del);
		if (nid==res_del->nid) {
			parent=res_del->n_parent;
			child=res_del->n_child;
			memcpy(&cur_param, &(res_del->param), sizeof(DIR_Param));
			set_expire();
		}
	}

_exit:

	free_res(res);
	return ret;
}

int DIR_Base::set_param(char *path, DIR_Param *p)
{
	if (!good()) {
		return DIR_RES_ECLIENT;
	}

	if ((!path)||(!p)) {
		return DIR_RES_EINVALID;
	}

	if (!is_valid_local_path(path)) {
		return DIR_RES_EINVALID;
	}

	char n_std_lpath[MAX_PATH_LEN];
	if (dir_std_path(NULL, path, n_std_lpath, MAX_PATH_LEN)<0) { // bad path
		return DIR_RES_EINVALID;
	}

	char n_opt_lpath[MAX_PATH_LEN];
	if (dir_opt_path(pwd, n_std_lpath, n_opt_lpath, MAX_PATH_LEN)<0) {
# if defined(__DEBUG__)
		tmpi_error(DBG_INTERNAL, "dir_opt_path failed <%s, %s>.\n", pwd, n_std_lpath);
# endif
		strcpy(n_opt_lpath, n_std_lpath);
	}

	// construct a request.
	int req_len;
	DIR_REQ *req=new_req_setp(n_opt_lpath, p, &req_len);
	if (!req) {
		return DIR_RES_EMEM;
	}

	DIR_RES *res=hook_setp(req, req_len);
	DIR_RES_SETP *res_setp=NULL;
	if (!res) {
		return DIR_RES_ESERVER;
	}

	free(req);

	int ret=DIR_RES_OKAY;

	if (res->status!=DIR_RES_OKAY) {
		ret=res->status;
		goto _exit;
	}

	res_setp=&(res->u.d_setp);
	if (nid==res_setp->nid) {
		parent=res_setp->n_parent;
		child=res_setp->n_child;
		memcpy(&cur_param, &(res_setp->param), sizeof(DIR_Param));
		set_expire();
	}

_exit:

	free_res(res);
	return ret;
}
