// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: dir_cli_mt.cc,v $ - the Directory Server client, 
 *            multi-threaded version (implementation). 
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/09/12 00:29:15 $
 * Version :  $Revision: 1.9 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

# include "dir_common.h"
# include "dir_cli_mt.h"
# include "dir_cli_ap.h"

const DIR_Entry::node_t DIR_Entry::NIL=(DIR_Entry::node_t)DIRDB_NIL; /* 0 */
const DIR_Entry::node_t DIR_Entry::ROOT=(DIR_Entry::node_t)DIRDB_ROOT; /* 1 */
const DIR_Entry::node_t DIR_Entry::ERR=(DIR_Entry::node_t)DIRDB_ERR; /* 2 */
const int DIR_Entry::MAX_NAME_LEN=DIR_MAX_NAME_LEN;
const int DIR_Entry::MAX_PATH_LEN=DIR_MAX_PATH_LEN;

/** Default constructor, set as the root. */
DIR_Entry::DIR_Entry() :
	nid(ROOT),
	parent(ERR),
	child(ERR),
	next_node(NIL),
	expire(0),
	expired(true)
{
	cur_param.options=DIR_OPTION_DIR;
	pwd[0]='/';
	pwd[1]='\0';
}

/**
 * Check whether the cached attributes are expired or not. 
 * @param force_expire Force the expiration of attributes.
 * @return none.
 */
void DIR_Entry::check_expire(bool force_expire)
{
	if (expired)
		return;

	if (!force_expire) {
		time_t now;
		time(&now);

		if (now<=expire) { // not expired yet.
			return;
		}
	}

	child=ERR;
	if ((parent!=ROOT)||(nid!=parent))
		parent=ERR;
	expired=true;


	return;
}

# define CACHE_ATTRIB_EXPIRE 60

void DIR_Entry::set_expire()
{
	time_t now;
	time(&now);

	expire=now+CACHE_ATTRIB_EXPIRE; 
	expired=false;
}

/**
 * Refresh the cached attributes (parent, child, param)
 * @param force_expire Force the expiration of attributes.
 * @param ap Directory server access point.
 * @return none.
 */
int DIR_Entry::refresh_info(DIR_Cli_AP *ap)
{
	check_expire();

	if (expired) {
		QUERY_RES res;

		if ( (query(ap, ".", &res, NULL, NULL, 0, NULL)<0) &&
			 (query(ap, pwd, &res, NULL, NULL, 0, NULL)<0) ) {
			return res.status;
		}

		update(&res, NULL, NULL);
	}

	return DIR_RES_OKAY;
}

// path must have already been standardized
int DIR_Entry::query(DIR_Cli_AP *ap, const char *path, DIR_Entry::QUERY_RES *meta, \
		char *buf, int *len, unsigned int flags, const DIR_Param *param)
{
	// if ((!meta)) {
		// return -1;
	// }

	// if (!ap) {
		// meta->status=DIR_RES_ECLIENT;
		// return -1;
	// }

	// if (((!buf)&&(len))||((buf)&&(!len))) { // must both be NULL or none NULL
		// meta->status=DIR_RES_EINVALID;
		// return -1;
	// }

	int buf_len=(len)?*len:0;
	int req_len;

	if ((!path)&&(next_node==NIL)) {
		meta->status=DIR_RES_EEND;
		return -1;
	}

	DIR_REQ *req=new_req_query(path, buf_len, flags, &req_len, param);
	if (!req) {
		meta->status=DIR_RES_EMEM;
		return -1;
	}

	DIR_RES *res=ap->request(req, req_len); 

	if (!res) { /* all other errors will store the error code in the struct */
		meta->status=DIR_RES_ESERVER;
		return -1;
	}

	free(req);

	int ret=0;
	DIR_RES_QUERY *res_qry=NULL;

	if (res->status!=DIR_RES_OKAY) {
		ret=-1;
		meta->status=res->status;
		goto _exit;
	}

	res_qry=(DIR_RES_QUERY *)&(res->u.d_query);
	meta->nid=res_qry->nid;
	meta->child=res_qry->n_child;
	meta->parent=res_qry->n_parent;
	memcpy(&(meta->param), &(res_qry->param), sizeof(DIR_Param));

	if (buf_len>0) { // looks for child
		meta->next=res_qry->n_next;
		if (*len<res_qry->buf_len) {
			ret=-1;
			meta->status=DIR_RES_EBUFFER;
			goto _exit;
		}
		*len=res_qry->buf_len;
		if (*len>0)
			memcpy(buf, res_qry->buf, *len);
	}

_exit:
	ap->free_response(res);

	return ret;
}

/**
 * Allocate a query request.
 * @param path The path of the query node.
 * @param buf_len The size of the children node buffer. 
 * 	Used for query directory content.
 * @param flags The flags (whether to create node).
 * @param param The parameter of the end node (only 
 *  needed DIR_CREATE is specified in the flags).
 * @param len An pointer to an interger, which will 
 *  be used to store the size of the request structure.
 * @return The request structure.
 */
DIR_REQ *DIR_Entry::new_req_query(const char *path, int buf_len, unsigned int flags, int *len, const DIR_Param *param)
{
	if (!len)
		return NULL;

	if (path) {
		*len=DIR_REQ_HEADER_SZ+DIR_REQ_QUERY_HEADER_SZ+strlen(path)+1;
	}
	else {
		*len=DIR_REQ_HEADER_SZ+sizeof(DIR_REQ_QUERY);
	}
	DIR_REQ *ret=(DIR_REQ *)calloc(1, *len);
	if (!ret)
		return NULL;

	ret->type=DIR_REQ_TYPE_QUERY;

	DIR_REQ_QUERY *req=&(ret->u.d_query);

	req->buf_len=buf_len;
	req->flags=flags;
	if ((flags&DIR_CREATE) && (param)) {
		memcpy(&(req->param), param, sizeof(DIR_Param));
	}

	check_expire();

	if (path) {
		strcpy(req->path, path);
		req->base=nid;
		req->b_child=(child==NIL)?ERR:child;
		req->b_parent=parent;
		req->next_list=ERR;
	}
	else {
		req->path[0]='\0';
		req->base=ERR;
		req->b_child=ERR;
		req->b_parent=ERR;
		req->next_list=next_node;
	}

	return ret;
}

DIR_REQ *DIR_Entry::new_req_del(const char *path, unsigned int flags, int *len)
{
	if ((!len)||(!path))
		return NULL;

	*len=DIR_REQ_HEADER_SZ+DIR_REQ_DEL_HEADER_SZ+strlen(path)+1;

	DIR_REQ *ret=(DIR_REQ *)calloc(1, *len);

	if (!ret)
		return NULL;

	ret->type=DIR_REQ_TYPE_DEL;

	DIR_REQ_DEL *req=&(ret->u.d_del);

	req->flags=flags;

	strcpy(req->path, path);
	check_expire();
	req->base=nid;
	req->b_child=(child==NIL)?ERR:child;
	req->b_parent=parent;

	return ret;
}

DIR_REQ *DIR_Entry::new_req_setp(const char *path, const DIR_Param *param, int *len)
{
	if ((!len)||(!path))
		return NULL;

	*len=DIR_REQ_HEADER_SZ+DIR_REQ_SETP_HEADER_SZ+strlen(path)+1;

	DIR_REQ *ret=(DIR_REQ *)calloc(1, *len);

	if (!ret)
		return NULL;

	ret->type=DIR_REQ_TYPE_SETP;

	DIR_REQ_SETP *req=&(ret->u.d_setp);

	strcpy(req->path, path);
	check_expire();
	req->base=nid;
	req->b_child=(child==NIL)?ERR:child;
	req->b_parent=parent;
	memcpy(&(req->param), param, sizeof(DIR_Param));

	return ret;
}

/**
 * More rigorous checking?
 */
bool DIR_Entry::is_valid_local_path(const char *path)
{
	if (!path)
		return false;

	if (strchr(path, ':'))
		return false;
	return true;
}

int DIR_Entry::del(DIR_Cli_AP *ap, const char *path, unsigned int flags)
{
	// if (!ap) {
		// return DIR_RES_ECLIENT;
	// }

	// if (!path) {
		// return DIR_RES_EINVALID;
	// }

	// if (!is_valid_local_path(path)) { 
		// return DIR_RES_EINVALID;
	// }

	// char n_std_lpath[MAX_PATH_LEN];
	// if (dir_std_path(NULL, path, n_std_lpath, MAX_PATH_LEN)<0) { // bad path
		// return DIR_RES_EINVALID;
	// }

	// construct a request.
	int req_len;
	DIR_REQ *req=new_req_del(path, flags, &req_len);
	if (!req) {
		return DIR_RES_EMEM;
	}

	DIR_RES *res=ap->request(req, req_len);
	if (!res) {
		return DIR_RES_ESERVER;
	}

	free(req);

	int ret=DIR_RES_OKAY;
	DIR_RES_DEL *res_del=NULL;

	if ((res->status!=DIR_RES_OKAY)&&(res->status!=DIR_RES_ESELF)) {
		ret=res->status;
		goto _exit;
	}

	// succeeded, but current directory no longer exists.
	if (res->status==DIR_RES_ESELF) {
		check_expire(true);
		nid=ROOT;
		strcpy(pwd,"/");
		parent=ROOT;
		next_node=NIL;
		cur_param.options=DIR_OPTION_DIR;

		ret=DIR_RES_ESELF;
	}

	res_del=&(res->u.d_del);
	if (nid==res_del->nid) {
		parent=res_del->n_parent;
		child=res_del->n_child;
		next_node=NIL;
		memcpy(&cur_param, &(res_del->param), sizeof(DIR_Param));
		set_expire();
	}

_exit:

	ap->free_response(res);
	return ret;
}

int DIR_Entry::setp(DIR_Cli_AP *ap, const char *path, const DIR_Param *p)
{
	// if (!ap) {
		// return DIR_RES_ECLIENT;
	// }

	// if ((!path)||(!p)) {
		// return DIR_RES_EINVALID;
	// }

	// if (!is_valid_local_path(path)) {
		// return DIR_RES_EINVALID;
	// }

	// char n_std_lpath[MAX_PATH_LEN];
	// if (dir_std_path(NULL, path, n_std_lpath, MAX_PATH_LEN)<0) { // bad path
		// return DIR_RES_EINVALID;
	// }

	// construct a request.
	int req_len;
	DIR_REQ *req=new_req_setp(path, p, &req_len);
	if (!req) {
		return DIR_RES_EMEM;
	}

	DIR_RES *res=ap->request(req, req_len);
	DIR_RES_SETP *res_setp=NULL;
	if (!res) {
		return DIR_RES_ESERVER;
	}

	free(req);

	int ret=DIR_RES_OKAY;

	if (res->status!=DIR_RES_OKAY) {
		ret=res->status;
		goto _exit;
	}

	res_setp=&(res->u.d_setp);
	if (nid==res_setp->nid) {
		parent=res_setp->n_parent;
		child=res_setp->n_child;
		next_node=NIL;
		memcpy(&cur_param, &(res_setp->param), sizeof(DIR_Param));
		set_expire();
	}

_exit:

	ap->free_response(res);
	return ret;
}

// Refresh the info using the data in res.
// Important, I did not refresh next_node. Need a full inspection
// of code to make sure it is okay. -Hong 09/11/03.
int DIR_Entry::update(QUERY_RES *res, const char *base, const char *path)
{
	char n_pwd[MAX_PATH_LEN];
	if (base) {
		if (dir_std_path(base, path, n_pwd, MAX_PATH_LEN)<0) {
			return DIR_RES_EBUFFER;
		}
	}

	nid=res->nid;
	parent=res->parent;
	child=res->child;
	next_node=NIL; //* <== TODO: maybe it is okay to change it to res.next?
	memcpy(&cur_param, &(res->param), sizeof(DIR_Param));
	if (base) {
		strcpy(pwd, n_pwd);
	}
	set_expire();

	return DIR_RES_OKAY;
}

/**< Look up a name relative to myself. */
int DIR_Entry::lookup(DIR_Cli_AP *ap, const char *path, DIR_Entry *entry)
{
	if  (!ap) {
		return DIR_RES_ECLIENT;
	}

	if ( (!path) || (path[0]=='\0') || (!is_valid_local_path(path)) || (!entry) ) {
		return DIR_RES_EINVALID;
	}

	if (!DIR_IS_DIR(cur_param.options)) {
		return DIR_RES_ENOTDIR;
	}

	char n_std_lpath[MAX_PATH_LEN];
	if (dir_std_path(NULL, path, n_std_lpath, MAX_PATH_LEN)<0) {
		return DIR_RES_EINVALID;
	}

	if (strcmp(n_std_lpath, ".")==0) {
		memcpy(entry, this, sizeof(DIR_Entry));
		// entry.refresh_info(ap);
		return DIR_RES_OKAY;
	}

	QUERY_RES res;
	if (query(ap, n_std_lpath, &res, NULL, NULL, 0, NULL)<0) {
		return res.status;
	}

	if (res.nid == nid) {
		update(&res, NULL, NULL);
	}

	return entry->update(&res, pwd, n_std_lpath);
}

/**< Create a new directory. */
int DIR_Entry::mkdir(DIR_Cli_AP *ap, const char *path, DIR_Entry *entry)
{
	if  (!ap) {
		return DIR_RES_ECLIENT;
	}

	if ( (!path) || (path[0]=='\0') || (!is_valid_local_path(path)) || (!entry) ) {
		return DIR_RES_EINVALID;
	}

	if (!DIR_IS_DIR(cur_param.options)) {
		return DIR_RES_ENOTDIR;
	}

	char n_std_lpath[MAX_PATH_LEN];
	if (dir_std_path(NULL, path, n_std_lpath, MAX_PATH_LEN)<0) {
		return DIR_RES_EINVALID;
	}

	if ( (strcmp(n_std_lpath, ".")==0) || (strcmp(n_std_lpath, "..")==0) 
			|| (strcmp(n_std_lpath, "/")==0) ) {
		return DIR_RES_EEXIST;
	}

	QUERY_RES res;
	DIR_Param param;
	param.reset();
	param.options=DIR_OPTION_DIR;

	if (query(ap, n_std_lpath, &res, NULL, NULL, DIR_CREATE|DIR_EXCLUSIVE, &param)<0) {
		return res.status;
	}

	return entry->update(&res, pwd, n_std_lpath);
}

/**< Delete a directory. */
/**
 * TODO: better to add some guard mechanism.
 */ 
int DIR_Entry::rmdir(DIR_Cli_AP *ap, const char *path)
{
	if  (!ap) {
		return DIR_RES_ECLIENT;
	}

	if ( (!path) || (path[0]=='\0') || (!is_valid_local_path(path)) ) {
		return DIR_RES_EINVALID;
	}

	if (!DIR_IS_DIR(cur_param.options)) {
		return DIR_RES_ENOTDIR;
	}

	char n_std_lpath[MAX_PATH_LEN];
	if (dir_std_path(NULL, path, n_std_lpath, MAX_PATH_LEN)<0) {
		return DIR_RES_EINVALID;
	}

	/// cannot delete root or parent.
	if ( (strcmp(n_std_lpath, "/")==0) || (strcmp(n_std_lpath, "..")==0) ) {
		return DIR_RES_EEXIST;
	}

	if ( strcmp(n_std_lpath, ".") ) {
		/** check whether this is a directory */
		QUERY_RES res;
		if (query(ap, n_std_lpath, &res, NULL, NULL, 0, NULL)<0) {
			return res.status;
		}
		if (!DIR_IS_DIR(res.param.options)) {
			return DIR_RES_ENOTDIR;
		}
	}

	return del(ap, n_std_lpath, 0);
}

/**< Create a new file. */
int DIR_Entry::create(DIR_Cli_AP *ap, const char *path, const DIR_Param *param, DIR_Entry *entry)
{
	if  (!ap) {
		return DIR_RES_ECLIENT;
	}

	if ( (!path) || (path[0]=='\0') || (!is_valid_local_path(path)) || (!param) || (!entry) ) {
		return DIR_RES_EINVALID;
	}

	if (!DIR_IS_DIR(cur_param.options)) {
		return DIR_RES_ENOTDIR;
	}

	if (DIR_IS_DIR(param->options)) {
		return DIR_RES_EISDIR;
	}

	char n_std_lpath[MAX_PATH_LEN];
	if (dir_std_path(NULL, path, n_std_lpath, MAX_PATH_LEN)<0) {
		return DIR_RES_EINVALID;
	}

	if ( (strcmp(n_std_lpath, ".")==0) || (strcmp(n_std_lpath, "..")==0) 
			|| (strcmp(n_std_lpath, "/")==0) ) {
		return DIR_RES_EEXIST;
	}

	QUERY_RES res;

	if (query(ap, n_std_lpath, &res, NULL, NULL, DIR_CREATE|DIR_EXCLUSIVE, param)<0) {
		return res.status;
	}

	return entry->update(&res, pwd, n_std_lpath);
}

/**< Remove a file. */
int DIR_Entry::remove(DIR_Cli_AP *ap, const char *path)
{
	if  (!ap) {
		return DIR_RES_ECLIENT;
	}

	if ( (!path) || (path[0]=='\0') || (!is_valid_local_path(path)) ) {
		return DIR_RES_EINVALID;
	}

	if (!DIR_IS_DIR(cur_param.options)) {
		return DIR_RES_ENOTDIR;
	}

	char n_std_lpath[MAX_PATH_LEN];
	if (dir_std_path(NULL, path, n_std_lpath, MAX_PATH_LEN)<0) {
		return DIR_RES_EINVALID;
	}

	if ( (strcmp(n_std_lpath, ".")==0) || (strcmp(n_std_lpath, "..")==0) 
			|| (strcmp(n_std_lpath, "/")==0) ) {
		return DIR_RES_EISDIR;
	}

	QUERY_RES res;

	if (query(ap, n_std_lpath, &res, NULL, NULL, 0, NULL)<0) {
		return res.status;
	}

	if (DIR_IS_DIR(res.param.options)) {
			return DIR_RES_EISDIR;
	}

	return del(ap, n_std_lpath, 0);
}

/**
int DIR_Entry::deltree(DIR_Cli_AP *ap, const char *path)
{
	if  (!ap) {
		return DIR_RES_ECLIENT;
	}

	if ( (!path) || (path[0]=='\0') || (!is_valid_local_path(path)) ) {
		return DIR_RES_EINVALID;
	}

	if (!DIR_IS_DIR(cur_param.options)) {
		return DIR_RES_ENOTDIR;
	}

	char n_std_lpath[MAX_PATH_LEN];
	if (dir_std_path(NULL, path, n_std_lpath, MAX_PATH_LEN)<0) {
		return DIR_RES_EINVALID;
	}

	return del(ap, n_std_lpath, DIR_RECURSIVE);
}
**/

/**< read the children nodes of myself. */
int DIR_Entry::readdir(DIR_Cli_AP *ap, DIR_List &list, int buf_len)
{
	if  (!ap) {
		return DIR_RES_ECLIENT;
	}

	if ( (buf_len<=0) ) {
		return DIR_RES_EINVALID;
	}

	if (!DIR_IS_DIR(cur_param.options)) {
		return DIR_RES_ENOTDIR;
	}

	QUERY_RES res;
	char buf[buf_len];
	int bl1=buf_len, bl2=buf_len;

	if ( (query(ap, ".", &res, buf, &bl1, 0, NULL)<0) && 
		 (query(ap, pwd, &res, buf, &bl2, 0, NULL)<0) )
	{
		return res.status;
	}

	update(&res, NULL, NULL);
	next_node=res.next;

	char *p=NULL;
	DIR_Item item;
	char *iname;
	int bl = (bl1<=bl2)?bl1:bl2;
	while (read_buf(buf, bl, &p, &iname, &(item.param))==DIR_RES_OKAY) {
		item.name=iname;
		list.push_back(item);
	}

	return DIR_RES_OKAY;
}

int DIR_Entry::read_buf(char *buf, int len, char **p, char **iname, DIR_Param *ipm)
{
	static const int min_item_sz=sizeof(DIR_Param)+sizeof(node_t)+2;

	if ((!buf)||(len<0)||(!p)||(!iname)||(!ipm)) {
		return DIR_RES_EINVALID;
	}

	if (*p==NULL) { /* start call */
		*p=buf;
	}
	
	char *buf_end=buf+len;
	if (*p+min_item_sz>buf_end) {
		return DIR_RES_EEND;
	}
	memcpy(ipm, *p+sizeof(node_t), sizeof(DIR_Param));
	*iname=*p+sizeof(node_t)+sizeof(DIR_Param);
	*p=*iname;
	while (*p<buf_end) {
		if (**p=='\0') {
			break;
		}
		(*p)++;
	}

	if (*p<buf_end) {
		(*p)++;
		return DIR_RES_OKAY;
	}

	return DIR_RES_EEND;
}

/**< continue read the children nodes of myself. */
int DIR_Entry::readdir_cont(DIR_Cli_AP *ap, DIR_List &list, int buf_len)
{
	if  (!ap) {
		return DIR_RES_ECLIENT;
	}

	if ( (buf_len<=0) ) {
		return DIR_RES_EINVALID;
	}

	if (!DIR_IS_DIR(cur_param.options)) {
		return DIR_RES_ENOTDIR;
	}

	QUERY_RES res;
	char buf[buf_len];

	if ( (query(ap, NULL, &res, buf, &buf_len, 0, NULL)<0) )
	{
		return res.status;
	}

	next_node=res.next;

	char *p=NULL;
	DIR_Item item;
	char *iname;
	while (read_buf(buf, buf_len, &p, &iname, &(item.param))==DIR_RES_OKAY) {
		item.name=iname;
		list.push_back(item);
	}

	return DIR_RES_OKAY;
}

/**< set the attribute of the node (only size matters). */
int DIR_Entry::setattr(DIR_Cli_AP *ap, const DIR_Param *param)
{
	if  (!ap) {
		return DIR_RES_ECLIENT;
	}

	if (DIR_IS_DIR(cur_param.options)) {
		return DIR_RES_EISDIR;
	}

	return setp(ap, ".", param);
}

/**< get the attribute of the node. */
int DIR_Entry::getattr(DIR_Cli_AP *ap, DIR_Param *param)
{
	if (!param) {
		return DIR_RES_EINVALID;
	}

	int ret=refresh_info(ap);
	if (ret==DIR_RES_OKAY) {
		memcpy(param, &cur_param, sizeof(DIR_Param));
	}

	return ret;
}

/**< the type of the entry: dir or file. */
int DIR_Entry::get_type(DIR_Cli_AP *ap, int &type)
{
	int ret=refresh_info(ap);
	if (ret==DIR_RES_OKAY) {
		type=DIR_IS_DIR(cur_param.options)?DIR_TYPE_DIR:DIR_TYPE_FILE;
	}

	return ret;
}

/**< the guid. */
int DIR_Entry::get_guid(DIR_Cli_AP *ap, guid_t &guid)
{
	int ret=refresh_info(ap);
	if (ret==DIR_RES_OKAY) {
		memcpy(&guid, &(cur_param.guid), sizeof(guid_t));
	}

	return ret;
}

/**< the creation time */
int DIR_Entry::get_ctime(DIR_Cli_AP *ap, time_t &ctime)
{
	int ret=refresh_info(ap);
	if (ret==DIR_RES_OKAY) {
		ctime=cur_param.t_create;
	}

	return ret;
}

/**< the size */
int DIR_Entry::get_size(DIR_Cli_AP *ap, off64_t &size)
{
	int ret=refresh_info(ap);
	if (ret==DIR_RES_OKAY) {
		size=cur_param.size;
	}

	return ret;
}
