// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: dir_cli_mt.h,v $ - the Directory Server client, 
 *            multi-threaded version. We separate the directory
 *            server access point from the directory tree node 
 *            state (an m-to-n relationship).
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/08/01 11:21:20 $
 * Version :  $Revision: 1.6 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

# ifndef _DIR_CLI_MT_H_
# define _DIR_CLI_MT_H_

// # include "udp_cli.h"
# include "dir_common.h"

// # include "dirdb.h"
// # include "dir_cli_ap.h"

class DIR_Cli_AP;

# include <vector>
# include <string>

/** Each item in a directory is a DIR_Item. */
struct DIR_Item {
	std::string name;
	DIR_Param param;
};

/** Represent a list of directory items. */
typedef std::vector<DIR_Item>  DIR_List;


/** It is a bad news that I cannot implement DIR_Entry 
 * as a subclass of DIR_Base. Although they share lots
 * of commonalities. The design philosophy is quite 
 * different. For instance, here we need to distinguish
 * between the directories and files, while in DIR_Base,
 * everything is considered a node. Another example,
 * we want to always keep the attributes up-to-date,
 * so even an optimizable "chdir("..")" will result
 * in a network operation.
 */
class DIR_Entry {
public:
	typedef dir_node_t node_t; /**< type */
	static const node_t NIL; /**< 0 */
	static const node_t ROOT; /**< 1 */
	static const node_t ERR; /**< 2 */
	static const int MAX_NAME_LEN; /**< maximum name component length */
	static const int MAX_PATH_LEN; /**< maximum total path length */

 private:
	/** current absolute working direcoty (for path optimization) */
	char pwd[DIR_MAX_PATH_LEN]; 
	node_t nid; /**< The current node id */
	node_t parent; /**< cached attributes */
	node_t child; /**< cached attributes */
	node_t next_node; /**< only used by list() when path==NULL. */
	DIR_Param cur_param; /**< parameters for the current directory */
	time_t expire; /**< the two cached values are only valid for 60 seconds. */
	bool expired; /**< are cached info expired? */

	/**
	 * The structure for directory server query processing.
	 */
	struct QUERY_RES {
		int status; /**< Status code */
		node_t nid; /**< base node id */
		node_t parent; /**< parent node id */
		node_t child; /**< child node id */
		node_t next; /**< next node id (for list only) */
		DIR_Param param; /**< parameter for base node */
	};

	/** 
	 * @defgroup genop Generic protocol operations.
	 * The directory server protocol consists of three types of operations
	 * and are encapsulated by the following three function calls.
	 */
	/** @{ */
	/** Perform the actual query operation, the most versatile op. */
	int query(DIR_Cli_AP *ap, const char *node_path, QUERY_RES *meta, char *buf, int *len, unsigned int flags, const DIR_Param *param=NULL);
	/** Perform the del operation. */
	int del(DIR_Cli_AP *ap, const char *path, unsigned int flags);
	/** Perform the setp operation. */
	int setp(DIR_Cli_AP *ap, const char *path, const DIR_Param *p);
	/** @} */

	/** @defgroup reqalloc Request structure allocation. */
	/** @{ */
	DIR_REQ *new_req_query(const char *path, int buf_len, unsigned int flags, int *len, const DIR_Param *param=NULL);
	DIR_REQ *new_req_del(const char *path, unsigned int flags, int *len);
	DIR_REQ *new_req_setp(const char *path, const DIR_Param *p, int *len);
	/** @} */

	/** @defgroup caching Attribute caching. */
	/** @{ */
	void check_expire(bool force_expire=false); /**< called before the use of a cached attribute. */
	void set_expire(); /**< called after a del operation */
	/** use QUERY_RES to update myself, called after a query() */
	int update(QUERY_RES *res, const char *base, const char *path);
	/** @} */

	/** check the validity of the path */
	bool is_valid_local_path(const char *path);
	/** parsing the readdir buffer */
	int read_buf(char *buf, int len, char **p, char **iname, DIR_Param *ipm);

 public:
	DIR_Entry(); /**< The default constructor */
	/**
	 * @defgroup operations DIR_Entry operations.
	 * The set of operations and their format follow closely
	 * with the NFS3 protocol (you may imagine DIR_Entry is
	 * a file handle). The getattr procedure is replaced with
	 * a number of individual calls. We further add a deltree
	 * operation to facilitate fast deletion of a whole 
	 * directory tree.
	 */
	/** @{ */
	int lookup(DIR_Cli_AP *ap, const char *path, DIR_Entry *entry); /**< Look up a name relative to myself. */
	int mkdir(DIR_Cli_AP *ap, const char *path, DIR_Entry *entry); /**< Create a new directory. */
	int rmdir(DIR_Cli_AP *ap, const char *path); /**< Delete a directory. */
	// int deltree(DIR_Cli_AP *ap, const char *path); /**< Delete a directory tree. */
	int create(DIR_Cli_AP *ap, const char *path, const DIR_Param *param, DIR_Entry *entry); /**< Create a new file. */
	int remove(DIR_Cli_AP *ap, const char *path); /**< Remove a file. */
	int readdir(DIR_Cli_AP *ap, DIR_List &list, int buf_len); /**< read the children nodes of myself. */
	/** Is end-of-directory during a sequence of readdir_cont calls */
	bool is_eod() const { return (next_node == DIR_Entry::NIL); }
	int readdir_cont(DIR_Cli_AP *ap, DIR_List &list, int buf_len); /**< continue read the children nodes of myself. */
	int setattr(DIR_Cli_AP *ap, const DIR_Param *param); /**< update the attribute of the node (only size matters). */
	int getattr(DIR_Cli_AP *ap, DIR_Param *param); /**< get the attribute of the node. */
	/** @} */
	const char *path() const { return pwd; } /**< absolute path */
	/**
	 * @defgroup getattr Get Attributes.
	 * The following set of functions get the UP-TO-DATE 
	 * attributes of DIR_Entry.
	 */
	/** @{ */
	int get_type(DIR_Cli_AP *ap, int &type); /**< the type of the entry: dir or file. */
	int get_guid(DIR_Cli_AP *ap, guid_t &guid); /**< the guid. */
	int get_ctime(DIR_Cli_AP *ap, time_t &ctime); /**< the creation time */
	int get_size(DIR_Cli_AP *ap, off64_t &size); /**< the size */
	/** @} */
	int refresh_info(DIR_Cli_AP *ap); /**< called when we absolutely need the fresh info. */
	const DIR_Param * get_param() const { 
		if (expired) 
			return NULL; 
		return &cur_param; 
	}

	bool is_dir(DIR_Cli_AP *ap) { 
		if (expired) refresh_info(ap);
		return DIR_IS_DIR(cur_param.options);
	}
};

# endif // _DIR_CLI_MT_H_
