// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: dir_cli_opt.cc,v $ - directory server client command-line utilities' 
 *            option parser.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/05/06 09:19:22 $
 * Version :  $Revision: 1.7 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 
# include <stdio.h>
# include <string.h>
# include <getopt.h>
# include <ctype.h>
# include "tmpi_debug.h"
# include "dir_cli_opt.h"

# if defined(__DEBUG__)
# include "asmalloc.h"
# endif

static char *program_name=NULL;

static char *b2s(bool b)
{
	if (b) {
		return "TRUE";
	}
	return "FALSE";
}

void dump_dir_cli_opt(DIR_Cli_opt *opt)
{
	if (!opt) {
		printf("  <nul>.\n");
		return;
	}
	if (strlen(opt->srv)>0) {
		printf("  DIR Server: \"%s:%d\".\n", opt->srv, opt->port);
	}

	if (strlen(opt->bsf)>0) {
		printf("  DIR DB File: \"%s\".\n", opt->bsf);
	}

	printf("  Create-on-Demand: %s.\n", b2s(opt->creat));
	printf("  Create-Exclusion: %s.\n", b2s(opt->excl));
	printf("  Recursive-Deletion: %s.\n", b2s(opt->recur));
}

int DIR_Cli_opt::set_var(char *s, bool cpref, bool set_init, char *&var, int init_mask, int free_mask)
{
	if (!s) {
		tmpi_error(DBG_CALLER, "NULL Pointer.\n");
		return -1;
	}

	if (flag&init_mask) {
		// already initialized
		return 0;
	}

	if (strcmp(s, var)) {
		if (cpref) {
			var=s;
		}
		else {
			var=strdup(s);
			if (!var) {
				tmpi_error(DBG_FATAL, "Out of memory.\n");
				return -1;
			}
			flag|=free_mask;
		}
	}

	if (set_init) {
		flag|=init_mask;
	}

	return 0;
}

DIR_Cli_opt::~DIR_Cli_opt()
{
	if ((idir)&&(flag&IDIR_FREE)) {
		free(idir);
	}
	if ((bsf)&&(flag&BSF_FREE)) {
		free(bsf);
	}
	if ((srv)&&(flag&SRV_FREE)) {
		free(srv);
	}
}

bool DIR_Cli_opt::is_valid_num(char *p)
{
	if (!p)
		return false;

	while (*p) {
		if (!isdigit(*p)) {
			return false;
		}
		p++;
	}

	return true;
}

int DIR_Cli_opt::set_bool(char *b, bool set_init, bool &boolean, int init_mask)
{
	if (flag&init_mask) {
		return 0;
	}

	if (strcasecmp(b, "true")==0) {
		boolean=true;
	}
	else if (strcasecmp(b, "yes")==0) {
		boolean=true;
	}
	else if (strcasecmp(b, "false")==0) {
		boolean=false;
	}
	else if (strcasecmp(b, "no")==0) {
		boolean=false;
	}
	else if (is_valid_num(b)) {
		int n=atoi(b);
		if (n==0)
			boolean=false;
		else
			boolean=true;
	}
	else {
		tmpi_error(DBG_CALLER, "Bad boolean expression: %s.\n", b);
		return -1;
	}

	if (set_init) {
		flag|=init_mask;
	}

	return 0;
}

int DIR_Cli_opt::set_num(char *n, bool set_init, int &num, int init_mask)
{
	if (!n) {
		tmpi_error(DBG_CALLER, "NULL Pointer.\n");
		return -1;
	}

	if (!is_valid_num(n)) {
		tmpi_error(DBG_CALLER, "Bad numeric format: %s.\n", n);
		return -1;
	}

	if (flag&init_mask) {
		// already initialized
		return 0;
	}

	num=atoi(n);

	if (set_init) {
		flag|=init_mask;
	}

	return 0;
}

int DIR_Cli_opt::set_var_val(char *var, char *val)
{
	static struct look_table {
		char *var;
		int (DIR_Cli_opt::*f)(char *, bool, bool);
	} *p, table[] = {
		{"INIT_DIR", &DIR_Cli_opt::set_idir},
		{"BACKSTORE", &DIR_Cli_opt::set_bsf},
		{"SERVER", &DIR_Cli_opt::set_srv},
		{"PORT", &DIR_Cli_opt::set_port},
		{"CREATE", &DIR_Cli_opt::set_creat},
		{"EXCLUSIVE", &DIR_Cli_opt::set_excl},
		{"RECURSIVE", &DIR_Cli_opt::set_recur},
		{NULL, NULL}
	};

	p=table;
	while (p->var) {
		if (strcasecmp(p->var, var)==0) {
			return (this->*(p->f))(val, true, false);
		}
		p++;
	}
	return -1;
}

static char *chop(char *line)
{
	while (*line) {
		if (isspace(*line))
			line++;
		else
			break;
	}

	char *p=line+strlen(line)-1;
	while ((p>=line)&&(isspace(*p))) {
		*p='\0';
		p--;
	}

	return line;
}

void dir_cli_usage(int status)
{
	if (status != 0) {
		fprintf(stderr, "Try '%s --help' for more information.\n", program_name);
	}
	else {
		printf ("Usage: %s [options].\n", program_name);
		printf ("\
A DIR Client Shared Command Line Interface.\n\
\n\
1. DIR Database Location:\n\
    -f, --dir-db <db_file>    Set dir_db file (cannot combine with -S).\n\
    -S, --dir-srv <server>    Set dir-server address (cannot combine with -f).\n\
    -p, --port <port>         Set dir-server port (only combine with -S).\n\
\n\
2. Initial Settings:\n\
    -I, --dir <init_dir>      Set the initial directory.\n\
    -c, --create              Set default to create-on-demond.\n\
    -C, --no-create           Set default to no-create.\n\
    -e, --exclusive           Set default to fail-if-exist.\n\
    -E, --no-exclusive        Set default to no-fail-if-exist.\n\
    -r, --recursive           Set default to recursive (deletion).\n\
    -R, --no-recursive        Set default to no-recursive (deletion).\n\
 \n\
3. Debugging:\n\
    -s, --silent              Display/log only fatal information.\n\
    -d, --debug               Keep a detailed log.\n\
  \n\
4. Information:\n\
    -h, --help                Display this help screen.\n\
    -V, --version             Display version number.\n\
");
	}

	exit(status);
}

void parse_dir_cli_opt(int argc, char *argv[], DIR_Cli_opt *opt, 
		void (*version)()) 
{
	static struct option const long_options[] = {
		{"dir", required_argument, 0, 'I'},
		{"dir-db", required_argument, 0, 'f'},
		{"dir-srv", required_argument, 0, 'S'},
		{"port", required_argument, 0, 'p'},
		{"create", no_argument, 0, 'c'},
		{"no-create", no_argument, 0, 'C'},
		{"exclusive", no_argument, 0, 'e'},
		{"no-exclusive", no_argument, 0, 'E'},
		{"recursive", no_argument, 0, 'r'},
		{"no-recursive", no_argument, 0, 'R'},
		{"silent", no_argument, 0, 's'},
		{"debug", no_argument, 0, 'd'},
		{"help", no_argument, 0, 'h'},
		{"version", no_argument, 0, 'V'},
		{NULL, 0, 0, 0}
	};

	int optc;

	program_name=argv[0];
	while ((optc=getopt_long(argc, argv, \
					"I:f:S:p:cCeErRsdhV", long_options, NULL)) !=-1) {
		switch (optc) {
			case 0:
				break;
			case 'I':
				if (opt->set_idir(optarg, true, true)<0) {
					dir_cli_usage(0);
				}
				break;
			case 'f':
				if (opt->set_bsf(optarg, true, true)<0) {
					dir_cli_usage(0);
				}
				break;
			case 'S':
				if (opt->set_srv(optarg, true, true)<0) {
					dir_cli_usage(0);
				}
				break;
			case 'p':
				if (opt->set_port(optarg, true)<0) {
					dir_cli_usage(0);
				}
				break;
			case 'c':
				if (opt->set_creat("1", true)<0) {
					dir_cli_usage(0);
				}
				break;
			case 'C':
				if (opt->set_creat("0", true)<0) {
					dir_cli_usage(0);
				}
				break;
			case 'e':
				if (opt->set_excl("1", true)<0) {
					dir_cli_usage(0);
				}
				break;
			case 'E':
				if (opt->set_excl("0", true)<0) {
					dir_cli_usage(0);
				}
				break;
			case 'r':
				if (opt->set_recur("1", true)<0) {
					dir_cli_usage(0);
				}
				break;
			case 'R':
				if (opt->set_recur("0", true)<0) {
					dir_cli_usage(0);
				}
				break;
			case 's':
				tmpi_set_dbg_level(DBG_FATAL-1);
				tmpi_hide_location();
				break;
			case 'd':
				tmpi_set_dbg_level(DBG_ALL);
				tmpi_show_location();
				break;
			case 'V':
				(*version)();
				break;
			case 'h':
				dir_cli_usage(0);
				break;
			default:
				dir_cli_usage(1);
		}
	}

	if (optind<argc) {
		if (optind<argc-1) {
			dir_cli_usage(1);
		}

		if (opt->set_bsf(argv[optind], true, true)<0) {
			dir_cli_usage(0);
		}
	}

	if ((strlen(opt->srv)>0)&&(strlen(opt->bsf)>0)) {
		dir_cli_usage(0);
	}

	if ((strlen(opt->srv)==0)&&(strlen(opt->bsf)==0)) {
		dir_cli_usage(0);
	}

	return;
}

static bool is_empty_line(char *line)
{
	while (*line) {
		if (!isspace(*line))
			return false;
		line++;
	}

	return true;
}

int parse_dir_cli_config(DIR_Cli_opt *opt, char *config)
{
	if ((!opt)||(config))
		return -1;

	FILE *fp=fopen(config, "r");

	if (!fp) {
		tmpi_error(DBG_CALLER, "Cannot open configuration file %s.\n", config);
		return -1;
	}

	char buf[1024];
	char *var, *val;
	int lineno=0;

	while (fgets(buf, 1024, fp)) {
		char *p;

		lineno++;
		if ((p=strchr(buf, '#'))) {
			*p='\0';
		}

		if (is_empty_line(buf)) {
			continue;
		}
		
		if ((p=strchr(buf, '='))) {
			*p='\0';
			var=chop(buf);
			val=chop(p+1);
			if (opt->set_var_val(var, val)<0) {
				tmpi_error(DBG_VERBOSE, "Error encountered @ config file line %d.\n", 
						lineno);
			}
		}
		else {
			tmpi_error(DBG_CALLER, "Bad config file format @ line %d.\n", lineno);
		}
	}

	return 0;
}
