// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: dir_common.h,v $ - commonly shared constants and structs between
 *            directory server and clients.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/08/01 11:21:20 $
 * Version :  $Revision: 1.21 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 
# ifndef _DIR_COMMON_H_
# define _DIR_COMMON_H_

# include <sys/types.h>
# include "guid.h"
# include "pstree.h"

#if !defined __off64_t_defined
typedef __off64_t off64_t;
# define __off64_t_defined
#endif

typedef unsigned int dir_node_t; 

/* request ID */
# define DIR_REQ_TYPE_IDENT 0x10000 
# define DIR_REQ_TYPE_QUERY 0x20000
# define DIR_REQ_TYPE_DEL 0x40000
# define DIR_REQ_TYPE_SETP 0x80000

/* response status */
# define DIR_RES_OKAY 0x0
# define DIR_RES_ECREATE 0x1
# define DIR_RES_EEXIST 0x2
# define DIR_RES_ENONEXIST 0x3
# define DIR_RES_ESERVER 0x4
# define DIR_RES_ENETWORK 0x5
# define DIR_RES_ETOUT 0x6
# define DIR_RES_EDATA 0x7
# define DIR_RES_ECLIENT 0x8
# define DIR_RES_EINVALID 0x9
# define DIR_RES_EBUFFER 0x10
# define DIR_RES_EMEM 0x11
# define DIR_RES_EAGAIN 0x12
# define DIR_RES_ERAGAIN 0x13
# define DIR_RES_EEND 0x14
# define DIR_RES_ESELF 0x15
# define DIR_RES_EISDIR 0x16
# define DIR_RES_ENOTDIR 0x17
# define DIR_RES_MAX_ERROR 0x100

# define DIRSRV_PORT_DEFAULT 3832

/* string length */
# define DIR_MAX_NAME_LEN 256
# define DIR_MAX_PATH_LEN 1024

/*---------------------------------*/
/*             flags               */
/*---------------------------------*/

/* create if not existent */
# define DIR_CREATE 0x1

/* DIR_EXCL can only be used with DIR_CREATE */
# define DIR_EXCLUSIVE 0x2 

/* recursive removal */
# define DIR_RECURSIVE 0x4 

/*---------------------------------*/
/*   path string handling          */
/*---------------------------------*/

/* parse path return flags */
# define DIR_PATH_HOST 0x1
# define DIR_PATH_PORT 0x2
# define DIR_PATH_LPATH 0x4

// return value is a bit-wise OR of the above flags.
int dir_parse_path(char *path, char *host, int len, int *port, char *lpath, int len2);
// returns 0 on success and -1 otherwise.
int dir_std_path(const char *base, const char *path, char *std_path, int len);
int dir_opt_path(const char *base, const char *std_path, char *opt_path, int len);

/*---------------------------------*/
/*   Data structures in query      */
/*---------------------------------*/

struct DIR_REQ_IDENT {
	int dummy; // nothing.
};

struct DIR_Param {
	guid_t guid;
	unsigned int options; // the highest bit of the options is reserved.
						  // and must be 0.
	off64_t size; /**< The size of the node. */
	time_t t_create; // readonly attribute
public:
	// DIR_Param(bool dummy) : guid(Guid::GUID_NIL), t_create(0) {}
	void reset () 
	{ 
		memcpy(&guid, &(Guid::GUID_NIL), sizeof(guid_t)); 
		options=0;
		size=0;
		t_create=0; 
		return; 
	}
};

struct DIR_REQ_QUERY {
	dir_node_t base;
	dir_node_t b_child;
	dir_node_t b_parent;
	dir_node_t next_list; // if != NIL, base, b_child, b_parent, path are ignored.
	int buf_len; // client buffer to hold the result. 
				 // if less than one entry, nothing will be returned.
	unsigned int flags;
	DIR_Param param; /**< Only examined when flags&DIR_CREATE is true */
	char path[1];
};

# define DIR_REQ_QUERY_HEADER_SZ ((int)&(((DIR_REQ_QUERY *)0L)->path))

/**
 * Note that if the cwd has been deleted,
 * return code would be DIR_RES_ESELF.
 * Even though it is an error code, the directory
 * actually has been deleted, it is necessary
 * to notify the client that the current
 * directory is no longer valid.
 */
struct DIR_REQ_DEL {
	dir_node_t base;
	dir_node_t b_child;
	dir_node_t b_parent;
	unsigned int flags;
	char path[1];
};

# define DIR_REQ_DEL_HEADER_SZ ((int)&(((DIR_REQ_DEL *)0L)->path))

// the lower 16 bits are user-modifiable, the top 16 bits
// are system reserved. specifically, the highest bit
// is reserved for future expansion.
// The lowest 8 bits are for the node types
// Not used yet.
// uninitialized
# define DIR_OPTION_INIT 0x0
# define DIR_OPTION_FILE 0x1
// directory
# define DIR_OPTION_DIR 0x2
// symbolic link
# define DIR_OPTION_LINK 0x3

# define DIR_TYPE_FILE 0x1
# define DIR_TYPE_DIR 0x2

# define DIR_IS_DIR(option) (option & DIR_OPTION_DIR)

// sweeping flags

# define DIR_OPTION_SWEEP 0x4000
/**
# define DIR_OPTION_SWEEP2 0x8000
# define DIR_OPTION_SWEEP_MASK 0xc000
**/

// loaded (only for the root record
# define DIR_OPTION_LOAD 0x2000

// error
# define DIR_OPTION_ERR 0xff
// delete
# define DIR_OPTION_DEL 0x40000000
# define DIR_OPTION_USR_MASK 0xffff
# define DIR_OPTION_SYS_MASK 0xffff0000

struct DIR_REQ_SETP {
	dir_node_t base;
	dir_node_t b_child;
	dir_node_t b_parent;
	DIR_Param param;
	char path[1];
};

# define DIR_REQ_SETP_HEADER_SZ ((int)&(((DIR_REQ_SETP *)0L)->path))

struct DIR_REQ {
	int type;
	union {
		struct DIR_REQ_IDENT d_ident;
		struct DIR_REQ_QUERY d_query;
		struct DIR_REQ_DEL d_del;
		struct DIR_REQ_SETP d_setp;
	} u;
};

# define DIR_REQ_HEADER_SZ ((int)&(((DIR_REQ *)0L)->u))

struct DIR_RES_IDENT {
	char ident[1];
};

struct DIR_RES_QUERY {
	dir_node_t nid;  // node ID of THIS directory, invalid when request's next_list !=NIL.
	dir_node_t n_child; // invalid when the request's next_list != NIL
	dir_node_t n_parent; // invalid when the request's next_list != NIL
	dir_node_t n_sibling; // invalid when the request's next_list != NIL
	dir_node_t n_next; // the next node that is not stored in the buf.
	DIR_Param param; // invalid when the request's next_list != NIL
	int buf_len;
	char buf[1]; /* packed data, [nid, dir_param, name], in fact, 
				  * nid will probably be ignored by the 
				  * client, but it is returned by the
				  * server just for further optimization
				  */ 
};

# define DIR_RES_QUERY_HEADER_SZ ((int)&(((DIR_RES_QUERY *)0L)->buf))

struct DIR_RES_DEL {
	// info for the node that links to the deleted node, 
	dir_node_t nid;  
	dir_node_t n_child;
	dir_node_t n_parent; 
	dir_node_t n_sibling; 
	DIR_Param param; 
	// int dummy;
};

struct DIR_RES_SETP {
	// info for the node that resets its parameter.
	dir_node_t nid;  
	dir_node_t n_child;
	dir_node_t n_parent; 
	dir_node_t n_sibling; 
	DIR_Param param; 
	// int dummy;
};

struct DIR_RES {
	int status;
	union {
		struct DIR_RES_IDENT d_ident; 
		struct DIR_RES_QUERY d_query; 
		struct DIR_RES_DEL d_del; 
		struct DIR_RES_SETP d_setp; 
	} u;
};

# if defined(__PSTREE_DBG__)
# 	include "pstree_core.h"
# 	define PSTREE PsTree_Core
# else
# 	include "pstree.h"
# 	define PSTREE PsTree
# endif

# define DIRDB_ROOT	((dir_node_t)PSTREE::ROOT)
# define DIRDB_ERR	((dir_node_t)PSTREE::FREE)
# define DIRDB_NIL	((dir_node_t)PSTREE::NIL)

# define DIR_RES_HEADER_SZ ((int)&(((DIR_RES *)0L)->u))

# endif
