// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: dir_srv.cc,v $ - directory server implementation.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/07/16 06:40:22 $
 * Version :  $Revision: 1.14 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 


# include <stdio.h>
# include <ctype.h>
# include <string.h>
# include <getopt.h>
# include <signal.h>
# include <unistd.h>

# include "app.h"
# include "atomic.h"
# include "tmpi_debug.h"
# include "sock_sos.h"
# include "db_worker.h"

static char *program_name=NULL;
static DB_Worker *server=NULL;
static char *config_file;
static unsigned int server_stop=0;
char *Id="$Id: dir_srv.cc,v 1.14 2003/07/16 06:40:22 htang Exp $";

static char *chop(char *line)
{
	while (*line) {
		if (isspace(*line))
			line++;
		else
			break;
	}

	char *p=line+strlen(line)-1;
	while ((p>=line)&&(isspace(*p))) {
		*p='\0';
		p--;
	}

	return line;
}

static char *get_RCS_val(char *var)
{
	static char buf[1024];
	char *p;

	if ((p=strchr(var, ':'))==NULL) {
		return var;
	}
	buf[1023]='\0';
	strncpy(buf, p+1, 1024);
	p=buf+strlen(buf)-1;
	while ((p>=buf)&&(*p=='$')) {
		*p='\0';
		p--;
	}

	return chop(buf);
}

static void usage(int status)
{
	if (status != 0) {
		fprintf(stderr, "Try '%s --help' for more information.\n", program_name);
	}
	else {
		printf ("Usage: %s [options]\n", program_name);
		printf ("\
The directory server daemon.\n\
   -f, --config <cfg_file>   Set directory server configuration file.\n\
   -s, --silent              Display/log only fatal information.\n\
   -d, --debug               Keep a detailed log.\n\
   -h, --help                Display this help screen.\n\
   -V, --version             Display version number.\n\
");
	}

	exit(status);
}

static void do_version()
{
	static char *version="$Revision: 1.14 $";
	static char *author="Hong Tang";
	static char *date="$Date: 2003/07/16 06:40:22 $";

	{
		printf("Directory Server Daemon (%s) %s.\n", program_name, get_RCS_val(version));
		printf("Copyright (c) University of California, Santa Barbara.\n");
		printf("Author: %s.\n", get_RCS_val(author));
		printf("Last Modification Date: %s GMT.\n", get_RCS_val(date));
		printf("Specify the --help option for further information.\n");
	}

	exit(0);
}

static void stop_server(int signum)
{
	tmpi_error(DBG_VERBOSE, "Got signal %d.\n", signum);
	server_stop|=APP_SYS_QUIT;
	commit32(&server_stop);
	if (server)
		server->stop();
	signal(signum, &stop_server);
}

static void set_signals()
{
	// set ignore
	static const int ignore_signals[] = {
		SIGHUP,
		SIGPIPE,
		SIGALRM,
		SIGUSR1,
		SIGUSR2,
		SIGTTIN,
		SIGTTOU,
		SIGURG,
		SIGIO,
		SIGPOLL,
		SIGVTALRM,
		SIGPROF,
		-1
	};

	// set signals
	static const int stop_signals[] = {
		SIGINT,
		SIGQUIT,
		SIGILL,
		SIGTRAP,
		SIGABRT,
		SIGIOT,
		SIGFPE,
		SIGBUS,
		SIGSEGV,
		SIGSTKFLT,
		SIGTERM,
		SIGXCPU,
		SIGXFSZ,
		SIGPWR,
		-1
	};

	static const int default_signals[] = {
		SIGSTOP,
		SIGCONT,
		SIGTSTP,
		SIGCHLD,
		SIGCLD,
		-1
	};

	const int *p=ignore_signals;
	while (*p>=0) {
		signal(*p, SIG_IGN);
		p++;
	}

	p=stop_signals;
	while (*p>=0) {
		signal(*p, stop_server);
		p++;
	}
	
	p=default_signals;
	while (*p>=0) {
		signal(*p, SIG_DFL);
		p++;
	}
}

static void parse_cmd_opt(int argc, char *argv[])
{
	static struct option const long_options[] = {
		{"config", required_argument, 0, 'f'},
		{"help", no_argument, 0, 'h'},
		{"silent", no_argument, 0, 's'},
		{"debug", no_argument, 0, 'd'},
		{"version", no_argument, 0, 'V'},
		{NULL, 0, 0, 0}
	};

	int optc;

	while ((optc=getopt_long(argc, argv, \
					"f:hsdV", long_options, NULL)) !=-1) {
		switch (optc) {
			case 'f':
				config_file=strdup(optarg);
				break;
			case 'h':
				usage(0);
				break;
			case 'V':
				do_version();
				break;
			case 's':
				tmpi_set_dbg_level(DBG_FATAL-1);
				tmpi_hide_location();
				break;
			case 'd':
				tmpi_set_dbg_level(DBG_ALL);
				tmpi_show_location();
				break;
			default:
				usage(1);
		}
	}

	if (config_file==NULL) {
		usage(0);
	}
}

int main(int argc, char *argv[])
{
	program_name=argv[0];

# if defined(__DEBUG__)
	tmpi_set_dbg_level(DBG_ALL);
	tmpi_show_location();
# else
	tmpi_set_dbg_level(DBG_VERBOSE);
	tmpi_hide_location();
# endif

	parse_cmd_opt(argc, argv);

	int ret=-1;
	FILE *fp=NULL;

	try {
		DirSrvConf dconf(config_file);

		{
			// set log
			fp=fopen(dconf.log_file(), "a+");
			if (!fp) {
				tmpi_error(DBG_CALLER, "Cannot open log file %s.\n", dconf.log_file());
			}
			else {
				tmpi_error(DBG_APP, "Redirecting log output to %s\n", dconf.log_file());
				tmpi_set_output(fp);
			}
		}

		tmpi_error(DBG_APP, "Preparing to set SOS pointer.\n");
		if (sock_set_sos_ptr(&server_stop)<0) {
			throw std::string("Cannot set application SOS flag.");
		}

		tmpi_error(DBG_APP, "Initializing servers ...\n");
		server=new DB_Worker(dconf);

		tmpi_error(DBG_APP, "Set signal handlers ...\n");
		set_signals();

		tmpi_error(DBG_APP, "Starting servers ...\n");
		if (server->start() == false) {
			throw std::string("Cannot start the servers.");
		}

		tmpi_error(DBG_APP, "%s running now at port %d.\n", program_name, server->get_port());
	
		while (1)  {
			if (!server->is_running()) // health check
				break;
			sleep(1);
		}

		ret=0;
	}
	catch (std::string &msg) {
		tmpi_error(DBG_FATAL, "Error: %s.\n", msg.c_str());
	}

	if (server) {
		tmpi_error(DBG_APP, "Sending signals to servers...\n");
		server->stop();
		tmpi_error(DBG_APP, "Waiting for servers to stop...\n");
		server->finalize();
		tmpi_error(DBG_APP, "Server stopped.\n");
	}


	if (fp) {
		fclose(fp);
	}

	return ret;
}
