// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: dirdb.cc,v $ - directory server database.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/09/24 01:20:32 $
 * Version :  $Revision: 1.31 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 
# include <sys/stat.h>
# include <stdlib.h>
# include <pthread.h>
# include <time.h>
# include <queue>
# include "tmpi_debug.h"
# include "resmon.h"
# include "guid.h"
# include "dirdb.h"
# include "dir_common.h"

# if defined(__DEBUG__)
# include "asmalloc.h"
# endif

using namespace std;

/**
 * initialize the DIR_REC record for it
 * to be written to the database. this
 * is used when a new node is created.
 */
int DIR_DB::set_rec(DIR_REC *rec, guid_t *guid, char *path, int options, int *len, off64_t size)
{
	if ((!len)||(!path))
		return -1;

	int plen=strlen(path);

	if (plen>=DIR_MAX_NAME_LEN)
		return -1;

	if (guid==NULL) {
		Guid root_guid;
		root_guid.gen();
		root_guid.pack(rec->param.guid);
	}
	else if (memcmp(guid, &(Guid::GUID_NIL), GUID_LEN)==0) {
		Guid root_guid;
		root_guid.gen();
		root_guid.pack(rec->param.guid);
	}
	else {
		memcpy(&(rec->param.guid), guid, GUID_LEN);
		// fprintf(stderr, "User specified GUID used.\n");
	}
	time(&(rec->param.t_create));
	rec->param.options=options;
	rec->param.size=size;
	strcpy(rec->path, path);
	*len=DIR_REC_HEADER_SZ+plen+1;

	return 0;
}

/**
 * adjusting the parameter of an existing record. Only GUID, 
 * user specifiable flag options, and size are modifiable. 
 */
int DIR_DB::set_param(DIR_Param *dest, DIR_Param *src)
{
	if ((!dest)||(!src))
		return -1;

	if (memcmp(&(src->guid), &(Guid::GUID_NIL), GUID_LEN)==0) {
		Guid guid;
		guid.gen();
		guid.pack(dest->guid);
	}
	else {
		memcpy(&(dest->guid), &(src->guid), GUID_LEN);
	}

	dest->options&=DIR_OPTION_SYS_MASK;
	dest->options|=(src->options&DIR_OPTION_USR_MASK);
	dest->size=src->size;

	return 0;
}

/* 128 MB */
# define CACHE_SZ_DEFAULT 0x20000
/* 1 GB */
# define CACHE_SZ_MAX 0x100000

DIR_DB::DIR_DB(char *db_file) : 
	PSTREE(sizeof(DIR_REC)),
	dir_cache(true)
{
	unsigned int cache_sz; // we try to use as much cache memory as possible.
				  // maximum of (minimum of current free memory 
				  // and 3/4 of total physical memory) and half of the 
				  // total physical memory
	ResMon::MEM mem;
	memset(&mem, 0, sizeof(mem));
	if (ResMon::get_mem(&mem)<0) {
		cache_sz=CACHE_SZ_DEFAULT;
	}
	else {
		cache_sz=(mem.mfree>mem.mtotal/2)?mem.mfree:(mem.mtotal/2);
		cache_sz=(cache_sz>(mem.mtotal*3/4))?(mem.mtotal*3/4):cache_sz;
	}

	if (cache_sz > CACHE_SZ_MAX) {
		cache_sz = CACHE_SZ_MAX;
	}

	struct stat ss;
	if (stat(db_file, &ss)) { // does not exist, create it.
		DIR_REC rec;
		int len;
		set_rec(&rec, NULL, "/", DIR_OPTION_DIR, &len, 0);
		create(db_file, &rec, len, cache_sz);
		PSTREE::sync();
	}
	else {
		load(db_file, cache_sz);
	}

	ident=string(db_file);

	pthread_mutex_init(&lock, NULL);
}


void DIR_DB::DIRDB_REC::init(node_t par, node_t chld)
{
	parent=par;
	child=chld;

	sibling=DIRDB_ERR;
	memset(&(guid()), 0, GUID_LEN);
	t_create()=0;
	options()=0;
	size()=0;
	path()[0]='\0';
}

DIR_DB::Traverse_cache::Traverse_cache(node_t base, node_t parent, node_t child, char *path) :
	n_prev(DIRDB_ERR),
	r_prev(NULL),
	l_prev(0),
	n_cur(DIRDB_ERR),
	r_cur(NULL),
	l_cur(0),
	s(NULL),
	p(NULL)
{
	if (!path)
		return;

	r_prev=(DIRDB_REC *)malloc(REC_HEADER_SZ+sizeof(DIR_REC));
	r_cur=(DIRDB_REC *)malloc(REC_HEADER_SZ+sizeof(DIR_REC));

	if ((!r_prev)&&(!r_cur)) {
		if (r_prev) {
			free(r_prev);
			r_prev=NULL;
		}
		if (r_cur) {
			free(r_cur);
			r_cur=NULL;
		}

		return;
	}

	n_prev=DIRDB_ERR;
	r_prev->init();

	n_cur=base;
	r_cur->init(parent, child);

	s=strdup(path);
	p=s;
}

void DIR_DB::Traverse_cache::release()
{
	if (r_prev) {
		free(r_prev);
		r_prev=NULL;
	}

	if (r_cur) {
		free(r_cur);
		r_cur=NULL;
	}

	if (s) {
		free(s);
		p=NULL;
	}
}

DIR_DB::Traverse_cache::~Traverse_cache()
{
	release();
}

void DIR_DB::Traverse_cache::refresh_prev(DIRDB_REC *node, int len)
{
	if ((!node)||(len<=0))
		return;

# if defined(__DEBUG__)
	if (len>(int)sizeof(DIR_REC)) {
		tmpi_error(DBG_INTERNAL, "buffer overflow %d>%d.\n", len, sizeof(DIR_REC));
		len=sizeof(DIR_REC);
	}
# endif

	memcpy(r_prev, node, REC_HEADER_SZ+len);
	l_prev=len;
}

void DIR_DB::Traverse_cache::refresh(DIRDB_REC *node, int len)
{
	if ((!node)||(len<=0))
		return;

# if defined(__DEBUG__)
	if (len>(int)sizeof(DIR_REC)) {
		tmpi_error(DBG_INTERNAL, "buffer overflow %d>%d.\n", len, sizeof(DIR_REC));
		len=sizeof(DIR_REC);
	}
# endif

	memcpy(r_cur, node, REC_HEADER_SZ+len);
	l_cur=len;
}

void DIR_DB::Traverse_cache::dup(Traverse_cache *cache)
{
	if ((!cache)||(!cache->r_prev)||(!cache->r_cur))
		return;
	n_prev=cache->n_prev;
	n_cur=cache->n_cur;
	refresh_prev(cache->r_prev, cache->l_prev);
	refresh(cache->r_cur, cache->l_cur);

	if (s)
		free(s);
	s=strdup(cache->p);
	p=s;
}

void DIR_DB::Traverse_cache::shift(node_t n, DIRDB_REC *next, int len)
{
	if ((!next)||(len<=0))
		return;

	n_prev=n_cur;
	l_prev=l_cur;

	DIRDB_REC *tmp=r_prev;
	r_prev=r_cur;

	n_cur=n;
	r_cur=tmp;

	refresh(next, len);

	return;
}

# if defined(__DEBUG__)
# undef get_node
DIR_DB::DIRDB_REC * DIR_DB::get_node1(node_t n, unsigned int *len, char *f, int l)
{
	static bool first_time=true;
	DIRDB_REC *ret=(DIRDB_REC *)PSTREE::get_node(n, len);
	if (first_time) {
		first_time=false;
		tmpi_error(DBG_INTERNAL, "get_node() checked in debug mode.\n");
	}
	
	tmpi_error(DBG_INTERNAL, "get_node(%lu) called from [%s:%d].\n", n, f, l);

	if (ret) {
		/* format check */
		if (ret->parent==DIRDB_ERR) {
			tmpi_error(DBG_INTERNAL, "<%lu>.parent==ERR.\n", n);
		}

		if (ret->parent==DIRDB_NIL) {
			tmpi_error(DBG_INTERNAL, "<%lu>.parent==NIL.\n", n);
		}

		if (ret->child==DIRDB_ERR) {
			tmpi_error(DBG_INTERNAL, "<%lu>.child==ERR.\n", n);
		}

		if (ret->sibling==DIRDB_ERR) {
			tmpi_error(DBG_INTERNAL, "<%lu>.sibling==ERR.\n", n);
		}

		if (ret->t_create()==(time_t)0) {
			tmpi_error(DBG_INTERNAL, "<%lu>.t_create==0.\n", n);
		}

		if (ret->path()[0]=='\0') {
			tmpi_error(DBG_INTERNAL, "<%lu>.path==NULL.\n", n);
		}

		if (memcmp((ret->guid()), Guid::GUID_NIL, GUID_LEN)==0) {
			tmpi_error(DBG_INTERNAL, "<%lu>.guid==NIL.\n", n);
		}
	}

	return ret;
}
# define get_node(n, pl) get_node1(n, pl, __FILE__, __LINE__)
# endif

bool DIR_DB::refresh_cache_prev(DIR_DB::Traverse_cache *cache)
{
	unsigned int len;
	DIRDB_REC *node=(DIRDB_REC *)get_node(cache->n_prev, &len);

	if (node) {
		if (!(node->options()&DIR_OPTION_DEL))
			dir_cache.ins(node->parent, node->path(), cache->n_prev);
		cache->refresh_prev(node, len);
		free_node((Rec *)node);
		return true;
	}

	return false;
}

bool DIR_DB::refresh_cache_current(DIR_DB::Traverse_cache *cache)
{
	unsigned int len;
	DIRDB_REC *node=(DIRDB_REC *)get_node(cache->n_cur, &len);
	if (node) {
		if (!(node->options()&DIR_OPTION_DEL))
			dir_cache.ins(node->parent, node->path(), cache->n_cur);
		cache->refresh(node, len);
		free_node((Rec *)node);
		return true;
	}

	return false;
}

bool DIR_DB::check_cache_parent(DIR_DB::Traverse_cache *cache)
{
	if (cache->r_cur->parent==DIRDB_ERR)
		return refresh_cache_current(cache);
	return true;
}

bool DIR_DB::check_cache_child(DIR_DB::Traverse_cache *cache)
{
	if (cache->r_cur->child==DIRDB_ERR)
		return refresh_cache_current(cache);
	return true;
}

bool DIR_DB::check_cache(DIR_DB::Traverse_cache *cache)
{
	if ((cache->r_cur->child==DIRDB_ERR)||(cache->r_cur->parent==DIRDB_ERR)||
		(cache->r_cur->t_create()==(time_t)0) 
# if defined(__DEBUG__)
		||(memcmp((cache->r_cur->guid()), Guid::GUID_NIL, GUID_LEN)==0)
# endif
		)

		return refresh_cache_current(cache);
	return true;
}

/**
 * traverse the tree till the path no longer exists.
 * returns the number of levels it traversed, or -1
 * on error. The only flag it cares about is DIR_CREATE.
 */
int DIR_DB::traverse(DIR_DB::Traverse_cache *cache, bool use_dir_cache, \
		bool update_dir_cache, int flags, DIR_Param *param)
{
	if (!cache)
		return -1;

	// set the base
	if (cache->p[0]=='/') {
		cache->r_cur->init(DIRDB_ROOT, DIRDB_ERR);
		cache->p++;
		cache->n_cur=DIRDB_ROOT;
	}

	char *end_p=cache->p+strlen(cache->p);

	int levels=0;
	bool last_level=false;
	do {
		if (cache->p[0]=='\0')
			break;

		// get next component
		char *next_p=strchr(cache->p, '/');
		if (!next_p) {
			next_p=end_p;
			last_level=true;
		}
		else {
			*next_p='\0';
			next_p++;
			if (&next_p=='\0')
				last_level=true;
		}

		if ((cache->p[0]=='.')&&(cache->p[1]=='\0')) {
			/* current path, do nothing */
		}
		else if ((cache->p[0]=='.')&&(cache->p[1]=='.')&&(cache->p[2]=='\0')) {
			if (check_cache_parent(cache)==false) {
				tmpi_error(DBG_INTERNAL, "DIR_DB data structure broken.\n");
				return -1;
			}

			node_t n=cache->r_cur->parent;
			DIRDB_REC *node;
			unsigned int len;
			node=(DIRDB_REC *)get_node(n, &len);
			if (!node) {
				return -1;
			}
			else {
				if (update_dir_cache && (!(node->options()&DIR_OPTION_DEL)) )
					dir_cache.ins(node->parent, node->path(), n);
				cache->shift(n, node, len);
				free_node((Rec *)node);
			}
		}
		else {
			// normal path component.

			// look in the cache
			node_t n_cur=cache->n_cur;
			node_t n=DIRDB_NIL;
			DIRDB_REC *node;
			unsigned int len;

			if (use_dir_cache && (dir_cache.find(n_cur, cache->p, n)==0)) {
				// cache hit.
				node=(DIRDB_REC *)get_node(n, &len);
				if (node) {
					if ((node->parent!=n_cur)||(strcmp(node->path(), cache->p))) {
						// I do not remove the old one
						// which might just be a normal conflict
						if (update_dir_cache && (!(node->options()&DIR_OPTION_DEL)) )
							dir_cache.ins(node->parent, node->path(), n);
						n=DIRDB_NIL;
					}
					else {
						cache->shift(n, node, len);
					}

					free_node((Rec *)node);
				}
				else {
					n=DIRDB_NIL;
				}
			}

			// cache miss
			if (n==DIRDB_NIL) {
				if (check_cache_child(cache)==false) {
					tmpi_error(DBG_INTERNAL, "DIR_DB data structure broken.\n");
					return -1;
				}

				n=cache->r_cur->child;

				while (n!=DIRDB_NIL) {
					node=(DIRDB_REC *)get_node(n, &len);
					if (!node) {
						return -1;
					}
					else {
						if (update_dir_cache && (!(node->options()&DIR_OPTION_DEL)) )
							dir_cache.ins(node->parent, node->path(), n);
						cache->shift(n, node, len);
						if (strcmp(node->path(), cache->p)==0) {
							free_node((Rec *)node);
							break;
						}
						n=node->sibling;
						free_node((Rec *)node);
					}
				}

				if (n==DIRDB_NIL) { // not found.
					if (cache->n_cur!=n_cur) {
						// reset the r_cur to be the parent.
						node=(DIRDB_REC *)get_node(n_cur, &len);
						if (!node)
							return -1;
						if (update_dir_cache && (!(node->options()&DIR_OPTION_DEL)) )
							dir_cache.ins(node->parent, node->path(), n_cur);
						cache->shift(n_cur, node, len);
						free_node((Rec *)node);
					}

					/**
					 * If cur.child == DIRDB_NIL, then add to the first.
					 * Else, prev should be the last node and add to prev.
					 */
					if (flags&DIR_CREATE) {
						DIR_REC rec;
						int rlen;
						if (!last_level || (param == NULL)) {
							if (set_rec(&rec, NULL, cache->p, DIR_OPTION_DIR, &rlen, 0)) {
								return -1; /* should be DIR_RES_EINVALID */
							}
						}
						else { /* last level, honor the user specified parameters. */
							if (set_rec(&rec, &(param->guid), cache->p, param->options, &rlen, param->size)) {
								return -1; /* should be DIR_RES_EINVALID */
							}
						}

						if (check_cache_child(cache)==false) {
							tmpi_error(DBG_INTERNAL, "DIR_DB data structure broken.\n");
							return -1;
						}

						node_t nc;
						if (cache->r_cur->child==DIRDB_NIL) {
							nc=make_child(n_cur, &rec, rlen, DIRDB_NIL);
						}
						else {
							nc=make_child(n_cur, &rec, rlen, cache->n_prev);
						}

						if (nc==DIRDB_ERR)
							return -1;
						else {
							node=(DIRDB_REC *)get_node(nc, &len);
							if (!node)
								return -1;

							if (update_dir_cache && (!(node->options()&DIR_OPTION_DEL)) )
								dir_cache.ins(node->parent, node->path(), nc);
							cache->shift(nc, node, len);
							free_node((Rec *)node);
						}
					}
					else {
						// restore last component.
						if (*next_p!='\0')
							*(next_p-1)='/';
						break;
					}
				}
			}
		}

		cache->p=next_p;
		levels++;
	} while (1);
	
	if (check_cache(cache)==false) {
		tmpi_error(DBG_INTERNAL, "DIR_DB data structure broken.\n");
		return -1;
	}

	return levels;
}

int DIR_DB::query(DIR_REQ_QUERY *req, DIR_RES_QUERY *res)
{
	/**
# if defined(__DEBUG__)
	tmpi_error(DBG_INTERNAL, "In DIR_DB::query.\n");
# endif
	*/

	if (!is_loaded())
		return DIR_RES_ESERVER;

	node_t next_node=DIRDB_ERR;

	if (req->next_list==DIRDB_ERR) { /* search path */
		/**
# if defined(__DEBUG__)
		tmpi_error(DBG_INTERNAL, "DIR_DB::query:seach_path=\"%s\".\n", req->path);
# endif
		*/
		char std_path[DIR_MAX_PATH_LEN];
		if (dir_std_path(NULL, req->path, std_path, DIR_MAX_PATH_LEN)) {
			return DIR_RES_EINVALID;
		}

		Traverse_cache cache(req->base, req->b_parent, req->b_child, std_path);
		if (!cache.good()) {
			tmpi_error(DBG_APP, "Cannot initialize Traverse_cache (out of memory?).\n");
			return DIR_RES_ESERVER;
		}

		if (traverse(&cache, true, true)<0) {
			return DIR_RES_ESERVER;
		}
		
		if (cache.p[0]!='\0') {
			if (req->flags&DIR_CREATE) {
				if (!DIR_IS_DIR(cache.r_cur->options())) {
					return DIR_RES_ENOTDIR;
				}

				// recursively create the path.
				// acquire lock before calling traverse with DIR_CREATE flag
				if (pthread_mutex_lock(&lock)<0) {
					tmpi_error(DBG_INTERNAL, "Cannot acquire lock.\n");
					return DIR_RES_ESERVER;
				}

				if (refresh_cache_current(&cache)==false) {
					tmpi_error(DBG_INTERNAL, "Failed to refresh current cache.\n");
				}

				if (cache.r_cur->options()&DIR_OPTION_DEL) {
					pthread_mutex_unlock(&lock);
					return DIR_RES_ENONEXIST; /* client might try again. */
				}

				if (traverse(&cache, false, true, req->flags, &(req->param))<0) {
					pthread_mutex_unlock(&lock);
					return DIR_RES_ESERVER;
				}

				pthread_mutex_unlock(&lock);
			}
			else {
				return DIR_RES_ENONEXIST;
			}
		}
		else {
			if ((req->flags&DIR_CREATE)&&(req->flags&DIR_EXCLUSIVE)) {
				return DIR_RES_EEXIST;
			}
		}

		// copy cache.r_cur to res;
		DIRDB_REC *r_cur=cache.r_cur;
		res->nid=cache.n_cur;
		res->n_child=r_cur->child;
		res->n_parent=r_cur->parent;
		res->n_sibling=r_cur->sibling;
		memcpy(&(res->param), &(r_cur->param()), sizeof(DIR_Param));

		if (req->buf_len>0) {
			next_node=res->n_child;
		}
	}
	else {
# if defined(__DEBUG__)
		tmpi_error(DBG_INTERNAL, "DIR_DB::query:seach_next=%lu.\n", req->next_list);
# endif
		if (req->buf_len>0) {
			next_node=req->next_list;
		}
	}

	if ((next_node!=DIRDB_ERR)&&(req->buf_len>0)) {
		// starting from next_node, read till buffer full or 
		// sibling=DIRDB_NIL, skip DELeted record.
		// set res->next_list and the real buf_len;
		char *b=res->buf;
		char *b_end=res->buf+req->buf_len;
		char *p=res->buf;

		while (next_node!=DIRDB_NIL) {
			static const int min_item_sz=sizeof(DIR_Param)+sizeof(node_t)+2;
			int rem_len=b_end-p;
			if (rem_len<min_item_sz) {
				break;
			}
			unsigned int len;
			DIRDB_REC *node=(DIRDB_REC *)get_node(next_node, &len);

			if (!node)
				return DIR_RES_ESERVER;

			if (node->options()&DIR_OPTION_DEL) {
				next_node=node->sibling;
				free_node((Rec *)node);
				continue;
			}

			dir_cache.ins(node->parent, node->path(), next_node);

			int p_len=strlen(node->path())+1;
			int r_len=sizeof(DIR_Param)+sizeof(node_t)+p_len;

			if (rem_len<r_len) {
				break;
			}

			memcpy(p, &next_node, sizeof(node_t));
			memcpy(p+sizeof(node_t), &(node->param()), sizeof(DIR_Param));
			strcpy(p+sizeof(node_t)+sizeof(DIR_Param), node->path());
			p+=r_len;
			next_node=node->sibling;
			free_node((Rec *)node);
		}

		res->n_next=next_node;
		res->buf_len=p-b;
	}
	else {
		res->buf_len=1;
		res->buf[0]='\0';
	}

	return DIR_RES_OKAY;
}

int DIR_DB::deltree(node_t root, DIRDB_REC *root_rec, node_t base, bool & base_deleted)
{
	node_t list;
	if (root_rec!=NULL) {
		list=root_rec->child;
	}
	else {
		list=root;
	}

	int ret=0;

	while (list!=DIRDB_NIL) {
		unsigned int len;
		DIRDB_REC *node=(DIRDB_REC *)get_node(list, &len);

		if (!node) {
			return -1;
		}

		if (node->options()&DIR_OPTION_DEL) {
			// weird, we should not be here.
			tmpi_error(DBG_INTERNAL, "Unreachable code.\n");
		}
		else {
			node->options()|=DIR_OPTION_DEL;
			if (set_node(list, node, len)<0) {
				ret=-1;
			}

			dir_cache.del(node->parent, node->path());
			if (base==list)
				base_deleted=true;

			if (deltree(list, node, base, base_deleted)<0) {
				ret=-1;
			}
		}

		list=node->sibling;
		free_node((Rec *)node);
	}

	return ret;
}

int DIR_DB::del(DIR_REQ_DEL *req, DIR_RES_DEL *res)
{
	if (!is_loaded())
		return DIR_RES_ESERVER;

	char std_path[DIR_MAX_PATH_LEN];
	if (dir_std_path(NULL, req->path, std_path, DIR_MAX_PATH_LEN)) {
		return DIR_RES_EINVALID;
	}

	/**
	if ((std_path[0]=='.') && (std_path[1]=='\0')) {
		unsigned int len;
		DIRDB_REC *node=(DIRDB_REC *)get_node(req->base, &len);
		sprintf(std_path, "../%s", node->path());
	}
	**/

	Traverse_cache cache(req->base, req->b_parent, req->b_child, std_path);
	if (!cache.good()) {
		tmpi_error(DBG_APP, "Cannot initialize Traverse_cache (out of memory?).\n");
		return DIR_RES_ESERVER;
	}

	if (traverse(&cache, true, true)<0) {
		return DIR_RES_ESERVER;
	}
	
	if (cache.p[0]!='\0') {
		return DIR_RES_ENONEXIST;
	}

	if (pthread_mutex_lock(&lock)<0) {
		return DIR_RES_ESERVER;
	}

	int ret=DIR_RES_OKAY;
	DIRDB_REC *r_prev;
	DIRDB_REC *r_cur;

	if (!refresh_cache_current(&cache)) {
		ret=DIR_RES_ESERVER;
		goto _exit;
	}

	r_cur=cache.r_cur;
	if (r_cur->options()&DIR_OPTION_DEL) {
		ret=DIR_RES_ENONEXIST;
		goto _exit;
	}

	if ((!(req->flags&DIR_RECURSIVE))&&(r_cur->child!=DIRDB_NIL)) {
		// non-empty directory.
		ret=DIR_RES_EEXIST;
		goto _exit;
	}

	if (cache.n_cur==DIRDB_ROOT) {
		// special handling for root deletion
		// check if req->base is in the nodes that to be deleted.

		DIRDB_REC *r_cur=cache.r_cur;
		node_t list=r_cur->child;
		if (list!=DIRDB_NIL) {
			r_cur->child=DIRDB_NIL;
			if (set_node(DIRDB_ROOT, (Rec *)r_cur, cache.l_cur)<0) {
				ret=DIR_RES_ESERVER;
				goto _exit;
			}

			res->nid=DIRDB_ROOT;
			res->n_child=DIRDB_NIL;
			res->n_parent=DIRDB_ROOT;
			res->n_sibling=DIRDB_NIL;
			memcpy(&(res->param), &(r_cur->param()), sizeof(DIR_Param));
			
			bool base_deleted=false;

			if (deltree(list, NULL, req->base, base_deleted)<0) {
				ret|=DIR_RES_ESERVER;
				goto _exit;
			}

			if (base_deleted)
				ret=DIR_RES_ESELF;
		}
	}
	else {
		if (cache.n_prev!=DIRDB_ERR) {
			if (!refresh_cache_prev(&cache)) {
				ret=DIR_RES_ESERVER;
				goto _exit;
			}
		}

		r_prev=cache.r_prev;
		if (
			(cache.n_prev==DIRDB_ERR) || 
			(r_prev->options()&DIR_OPTION_DEL) || 
			((r_prev->sibling!=cache.n_cur)&&(r_prev->child!=cache.n_cur))  )
		{
			// cached value of prev is wrong.
			char p[DIR_MAX_NAME_LEN];
			strcpy(p, r_cur->path());
			Traverse_cache cache1(r_cur->parent, DIRDB_ERR, DIRDB_ERR, p);
	
			if (!cache1.good()) {
				ret=DIR_RES_ESERVER;
				goto _exit;
			}
	
			if (traverse(&cache1, false, false)<0) {
				ret=DIR_RES_ESERVER;
				goto _exit;
			}
	
			if (cache1.p[0]!='\0') { // very weird
				ret=DIR_RES_ENONEXIST;
				goto _exit;
			}
	
			cache.dup(&cache1);
		}

		if ((r_prev->options()&DIR_OPTION_DEL) || 
			((r_prev->sibling!=cache.n_cur)&&(r_prev->child!=cache.n_cur))  )
		{
			ret=DIR_RES_ESERVER;
			goto _exit;
		}
	
		// modify r_prev to skip cur and write to the disk.
		// mark the r_cur DEL, write to the disk
		// if recursive, make DELs for the children.
		
		if (r_prev->sibling==cache.n_cur) {
			r_prev->sibling=r_cur->sibling;
			if (set_node(cache.n_prev, r_prev, cache.l_prev)<0) {
				ret=DIR_RES_ESERVER;
				goto _exit;
			}
		}
		else {
			r_prev->child=r_cur->sibling;
			if (set_node(cache.n_prev, r_prev, cache.l_prev)<0) {
				ret=DIR_RES_ESERVER;
				goto _exit;
			}
		}

		res->nid=cache.n_prev;
		res->n_child=r_prev->child;
		res->n_parent=r_prev->parent;
		res->n_sibling=r_prev->sibling;
		memcpy(&(res->param), &(r_prev->param()), sizeof(DIR_Param));
	
		r_cur->options()|=DIR_OPTION_DEL;
		if (set_node(cache.n_cur, r_cur, cache.l_cur)<0) {
			ret|=DIR_RES_ESERVER;
		}

		dir_cache.del(r_cur->parent, r_cur->path());

		bool base_deleted=(cache.n_cur==req->base);

		if (deltree(cache.n_cur, r_cur, req->base, base_deleted)<0) {
			ret|=DIR_RES_ESERVER;
			goto _exit;
		}

		if (base_deleted)
			ret=DIR_RES_ESELF;
	}

_exit:
	pthread_mutex_unlock(&lock);
	return ret;
}


int DIR_DB::setp(DIR_REQ_SETP *req, DIR_RES_SETP *res)
{
	if (!is_loaded())
		return DIR_RES_ESERVER;

	char std_path[DIR_MAX_PATH_LEN];
	if (dir_std_path(NULL, req->path, std_path, DIR_MAX_PATH_LEN)) {
		return DIR_RES_EINVALID;
	}

	Traverse_cache cache(req->base, req->b_parent, req->b_child, std_path);
	if (!cache.good()) {
		tmpi_error(DBG_APP, "Cannot initialize Traverse_cache (out of memory?).\n");
		return DIR_RES_ESERVER;
	}

	if (traverse(&cache, true, true)<0) {
		return DIR_RES_ESERVER;
	}
	
	if (cache.p[0]!='\0') {
		return DIR_RES_ENONEXIST;
	}

	if (pthread_mutex_lock(&lock)<0) {
		return DIR_RES_ESERVER;
	}

	int ret=DIR_RES_OKAY;
	DIRDB_REC *node;
	if (!refresh_cache_current(&cache)) {
		ret=DIR_RES_ESERVER;
		goto _exit;
	}

	node=cache.r_cur;
	if (node->options()&DIR_OPTION_DEL) {
		ret=DIR_RES_ENONEXIST;
		goto _exit;
	}

	set_param(&(node->param()), &(req->param));
	if (set_node(cache.n_cur, (Rec *)node, cache.l_cur)) {
		ret=DIR_RES_ESERVER;
	}
	res->nid=cache.n_cur;
	res->n_child=node->child;
	res->n_parent=node->parent;
	res->n_sibling=node->sibling;
	memcpy(&(res->param), &(node->param()), sizeof(DIR_Param));

_exit:
	pthread_mutex_unlock(&lock);
	return ret;
}

void DIR_DB::dump_header()
{
	printf("   No      NODE     PARENT   CHILD   SIBLING  PATH\n");
}

void DIR_DB::dump_record(node_t id, DIRDB_REC *node)
{
	static int cn=0;

	if (node==NULL) {
		cn=0;
	}
	else {
		printf("%5d %8u %8u %8u %8u     %s", 
			   cn++, id, node->parent, node->child, node->sibling, node->path());
		if (node->options()&DIR_OPTION_DEL) {
			printf(" (deleted)");
		}
		printf("\n");
	}
}

void DIR_DB::dump_sequence()
{
	if (!is_loaded())
		return; 

	dump_header();
	dump_record(DIRDB_ERR, NULL);

	for (node_t i=DIRDB_ROOT; ; i++) {
		unsigned int len;
		if (i==DIRDB_ERR)
			continue;
		DIRDB_REC *node=(DIRDB_REC *)get_node(i, &len);
		if (!node)
			break;
		dump_record(i, node);
		free_node((Rec *)node);
	}
}

void DIR_DB::dump_depth()
{
	if (!is_loaded())
		return; 

	dump_header();
	dump_record(DIRDB_ERR, NULL);
	dump_depth0(DIRDB_ROOT);
}

DIR_DB::node_t DIR_DB::dump_depth0(node_t root)
{
	unsigned int len;
	DIRDB_REC *node=(DIRDB_REC *)get_node(root, &len);
	dump_record(root, node);
	node_t list=node->child;
	node_t next=node->sibling;
	free_node((Rec *)node);

	while (list!=DIRDB_NIL) {
		list=dump_depth0(list);
	}

	return next;
}

void DIR_DB::dump_breadth()
{
	if (!is_loaded())
		return; 
	dump_header();
	dump_record(DIRDB_ERR, NULL);

	queue<node_t> tqueue;
	tqueue.push(DIRDB_ROOT);
	while (!tqueue.empty()) {
		node_t nid=tqueue.front();
		tqueue.pop();
		unsigned int len;
		DIRDB_REC *node=(DIRDB_REC *)get_node(nid, &len);
		dump_record(nid, node);
		node_t list=node->child;
		free_node((Rec *)node);
		while (list!=DIRDB_NIL) {
			tqueue.push(list);
			node=(DIRDB_REC *)get_node(list, &len);
			list=node->sibling;
			free_node((Rec *)node);
		}
	}
}

void DIR_DB::dump_full()
{
	if (!is_loaded())
		return; 

	dump_full0(NULL, DIRDB_ROOT);
}

static int time_to_str(time_t tt, char *s, int l)
{
	if (!s)
		return -1;

	struct tm ttm;
	memcpy(&ttm, gmtime(&tt), sizeof(struct tm));
	if (strftime(s, l, "%b %e %T %Y", &ttm)==0)
		return -1;
	return 0;
}

void DIR_DB::dump_full0(char *prefix, node_t root)
{
	char npath[DIR_MAX_PATH_LEN];
	unsigned int len;
	DIRDB_REC *node=(DIRDB_REC *)get_node(root, &len);
	if (prefix) {
		sprintf(npath, "%s%s/", prefix, node->path());
	}
	else {
		strcpy(npath, node->path());
	}

	printf("%s:\n", npath);

	queue<node_t> tqueue;

	node_t list=node->child;
	free_node((Rec *)node);

	while (list!=DIRDB_NIL) {
		node=(DIRDB_REC *)get_node(list, &len);
		char ts[256];
		char guid_str[256];
		bool is_dir=DIR_IS_DIR(node->options());
		if (is_dir)
			tqueue.push(list);
		Guid guid(node->guid());
		guid.to_string(guid_str, 256);
		time_to_str(node->t_create(), ts, 256);
		printf("%crw%c  %10lld %20s %s [%s]\n", is_dir?'d':'-', is_dir?'x':'-', node->size(), ts, node->path(),  guid_str);
		list=node->sibling;
		free_node((Rec *)node);
	}
	printf("\n");

	while (!tqueue.empty()) {
		node_t nid=tqueue.front();
		tqueue.pop();
		dump_full0(npath, nid);
	}

	return;
}

void DIR_DB::dump_plain()
{
	if (!is_loaded())
		return; 

	printf("/\n");
	dump_plain0(NULL, DIRDB_ROOT);
}

void DIR_DB::dump_plain0(char *prefix, node_t root)
{
	char npath[DIR_MAX_PATH_LEN];
	unsigned int len;
	DIRDB_REC *node=(DIRDB_REC *)get_node(root, &len);
	if (prefix) {
		sprintf(npath, "%s%s/", prefix, node->path());
	}
	else {
		strcpy(npath, node->path());
	}

	queue<node_t> tqueue;

	node_t list=node->child;
	free_node((Rec *)node);

	while (list!=DIRDB_NIL) {
		tqueue.push(list);
		node=(DIRDB_REC *)get_node(list, &len);
		printf("%s%s\n", npath, node->path());
		list=node->sibling;
		free_node((Rec *)node);
	}

	while (!tqueue.empty()) {
		node_t nid=tqueue.front();
		tqueue.pop();
		dump_plain0(npath, nid);
	}

	return;
}

struct Tuple { 
	DIR_DB::node_t sid, did; 
	Tuple(DIR_DB::node_t a, DIR_DB::node_t b) : sid(a), did(b) {}
};

int DIR_DB::copy(char *dest, bool force)
{
	DIR_DB *old_db=this;

	unsigned int root_len;
	DIRDB_REC *root_node=NULL;
	
	root_node=(DIRDB_REC *)old_db->get_node(DIRDB_ROOT, &root_len);

	if (!root_node) {
		return -1;
	}

	if (force==false) {
		// check if the DB is valid.
		if (root_node->options()&DIR_OPTION_LOAD) {
			old_db->free_node((Rec *)root_node);
			tmpi_error(DBG_FATAL, "DB might be used by another process.\n");
			return -1;
		}
	}

	// set LOAD flag
	if (root_node->options()&DIR_OPTION_LOAD) {
		// nothing
	}
	else {
		root_node->options()|=DIR_OPTION_LOAD;
		if (old_db->set_node(DIRDB_ROOT, root_node, root_len)) {
			old_db->free_node((Rec *)root_node);
			return -1;
		}
		old_db->sync();
	}

	// clear SWEEP flags of non-root nodes.
	if (root_node->options()&DIR_OPTION_SWEEP) {
		for (node_t i=DIRDB_ERR+1; ; i++) {
			unsigned int len;
			DIRDB_REC *node=(DIRDB_REC *)old_db->get_node(i, &len);
			if (!node)
				break;

			if (node->options()&DIR_OPTION_SWEEP) {
				node->options()&=(~DIR_OPTION_SWEEP);
				if (old_db->set_node(i, node, len)) {
					old_db->free_node((Rec *)root_node);
					old_db->free_node((Rec *)node);
					return -1;
				}
			}
			free_node((Rec *)node);
		}

		// make sure that all SWEEP flags are cleared (on disk) before
		// we clear the SWEEP flag of the root.
		old_db->sync();

		root_node->options()&=(~DIR_OPTION_SWEEP);
		if (old_db->set_node(DIRDB_ROOT, root_node, root_len)) {
			old_db->free_node((Rec *)root_node);
			return -1;
		}
	}

	// DIRDB_ROOT: LOAD set. all nodes: SWEEP cleared.
	
	queue<Tuple> q;
	int ret=0;
	node_t list=DIRDB_NIL;
	node_t nextid=DIRDB_ERR+1;
	unsigned int len;
	DIRDB_REC *node=NULL;

	DIR_DB *new_db=new DIR_DB(dest);
	if (!new_db) {
		old_db->free_node((Rec *)root_node);
		ret=-1;
		goto _exit;
	}

	if (!new_db->good()) {
		old_db->free_node((Rec *)root_node);
		ret=-1;
		goto _exit;
	}
	
	// set root record to be the same as the old_db
	// except that flags are cleared.
	root_node->options() &=(~DIR_OPTION_LOAD);
	list=root_node->child;
	root_node->child=DIRDB_NIL;
	root_node->parent=DIRDB_ROOT;
	root_node->sibling=DIRDB_NIL;
	if (new_db->set_node(DIRDB_ROOT, root_node, root_len)) {
		old_db->free_node((Rec *)root_node);
		ret=-1;
		goto _exit;
	}

	old_db->free_node((Rec *)root_node);
	
	{
		// Tuple tp(DIRDB_ROOT, DIRDB_ROOT);

		// put the root into the queue.
		if (list!=DIRDB_NIL) {
			q.push(Tuple(DIRDB_ROOT, DIRDB_ROOT));
		}
	}

	while (!q.empty()) {
		Tuple tp=q.front();
		q.pop();
			
		// enqueue the children of the node tp.sid
			
		// validity check
		node=(DIRDB_REC *)old_db->get_node(tp.sid, &len);
		if (!node)
			continue;
			
		if ( (node->options()&DIR_OPTION_DEL) ||
			 (node->options()&DIR_OPTION_SWEEP)) {
			old_db->free_node((Rec *)node);
			continue;
		}
			
		// set sweep flag for it.
		node->options()|=DIR_OPTION_SWEEP;
		if (old_db->set_node(tp.sid, node, len)) {
			old_db->free_node((Rec *)node);
			ret=-1;
			goto _exit;
		}

		list=node->child;
		old_db->free_node((Rec *)node);

		bool first_child=true;

		while (list != DIRDB_NIL) {
			node=(DIRDB_REC *)old_db->get_node(list, &len);

			if (!node) {
				break;;
			}

			if (node->options()&DIR_OPTION_SWEEP) {
				old_db->free_node((Rec *)node);
				break;
			}

			if (node->options()&DIR_OPTION_DEL) {
				node->options()|=DIR_OPTION_SWEEP;
				if (old_db->set_node(list, node, len)) {
					old_db->free_node((Rec *)node);
					ret=-1;
					goto _exit;
				}
				list=node->sibling;
				old_db->free_node((Rec *)node);
				continue;
			}

			node_t this_id=nextid++;

			// first child, adjust parent 
			if (first_child==true) {
				unsigned int len_par;
				DIRDB_REC *node_par=(DIRDB_REC *)new_db->get_node(tp.did, &len_par);
				if (!node_par) {
					old_db->free_node((Rec *)node);
					ret=-1;
					goto _exit;
				}
				node_par->child=this_id;
				if (new_db->set_node(tp.did, node_par, len_par)) {
					old_db->free_node((Rec *)node);
					new_db->free_node((Rec *)node_par);
					ret=-1;
					goto _exit;
				}
				new_db->free_node((Rec *)node_par);
			}

			// enqueue this item 
			if (node->child != DIRDB_NIL) {
				Tuple tp2(list, this_id);
				q.push(tp2);
			}
			else {
				node->options()|=DIR_OPTION_SWEEP;
				if (old_db->set_node(list, node, len)) {
					old_db->free_node((Rec *)node);
					ret=-1;
					goto _exit;
				}
				node->options()&=(~DIR_OPTION_SWEEP);
			}

			list=node->sibling;

			node->parent=tp.did;
			node->child=DIRDB_NIL;
			node->sibling=DIRDB_NIL;

			if (new_db->set_node(this_id, node, len)) {
				old_db->free_node((Rec *)node);
				ret=-1;
				goto _exit;
			}

			old_db->free_node((Rec *)node);

			if (!first_child) {
				node=(DIRDB_REC *)new_db->get_node(this_id-1, &len);
				if (!node) {
					ret=-1;
					goto _exit;
				}
				node->sibling=this_id;
				if (new_db->set_node(this_id-1, node, len)) {
					new_db->free_node((Rec *)node);
					ret=-1;
					goto _exit;
				}
				new_db->free_node((Rec *)node);
			}
			else {
				first_child=false;
			}
		}
	}

 _exit:
	node=(DIRDB_REC *)old_db->get_node(DIRDB_ROOT, &len);
	node->options()&=~DIR_OPTION_LOAD;
	if (old_db->set_node(DIRDB_ROOT, node, len)) {
		ret=-1;
	}
	old_db->free_node((Rec *)node);
	old_db->sync();

	if (new_db)
		delete new_db;

	return ret;
}
