// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: dirdb.h,v $ - directory server db object
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/08/01 11:21:20 $
 * Version :  $Revision: 1.27 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 
# ifndef _DIRDB_H_
# define _DIRDB_H_

/* # include <pthread.h> */
# include <string>

# include "dir_common.h"
# include "dirdb_cache.h"
/* # include "rwlock.h" */

# if defined(__CHECK_MEM__)
# include "asmalloc.h"
# endif

class DIR_DB : public PSTREE {
private:
	std::string ident;
	pthread_mutex_t lock;
	DIRDB_Cache dir_cache;

	struct DIR_REC {
		DIR_Param param;
		char path[DIR_MAX_NAME_LEN];
	};
# define DIR_REC_HEADER_SZ ((int)&(((DIR_DB::DIR_REC *)0L)->path))

	struct DIRDB_REC : Rec {
		private:
			DIR_REC *dir_rec() { return (DIR_REC *)user; }
		public:
			// parent, child, and sibling will be accessed as usual.
			guid_t &guid() { return dir_rec()->param.guid; }
			unsigned int &options() { return dir_rec()->param.options; }
			off64_t &size() { return dir_rec()->param.size; }
			time_t &t_create() { return dir_rec()->param.t_create; }
			char * path() { return dir_rec()->path; }

			void init(node_t par=DIRDB_ERR, node_t chld=DIRDB_ERR);
			DIR_Param &param() { return dir_rec()->param; }
	};

	struct Traverse_cache { // double buffer
		/*---------------*/
		node_t n_prev;
		DIRDB_REC *r_prev;
		int l_prev;
		/*---------------*/
		node_t n_cur;
		DIRDB_REC *r_cur;
		int l_cur;
		/*---------------*/
		char *s;
		char *p;

	public:
		// constructor
		Traverse_cache(node_t base, node_t parent, node_t child, char *path);
		~Traverse_cache();
		void shift(node_t n, DIRDB_REC *next, int len); // traverse one step
		void refresh(DIRDB_REC *node, int len);
		void refresh_prev(DIRDB_REC *node, int len);
		bool good() { return ((r_prev)&&(r_cur)&&(s)); }
		void dup(Traverse_cache *cache);
		void release();
	};

	bool refresh_cache_current(Traverse_cache *cache);
	bool refresh_cache_prev(Traverse_cache *cache);
	bool check_cache_parent(Traverse_cache *cache);
	bool check_cache_child(Traverse_cache *cache);
	bool check_cache(DIR_DB::Traverse_cache *cache);
	int traverse(Traverse_cache *cache, bool use_dir_cache, \
			bool update_dir_cache, int flags=0, DIR_Param *param=NULL);
	int deltree(node_t root, DIRDB_REC *root_rec, node_t base, bool & base_deleted);
	int set_rec(DIR_REC *rec, guid_t *guid, char *path, int options, int *len, off64_t size);
	int set_param(DIR_Param *dest, DIR_Param *src);
	node_t dump_depth0(node_t root);
	void dump_full0(char *prefix, node_t root);
	void dump_plain0(char *prefix, node_t root);
	// typedef PSTREE::node_t node_t_tmp;

public:
	DIR_DB(char *db_file);
	bool good() { return is_loaded(); }
	~DIR_DB() {  pthread_mutex_destroy(&lock); }

	int query(DIR_REQ_QUERY *req, DIR_RES_QUERY *res);
	int del(DIR_REQ_DEL *req, DIR_RES_DEL *res);
	int setp(DIR_REQ_SETP *req, DIR_RES_SETP *res);
	bool sync() { return PSTREE::sync();}
	void cache_report() {dir_cache.report();}
	void dump_header();
	void dump_record(node_t nid, DIRDB_REC *rec);
	void dump_sequence();
	void dump_breadth();
	void dump_depth();
	void dump_full();
	void dump_plain();
	int copy(char *dest, bool force=false);
	const char *get_ident() { return ident.c_str(); }

# if defined(__DEBUG__)
	DIRDB_REC *get_node1(node_t n, unsigned int *len, char *f, int l);
# define get_node(n, pl) get_node1(n, pl, __FILE__, __LINE__)
# endif
};

# endif
