// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: dirdb_cache.cc,v $ - directory server db cache.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/05/06 09:19:22 $
 * Version :  $Revision: 1.5 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 
# include <stdlib.h>
# include <stdio.h>
# include <pthread.h>

# if defined(__USE_MD5__)
# include "md5.h"
# endif

# include "dirdb_cache.h"

DIRDB_Cache::DIRDB_Cache(bool enable) :
	cache(NULL),
	cache_on(enable),
	n_ins(0),
	n_del(0),
	n_miss(0),
	n_hit(0)
{
	if (cache_on) {
		cache=(Entry *)calloc(DIRDB_CACHE_SIZE, sizeof(Entry));
		pthread_mutex_init(&lock, NULL);
	}
}

DIRDB_Cache::~DIRDB_Cache()
{
	if (cache_on) {
		if (cache)
			free(cache);

		pthread_mutex_destroy(&lock);
	}
}

# define HIGH_BITS (~((unsigned int)(~0)>>4))

void DIRDB_Cache::pjw(node_t parent, char *path, unsigned int &sum)
{
	unsigned int h=parent;
	unsigned int g;

	for (; *path; ++path) {
		if ( (g = h & HIGH_BITS) )
			h = (h^(g>>24)) & ~HIGH_BITS;

		h = ( h << 4) + *path;
	}

	sum=h;
}

# if defined(__USE_MD5__)
void DIRDB_Cache::md5(node_t parent, char *path, unsigned char *digest)
{
	Md5 md5;
	md5.stuff((char *)&parent, sizeof(node_t));
	md5.stuff(path, strlen(path));
	md5.close();
	md5.pack(digest);
	/**
	unsigned int *ints=(unsigned int *)digest;
	sum=ints[0]^ints[1]^ints[2]^ints[3];
	**/
}
# endif

void DIRDB_Cache::ins(node_t parent, char *path, node_t node)
{
	if (cache_on) {

# if defined (__USE_MD5__)
		unsigned char digest[MD5_LEN];
# endif
		unsigned int sum;

		pjw(parent, path, sum);

# if defined(__USE_MD5__)
		md5(parent, path, digest);
# endif

		int idx=sum%DIRDB_CACHE_SIZE;

		pthread_mutex_lock(&lock);
		cache[idx].set(
# if defined(__USE_MD5__)
				digest, 
# endif
				sum, node);
		pthread_mutex_unlock(&lock);
		n_ins++;
	}
}

void DIRDB_Cache::del(node_t parent, char *path)
{
	if (cache_on) {
# if defined(__USE_MD5__)
		unsigned char digest[MD5_LEN];
# endif
		unsigned int sum;

		pjw(parent, path, sum);
# if defined(__USE_MD5__)
		md5(parent, path, digest);
# endif

		int idx=sum%DIRDB_CACHE_SIZE;

		pthread_mutex_lock(&lock);
		if (cache[idx].sum!=sum)
			goto _exit;

# if defined(__USE_MD5__)
		if (memcmp(cache[idx].tag, digest, MD5_LEN))
			goto _exit;
# endif

		cache[idx].clear();

_exit:
		pthread_mutex_unlock(&lock);
		n_del++;
	}
}

int DIRDB_Cache::find(node_t parent, char *path, node_t &node)
{
	if (cache_on) {
		unsigned int sum;
		pjw(parent, path, sum);
		int idx=sum%DIRDB_CACHE_SIZE;

		if (cache[idx].sum!=sum) {
			n_miss++;
			return -1;
		}

# if defined(__USE_MD5__)
		unsigned char digest[MD5_LEN];
		md5(parent, path, digest);

		if (memcmp(cache[idx].tag, digest, MD5_LEN)) {
			n_miss++;
			return -1;
		}
# endif

		node=cache[idx].inode;
		n_hit++;

		return 0;
	}

	return -1;
}

void DIRDB_Cache::report()
{
	if (cache_on)
		printf("INS=%d, DEL=%d, HIT=%d, MISS=%d\n", n_ins, n_del, n_hit, n_miss);
}
