// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-
/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: t_dirdbtr.cc,v $ - directory server trace-driven tester.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/05/06 09:19:22 $
 * Version :  $Revision: 1.4 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 
# include <stdio.h>
# include <stdlib.h>
# include <ctype.h>
# include <unistd.h>

extern "C" {
# include <readline/readline.h>
# include <readline/history.h>
}

# include "dir_cli_simple.h"
# include "dirdb.h"
# include "dir_cli_opt.h"
# include "helper.h"

# if defined(__DEBUG__)
# include "asmalloc.h"
# endif

static char *program_name=NULL;

static char *chop(char *line)
{
	while (*line) {
		if (isspace(*line))
			line++;
		else
			break;
	}

	char *p=line+strlen(line)-1;
	while ((p>=line)&&(isspace(*p))) {
		*p='\0';
		p--;
	}

	return line;
}

static char *get_RCS_val(char *var)
{
	static char buf[1024];
	char *p;

	if ((p=strchr(var, ':'))==NULL) {
		return var;
	}
	buf[1023]='\0';
	strncpy(buf, p+1, 1024);
	p=buf+strlen(buf)-1;
	while ((p>=buf)&&(*p=='$')) {
		*p='\0';
		p--;
	}

	return chop(buf);
}

enum CMDS {
	CMD_CD,
	CMD_LS,
	CMD_PWD,
	CMD_MD,
	CMD_RM,
	CMD_HELP,
	CMD_QUIT,
	CMD_SYNC
};

static void do_help(char *cmd)
{
	if (cmd) {
		printf("Unknow command: %s. Type help for more info.\n", cmd);
	}
	else {
		printf("\n\
The shell suports the following commands:\n\
    cd <path>       Change the current directory.\n\
    ls <path>       List the attribute and content of a node.\n\
    pwd             Show the current working directory.\n\
    md <path>       Create a node.\n\
    rm [-r] <path>  Remove a node.\n\
    sync            Synchronize local settings with the server.\n\
    help            Print this message.\n\
    quit            Quit the console.\n\
");
	}

	return;
}

static int handle_cd(DIR *dir_cli, char *path, unsigned int flags)
{
	if (!path) {
		printf("No path specified.\n");
		return 0;
	}

	int ret=dir_cli->change(path);
	switch (ret) {
		case DIR_RES_OKAY: 
			return 0;
		case DIR_RES_ENONEXIST: 
			printf("Node does not exist.\n"); return 0;
		case DIR_RES_EINVALID: 
			printf("Bad argument.\n"); return 0;
		case DIR_RES_EBUFFER: 
		case DIR_RES_EMEM: 
			tmpi_error(DBG_INTERNAL, "Out of memory.\n"); return 0;
		default:
			tmpi_error(DBG_INTERNAL, "Internal error. Code: %d.\n", ret);
			return 1;
	}
}

static int time_to_str(time_t tt, char *s, int l)
{
	if (!s)
		return -1;

    struct tm ttm;

    memcpy(&ttm, gmtime(&tt), sizeof(struct tm));
    if (strftime(s, l, "%b %e %T %Y", &ttm)==0)
		return -1;

    return 0;
}

static int handle_md(DIR *dir_cli, char *path, unsigned int flags)
{
	if (!path) {
		printf("No path specified.\n");
		return 0;
	}

	DIR_Param param;
	int ret=dir_cli->list(path, &param, NULL, NULL, DIR_CREATE|DIR_EXCLUSIVE);
	switch (ret) {
		case DIR_RES_OKAY: 
			{
				char guid_str[256];
				Guid guid(param.guid);
				guid.to_string(guid_str, 256);
				char ts[256];
				time_to_str(param.t_create, ts, 256);
				printf("Node created, GUID=%s (%s).\n", guid_str, ts);
				return 0;
			}
		case DIR_RES_EEXIST: 
			printf("Node already exists.\n"); return 0;
		case DIR_RES_ENONEXIST: 
			printf("Unable to create node at this point, try again.\n"); return 0;
		case DIR_RES_EINVALID: 
			printf("Bad argument.\n"); return 0;
		case DIR_RES_EBUFFER: 
		case DIR_RES_EMEM: 
			tmpi_error(DBG_INTERNAL, "Out of memory.\n"); return 0;
		default:
			tmpi_error(DBG_INTERNAL, "Internal error. Code: %d.\n", ret);
			return 1;
	}
}

static int handle_rm(DIR *dir_cli, char *path, unsigned int flags)
{
	if (!path) {
		printf("No path specified.\n");
		return 0;
	}
	
	int ret=dir_cli->del(path, flags);
	switch (ret) {
		case DIR_RES_OKAY: 
			return 0;
		case DIR_RES_EEXIST: 
			printf("Directory not empty.\n"); return 0;
		case DIR_RES_ENONEXIST: 
			printf("Node does not exist.\n"); return 0;
		case DIR_RES_EINVALID: 
			printf("Bad argument.\n"); return 0;
		case DIR_RES_ESELF: 
			printf("Current directory no longer exists.\n"); return 0;
		case DIR_RES_EBUFFER: 
		case DIR_RES_EMEM: 
			tmpi_error(DBG_INTERNAL, "Out of memory.\n"); return 0;
		default:
			tmpi_error(DBG_INTERNAL, "Internal error. Code: %d.\n", ret);
			return 1;
	}
}

static int handle_ls(DIR *dir_cli, char *path, unsigned int flags)
{
	if (!path) {
		path=".";
	}

	int ret=0;
	bool first_time=true;
	const int buf_len=60000;
	char buf[buf_len];
	int i=1;
	DIR_Param param;

	while (1) {
		int len=buf_len;
		if (first_time) {
			ret=dir_cli->list(path, &param, buf, &len, 0);
		}
		else {
			ret=dir_cli->list(NULL, NULL, buf, &len, 0);
		}

		if ((ret!=DIR_RES_OKAY)&&(ret!=DIR_RES_EEND)) {
			switch (ret) {
				case DIR_RES_ENONEXIST: 
					printf("Node does not exist.\n"); return 0;
				case DIR_RES_EINVALID: 
					printf("Bad argument.\n"); return 0;
				case DIR_RES_EBUFFER: 
				case DIR_RES_EMEM: 
					tmpi_error(DBG_INTERNAL, "Out of memory.\n"); return 0;
				default:
					tmpi_error(DBG_INTERNAL, "Internal error. Code: %d.\n", ret);
					return 1;
			}
		}

		if (first_time) {
			first_time=false;
			char guid_str[256];
			Guid guid(param.guid);
			guid.to_string(guid_str, 256);
			char ts[256];
			time_to_str(param.t_create, ts, 256);
			printf("List directory %s /%s/ (%s):\n", path, ts, guid_str);
		}

		if (ret==DIR_RES_EEND)
			break;

		char *p=NULL;
		int k=0;
		while (1) {
			char *name;
			DIR_Param gg;
			ret=dir_cli->read_list(buf, len, &p, &name, &gg);
			if (ret!=DIR_RES_EEND) {
				char guid_str[256];
				Guid guid(gg.guid);
				guid.to_string(guid_str, 256);
				char ts[256];
				time_to_str(gg.t_create, ts, 256);
				printf("[%d] %s /%s/ (%s)\n", i++, name, ts, guid_str);
				k++;
			}
			else
				break;
		}

		if (k==0)
			break;
 	}

	return DIR_RES_OKAY;
}

static int handle_pwd(DIR *dir_cli, char *path, unsigned int flags)
{
	char *p=dir_cli->cwd();
	if (!p) {
		printf("Cannot determine current directory.\n");
	}
	else {
		printf("%s\n", p);
	}

	return 0;
}

static int handle_sync(DIR *dir_cli, char *path, unsigned int flags)
{
	dir_cli->sync();

	return 0;
}

struct CMD_ARRAY {
	char *cmd;
	int id;
	int (*f)(DIR *, char *, unsigned int);
} cmd_array[] = {
	{ "cd", CMD_CD, &handle_cd},
	{ "ls", CMD_LS, &handle_ls},
	{ "pwd", CMD_PWD, &handle_pwd},
	{ "md", CMD_MD, &handle_md},
	{ "rm", CMD_RM, &handle_rm},
	{ "sync", CMD_SYNC, &handle_sync},
	{ "help", CMD_HELP, NULL},
	{ "quit", CMD_QUIT, NULL},
	{ NULL, -1, NULL}
};

static int handle_cmd(DIR *dir_cli, char *cmd, char *path, unsigned int flags)
{
	if (!cmd)
		return 0;

	struct CMD_ARRAY *p=cmd_array;

	while (p->cmd) {
		if (strcmp(cmd, p->cmd)==0) {
			break;
		}
		p++;
	}

	if (p->cmd) {
		if (p->f) {
			return (*p->f)(dir_cli, path, flags);
		}

		switch (p->id) {
			case CMD_HELP : 
				do_help(NULL); return 0;
			case CMD_QUIT : 
				printf("Good bye!\n"); return 1;
			default : 
				tmpi_error(DBG_APP, "Internal Error: Bad Command Code.\n"); 
				return 0;
		}
	}
	else {
		do_help(cmd);
	}

	return 0;
}

char *Id="$Id: t_dirdbtr.cc,v 1.4 2003/05/06 09:19:22 htang Exp $";

static void do_version()
{
	static char *version="$Revision: 1.4 $";
	static char *author="Hong Tang";
	static char *date="$Date: 2003/05/06 09:19:22 $";

	{
		printf("Directory Server/DB Client (%s) %s.\n", program_name, get_RCS_val(version));
		printf("Copyright (c) University of California, Santa Barbara.\n");
		printf("Author: %s.\n", get_RCS_val(author));
		printf("Last Modification Date: %s GMT.\n", get_RCS_val(date));
		printf("Specify the --help option for further information.\n");
	}

	exit(0);
}

static unsigned int make_flags(bool creat, bool excl, bool recur)
{
	unsigned int flags=0;

	if (creat)
		flags|=DIR_CREATE;

	if (excl)
		flags|=DIR_EXCLUSIVE;

	if (recur)
		flags|=DIR_RECURSIVE;

	return flags;
}

static bool is_empty_line(char *line)
{
	while (*line) {
		if (!isspace(*line))
			return false;
		line++;
	}

	return true;
}

static char *get_token(char * &t, char * & t_end)
{
	if (!t)
		return NULL;

	while ((*t)&&(isspace(*t))) {
		t++;
	}

	t_end=t;
	while ((*t_end)&&(!isspace(*t_end))) {
		t_end++;
	}
	
	if (t==t_end)
		return NULL;

	return t;
}

static	DIR_Cli_opt opt;

# if (__DEBUG__)
static char *b2s(bool b)
{
	if (b)
		return "true";
	else 
		return "false";
}
# endif

static int	parse_line(char *line, char *&cmd, char *&path, unsigned int &flags)
{
	if (!line)
		return -1;
	
	cmd=NULL;
	path=NULL;

	line=chop(line);

	if (is_empty_line(line)) {
		return 0;
	}

	char *cmd_begin;
	char *cmd_end;
	char *line_end=line+strlen(line);

	cmd_begin=line;
	cmd_begin=get_token(cmd_begin, cmd_end);
	
	if (cmd_begin) { /* check for arguments */
		*cmd_end='\0';
		char *p=cmd_end+1;
		char *p_end;
		bool creat=opt.creat;
		bool excl=opt.excl;
		bool recur=opt.recur;
		while ((p<line_end)&&(get_token(p, p_end))) {
			*p_end='\0';
			if (*p=='-') { /* options */
				p++;
				while (*p) {
					switch (*p) {
					case 'c' : creat=true; break;
					case 'C' : creat=false; break;
					case 'e' : excl=true; break;
					case 'E' : excl=false; break;
					case 'r' : recur=true; break;
					case 'R' : recur=false; break;
					default : return -1; /* bad options */
					}
					p++;
				}
			}
			else { /* take it as path */
				if (path)
					return -1;
				path=p;
			}

			p=p_end+1;
		}
		flags=make_flags(creat, excl, recur);
		cmd=cmd_begin;
# if (__DEBUG__)
		// tmpi_error(DBG_VERBOSE, "CMD=%s, creat=%s, excl=%s, recur=%s, path=%s\n",
		//	   cmd_begin, b2s(creat), b2s(excl), b2s(recur), (path)?path:"<nul>");
# endif
	}

	return 0;
}

int main(int argc, char *argv[])
{
	program_name=argv[0];
# if defined(__DEBUG__)
	AS_InitMem();
# endif

# if defined(__DEBUG__)
	tmpi_set_dbg_level(DBG_ALL);
	tmpi_show_location();
# else
	tmpi_set_dbg_level(DBG_VERBOSE);
	tmpi_hide_location();
# endif
	

	// might be no return.
	parse_dir_cli_opt(argc, argv, &opt, &do_version); 

	if (strlen(opt.bsf)<=0) {
		tmpi_error(DBG_CALLER, "NO DB file specified. This version only supports file-based dir-db.\n");
		return -1;
	}

	DIR_DB db(opt.bsf);
	if (!db.good()) {
		tmpi_error(DBG_APP, "Cannot initialize DIR DB file <%s>.\n", opt.bsf);
		return -1;
	}

	dump_dir_cli_opt(&opt);

	char *init_dir=(strlen(opt.idir)>0)?opt.idir:NULL;
	unsigned int flags=make_flags(opt.creat, opt.excl, opt.recur);
	DIR dir_cli(&db, init_dir, flags);

	char *cmd; 
	char *path;
	char pwd[100]="pwd";
	parse_line(pwd, cmd, path, flags);
	handle_cmd(&dir_cli, cmd, path, flags);

	long count=0;

	while (1) {
		char line[1024];
		fgets(line, 1024, stdin);
		count ++;

		if (line&&(line[0]))
			add_history(line);

		if (parse_line(line, cmd, path, flags)<0) {
			printf("Bad command.\n");
			free_line(line);
			continue;
		}

		if (handle_cmd(&dir_cli, cmd, path, flags))  {
			free_line(line);
			break;
		}
		// db.dump();
	}

	return 0;
}
