// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/*
    FUSE: Filesystem in Userspace
    Copyright (C) 2001  Miklos Szeredi (mszeredi@inf.bme.hu)

    This program can be distributed under the terms of the GNU GPL.
    See the file COPYING.
*/

#include "fuse_i.h"
#include "sor_proc.h"

#include <linux/poll.h>
#include <linux/proc_fs.h>
#include <linux/file.h>
#include <linux/vmalloc.h>


/* If more requests are outstanding, then the operation will block */
#define MAX_OUTSTANDING 10


/* ---------------------------------------------------------------
 * Code adopted from PVFS
 * --------------------------------------------------------------- */
/** Wait queue used to allow process to sleep if not enough buffer */
static DECLARE_WAIT_QUEUE_HEAD(sordev_buffer_pool_wait);

/* Maximum pool size 64MB */
#define SOR_BUFFER_POOL_SIZE 64*1024*1024

/** buffer pool lock */
static spinlock_t sordev_buffer_pool_lock = SPIN_LOCK_UNLOCKED;

#define LOCK_BUFFER_POOL    spin_lock(&sordev_buffer_pool_lock)
#define UNLOCK_BUFFER_POOL  spin_unlock(&sordev_buffer_pool_lock)

/** current buffer pool size */
static unsigned long sordev_buffer_size = 0;

/** 
 * Check for pending signals after a process has woken up.
 * @return 0 is everything is ok, -EINTR otherwise
 */
static inline
int check_bad_signals(void)
{
	if (signal_pending(current)) {
		if (sigismember(&(current->pending.signal), SIGKILL) ||
			sigismember(&(current->pending.signal), SIGINT) ) {
			return -EINTR;
		}
	}
	return 0;
}

/*
 * The PVFS's BUFMAP technique is not ported here because it is
 * not working with high memory systems. For high memory (>1G),
 * the memory region has to be mapped to normal ZONE so that
 * kernel can direct address it. The BUFMAP technique pins client
 * memory and copy buffer from daemon to that memory directly.
 */


/** 
 * Allocate a kernel buffer 
 * @param copybuf On success holds address of allocated buffer
 * @param size Requested buffer size
 */
int 
get_pool_buffer(void **copybuf, unsigned long size)
{
	struct task_struct *tsk = current;
	DECLARE_WAITQUEUE(my_wait, current);

	LOCK_BUFFER_POOL;
	if ( (sordev_buffer_size + size) <= SOR_BUFFER_POOL_SIZE) {
		sordev_buffer_size += size;
		UNLOCK_BUFFER_POOL;
	} else {
		UNLOCK_BUFFER_POOL;
		my_wait.task = tsk;
		add_wait_queue(&sordev_buffer_pool_wait, &my_wait);
	REPEAT:
		set_current_state(TASK_INTERRUPTIBLE);

		/* check for signals */
		if (check_bad_signals()) {
			printk("sordev: received signal and bail out\n");
			remove_wait_queue(&sordev_buffer_pool_wait, &my_wait);
			set_current_state(TASK_RUNNING);
			return -EINTR;
		}

		/* try again */
		LOCK_BUFFER_POOL;
		if ( (sordev_buffer_size + size) > SOR_BUFFER_POOL_SIZE) {
			UNLOCK_BUFFER_POOL;
			schedule();
			goto REPEAT;
		}

		/* memory is available now */
		sordev_buffer_size += size;
		UNLOCK_BUFFER_POOL;

		remove_wait_queue(&sordev_buffer_pool_wait, &my_wait);
		set_current_state(TASK_RUNNING);
	}

	/* now allocate the buffer */
	*copybuf = (void*) vmalloc(size);
	if (*copybuf == NULL) {
		printk("sordev: could not allocate memory\n");
		LOCK_BUFFER_POOL;
		sordev_buffer_size -= size;
		UNLOCK_BUFFER_POOL;
		return -ENOMEM;
	}

	return 0;
}

/** Free dynamic buffer and wake up sleeping process */
void
free_pool_buffer(void **buf, unsigned long size)
{
	vfree(*buf);
	*buf = NULL;

	LOCK_BUFFER_POOL;
	sordev_buffer_size -= size;
	UNLOCK_BUFFER_POOL;

	wake_up_interruptible(&sordev_buffer_pool_wait);
}
/* -------------------End of adopted code -----------------------*/





struct proc_dir_entry *proc_fuse_dev;
static kmem_cache_t *fuse_req_cachep;

static struct fuse_req *request_new(void)
{
	struct fuse_req *req;

	req = (struct fuse_req *) kmem_cache_alloc(fuse_req_cachep, SLAB_NOFS);
	if(req) {
		INIT_LIST_HEAD(&req->list);
		req->issync = 0;
		req->locked = 0;
		req->interrupted = 0;
		req->sent = 0;
		req->finished = 0;
		req->in = NULL;
		req->out = NULL;
		init_waitqueue_head(&req->waitq);
	}

	return req;
}

static void request_free(struct fuse_req *req)
{
	kmem_cache_free(fuse_req_cachep, req);
}

static int request_restartable(enum fuse_opcode opcode)
{
	switch(opcode) {
	case FUSE_LOOKUP:
	case FUSE_GETATTR:
	case FUSE_READLINK:
	case FUSE_GETDIR:
	case FUSE_OPEN:
	case FUSE_READ:
	case FUSE_WRITE:
		return 1;

	default:
		return 0;
	}
}

/* Called with fuse_lock held.  Releases, and then reaquires it. */
static void request_wait_answer(struct fuse_req *req)
{
	int intr;
	
	spin_unlock(&fuse_lock);
	intr = wait_event_interruptible(req->waitq, req->finished);
	spin_lock(&fuse_lock);
	if(!intr)
		return;

	/* Request interrupted... Wait for it to be unlocked */
	if(req->locked) {
		req->interrupted = 1;
		spin_unlock(&fuse_lock);
		wait_event(req->waitq, !req->locked);
		spin_lock(&fuse_lock);
	}
	
	/* Operations which modify the filesystem cannot safely be
	   restarted, because it is uncertain whether the operation has
	   completed or not... */
	if(req->sent && !request_restartable(req->in->h.opcode))
		req->out->h.error = -EINTR;
	else
		req->out->h.error = -ERESTARTSYS;
}

static int get_unique(struct fuse_conn *fc)
{
	do fc->reqctr++;
	while(!fc->reqctr);
	return fc->reqctr;
}

/* Must be called with fuse_lock held, and unlocks it */
static void request_end(struct fuse_conn *fc, struct fuse_req *req)
{
	fuse_reqend_t endfunc = req->end;

	if(!endfunc) {
		wake_up(&req->waitq);
		spin_unlock(&fuse_lock);
	} else {
		spin_unlock(&fuse_lock);
		endfunc(fc, req->in, req->out, req->data);
		request_free(req);
		up(&fc->outstanding);
	}
}

/** In client's context, wait until the results are ready */
void 
request_send(struct fuse_conn *fc, 
			 struct fuse_in *in,
			 struct fuse_out *out)
{
	struct fuse_req *req;

	out->h.error = -ERESTARTSYS;
	if(down_interruptible(&fc->outstanding))
		return;

	out->h.error = -ENOMEM;
	req = request_new();
	if(req) {
		req->in = in;
		req->out = out;
		req->issync = 1;
		req->end = NULL;
		
		spin_lock(&fuse_lock);
		out->h.error = -ENOTCONN;
		if(fc->file) {
			in->h.unique = get_unique(fc);		
			list_add_tail(&req->list, &fc->pending);
			wake_up(&fc->waitq);
			request_wait_answer(req);
			list_del(&req->list);
		}
		spin_unlock(&fuse_lock);
		request_free(req);
	}

	up(&fc->outstanding);
}


static inline void destroy_request(struct fuse_req *req)
{
	if(req) {
		kfree(req->in);
		request_free(req);
	}
}

/* This one is currently only used for sending FORGET and RELEASE,
   which are kernel initiated request.  So the outstanding semaphore
   is not used.  */
int request_send_noreply(struct fuse_conn *fc, struct fuse_in *in)
{
	struct fuse_req *req;

	req = request_new();
	if(!req)
		return -ENOMEM;

	req->in = in;
	req->issync = 0;

	spin_lock(&fuse_lock);
	if(!fc->file) {
		spin_unlock(&fuse_lock);
		request_free(req);
		return -ENOTCONN;
	}

	list_add_tail(&req->list, &fc->pending);
	wake_up(&fc->waitq);
	spin_unlock(&fuse_lock);
	return 0;
}

int request_send_nonblock(struct fuse_conn *fc, struct fuse_in *in,
			  struct fuse_out *out, fuse_reqend_t end, void *data)
{
	int err;
	struct fuse_req *req;

	//BUG_ON(!end);

	if(down_trylock(&fc->outstanding))
		return -EWOULDBLOCK;

	err = -ENOMEM;
	req = request_new();
	if(req) {
		req->in = in;
		req->out = out;
		req->issync = 1;
		req->end = end;
		req->data = data;

		spin_lock(&fuse_lock);
		err = -ENOTCONN;
		if(fc->file) {
			in->h.unique = get_unique(fc);		
			list_add_tail(&req->list, &fc->pending);
			wake_up(&fc->waitq);
			spin_unlock(&fuse_lock);
			return 0;
		}
		spin_unlock(&fuse_lock);
		request_free(req);
	}
	up(&fc->outstanding);
	return err;
}

/** User-land daemon wait here until there is request pending */
static void 
request_wait(struct fuse_conn *fc)
{
	DECLARE_WAITQUEUE(wait, current);

	add_wait_queue_exclusive(&fc->waitq, &wait);
	while(fc->sb != NULL && list_empty(&fc->pending)) {
		set_current_state(TASK_INTERRUPTIBLE);
		if(signal_pending(current))
			break;

		spin_unlock(&fuse_lock);
		schedule();
		spin_lock(&fuse_lock);
	}
	set_current_state(TASK_RUNNING);
	remove_wait_queue(&fc->waitq, &wait);
}

/** Copy an argument to user space */
static inline int 
copy_arg_to_user(const void *src, 
				 size_t srclen, 
				 char **dstp, 
				 size_t *dstlenp)
{
	if(*dstlenp < srclen) {
		printk("fuse_dev_read: buffer too small\n");
		return -EINVAL;
	}
			
	if(copy_to_user(*dstp, src, srclen))
		return -EFAULT;

	*dstp += srclen;
	*dstlenp -= srclen;

	return 0;
}

/** Copy request to user space */
static inline int 
copy_req_to_user(struct fuse_in *in, char *buf, size_t nbytes)
{
	int err;
	int i;
	size_t orignbytes = nbytes;
		
	err = copy_arg_to_user(&in->h, sizeof(in->h), &buf, &nbytes);
	if(err)
		return err;

	for(i = 0; i < in->numargs; i++) {
		struct fuse_in_arg *arg = &in->args[i];
		err = copy_arg_to_user(arg->value, arg->size, &buf, &nbytes);
		if(err)
			return err;
	}

	return orignbytes - nbytes;
}

/** Called when user daemon read to the proc file */
static ssize_t 
fuse_dev_read(struct file *file, 
			  char *buf, 
			  size_t nbytes,
			  loff_t *off)
{
	ssize_t ret;
	struct fuse_conn *fc = DEV_FC(file);
	struct fuse_req *req = NULL;

	spin_lock(&fuse_lock);
	request_wait(fc);
	if(fc->sb != NULL && !list_empty(&fc->pending)) {
		req = list_entry(fc->pending.next, struct fuse_req, list);
		list_del_init(&req->list);
		req->locked = 1;
	}
	spin_unlock(&fuse_lock);
	if(fc->sb == NULL)
		return -ENODEV;
	if(req == NULL)
		return -EINTR;

	ret = copy_req_to_user(req->in, buf, nbytes);
	spin_lock(&fuse_lock);
	if(req->issync) {
		if(ret < 0) {
			req->out->h.error = -EPROTO;
			req->finished = 1;
		} else {
			list_add_tail(&req->list, &fc->processing);
			req->sent = 1;
		}
		req->locked = 0;
		if(ret < 0 || req->interrupted)
			/* Unlocks fuse_lock: */
			request_end(fc, req);
		else
			spin_unlock(&fuse_lock);
	} else {
		spin_unlock(&fuse_lock);
		destroy_request(req);
	}
	return ret;
}

static struct fuse_req *request_find(struct fuse_conn *fc, unsigned int unique)
{
	struct list_head *entry;
	struct fuse_req *req = NULL;

	list_for_each(entry, &fc->processing) {
		struct fuse_req *tmp;
		tmp = list_entry(entry, struct fuse_req, list);
		if(tmp->in->h.unique == unique) {
			req = tmp;
			break;
		}
	}

	return req;
}

static void process_getdir(struct fuse_req *req)
{
	struct fuse_getdir_out *arg;
	arg = (struct fuse_getdir_out *) req->out->args[0].value;
	arg->file = fget(arg->fd);
}

static inline 
int copy_one_downcall_arg(struct fuse_out_arg *arg, const char **srcp,
						  size_t *srclenp, int allowvar)
{
	size_t dstlen = arg->size;
	if(*srclenp < dstlen) {
		if(!allowvar) {
			printk("fuse_dev_write: write is short\n");
			return -EINVAL;
		}
		dstlen = *srclenp;
	}

	if(dstlen) {
		if(copy_from_user(arg->value, *srcp, dstlen))
			return -EFAULT;
	}

	*srcp += dstlen;
	*srclenp -= dstlen;
	arg->size = dstlen;

	return 0;
}

/** Copy downcall arguments from file server */
static inline int 
copy_downcall_args(struct fuse_out *out,
				   const char *buf,
				   size_t nbytes)
{
	int err;
	int i;

	buf += sizeof(struct fuse_out_header);
	nbytes -= sizeof(struct fuse_out_header);
		
	if(!out->h.error) {
		for(i = 0; i < out->numargs; i++) {
			struct fuse_out_arg *arg = &out->args[i];
			int allowvar;

			if(out->argvar && i == out->numargs - 1)
				allowvar = 1;
			else
				allowvar = 0;

			err = copy_one_downcall_arg(arg, &buf, &nbytes, allowvar);
			if(err)
				return err;
		}
	}

	if(nbytes != 0) {
		printk("fuse_dev_write: write is long\n");
		return -EINVAL;
	}

	return 0;
}

/** Copy downcall header from file server */
static inline int 
copy_downcall_header(struct fuse_out_header *oh, 
			  const char *buf,
			  size_t nbytes)
{
	if(nbytes < sizeof(struct fuse_out_header)) {
		printk("fuse_dev_write: write is short\n");
		return -EINVAL;
	}
	
	if(copy_from_user(oh, buf, sizeof(struct fuse_out_header)))
		return -EFAULT;

	return 0;
}

#ifdef KERNEL_2_6
static int fuse_invalidate(struct fuse_conn *fc, struct fuse_user_header *uh)
{
	struct inode *inode = ilookup(fc->sb, uh->ino);
	if (!inode)
		return -ENOENT;
	invalidate_inode_pages(inode->i_mapping);
	iput(inode);
	return 0;
}
#else 
static int fuse_invalidate(struct fuse_conn *fc, struct fuse_user_header *uh)
{
	struct inode *inode = iget(fc->sb, uh->ino);
	int err = -ENOENT;
	if(inode) {
		if(inode->u.generic_ip) {
			invalidate_inode_pages(inode);
			err = 0;
		}
		iput(inode);
	}
	return err;
}
#endif

static int fuse_user_request(struct fuse_conn *fc, const char *buf,
			     size_t nbytes)
{
	struct fuse_user_header uh;
	int err;

	if (nbytes < sizeof(struct fuse_user_header)) {
		printk("fuse_dev_write: write is short\n");
		return -EINVAL;
	}

	if(copy_from_user(&uh, buf, sizeof(struct fuse_out_header)))
		return -EFAULT;
	
	switch(uh.opcode) {
	case FUSE_INVALIDATE:
		err = fuse_invalidate(fc, &uh);
		break;

	default:
		err = -ENOSYS;
	}
	return err;
}
    

/** Called when daemon write to the proc file */
static ssize_t 
fuse_dev_write(struct file *file, const char *buf,
			   size_t nbytes, loff_t *off)
{
	int err;
	struct fuse_conn *fc = DEV_FC(file);
	struct fuse_req *req;
	struct fuse_out_header oh;

	if(!fc->sb)
		return -EPERM;

	err = copy_downcall_header(&oh, buf, nbytes);
	if(err)
		return err;

	if (!oh.unique)	{
		err = fuse_user_request(fc, buf, nbytes);
		goto out;
	}     

	if (oh.error <= -512 || oh.error > 0) {
			printk("fuse_dev_write: bad error value\n");
			return -EINVAL;
	}

	spin_lock(&fuse_lock);
	req = request_find(fc, oh.unique);
	if(req != NULL) {
		list_del_init(&req->list);
		req->locked = 1;
	}
	spin_unlock(&fuse_lock);
	if(!req)
		return -ENOENT;

	req->out->h = oh;
	err = copy_downcall_args(req->out, buf, nbytes);

	spin_lock(&fuse_lock);
	if(err)
		req->out->h.error = -EPROTO;
	else {
		/* fget() needs to be done in this context */
		if(req->in->h.opcode == FUSE_GETDIR && !oh.error)
			process_getdir(req);
	}	
	spin_unlock(&fuse_lock);

	/* check if it is first write downcall */
	if (req->in->h.opcode == FUSE_WRITE && req->out->numargs == 1) {
		struct sor_write_buf *p = req->out->args[0].value;

		req->out->numargs = 0; /* next downcall don't take argument */
		if (req->in->args[1].size != p->buf_size) {
			printk("Write 1st downcall, size doesn't match %d %d %p\n",
				   req->in->args[1].size, p->buf_size, p->buf);
			err = -EINVAL;
		} else if (access_ok(VERIFY_WRITE, p->buf, p->buf_size) == 0) {
			err = -EFAULT;
		} else {
			if ( copy_to_user(p->buf, req->in->args[1].value, p->buf_size) )
				err = -EFAULT;
		}

		//printk("now err is: %d\n", err);
		if (!err) {
			/* no error, put it back to processing list */
			spin_lock(&fuse_lock);
			req->locked = 0;
			list_add_tail(&req->list, &fc->processing);
			spin_unlock(&fuse_lock);

			return nbytes;
		}
	}

	spin_lock(&fuse_lock);
	req->finished = 1;
	req->locked = 0;
	/* Unlocks fuse_lock: */
	request_end(fc, req);

  out:
	if(!err)
		return nbytes;
	else
		return err;
}


static unsigned int fuse_dev_poll(struct file *file, poll_table *wait)
{
	struct fuse_conn *fc = DEV_FC(file);
	unsigned int mask = POLLOUT | POLLWRNORM;

	if(!fc->sb)
		return -EPERM;

	poll_wait(file, &fc->waitq, wait);

	spin_lock(&fuse_lock);
	if (!list_empty(&fc->pending))
                mask |= POLLIN | POLLRDNORM;
	spin_unlock(&fuse_lock);

	return mask;
}

static struct fuse_conn *new_conn(void)
{
	struct fuse_conn *fc;

	fc = kmalloc(sizeof(*fc), GFP_KERNEL);
	if(fc != NULL) {
		fc->sb = NULL;
		fc->file = NULL;
		fc->flags = 0;
		fc->uid = 0;
		init_waitqueue_head(&fc->waitq);
		INIT_LIST_HEAD(&fc->pending);
		INIT_LIST_HEAD(&fc->processing);
		sema_init(&fc->outstanding, MAX_OUTSTANDING);
		fc->reqctr = 1;
	}
	return fc;
}

static int fuse_dev_open(struct inode *inode, struct file *file)
{
	struct fuse_conn *fc;

	fc = new_conn();
	if(!fc)
		return -ENOMEM;

	fc->file = file;
	file->private_data = fc;

	return 0;
}

static void end_requests(struct fuse_conn *fc, struct list_head *head)
{
	while(!list_empty(head)) {
		struct fuse_req *req;
		req = list_entry(head->next, struct fuse_req, list);
		list_del_init(&req->list);
		if(req->issync) {
			req->out->h.error = -ECONNABORTED;
			req->finished = 1;
			/* Unlocks fuse_lock: */
			request_end(fc, req);
			spin_lock(&fuse_lock);
		} else
			destroy_request(req);
	}
}

static int fuse_dev_release(struct inode *inode, struct file *file)
{
	struct fuse_conn *fc = DEV_FC(file);

	spin_lock(&fuse_lock);
	fc->file = NULL;
	end_requests(fc, &fc->pending);
	end_requests(fc, &fc->processing);
	fuse_release_conn(fc);
	spin_unlock(&fuse_lock);
	return 0;
}

static struct file_operations fuse_dev_operations = {
	.owner		= THIS_MODULE,
	.read		= fuse_dev_read,
	.write		= fuse_dev_write,
	.poll		= fuse_dev_poll,
	.open		= fuse_dev_open,
	.release	= fuse_dev_release,
};


extern struct proc_dir_entry *proc_fs_sor;

int fuse_dev_init()
{
	int ret;

	proc_fuse_dev = NULL;

	fuse_req_cachep = kmem_cache_create("sorrento_request",
					     sizeof(struct fuse_req),
					     0, 0, NULL, NULL);
	if(!fuse_req_cachep)
		return -ENOMEM;

	ret = -ENOMEM;

	/* init Sorrento FS in /proc */
	if (sor_proc_init())
	   	return ret;

	proc_fuse_dev = create_proc_entry("dev", S_IFSOCK | 0600, proc_fs_sor);
	if(!proc_fuse_dev) {
		printk(KERN_INFO "Failed to create entry in /proc/fs/sorrento\n");
		goto err;
	}
	proc_fuse_dev->proc_fops = &fuse_dev_operations;

	return 0;

  err:
	fuse_dev_cleanup();
	return ret;
}

void fuse_dev_cleanup()
{
	if(proc_fs_sor) {
		remove_proc_entry("dev", proc_fs_sor);
	}
	
	sor_proc_cleanup();

	kmem_cache_destroy(fuse_req_cachep);
}
