/* -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*- */

/*
    FUSE: Filesystem in Userspace
    Copyright (C) 2001  Miklos Szeredi (mszeredi@inf.bme.hu)

    This program can be distributed under the terms of the GNU GPL.
    See the file COPYING.
*/
#include "fuse_i.h"
#include "sor_proc.h"
#include "dev.h"

#include <linux/poll.h>
#include <linux/pagemap.h>
#include <linux/slab.h>
#ifdef KERNEL_2_6
#include <linux/backing-dev.h>
#include <linux/writeback.h>
#endif

#ifndef KERNEL_2_6
#define PageUptodate(page) Page_Uptodate(page)
#define filemap_fdatawrite filemap_fdatasync
#endif

static int fuse_open(struct inode *inode, struct file *file)
{
	struct fuse_conn *fc = INO_FC(inode);
	struct fuse_in in = FUSE_IN_INIT;
	struct fuse_out out = FUSE_OUT_INIT;
	struct fuse_open_in inarg;

	sor_vfs_stats.open++;

	/* If opening the root node, no lookup has been performed on
	   it, so the attributes must be refreshed */
	if(inode->i_ino == FUSE_ROOT_INO) {
		int err = fuse_do_getattr(inode);
		if(err)
		 	return err;
	}

	memset(&inarg, 0, sizeof(inarg));
	inarg.flags = file->f_flags & ~O_EXCL;

	in.h.opcode = FUSE_OPEN;
	in.h.ino = inode->i_ino;
	in.numargs = 1;
	in.args[0].size = sizeof(inarg);
	in.args[0].value = &inarg;
	request_send(fc, &in, &out);
	if(!out.h.error && !(fc->flags & FUSE_KERNEL_CACHE)) {
#ifdef KERNEL_2_6
		invalidate_inode_pages(inode->i_mapping);
#else
		invalidate_inode_pages(inode);
#endif
	}

	return out.h.error;
}

/** 
 * When a process calls close() system call, this is invoked to force
 * file system daemon to write back dirty buffers.
 */
static int
sor_file_flush(struct file *file)
{
	struct inode *inode;
	struct fuse_conn *fc;
	struct fuse_in in = FUSE_IN_INIT;
	struct fuse_out out = FUSE_OUT_INIT;

	inode = file->f_dentry->d_inode;
	if (inode == NULL) {
		printk("%s:%d inode = NULL\n", __FUNCTION__, __LINE__);
		return 0;
	}
	fc = INO_FC(inode);

	in.h.opcode = SOR_FILE_FLUSH;
	in.h.ino = inode->i_ino;
	in.numargs = 0;
	request_send(fc, &in, &out);
	return out.h.error;
}

static int fuse_release(struct inode *inode, struct file *file)
{
	struct fuse_conn *fc = INO_FC(inode);
	struct fuse_in *in = NULL;
	struct fuse_open_in *inarg = NULL;
	unsigned int s = sizeof(struct fuse_in) + sizeof(struct fuse_open_in);

	sor_vfs_stats.release++;

	if(file->f_mode & FMODE_WRITE)
		filemap_fdatawrite(inode->i_mapping);

	in = kmalloc(s, GFP_NOFS);
	if(!in)
		return -ENOMEM;
	memset(in, 0, s);
	inarg = (struct fuse_open_in *) (in + 1);
	inarg->flags = file->f_flags & ~O_EXCL;

	in->h.opcode = FUSE_RELEASE;
	in->h.ino = inode->i_ino;
	in->numargs = 1;
	in->args[0].size = sizeof(struct fuse_open_in);
	in->args[0].value = inarg;
	if(!request_send_noreply(fc, in))
		return 0;

	kfree(in);
	return 0;
}

static int fuse_fsync(struct file *file, struct dentry *de, int datasync)
{
	struct inode *inode = de->d_inode;
	struct fuse_conn *fc = INO_FC(inode);
	struct fuse_in in = FUSE_IN_INIT;
	struct fuse_out out = FUSE_OUT_INIT;
	struct fuse_fsync_in inarg;
	
	sor_vfs_stats.fsync++;

	memset(&inarg, 0, sizeof(inarg));
	inarg.datasync = datasync;

	in.h.opcode = FUSE_FSYNC;
	in.h.ino = inode->i_ino;
	in.numargs = 1;
	in.args[0].size = sizeof(inarg);
	in.args[0].value = &inarg;
	request_send(fc, &in, &out);
	return out.h.error;

	/* FIXME: need to ensure, that all write requests issued
           before this request are completed.  Should userspace take
           care of this? */
}

static int fuse_readpage(struct file *file, struct page *page)
{
	struct inode *inode = page->mapping->host;
	struct fuse_conn *fc = INO_FC(inode);
	struct fuse_in in = FUSE_IN_INIT;
	struct fuse_out out = FUSE_OUT_INIT;
	struct fuse_read_in inarg;
	char *buffer;

	sor_vfs_stats.readpage++;
	buffer = kmap(page);
	
	memset(&inarg, 0, sizeof(inarg));
	inarg.offset = (unsigned long long) page->index << PAGE_CACHE_SHIFT;
	inarg.size = PAGE_CACHE_SIZE;

	in.h.opcode = FUSE_READ;
	in.h.ino = inode->i_ino;
	in.numargs = 1;
	in.args[0].size = sizeof(inarg);
	in.args[0].value = &inarg;
	out.argvar = 1;
	out.numargs = 1;
	out.args[0].size = PAGE_CACHE_SIZE;
	out.args[0].value = buffer;

	request_send(fc, &in, &out);
	if(!out.h.error) {
		size_t outsize = out.args[0].size;
		if(outsize < PAGE_CACHE_SIZE) 
			memset(buffer + outsize, 0, PAGE_CACHE_SIZE - outsize);
		SetPageUptodate(page);
	}

	kunmap(page);
	unlock_page(page);

	return out.h.error;
}



#if 1
int sor_maxsz = SOR_DEFAULT_IO_SIZE;

/**
 * @param buf User land buffer to be filled in
 * @param count Buffer size
 * @param offp On success, it is increased by amount read
 * @return number of bytes read on success, or error (<0) on
 * failure.
 */
static int
ll_sor_file_read(struct inode *inode, struct fuse_conn *fc,
				 char *buf, int count, loff_t *offp)
{
	struct fuse_in in = FUSE_IN_INIT;
	struct fuse_out out = FUSE_OUT_INIT;
	struct fuse_read_in inarg;
	int err;

	memset(&inarg, 0, sizeof(inarg));
	inarg.offset = *offp;
	inarg.size = count;

	in.h.opcode = FUSE_READ;
	in.h.ino = inode->i_ino;
	in.numargs = 1;
	in.args[0].size = sizeof(inarg);
	in.args[0].value = &inarg;


	out.argvar = 1;
	out.numargs = 1;
	/* allocate a new kernel buffer */
	out.args[0].size = count;
	err = get_pool_buffer(&(out.args[0].value), count);
	if (err < 0)
		return err;

	request_send(fc, &in, &out);
	if(!out.h.error) {
		/* copy to user the results got and free the kernel buffer */
		if (copy_to_user(buf, out.args[0].value, out.args[0].size)) {
			free_pool_buffer( &(out.args[0].value), count);
			return -EFAULT;
		}

		free_pool_buffer( &(out.args[0].value), count);
		*offp += out.args[0].size;
		return out.args[0].size;
	}

	return out.h.error;
}

/** 
 * Sorrento FS specific read function 
 * @return number of bytes read on success, -errno otherwise
 */
static ssize_t 
sor_file_read(struct file *filp,
			  char *buf,
			  size_t count,
			  loff_t *ppos)
{
	struct address_space *mapping = filp->f_dentry->d_inode->i_mapping;
	struct inode *inode = mapping->host;
	struct fuse_conn *fc = INO_FC(inode);
	int retsz = 0, xfersize, error;
	loff_t sor_pos = *ppos;

	sor_vfs_stats.read++;

	if (access_ok(VERIFY_WRITE, buf, count) == 0) {
		return -EFAULT;
	}

	if (count == 0) {
		return 0;
	}

	/* split read operation into blocks of sor_maxsz or smaller */
	do {
		xfersize = (count < sor_maxsz) ? count: sor_maxsz;
		error = ll_sor_file_read(inode, fc, buf, xfersize, &sor_pos);
		if (error < 0) {
			return error;
		}
		retsz += error;

		/* positition is updated inside ll_sor_file_read() */
		count -= error;
		buf += error;   
		/* pvfs used xfersize, which may be not the same as xfersize.
		 * Since sor_pos is upaded as real number of bytes read, we 
		 * should be consistent on the number of bytes read.
		 */
	} while (count > 0 && error != 0);

	*ppos = sor_pos;
	return retsz;
}

/**
 * @param buf User land buffer to read from
 * @param count Buffer size
 * @param offp On success, it is increased by amount read
 * @return number of bytes write on success, or error (<0) on
 * failure.
 */
static int
ll_sor_file_write(struct inode *inode, struct fuse_conn *fc,
				  char *buf, int count, loff_t *offp)
{
	struct fuse_in in = FUSE_IN_INIT;
	struct fuse_out out = FUSE_OUT_INIT;
	struct fuse_read_in inarg;
	struct sor_write_buf outarg; /* hold first downcall data */
	int err;

	memset(&inarg, 0, sizeof(inarg));
	inarg.offset = *offp;
	inarg.size = count;

	in.h.opcode = FUSE_WRITE;
	in.h.ino = inode->i_ino;

	/* Write is in several steps. First, the buffer size is sent to
	 * user.  When file server notify the kernel that user buffer is
	 * ready, we copy the data to user. Finally, the file server tell
	 * kernel the completion, then we can return.
	 *
	 * Here we prepare two arguments, but only the first is copied to
	 * the file server. The second is kernel buffer used when the file
	 * server says user buffer is ready.
	 */
	in.numargs = 1;
	in.args[0].size = sizeof(inarg);
	in.args[0].value = &inarg;

	/* allocate a new kernel buffer */
	in.args[1].size = count;
	err = get_pool_buffer( (void**)&(in.args[1].value), count);
	if (err < 0)
		return err;

	/* copy from user buffer to the kernel buffer */
	err = copy_from_user((void*)in.args[1].value, buf, count);
	if (err) {
		free_pool_buffer( (void**)&(in.args[1].value), count);
		return err;
	}

	/* set buffer for downcall data */
	out.numargs = 1;
	out.args[0].size = sizeof(outarg);
	out.args[0].value = &outarg;

	//printk("ll_sor_file_write: count %d, offset %d\n", count, *offp);
	request_send(fc, &in, &out);

	free_pool_buffer( (void**)&(in.args[1].value), count);
	if(!out.h.error) {
		*offp += count;
		return count;
	}

	return out.h.error;
}

/** 
 * Sorrento FS specific write function 
 * @return number of bytes read on success, -errno otherwise
 */
static ssize_t 
sor_file_write(struct file *filp,
			   const char *buf,
			   size_t count,
			   loff_t *ppos)
{
	struct address_space *mapping = filp->f_dentry->d_inode->i_mapping;
	struct inode *inode = mapping->host;
	struct fuse_conn *fc = INO_FC(inode);
	int retsz = 0, xfersize, error;
	loff_t sor_pos = *ppos;

	sor_vfs_stats.write++;

	if (access_ok(VERIFY_READ, buf, count) == 0) {
		return -EFAULT;
	}

	if (count == 0) {
		return 0;
	}

	/* split write operation into blocks of sor_maxsz or smaller */
	do {
		xfersize = (count < sor_maxsz) ? 
			count: sor_maxsz;
		error = ll_sor_file_write(inode, fc, (char*)buf, xfersize, &sor_pos);
		if (error < 0) {
			return error;
		}
		retsz += error;

		/* positition is updated inside ll_sor_file_write() */
		count -= error;
		buf += error;   
		/* pvfs used xfersize, which may be not the same as xfersize.
		 * Since sor_pos is upaded as real number of bytes read, we 
		 * should be consistent on the number of bytes read.
		 */
	} while (count > 0 && error != 0);

	*ppos = sor_pos;
	return retsz;
}
#endif

#if 0
static ssize_t fuse_file_read(struct file *filp, char *buf,
		size_t count, loff_t * ppos)
{
	struct address_space *mapping = filp->f_dentry->d_inode->i_mapping;
	struct inode *inode = mapping->host;
	struct fuse_conn *fc = INO_FC(inode);

	sor_vfs_stats.read++;

	if(fc->flags & FUSE_LARGE_READ)
		fuse_file_bigread(mapping, inode, *ppos, count);

	return generic_file_read(filp, buf, count, ppos);
}  

static int fuse_prepare_write(struct file *file, struct page *page,
			      unsigned offset, unsigned to)
{
	sor_vfs_stats.prepare_write++;

	/* No op */
	return 0;
}

static int fuse_commit_write(struct file *file, struct page *page,
			     unsigned offset, unsigned to)
{
	int err;
	struct inode *inode = page->mapping->host;

	sor_vfs_stats.commit_write++;

	err = write_buffer(inode, page, offset, to - offset);
	if(!err) {
		loff_t pos = (page->index << PAGE_CACHE_SHIFT) + to;
		if(pos > inode->i_size)
			inode->i_size = pos;
	}
	return err;
}

#ifdef KERNEL_2_6
static void write_buffer_end(struct fuse_conn *fc, struct fuse_in *in,
			     struct fuse_out *out, void *_page)
{
	struct page *page = (struct page *) _page;
	
	lock_page(page);
	if(out->h.error) {
		SetPageError(page);
		if(out->h.error == -ENOSPC)
			set_bit(AS_ENOSPC, &page->mapping->flags);
		else
			set_bit(AS_EIO, &page->mapping->flags);
	}
	end_page_writeback(page);
	kunmap(page);
	unlock_page(page);
	kfree(in);	
}

static int write_buffer_nonblock(struct inode *inode, struct page *page,
				 unsigned offset, size_t count)
{
	int err;
	struct fuse_conn *fc = INO_FC(inode);
	struct fuse_in *in = NULL;
	struct fuse_out *out = NULL;
	struct fuse_write_in *inarg = NULL;
	char *buffer;
	unsigned int s = sizeof(struct fuse_in) + sizeof(struct fuse_out) +
		sizeof(struct fuse_write_in);

	in = kmalloc(s, GFP_NOFS);
	if(!in)
		return -ENOMEM;
	memset(in, 0, s);
	out = (struct fuse_out *)(in + 1);
	inarg = (struct fuse_write_in *)(out + 1);
	
	buffer = kmap(page);

	inarg->offset = ((unsigned long long) page->index << PAGE_CACHE_SHIFT) + offset;
	inarg->size = count;
	
	in->h.opcode = FUSE_WRITE;
	in->h.ino = inode->i_ino;
	in->numargs = 2;
	in->args[0].size = sizeof(struct fuse_write_in);
	in->args[0].value = inarg;
	in->args[1].size = count;
	in->args[1].value = buffer + offset;
	err = request_send_nonblock(fc, in, out, write_buffer_end, page);
	if(err) {
		if(err != -EWOULDBLOCK)
			SetPageError(page);
		kunmap(page);
		kfree(in);
	}
	return err;
}

static int fuse_writepage(struct page *page, struct writeback_control *wbc)
{
	int err;
	struct inode *inode = page->mapping->host;
	unsigned count = get_write_count(inode, page);

	sor_vfs_stats.writepage++;

	err = -EINVAL;
	if(count) {
		/* FIXME: check sync_mode, and wait for previous writes (or
		   signal userspace to do this) */
		if(wbc->nonblocking) {
			err = write_buffer_nonblock(inode, page, 0, count);
			if(!err)
				SetPageWriteback(page);
			else if(err == -EWOULDBLOCK) {
				__set_page_dirty_nobuffers(page);
				err = 0;
			}
		} else
			err = write_buffer(inode, page, 0, count);
	}

	unlock_page(page);
	return err;
}
#else /* 2.4 kernel follows */
static int fuse_writepage(struct page *page)
{
	int err;
	struct inode *inode = page->mapping->host;
	int count = get_write_count(inode, page);
	sor_vfs_stats.writepage++;

	err = -EINVAL;
	if(count)
		err = write_buffer(inode, page, 0, count);

	unlock_page(page);
	return err;
}
#endif /* kernel 2.6 */

static int write_buffer(struct inode *inode, struct page *page,
			unsigned offset, size_t count)
{
	struct fuse_conn *fc = INO_FC(inode);
	struct fuse_in in = FUSE_IN_INIT;
	struct fuse_out out = FUSE_OUT_INIT;
	struct fuse_write_in inarg;
	char *buffer;

	buffer = kmap(page);

	memset(&inarg, 0, sizeof(inarg));
	inarg.offset = ((unsigned long long) page->index << PAGE_CACHE_SHIFT) +
		offset;
	inarg.size = count;
	
	in.h.opcode = FUSE_WRITE;
	in.h.ino = inode->i_ino;
	in.numargs = 2;
	in.args[0].size = sizeof(inarg);
	in.args[0].value = &inarg;
	in.args[1].size = count;
	in.args[1].value = buffer + offset;
	request_send(fc, &in, &out);
	kunmap(page);
	if(out.h.error)
		SetPageError(page);

	return out.h.error;
}

static int get_write_count(struct inode *inode, struct page *page)
{
	unsigned long end_index;
	int count;
	
	end_index = inode->i_size >> PAGE_CACHE_SHIFT;
	if(page->index < end_index)
		count = PAGE_CACHE_SIZE;
	else {
		count = inode->i_size & (PAGE_CACHE_SIZE - 1);
		if(page->index > end_index || count == 0)
			return 0;
	}
	return count;
}

#endif /* if 0 */



static struct file_operations fuse_file_operations = {
	.open =		fuse_open,
	.flush =    sor_file_flush, /* invoked by sys_close() when closing file descriptor */
	.release =	fuse_release,   /* called much later than sys_close() */
	.fsync =	fuse_fsync,
	//.read =		fuse_file_read,
	.read =     sor_file_read,
	//.write =	generic_file_write,
	.write = sor_file_write,
	.mmap =		generic_file_mmap,
};

static struct address_space_operations fuse_file_aops  = {
	.readpage =		fuse_readpage,
	//.writepage =		fuse_writepage,
	//.prepare_write =	fuse_prepare_write,
	//.commit_write =		fuse_commit_write,
};

void fuse_init_file_inode(struct inode *inode)
{
#ifdef KERNEL_2_6
	struct fuse_conn *fc = INO_FC(inode);
	/* Readahead somehow defeats big reads on 2.6 (says Michael
           Grigoriev) */
	if(fc->flags & FUSE_LARGE_READ)
		inode->i_mapping->backing_dev_info->ra_pages = 0;
#endif
	inode->i_fop = &fuse_file_operations;
	inode->i_data.a_ops = &fuse_file_aops;
}

/* 
 * Local Variables:
 * indent-tabs-mode: t
 * c-basic-offset: 4
 * End:
 */
