// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

# ifndef _SORRENTO_CLIENT_H_
# define _SORRENTO_CLIENT_H_

# include "sorrento_types.h"
# include "host_map.h"
# include "../dir_srv/dir_cli_mt.h"
# include "io_cache.h"

class DIR_Cli_AP;
class Client;
class SorFile;
class SorFS;



/** maintaing a mapping from int to SorFile pointer */
#define FILE_TABLE_SIZE 128

/** Default invalid inode number */
#define INVALID_INO     -1

#define AP_TABLE_SIZE 8
#define CLIENT_POOL_SIZE 8
#define THREAD_UNUSED 0

/** Client pool: storage access point pool */
class ClientPool {
private:
	struct CP {
		Client *cli;            /**< Provider access point */
		pthread_mutex_t lock;   /**< lock for each AP */
		pthread_t thread;       /**< the thread current using the AP */
		bool inuse;             /**< If the AP is in use */
	};

	std::vector<struct CP*> cp; /**< Client Pool */
	pthread_mutex_t cp_lock;    /**< protect operations on cp that affect its size */
	struct CP* allocate_CP();

public:
	ClientPool();
	~ClientPool();

	Client* get_client();       /**< Get a client AP */
	void put_client();          /**< Release an AP */
};



/**
 * Directory Access Point Pool: different threads HAVE to use
 * different access point when accessing the directory server. They
 * will need to obtain one from this table. Only used by SClientStub.
 * 
 * TODO: detect directory server failure and update its access
 * point. Now we just assume the directory server never fails. If it
 * happens, the daemon just hangs due to timeout of UDP request to
 * unknown port. The callback interface of Hmap can't update event of
 * directory server. We can only periodically check the liveness of
 * the directory server.
 */
class APPool {
private:
	struct AP {
		DIR_Cli_AP *ap;        /**< The directory server Access Point (AP) */
		pthread_mutex_t lock;  /**< locks for each AP */
		pthread_t thread;      /**< the thread current using the AP */
		bool inuse;            /**< If the AP is in use */
	};

	SorFS *sp;                 /**< Sorrento FS handle, for allocate AP */
	std::vector<struct AP*> aps;  /**< directory access pool */
	pthread_mutex_t ap_lock;   /**< protect operation on aps that affect its size */

	struct AP* allocate_AP();

public:
	APPool(SorFS *);
	~APPool();

	DIR_Cli_AP* get_ap();     /**< Get an AP */
	void put_ap();            /**< Release an AP */
};


/* forward declaration */
class SClientStub;

/** 
 * File table: maintains a mapping from integer to SorFile objects.
 * Multi-thread safety is ensured by locking before access a SorFile
 * object. Should be private class of SClientStub.
 */
class FileTable {
private:
	/* passed from SClientStub */
	HostMap *hmap;   /**< The host map structure */
	APPool *apt;     /**< The Directory Server Access Point pool */
	ClientPool *cp;  /**< The stoarge access point. */
	SClientStub *client; /**< main client, for adding async. I/O jobs */

	struct File {
		SorFile *file;         /**< Sorrento File Handle */
		pthread_mutex_t lock;  /**< locks for each file */
		int flags;             /**< file flags */
		long ino;              /**< inode number of this file */
		bool valid;            /**< If the file is valid */
		struct io_job jobs;    /**< A list of jobs pending */

		File() :file(NULL){} /* not used, just make compiler happy */
	};
	struct File files[FILE_TABLE_SIZE]; /**< file table */

	friend class IOCache;
	/** Synchronous Write a number of bytes from a certain offset */
	int s_write(int fd, off64_t offset, char *buf, off_seg_t len, 
				off_seg_t& written_bytes, bool sync=true);

	inline int flush_buffers(int fd);

public:
	FileTable(HostMap*, APPool*, ClientPool*, SClientStub*);
	~FileTable();  /**< Destructor */

	/** Add a new SorFile to the table */
	int put_file(SorFile *fp, int flags, long ino);

	/** Get a SorFile object by index */
	SorFile* get_file(int fd);

	/** Find file descriptor for a given inode */
	int get_fd(long ino);

	/** Close a SorFile object by index */
	int close_file(int fd);

	/** Read a number of bytes from a certain offset */
	int read(int fd, off64_t offset, char *buf, off_seg_t len, off_seg_t & read_bytes); 

	/** Write a number of bytes from a certain offset */
	int write(int fd, off64_t offset, char *buf, off_seg_t len, off_seg_t & written_bytes);


	/* Unlink an asynchronous job from pending list */
	int remove_job(int fd, struct io_job *job);

	/* flush all delayed writes */
	int flush(int fd);
};



/**
 * Sorrento Client Stub - A multi-thread-safe Sorrento access point.
 *
 * maintaining per-client state information such
 * as current working directory (within a certain Sorrento file 
 * system). We also keep a DIR_Srv_AP pointer that can be used
 * to request dir_srv related functionalities. This
 */
class SClientStub {
 private:
	HostMap *hmap;   /**< The host map structure */
	DIR_Entry cwd;   /**< The current working directory. */
	int errnum;      /**< per client previous error number */
	int callid;      /**< per client previous call ID */

	FileTable *ft;   /**< Sorrento File handle table */
	APPool *apt;     /**< Directory server access point pool */
	ClientPool *cp;  /**< Storage access point pool */
#if 0
	HostCB *hostCB;  /**< Call back functions of host joins/leaves */
#endif
	IOCache *cache;  /**< Cache object - for asynchronous operations */

 public:
	typedef DIR_Entry Cookie;

	SClientStub(SorFS *sp); /**< The constructor. */
	~SClientStub(); /**< The destructor. */


	int mkdir(const char *path); /**< Make a new directory. */
	int rmdir(const char *path); /**< Remove an existing directory. */

	/** Read the content of a directory. */
	int readdir(const char *path, DIR_List &list, int buf_len, Cookie *cookie);
	/** 
	 * Is end-of-directory? Check whether we need to make another 
	 * readdir_cont() call. 
	 */
	bool is_eod(const DIR_Entry *cookie) const;
	/** Continue the previous reading. */
	int readdir_cont(Cookie *cookie, DIR_List &list, int buf_len);


	int unlink(const char *path); /**< remove an existing file node. */
	SorFile *create(const char *path); 
	int open(const char *path, int flags, long ino); /**< Open a file for access */
	int close(SorFile *sf); /**< Close a file */
	int release(long ino);   /**< Close a file after kernel reference count becomes 0 */

	/** Read a number of bytes from a certain offset */
	int read(long ino, off64_t offset, char *buf, off_seg_t len, off_seg_t & read_bytes); 

	/** Write a number of bytes from a certain offset */
	int write(long ino, off64_t offset, char *buf, off_seg_t len, off_seg_t & written_bytes);

	/** Change the size of the current file */
	int ftruncate(SorFile *sf, off64_t offset);
	int truncate(const char *path, off64_t offset);

	/** Get attribute of a certain file. */
	int stat(const char *path, DIR_Param *param);

	const char *get_pwd(void) const { return cwd.path(); }
	const char *ident(void) const { return hmap->ident(); }

	/** Print out a string corresponding to previous error state. */
	std::string str_error(); 
	void dump_view(void);
	void stop_service(void);


	/* Asynchronous stuff, called by FileTable */
	bool add_job(struct io_job *job, int type) {
		return cache->add_job(job, type);
	}

	/* flush delayed write buffer */
	int flush(long ino);
};

#if 0
class HostCB : public SrvCBInterface {
public:
	virtual void notify_host_join(HostID &hid) {}

	virtual void notify_host_leave(HostID &hid) {
		printf("host leaves\n");
	}

	virtual bool is_self(HostID &hid) {return false;}
};
#endif

# endif // _SORRENTO_CLIENT_H_
