/* -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*- */

#include <errno.h>
#include <sys/time.h>

#include "atomic.h"
#include "tmpi_debug.h"
#include "io_cache.h"
#include "SClientStub.h"

/** Constructor */
IOCache::IOCache(FileTable *p, unsigned limit)
{
	ftable = p;
	pthread_mutex_init(&dirty_lock, NULL);
	pthread_cond_init(&dirty_cond, NULL);

	/* minimum set to 1MB */
	_limit = (limit >= 1024*1024) ? limit : 1024 * 1024;
	_csize = 0;
	pthread_mutex_init(&size_lock, NULL);
	pthread_cond_init(&size_cond, NULL);
}

/** Destructor */
IOCache::~IOCache()
{
	/* XXX: should finish pending requests before exit */
	pthread_mutex_destroy(&dirty_lock);
	pthread_cond_destroy(&dirty_cond);

	pthread_mutex_destroy(&size_lock);
	pthread_cond_destroy(&size_cond);
}

/** 
 * Add a new job: here we just use a simple FIFO policy and assume requests
 * is write (so that it will be added to dirty list).
 * @param job I/O job
 * @param type job type
 * @return true if successful
 */
bool
IOCache::add_job(struct io_job *job, int type)
{
	/* simple validity check */
	if (job == NULL || job->buf == NULL)
		return false;

	/* wait until we have enough buffering space */
	pthread_mutex_lock(&size_lock);
	while ( _csize + job->len > _limit ) {
		pthread_cond_signal(&dirty_cond); /* ask cache thread do jobs */

		pthread_cond_wait(&size_cond, &size_lock); /* wait */
	}
	_csize += job->len;
	pthread_mutex_unlock(&size_lock);

	pthread_mutex_lock(&dirty_lock);
	dirty_list.push_back(job);
	pthread_cond_signal(&dirty_cond);
	pthread_mutex_unlock(&dirty_lock);

	return true;
}

/** Perform asychrnnous I/O */
void*
IOCache::run()
{
	struct timeval now;
	struct timespec timeout;
	int ret;


	while (1) {
		gettimeofday(&now, NULL);
		timeout.tv_sec = now.tv_sec + 1; /* timeout every 1 second */
		timeout.tv_nsec = now.tv_usec * 1000;

		ret = pthread_cond_timedwait(&dirty_cond, &dirty_lock, &timeout);
		if (should_quit())
			break;

		/* if (ret == ETIMEDOUT), the lock is acquired */
		if (ret == EINTR)
			continue;

		/* we have successfully got the lock from now on */
		if (dirty_list.size() == 0) {
			pthread_mutex_unlock(&dirty_lock);
			continue;
		}

		/* find the first unfinished job */
		struct io_job *job = NULL;
		while (dirty_list.size() > 0) {
			job = dirty_list[0];
			dirty_list.erase(dirty_list.begin());
			if (!job->done) {
				break;
			}

			/* signal thread waiting for buffering space */
			pthread_mutex_lock(&size_lock);
			_csize -= job->len;
			if (_csize < _limit)
				pthread_cond_signal(&size_cond);
			pthread_mutex_unlock(&size_lock);

			//free(job->buf);
			delete job; /* the job is done in flush_buffer() call */

			job = NULL;
		}
		pthread_mutex_unlock(&dirty_lock);

		if (job == NULL)
			continue;

		/* skip this job if another thread (flush or relese call) is using it */
		if ( cas16(&(job->busy), IO_JOB_IDLE, IO_JOB_BUSY) == 0 )
			continue; 

		off_seg_t written_bytes;
		/* how about async. write and push back job? will signal be lost? */
		ret = ftable->s_write(job->idx, job->offset, job->buf, job->len,
							  written_bytes, true); 
		if (ret != 0) {
			/* TODO: write error. What should we do here? */
			tmpi_error(DBG_INTERNAL, "delayed write failed! (%d, %lld, %d, %d)\n",
					   job->idx, job->offset, job->len, ret);
		}

		/* signal thread waiting for buffering space */
		pthread_mutex_lock(&size_lock);
		_csize -= job->len;
		if (_csize < _limit)
			pthread_cond_signal(&size_cond);
		pthread_mutex_unlock(&size_lock);

		/* unlink from file's pending list */
		ftable->remove_job(job->idx, job);

		/* free memory */
		if (job->buf != NULL)
			free(job->buf); /* allocated in the lib */
		delete job;
	}

	return NULL;
}
