/* -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*- */

#include <fuse.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <errno.h>
#include <sys/statfs.h>

#include <signal.h>
#include <assert.h>

#include "SClientStub.h"
#include "tmpi_debug.h"
#include "debug.h"

#define SOR_ENV  "SORRENTO_CONF"
static SClientStub *scli = NULL;



static int 
sor_getattr(const char *path, struct stat *stbuf)
{
    DIR_Param param;
	int ret;
	PENTRY;

	if (sor_debug_flag & DEBUG_STATS)
		sor_stats.getattr++;

	//tmpi_error(DBG_VERBOSE, "getattr %s\n", path);
	tmpi_error(DBG_ALL, "getattr T[%d] %s %f\n", pthread_self(), path, getTime());
    assert(scli != NULL);
    if ( (ret = scli->stat(path, &param)) != 0 ) {
		return -ret; /* file not exist */
    }
    
    memset(stbuf, 0, sizeof(struct stat));
    /* set mode for the file or directory */
    stbuf->st_mode = 0777;
    stbuf->st_mode |= DIR_IS_DIR(param.options) ? S_IFDIR:S_IFREG;
    /* the following doesn't really matter */
    stbuf->st_nlink = 1;
    stbuf->st_size = param.size;
    stbuf->st_blksize = 512;
    stbuf->st_blocks = param.size / 512;
    stbuf->st_atime = param.t_create;
    stbuf->st_mtime = param.t_create;
    stbuf->st_ctime = param.t_create;

	PEXIT;
    return 0;
}

static int 
sor_getdir(const char *path, fuse_dirh_t h, fuse_dirfil_t filler)
{
	DIR_Param param;
	int ret;
	PENTRY;

	if (sor_debug_flag & DEBUG_STATS)
		sor_stats.getdir++;

	tmpi_error(DBG_ALL, "getdir %s\n", path);
	assert(scli != NULL);
	if ( (ret = scli->stat(path, &param)) != 0) {
		PEXIT;
		return -ret;
    }
	
    bool is_dir = DIR_IS_DIR(param.options);
    if (is_dir) {
		SClientStub::Cookie cookie;
		DIR_List list;
		
		if ( (ret = scli->readdir(path, list, 30000, &cookie)) != 0) {
			tmpi_error(DBG_WARN1, "List/readdir\n");
			return 0;
		}
		
		while (!scli->is_eod(&cookie)) {
			if (scli->readdir_cont(&cookie, list, 30000)) {
				tmpi_error(DBG_WARN1, "List/readdir_cont\n");
				return 0;
			}
		}

		for (unsigned int i=0; i < list.size(); i++) {
			is_dir = DIR_IS_DIR(list[i].param.options);
			filler(h, list[i].name.c_str(), is_dir ? DT_DIR : DT_REG);
		}
    }

	PEXIT;	
    return 0;
}

static int
sor_mknod(const char *path, mode_t mode, dev_t rdev)
{
	SorFile *fp;
	PENTRY;
	
	if (sor_debug_flag & DEBUG_STATS)
		sor_stats.mknod++;

	//tmpi_error(DBG_VERBOSE, "mknod T[%d] %s %f\n", pthread_self(), path, getTime());
	tmpi_error(DBG_INFO, "mknod T[%d] %s %f\n", pthread_self(), path, getTime());
	assert(scli != NULL);
    fp = scli->create(path);
    if (fp == NULL) {
		tmpi_error(DBG_WARN1, "Create %s error\n", path);
		return -1;
    }

	scli->close(fp);

	PEXIT;
    return 0;
}

static int 
sor_mkdir(const char *path, mode_t mode)
{
    int ret;
	PENTRY;

	if (sor_debug_flag & DEBUG_STATS)
		sor_stats.mkdir++;

    assert(scli != NULL);
    if ((ret = scli->mkdir(path)) != 0) {
		tmpi_error(DBG_WARN1, "Mkdir %s error\n", path);
        return -ret;
    }
	
	PEXIT;
    return 0;
}

static int 
sor_unlink(const char *path)
{
	int ret;
	PENTRY;

	if (sor_debug_flag & DEBUG_STATS)
		sor_stats.unlink++;

    assert(scli != NULL);
    if ( (ret = scli->unlink(path)) != 0) {
		tmpi_error(DBG_WARN1, "Unlink %s error\n", path);
		return -ret;
    }

	PEXIT;
    return 0;
}


static int 
sor_rmdir(const char *path)
{
    int ret;
	PENTRY;

	if (sor_debug_flag & DEBUG_STATS)
		sor_stats.rmdir++;

    assert(scli != NULL);
    if ( (ret = scli->rmdir(path)) != 0) {
		tmpi_error(DBG_WARN1, "Rmdir %s error\n", path);
		return -ret;
    }

	PEXIT;
    return 0;
}

#if 0
static int xmp_rename(const char *from, const char *to)
{
    int res;

    res = rename(from, to);
    if(res == -1)
        return -errno;

    return 0;
}
#endif

static int 
sor_truncate(const char *path, off_t size)
{
    int ret;
	PENTRY;

	if (sor_debug_flag & DEBUG_STATS)
		sor_stats.truncate++;

	assert(scli != NULL);
    if ( (ret = scli->truncate(path, size)) != 0) {
		tmpi_error(DBG_WARN1, "%s %s error\n", __FUNCTION__, path);
		return -ret;
    }

	PEXIT;
	return 0;
}

static int
sor_utime(const char *path, struct utimbuf *buf)
{
	tmpi_error(DBG_VERBOSE, "utime\n");
    assert(scli != NULL);
    return 0;
}


/**
 * @return status (0 for success, -ERRNO)
 */
static int 
sor_open(const char *path, int flags, long ino)
{
	int ret;
	PENTRY;

	if (sor_debug_flag & DEBUG_STATS)
		sor_stats.open++;

	tmpi_error(DBG_ALL, "open %s ino %ld T[%d]\n", path, ino, pthread_self());
    assert(scli != NULL);
#if 0
	if (flags & O_CREAT) {
		/* create a new file */
		file = scli->create(path);
		if (file == NULL) {
			sor_error2("Create", path);
			return -1;
		}
	} else {
#endif
		/* open an existing file */
	    ret = scli->open(path, flags, ino);
		if(ret != 0) {
			tmpi_error(DBG_WARN1, "%s %s error\n", __FUNCTION__, path);
			return -ret;
		}
#if 0
    }
#endif

	PEXIT;
	if (sor_debug_flag & DEBUG_STATS)
		sor_stats.t_open += __t_duration;
    return 0;
}

/**
 * Read from a file 
 * @return number of bytes read, on error -ERRNO is returned
 */
static int 
sor_read(long ino, char *buf, size_t size, off_t offset)
{
	off64_t off = offset;
	int ret;
	off_seg_t nread;
	PENTRY;

	if (sor_debug_flag & DEBUG_STATS)
		sor_stats.read++;

	tmpi_error(DBG_ALL, "read ino %ld T[%d]\n", ino, pthread_self());
    assert(scli != NULL);

    if ( (ret =scli->read(ino, off, buf, size, nread)) != 0) {
		tmpi_error(DBG_WARN1, "%s error (%d:%s)\n", __FUNCTION__, 
				   ret, strerror(ret));
		return -ret;
    }

	PEXIT;
	if (sor_debug_flag & DEBUG_STATS)
		sor_stats.t_read += __t_duration;
	return nread;
}

/**
 * Write to a file 
 * @return number of bytes write, on error -ERRNO is returned
 */
static int 
sor_write(long ino, const char *buf, size_t size, off_t offset)
{
    off_seg_t len;
    int ret;
	PENTRY;

	if (sor_debug_flag & DEBUG_STATS)
		sor_stats.write++;

	tmpi_error(DBG_ALL, "write to ino %ld, size %d, offset %d\n", 
			   ino, size, offset);
    assert(scli != NULL);

#define NO_DEBUG_ROUND_TRIP_WITH_BUFFER
#ifdef NO_DEBUG_ROUND_TRIP_WITH_BUFFER
    if ( (ret = scli->write(ino, offset, (char*)buf, size, len)) != 0) {
		tmpi_error(DBG_WARN1, "%s error\n", __FUNCTION__);
		return -ret;
    }
#else
	len = size;
#endif

	PEXIT;
	if (sor_debug_flag & DEBUG_STATS)
		sor_stats.t_write += __t_duration;
    return len;
}

static int 
sor_release(long ino, int flags)
{
	int ret;
	PENTRY;

	if (sor_debug_flag & DEBUG_STATS)
		sor_stats.release++;

	tmpi_error(DBG_ALL, "release ino %ld T[%d]\n", ino, pthread_self());
    ret = scli->release(ino);

	PEXIT;
	if (sor_debug_flag & DEBUG_STATS)
		sor_stats.t_release += __t_duration;
	return ret;
}

#if 0
static int xmp_statfs(struct fuse_statfs *fst)
{
    struct statfs st;
    int rv = statfs("/",&st);
    if(!rv) {
    	fst->block_size  = st.f_bsize;
    	fst->blocks      = st.f_blocks;
    	fst->blocks_free = st.f_bavail;
    	fst->files       = st.f_files;
    	fst->files_free  = st.f_ffree;
    	fst->namelen     = st.f_namelen;
    }
    return rv;
}
#endif


static int 
sor_fsync(const char *path, int isdatasync)
{
	if (sor_debug_flag & DEBUG_STATS)
		sor_stats.fsync++;

    tmpi_error(DBG_VERBOSE, "%s %s\n", __FUNCTION__, path);
    return 0;
}

/** flush delayed write buffer */
static int
sor_flush(int ino)
{
	return scli->flush(ino);
}

/** Sorrento file operation table */
static struct fuse_operations sor_ops = {
    sor_getattr,
    NULL,         // readlink
    sor_getdir,
    sor_mknod,
    sor_mkdir,
    sor_unlink,
    sor_rmdir,
    NULL,         // symlink
    NULL,         // rename
    NULL,         // link
    NULL,         // chmod
    NULL,         // chown
    sor_truncate,
    sor_utime,
    sor_open,
    sor_read,
    sor_write,
    NULL,         // statfs
    sor_release,
    sor_fsync,
	sor_flush     // flush delayed write buffer
};


extern SClientStub* init_sorrento(char *);
extern void cleanup_sorrento();


inline void exiterror(char *msg)
{
	tmpi_error(DBG_APP, "%s\n", msg);
	exit(-1);
}

/** Signal handler */
static void
do_signal(int sig)
{
	switch (sig) {
	case SIGTERM:
		cleanup_sorrento();
		exiterror("shutdown due to SIGTERM.");

	case SIGSEGV:
		tmpi_error(DBG_INTERNAL, "SEG FAULT!\n");
		//cleanup_sorrento();
		exiterror("shutdown due to SIGSEGV.");

	case SIGPIPE:
		cleanup_sorrento();
		exiterror("shutdown due to SIGPIPE.");

	case SIGUSR1:
		/* debugging purpose, print out file operation stastistics */
		print_sor_stats();
		break;

	case SIGUSR2:
		/* debugging purpose, print out file operation stastistics */
		reset_sor_stats();
		break;

	default:
		cleanup_sorrento();
		tmpi_error(DBG_APP, "shutdown due to unexpected signal(%d).\n", sig);
		exit(-1);
	}
}

/** set up signal handler */
static void 
setup_sig_handler()
{
	if (signal(SIGTERM, do_signal) == SIG_ERR)
		exiterror("Could not set up SIGTERM handler");

	if (signal(SIGSEGV, do_signal) == SIG_ERR)
		exiterror("Could not set up SIGSEGV handler");

	if (signal(SIGPIPE, do_signal) == SIG_ERR)
		exiterror("Could not set up SIGPIPE handler");

	if (signal(SIGUSR1, do_signal) == SIG_ERR)
		exiterror("Could not set up SIGUSR1 handler");

	if (signal(SIGUSR2, do_signal) == SIG_ERR)
		exiterror("Could not set up SIGUSR2 handler");
}

int 
main(int argc, char *argv[])
{
	char *sor_conf = getenv(SOR_ENV);
    
	if (sor_conf == NULL) {
		printf("SORRENTO_CONF is not set.\n");
		return -1;
	}
    
	scli = init_sorrento(sor_conf);
	if (scli == NULL) {
		printf("Sorrento client initialization failed.\n");
		return -1;
	}

	/* set up signal */
	setup_sig_handler();

	sor_debug_flag = DEBUG_STATS;
	reset_sor_stats();

	/* set debug level */
	//tmpi_set_dbg_level(DBG_INFO-1);
	tmpi_set_dbg_level(DBG_WARN1-1);

    fuse_main(argc, argv, &sor_ops);
	
	
    cleanup_sorrento();
	
    return 0;
}
