// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: cbs_bf.cc,v $ - Bloom filters with incremental updating.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2002/10/08 07:32:51 $
 * Version :  $Revision: 1.7 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 
# include <string.h>
# include <stdio.h>

# include "tmpi_debug.h"
# include "cbs_bf.h"
/**
# include "bloom.h"
**/

/**
# include "guid.h"
**/


void CBS_BF_PKT::dump()
{
	tmpi_error(DBG_VERBOSE, "PKT: n_bits=%u, size=%d, lab_entry=%d, bf_pages=%d.\n", n_bits, pkt_size(), n_lab_entries, n_bf_pages);
	if (n_lab_entries>0) {
		unsigned int *lab=lab_entries();
		char buffer[655360];
		char item[20];
		snprintf(buffer, 65536, "lab entries: %u", lab[0]);
		for (int i=1; i<n_lab_entries; i++) {
			snprintf(item, 20, ", %u", lab[i]);
			strcat(buffer, item); 
		}
		tmpi_error(DBG_VERBOSE, "%s\n", buffer);
	}
	if (n_bf_pages>0) {
		unsigned int *bf_idx=bf_pg_idx();
		char buffer[655360];
		char item[20];
		snprintf(buffer, 65536, "bf pages: %u", bf_idx[0]);
		for (int i=1; i<n_bf_pages; i++) {
			snprintf(item, 20, ", %u", bf_idx[i]);
			strcat(buffer, item); 
		}
		tmpi_error(DBG_VERBOSE, "%s\n", buffer);
	}
}

unsigned int CBS_BF_PKT::pkt_size(int n_lab_ent, int n_bf_pg) 
{
	unsigned int header_len=packet_header_len();
	return header_len+n_lab_ent*sizeof(unsigned int)+
		(BF_PAGE_SIZE+sizeof(unsigned int))*n_bf_pg;
}


bool CBS_BF_PKT::fit_packet
(unsigned int pkt_len, int &n_lab_ent, int &n_bf_pg)
{
	if ((n_lab_ent==-1)&&(n_bf_pg==-1))
		return false;
	if ((n_lab_ent!=-1)&&(n_bf_pg!=-1))
		return false;

	unsigned int header_len=packet_header_len();
	if (pkt_len<header_len)
		return false;
	unsigned int dat_len=pkt_len-header_len;
	if (n_lab_ent==-1) {
		unsigned int bf_pg_dat_sz=(BF_PAGE_SIZE+sizeof(unsigned int))*n_bf_pg;
		if (dat_len<bf_pg_dat_sz)
			return false;
		n_lab_ent=(int)((dat_len-bf_pg_dat_sz)/sizeof(unsigned int));
		return true;
	}
	else {
		unsigned int lab_ent_sz=sizeof(unsigned int)*n_lab_ent;
		if (dat_len<lab_ent_sz)
			return false;
		n_bf_pg=(int)(dat_len-lab_ent_sz)/(BF_PAGE_SIZE+sizeof(unsigned int));
		return true;
	}
}

bool CBS_BaseBF::update_from_packet(CBS_BF_PKT *pkt)
{
	if (!pkt)
		return false;

	if (pkt->n_bits != fsize)
		return false;

	bool ret=true;
	ret = ret && update_bits(pkt->lab_entries(), (int)pkt->n_lab_entries);
# ifndef __NO_CHUNK_UPDATE__
	ret = ret && update_blocks(pkt->bf_pages(), pkt->bf_pg_idx(), (int)pkt->n_bf_pages);
# endif
	return ret;
}

char *CBS_BF::bit_flip_policy=
# ifdef __OVERLAP_UPDATE__
	"overlapped bit flipping history";
# else
	"un-overlapped bit flipping history";
# endif

char *CBS_BF::chunk_update_policy=
# ifdef __NO_CHUNK_UPDATE__
	"no chunked update";
# elif defined(__CONST_FREQ_UPDATE__)
	"constant frequency update";
# else
	"variable frequency update";
# endif

bool CBS_BF::make_bf_update_packet(CBS_BF_PKT *pkt, unsigned short &pkt_len)
{
	if (!pkt)
		return false;

	pkt->n_bits=fsize;

	// half of the data part is used to store LAB entries.
	int max_lab_entries=
		((pkt_len-pkt->packet_header_len())/2/sizeof(unsigned int));

	int iLabEntries=get_usable_lab_size();
# ifdef __OVERLAP_UPDATE__
	iLabEntries+=n_overlap_lab_entries;
# endif

	if (iLabEntries>max_lab_entries)
		iLabEntries=max_lab_entries;

	int iBFPages=-1;
	if (CBS_BF_PKT::
		fit_packet(pkt_len, iLabEntries, iBFPages)==false)
		return false;
	iLabEntries=-1;
	if (CBS_BF_PKT::
		fit_packet(pkt_len, iLabEntries, iBFPages)==false)
		return false;

	pkt->n_lab_entries=(unsigned short)iLabEntries;
	pkt->n_bf_pages=(unsigned short)iBFPages;

# if defined(__DEBUG__)
	unsigned int cpkt_len=pkt->pkt_size();
	if (cpkt_len>pkt_len)
		return false;
# endif

	int ret=0;
# ifndef __OVERLAP_UPDATE__
	ret=get_lab_entries(pkt->lab_entries(), pkt->n_lab_entries);
	if (ret<0)
		return false;
	pkt->n_lab_entries=ret;
# else
	unsigned int *pkt_lab_buf=pkt->lab_entries();
	unsigned int pkt_lab_buf_size=pkt->n_lab_entries;
	unsigned int overlap=(n_overlap_lab_entries>pkt_lab_buf_size)?
		pkt_lab_buf_size:n_overlap_lab_entries;

	pkt->n_lab_entries=0;
	if (overlap>0) {
		memcpy(pkt_lab_buf, overlap_lab_entries, overlap*sizeof(int));
		pkt_lab_buf+=overlap;
		pkt_lab_buf_size-=overlap;
		pkt->n_lab_entries+=overlap;
		n_overlap_lab_entries-=overlap;
		if (n_overlap_lab_entries>0) {
			memmove(overlap_lab_entries, 
					overlap_lab_entries+overlap, 				
					n_overlap_lab_entries*sizeof(int));
		}
	}
	if (pkt_lab_buf_size>0) {
		ret=get_lab_entries(pkt_lab_buf, pkt_lab_buf_size);
		if (ret<0)
			return false;
		memcpy(overlap_lab_entries+n_overlap_lab_entries, 
				pkt_lab_buf, sizeof(int)*ret);
		n_overlap_lab_entries+=ret;
		pkt->n_lab_entries+=ret;
	}
# endif

# if defined(__DEBUG__)
	cpkt_len=pkt->pkt_size();
	if (cpkt_len>pkt_len)
		return false;
# endif

# ifdef __CONST_FREQ_UPDATE__
	ret=get_pages(pkt->bf_pages(), next_chunk_id, pkt->n_bf_pages);
	if (ret<0)
		return false;
	unsigned int *pkt_idx_buf=pkt->bf_pg_idx();
	for (int i=0; i<ret; i++, next_chunk_id++) {
		next_chunk_id=next_chunk_id%msize;
		pkt_idx_buf[i]=next_chunk_id;
	}
	next_chunk_id=next_chunk_id%msize;
# else
	ret=get_mod_pages(pkt->bf_pages(), pkt->bf_pg_idx(), pkt->n_bf_pages);
	if (ret<0)
		return false;
# endif

	char *src=(char *)pkt->bf_pg_idx();
	pkt->n_bf_pages=(unsigned short)ret;
	char *dst=(char *)pkt->bf_pg_idx();
	if ((src != dst)&&(pkt->n_bf_pages>0)) {
		memmove(dst, src, sizeof(unsigned int)*pkt->n_bf_pages);
	}

# if defined(__DEBUG__)
	cpkt_len=pkt->pkt_size();
	if (cpkt_len>pkt_len)
		return false;
	pkt_len=(unsigned short)cpkt_len;
# else
	pkt_len=(unsigned short)pkt->pkt_size();
# endif

	return true;
}
