// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: scsd.cc,v $ - provider daemon.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/05/06 09:19:22 $
 * Version :  $Revision: 1.20 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 
# include <stdio.h>
# include <ctype.h>
# include <string.h>
# include <getopt.h>
# include <signal.h>

# include "app.h"
# include "tmpi_debug.h"
# include "sock_sos.h"
# include "resmon.h"
# include "load_announcer.h"
# include "cbs_bf.h"
# include "udp_srv.h"
# include "../shared/daemon_common.h"

struct Scsdopt {
public:
	char *home;
	char *cfg;
	char *log;
	char *mgrp;
	int port;
	int con_port;
# define MAX_MOUNTS 1024
	char *mounts[MAX_MOUNTS];
	char *rootdir[MAX_MOUNTS];
	int n_mounts;
	int annc_int;

private:
	enum {
		HOME_INIT=0x1, 
		HOME_FREE=0x2, 
		CFG_INIT=0x4, 
		CFG_FREE=0x8, 
		LOG_INIT=0x10, 
		LOG_FREE=0x20,
		MGRP_INIT=0x40, 
		MGRP_FREE=0x80, 
		PORT_INIT=0x100,
		CPRT_INIT=0x200,
		AINT_INIT=0x400 
	};

	int flag;

# define HOME_DEFAULT ""
# define CFG_DEFAULT "/etc/scsd.conf"
# define LOG_DEFAULT "/var/log/scsd.log"
# define MGRP_DEFAULT CBS_MGRP_DEFAULT
# define PORT_DEFAULT CBS_PORT_DEFAULT
# define CPRT_DEFAULT 3802
# define ANNC_INT_DEFAULT 800

	int set_var(char *s, bool cpref, bool set_init, char *&var, int init_mask, int free_mask)
	{
		if (!s) {
			tmpi_error(DBG_CALLER, "NULL Pointer.\n");
			return -1;
		}

		if (flag&init_mask) {
			// already initialized
			return 0;
		}

		if (strcmp(s, var)) {
			if (cpref) {
				var=s;
			}
			else {
				var=strdup(s);
				if (!var) {
					tmpi_error(DBG_FATAL, "Out of memory.\n");
					return -1;
				}
				flag|=free_mask;
			}
		}

		if (set_init) {
			flag|=init_mask;
		}

		return 0;
	}

	bool is_valid_num(char *p)
	{
		if (!p)
			return false;

		while (*p) {
			if (!isdigit(*p)) {
				return false;
			}
			p++;
		}

		return true;
	}

	int set_num(char *n, bool set_init, int &num, int init_mask)
	{
		if (!n) {
			tmpi_error(DBG_CALLER, "NULL Pointer.\n");
			return -1;
		}

		if (!is_valid_num(n)) {
			tmpi_error(DBG_CALLER, "Bad numeric format.\n");
			return -1;
		}

		if (flag&init_mask) {
			// already initialized
			return 0;
		}

		num=atoi(n);

		if (set_init) {
			flag|=init_mask;
		}

		return 0;
	}

public:
	Scsdopt() : 
		home(HOME_DEFAULT),
		cfg(CFG_DEFAULT),
		log(LOG_DEFAULT),
		mgrp(MGRP_DEFAULT),
		port(PORT_DEFAULT),
		con_port(CPRT_DEFAULT),
		n_mounts(0),
		annc_int(ANNC_INT_DEFAULT),
		flag(0)
	{
		memset(mounts, 0, sizeof(char *)*MAX_MOUNTS);
		return;
	}

	~Scsdopt() {
		if ((home)&&(flag&HOME_FREE)) {
			free(home);
		}
		if ((cfg)&&(flag&CFG_FREE)) {
			free(cfg);
		}
		if ((log)&&(flag&LOG_FREE)) {
			free(log);
		}
		if ((mgrp)&&(flag&MGRP_FREE)) {
			free(mgrp);
		}

		for (int i=0; i<n_mounts; i++) {
			if (mounts[i]) {
				free(mounts[i]);
			}
			if (rootdir[i]) {
				free(rootdir[i]);
			}
		}
	}

	int set_home(char *s, bool set_init, bool cpref)
	{
		if (!s) {
			tmpi_error(DBG_CALLER, "NULL Pointer.\n");
			return -1;
		}

		if (s[0]!='/') {
			tmpi_error(DBG_CALLER, "SCS_HOME must be an absolute path.\n");
			return -1;
		}

		if (flag&HOME_INIT) {
			// already initialized
			return 0;
		}

		if (strcmp(home, s)) {
			if (cpref) {
				home=s;
			}
			else {
				home=strdup(s);
				if (!home) {
					tmpi_error(DBG_FATAL, "Out of memory.\n");
					return -1;
				}
				flag|=HOME_FREE;
			}
		}

		if (set_init) {
			flag|=HOME_INIT;
		}

		// adjust the default cfg directory
		if (!(flag&CFG_INIT)) {
			char new_cfg[1024];
			strncpy(new_cfg, home, 1024);
			strncat(new_cfg, cfg, 1024);
			set_cfg(new_cfg, false, false);
		}

		// adjust the default log directory
		// strip out the "/var" of the default log path.
		if (!(flag&LOG_INIT)) {
			char new_log[1024];
			char *log1=log;
			strncpy(new_log, home, 1024);
			if (strncmp(log, "/var", 4)==0) {
				log1+=4;
			}
			strncat(new_log, log1, 1024);
			set_log(new_log, false, false);
		}

		return 0;
	}

	int set_cfg(char *s, bool set_init, bool cpref)
	{
		if (s[0]!='/') {
			tmpi_error(DBG_CALLER, "Configuration file must start with an absolute path.\n");
			return -1;
		}

		return set_var(s, cpref, set_init, cfg, CFG_INIT, CFG_FREE);
	}

	int set_log(char *s, bool set_init, bool cpref)
	{
		if (s[0]!='/') {
			tmpi_error(DBG_CALLER, "Log file must start with an absolute path.\n");
			return -1;
		}

		return set_var(s, cpref, set_init, log, LOG_INIT, LOG_FREE);
	}
	
	int set_mgrp(char *s, bool set_init, bool cpref)
	{
		return set_var(s, cpref, set_init, mgrp, MGRP_INIT, MGRP_FREE);
	}

	int set_aint(char *p, bool set_init=true, bool dummy=true)
	{
		return set_num(p, set_init, annc_int, AINT_INIT);
	}

	int set_port(char *p, bool set_init=true, bool dummy=true)
	{
		return set_num(p, set_init, port, PORT_INIT);
	}

	int set_cprt(char *p, bool set_init=true, bool dummy=true)
	{
		return set_num(p, set_init, con_port, CPRT_INIT);
	}
	
	int add_mount(char *mnt_ptr, bool dummy1=true, bool dummy2=true)
	{
		if (!mnt_ptr) {
			tmpi_error(DBG_CALLER, "NULL Point.\n");
			return -1;
		}

		char buf[1024];

		if (ResMon::find_mount_point(mnt_ptr, buf, 1024)<0) {
			tmpi_error(DBG_CALLER, "Invalid Mount Point.\n");
			return -1;
		}

		for (int i=0; i<n_mounts; i++) {
			if (strcmp(buf, mounts[i])==0) {
				tmpi_error(DBG_VERBOSE, "Mount Point Already Added.\n");
				return 0;
			}
		}

		if (n_mounts<MAX_MOUNTS) {
			if (!(mounts[n_mounts]=strdup(buf))) {
				tmpi_error(DBG_INTERNAL, "Out of memory.\n");
				return -1;
			}
			if (!(rootdir[n_mounts]=strdup(mnt_ptr))) {
				tmpi_error(DBG_INTERNAL, "Out of memory.\n");
				free(mounts[n_mounts]);
				mounts[n_mounts]=NULL;
				return -1;
			}
			n_mounts++;
			return 0;
		}

		tmpi_error(DBG_INTERNAL, "Internal mount table full.\n");
		return -1;
	}

	// for configuration file parsing
	int set_var_val(char *var, char *val)
	{
		static struct look_table {
			char *var;
			int (Scsdopt::*f)(char *, bool, bool);
		} *p, table[] = {
			{"SCS_HOME", &Scsdopt::set_home},
			{"LOG", &Scsdopt::set_log},
			{"Multicast", &Scsdopt::set_mgrp},
			{"CONPORT", &Scsdopt::set_cprt},
			{"PORT", &Scsdopt::set_port},
			{"PARTITION", &Scsdopt::add_mount},
			{"ANNC_INT", &	Scsdopt::set_aint},
			{NULL, NULL}
		};

		p=table;
		while (p->var) {
			if (strcasecmp(p->var, var)==0) {
				return (this->*(p->f))(val, true, false);
			}
			p++;
		}
		return -1;
	}
};

static char *chop(char *line)
{
	while (*line) {
		if (isspace(*line))
			line++;
		else
			break;
	}

	char *p=line+strlen(line)-1;
	while ((p>=line)&&(isspace(*p))) {
		*p='\0';
		p--;
	}

	return line;
}

static char *program_name=NULL;
static char *get_RCS_val(char *var)
{
	static char buf[1024];
	char *p;

	if ((p=strchr(var, ':'))==NULL) {
		return var;
	}
	buf[1023]='\0';
	strncpy(buf, p+1, 1024);
	p=buf+strlen(buf)-1;
	while ((p>=buf)&&(*p=='$')) {
		*p='\0';
		p--;
	}

	return chop(buf);
}

static void usage(int status)
{
	if (status != 0) {
		fprintf(stderr, "Try '%s --help' for more information.\n", program_name);
	}
	else {
		printf ("Usage: %s [options]\n", program_name);
		printf ("\
Starting the Storage Cluster Server (SCS) daemon.\n\
   -r, --home <scs_home>     Set SCS root directory.\n\
   -l, --log <log_file>      Set SCS log file.\n\
   -c, --config <cfg_file>   Set SCS configuration file.\n\
   -p, --port <port>         Set SCS service port.\n\
   -q, --conport <con_port>  Set SCS service daemon control port.\n\
   -g, --multicast <group>   Set SCS service multicast group.\n\
   -i, --interval <interval> Set SCS service multicast interval.\n\
   -m, --partition <mnt_ptr> Let SCS service manage the partition.\n\
   -s, --silent              Display/log only fatal information.\n\
   -d, --debug               Keep a detailed log.\n\
   -h, --help                Display this help screen.\n\
   -V, --version             Display version number.\n\
");
	}

	exit(status);
}

char *Id="$Id: scsd.cc,v 1.20 2003/05/06 09:19:22 htang Exp $";

static void do_version()
{
	static char *version="$Revision: 1.20 $";
	static char *author="Hong Tang";
	static char *date="$Date: 2003/05/06 09:19:22 $";

	{
		printf("Storage Cluster Server (SCS) Daemon (%s) %s.\n", program_name, get_RCS_val(version));
		printf("Copyright (c) University of California, Santa Barbara.\n");
		printf("Author: %s.\n", get_RCS_val(author));
		printf("Last Modification Date: %s GMT.\n", get_RCS_val(date));
		printf("Specify the --help option for further information.\n");
	}

	exit(0);
}

/**
static struct SCS_STATE {
	Scsdopt *opt; // parameters.
	CBS_BF *ss; // local bloom filter.
	StoreHash *sh; // persistent hash table for all the objects.
	FileCache *fc; // file cache.
} scs_state;
**/

static LoadAnnouncer *la=NULL;
static Udp_srv *srv=NULL;

static int init_servers(Scsdopt *scsdopt)
{
	// start myself as a daemon.
	srv=new Udp_srv(NULL, scsdopt->con_port);

	if ((!srv)||(!srv->good())) {
		tmpi_error(DBG_FATAL, "Cannot start UDP server.\n");
		return -1;
	}

	if (scsdopt->con_port==0) {
		scsdopt->con_port=srv->port();
	}

	// start servers
	la=new LoadAnnouncer(scsdopt->mgrp, scsdopt->port, scsdopt->mounts, scsdopt->n_mounts, scsdopt->con_port, scsdopt->annc_int);

	if ((!la)||(!la->good())) {
		tmpi_error(DBG_FATAL, "Cannot init load announcer.\n");
		return -1;
	}

	return 0;
}

static int start_servers()
{
	if (la) {
		if (!la->start()) {
			tmpi_error(DBG_FATAL, "Cannot run load announcer.\n");
			return -1;
		}
	}
	else {
		return -1;
	}
	
	return 0;
}

static void shutdown_servers()
{
	// shutdown servers
	if (la) {
		la->stop();
	}

	if (srv) {
		srv->stop();
	}
}

static void finalize_servers()
{
	if (la) {
		la->finalize();
		delete la;
		la=NULL;
	}

	if (srv) {
		delete srv;
		srv=NULL;
	}
}

static unsigned int server_stop=0;
static void stop_server(int signum)
{
	tmpi_error(DBG_VERBOSE, "Got signal %d.\n", signum);
	server_stop|=APP_SYS_QUIT;
	commit32(&server_stop);
	shutdown_servers();
	signal(signum, &stop_server);
}

static void set_signals()
{
	// set ignore
	static const int ignore_signals[] = {
		SIGHUP,
		SIGPIPE,
		SIGALRM,
		SIGUSR1,
		SIGUSR2,
		SIGTTIN,
		SIGTTOU,
		SIGURG,
		SIGIO,
		SIGPOLL,
		SIGVTALRM,
		SIGPROF,
		-1
	};

	// set signals
	static const int stop_signals[] = {
		SIGINT,
		SIGQUIT,
		SIGILL,
		SIGTRAP,
		SIGABRT,
		SIGIOT,
		SIGFPE,
		SIGBUS,
		SIGSEGV,
		SIGSTKFLT,
		SIGTERM,
		SIGXCPU,
		SIGXFSZ,
		SIGPWR,
		-1
	};

	static const int default_signals[] = {
		SIGSTOP,
		SIGCONT,
		SIGTSTP,
		SIGCHLD,
		SIGCLD,
		-1
	};

	const int *p=ignore_signals;
	while (*p>=0) {
		signal(*p, SIG_IGN);
		p++;
	}

	p=stop_signals;
	while (*p>=0) {
		signal(*p, stop_server);
		p++;
	}
	
	p=default_signals;
	while (*p>=0) {
		signal(*p, SIG_DFL);
		p++;
	}
}

static void parse_cmd_opt(int argc, char *argv[], Scsdopt *scsdopt)
{
	static struct option const long_options[] = {
		{"home", required_argument, 0, 'r'},
		{"log", required_argument, 0, 'l'},
		{"config", required_argument, 0, 'c'},
		{"port", required_argument, 0, 'p'},
		{"conport", required_argument, 0, 'q'},
		{"multicast", required_argument, 0, 'g'},
		{"interval", required_argument, 0, 'i'},
		{"partition", required_argument, 0, 'm'},
		{"help", no_argument, 0, 'h'},
		{"silent", no_argument, 0, 's'},
		{"debug", no_argument, 0, 'd'},
		{"version", no_argument, 0, 'V'},
		{NULL, 0, 0, 0}
	};

	int optc;

	while ((optc=getopt_long(argc, argv, \
					"r:l:c:p:q:g:i:m:sdhV", long_options, NULL)) !=-1) {
		switch (optc) {
			case 0:
				break;
			case 'r':
				if (scsdopt->set_home(optarg, true, true)<0) {
					usage(0);
				}
				break;
			case 'l':
				if (scsdopt->set_log(optarg, true, true)<0) {
					usage(0);
				}
				break;
			case 'c':
				if (scsdopt->set_cfg(optarg, true, true)<0) {
					usage(0);
				}
				break;
			case 'p':
				if (scsdopt->set_port(optarg, true)<0) {
					usage(0);
				}
				break;
			case 'q':
				if (scsdopt->set_cprt(optarg, true)<0) {
					usage(0);
				}
				break;
			case 'g':
				if (scsdopt->set_mgrp(optarg, true, true)<0) {
					usage(0);
				}
				break;
			case 'i':
				if (scsdopt->set_aint(optarg, true, true)<0) {
					usage(0);
				}
				break;
			case 'm':
				if (scsdopt->add_mount(optarg)<0) {
					usage(0);
				}
				break;
			case 'h':
				usage(0);
				break;
			case 'V':
				do_version();
				break;
			case 's':
				tmpi_set_dbg_level(DBG_FATAL-1);
				tmpi_hide_location();
				break;
			case 'd':
				tmpi_set_dbg_level(DBG_ALL);
				tmpi_show_location();
				break;
			default:
				usage(1);
		}
	}
}

static bool is_empty_line(char *line)
{
	while (*line) {
		if (!isspace(*line))
			return false;
		line++;
	}

	return true;
}

static int parse_config(Scsdopt *scsdopt)
{
	if ((!scsdopt)||(!scsdopt->cfg))
		return -1;

	FILE *fp=fopen(scsdopt->cfg, "r");

	if (!fp) {
		tmpi_error(DBG_CALLER, "Cannot open configuration file %s.\n", scsdopt->cfg);
		return -1;
	}

	char buf[1024];
	char *var, *val;
	int lineno=0;

	while (fgets(buf, 1024, fp)) {
		char *p;

		lineno++;
		if ((p=strchr(buf, '#'))) {
			*p='\0';
		}

		if (is_empty_line(buf)) {
			continue;
		}
		
		if ((p=strchr(buf, '='))) {
			*p='\0';
			var=chop(buf);
			val=chop(p+1);
			if (scsdopt->set_var_val(var, val)<0) {
				tmpi_error(DBG_VERBOSE, "Error encountered @ config file line %d.\n", 
						lineno);
			}
		}
		else {
			tmpi_error(DBG_CALLER, "Bad config file format @ line %d.\n", lineno);
		}
	}

	return 0;
}

static void	dump_scsdopt(Scsdopt *scsdopt)
{
	printf("  SCS_HOME=\"%s\".\n", scsdopt->home);
	printf("  Configuration file=\"%s\".\n", scsdopt->cfg);
	printf("  Log file=\"%s\".\n", scsdopt->log);
	printf("  Multicast group=\"%s\".\n", scsdopt->mgrp);
	printf("  Port=%d\n", scsdopt->port);
	printf("  Multicast interval=%dms.\n", scsdopt->annc_int);
	printf("  Control port=%d\n", scsdopt->con_port);
	if (scsdopt->n_mounts) {
		printf("  Broadcast disk info on %d anc->mounts:\n", scsdopt->n_mounts);
		for (int i=0; i<scsdopt->n_mounts; i++) {
			printf("    [%d] %s.\n", i+1, /* scsdopt->mounts[i], */ scsdopt->rootdir[i]);
		}
	}
}

int main(int argc, char *argv[])
{
	program_name=argv[0];

# if defined(__DEBUG__)
	tmpi_set_dbg_level(DBG_ALL);
	tmpi_show_location();
# else
	tmpi_set_dbg_level(DBG_VERBOSE);
	tmpi_hide_location();
# endif

	struct Scsdopt *scsdopt=new Scsdopt();
	parse_cmd_opt(argc, argv, scsdopt);

	{
		// read configuration from file.
		char buf[1024];
		char *p;

		if ((p=getenv("SCS_HOME"))!=NULL) {
			buf[1023]='\0';
			strncpy(buf, p, 1024);
			if (scsdopt->set_home(buf, true, false)<0) {
				usage(0);
			}
		}

		if (parse_config(scsdopt)<0) {
			tmpi_error(DBG_INTERNAL, "No configuration file.\n");
			// usage(0);
		}
		else {
			tmpi_error(DBG_VERBOSE, "Configuration file %s parsed.\n", scsdopt->cfg);
		}
	}

	if (tmpi_debug(DBG_VERBOSE)) {
		tmpi_error(DBG_VERBOSE, "SCSD setting:\n");
		dump_scsdopt(scsdopt);
	}

	int ret=0;
	FILE *fp=NULL;

	{
		// set log
		fp=fopen(scsdopt->log, "a+");
		if (!fp) {
			tmpi_error(DBG_CALLER, "Cannot open log file %s.\n", scsdopt->log);
			// usage(0);
		}
		else {
			tmpi_error(DBG_APP, "Redirecting log output to %s\n", scsdopt->log);
			tmpi_set_output(fp);
		}
	}

	tmpi_error(DBG_APP, "Preparing to set SOS pointer.\n");
	if (sock_set_sos_ptr(&server_stop)<0) {
		tmpi_error(DBG_INTERNAL, "Cannot set application SOS flag.\n");
		ret=-1;
		goto _exit;
	}

	tmpi_error(DBG_APP, "Initializing servers ...\n");
	if (init_servers(scsdopt)<0) {
		ret=-1;
		goto _exit;
	}

	tmpi_error(DBG_APP, "Set signal handlers ...\n");
	set_signals();

	tmpi_error(DBG_APP, "Starting servers ...\n");

	if (start_servers()<0) {
		ret=-1;
		goto _exit;
	}

	tmpi_error(DBG_APP, "%s running now at %s:%d::%d.\n", program_name, \
			scsdopt->mgrp, scsdopt->port, scsdopt->con_port);

	delete scsdopt; // no more usage.
	scsdopt=NULL;

	while (1) {
		Udp_srv_req *q=srv->wait_for_request();

		if (server_stop) {
			tmpi_error(DBG_APP, "Server going to shut down.\n");
			if (q)
				delete q;

			break;
		}

		if (!q) {
			tmpi_error(DBG_INTERNAL, "Cannot get request, server stop.\n");
			break;
		}

		if (q->is_stop_req()) {
			tmpi_error(DBG_APP, "Got shutdown request, server stop.\n");
			delete q;
			break;
		}

		tmpi_error(DBG_VERBOSE, "Got unhandled request.\n");
		delete q;
	}

_exit:

	tmpi_error(DBG_APP, "Sending signals to servers...\n");
	shutdown_servers();

	tmpi_error(DBG_APP, "Waiting for servers to stop...\n");
	finalize_servers();

	tmpi_error(DBG_APP, "Server stopped.\n");

	if (fp) {
		fclose(fp);
	}

	if (scsdopt)
		delete scsdopt;

	return ret;
}
