// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.
 
  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
 
   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.
 
   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
 
   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.
 
   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.
 
  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: client.cc,v $ - Interface to send requests to the 
              server.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Aziz Gulbeden, gulbeden@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2004/02/19 13:14:51 $ by $Author: gulbeden $
 * Version :  $Revision: 1.28 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu, gulbeden@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */

# include <unistd.h>			/* sleep */
# include <vector>
# include <stdio.h>
# include <iostream>
# include <string>

# include "protocol.h"			/* protocol that server and client talk */
# include "client.h"
# include "tmpi_debug.h"
# include "sock_conn.h"

using namespace std;

# define STG_CLI_UDP_TOUT 10000
/**
 * costructor - starts listening.
 * @throws an error message if construction fails
 */
Client::Client()
{
	cli=new Udp_cli(NULL, SOCK_PORT_ANY, STG_CLI_UDP_TOUT);

	if (!cli) {
		throw string("Client::Client: Cannot create a UDP client.");
	}
}

/**
 * appends data at the end.
 * @return the value returned by the remote file system.
 */
int Client::append(Descriptor &desc, char *buffer, off_seg_t length, \
	off_seg_t &append_bytes, int max_redir_loop)
{
	while (1) {
		Stg_Request *req=new_stg_req_append(desc.desc, buffer, length);
		if (!req)
			return STG_RES_EMEM;
		
		Stg_Req_Handle reqh;
		int ret=stg_start_request_with_payload(cli, desc.saddr, req, buffer, length, reqh, STG_USE_TCP);
		stg_req_free(req);

		if (ret != STG_RES_OKAY) {
			return ret;
		}

		Stg_Res_Handle resh;
		ret=stg_wait_response_no_payload(cli, reqh, resh);
		if (ret != STG_RES_OKAY) {
			return ret;
		}

		Stg_Response *res=resh.response();
		if (res->header.tag == STG_RES_REDIR2) {
			// redirection.
			if (max_redir_loop==0) {
				return STG_RES_ENOTFOUND;
			}
			desc=Descriptor(res->u.u_redir2.saddr, res->u.u_redir2.desc);
			max_redir_loop--;
			continue;
		}

		if (res->header.tag != STG_RES_APPEND) {
			tmpi_error(DBG_VERBOSE, "Unexpected reply.\n");
			return STG_RES_ECLIENT;
		}

		append_bytes = res->u.u_append.append_bytes;

		return STG_RES_OKAY;
	}

	// unreachable.
	return STG_RES_ECLIENT;
}

/**
 * Sends close request.
 */
int Client::close(Descriptor &desc, int max_redir_loop)
{
	while (1) {
		Stg_Request *req=new_stg_req_close(desc.desc, 0);
		if (!req) {
			tmpi_error(DBG_VERBOSE, "new_stg_req_close failed.\n");
			return STG_RES_EMEM;
		}
		
		Stg_Req_Handle reqh;
		int ret=stg_start_request_no_payload(cli, desc.saddr, req, reqh, STG_USE_TCP);
		stg_req_free(req);

		if (ret != STG_RES_OKAY) {
			tmpi_error(DBG_VERBOSE, "start_request failed (%d)\n", ret);
			return ret;
		}

		Stg_Res_Handle resh;
		ret=stg_wait_response_no_payload(cli, reqh, resh);
		if (ret != STG_RES_OKAY) {
			tmpi_error(DBG_VERBOSE, "wait_response failed (%d)\n", ret);
			return ret;
		}

		Stg_Response *res=resh.response();

		if (res->header.tag == STG_RES_REDIR2) {
			// redirection.
			if (max_redir_loop==0) {
				return STG_RES_ENOTFOUND;
			}
			desc=Descriptor(res->u.u_redir2.saddr, res->u.u_redir2.desc);
			max_redir_loop--;
			continue;
		}

		if (res->header.tag != STG_RES_CLOSE) {
			tmpi_error(DBG_VERBOSE, "Unexpected reply.\n");
			return STG_RES_ECLIENT;
		}

		return STG_RES_OKAY;
	}

	return STG_RES_ECLIENT;
}

/**
 * Creates the object.
 * @return descriptor to the opened file. Negative if open fails.
 */
int Client::create(ServerAddr &saddr_in, ObjID &oid, Descriptor &desc, int max_redir_loop)
{
	ServerAddr saddr=saddr_in; // create a local copy.

	while (1) {
		Stg_Request *req=new_stg_req_create1(oid);
		if (!req)
			return STG_RES_EMEM;
		
		Stg_Req_Handle reqh;
		int ret=stg_start_request_no_payload(cli, saddr, req, reqh, STG_USE_TCP);
		stg_req_free(req);

		if (ret != STG_RES_OKAY) {
			return ret;
		}

		Stg_Res_Handle resh;
		ServerAddr saddrs[64];	// I am assuming there can be at most 64 
								// replicas, this array will be filled in 
								// case of a redirection.

		ret=stg_wait_response_with_payload(cli, reqh, resh, (char *) saddrs, 64*sizeof(ServerAddr));

		if (ret == STG_RES_ESIZE) {
			tmpi_error(DBG_VERBOSE, "STG_RES_ESIZE (%d).\n", ret);
			return ret;
		}

		if (ret==STG_RES_EBUFFER) {
			tmpi_error(DBG_VERBOSE, "STG_RES_EBUFFER (%d).\n", ret);
			return ret;
		}

		if (ret != STG_RES_OKAY) {
			tmpi_error(DBG_VERBOSE, "CREATE FAILED (%d).\n", ret);
			return ret;
		}

		Stg_Response *res=resh.response();

		if (res->header.tag == STG_RES_REDIR1_SMALL) {
			// redirection.
			if (max_redir_loop==0) {
				return STG_RES_ENOTFOUND;
			}
			saddr=saddrs[0];	// FORWARD TO THE FIRST SERVER SENT IN
								// RESPONSE
			max_redir_loop--;
			continue;
		}

		else if (res->header.tag == STG_RES_REDIR1_BIG) {
			// redirection.
			if (max_redir_loop==0) {
				return STG_RES_ENOTFOUND;
			}

			saddr=saddrs[0];	// FORWARD TO THE FIRST SERVER SENT IN
								// RESPONSE
			max_redir_loop--;
			continue;
		}

		if (res->header.tag != STG_RES_CREATE) {
			tmpi_error(DBG_VERBOSE, "Unexpected reply.\n");
			return STG_RES_ECLIENT;
		}

		desc = Descriptor(saddr, res->u.u_create.desc);

		return STG_RES_OKAY;
	}

	return STG_RES_ECLIENT;
}

/**
 * Opens object for read/write/append
 * @return descriptor to the opened file. NULL if open fails.
 */
int Client::open(ServerAddr &saddr_in, ObjID &oid, Descriptor &desc, int flags, int max_redir_loop)
{
	ServerAddr saddr=saddr_in; // create a local copy.

	while (1) {
		Stg_Request *req=new_stg_req_open(oid, flags);
		if (!req)
			return STG_RES_EMEM;
		
		Stg_Req_Handle reqh;
		int ret=stg_start_request_no_payload(cli, saddr, req, reqh, STG_USE_TCP);
		stg_req_free(req);

		if (ret != STG_RES_OKAY) {
			return ret;
		}

		Stg_Res_Handle resh;
		ServerAddr saddrs[64];	// I am assuming there can be at most 64 
								// replicas, this array will be filled in 
								// case of a redirection.

		// res_open res is without payload, res_redir is with payload, 
		// therefore wait for "with_payload"
		ret=stg_wait_response_with_payload(cli, reqh, resh, (char *) saddrs, 64*sizeof(ServerAddr));

		if (ret == STG_RES_ESIZE) {
			tmpi_error(DBG_VERBOSE, "STG_RES_ESIZE (%d).\n", ret);
			return ret;
		}

		if (ret==STG_RES_EBUFFER) {
			tmpi_error(DBG_VERBOSE, "STG_RES_EBUFFER (%d).\n", ret);
			return ret;
		}

		if (ret != STG_RES_OKAY) {
			tmpi_error(DBG_VERBOSE, "OPEN FAILED (%d).\n", ret);
			return ret;
		}

		Stg_Response *res=resh.response();

		if (res->header.tag == STG_RES_REDIR1_SMALL) {
			// redirection.
			if (max_redir_loop==0) {
				return STG_RES_ENOTFOUND;
			}
			saddr=saddrs[0];	// FORWARD TO THE FIRST SERVER SENT IN
								// RESPONSE
			max_redir_loop--;
			continue;
		}
		
		else if (res->header.tag == STG_RES_REDIR1_BIG) {
			// redirection.
			if (max_redir_loop==0) {
				return STG_RES_ENOTFOUND;
			}
			saddr=saddrs[0];	// FORWARD TO THE FIRST SERVER SENT IN
								// RESPONSE
			max_redir_loop--;
			continue;
		}

		if (res->header.tag != STG_RES_OPEN) {
			tmpi_error(DBG_VERBOSE, "Unexpected reply.\n");
			return STG_RES_ECLIENT;
		}

		desc = Descriptor(saddr, res->u.u_open.desc);

		return STG_RES_OKAY;
	}

	return STG_RES_ECLIENT;
}

/**
 * Reads the requested part.
 * @return the value returned by the remote file system.
 */
int Client::read(Descriptor &desc, off_seg_t offset, char *buffer, off_seg_t length, off_seg_t &read_bytes, int max_redir_loop)
{
	int hint=(length>STG_MAX_UDP_SIZE)?STG_USE_TCP:0;

	while (1) {
		Stg_Request *req=new_stg_req_read(desc.desc, offset, length);
		if (!req)
			return STG_RES_EMEM;
		
		Stg_Req_Handle reqh;
		int ret=stg_start_request_no_payload(cli, desc.saddr, req, reqh, hint);
		stg_req_free(req);

		if (ret != STG_RES_OKAY) {
			return ret;
		}

		Stg_Res_Handle resh;
		ret=stg_wait_response_with_payload(cli, reqh, resh, buffer, length);
		if ((ret == STG_RES_ESIZE) && (!reqh.use_tcp())) {
			hint=STG_USE_TCP;
			continue;
		}

		// special handling for buffer-too-small error.
		// should never happen though.
		if (ret==STG_RES_EBUFFER) {
			// store the actual size into read_bytes.
			read_bytes=resh.response()->payload_size();
			return ret;
		}

		if (ret != STG_RES_OKAY) {
			return ret;
		}

		Stg_Response *res=resh.response();

		if (res->header.tag == STG_RES_REDIR2) {
			// redirection.
			if (max_redir_loop==0) {
				return STG_RES_ENOTFOUND;
			}
			desc=Descriptor(res->u.u_redir2.saddr, res->u.u_redir2.desc);
			max_redir_loop--;
			continue;
		}

		if ( (res->header.tag != STG_RES_READ_SMALL) && 
			 (res->header.tag != STG_RES_READ_BIG) ) {
			tmpi_error(DBG_VERBOSE, "Unexpected reply.\n");
			return STG_RES_ECLIENT;
		}

		if (res->header.tag == STG_RES_READ_SMALL) {
			read_bytes=res->u.u_read_small.read_bytes;
		}
		else {
			read_bytes=res->u.u_read_big.read_bytes;
		}

		return STG_RES_OKAY;
	}

	return STG_RES_ECLIENT;
}

/**
 * Removes the object.
 * @return the value returned by the remote file system.
 */
int Client::remove(ServerAddr &saddr_in, ObjID &oid, int max_redir_loop)
{
	ServerAddr saddr=saddr_in;
	while (1) {
		Stg_Request *req=new_stg_req_remove(oid);
		if (!req)
			return STG_RES_EMEM;
		
		Stg_Req_Handle reqh;
		int ret=stg_start_request_no_payload(cli, saddr, req, reqh, STG_USE_TCP);
		stg_req_free(req);

		if (ret != STG_RES_OKAY) {
			return ret;
		}

		Stg_Res_Handle resh;
		ServerAddr saddrs[64];	// I am assuming there can be at most 64 
								// replicas, this array will be filled in 
								// case of a redirection.

		ret=stg_wait_response_with_payload(cli, reqh, resh, (char *) saddrs, 64*sizeof(ServerAddr));

		if (ret == STG_RES_ESIZE) {
			tmpi_error(DBG_VERBOSE, "STG_RES_ESIZE (%d).\n", ret);
		}

		if (ret==STG_RES_EBUFFER) {
			tmpi_error(DBG_VERBOSE, "STG_RES_EBUFFER (%d).\n", ret);
			return ret;
		}

		if (ret != STG_RES_OKAY) {
			tmpi_error(DBG_VERBOSE, "Remove FAILED (%d).\n", ret);
			return ret;
		}

		Stg_Response *res=resh.response();

		if (res->header.tag == STG_RES_REDIR1_SMALL) {
			// redirection.
			if (max_redir_loop==0) {
				return STG_RES_ENOTFOUND;
			}
			saddr=saddrs[0];	// FORWARD TO THE FIRST SERVER SENT IN
								// RESPONSE
			max_redir_loop--;
			continue;
		}
		
		if (res->header.tag == STG_RES_REDIR1_BIG) {
			// redirection.
			if (max_redir_loop==0) {
				return STG_RES_ENOTFOUND;
			}
			saddr=saddrs[0];	// FORWARD TO THE FIRST SERVER SENT IN
								// RESPONSE
			max_redir_loop--;
			continue;
		}

		if (res->header.tag != STG_RES_OKAY) {
			tmpi_error(DBG_VERBOSE, "Unexpected reply.\n");
			return STG_RES_ECLIENT;
		}

		return STG_RES_OKAY;
	}
	return STG_RES_ECLIENT;
}

/**
 * Sends stop request to the server
 */
int Client::stop_server_normal(const ServerAddr & saddr)
{
	Stg_Request *req=new_stg_req_stop();
	if (!req)
		return STG_RES_EMEM;
	
	Stg_Req_Handle reqh;
	int ret=stg_start_request_no_payload(cli, saddr, req, reqh, STG_USE_UDP);
	stg_req_free(req);

	if (ret != STG_RES_OKAY) {
		return ret;
	}

	Stg_Res_Handle resh;
	ret=stg_wait_response_no_payload(cli, reqh, resh);
	if (ret != STG_RES_OKAY) {
		return ret;
	}

	Stg_Response *res=resh.response();
	if (res->header.tag != STG_RES_OKAY) {
		tmpi_error(DBG_VERBOSE, "Unexpected reply.\n");
		return STG_RES_ECLIENT;
	}

	return STG_RES_OKAY;
}

int Client::stop_server(const ServerAddr & saddr)
{
	int ret=stop_server_normal(saddr);
	char *auth="";

	if (ret != STG_RES_OKAY) {
		// try a direct way of shutting down the server.
		ret = cli->stop_srv(saddr.ip, saddr.udp_port, auth, 1, 3);
		ret = (ret<0)?STG_RES_ECLIENT:STG_RES_OKAY;
	}

	return ret;
}

/**
 * Truncates to the specified length.
 * @return the value returned by the remote file system.
 */
int Client::truncate(Descriptor &desc, off_seg_t length, off_seg_t &new_len, int max_redir_loop)
{
	while (1) {
		Stg_Request *req=new_stg_req_truncate(desc.desc, length);
		if (!req)
			return STG_RES_EMEM;
		
		Stg_Req_Handle reqh;
		int ret=stg_start_request_no_payload(cli, desc.saddr, req, reqh, 0);
		stg_req_free(req);

		if (ret != STG_RES_OKAY) {
			return ret;
		}

		Stg_Res_Handle resh;
		ret=stg_wait_response_no_payload(cli, reqh, resh);
		if (ret != STG_RES_OKAY) {
			return ret;
		}

		Stg_Response *res=resh.response();

		if (res->header.tag == STG_RES_REDIR2) {
			// redirection.
			if (max_redir_loop==0) {
				return STG_RES_ENOTFOUND;
			}
			desc=Descriptor(res->u.u_redir2.saddr, res->u.u_redir2.desc);
			max_redir_loop--;
			continue;
		}

		if (res->header.tag != STG_RES_TRUNCATE) {
			tmpi_error(DBG_VERBOSE, "Unexpected reply, tag = %d.\n", res->header.tag);
			return STG_RES_ECLIENT;
		}

		new_len=res->u.u_truncate.new_size;

		return STG_RES_OKAY;
	}

	return STG_RES_ECLIENT;
}

/**
 *
 * @return the value returned by the remote file system.
 */
int Client::write(Descriptor &desc, off_seg_t offset, char *buffer, off_seg_t length, off_seg_t &written_bytes, int max_redir_loop)
{
	while (1) {
		Stg_Request *req=new_stg_req_write(desc.desc, offset, buffer, length);
		if (!req)
			return STG_RES_EMEM;
		
		Stg_Req_Handle reqh;
		int ret=stg_start_request_with_payload(cli, desc.saddr, req, buffer, length, reqh, 0);
		stg_req_free(req);

		if (ret != STG_RES_OKAY) {
			return ret;
		}

		Stg_Res_Handle resh;
		ret=stg_wait_response_no_payload(cli, reqh, resh);
		if (ret != STG_RES_OKAY) {
			return ret;
		}

		Stg_Response *res=resh.response();

		if (res->header.tag == STG_RES_REDIR2) {
			// redirection.
			if (max_redir_loop==0) {
				return STG_RES_ENOTFOUND;
			}
			desc=Descriptor(res->u.u_redir2.saddr, res->u.u_redir2.desc);
			max_redir_loop--;
			continue;
		}

		if (res->header.tag != STG_RES_WRITE) {
			tmpi_error(DBG_VERBOSE, "Unexpected reply.\n");
			return STG_RES_ECLIENT;
		}

		written_bytes = res->u.u_write.written_bytes;

		return STG_RES_OKAY;
	}

	return STG_RES_ECLIENT;
}
