// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*- 

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*************************************************************************/

/**
 * Name    :  $RCSfile: host_load.h,v $ - header file for host load. 
 *            The HostLoad class maintains the load information
 *            of a host's load condition.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/10/02 20:24:16 $
 * Version :  $Revision: 1.8 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

# ifndef _HOST_LOAD_H_
# define _HOST_LOAD_H_

# include <time.h>
# include <string>
# include "sorrento_types.h"
# include "sor_dll.h"
# include "sock_conn.h"

# define MAX_DISKS 16
# define EWMA_ALPHA 0.8

/**
 * We may need to extend the HostLoad class to be a
 * subclass of an item that can be kept in a double
 * link list (sorted based on the t_recv) to make 
 * speed up the detection of dead hosts.
 */

class HostLoad : public DoubleLinkListable {
 public:
	std::string ident;
	HostID hid;
	time_t t_recv; // receiving time, in seconds.
	time_t uptime; // the time when the host is up.
	ServerAddr addr; // the server's address.
	// char addr[16]; // the host's dot address (as a string)
	// int port; // the port for client requests.

	// CPU statistics
	// The calculation of these statistics are a bit different from
	// what we receive from the multicast packets.
	// cpu_comp=(cpu_user+cpu_kernel)/cpu_total.
	// cpu_io=(cpu_io)/cpu_total

	float cpu_ewma_comp; // fraction of CPU for computation (EWMA)
	float cpu_ewma_io; // fraction of CPU for I/O (EWMA)
	float cpu_comp; // fraction of CPU for computation (previous)
	float cpu_io; // fraction of CPU for I/O (previous)
	int n_cpu; // number of CPUs.
	int cpu_speed; // CPU speed in MHz.
	
	// memory statistics
	// mem_used=mused/mem_total.
	float mem_ewma_used; // fraction of memory used (EWMA).
	float mem_used; // fraction of memory used (previous).
	int mem_total; // total memory (MB)

	// swap statistics
	// swap_used=sused/mem_total.
	float swap_ewma_used; // fraction of used swap space (EWMA);
	float swap_used; // fraction of used swap space (previous);
	int swap_total; // total swap space (MB)
	
	// server queue length
	int qlen; 

	// disk statistics
	int ndisks;
	int vnodes;
	// int tcp_port;
	// int udp_port;
	DiskInfo disks[MAX_DISKS];

	bool never_expire;

	HostLoad() {}

	HostLoad(std::string &i, HostID &h) :
		ident(i),
		hid(h),
		t_recv(0),
		uptime(0),
		// port(-1), // to-be-added from publisher
		cpu_ewma_comp(0),
		cpu_ewma_io(0),
		cpu_comp(0),
		cpu_io(0),
		n_cpu(0),
		cpu_speed(0), // to-be-added from publisher
		mem_ewma_used(0),
		mem_used(0),
		mem_total(0),
		swap_ewma_used(0),
		swap_used(0),
		swap_total(0),
		qlen(0), // to-be-added from publisher
		ndisks(0),
		vnodes(32),
		// tcp_port(SOCK_PORT_ANY),
		// udp_port(SOCK_PORT_ANY),
		never_expire(false)
	{
		time(&t_recv);
	}

	// use the old load info to update the new load info
	void update(const HostLoad *old)
	{
		cpu_ewma_comp=EWMA_ALPHA*old->cpu_ewma_comp+(1-EWMA_ALPHA)*cpu_comp;
		cpu_ewma_io=EWMA_ALPHA*old->cpu_ewma_io+(1-EWMA_ALPHA)*cpu_io;
		mem_ewma_used=EWMA_ALPHA*old->mem_ewma_used+(1-EWMA_ALPHA)*mem_used;
		swap_ewma_used=EWMA_ALPHA*old->swap_ewma_used+(1-EWMA_ALPHA)*swap_used;

		return;
	}

	~HostLoad() { }
};

# endif
