// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*- 

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*************************************************************************/

/**
 * Name    :  $RCSfile: host_map.cc,v $ - host map implementation. 
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2004/04/16 23:33:06 $
 * Version :  $Revision: 1.10 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

# include "tmpi_debug.h"
# include "host_map.h"
// # include "server.h"

# define UPTO_FACTOR 3

void HostMap::synchronize()
{
	assert(m_lock.read_lock());
	if (v) delete v;
	v = new VHostLoad(id.c_str());
	for (HostMap::iterator pi=this->begin(); pi!=this->end(); pi++) {
		v->push_back(*(pi->second));
	}

	vec_ver = map_ver;
	m_lock.release();
}

void HostMap::check_synchrony() 
{
	time_t now;
	time(&now);

	// don't check too often.
	if (now==last_check)
		return;

	if (vec_ver < map_ver) {
		assert(v_lock.promote_lock());
		if (vec_ver < map_ver)
			synchronize();
		v_lock.demote_lock();
	}

	time(&now);
	last_check=now;
}

static int can_accommodate(HostLoad *hl, int size_in_kb)
{
	int ret=0;

	for (int i=0; i<hl->ndisks; i++) {
		if (size_in_kb <= hl->disks[i].bfree) {
			ret+=hl->disks[i].bfree/1024;
		}
	}

	return ret;
}

HostMap::HostMap(const char *mid) : 
	id(mid), 
	map_ver(0), 
	vec_ver(0), 
	v(NULL),
	host_ring(true),
	cbi(NULL)
{ 
	unsigned long seed=(unsigned long)time(NULL);
	rng_gen=gsl_rng_alloc(gsl_rng_default);
	gsl_rng_set(rng_gen, seed);
	pthread_mutex_init(&cb_lock, NULL);
}

HostMap::~HostMap()
{
	pthread_mutex_destroy(&cb_lock);

	if (v) delete v;
	gsl_rng_free(rng_gen);

	// release all HostLoad structures.
	for (HostMap::iterator pi=this->begin(); pi!=this->end(); pi++) {
		if (pi->second) {
			pi->second->unlink();
			delete pi->second;
		}
		// get rid of myself from the double link list.
		pi->second=NULL;
	}
}

// returns true when a new host is added, false
// when the host already exists.
bool HostMap::update(HostLoad *load)
{
	bool ret;

	assert(m_lock.write_lock());

	// check whether we should set the never_expire flag on.
	if (cbi) {
		if (cbi->is_self(load->hid)) {
			load->never_expire=true;
		}
	}

	HostMap::iterator pi=this->find(load->hid);
	if (pi == this->end()) {
		ret=true;
		this->insert(HostMap::value_type(load->hid, load));
		host_ring.add(load->hid, load->vnodes);
		if (false) {
			Guid tmp(load->hid);
			char hid_str[128];
			tmp.to_string(hid_str, 129);
			tmpi_error(DBG_VERBOSE, "Host %s added to host_ring.\n", hid_str);
		}
		if (cbi) {
			pthread_mutex_lock(&cb_lock);
			if (cbi) {
				cbi->notify_host_join(load->hid);
			}
			pthread_mutex_unlock(&cb_lock);
		}
	}
	else {
		ret=false;
		HostLoad *old_load=pi->second;
		load->update(old_load);
		pi->second=load;
		// get rid of myself from the double link list.
		old_load->unlink(); 
		delete old_load;
	}

	map_ver++;

	m_lock.release();

	return ret;
}

bool HostMap::remove(HostLoad *load)
{
	bool ret=true;

	assert(m_lock.write_lock());

	HostMap::iterator pi=this->find(load->hid);
	if (pi == this->end())
		ret=false;
	else {
		// weird things happened.
		if (pi->second != load)
			ret = false;
		else {
			host_ring.remove(load->hid);
			this->erase(pi);
			if (false) {
				Guid tmp(load->hid);
				char hid_str[128];
				tmp.to_string(hid_str, 129);
				tmpi_error(DBG_VERBOSE, "Host removed from  host_ring.\n");
			}
			if (cbi) {
				pthread_mutex_lock(&cb_lock);
				if (cbi) {
					cbi->notify_host_leave(load->hid, load->addr);
				}
				pthread_mutex_unlock(&cb_lock);
			}
			map_ver++;
		}
	}

	m_lock.release();

	return ret;
}


/**
 * Translate a host ID to the network address.
 */
bool HostMap::get_address(const HostID &hid, ServerAddr &addr)
{
	bool ret;

	assert(m_lock.read_lock());
	HostMap::iterator pi=this->find(hid);
	if (pi == this->end()) {
		ret=false;
	}
	else {
		ret=true;
		addr=pi->second->addr;
	}
	
	m_lock.release();

	return ret;
}

/**
 * Returning a set of candidate hosts that may host a certain object.
 */
bool HostMap::lookup(ObjID &guid, VecServerAddr &candidates) 
{
	assert(m_lock.read_lock());

	HostID hids[UPTO_FACTOR];
	int n_nb=host_ring.get_neighbors(hids, UPTO_FACTOR, guid);
	if (false) {
		if (n_nb < UPTO_FACTOR) {
			tmpi_error(DBG_VERBOSE, "Too few neighbors (%d) found (host_ring size is %d).\n", n_nb, host_ring.get_size());
		}
	}
	bool ret=false;
	for (int i=0; i<n_nb; i++) {
		HostMap::iterator pi=this->find(hids[i]);
		if (pi == this->end())
			continue;
		HostLoad *hl=pi->second;
		candidates.push_back(hl->addr);
		ret=true;
	}

	m_lock.release();

	return ret;
}

/**
 * Find a set of hosts that are suitable for keeping a certain object.
 */
bool HostMap::place(ObjID &guid, off_seg_t seg_size, VecServerAddr &candidates)
{
	off_seg_t seg_size_in_kb = (seg_size)/1024+1;
	bool ret=false;

	assert(m_lock.read_lock());

	/** we seek to find 3 candidates. */
	int n_cand=3;

	/**
	 * First we try to see if the home host is able
	 * to be able to accommodate the object.
	 */
	HostID home=Guid::GUID_NIL;
	int n_nb=host_ring.get_neighbors(&home, 1, guid);
	if (n_nb>0) {
		HostMap::iterator pi=this->find(home);
		if (pi != this->end()) {
			HostLoad *hl=pi->second;
			if (can_accommodate(hl, (int)(seg_size_in_kb*3))) {
				candidates.push_back(hl->addr);
				n_cand--;
				ret=true;
			}
		}
	}

	/**
	 * Temporary structure used to do host selection.
	 */
	struct Selector {
		int my_cap;
		HostLoad *hl;
	};

	Selector selectors[this->size()];
	int n_selectors = 0;
	int total_cap = 0;

	/**
	 * get all hosts that can host the object and
	 * put them into the selectors array.
	 */
	for (HostMap::const_iterator pi=this->begin();
			pi != this->end(); pi++) {
		// do not compare with the good home host.
		if (memcmp(&(pi->second->hid), &home, sizeof(HostID))==0)
			continue;

		int my_cap;
		if ((my_cap=can_accommodate(pi->second, seg_size_in_kb))) {
			selectors[n_selectors].my_cap=my_cap;
			total_cap+=my_cap;
			selectors[n_selectors].hl=pi->second;
			n_selectors++;
		}
	}

	if (n_cand>=n_selectors) {
		// just dump everything into the candidates.
		for (int i=0; i<n_selectors; i++) {
			candidates.push_back(selectors[i].hl->addr);
			ret=true;
		}
	}
	else {
		// use a randomized algorithm to select candidates.
		for (int i=0; i<n_cand; i++) {
			int tester=(int)((unsigned)gsl_rng_get(rng_gen)%(unsigned)total_cap);
			int lb=total_cap;
			int j=n_selectors-1;
			for (; j>=0; j--) {
				lb-=selectors[j].my_cap;
				if (lb<=tester)
					break;
			}
			if (j<0) {
				tmpi_error(DBG_INTERNAL, "lb=%d, tester=%d, total_cap=%d.\n", lb, tester, total_cap);
				assert(j>=0);
			}
			candidates.push_back(selectors[j].hl->addr);
			ret=true;
			// swap value with the last one.
			n_selectors--;
			total_cap-=selectors[j].my_cap;
			selectors[j].my_cap=selectors[n_selectors].my_cap;
			selectors[j].hl=selectors[n_selectors].hl;
		}
	}

	m_lock.release();

	return ret;
}
