// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*- 

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*************************************************************************/

/**
 * Name    :  $RCSfile: host_map.h,v $ - header file for host map. 
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2004/01/26 16:24:27 $
 * Version :  $Revision: 1.15 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

# ifndef _HOST_MAP_H_
# define _HOST_MAP_H_

# include "sorrento_types.h"
# include "host_load.h"
# include "rwlock.h"
# include "unit_ring.h"
# include "srv_callback.h"

# include <map>
# include <vector>
# include <string>
# include <assert.h>
# include <pthread.h>
# include "gsl/gsl_rng.h"

/**
 *  A map of HostLoad structures indexed by HostID, it also 
 *  maintains a read-only snapshot of HostLoads as a vector. 
 *  The HostLoad map structure is updated by the subscriber. 
 *  The vector structure is read by proxy threads to perform
 *  various functions (data location, placement, etc).
 *  We do not maintain the synchrony between the map and
 *  the vector. Instead, we keep the versions for the vector
 *  and the map. Each time when the subscriber updates
 *  the map, it also increments the map version. When a reader
 *  finds that the map version is higher than the vector 
 *  version, then it will try to do lock promotion and acquire
 *  a write lock, re-compute the vector, and synchronize
 *  the versions. 
 */
class VHostLoad : public std::vector<HostLoad> {
 private:
	 std::string id;
 public:
	VHostLoad(const char *vid) : id(vid) {}
	const char *ident() const { return id.c_str(); }
};

typedef std::vector<ServerAddr> VecServerAddr;

class HostMap : public std::map<HostID, HostLoad *, HostID_cmp> {
 private:
	std::string id;

	unsigned int map_ver;
	unsigned int vec_ver;
	time_t last_check;

	RWLock m_lock;
	RWLock v_lock;
	VHostLoad *v;

	void synchronize();
	void check_synchrony();

	UnitRing host_ring;
	gsl_rng *rng_gen;

	pthread_mutex_t cb_lock;
	SrvCBInterface *cbi;
	HostAddr dir_addr;
	int expire;

 public:
	HostMap(const char *mid);
	~HostMap();

	const char *ident() const { return id.c_str(); }
	const VHostLoad *get_host_loads() {
		assert(v_lock.read_lock());
		check_synchrony();
		return v;
	}

	bool release_host_loads(const VHostLoad *vhl) {
		assert(v==vhl);
		return v_lock.release();
	}

	// returns true when a new host is added, false
	// when the host already exists.
	bool update(HostLoad *load);

	// returns true when a host is removed, false
	// when the host does not exists.
	bool remove(HostLoad *load);

	/** Get the home host ID of an object. */
	bool get_home(ObjID &guid, HostID &hid) {
		int n_nb=host_ring.get_neighbors(&hid, 1, guid);
		return (n_nb>0);
	}

	/** Translate a host ID to the network address. */
	bool get_address(const HostID &hid, ServerAddr &addr);

	/**
	 * Returning a set of candidate hosts that may host a certain object.
	 */
	bool lookup(ObjID &guid, VecServerAddr &candidates);

	/**
	 * Find a set of hosts that are suitable for keeping a certain object.
	 */
	bool place(ObjID &guid, off_seg_t seg_size, VecServerAddr &candidates);

	bool get_control_ranges(HostID & hid, VecURRange & ranges) {
		return host_ring.get_ranges(hid, ranges);
	}

	void set_cbi(SrvCBInterface *srv_cbi) {
		pthread_mutex_lock(&cb_lock);
		cbi=srv_cbi;
		pthread_mutex_unlock(&cb_lock);
	}

	SrvCBInterface * clr_cbi(void) {
		SrvCBInterface *ret=NULL;
		pthread_mutex_lock(&cb_lock);
		ret=cbi;
		cbi=NULL;
		pthread_mutex_unlock(&cb_lock);
		return ret;
	}

	bool set_dir_addr(HostAddr &ha) {
		time_t now;
		now=time(NULL);
		expire=now+5;
		dir_addr = ha;
		return true;
	}

	bool get_dir_addr(HostAddr &da) {
		if (expire == (time_t)0)
			return false;

		if (dir_addr.valid()) {
			time_t now;
			now=time(NULL);
			if (now > expire) {
				expire=(time_t)0;
				return false;
			}
			da=dir_addr;
			return true;
		}

		return false;
	}
	void dump()
	{
		for (HostMap::const_iterator pi = this->begin(); pi != this->end(); pi++) {
			char guid_str[128];
			Guid guid(pi->first);
			guid.to_string(guid_str, 128);
			printf("%s\n", guid_str);
		}

		return;
	}
};

# endif
