// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: myterm.cc,v $ - a simple terminal manipulate
 *            class.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/08/02 00:06:25 $
 * Version :  $Revision: 1.6 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

# include <string>
# include <stdlib.h>
# include <stdio.h>
# include <errno.h>
# include <sys/ioctl.h>

# include "myterm.h"

# include <termios.h>
# include <curses.h>
# include <term.h>

using namespace std;

typedef int (*PTYPE)(int);
#define PUTP(x) (tputs(x,1,(PTYPE)putchar))

MyTerm::MyTerm(char *termtype)
{
	struct termios newtty;

	if (!termtype) {
		termtype = getenv("TERM");
	}
	if (!termtype) {
		termtype = "VT100";
	}

	/*
	 * Get termcap entries and window size.
	 */
	if(tgetent(NULL, termtype) != 1) {
		throw string("MyTerm::MyTerm: Unknown terminal: ")+termtype+".";
	}

	cm = tgetstr("cm", 0);
	cd = tgetstr("cd", 0);
	cl = tgetstr("cl", 0);
	ce = tgetstr("ce", 0);
	ho = tgetstr("ho", 0);
	md = tgetstr("md", 0);
	mr = tgetstr("mr", 0);
	me = tgetstr("me", 0);

	if (tcgetattr(0, &Savetty) == -1) {
		throw string("MyTerm::MyTerm: tcgetattr() failed: ")+strerror(errno)+".";
	}

	newtty = Savetty;
	newtty.c_lflag &= ~ICANON;
	newtty.c_lflag &= ~ECHO;
	newtty.c_cc[VMIN] = 1;
	newtty.c_cc[VTIME] = 0;

	if (tcsetattr(0, TCSAFLUSH, &newtty) == -1) {
		throw string("MyTerm::MyTerm: Cannot put tty into raw mode.");
	}

	tcgetattr(0, &Rawtty);

	adj_window_size();

	return;
}
	
MyTerm::~MyTerm()
{
    tcsetattr(0, TCSANOW, &Savetty);
}

void MyTerm::adj_window_size()
{
	struct winsize ws;

	if ((ioctl(1, TIOCGWINSZ, &ws) != -1)){
		Cols = ws.ws_col;
		Lines = ws.ws_row;
	}
	else {
		Cols = tgetnum("co");
		Lines = tgetnum("li");
	}
}

void MyTerm::clr_scr()
{
	PUTP(cl); // clear the screen.
	// PUTP(ho); // goto cursor home.
}

// normal line is displayed on the third line and below (left aligned, normal).
void MyTerm::display_line(string &line, int line_no)
{
	if (line_no<2)
		return;

	PUTP(tgoto(cm, 0, line_no));
	printf ("%s",line.c_str());
	fflush(stdout);
	// PUTP(ce);
}

// message is displayed on the second line (left aligned, bold, inversed).
void MyTerm::display_msg(string &msg)
{
	string s;
	if (msg.length()>(unsigned)Cols) {
		s=string(msg, 0, Cols-3);
		s=s+"...";
	}
	else {
		s=msg;
		for (int i=msg.length(); i<Cols; i++) {
			s=s+' ';
		}
	}
	PUTP(tgoto(cm, 0, 1));
	PUTP(md);
	PUTP(mr);
	printf ("%s",s.c_str());
	fflush(stdout);
	// PUTP(ce);
	PUTP(me);
}

// title is displayed on the top line (center aligned, bold).
void MyTerm::display_title(string &title)
{
	// PUTP(ho); // cursor home
	int mid=0;

	mid = (Cols - (signed)title.length())/2;

	if (mid>0) {
		PUTP(tgoto(cm, mid, 0)); // goto (mid, 0) (x, y)
		PUTP(md); // bold mode
		printf("%s", title.c_str());
	}
	else {
		PUTP(tgoto(cm, 0, 0));
		PUTP(md); // bold mode
		string new_title(title, 0, Cols-3);
		new_title=new_title+"...";
		printf("%s", new_title.c_str());
	}
	fflush(stdout);
	// PUTP(ce); // clear till the end of the line.
	PUTP(me); // turn off attributes.
}

void MyTerm::display_body(string *str_lines, int n_line)
{
	if ((str_lines==NULL)||(n_line<1))
		return;

	PUTP(tgoto(cm, 0, 2));
	for (int i=0; i<n_line; i++) {
		if (str_lines[i].length()>(unsigned)Cols) {
			string s(str_lines[i], 0, Cols-3);
			s=s+"...";
			printf("%s", s.c_str());
		}
		else {
			printf("%s", str_lines[i].c_str());
		}
		if (i<n_line-1)
			printf("\n");
	}
	fflush(stdout);
}
