// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: publisher.cc,v $ - load publisher.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/09/07 01:09:40 $
 * Version :  $Revision: 1.14 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 
# include <sys/types.h>
# include <sys/stat.h>
# include <fcntl.h>
# include <netinet/in.h>
# include <unistd.h>
# include <stdio.h>
# include <mntent.h>
# include <string.h>
# include <stdio.h>
# include <stdlib.h>
# include <string>

# include "tmpi_debug.h"
# include "sock_conn.h"
# include "mcast_annc.h"
# include "resmon.h"
# include "../shared/daemon_common.h"
# include "sorrento_conf.h"
# include "publisher.h"

using namespace std;
using std::string;


Publisher::Publisher(SorrentoConf &sconf, int tport, int uport) : 
	ident(NULL),
	ancer(NULL),
	mnt_ptrs(NULL),
	n_mnt(0),
	interval(1000),
	vnodes(32),
	tcp_port(tport),
	udp_port(uport),
	diskInfos(NULL),
	verfs(NULL)
{
	ident=sconf.ident();
	if (!ident) {
		throw string("Publisher::Publisher: No identity string defined.");
	}

	char *homeDir = sconf.home_dir();
	if (!homeDir) {
		throw string("Publisher::Publisher: Cannot determine Sorrento home.");
	}

	char host_id_file[1024];
	sprintf(host_id_file, "%s/etc/host_id.%s", homeDir, ident);
	free(homeDir);

	if (access(host_id_file, R_OK)) {
		// host-ID file not found, create it.
		int fd=open(host_id_file, O_CREAT|O_EXCL|O_WRONLY, S_IRUSR|S_IRGRP);
		if (fd<0) {
			char *hDir = sconf.home_dir();
			string sHDir (hDir);
			free(hDir);
			throw string( "Publisher::Publisher: Failed to create host-ID file: permission denied.\n(Check if the directory: \""+ sHDir +"\" exists)");
		}

		Guid guid;
		guid.gen();
		guid.pack(&host_id, sizeof(HostID));

		char guid_str[128];
		guid.to_string(guid_str, 128);
		write(fd, guid_str, strlen(guid_str));
		close(fd);
	}
	else {
		int fd=open(host_id_file, O_RDONLY);
		if (fd<0) {
			throw string("Publisher::Publisher: Failed to read host-ID file: permission denied.");
		}

		char guid_str[128];
		int n=read(fd, guid_str, 127);

		close(fd);

		guid_str[n]='\0';

		Guid guid;
		guid.parse(guid_str);
		guid.pack(&host_id, sizeof(HostID));
	}

	char *mc_group=sconf.mc_group();
	int mc_port=sconf.mc_port();

	struct sockaddr_in sin;

	if ((!mc_group)||(mc_port<=0)) {
		tmpi_error(DBG_CALLER, "Bad parameters: mc_group=%s, mc_port=%d.\n", mc_group, mc_port);
		throw string("Publisher::Publisher: Bad parameters to Publisher().");
	}

	if (sock_filladdr(mc_group, mc_port, &sin)<0) {
		tmpi_error(DBG_CALLER, "Bad address: %s:%d.\n", mc_group, mc_port);
		throw string("Publisher::Publisher: Bad parameters to Publisher().");
	}

	if (sock_is_mcast_addr(&sin, sizeof(sin))!=1) {
		tmpi_error(DBG_CALLER, "Bad multi-cast address: %s:%d.\n", mc_group, mc_port);
		throw string("Publisher::Publisher: Bad parameters to Publisher().");
	}

	mnt_ptrs=sconf.partitions();
	if (mnt_ptrs) {
		char **p=mnt_ptrs;
		while (*p) {
			p++;
			n_mnt++;
		}
	}

	diskInfos = NULL;
	
	ancer=new Mc_ancer(mc_group, mc_port);

	free(mc_group);

	if (!ancer->good()) {
		tmpi_error(DBG_CALLER, "Mc_ancer() failed.\n");
		throw string("Publisher::Publisher: Cannot create multicast publisher.");
	}

	interval=sconf.mc_interval();
	vnodes=sconf.vnodes();

	return;
}

Publisher::~Publisher(void)
{
	release();
}

void Publisher::release(void)
{
	if (mnt_ptrs) {
		for (int i=0; i<n_mnt; i++) {
			if (mnt_ptrs[i]) {
				free(mnt_ptrs[i]);
				mnt_ptrs[i]=NULL;
			}
		}
		free(mnt_ptrs);
		mnt_ptrs=NULL;
	}

	if (ancer) {
		delete ancer;
		ancer=NULL;
	}

	free(ident);
	//delete diskInfos;			/**/
}

static void random_usleep(unsigned long usec)
{
	int least=usec/2;
	usleep(least+random()%usec);
}

# define MAX_ERRORS 1000

void *Publisher::run()
{
	int errors=0;

	//tmpi_error(DBG_VERBOSE, "Publisher enters the run() loop.\n");

	while (!should_quit()) {
		if (errors>MAX_ERRORS) {
			tmpi_error(DBG_APP, "Load Publisher is quitting due to too many errors ...\n");
			return NULL;
		}

		int annc_error=0;
		if ((annc_error=announce())) {
			tmpi_error(DBG_INTERNAL, "Publisher.announce failed.\n");
			errors+=annc_error;
		}
		else {
			//tmpi_error(DBG_VERBOSE, "Load announced.\n");
		}

		random_usleep(interval*1000);
	}

	//tmpi_error(DBG_VERBOSE, "Load Publisher is quitting normally ...\n");

	return NULL;
}

// TO-DO: it might be better to use XDR to marshal data, currently there is
// an obvious space waste.
int Publisher::announce()
{
	int errors=0;

	int use_cpu=1;
	int use_mem=1;

	ResMon::CPU cpu;

	if (ResMon::get_cpu(&cpu)<0) {
		tmpi_error(DBG_INTERNAL, "Cannot obtain CPU load.\n");
		use_cpu=0;
		errors++;
	}

	ResMon::MEM mem;
	if (ResMon::get_mem(&mem)<0) {
		tmpi_error(DBG_INTERNAL, "Cannot obtain mem load.\n");
		use_mem=0;
		errors++;
	}

	Mc_anc_data data;

	if (data.add(KEY_SOR_IDENT, ident, strlen(ident)+1)<0) {
		tmpi_error(DBG_INTERNAL, "Mc_annc_data.add(%d) failed.\n", KEY_SOR_IDENT);
		errors++;
	}

	if (data.add(KEY_SVR_IDENT, &host_id, sizeof(HostID))<0) {
		tmpi_error(DBG_INTERNAL, "Mc_annc_data.add(%d) failed.\n", KEY_SVR_IDENT);
		errors++;
	}

	if (data.add(KEY_SVR_VNODES, &vnodes, sizeof(int))<0) {
		tmpi_error(DBG_INTERNAL, "Mc_annc_data.add(%d) failed.\n", KEY_SVR_VNODES);
		errors++;
	}

	if (data.add(KEY_SVR_TCP_PORT, &tcp_port, sizeof(int))<0) {
		tmpi_error(DBG_INTERNAL, "Mc_annc_data.add(%d) failed.\n", KEY_SVR_TCP_PORT);
		errors++;
	}

	if (data.add(KEY_SVR_UDP_PORT, &udp_port, sizeof(int))<0) {
		tmpi_error(DBG_INTERNAL, "Mc_annc_data.add(%d) failed.\n", KEY_SVR_UDP_PORT);
		errors++;
	}

	struct {
		unsigned char key;
		void *buf;
		unsigned int len;
	} *p, 

	recs_cpu[] = {
		{KEY_CPU_TOT, &(cpu.cputotal), sizeof(cpu.cputotal)},
		{KEY_CPU_USR, &(cpu.user), sizeof(cpu.user)},
		{KEY_CPU_KER, &(cpu.kernel), sizeof(cpu.kernel)},
		{KEY_CPU_IOW, &(cpu.iowait), sizeof(cpu.iowait)},
		{KEY_CPU_IDL, &(cpu.idle), sizeof(cpu.idle)},
		{KEY_CPU_NUM, &(cpu.cpus), sizeof(cpu.cpus)},
		{KEY_CPU_CXW, &(cpu.cswitches), sizeof(cpu.cswitches)},
		{0, NULL, 0}
	}, 

	recs_mem[] = {
		{KEY_MEM_TOT, &(mem.mtotal), sizeof(mem.mtotal)},
		{KEY_MEM_USE, &(mem.mused), sizeof(mem.mused)},
		{KEY_MEM_FRE, &(mem.mfree), sizeof(mem.mfree)},
		/* {KEY_MEM_SHR, &(mem.mshared), sizeof(mem.mshared)}, */
		{KEY_MEM_BUF, &(mem.mbuffers), sizeof(mem.mbuffers)},
		{KEY_MEM_CCH, &(mem.mcached), sizeof(mem.mcached)},
		{KEY_SWP_TOT, &(mem.stotal), sizeof(mem.stotal)},
		{KEY_SWP_FRE, &(mem.sfree), sizeof(mem.sfree)},
		{0, NULL, 0}
	};

	if (use_cpu) {
		p=recs_cpu;
		while (p->buf) {
			if (data.add(p->key, p->buf, p->len)<0) {
				tmpi_error(DBG_INTERNAL, "Mc_annc_data.add(%d) failed.\n", p->key);
				errors++;
			}
			p++;
		}
	}

	if (use_mem) {
		p=recs_mem;
		while (p->buf) {
			if (data.add(p->key, p->buf, p->len)<0) {
				tmpi_error(DBG_INTERNAL, "Mc_annc_data.add(%d) failed.\n", p->key);
				errors++;
			}
			p++;
		}
	}

# if (0)
	/**
	 * The following has to be modified to read the
	 * configuration file and send out only the 
	 * managed volumns. 
	 */
	static const char *MTAB_NAME = "/etc/mtab";
	FILE *fp;
	struct mntent *mnt;
	int ndisks=0;
	int use_disks=1;

	if ((fp=setmntent(MTAB_NAME, "r"))==NULL) {
		tmpi_error(DBG_INTERNAL, "Cannot open mtab.\n");
		errors++;
		use_disks=0;
	}

	if (use_disks) {
		while ((mnt=getmntent(fp))) {
			if (strcmp(mnt->mnt_fsname, "none")==0) {
				continue;
			}
			ResMon::DISK *d=ResMon::get_disk(mnt->mnt_dir);
			if ((d->good)&&(d->btotal>0)) {
				ndisks++;

				DiskInfo dsk;
				dsk.btotal=d->btotal;
				dsk.bfree=d->bavail;

				if (data.add(KEY_DSK_FRE, &dsk, sizeof(dsk))<0) {
					tmpi_error(DBG_INTERNAL, "Mc_annc_data.add(%d) failed.\n", KEY_DSK_FRE);
					errors++;
				}

				delete d;
			}
		}
		endmntent(fp);
	}


	if (ndisks>0) {
		if (data.add(KEY_DSK_NUM, &ndisks, sizeof(ndisks))<0) {
			tmpi_error(DBG_INTERNAL, "Mc_annc_data.add(%d) failed.\n", KEY_DSK_NUM);
			errors++;
		}
	}

# endif


	int ndisks=0;

	if (n_mnt>0) {

		verfs->refreshDiskInfos();

		for (int i=0; i<n_mnt; i++) {

			if(diskInfos[i].btotal > 0 && diskInfos[i].bfree >=0){

				ndisks++;
				
				if (data.add(KEY_DSK_FRE, &diskInfos[i], sizeof(DiskInfo))<0) {
					tmpi_error(DBG_INTERNAL, "Mc_annc_data.add(%d) failed.\n", KEY_DSK_FRE);
					errors++;
				}
			}
			else {
				tmpi_error(DBG_VERBOSE, "Mount point not found: %s.\n", mnt_ptrs[i]);
			}
		}
	}

	if (ndisks>0) {
		if (data.add(KEY_DSK_NUM, &(ndisks), sizeof(ndisks))<0) {
			tmpi_error(DBG_INTERNAL, "Mc_annc_data.add(%d) failed.\n", KEY_DSK_NUM);
			errors++;
		}
	}

# if defined(__DEBUG__)
	else {
		// dump_anc(&data);
	}
# endif

	return ancer->announce(&data);


#if (0)							// old code that updates disk info
								// now this info is taken from verfs
	int ndisks=0;

	if (n_mnt>0) {
		for (int i=0; i<n_mnt; i++) {
			ResMon::DISK *d=ResMon::get_disk(mnt_ptrs[i]);
			if ((d->good)&&(d->btotal>0)) {
				ndisks++;

				DiskInfo dsk;
				dsk.btotal=d->btotal;
				dsk.bfree=d->bavail;

				/* added so that verfs can get free disk info */
				diskInfos[i] = dsk;
				/**/
				
				if (data.add(KEY_DSK_FRE, &dsk, sizeof(dsk))<0) {
					tmpi_error(DBG_INTERNAL, "Mc_annc_data.add(%d) failed.\n", KEY_DSK_FRE);
					errors++;
				}
				
				delete d;
			}
			else {
				
				/* added so that verfs can get free disk info */
				diskInfos[i].btotal = -1;
				diskInfos[i].bfree = -1;
				/**/

				tmpi_error(DBG_VERBOSE, "Mount point not found: %s.\n", mnt_ptrs[i]);
			}
		}
	}

	if (ndisks>0) {
		if (data.add(KEY_DSK_NUM, &(ndisks), sizeof(ndisks))<0) {
			tmpi_error(DBG_INTERNAL, "Mc_annc_data.add(%d) failed.\n", KEY_DSK_NUM);
			errors++;
		}
	}

# if defined(__DEBUG__)
	else {
		// dump_anc(&data);
	}
# endif

	return ancer->announce(&data);

#endif

}
