// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: replicator.cc,v $ - Module that is responsible from
 *            replication. 
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Aziz Gulbeden, gulbeden@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2004/04/29 02:17:14 $ by $Author: gulbeden $
 * Version :  $Revision: 1.11 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

# include "client.h"			// to connect to other servers
# include "host_map.h"
# include "protocol.h"
# include "replicator.h"
# include "tmpi_debug.h"
# include "udp_cli.h"
# include "verfs.h"

int Replicator::repDegree = 3;

bool Replicator::setRepDegree(int rd){
	repDegree = rd;
	return true;
}

/**
 * adds an object-server-version tuple to the replicator's table. It does not
 * allow same oid to be inserted more than once.
 * @param oid Object ID of the object
 * @param saadr Server Address of the serve that stores the object
 * @param v version of the segment stored.
 */
bool Replicator::addSegment(const ObjID &oid, const ServerAddr &saddr, int v){

	ownSegsType::iterator curr;

	if(pthread_mutex_lock(&ownSegLock) != 0)
		return false;

	curr = checkOidSaddr(oid, saddr);

	if(curr == ownSegs.upper_bound(oid)){
		// does not exist, insert
		
		SadVer sv(saddr,v);
		pair<const ObjID, SadVer> p2(oid, sv);
		ownSegs.insert(p2);
	}
	
	else 
		// this oid,server pair already inserted, just update version
		curr->second.version = v;

	setRepDegree(oid, Replicator::repDegree, v);
	pthread_mutex_unlock(&ownSegLock);

	return true;
}

/**
 * Checks if the oid,saddr pair is present in the OwnSegs table.
 * returns an iterator that points to the pair if found or to the upper bound 
 * if not found.
 */
ownSegsType::iterator Replicator::checkOidSaddr(const ObjID &oid, const ServerAddr &saddr){

	ownSegsType::iterator low;
	ownSegsType::iterator up;
	ownSegsType::iterator curr;
	
	low = ownSegs.lower_bound(oid);
	up = ownSegs.upper_bound(oid);

	// find the specified oid-server pair
	for(curr = low; curr!=up; curr++){
		if (memcmp(&((*curr).second.sa), &saddr, sizeof(ServerAddr)) == 0){
			return curr;
		}
	}

	return curr;
}


/**
 * Finds a host to replicate segment, the server to replicate at is stored at
 * sa.
 */
bool Replicator::determineReceiver(ObjID &oid, ServerAddr &sa) {

	VecServerAddr candidates;
	VecServerAddr::iterator sit;
	ownSegsType::iterator low;
	ownSegsType::iterator up;
	ownSegsType::iterator curr;
	
	// TODO: determine max_size in a better way
	hm->place(oid, 512*1024*1024, candidates);

	if (candidates.size() <2)
		return false;

	if(pthread_mutex_lock(&ownSegLock) != 0)
		return false;

	low = ownSegs.lower_bound(oid);
	up = ownSegs.upper_bound(oid);

	// find the first one that is not self
	for(sit=candidates.begin(); sit != candidates.end(); sit++){

		for(curr = low; curr != up; curr++){
			if (memcmp(&((*curr).second.sa), &(*sit), sizeof(ServerAddr))==0){
				// this host already has it - not suitable
				break;
			}
		}
		if(curr == up) break;
	}

	pthread_mutex_unlock(&ownSegLock);

	if (sit == candidates.end()) return false; 
	// because this candidate is not suitable

	memcpy(&sa, &(*sit), sizeof(ServerAddr));

	return true;
}

/**
 * Determines a server that will send data to an older version replica.
 */
bool Replicator::determineSender(ObjID &oid, ServerAddr &sa, int &latestVer){

	ownSegsType::iterator low;
	ownSegsType::iterator up;
	ownSegsType::iterator curr;
	
	if(pthread_mutex_lock(&ownSegLock) != 0){
		tmpi_error(DBG_INTERNAL, "Unable to get lock\n");
		return false;
	}

	low = ownSegs.lower_bound(oid);
	up = ownSegs.upper_bound(oid);

	map<const ObjID, DegVer, ObjID_cmp>::iterator i;
	i = repDegrees.find(oid);
	if (i==repDegrees.end()){
		tmpi_error(DBG_INTERNAL, "Object not in rep degrees map (this should never happen)\n");
		pthread_mutex_unlock(&ownSegLock);
		return false;
	}

	latestVer = (*i).second.version;

	// find somebody who has the latest version
	for(curr=low; curr!=up; curr++){
		if((*curr).second.version == latestVer) // found
			break;
	}

	memcpy(&sa, &(*curr).second.sa, sizeof(ServerAddr));

	pthread_mutex_unlock(&ownSegLock);

	if (curr == up) {
		//return false; 

		tmpi_error(DBG_INTERNAL, "(Continuing to run by the help of the patch)\n");

		// we actually should return false, but I (Aziz) am assuming
		// that the object needs to be at the current host so I copy
		// its address and return true. I don't know why somebody with
		// latest version is not found.

		memcpy(&sa, &(*myAddr), sizeof(ServerAddr));
	}

	return true;
}


/**
 * This function is called in case of a change in version at a replica.
 */
bool Replicator::setVersion(const ObjID &oid, const ServerAddr &saddr, int v){

	ownSegsType::iterator curr;
	
	if(pthread_mutex_lock(&ownSegLock) != 0)
		return false;

	curr = checkOidSaddr(oid, saddr);
	
	if (curr == ownSegs.upper_bound(oid)){ // not found so add
		pthread_mutex_unlock(&ownSegLock);
		addSegment(oid, saddr, v);		
	}

	else{
		(*curr).second.version = v;
		pthread_mutex_unlock(&ownSegLock);
	}

	bool ret = setLatestVersion(oid, v);

	return ret;
}

/**
 * This function sets the latest version to be the one sent as the parameter
 * if it is greater than the current latest version.
 */
bool Replicator::setLatestVersion(const ObjID &oid, int ver){

	repDegreesType::iterator i;
	i = repDegrees.find(oid);
	if (i==repDegrees.end()){
		tmpi_error(DBG_INTERNAL, "Object not in rep-degrees (this should never happen)\n");
		setRepDegree(oid, Replicator::repDegree, ver);
		i = repDegrees.find(oid);
		return true;
	}

	if (ver <= (*i).second.version)
		// update not necessary
		return true;

	(*i).second.version = ver;

	return true;
}


bool Replicator::setRepDegree(const ObjID &oid, int deg, int v){
	bool succ;
	DegVer dv(deg, v);

	// modify version number in the table
	map<const ObjID, DegVer, ObjID_cmp>::iterator i;
	i = repDegrees.find(oid);

	if(pthread_mutex_lock(&rdLock) != 0) return false;

	if(i!=repDegrees.end()){
		(*i).second.degree = deg;
		pthread_mutex_unlock(&rdLock);
		return true;
	}

	// insert dv
	pair<const ObjID, DegVer> p(oid, dv);
	succ = (repDegrees.insert(p)).second;
	
	pthread_mutex_unlock(&rdLock);

	if (succ==false){
		tmpi_error(DBG_INTERNAL, "Insert into rep degrees map failed\n");
		return false;
	}
	return true;
}

bool Replicator::start(){
	// first call parent class
	CBS_Thread::start();

	// now learn this server's address
	int MAX_TRY = 60;
	int i;
	
	for(i=0; i<MAX_TRY; i++){
		// The Replicator has to know what is this server's address so that 
		// it can work correctly, if the following call fails because hostmap
		// is not up-to-date, we retry. 
		if (hm->get_address(*myHid, *myAddr) == true)
			break;
		sleep(1);
	}
	if (i==MAX_TRY) return false;

	return true;
}

bool Replicator::removeSegment(const ObjID &oid){
	if(pthread_mutex_lock(&rdLock) != 0) return false;
	repDegrees.erase(oid);
	if(pthread_mutex_lock(&ownSegLock) != 0) return false;
	ownSegs.erase(ownSegs.lower_bound(oid), ownSegs.upper_bound(oid));
	pthread_mutex_unlock(&ownSegLock);
	pthread_mutex_unlock(&rdLock);
	return true;
}

# define STG_SCU_UDP_TOUT 2000

Replicator::Replicator(HostMap *hostMap, Verfs *v, HostID *hid, ServerAddr *ma) :
	hm(hostMap),
	verfs(v),
	myHid(hid),
	myAddr(ma)
{
	cli = new Client();
	if (pthread_mutex_init(&ownSegLock, NULL) != 0) 
		throw -1;
	if (pthread_mutex_init(&rdLock, NULL) != 0) 
		throw -1;

	udpCli = new Udp_cli(NULL, SOCK_PORT_ANY, STG_SCU_UDP_TOUT);
	
}


Replicator::~Replicator(){
	delete cli;
	pthread_mutex_destroy(&ownSegLock);
	pthread_mutex_destroy(&rdLock);
}


/**
 * goes over the tables to remove the objects which no longer belong here.
 */
bool Replicator::notify_host_join(){
	repDegreesType::iterator iter;
	
	if(pthread_mutex_lock(&rdLock) != 0) return false;	

	for(iter = repDegrees.begin(); iter !=repDegrees.end(); iter++){
		HostID home_host;
		if (hm->get_home((ObjID&)((*iter).first), home_host) == false) {
			return false;		// something went wrong
		}
		
		if (memcmp(&home_host, myHid, sizeof(HostID)) != 0) {
			// this host is not the homehost anymore 
			// delete from the table
			
			if(pthread_mutex_lock(&ownSegLock) != 0) return false;
			ownSegs.erase(ownSegs.lower_bound((ObjID&)(iter->first)), ownSegs.upper_bound((ObjID&)(iter->first)));
			pthread_mutex_unlock(&ownSegLock);

			repDegrees.erase(iter);
			tmpi_error(DBG_VERBOSE, "Deleted.\n");
		}
	}

	pthread_mutex_unlock(&rdLock);

	return true;
}

bool Replicator::notify_host_leave(ServerAddr &sad){

	ownSegsType::iterator osIter;

	for(osIter = ownSegs.begin(); osIter != ownSegs.end(); osIter++){

		if( memcmp( &((*osIter).second.sa), &sad, sizeof(ServerAddr)) == 0){
			ownSegs.erase(osIter);
			tmpi_error(DBG_INTERNAL, "Host left, index entry deleted in replicator.\n");
		}
	}
	return true;

}

/**
 * Sends data to the remote server.
 * @param oid id of the object to send.
 * @param sa address of the server to send the data to.
 * @return true if successful, false otherwise.
 *
 * NOT USED ANYMORE BECAUSE SERVER HANDLES THIS
 */
bool Replicator::sendData(ObjID &oid, ServerAddr &sa){

	unsigned int bufSize = 1024*1024*8;
	char buffer[bufSize];
	
	int offset = 0;
	Descriptor remoteDesc;
	int localDesc;
	int response;

	// open for read
	int version;
	localDesc = verfs->open(oid, "r", version);
	if (localDesc < 0){
		tmpi_error(DBG_INTERNAL, "Verfs failed in replicator\n");
		return false;
	}
	
	response = cli->create(sa, oid, remoteDesc);
	
	if(response == STG_RES_OKAY){
		bool lastTime = false;
		
		for(offset=0; lastTime != true; offset+=bufSize){
			int len;
			// verfs read
			len = verfs->read(localDesc, offset, buffer, bufSize);
			
			if((unsigned int)len < bufSize) // ignores one case where size%bufSize= 0
				// there is bug here!
				lastTime = true;
			
			// send write request
			off_seg_t writtenBytes;
			cli->write(remoteDesc,offset,buffer,len, writtenBytes);
		}
		
		// notify that transfer finished so that that host can notify
		// namespace server and this server
	}

	return true;
}

bool Replicator::sendSyncReq(ServerAddr &saSend, ServerAddr &saRecv,
							 const ObjID &oid, const int latestVer) const{

	sleep(1);
	// I added a sleep call here because when a big file is
	// closed a large number of synchronize requests will be
	// sent, if we don't sleep here, the rate of requests
	// generated is above the amount that remote server can
	// handle.

	Stg_Request *req=new_stg_req_sync_host(saSend, oid, latestVer);
	
	if (!req) {
		tmpi_error(DBG_INTERNAL, "Out of memory.\n");
		return false;
	}
	
	Stg_Req_Handle reqh;
	bool use_tcp=false;
	int ret=stg_start_request_no_payload(udpCli, saRecv, req, reqh, use_tcp);
	stg_req_free(req);
	
	if (ret != STG_RES_OKAY) {
		tmpi_error(DBG_VERBOSE, "Failed to send request (err=%d).\n", ret);
		return false;
	}

	// server should send a reply immediately and then do the
	// synchronization

	Stg_Res_Handle resh;
	ret=stg_wait_response_no_payload(udpCli, reqh, resh);
	if (ret != STG_RES_OKAY) {
		return false;
	}
	
	Stg_Response *res=resh.response();
	if (res->header.tag != STG_RES_OKAY) {
		tmpi_error(DBG_VERBOSE, "Unexpected reply.\n");
		return false;
	}

	return true;
}


void *Replicator::run() 
{

	repDegreesType::iterator iter;
	ObjID oid;

	while (!should_quit()) {

		sleep(10);

		for(iter = repDegrees.begin(); iter !=repDegrees.end(); iter++){
			
			oid = (ObjID&)(*iter).first;
			syncObject(oid);
			// Not the best way, sending the iterator would be more
			// useful but syncObject has to take an oid because it is
			// also called direcly by the server.
	    }
	}

	tmpi_error(DBG_VERBOSE, "Replicator stopped\n");
	
    return NULL;
}


double Replicator::compAvg(){

	int sum=0;
	int count = 0;

	repDegreesType::iterator iter;
	ObjID oid;

	for(iter = repDegrees.begin(); iter !=repDegrees.end(); iter++, count++){
		oid = (ObjID&)(*iter).first;
		sum += (int)(ownSegs.count(oid));
	}

	if (count == 0) return -1;
	else return ((double)sum/(double)count);

}


/**
 * This fuction makes sure that the changes performed on the object oid
 * is transferred to all of the replicas.
 */
bool Replicator::syncObject(/*const */ObjID &oid){

	bool status;
	repDegreesType::iterator iter;
	ServerAddr saRecv, saSend;
	int latestVer;
	char repStr[32];
	struct timeval tv;

	iter = repDegrees.find(oid);

	if(iter==repDegrees.end()){
		tmpi_error(DBG_INTERNAL,"Object to synchronize not found.\n");
		return false;
	}
	
	if((int)(ownSegs.count(oid)) < iter->second.degree){
		tmpi_error(DBG_VERBOSE, "object not replicated enough\n");
		
		// determine where to replicate
		status = determineReceiver(oid, saRecv);
		
		if (status == false){
			tmpi_error(DBG_VERBOSE, "Nowhere to replicate\n");
			return false;	// move to the next object
		}
		
		tmpi_error(DBG_VERBOSE, "will replicate at %s\n", \
				   saRecv.toString(repStr));
		
		// now determine who will send data there - one of the hosts
		// that has the latest version.
		status = determineSender(oid, saSend, latestVer);
		
		if (status == false){
			tmpi_error(DBG_VERBOSE, "No node has latest version (this should never happen)\n");
			return false;	// move to the next object
		}
		
		tmpi_error(DBG_VERBOSE, "Sender is: %s (%d)\n", \
				   saSend.toString(repStr), latestVer);
		
		// Now send message to the host that has LOWER version
		sendSyncReq(saSend,saRecv, oid, latestVer);
	}
	
	// Check if everybody has the latest version
	else {
		ownSegsType::iterator osIter;
		ownSegsType::iterator low;
		ownSegsType::iterator up;
				
		// ?????
		//if(pthread_mutex_lock(&ownSegLock) != 0){
		//tmpi_error(DBG_INTERNAL, "UNABLE TO GET LOCK\n");
		//return false;
		//}
		
		low = ownSegs.lower_bound(oid);
		up = ownSegs.upper_bound(oid);
		
		for(osIter = low; osIter !=up; osIter++){
			
			if ((*osIter).second.version < (*iter).second.version){
				
				tmpi_error(DBG_VERBOSE, "there is a server which does not have the latest version %d %d\n", (*osIter).second.version, (*iter).second.version);
				
				// determine who will send data there - one of
				// the hosts that has the latest version.
				status = determineSender(oid, saSend, (*iter).second.version);
				
				if (status == false){
					tmpi_error(DBG_VERBOSE, "No node has latest version:%d (this should never happen)\n", (*iter).second.version);
					return false;	// move to the next object
				}
				
				tmpi_error(DBG_VERBOSE, "Sender is: %s\n", \
						   saSend.toString(repStr));
				
				// Now send message to the host that has LOWER version
				sendSyncReq(saSend, (*osIter).second.sa,oid,
							(*iter).second.version);
			}
		}
	}

	//gettimeofday(&tv, NULL);
	//tmpi_error(DBG_INTERNAL, "Rep deg average: %lf at time %d \n", compAvg(), tv.tv_sec);

	return true;
}
