// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: server.cc,v $ - Sorrento storage server. 
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Aziz Gulbeden, gulbeden@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2004/04/29 02:17:14 $ by $Author: gulbeden $
 * Version :  $Revision: 1.53 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

#include "client.h"			/* to read data from other servers */
#include "guid.h"
#include "host_map.h"
#include "protocol.h"		/* protocol that server and client talk */
#include "publisher.h"
#include "replicator.h"
//#include "xput_tracker.h"
#include "server.h"
#include "sock_sr.h"		/* sock_read_tout, sock_write */
#include "srvcatalog.h"
#include "srvcatalogupdater.h"
#include "subscriber.h"
#include "tmpi_debug.h"
#include "udp_cli.h"
#include "unit_ring.h"
#include "verfs.h"

using namespace std;

# define STG_SRV_UDP_TOUT 2000

/* how should the replicator work: 1 means syncronous (slow) 2 means 
 * lazy and asynchronous */
# define PUSH_TO_REPLICAS 0

class SrvCBImp : public SrvCBInterface {
 private:
	Server *srv;
	HostID shid;
 public:
	SrvCBImp(Server *s) : srv(s)  { shid = srv->get_hid(); }
	virtual void notify_host_join(HostID & hid) { 
		if (!is_self(hid))
			srv->notify_host_join(hid);
	}
	virtual void notify_host_leave(HostID & hid, ServerAddr &sad) { 
		if (!is_self(hid))
			srv->notify_host_leave(hid, sad);
	}
	virtual bool is_self(HostID & hid) {
		return (memcmp(&shid, &hid, sizeof(HostID)) == 0);
	}
};

/**
 * finalizes the threads
 */
bool Server::finalize()
{
	udpDispatcher->finalize();
	tcpDispatcher->finalize();
	scUpdater->finalize();
	publisher->finalize();
	tp->finalize();
	if (replicator) replicator->finalize();

	return true;
}

// all errors during add_entry will be presented to the
// remote client as server error.
bool Server::request_add_entry(ServerAddr &saddr, ObjID &oid, int myid, int version)
{
	Stg_Request *req=new_stg_req_add_entry(myHid, oid, version);
	if (!req) {
		return false;
	}

	Stg_Req_Handle reqh;
	int ret=stg_start_request_no_payload(udp_clis[myid], saddr, req, reqh, 0);
	stg_req_free(req);
	if (ret != STG_RES_OKAY) {
		return false;
	}

	Stg_Res_Handle resh;
	ret=stg_wait_response_no_payload(udp_clis[myid], reqh, resh);
	if (ret != STG_RES_OKAY) {
		return false;
	}

	Stg_Response *res=resh.response();

	if (res->header.tag != STG_RES_OKAY) {
		return false;
	}

	return true;
}

// all errors during del_entry will be presented to the
// remote client as server error.
bool Server::request_del_entry(ServerAddr &saddr, ObjID &oid, int myid)
{
	Stg_Request *req=new_stg_req_del_entry(myHid, oid);
	if (!req) {
		return false;
	}

	Stg_Req_Handle reqh;
	int ret=stg_start_request_no_payload(udp_clis[myid], saddr, req, reqh, 0);
	stg_req_free(req);
	if (ret != STG_RES_OKAY) {
		return false;
	}

	Stg_Res_Handle resh;
	ret=stg_wait_response_no_payload(udp_clis[myid], reqh, resh);
	if (ret != STG_RES_OKAY) {
		return false;
	}

	Stg_Response *res=resh.response();

	if (res->header.tag != STG_RES_OKAY) {
		return false;
	}

	return true;
}


int Server::request_notify_version_ch(ObjID &oid, int version)
{
	// notify home host
	HostID home_host;
	if (hmap->get_home(oid, home_host) == false) {
		tmpi_error(DBG_INTERNAL, "Don't know the home host.\n");
		return STG_RES_ESERVER;
	}

	if (memcmp(&home_host, &myHid, sizeof(HostID)) == 0) {
		// this host is the homehost - make local function call
		if(replicator->setVersion(oid, *myAddr, version) == false){
			tmpi_error(DBG_INTERNAL, "Set version failed.\n");
			return STG_RES_ESERVER;
		}
		
		if(PUSH_TO_REPLICAS){
			replicator->syncObject(oid);
		}
	}

	else{
		Stg_Request *req2=new_stg_req_notify_ver_ch(*myAddr, oid, version);
		
		if (!req2) {
			tmpi_error(DBG_INTERNAL, "Out of memory.\n");
			return STG_RES_EMEM; 
		}
		
		Stg_Req_Handle reqh;
		bool use_tcp=false;

		ServerAddr saHome;
		if (hmap->get_address(home_host, saHome) == false) {
			return STG_RES_ESERVER;
		}
		
		// TODO: don't use udp_clis[0] but create another udp_cli (possible
		// reason of failure during a stress test)
		int ret=stg_start_request_no_payload(udp_clis[0], saHome, req2, reqh, use_tcp);

		stg_req_free(req2);
		
		if (ret != STG_RES_OKAY) {
			tmpi_error(DBG_VERBOSE, "Failed to send request (err=%d).\n", ret);
			return ret;
		}
		
		Stg_Res_Handle resh;
		ret=stg_wait_response_no_payload(udp_clis[0], reqh, resh);
		if (ret != STG_RES_OKAY) {
			return ret;
		}
	}
	// should I contact namespace server?
	return 0;
}


int Server::handle_create1(Stg_Srv_Request *req, int myid)
{
	try {
		tmpi_error(DBG_VERBOSE, "create request\n");
		ObjID oid=req->request()->u.u_create1.oid;
		//ClientID cid=req->request()->u.u_create1.cid;
		int version = -1;

		// TODO: what if creation fails? Shouldn't we return something?
		int desc = verfs->create(oid, version);
		if (desc < 0) {
			return STG_RES_ESERVER;
		}

		if(version<0){
			tmpi_error(DBG_INTERNAL, "Version is negative\n");
			return STG_RES_ESERVER;
		}

		if (si->add(oid) == false) {
			// the desc will be garbage collected
			// by verfs.
			return STG_RES_ESERVER;
		}

		HostID home_host;
		if (hmap->get_home(oid, home_host) == false) {
			return STG_RES_ESERVER;
		}

		if (memcmp(&home_host, &myHid, sizeof(HostID)) != 0) {
			ServerAddr saddr;
			if (hmap->get_address(home_host, saddr) == false) {
				return STG_RES_ESERVER;
			}
			if (request_add_entry(saddr, oid, myid, version) == false) {
				return STG_RES_ESERVER;
			}
		}

		else{
			// replicator has to keep track of this segment, since this node
			// is the home host.
			if(replicator->addSegment(oid, *myAddr, version)==false){ 
				// the desc will be garbage collected
				// by verfs.
				return STG_RES_ESERVER;
			}

			// newly created
			if(replicator->setRepDegree(oid, Replicator::repDegree, version)== false){
				tmpi_error(DBG_INTERNAL, "Set rep degree failed\nnot good\n");
				replicator->removeSegment(oid);
				// the desc will be garbage collected
				// by verfs.
				return STG_RES_ESERVER;
			}

			if(PUSH_TO_REPLICAS){
				replicator->syncObject(oid);
			}
		}
		Stg_Response *res =new_stg_res_create(desc, version);
        /*

        ClientID *cid = new ClientID(11);
        
		if(xput_tracker->addSegment(desc,*cid)){
            tmpi_error(DBG_INTERNAL, "addSegment Success\n");
        }else{
            tmpi_error(DBG_INTERNAL, "addSegment Failure!\n");
            return STG_RES_ESERVER;
        }
        */

        if (!res) {
			return STG_RES_EMEM;
		}
		req->respond(srv, res);
		stg_res_free(res);
	}
	catch (...) {
		return STG_RES_ESERVER;
	}

	return STG_RES_OKAY;
}

int Server::handle_open(Stg_Srv_Request *req, int myid)
{
	try {
		tmpi_error(DBG_VERBOSE, "open request\n");
		ObjID oid=req->request()->u.u_open.oid;
		int flags = req->request()->u.u_open.flags;
		int version = -1;

		int desc;

		if (flags == SOR_RD_ONLY)
			desc = verfs->open(oid, "r", version);
		else 
			desc = verfs->open(oid, "w", version);

		if (desc >= 0) {
			// opened successfully.
			Stg_Response *res = new_stg_res_open(desc);
			if (!res) {
				return STG_RES_EMEM;
			}
			req->respond(srv, res);
			stg_res_free(res);
		}
		else {
			// try to locate it in the local store:
			tmpi_error(DBG_VERBOSE, "Open miss, checking srv catalog.\n");

			vector<HostID> hids;
			if (!sc->get_hid(oid, hids)) {
				// TODO: might some other errors? Such as "NOT-FOUND"?
				tmpi_error(DBG_VERBOSE, "Not found \n");
				return STG_RES_ENOTFOUND;
			}

			bool use_tcp=req->use_tcp();

			// hid to address conversion:
			ServerAddr0 *saddrs = new ServerAddr0[hids.size()];
			int index = 0;
			for(vector<HostID>::const_iterator i=hids.begin(); 
				i!=hids.end(); i++)
			{
				if (hmap->get_address((HostID&)(*i), *(ServerAddr *)(&saddrs[index])) == true){
					// knows the address of that host.
					index++;
				}
			}

			Stg_Response *res = new_stg_res_redir1(saddrs, index, use_tcp);

			if (!res) {
				return STG_RES_EMEM;
			}
			
			req->respond(srv, res);
			stg_res_free(res);

			// send the payload if tcp is being used
 			if (use_tcp) {
				off_seg_t buf_len = index * sizeof(ServerAddr);

 				if (req->write_more((char *)saddrs, buf_len) == false)
 				{
 					tmpi_error(DBG_VERBOSE, "Cannot send data, network problem?\n");
				}				
 			}
		}
	}
	catch (...) {
		return STG_RES_ESERVER;
	}

	return STG_RES_OKAY;
}

int Server::handle_close(Stg_Srv_Request *req, int myid)
{
	ObjID oid;
	int version = -1;
	int oldVersion = -1;

	try {
		tmpi_error(DBG_VERBOSE, "close request");
		int desc=req->request()->u.u_close.desc;

		oid = verfs->getGUID(desc);
		oldVersion = verfs->getVersion(oid);

		if (verfs->close(desc, version) == 0) {
			tmpi_error(DBG_VERBOSE, "(%d)\n", version);
			// send the new version number back (may be useful, but
			// not used now)
			Stg_Response *res = new_stg_res_close(version);

			if (!res) {
				return STG_RES_EMEM;
			}

			if(!PUSH_TO_REPLICAS){
				// respond immediately
				req->respond(srv, res);
				stg_res_free(res);
			}

			// now the version may be changed, we need to notify the homehost
			// about this change.
			if(oldVersion != version){
				request_notify_version_ch(oid, version);
			}			

			if(PUSH_TO_REPLICAS){
				// dont respond immediately
				req->respond(srv, res);
				stg_res_free(res);
				
			}

			// there is a problem here but the server works fine by chance in
			// simple situations, that is: suppose a node joins and receives
			// list of objects, it thinks all other servers have version 0,
			// because the version numbers are not sent (TODO), then it chooses
			// one of the existing nodes to read from, and at the time of
			// close it learns the real version, then makes an update
			// again. Solution: handle_get_range should also send the version
			// numbers together with the ObjIDs. (Aziz)
			
		}
		else {
			// TODO: are there other possible errors?
			return STG_RES_ESERVER;
		}
	}
	catch (...) {
		return STG_RES_ESERVER;
	}

	return STG_RES_OKAY;
}

int Server::handle_write_internal(int desc, off_seg_t offset, char *buf, off_seg_t len, off_seg_t & written_bytes)
{
	// TODO:
	// what are possible return codes? 
	// do some convertion here? 
	// How about the handling of redirection?
	int ret = verfs->write(desc, offset, buf, len);
	if (ret < 0) {
		return STG_RES_ESERVER;
	}
	written_bytes = (off_seg_t)ret;

	return STG_RES_OKAY;
}

int Server::handle_write_small(Stg_Srv_Request *srv_req, int myid)
{
	try {
		tmpi_error(DBG_VERBOSE, "small write request\n");
		Stg_Request *req=srv_req->request();
		int desc=req->u.u_write_small.desc;
		off_seg_t offset=req->u.u_write_small.offset;
		off_seg_t len=req->u.u_write_small.length;
		char *buf=req->u.u_write_small.data;
		//ClientID cid=req->u.u_write_small.cid;

		off_seg_t written_bytes;

		int ret = handle_write_internal(desc, offset, buf, len, written_bytes);
		if (ret != STG_RES_OKAY) {
			// TODO: what happens?
			return ret;
		}else{
           /*
            ClientID *cid = new ClientID(11);
		    if(xput_tracker->addXput(desc,*cid,written_bytes)){
                tmpi_error(DBG_INTERNAL, "addXput Success\n");
            }else{
                tmpi_error(DBG_INTERNAL, "addXput Failure!\n");
                return STG_RES_EMEM;
            }
            */
		}

		// successful
		Stg_Response *res = new_stg_res_write(written_bytes);
		if (!res) {
			return STG_RES_EMEM;
		}

		//update the data tracker
		
		srv_req->respond(srv, res);
		stg_res_free(res);
	}
	catch (...) {
		return STG_RES_ESERVER;
	}

	return STG_RES_OKAY;
}

# define STG_SRV_TCP_READ_MAX_RETRY 3

int Server::handle_write_big(Stg_Srv_Request *srv_req, int myid)
{
	try {
		tmpi_error(DBG_VERBOSE, "big write request\n");
		Stg_Request *req=srv_req->request();
		bool use_tcp=srv_req->use_tcp();
		if (!use_tcp) {
			// bad request, should transfer through TCP.
			return STG_RES_ESIZE;
		}
		int desc=req->u.u_write_big.desc;
		off_seg_t offset=req->u.u_write_big.offset;
		const off_seg_t buf_size=128*1024; // 128K buffer.
		char buf[buf_size];
		off_seg_t length=req->u.u_write_big.length;
		off_seg_t remain_len=length;
		//lientID cid = req->u.u_write_big.cid;
		int retry = 0;
		int errnum=STG_RES_WRITE;
		while (remain_len>0) {
			off_seg_t len=(remain_len>buf_size)?buf_size:remain_len;
			off_seg_t read_bytes;
			// we keep reading from network and writing to the VFS until
			// we cannot do so any more.
			if (srv_req->read_more(buf, len, read_bytes) == false) {
				errnum=STG_RES_ENETWORK;
				break;
			}
			if (read_bytes > 0) {
				retry = 0; // reset retry.
				off_seg_t written_bytes;
				int ret = handle_write_internal(desc, offset, buf, read_bytes, written_bytes);
				// cannot write?
				if (ret != STG_RES_OKAY) {
					errnum = ret;
					break;
				}else{
				        //xput_tracker->addXput(desc,cid,written_bytes);
				}
				remain_len -= written_bytes;
				offset+=written_bytes;
				if (written_bytes != read_bytes) {
					// should never happen like this, out of space?
					errnum = STG_RES_ESPACE;
					break;
				}
			}
			else {
				// got zero bytes data, retry again.
				retry++;
				// avoiding client hang the server by keep sending
				// 0-bytes packets.
				if (retry > STG_SRV_TCP_READ_MAX_RETRY) {
					errnum = STG_RES_ETOUT;
					break;
				}
			}
		}

		if (remain_len > 0) {
			srv_req->set_status(errnum);
		}

		off_seg_t written_bytes = length - remain_len;
		if (written_bytes == 0) {
			return errnum;
		}

		// successful
		Stg_Response *res = new_stg_res_write(written_bytes);
		if (!res) {
			return STG_RES_EMEM;
		}
		
		srv_req->respond(srv, res);
		stg_res_free(res);
	}
	catch (...) {
		return STG_RES_ESERVER;
	}

	return STG_RES_OKAY;
}

int Server::handle_append_internal(int desc, char *buf, off_seg_t len, off_seg_t & append_bytes)
{
	// TODO:
	// what are possible return codes? 
	// do some convertion here? 
	// How about the handling of redirection?
	int ret = verfs->append(desc, buf, len);
	if (ret < 0) {
		return STG_RES_ESERVER;
	}
	append_bytes = (off_seg_t)ret;

	return STG_RES_OKAY;
}

int Server::handle_append_small(Stg_Srv_Request *srv_req, int myid)
{
	try {
		tmpi_error(DBG_VERBOSE, "small append request\n");
		Stg_Request *req=srv_req->request();
		int desc=req->u.u_append_small.desc;
		off_seg_t len=req->u.u_append_small.length;
		char *buf=req->u.u_append_small.data;
		//ClientID cid = req->u.u_append_small.cid;
		off_seg_t append_bytes;

		int ret = handle_append_internal(desc, buf, len, append_bytes);
		if (ret != STG_RES_OKAY) {
			// TODO: what happens?
			return ret;
		}else{
			//xput_tracker->addXput(desc, cid, append_bytes);
		}

		// successful
		Stg_Response *res = new_stg_res_append(append_bytes);
		if (!res) {
			return STG_RES_EMEM;
		}
		
		srv_req->respond(srv, res);
		stg_res_free(res);
	}
	catch (...) {
		return STG_RES_ESERVER;
	}

	return STG_RES_OKAY;
}

int Server::handle_append_big(Stg_Srv_Request *srv_req, int myid)
{
	try {
		tmpi_error(DBG_VERBOSE, "big append request\n");
		Stg_Request *req=srv_req->request();
		bool use_tcp=srv_req->use_tcp();
		if (!use_tcp) {
			// bad request, should transfer through TCP.
			return STG_RES_ESIZE;
		}
		int desc=req->u.u_append_big.desc;
		const off_seg_t buf_size=128*1024; // 128K buffer.
		char buf[buf_size];
		off_seg_t length=req->u.u_append_big.length;
		off_seg_t remain_len=length;
		//ClientID cid = req->u.u_append_big.cid;
		int retry = 0;
		int errnum=STG_RES_APPEND;
		while (remain_len>0) {
			off_seg_t len=(remain_len>buf_size)?buf_size:remain_len;
			off_seg_t read_bytes;
			// we keep reading from network and writing to the VFS until
			// we cannot do so any more.
			if (srv_req->read_more(buf, len, read_bytes) == false) {
				errnum=STG_RES_ENETWORK;
				break;
			}
			if (read_bytes > 0) {
				retry = 0; // reset retry.
				off_seg_t append_bytes;
				int ret = handle_append_internal(desc, buf, read_bytes, append_bytes);
				// cannot append?
				if (ret != STG_RES_OKAY) {
					errnum = ret;
					break;
				}else{
					//xput_tracker->addXput(desc, cid, append_bytes);
				}
				remain_len -= append_bytes;
				if (append_bytes != read_bytes) {
					// should never happen like this, out of space?
					errnum = STG_RES_ESPACE;
					break;
				}
			}
			else {
				// got zero bytes data, retry again.
				retry++;
				// avoiding client hang the server by keep sending
				// 0-bytes packets.
				if (retry > STG_SRV_TCP_READ_MAX_RETRY) {
					errnum = STG_RES_ETOUT;
					break;
				}
			}
		}

		if (remain_len > 0) {
			srv_req->set_status(errnum);
		}

		off_seg_t append_bytes = length - remain_len;
		if (append_bytes == 0) {
			return errnum;
		}

		// successful
		Stg_Response *res = new_stg_res_append(append_bytes);
		if (!res) {
			return STG_RES_EMEM;
		}
		
		srv_req->respond(srv, res);
		stg_res_free(res);
	}
	catch (...) {
		return STG_RES_ESERVER;
	}

	return STG_RES_OKAY;
}

int Server::handle_remove(Stg_Srv_Request *req, int myid)
{
	try {
		tmpi_error(DBG_VERBOSE, "remove request\n");

		ObjID oid=req->request()->u.u_remove.oid;
		int ret = verfs->remove(oid);

		if (ret == 0) {
			// removed successfully.
			if (si->remove(oid) == false) {
				return STG_RES_ESERVER;
			}

			HostID home_host;
			if (hmap->get_home(oid, home_host) == false) {
				return STG_RES_ESERVER;
			}

			if (memcmp(&home_host, &myHid, sizeof(HostID)) != 0) {
				ServerAddr saddr;
				if (hmap->get_address(home_host, saddr) == false) {
					return STG_RES_ESERVER;
				}
				if (request_del_entry(saddr, oid, myid) == false) {
					return STG_RES_ESERVER;
				}
			}

			req->respond_okay(srv);
		}
		else {
			// try to locate it in the local store:
			tmpi_error(DBG_VERBOSE, "Not found locally, checking srv catalog.\n");

			vector<HostID> hids;
			if (!sc->get_hid(oid, hids)) {
				// TODO: might some other errors? 
				return STG_RES_ENOTFOUND;
			}

			bool use_tcp=req->use_tcp();

			// hid to address conversion:
			ServerAddr0 *saddrs = new ServerAddr0[hids.size()];
			int index = 0;
			for(vector<HostID>::const_iterator i=hids.begin(); 
				i!=hids.end(); i++)
			{
				if (hmap->get_address((HostID&)(*i), *(ServerAddr *)(&saddrs[index]))
					    == true){
					// knows the address of that host.
					index++;
				}
			}
			Stg_Response *res = new_stg_res_redir1(saddrs, index, use_tcp);
			
			if (!res) {
				return STG_RES_EMEM;
			}
			
			req->respond(srv, res);
			stg_res_free(res);

			// send the payload if tcp is being used
 			if (use_tcp) {
				off_seg_t buf_len = index * sizeof(ServerAddr);

 				if (req->write_more((char *)saddrs, buf_len) == false)
 				{
 					tmpi_error(DBG_VERBOSE, "Cannot send data, network problem?\n");
				}				
 			}
		}
	}
	catch (...) {
		return STG_RES_ESERVER;
	}

	return STG_RES_OKAY;
}

int Server::handle_truncate_internal(int desc, off_seg_t size, off_seg_t &new_size)
{
	int ret = verfs->truncate(desc, size);
	if (ret < 0) {
		tmpi_error(DBG_VERBOSE, "Truncate to %d bytes failed.\n", size);
		return STG_RES_ESERVER;
	}
	new_size=(off_seg_t)ret;
	return STG_RES_OKAY;
}

int Server::handle_truncate(Stg_Srv_Request *req, int myid)
{
	try {
		tmpi_error(DBG_VERBOSE, "truncate request\n");
		int desc=req->request()->u.u_truncate.desc;
		off_seg_t size=req->request()->u.u_truncate.size;
		off_seg_t new_size;

		int ret = handle_truncate_internal(desc, size, new_size);

		if (ret == STG_RES_OKAY) {
			// truncated successfully.
			Stg_Response *res = new_stg_res_truncate(new_size);
			if (!res) {
				return STG_RES_EMEM;
			}
			
			req->respond(srv, res);
			stg_res_free(res);
		}
		else {
			return ret;
		}
	}
	catch (...) {
		return STG_RES_ESERVER;
	}

	return STG_RES_OKAY;
}

int Server::handle_add_entry(Stg_Srv_Request *srv_req, int myid)
{
	try {
		tmpi_error(DBG_VERBOSE, "add entry request\n");
		Stg_Request *req=srv_req->request();
		HostID hid=req->u.u_add_entry.hid;
		ObjID oid=req->u.u_add_entry.oid;
		int version=req->u.u_add_entry.version;

		if (sc->add(oid, hid) == false) {
			// TODO: what happens?
			//return STG_RES_ESERVER;

			// I (Aziz) commented out this return because this happens
			// when a replica notifies this server(since the object is
			// already in the catalog)
		}

		// Checking if this host is the homehost - actually this check
		// shouldn't be necessary because receiving this request
		// should mean that this host is the home.
		HostID home_host;
		if (hmap->get_home(oid, home_host) == false) {
			return STG_RES_ESERVER;
		}

		if (memcmp(&home_host, &myHid, sizeof(HostID)) == 0) {
	
			// replicator has to keep track of this segment, since
			// this node is the home host.
			if(replicator->addSegment(oid, *myAddr, version)==false){
				// the desc will be garbage collected
				// by verfs.
				return STG_RES_ESERVER;
			}
			//xput_tracker->addSegment(oid, *myAddr);
		}
		srv_req->respond_okay(srv);
	}
	catch (...) {
		return STG_RES_ESERVER;
	}

	return STG_RES_OKAY;
}

int Server::handle_del_entry(Stg_Srv_Request *srv_req, int myid)
{
	try {
		tmpi_error(DBG_VERBOSE, "del entry request\n");
		Stg_Request *req=srv_req->request();
		HostID hid=req->u.u_del_entry.hid;
		ObjID oid=req->u.u_del_entry.oid;

		if (sc->remove(oid, hid) == false) { 
			// TODO: what happens?
			return STG_RES_ESERVER;
		}

		// Checking if this host is the homehost - actually this check
		// shouldn't be necessary because receiving this request
		// should mean that this host is the home.
		HostID home_host;
		if (hmap->get_home(oid, home_host) == false) {
			return STG_RES_ESERVER;
		}

		if (memcmp(&home_host, &myHid, sizeof(HostID)) == 0) {
	
			// replicator keeps track of this segment, since this node
			// is the home host.
			if(replicator->removeSegment(oid)==false){
				return STG_RES_ESERVER; // what shall I do?
			}

		}
		srv_req->respond_okay(srv);
	}
	catch (...) {
		return STG_RES_ESERVER;
	}

	return STG_RES_OKAY;
}

// It seems that there is no way for us to send the data out
// on-the-fly. (We need to implement something similar to
// http's chunked transfer to do that.)
// So I will just allocate a buffer big enough to hold everything.
// That is, if we use UDP, the buffer will be 60K; if it is TCP,
// the buffer will be sizeof(ObjID) * #local objects.
int Server::handle_get_range(Stg_Srv_Request *req, int myid)
{
	ObjID *buf=NULL;
	try {
		VecURRange ranges;

		if (hmap->get_control_ranges(req->request()->u.u_get_range.hid, \
				ranges) == false) {
			throw STG_RES_ENOTFOUND;
		}

		off_seg_t buf_len;
		bool use_tcp=req->use_tcp();
		if (use_tcp) {
			buf_len=sizeof(ObjID)*si->get_size();
		}
		else {
			buf_len=STG_MAX_UDP_SIZE;
		}

		buf=(ObjID *)malloc(buf_len);
		if (!buf) {
			throw STG_RES_EMEM;
		}
		ObjID *p=buf;
		int max_oids=(int)(buf_len/sizeof(ObjID));
		int remain_oids=max_oids;

		for (unsigned int i=0; (i<ranges.size()) && (remain_oids>0); i++) {
			int real_size = si->get_keys_in_range(ranges[i], p, remain_oids);
			p+=real_size;
			remain_oids-=real_size;
		}

		if ((!use_tcp) && (remain_oids <= 0)) {
			// UDP channel, response may be too big.
			throw STG_RES_ESIZE;
		}

		int n_objs=max_oids-remain_oids;
		Stg_Response *res=new_stg_res_get_range(myHid, buf, n_objs, use_tcp);
		if (!res) {
			throw STG_RES_EMEM;
		}
		req->respond(srv, res);
		stg_res_free(res);
		if (use_tcp) {
			buf_len=n_objs*sizeof(ObjID); // resize to the actual size.
			if (req->write_more((char *)buf, buf_len) == false)
			{
				tmpi_error(DBG_VERBOSE, "Cannot send data, network problem?\n");
			}
		}

		free(buf);
	}
	catch (int errnum) {
		if (buf)
			free(buf);
		return errnum;
	}
	catch (...) {
		if (buf)
			free(buf);
		return STG_RES_ESERVER;
	}

	return STG_RES_OKAY;
}

int Server::handle_read_internal(int desc, off_seg_t offset, char *buf, off_seg_t len, off_seg_t & read_bytes)
{
	// TODO:
	// what are possible return codes? 
	// do some convertion here? 
	// How about the handling of redirection?
	int ret = verfs->read(desc, offset, buf, len);
	if (ret < 0) {
		return STG_RES_ESERVER;
	}
	read_bytes = (off_seg_t)ret;

	return STG_RES_OKAY;
}

// TODO: Need to get from VerFS the actual size I will
// send. Then do on-the-fly send. (Or chunked-transfer).
// Now I just preallocate everything.
int Server::handle_read(Stg_Srv_Request *srv_req, int myid)
{
	Stg_Response *res=NULL;
	char *buf=NULL;
	char *p=NULL;
	try {
		bool use_tcp=srv_req->use_tcp();
		Stg_Request *req=srv_req->request();
		off_seg_t length=req->u.u_read.length;
		off_seg_t offset=req->u.u_read.offset;
		off_seg_t desc=req->u.u_read.desc;
		//ClientID cid=req->u.u_read.cid;

		// Addjust the actual length of the read request.
		// We may add desc checking here too.
		// verfs->adjust_length(desc, offset, length)
		res=new_stg_res_read(length, use_tcp);
		if (!res) {
			return STG_RES_EMEM;
		}
		// Setting the buffer pointer.
		if (use_tcp) {
			buf=(char *)malloc(length);
			if (!buf) {
				throw STG_RES_EMEM;
			}
			p=buf;
		}
		else {
			// use UDP
			if (length>STG_MAX_UDP_SIZE) {
				// cannot use UDP to transfer large data.
				throw STG_RES_ESIZE;
			}
			p=(char *)res->payload();
		}
		off_seg_t read_bytes;
		int ret=handle_read_internal(desc, offset, p, length, read_bytes);
		if (ret != STG_RES_OKAY) {
	  	        tmpi_error(DBG_VERBOSE, "read failed.\n");
			throw ret;
		}else{
                	//xput_tracker->addXput(desc,cid,read_bytes);
                }

        	// tmpi_error(DBG_VERBOSE, "%d bytes read.\n", read_bytes);
		adjust_stg_res_read(res, read_bytes, use_tcp);
		/*******
		*if (use_tcp) {
		*	res->u.u_read_big.read_bytes=read_bytes;
		*}
		*else {
		*	res->u.u_read_small.read_bytes=read_bytes;
		*}
		******/
		srv_req->respond(srv, res);
		stg_res_free(res);
		res=NULL;
		if (use_tcp) {
			srv_req->write_more(buf, read_bytes);
			free(buf);
			buf=NULL;
		}
	}
	catch (int errnum) {
        tmpi_error(DBG_VERBOSE, "read failed, error=%d.\n", errnum);
		if (res)
			stg_res_free(res);
		if (buf)
			free(buf);
		return errnum;
	}
	catch (...) {
		if (res)
			stg_res_free(res);
		if (buf)
			free(buf);
		return STG_RES_ESERVER;
	}

	return STG_RES_OKAY;
}

int Server::handle_sync_host(Stg_Srv_Request *srv_req, int myid)
{
	ObjID oid;		
	unsigned int bufSize = 1024*1024*8;
	char* buffer;
	buffer = new char[bufSize];
	
	try {
		tmpi_error(DBG_VERBOSE, "synchronize request received\n");
		Stg_Request *req=srv_req->request();

		// send OK response immediately if working lazily
		if(!PUSH_TO_REPLICAS){
			srv_req->respond_okay(srv);
		}
		
		// now start getting the updates (Assuming this host is the one 
		// whose version is lower)
		
		int offset = 0;
		Descriptor remoteDesc;
		int localDesc;
		int response;
		//int len;
		off_seg_t readBytes;
		int stat;
		int version;
		int newVersion = req->u.u_sync_host.otherV;
		
		oid = req->u.u_sync_host.oid;

		// create locally for write
		localDesc = verfs->create(oid, version);
		if (localDesc < 0){

			// the object exists so we open to update it with the new
			// version.
			localDesc = verfs->open(oid, "w", version);
			
			if(localDesc < 0){
				tmpi_error(DBG_INTERNAL, "Verfs failed\n");
				delete buffer;
				return STG_RES_ESERVER;
			}
		}
		
		// open the remote object (open for read only)
		ServerAddr sa = req->u.u_sync_host.sa;
		response = client->open(sa, oid, remoteDesc, SOR_RD_ONLY);
		
		if(response == STG_RES_OKAY){
			bool lastTime = false;
			
			for(offset=0; lastTime != true; offset+=bufSize){
				// send read request
				client->read(remoteDesc, offset, buffer, bufSize, readBytes);
				if (readBytes == 0)
					break;		// finished reading

				if((unsigned int)readBytes < bufSize) 
					lastTime = true; // read the last block
				
				stat = verfs->write(localDesc, offset, buffer, readBytes);
				if(stat < 0){
					tmpi_error(DBG_INTERNAL, "Verfs failed.\n");
					delete buffer;
					return STG_RES_ESERVER;
				}
			}
		}

		verfs->close(localDesc, version);

		// TODO:
		// We need to do version adjustment here by instructing verfs
		// the version number to set.

		response = client->close(remoteDesc);
		
		if(response != STG_RES_OKAY){
			tmpi_error(DBG_INTERNAL, "close failed.\n");
			delete buffer;
			return STG_RES_ESERVER;
		}

		if(PUSH_TO_REPLICAS){
			// now respond ok since not working lazily
			srv_req->respond_okay(srv);
		}

		// notify home host
		//tmpi_error(DBG_INTERNAL, "new version:%d.\n", newVersion);
		request_notify_version_ch(oid, newVersion);	
	
	}
	catch (...) {
		delete buffer;
		return STG_RES_ESERVER;
	}

	delete buffer;
	return STG_RES_OKAY;
}

int Server::handle_notify_ver_ch(Stg_Srv_Request *srv_req, int myid){

	try {
		Stg_Request *req=srv_req->request();

		ObjID oid = req->u.u_notify_ver_ch.oid;
		ServerAddr sa = req->u.u_notify_ver_ch.sa;
		int version = req->u.u_notify_ver_ch.version;

		tmpi_error(DBG_VERBOSE, "version change notification received (%d)\n", 
				   version);

		if(replicator->setVersion(oid, sa, version) == true){
			// send OK response
			
			if(PUSH_TO_REPLICAS){
				replicator->syncObject(oid);
			}
			
			srv_req->respond_okay(srv);
		}

		else{
			srv_req->respond_error(srv, STG_RES_ESERVER);
			tmpi_error(DBG_VERBOSE, "Replicator failed.\n");
		}

	}catch (...) {
		tmpi_error(DBG_VERBOSE, "Exception.\n");
		return STG_RES_ESERVER;
	}

	return STG_RES_OKAY;
}

void Server::handle_request(Stg_Srv_Request *srv_req, int myid)
{
	Stg_Request *req=srv_req->request();
	int ret=STG_RES_OKAY;

	switch (req->header.tag) {
		case STG_REQ_CREATE1:
			ret=handle_create1(srv_req, myid); break;
		case STG_REQ_OPEN:
			ret=handle_open(srv_req, myid); break;
		case STG_REQ_CLOSE:
			ret=handle_close(srv_req, myid); break;
		case STG_REQ_READ:
			ret=handle_read(srv_req, myid); break;
		case STG_REQ_WRITE_SMALL:
			ret=handle_write_small(srv_req, myid); break;
		case STG_REQ_WRITE_BIG:
			ret=handle_write_big(srv_req, myid); break;
		case STG_REQ_APPEND_SMALL:
			ret=handle_append_small(srv_req, myid); break;
		case STG_REQ_APPEND_BIG:
			ret=handle_append_big(srv_req, myid); break;
		case STG_REQ_REMOVE:
			ret=handle_remove(srv_req, myid); break;
		case STG_REQ_TRUNCATE:
			ret=handle_truncate(srv_req, myid); break;
		case STG_REQ_ADD_ENTRY:
			ret=handle_add_entry(srv_req, myid); break;
		case STG_REQ_DEL_ENTRY:
			ret=handle_del_entry(srv_req, myid); break;
		case STG_REQ_GET_RANGE:
			ret=handle_get_range(srv_req, myid); break;
    	case STG_REQ_SYNC_HOST:
			ret=handle_sync_host(srv_req, myid); break;
	   	case STG_REQ_NOTIFY_VER_CH:
			ret=handle_notify_ver_ch(srv_req, myid); break;

		case STG_REQ_CREATE2:
		case STG_REQ_DEL_RANGE:
        case STG_REQ_DUMP:
			tmpi_error(DBG_INTERNAL, "Del range not implemented.\n");
			ret=STG_RES_ESERVER;
			break;
		default:
            tmpi_error(DBG_VERBOSE, "Unknown client request.\n", req->header.tag);
			throw STG_RES_ECLIENT;
	}

	if (ret != STG_RES_OKAY) {
		// the request has not been responded yet.
		throw ret;
	}

	return;
}

/**
 * Checks if all of the threads is running
 * @return true if all threads are running, false otherwise
 */
bool Server::is_running()
{
	return (udpDispatcher->is_running() && 
			tcpDispatcher->is_running() && 
			scUpdater->is_running() &&
			tp->is_running() &&
			/*replicator->is_running() &&*/
			publisher->is_running());
}

/**
 * Checks if any of the threads is running
 * @return true if any thread is running, false otherwise
 */
bool Server::is_any_running()
{
	return (udpDispatcher->is_running() || 
			tcpDispatcher->is_running() || 
			scUpdater->is_running() ||	
			tp->is_running() ||
			/*replicator->is_running() ||*/
			publisher->is_running());
}

/**
 * Called by the publisher when a host joins the system. Adds host to the
 * server index updater's list.
 * @param hid id of the host to be added.
 */
void Server::notify_host_join(HostID & hid)
{
	if (memcmp(&hid, &myHid, sizeof(hid)) != 0)
		scUpdater->addHost(hid);
	replicator->notify_host_join();
}

/**
 * Called by the publisher when a host leaves the system. Removes host from the
 * server index updater's list.
 * @param hid id of the host to be added.
 */
void Server::notify_host_leave(HostID & hid, ServerAddr &sad)
{
	replicator->notify_host_leave(sad);
	scUpdater->removeHost(hid);	
}

HostID Server::get_hid(void) const 
{ 
	return myHid; 
}

/**
 * direct from static to instance 
 */
char *server_stub_handler(void *p, void *r, int myid)
{
	Stg_Srv_Request * req=(Stg_Srv_Request *)r;
	SharedState *ss=(SharedState *)p;
	try {
		// error are returned through exceptions.
		ss->server->handle_request(req, myid);
		delete req;
		return NULL;
	}
	catch (int errnum) {
		req->respond_error(ss->server->srv, errnum);
		delete req;
		return NULL;
	}
	catch (...) {
		tmpi_error(DBG_INTERNAL, "Caught unknown exception.\n");
		delete req;
		return NULL;
	}
}

/**
 * Constructs server
 */
Server::Server(SorrentoConf &sconf, Proxy *p) :
	udpDispatcher(NULL),
	tcpDispatcher(NULL),
	scUpdater(NULL),
	publisher(NULL),
	tp(NULL),
	replicator(NULL),
	//xput_tracker(NULL),
	srv(NULL),
	verfs(NULL),
	si(NULL),
	sc(NULL),
	hmap(NULL),
	ss(this),
	myAddr(NULL),
	client(NULL)
{
	if (!p)
		throw string("Server::Server: Invalid pointer to the Proxy object.");

	for (int i=0; i<STG_TPOOL_SIZE; i++) {
		udp_clis[i]= new Udp_cli(NULL, SOCK_PORT_ANY, STG_SRV_UDP_TOUT);
	}

	// ss.init(this);

	si = new UnitRing(false);
	sc = new SrvCatalog();
	srv = new Udp_srv(NULL, SOCK_PORT_ANY);

	if (!srv || srv->port()==-1) { 
		throw string("Server::Server: Cannot create UDP server.");
	}

	udpDispatcher = new UdpDispatcher(srv, this);
	tcpDispatcher = new TcpDispatcher(SOCK_PORT_ANY, this);
	publisher = new Publisher(sconf, get_tcp_port(), get_udp_port());

	myHid= publisher->get_host_id();

	verfs = new Verfs(publisher->get_mnt_ptrs(), sconf.home_dir());

	publisher->setVerfs(verfs);

	hmap = p->get_host_map_by_ident(publisher->get_ident()); 
	if (!hmap)
		throw string("Server::Server: Cannot get host map for ")+publisher->get_ident()+".";

	tp = new Thread_Pool(NULL, STG_TPOOL_SIZE, &ss, &server_stub_handler);

	// set server callback
	hmap->set_cbi(new SrvCBImp(this));

	char ipAdd[16];
	//hostent *h = gethostbyname(srv->host());
	//for (int i = 0; h->h_addr_list[i] != 0; i++){ 
	//	memcpy(ipAdd, inet_ntoa (*(struct in_addr *)(h->h_addr_list[i])), 16);
	//	tmpi_error(DBG_VERBOSE, "%s\n",ipAdd);
	//	int a = 0;
	//	while (h->h_aliases[a] != 0){
	//		memcpy(ipAdd, (h->h_aliases[a]), 16);
	//		tmpi_error(DBG_VERBOSE, "%s\n",ipAdd);
	//		a++;
	//	}
	//}

	// just allocate memory, replicator will fill this in when started.
	myAddr = new ServerAddr(ipAdd, get_tcp_port(), get_udp_port());

	replicator = new Replicator(hmap, verfs, &myHid, myAddr);
	Replicator::setRepDegree(sconf.rep_degree());
	tmpi_error(DBG_VERBOSE, "Replication degree: %d\n", Replicator::repDegree);

    //xput_tracker = new Xput_Tracker(hmap);

	// scUpdater passes updates to replicator
	scUpdater = new SrvCatalogUpdater(sc, hmap, myHid, replicator);

	// necessary for sync host
	client = new Client();
}

Server::~Server()
{
	// clear server callback
	SrvCBImp *cbi=(SrvCBImp *)hmap->clr_cbi();
	delete cbi;

	// should be no harm.
	stop();
	finalize();

	delete replicator;
 	delete sc;
 	delete si;
	//delete hmap;				// causes seg.fault
 	delete myAddr;
 	delete verfs;
 	delete publisher;
 	delete tp;
 	delete tcpDispatcher;
 	delete udpDispatcher;
 	delete srv;

	for (int i=0; i<STG_TPOOL_SIZE; i++) {
		delete udp_clis[i];
	}
	delete client;
}


/**
 * Fills the server index and starts the threads
 * @return false if start fails
 */
bool Server::start()
{
	/* Read the objects from hashtable and insert into server index */
	guid_t guid;
	int objCount = 0;
	int version;

	
	/* ask other servers for their data    ??? */
	/* notify other servers for their data ??? */


	if (publisher->start() == false) {
		tmpi_error(DBG_INTERNAL, "Failed to start multicast publisher.\n");
		goto _exit;
	}
	else {
		tmpi_error(DBG_VERBOSE, "Publisher started.\n");
	}

	if (udpDispatcher->start() == false) {
		tmpi_error(DBG_INTERNAL, "Failed to start UDP Dispatcher.\n");
		goto _exit;
	}
	else {
		tmpi_error(DBG_VERBOSE, "UDP Dispatcher started.\n");
	}

	if (tcpDispatcher->start() == false) {
		tmpi_error(DBG_INTERNAL, "Failed to start TCP Dispatcher.\n");
		goto _exit;
	}
	else {
		tmpi_error(DBG_VERBOSE, "TCP Dispatcher started.\n");
	}
    /*
	if (xput_tracker->start() == false) {
		tmpi_error(DBG_INTERNAL, "Failed to start Throughput Tracker.\n");
		goto _exit;
	}
	else {
		tmpi_error(DBG_VERBOSE, "Throughput Tracker started.\n");
	}
    */
	if (tp->start() == false) {
		tmpi_error(DBG_INTERNAL, "Failed to start thread pool.\n");
		goto _exit;
	}
	else {
		tmpi_error(DBG_VERBOSE, "Thread pool started.\n");
	}

	if (scUpdater->start() == false) {
		tmpi_error(DBG_INTERNAL, "Failed to start server index updater.\n");
		goto _exit;
	}
	else {
		tmpi_error(DBG_VERBOSE, "Server index updater started.\n");
	}

	if(Replicator::repDegree > 1){
		
		if ((replicator) && replicator->start() == false) {
			tmpi_error(DBG_INTERNAL, "Failed to start replicator.\n");
			goto _exit;
		}
		else {
			tmpi_error(DBG_VERBOSE, "Replicator started.\n");
		}
	}

	for (bool ret = verfs->getFirst(guid); ret; \
			ret=verfs->getNext(guid)) {
		objCount++;
		si->add(guid);
		version = verfs->getVersion(guid);
		replicator->addSegment(guid, *myAddr, version);
	}

	tmpi_error(DBG_VERBOSE, "Read %d objects from the disk.\n", objCount);


	
	cout << "Server started @ " << srv->host() << ":" << get_udp_port() 
		<<  ":" << get_tcp_port()
		<< endl;

	return true;

_exit:
	stop();
	return false;
}

bool Server::stop(bool right_now)
{
	if (!is_any_running()) {
		return true;
	}

	// the stopping of thread pool is a bit tricky.
	// We first stop the publisher, wait for 10 seconds
	// to make sure peers remove this server from the
	// host map. Then I stop dispatchers to ensure no
	// more requests will flow into the server.
	// Finally, we put an EOQ mark into the thread pool.
	// We then wait for another 10 seconds (at most)
	// if thread pool stops naturally, that will be great,
	// otherwise, I will send the stop signal explicitly.
	if (publisher->is_running()) {
		tmpi_error(DBG_VERBOSE, "Requesting publisher to stop.\n");
		publisher->stop();

		if (!right_now) {
			tmpi_error(DBG_VERBOSE, "Waiting for 10 seconds.\n");
			sleep(10);
		}
	}

	if (tcpDispatcher->is_running()) {
		tmpi_error(DBG_VERBOSE, "Requesting tcpDispatcher to stop.\n");
		tcpDispatcher->stop();
	}
	if (udpDispatcher->is_running()) {
		tmpi_error(DBG_VERBOSE, "Requesting udpDispatcher to stop.\n");
		udpDispatcher->stop();
	}

	if (scUpdater->is_running()) {
		tmpi_error(DBG_VERBOSE, "Requesting Index Updater to stop.\n");
		scUpdater->stop();
	}

	if (replicator && replicator->is_running()) {
		tmpi_error(DBG_VERBOSE, "Requesting replicator to stop.\n");
		replicator->stop();
	}

	if (tp->is_running()) {
		tmpi_error(DBG_VERBOSE, "Putting the EOQ mark into the thread pool.\n");
		tp->enqueue_eoq_mark();

		tmpi_error(DBG_VERBOSE, "Waiting for the thread pool to stop.\n");
		if (!right_now) {
			for (int i=0; i<10; i++) {
				if (tp->is_running())
					sleep(1);
			}
		}
	}

	if (tp->is_running()) {
		tmpi_error(DBG_VERBOSE, "Forcing thread pool to stop.\n");
		tp->stop();
	}

	return true;
}

/**
 * Implementation of UdpDispatcher
 */
void *UdpDispatcher::run()
{
	Udp_srv_req *q;

	while ( (q=udpSrv->wait_for_request()) ) {
		/* stop request is initiated inside server module */
		if (should_quit())
			break;

		// tmpi_error(DBG_VERBOSE, "got request of length %d.\n",  q->length());
		if (!q) {
			tmpi_error(DBG_APP, "System error, server shutdown.\n");
			break;
		}

		/* stop request is initiated outside server module */
		if (q->is_stop_req()) {
			tmpi_error(DBG_VERBOSE, "UDP: Got shutdown request, server shutdown.");
			server->stop();
			break;
		}

		Stg_Srv_Request *req = NULL;
		try {
			req = new Stg_Srv_Request(q);
			if (req->request()->header.tag == STG_REQ_STOP) {
				tmpi_error(DBG_VERBOSE, "UDP: Got shutdown request, server shutdown.");
				req->respond_okay(udpSrv);
				delete req;
				server->stop();
				break;
			}
			if (req->request()->header.tag & STG_REQ_PRIORITY) {
				server->tp->prio_enqueue(req);
			}
			else {
				server->tp->enqueue(req);
			}
		}
		catch (const string &str_err) {
			tmpi_error(DBG_APP, "%s.\n", str_err.c_str());
			if (req)
				delete req;
			else {
				delete q;
			}
		}
	}

	tmpi_error (DBG_VERBOSE, "UDP Dispatcher stoped\n");
	return NULL;
}


/**
 * Sends stop message to UdpDispatcher.  */
int UdpDispatcher::stop()
{
	if (is_running()) {
		CBS_Thread::stop(); // first apply the stop signal on the server.
		Udp_cli cli(NULL, SOCK_PORT_ANY, STG_SRV_UDP_TOUT);
		// server->stop_server_local(STG_USE_TCP, &cli);
		char auth[16]="anonymous";
		cli.stop_srv("localhost",udpSrv->port(), auth, strlen(auth)+1, 3);
	}

	return 0;
}

/**
 * Implementation of TcpDispatcher
 */
void *TcpDispatcher::run()
{
	while ((!should_quit())) {
		int sock = sp_server_get_one(1000);
		// sock = sock_accept(listenSock);	// This does not timeout
		if(sock<0) {
			// tmpi_error(DBG_APP, "failed to accept connections.\n");
			continue;
		}
		else {
			// tmpi_error(DBG_VERBOSE, "client connected.\n");
		}

		Stg_Srv_Request *req = NULL;

		try {
			req = new Stg_Srv_Request(sock);
			if (req->request()->header.tag == STG_REQ_STOP) {
				tmpi_error(DBG_VERBOSE, "TCP: Got shutdown request, server shutdown.");
				req->respond_okay(NULL);
				delete req;
				server->stop();
				break;
			}
			server->tp->enqueue(req);
		}
		catch (const string &err) {
			tmpi_error(DBG_APP, "%s.\n", err.c_str());
			if (req) {
				delete req;
			}
			else {
				// error in constructor
				::close(sock);
			}
		}
	}

	tmpi_error (DBG_VERBOSE, "TCP Dispatcher stoped\n");

	return NULL;
}

int Server::stop_server_local(int hint, Udp_cli *cli)
{
	Stg_Request *req=new_stg_req_stop();
	if (!req)
		return STG_RES_EMEM;
	
	ServerAddr saddr("127.0.0.1", get_tcp_port(), get_udp_port());
	Stg_Req_Handle reqh;
	int ret=stg_start_request_no_payload(cli, saddr, req, reqh, hint);
	stg_req_free(req);

	if (ret != STG_RES_OKAY) {
		return ret;
	}

	Stg_Res_Handle resh;
	ret=stg_wait_response_no_payload(cli, reqh, resh);
	if (ret != STG_RES_OKAY) {
		return ret;
	}

	Stg_Response *res=resh.response();
	if (res->header.tag != STG_RES_OKAY) {
		tmpi_error(DBG_VERBOSE, "Unexpected reply.\n");
		return STG_RES_ECLIENT;
	}

	return STG_RES_OKAY;
}

/**
 * Sends stop messsage to TcpDispatcher
 */
int TcpDispatcher::stop()
{
	if (is_running()) {
		CBS_Thread::stop(); // first apply the stop signal on the server.
		return server->stop_server_local(STG_USE_TCP, NULL);
	}
	return 0;
}
