// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.
 
  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
 
   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.
 
   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
 
   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.
 
   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.
 
  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: server.h,v $ - header file for Sorrento storage 
 *            server object. 
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Aziz Gulbeden, gulbeden@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2004/04/16 23:33:06 $ by $Author: gulbeden $
 * Version :  $Revision: 1.32 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */

# ifndef __SERVER_H__
# define __SERVER_H__

# include "cbs_thread.h"
# include "sock_conn.h"
# include "sorrento_conf.h"
# include "sorrento_types.h"
# include "sp_singleton.h"
# include "subscriber.h"
# include "thread_pool.h"
# include "udp_srv.h"
//# include "xput_tracker.h"

class Server;

struct SharedState {
	Server *server;
	SharedState() : server(NULL) {}
	SharedState(Server *s) : server(s) {}
};

class Client;
class HostMap;
class Publisher;
class Request;
class ServerAddr;
class SrvCatalog;
class SrvCatalogUpdater;
struct Stg_Srv_Request;
class Udp_cli;
class UnitRing;
class Replicator;
class Verfs;

/**
 * Class: ServerDispatcher
 * Listens on UDP for incoming requests and enqueues them.
 */
class UdpDispatcher : public CBS_Thread
{
public:
	UdpDispatcher(Udp_srv *s, Server *srv) : udpSrv(s), server(srv)
	{ }
	int stop();

private:
	virtual void *run();

	Udp_srv *udpSrv;			/* ro respond back */
	Server *server;				/* to enqueue */
};

/**
 * Class: TcpDispatcher
 * Listens on TCP for incoming write requests and enqueues them.
 */
class TcpDispatcher : public CBS_Thread
{
public:
	TcpDispatcher(int port, Server *s) :
		server(s)
	{
		if (sp_server_init(1024, port)<0) {
			throw std::string("TcpDispatcher::TcpDispatcher: Cannot initialize socket pool.");
		}
		
		listenPort = sp_server_get_port();

		if (listenPort < 0) {
			throw std::string("TcpDispatcher::TcpDispatcher: Cannot get bind port.");
		}
	}
	~TcpDispatcher()
	{
		sp_server_finalize();
	}

	int stop();
	int get_port() { return listenPort; }

private:
	virtual void *run();
	Server *server;
	int listenPort;
};

# define STG_TPOOL_SIZE 30

class Server
{
private:
	UdpDispatcher *udpDispatcher;
	TcpDispatcher *tcpDispatcher;
	SrvCatalogUpdater *scUpdater;
	Publisher *publisher;
	Thread_Pool *tp;
	Replicator *replicator;
    //Xput_Tracker *xput_tracker;
    //////////////////////////////////////////////////////////
    // I should have implemented Thread_Pool as an template, 
    // so that it can take any Thread implementation as its
    // working threads AS LONG AS it inherits the TP_Thread
    // interface.
    // But for now, I have to put all thread-specific data
    // structures as arrays in the server object. Possible
    // false sharing!
    Udp_cli *udp_clis[STG_TPOOL_SIZE]; 
	Udp_srv *srv;
	Verfs *verfs;
	UnitRing *si;				// sever index - no wrapper required
	SrvCatalog *sc;
	HostMap *hmap;
	HostID myHid;
	SharedState ss;
	ServerAddr *myAddr;
	Client *client;

	void enqueue_request(Udp_srv_req *q);
	void handle_request(Stg_Srv_Request *req, int myid);
	// small request, small response.
	int handle_create1(Stg_Srv_Request *req, int myid);
	int handle_open(Stg_Srv_Request *req, int myid);
	int handle_close(Stg_Srv_Request *req, int myid);
	int handle_write_small(Stg_Srv_Request *req, int myid);
	int handle_append_small(Stg_Srv_Request *req, int myid);
	int handle_remove(Stg_Srv_Request *req, int myid);
	int handle_truncate(Stg_Srv_Request *req, int myid);
	int handle_add_entry(Stg_Srv_Request *req, int myid);
	int handle_del_entry(Stg_Srv_Request *req, int myid);

	// small request, small/large response.
	int handle_read(Stg_Srv_Request *req, int myid);
	int handle_get_range(Stg_Srv_Request *req, int myid);

	// small/large request, small response.
	int handle_write_big(Stg_Srv_Request *req, int myid);
	int handle_append_big(Stg_Srv_Request *req, int myid);

	int handle_write_internal(int desc, off_seg_t offset, char *buf, off_seg_t len, off_seg_t & written_bytes);
	int handle_append_internal(int desc, char *buf, off_seg_t len, off_seg_t &append_bytes);
	int handle_read_internal(int desc, off_seg_t offset, char *buf, off_seg_t len, off_seg_t &read_bytes);
	int handle_truncate_internal(int desc, off_seg_t size, off_seg_t &new_size);

	int stop_server_local(int hint, Udp_cli *cli);
	bool request_add_entry(ServerAddr &saddr, ObjID &oid, int myid, int version);
	bool request_del_entry(ServerAddr &saddr, ObjID &oid, int myid);
	int  request_notify_version_ch(ObjID &oid, int version);
	int handle_sync_host(Stg_Srv_Request *srv_req, int myid);
	int handle_notify_ver_ch(Stg_Srv_Request *srv_req, int myid);


	friend void *TcpDispatcher::run(); /* so that can enqueue */
	friend int TcpDispatcher::stop(); /* so that can enqueue */
	friend void *UdpDispatcher::run(); /* so that can enqueue */
	friend char *server_stub_handler(void *ss, void *req, int myid);

public:
	Server(SorrentoConf &sconf, Proxy *p);
	~Server();
	bool start();
	bool stop(bool right_now=false);
	bool finalize();
	bool is_running();
	bool is_any_running();
	void notify_host_join(HostID & hid);
	void notify_host_leave(HostID & hid, ServerAddr &sad);
	int get_tcp_port() { return tcpDispatcher->get_port(); }
	int get_udp_port() { return srv->port(); }
	HostID get_hid(void) const;
};

#endif // __SERVER_H__
