// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: sor_client.cc,v $ - the Sorrento FS client class
 *            (implementation). 
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2004/02/19 13:14:51 $
 * Version :  $Revision: 1.13 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

# include <fcntl.h>

# include "sor_client.h"
# include "client.h"
# include "sor_fs.h"
# include "sor_file.h"
# include "../dir_srv/dir_cli_ap_udp.h"

# include <string>

# define CID_NULL 		0x0
# define CID_CHDIR 		0x1
# define CID_MKDIR 		0x2
# define CID_RMDIR 		0x3
# define CID_DELTREE 	0x4
# define CID_READDIR 	0x5
# define CID_READDIR_CONT 	0x6
# define CID_UNLINK 	0x7
# define CID_CREATE 	0x8
# define CID_OPEN 		0x9
# define CID_CLOSE 		0xa
# define CID_READ 		0xb
# define CID_WRITE 		0xc
# define CID_FTRUNCATE 	0xd
# define CID_FSTAT 		0xe
# define CID_STAT 		0xf

/**
 * Constructor.
 * @param ap Directory Service Access Point.
 */
SorClient::SorClient(SorFS *sp) :
	// sp(super),
	hmap(NULL),
	ap(NULL),
	cli(NULL),
	errnum(DIR_RES_OKAY),
	callid(CID_NULL)
{
	if (!sp) 
		throw std::string("SorClient::SorClient: NULL superblock argument.");
	hmap=sp->get_hmap();
	ap=sp->new_dir_cli_ap();
	cli=new Client();
}

/** Destructor. */
SorClient::~SorClient()
{
	SorFS::free_dir_cli_ap(ap);
	delete cli;
}

/**
 * Change the current directory. 
 * Namespace operation.
 * @param path the new directory.
 * @return status (true: success).
 */
bool SorClient::chdir(const char *path)
{
	DIR_Entry entry;
	int ret=cwd.lookup(ap, path, &entry);

	if (ret == DIR_RES_OKAY) {
		if (entry.is_dir(ap)) {
			memcpy(&cwd, &entry, sizeof(DIR_Entry));
			return true;
		}
		else {
			ret=DIR_RES_ENOTDIR; /**< rewrite return code. */
		}
	}

	errnum=ret;
	callid=CID_CHDIR;

	return false;
}

/**
 * Make a new directory.  Namespace operation.
 * @param path the directory to be created.
 * @return status (true: success).
 */
bool SorClient::mkdir(const char *path)
{
	DIR_Entry entry;
	int ret=cwd.mkdir(ap, path, &entry);
	if (ret == DIR_RES_OKAY) {
		return true;
	}

	errnum=ret;
	callid=CID_MKDIR;

	return false;
}

/**
 * Remove an existing directory.  Namespace operation.
 * @param path the directory to be removed.
 * @return status (true: success).
 */
bool SorClient::rmdir(const char *path)
{
	int ret=cwd.rmdir(ap, path);
	if (ret == DIR_RES_OKAY) {
		return true;
	}

	errnum=ret;
	callid=CID_RMDIR;

	return false;
}

/**
 * Remove a whole directory tree.  Namespace operation.
 * @param path the root of directory (or file) to be removed.
 * @return status (true: success).
 */
/**
bool deltree(char *path)
{
	int ret=cwd.deltree(ap, path);
	if (ret == DIR_RES_OKAY) {
		return true;
	}

	errnum=ret;
	callid=CID_DELTREE;

	return false;
}
**/

/**
 * Read the content of a directory. Namespace operation.
 * @param path the directory to be read.
 * @param list the list of entries in the directory.
 * @param buf_len the memory consumption constraint. 
 * @param cookie the cookie used for subsequent readdir_cont() calls.
 * @return status (true: success).
 */
bool SorClient::readdir(const char *path, DIR_List &list, int buf_len, DIR_Entry *cookie)
{
	int ret=cwd.lookup(ap, path, cookie);

	if (ret != DIR_RES_OKAY) {
		goto _exit;
	}

	if (!cookie->is_dir(ap)) {
		ret = DIR_RES_ENOTDIR;
		goto _exit;
	}

	ret=cookie->readdir(ap, list, buf_len);
	if ( (ret != DIR_RES_OKAY) ) {
		goto _exit;
	}

	return true;

_exit:
	errnum=ret;
	callid=CID_READDIR;

	return false;
}

/**
 * Check whether we need to make another readdir_cont() call. 
 * @param cookie the cookie from previous call of readdir or readdir_cont.
 * @return status (true: yes).
 */
bool SorClient::is_eod(const DIR_Entry *cookie) const
{
	return cookie->is_eod();
}

/** 
 * Continue the previous reading.
 * @param cookie the cookie used for subsequent readdir_cont() calls.
 * @param list the list of entries in the directory.
 * @param buf_len the memory consumption constraint. 
 * @return status (true: success).
 */
bool SorClient::readdir_cont(DIR_Entry *cookie, DIR_List &list, int buf_len)
{
	int ret=cookie->readdir_cont(ap, list, buf_len); 
	if ( (ret != DIR_RES_OKAY) && (ret != DIR_RES_EEND) ) {
		errnum=ret;
		callid=CID_READDIR;

		return false;
	}

	return true;
}

/**
 * Remove an existing file.  Namespace and storage operation.
 * @param path to the file to be removed.
 * @return status (true: success).
 */
bool SorClient::unlink(const char *path)
{
	// retrieve the entry first.
	DIR_Entry entry;
	int ret=cwd.lookup(ap, path, &entry);
	if (ret != DIR_RES_OKAY) {
		goto _exit;
	}

	// store away the guid
	// CHECK: should I check return value of get_guid??
	guid_t guid;
	entry.get_guid(ap, guid);
	// ObjID oid=guid;

	// then remove the directory entry (possibly with the guard 
	// of entry later).
	ret=cwd.remove(ap, path);
	if (ret != DIR_RES_OKAY) {
		goto _exit;
	}

	// finally, remove the meta object and the data segments.
	ret=SorFile::unlink(cli, hmap, guid);
	if (ret != FILE_RES_OKAY) {
		goto _exit;
	}

	return true;

_exit:
	errnum=ret;
	callid=CID_UNLINK;

	return false;
}

/**
 * Initialize the DIR_Param structure for use of close().
 * @param param the DIR_Param struct to be set.
 * @param guid the guid.
 * @param size the size.
 */
void SorClient::set_param(DIR_Param *param, ObjID &guid, off64_t size)
{
	param->reset();
	memcpy(&(param->guid), &guid, sizeof(guid_t));
	param->size=size;
	param->options=DIR_OPTION_FILE;
}

/**
 * Initialize the DIR_Param structure for use of create().
 */
void SorClient::init_param(DIR_Param *param)
{
	param->reset();
	param->options=DIR_OPTION_FILE;
}

/**
 * Create a new file.
 * @param path to of the file to be created.
 * @return file handle (NULL if failed.)
 */
SorFile *SorClient::create(const char *path)
{
	SorFile *sf=NULL;
	DIR_Entry entry;
	DIR_Param param;
	init_param(&param);

	int ret=cwd.create(ap, path, &param, &entry);
	if (ret != DIR_RES_OKAY) {
		goto _exit;
	}

	try {
		/** create a new file, create the meta object */
		entry.refresh_info(ap);
		sf=new SorFile(cli, hmap, ap, &entry, SOR_WR, true);
		return sf;
	}
	catch (int error) {
		ret=error;
		goto _exit;
	}

_exit:
	errnum=ret;
	callid=CID_CREATE;

	return NULL;
}

/**
 * Open an existing file.
 * @param path to the file to be opened.
 * @param flags (eg, read/write modes, ignored).
 * @return status (true: success).
 */
SorFile *SorClient::open(const char *path, int flags)
{
	DIR_Entry entry;
	SorFile *sf=NULL;

	int ret=cwd.lookup(ap, path, &entry);
	if (ret != DIR_RES_OKAY) {
		goto _exit;
	}
	else {
		if (entry.is_dir(ap)) {
			ret=DIR_RES_EISDIR;
			goto _exit;
		}
	}

	try {
		entry.refresh_info(ap);
		sf=new SorFile(cli, hmap, ap, &entry, flags); 

		return sf;
	} 
	catch (int error) {
		ret=error;
		goto _exit;
	}

_exit:
	errnum=ret;
	callid=CID_OPEN;

	return NULL;
}

/**
 * Close an opened file. 
 * @param sf Sorrento file handle.
 * @return status (true: success.)
 */
bool SorClient::close(SorFile *sf)
{
	/** 
	 * close the meta object, data segments, and update dir_srv entry. 
	 */
	int ret=sf->close(ap, cli, hmap);
	if (ret != FILE_RES_OKAY) {
		errnum=ret;
		callid=CID_CLOSE;
		return false;
	}

	delete sf;

	return true;
}

/** 
 * Read a number of bytes from a certain offset. 
 * @param sf Sorrento file handle.
 * @param offset the offset of the starting range.
 * @param buf the data buffer to be filled.
 * @param len the length of the buffer.
 * @param read_bytes the number of bytes read.
 * @return status (0: success; -1: error). 
 */
bool SorClient::read(SorFile *sf, off64_t offset, char *buf, off_seg_t len, off_seg_t & read_bytes)
{
	int ret=sf->read(cli, hmap, offset, buf, len, read_bytes);
	if (ret != FILE_RES_OKAY) {
		errnum=ret;
		callid=CID_READ;
		return false;
	}

	return true;
}

/** 
 * Write a number of bytes to a certain offset.
 * @param sf Sorrento file handle.
 * @param offset the offset of the starting range.
 * @param buf the data buffer to be filled.
 * @param len the length of the buffer.
 * @param written_bytes the number of bytes being written.
 * @return status (0: success; -1: error).
 */
bool SorClient::write(SorFile *sf, off64_t offset, char *buf, off_seg_t len, off_seg_t & written_bytes)
{
	int ret=sf->write(cli, hmap, offset, buf, len, written_bytes);
	if (ret != FILE_RES_OKAY) {
		errnum=ret;
		callid=CID_WRITE;
		return false;
	}

	return true;
}

/** 
 * Change the size of an opened file.
 * @param sf Sorrento file handle.
 * @param new_size the new size of the file.
 * @return status (true: success).
 */
bool SorClient::ftruncate(SorFile *sf, off64_t new_size)
{
	int ret=sf->truncate(cli, hmap, new_size);
	if (ret != FILE_RES_OKAY) {
		errnum=ret;
		callid=CID_FTRUNCATE;

		return false;
	}

	return true;
}

/** 
 * Change the size of a file.
 * @param path Sorrento file name.
 * @param new_size the new size of the file.
 * @return status (true: success).
 */
bool SorClient::truncate(const char *path, off64_t new_size)
{
	SorFile *sf=this->open(path, O_WRONLY);
	if (!sf)
		return false;

	bool ret=this->ftruncate(sf, new_size);

	bool ret2=this->close(sf);

	return ret && ret2;
}

/** 
 * Get attribute of a certain file. The attributes are
 * readily available from the SorFile structure.
 * @param sf the SorFile object.
 * @param param the attributes.
 * @return status (true: success).
 */
bool SorClient::fstat(SorFile *sf, DIR_Param *param)
{
	int ret=sf->stat(param);
	if (ret != FILE_RES_OKAY) {
		errnum=ret;
		callid=CID_FSTAT;

		return false;
	}

	return true;
}

/** 
 * Get attribute of a certain file or directory based on name. This is in fact a
 * namespace operation.
 * @param path the path to the file.
 * @param param the attributes.
 * @return status (true: success).
 */
bool SorClient::stat(const char *path, DIR_Param *param)
{
	DIR_Entry entry;
	int ret=cwd.lookup(ap, path, &entry);
	if (ret != DIR_RES_OKAY) {
		goto _exit;
	}

	// if (entry.is_dir(ap)) {
		// ret=DIR_RES_EISDIR;
		// goto _exit;
	// }

	ret=entry.getattr(ap, param);
	if (ret != DIR_RES_OKAY) {
		goto _exit;
	}

	return true;

_exit:
	errnum=ret;
	callid=CID_STAT;

	return false;
}

std::string SorClient::str_error()
{
	if (callid == CID_NULL) {
		return std::string("No error.");
	}

	std::string cmd, reason;

	switch (this->callid) {
		case CID_CHDIR: 		cmd="chdir"; 	break;
		case CID_MKDIR: 		cmd="mkdir"; 	break;
		case CID_RMDIR: 		cmd="rmdir"; 	break;
		case CID_DELTREE: 		cmd="deltree"; 	break;
		case CID_READDIR: 		cmd="readdir"; 	break;
		case CID_READDIR_CONT: 	cmd="readdir_cont"; break;
		case CID_UNLINK: 		cmd="unlink"; 	break;
		case CID_CREATE: 		cmd="create"; 	break;
		case CID_OPEN: 			cmd="open"; 	break;
		case CID_CLOSE: 		cmd="close"; 	break;
		case CID_READ: 			cmd="read"; 	break;
		case CID_WRITE: 		cmd="write"; 	break;
		case CID_FTRUNCATE: 	cmd="ftruncate"; break;
		case CID_FSTAT: 		cmd="fstat"; 	break;
		case CID_STAT: 			cmd="stat"; 	break;
		default: 				cmd="<unknown>"; break;
	}

	switch (this->errnum) {
		case DIR_RES_OKAY:
			reason="Okay"; break;
		case DIR_RES_ENONEXIST: 
			if ((callid==CID_MKDIR)||(callid==CID_CREATE))
				reason="Unable to create file or directory now, try again."; 
			else
				reason="File or directory does not exist."; 
			break;
		case DIR_RES_EEXIST: 
			if (callid==CID_RMDIR)
				reason="Directory not empty."; 
			else
				reason="File or directory already exists."; 
			break;
		case DIR_RES_ESELF: 
			reason="Current directory no longer exists."; break;
		case DIR_RES_EINVALID: 
			reason="Invalid argument."; break;
		case DIR_RES_EISDIR: 
			reason="Target is a directory."; break;
		case DIR_RES_ENOTDIR: 
			reason="Target is a not a directory."; break;
		case DIR_RES_EBUFFER: 
		case DIR_RES_EMEM: 
		case FILE_RES_EMEM :
			reason="Out of memory."; break;
		case DIR_RES_ESERVER: 
		case FILE_RES_ESERVER :
			reason="Server error."; break;
		case DIR_RES_ETOUT: 
			reason="Request timed out. Maybe network error."; break;
		case DIR_RES_ENETWORK: 
		case FILE_RES_ENETWORK :
			reason="Network error."; break;
		case DIR_RES_ECLIENT: 
		case FILE_RES_ECLIENT :
			reason="Client error."; break;
		case FILE_RES_ESPACE :
			reason="Server out of space."; break;
		case FILE_RES_EREAD :
			reason="Failed to read."; break;
		case FILE_RES_EWRITE :
			reason="Failed to write."; break;
		case FILE_RES_NOT_FOUND :
			reason="File not found."; break;
		case FILE_RES_EMETA :
			reason="Metadata corrupted."; break;
		case FILE_RES_EFILE :
			reason="File object corrupted."; break;
		case FILE_RES_EDATA :
			reason="Object data info corrupted."; break;
		case FILE_RES_ATTACHMENT :
			reason="Attachment corrupted."; break;
		case FILE_RES_EIO :
			reason="Server I/O error."; break;
		case FILE_RES_EEXIST :
			reason="File or object already exists."; break;
		case FILE_RES_ESCLIENT :
			reason="Subscriber error."; break;
		case FILE_RES_RD_ONLY :
			reason="File is read only."; break;
		default:
			{
				char buf[32];
				sprintf(buf, "%d", errnum);
				reason="Internal error. Code = ";
				reason= reason+buf;
			}

			break;
	}

	return cmd+": "+reason;
}

void SorClient::stop_service(void)
{
	const VHostLoad *hostloads=(hmap)?hmap->get_host_loads():NULL;
	if (hostloads) {
		for (unsigned int i=0; i<hostloads->size(); i++) {
			cli->stop_server((*hostloads)[i].addr);
		}
	}
	if (hmap)
		hmap->release_host_loads(hostloads); // Important: release_host_loads even if it is NULL.

	HostAddr dir_addr;
	if (hmap->get_dir_addr(dir_addr)) {
		Udp_cli udp_cli(NULL, SOCK_PORT_ANY, 1000);
		udp_cli.stop_srv(dir_addr.ip, dir_addr.port, "", 1, 10);
	}
}

void SorClient::dump_view(void)
{
	HostAddr dir_addr;
	if (hmap->get_dir_addr(dir_addr)) {
		printf("Directory server: %s.\n", dir_addr.to_string().c_str());
	}

	const VHostLoad *hostloads=(hmap)?hmap->get_host_loads():NULL;
	if (hostloads) {
		for (unsigned int i=0; i<hostloads->size(); i++) {
			printf(" %-16s %-4d %04.1f   %04.1f   %04.1f   %04.1f   %04.1f   %04.1f  <", 
					(*hostloads)[i].addr.ip,
					(*hostloads)[i].vnodes,
					(*hostloads)[i].cpu_comp*100,
					(*hostloads)[i].cpu_ewma_comp*100,
					(*hostloads)[i].cpu_io*100,
					(*hostloads)[i].cpu_ewma_io*100,
					(*hostloads)[i].mem_used*100,
					(*hostloads)[i].swap_used*100
			);

			for (int j=0; j<(*hostloads)[i].ndisks; j++) {
				if (j>0)
					printf(",");
				printf("%d", (*hostloads)[i].disks[j].bfree/1024); 
			}
			printf(">\n");
		}
	}
	if (hmap)
		hmap->release_host_loads(hostloads); // Important: release_host_loads even if it is NULL.
}
