// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: sor_file.cc,v $ - the Sorrento File class.
 *            It represents a logical file.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2004/02/19 13:14:51 $
 * Version :  $Revision: 1.23 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

# include <stddef.h>
# include <assert.h>
# include <string>

# include "tmpi_debug.h"
# include "sor_file.h"
# include "client.h"
# include "host_map.h"
# include "udp_common.h"
# include "protocol.h"
# include "../dir_srv/dir_cli_ap.h"

/** Maximum attachment size is 60KB *MINUS* the metadata header */
# define SOR_META_HEADER (offsetof(SorMetaData, payload))
# define BLOCK_SIZE (4*1024)
# define MAX_META_SIZE (60*1024)
# define MAX_ATTACHMENT_SIZE (UDP_SAFE_SZ-SOR_META_HEADER)
# define MIN_ATTACHMENT_SIZE BLOCK_SIZE
# define ROUNDUP(x, y) (((x)+(y)-1)/(y)*(y))
# define CEILING(x, y) (((x)+(y)-1)/(y))
# define MAX_BUFFER_SIZE ROUNDUP(MAX_ATTACHMENT_SIZE, BLOCK_SIZE)
# define META_MAGIC 0x75fa0d90b32cfaedLL
# define SIZE_KB 1024LL
# define SIZE_MB (SIZE_KB*SIZE_KB)
# define SIZE_GB (SIZE_MB*SIZE_KB)

/**
 * The metadata block structure. This is the on-disk image of the 
 * metadata structure. 
 *
 * In this version, we use a simplified algorithm to decide 
 * whether the data is within the object or not (aka, whether we
 * should use payload.data or use payload.data_segs).
 * If size is smaller than 60KB (minus header bytes), we use payload.data; otherwise, 
 * we use payload.data_segs. The sizes of the segments are ordered
 * as follows:
 * 0-7: 1MB, (up to 8MB)
 * 8-14: 8MB, (up to 64MB)
 * 15-21: 64MB, (up to 512MB)
 * 22-28: 512MB, (up to 4GB)
 * 29-...: 2GB
 * 60KB/16=3840 segments, of size 3811*2+4=7626GB.
 */
struct SorMetaData {
	off64_t magic; /**< Magic number to prevent data corruption */
	off64_t size; /**< The size of the logical object */
	union {
		ObjID data_segs[1];
		char data[1];
	} payload;
};

class Attachment {
 private:
	char *buf; /**< buffer pointer */
	off_seg_t b_size; /**< The actual buffer size */
	off_seg_t size; /**< size of the buffer. */
	bool dirty; /**< whether the buffer has been rewritten */
	int flags;
	void expand(off_seg_t new_size);
	bool adjust_size(off_seg_t sug_size, off_seg_t &new_size);

 public:
	// Attachment(off_seg_t init_size=0); /**< The default constructor */
	Attachment(char *a_data, off_seg_t a_size, int flags); /**< The constructor */
	~Attachment();
	int read(off_seg_t offset, char *data, off_seg_t len, off_seg_t & bytes_read);  /**< read */
	/** Can we fit new_size bytes at offset? */
	static bool can_fit(off_seg_t offset, off_seg_t new_size) 
	{ 
		return is_legal_size(offset+new_size); 
	} 
	int write(off_seg_t offset, char *data, off_seg_t len, off_seg_t & bytes_written );  /**< write */
	int truncate(off_seg_t new_size, off_seg_t & actual_size); /**< truncate */
	/** is the new_size a legal size? */
	static bool is_legal_size(off_seg_t new_size);
	off_seg_t get_size() const { return size; }
	const char * get_data() const { return buf; }
	bool is_dirty() const { return dirty; }
	int pack(SorMetaData *meta, int m_size);
};

class Segment {
 private:
	ObjID id; /**< The GUID of the segment */
	off_seg_t max_size; /**< the maximum segment size */
	bool created; /**< Whether the segment is created or not */
	Descriptor des; /**< Cached descriptor for the segment. */
	int flags;
	void refresh_des(Client *cli, HostMap *hmap);

 public:
	/** Constructor, creating a new segment */
	Segment(off_seg_t maxsz); 
	/** Constructor, an existing segment */
	Segment(ObjID &guid, off_seg_t maxsz, int flags, bool exist=true);
	~Segment(); /**< destructor */
	int read(Client *cli, HostMap *hmap, off_seg_t offset, char *buf, off_seg_t len, off_seg_t &bytes_read);
	int write(Client *cli, HostMap *hmap, off_seg_t offset, char *buf, off_seg_t len, off_seg_t &bytes_written);
	int truncate(Client *cli, HostMap *hmap, off_seg_t new_size, off_seg_t &actual_size);
	int close(Client *cli);
	int remove(Client *cli, HostMap *hmap);

	friend class DataSegments;
};

class DataSegments : std::vector<Segment *> {
	off64_t obj_size; /**< The size of the logical object */
	bool dirty; /**< whether the vector has been modified */
	int flags;
 public:
	DataSegments(); /**< construct with an empty vector. */
	DataSegments(SorMetaData *meta, int m_size, int flags); /**< construct from a meta object */
	~DataSegments(); /**< destructor */
	int truncate(Client *cli, HostMap *hmap, off64_t size);
	int read(Client *cli, HostMap *hmap, off64_t offset, char *buf, off_seg_t len, off_seg_t &read_bytes);
	int write(Client *cli, HostMap *hmap, off64_t offset, char *buf, off_seg_t len, off_seg_t &written_bytes);
	int close(Client *cli);
	int remove_all(Client *cli, HostMap *hmap);

	/** @defgroup attributes Getting attributes. */
	/** @{ */
	bool is_dirty() const { return dirty; }
	off64_t get_size() const { return obj_size; }
	/** pack everything into a meta-object */
	int pack(SorMetaData *meta, int m_size);
	/** @} */
};

bool Attachment::is_legal_size(off_seg_t x_size)
{
	if (x_size>MAX_ATTACHMENT_SIZE)
		return false;

	return true;
}

/**
 * Note: The data buffer for the attachment follows the following 
 * ideas during expansion. 
 * (1) The buffer size always expands and never shrinks.
 * (2) The buffer size is always a multiple of 4KB.
 * (3) The buffer size increases exponentially. If the current
 * buffer size is x 4KB blocks, then we first find y<x and y 
 * is a power of 2, we expand the buffer to min((x+y), 15) 4KB
 * blocks.
 */

/**
 * Perform an exponential increase upon the base_size.
 * @param base_size the base size, must be a multiple of BLOCK_SIZE.
 * @return the new size after the exponential expansion.
 */
static off_seg_t exp_increase(off_seg_t base_size)
{
	assert((base_size%BLOCK_SIZE) == 0);

	if (base_size < BLOCK_SIZE)
		return BLOCK_SIZE;

	unsigned int n_blks=base_size/BLOCK_SIZE;
	unsigned int n_blks_in_bits=0;
	while (n_blks) {
		n_blks=n_blks/2;
		n_blks_in_bits++;
	}

	unsigned int n_inc_blks=1<<(n_blks_in_bits-1);

	off_seg_t new_size=base_size+n_inc_blks*BLOCK_SIZE;
	
	if (new_size > MAX_BUFFER_SIZE)
		return MAX_BUFFER_SIZE;

	return new_size;
}

/**
 * @param sug_size the suggested size. The new_size should be larger
 * than sug_size.
 * @param new_size the adjusted size.
 * @return status (true: need adjustment; false: no need.)
 */
bool Attachment::adjust_size(off_seg_t sug_size, off_seg_t &new_buf_size)
{
	/** sug_size must be legal */
	assert(is_legal_size(sug_size));

	if (sug_size <= b_size) /**< nothing to do */
		return false;

	off_seg_t exp_size=exp_increase(b_size);
	if (exp_size>=sug_size) {
		new_buf_size=exp_size;
		return true;
	}

	// The following should never be reached.
	if (sug_size < MIN_ATTACHMENT_SIZE) {
		new_buf_size= MIN_ATTACHMENT_SIZE;
		return true;
	}

	new_buf_size=ROUNDUP(sug_size, BLOCK_SIZE);

	return true;
}

/** 
 * The constructor. 
 * @param a_data the data of the attachment.
 * @param a_size the size of the data portion.
 */
Attachment::Attachment(char *a_data, off_seg_t a_size, int f) :
	buf(NULL),
	b_size(0),
	size(0),
	dirty(false),
	flags(f)
{
	assert(is_legal_size(a_size));

	adjust_size(a_size, b_size); /**< adjust the size */

	if (b_size == 0) {
		b_size = (int)BLOCK_SIZE;
	}

	buf=(char *)calloc(b_size, 1);
	if (a_size > 0) {
		memcpy(buf, a_data, a_size);
		size=a_size;
	}
}

Attachment::~Attachment()
{
	if (buf)
		free(buf);
}

/** 
 * Read something from the attachment. 
 * @param offset the starting offset.
 * @param data the buffer to deposit the data.
 * @param len the size to be read.
 * @param bytes_read bytes read.
 * @return status (0: success; otherwise: error reason).
 */
int Attachment::read(off_seg_t offset, char *data, off_seg_t len, off_seg_t & bytes_read)
{
	if (offset >= size) {
		bytes_read = 0;
		return FILE_RES_OKAY;
	}

	if (offset+len>size) {
		len=size-offset;
	}

	memcpy(data, buf+offset, len);

	bytes_read = len;
	return FILE_RES_OKAY;
}

void Attachment::expand(off_seg_t size)
{
	off_seg_t new_size;
	if (adjust_size(size, new_size)) {
		buf=(char *)realloc(buf, new_size);
		if (!buf)
			throw FILE_RES_EMEM;
		b_size=new_size;
	}
}

/** 
 * Write something into the attachment. 
 * @param offset the starting offset.
 * @param data the data buffer.
 * @param len the size of data buffer to be written.
 * @param bytes_written bytes_written.
 * @return status (0: success; otherwise: error reason).
 */
int Attachment::write(off_seg_t offset, char *data, off_seg_t len, off_seg_t & bytes_written)
{
	if (flags == SOR_RD_ONLY)
		return FILE_RES_RD_ONLY;

	if (offset >= MAX_ATTACHMENT_SIZE) {
		bytes_written = 0;
		return FILE_RES_OKAY;
	}

	if (offset+len > MAX_ATTACHMENT_SIZE) {
		len=MAX_ATTACHMENT_SIZE - offset;
	}

	expand(offset+len); /**< make sure the buffer fits */

	memcpy(buf+offset, data, len);

	if (offset+len>size)
		size=offset+len;

	dirty=true;

	bytes_written = len;
	return FILE_RES_OKAY;
}

/** 
 * Truncate the buffer to a new size. 
 * @param new_size the new size of the attachment.
 * @param actual_size the actual size of the attachment now.
 * @return status (0: success; otherwise: error reason).
 */
int Attachment::truncate(off_seg_t new_size, off_seg_t & actual_size)
{
	if (new_size>MAX_ATTACHMENT_SIZE) {
		new_size=MAX_ATTACHMENT_SIZE;
	}

	expand(new_size); /**< make sure the buffer fits */
	actual_size=size=new_size; /**< adjust the useful data size */
	dirty=true; /**< mark dirty */

	return FILE_RES_OKAY;
}

/**
 * Pack the data segments into a SorMetaData object.
 * @param p_meta the pointer to the meta object.
 * @param m_size the size of the buffer.
 * @return the actual size of the meta object.  Or 
 * -1 on errors. Error reason: (1) buffer too small.
 * (2) the size of DataSegments too small.
 */
int Attachment::pack(SorMetaData *p_meta, int m_size)
{
	if (!p_meta)
		return -1;

	if (size > MAX_ATTACHMENT_SIZE)
		return -1;

	int real_size = (int)(SOR_META_HEADER + size);
	if (m_size < real_size)
		return -1;

	p_meta->magic = META_MAGIC;
	p_meta->size = size;
	memcpy(&(p_meta->payload.data), buf, size);

	return real_size;
}

/** Constructor, create a new sgement */
Segment::Segment(off_seg_t maxsz) : max_size(maxsz), created(false), flags(SOR_WR)
{
	Guid guid;
	guid.gen();
	guid.pack(id);
}

/** 
 * Constructor, open an existing sgement 
 * @param guid the guid of the segment.
 */
Segment::Segment(ObjID &guid, off_seg_t maxsz, int f, bool exist) : id(guid), max_size(maxsz), created(exist), flags(f)
{ }

static int stg_to_file_errno(int stg_err)
{
	switch (stg_err) {
		case STG_RES_OKAY : return FILE_RES_OKAY;
		case STG_RES_ESERVER : return FILE_RES_ESERVER;
		case STG_RES_ECLIENT : return FILE_RES_ECLIENT;
		case STG_RES_ENETWORK : return FILE_RES_ENETWORK;
		case STG_RES_ETOUT : return FILE_RES_ENETWORK;
		case STG_RES_EIO : return FILE_RES_EIO;
		case STG_RES_ESPACE : return FILE_RES_ESPACE;
		case STG_RES_EMEM : return FILE_RES_EMEM;
		case STG_RES_ENOTFOUND : return FILE_RES_NOT_FOUND;
		case STG_RES_EEXIST  : return FILE_RES_EEXIST;
		default : return FILE_RES_ECLIENT;
	}
}

/**
 * Refresh the descriptor, if not opened, open it.
 * @param cli the Storage access point.
 * @param hmap the host map.
 * @return no return. 
 * @throw throws out the error code (int).
 */
void Segment::refresh_des(Client *cli, HostMap *hmap)
{
	if (des.valid())
		return;

	if (!created) {
		VecServerAddr hosts;

		if (!hmap->place(id, max_size, hosts))
			throw FILE_RES_ESCLIENT;

		int final_err = STG_RES_OKAY;
		for (unsigned int i=0; i<hosts.size(); i++) {
			int ret=cli->create(hosts[i], id, des);
			if (ret == STG_RES_OKAY) {
				created = true;
				return;
			}
			if (ret == STG_RES_EEXIST) {
				created = true;
				ret = cli->open(hosts[i], id, des, flags);
				if (ret == STG_RES_OKAY)
					return;
				goto _try_open;
			}
			if (ret < final_err)
				final_err = ret;
		}

		throw stg_to_file_errno(final_err);
	}

_try_open:
	{
		VecServerAddr hosts;

		if (!hmap->lookup(id, hosts))
			throw FILE_RES_ESCLIENT;
		int final_err = STG_RES_OKAY;
		for (unsigned int i=0; i<hosts.size(); i++) {
			int ret = cli->open(hosts[i], id, des, flags);
			if (ret == STG_RES_OKAY) {
				return;
			}
			if (ret < final_err)
				final_err = ret;
		}

		throw stg_to_file_errno(final_err);
	}
}

/** 
 * Read data from within a segment.
 * @param cli the storage access point.
 * @param hmap the host map for data location.
 * @param offset the starting offset within the segment.
 * @param buf the buffer where we deposit data.
 * @param len the number of bytes to be read, must be a multiple
 * of BLOCK_SIZE.
 * @param bytes_read the actual number of bytes being read.
 * @return status (0: success; otherwise: error reason).
 */
int Segment::read(Client *cli, HostMap *hmap, off_seg_t offset, char *buf, off_seg_t len, off_seg_t &bytes_read)
{
	try {
		refresh_des(cli, hmap);

		int ret=cli->read(des, offset, buf, len, bytes_read);
		if (ret == STG_RES_OKAY) {
			return FILE_RES_OKAY;
		}

		return stg_to_file_errno(ret);
	}
	catch (int errnum) {
		return errnum;
	}
}

/** 
 * Write data within a segment.
 * @param cli the storage access point.
 * @param hmap the host map for data location.
 * @param offset the starting offset within the segment.
 * @param buf the buffer where stores the output data.
 * @param len the number of bytes to be written, must be a multiple
 * of BLOCK_SIZE.
 * @param bytes_written the actual number of bytes being written.
 * @return status (0: success; otherwise: error reason).
 */
int Segment::write(Client *cli, HostMap *hmap, off_seg_t offset, char *buf, off_seg_t len, off_seg_t &bytes_written)
{
	if (flags == SOR_RD_ONLY)
		return FILE_RES_RD_ONLY;

	try {
		refresh_des(cli, hmap);
		int ret=cli->write(des, offset, buf, len, bytes_written);
		if (ret == STG_RES_OKAY)
			return FILE_RES_OKAY;

		return stg_to_file_errno(ret);
	}
	catch (int errnum) {
		return errnum;
	}
}

/** 
 * Truncate a segment.
 * @param cli the storage access point.
 * @param hmap the host map for data location.
 * @param new_size the new size of the segment.
 * @param actual_size the acutal size after the truncation.
 * @return status (0: success; otherwise: error reason).
 */
int Segment::truncate(Client *cli, HostMap *hmap, off_seg_t new_size, off_seg_t &actual_size)
{
	try {
		refresh_des(cli, hmap);
		int ret=cli->truncate(des, new_size, actual_size);
		if (ret == STG_RES_OKAY)
			return FILE_RES_OKAY;
		return stg_to_file_errno(ret);
	}
	catch (int errnum) {
		return errnum;
	}
}

/** 
 * Close a segment.
 * @param cli the storage access point.
 * @return status (0: success; otherwise: error reason).
 */
int Segment::close(Client *cli)
{
	if (des.valid()) {
		int ret=cli->close(des);
		des.reset();
		if (ret == STG_RES_OKAY) {
			return FILE_RES_OKAY;
		}
		else {
			tmpi_error(DBG_VERBOSE, "Failed to close segment (%d).\n", ret);
			return stg_to_file_errno(ret);
		}
	}

	return FILE_RES_OKAY;
}

/**
 * remove a segment.
 * @param cli the storage access point.
 * @param hmap the host map.
 * @return status.
 */
int Segment::remove(Client *cli, HostMap *hmap)
{
	{
		int ret=close(cli);
		if (ret != 0) {
			des.reset(); // have to remove it anyway.
		}
	}

	VecServerAddr hosts;

	if (!hmap->lookup(id, hosts))
		return FILE_RES_ESCLIENT;

	int final_ret=FILE_RES_OKAY;
	for (unsigned int i=0; i<hosts.size(); i++) {
		int ret=cli->remove(hosts[i], id);
		if ((ret!=STG_RES_OKAY)&&(ret!=STG_RES_ENOTFOUND)) {
			final_ret=stg_to_file_errno(ret);
		}
		if (ret == STG_RES_OKAY) {
			break;
		}
	}

	return final_ret;
}

/** the destructor */
Segment::~Segment()
{
	if (des.valid()) {
		tmpi_error(DBG_INTERNAL, "Forgot to close a segment before destroy.\n");
		Client *cli = new Client;
		cli->close(des);
		delete cli;
	}
}

/** construct with an empty vector. */
DataSegments::DataSegments() : obj_size(0), dirty(false)
{ }

/** destructor. */
DataSegments::~DataSegments() 
{ 
	for (unsigned int i=0; i<this->size(); i++) {
		Segment *ps=(*this)[i];
		if (ps)
			delete ps;
	}
}

/**
 * Calculate the number of segments from the size.
 * @param size the size of the object.
 * @return the number of segments.
 */
static unsigned int size_to_segments(off64_t size)
{
	if (size > 4*SIZE_GB) {
		return (unsigned int)CEILING(size-4*SIZE_GB, 2*SIZE_GB)+29;
	}

	if (size > (512*SIZE_MB)) {
		return (unsigned int)CEILING(size-512*SIZE_MB, 512*SIZE_MB)+22;
	}

	if (size > (64*SIZE_MB)) {
		return (unsigned int)CEILING(size-64*SIZE_MB, 64*SIZE_MB)+15;
	}

	if (size > (8*SIZE_MB)) {
		return (unsigned int)CEILING(size-8*SIZE_MB, 8*SIZE_MB)+8;
	}

	return (unsigned int)CEILING(size, SIZE_MB);
}

/**
 * Calculate the maximum size for a given number of segments.
 * @param n_segs the number of segments. 
 * @return the maximum number of bytes that can be stored in these segments.
 */
static unsigned int segments_to_max_size(int n_segs)
{
	if (n_segs>=29) {
		return 4*SIZE_GB + (n_segs-29)*2*SIZE_GB;
	}

	if (n_segs>=22) {
		return 512*SIZE_MB + (n_segs-22)*512*SIZE_MB;
	}

	if (n_segs>=15) {
		return 64*SIZE_MB + (n_segs-15)*64*SIZE_MB;
	}

	if (n_segs>=8) {
		return 8*SIZE_MB + (n_segs-8)*8*SIZE_MB;
	}

	return n_segs*SIZE_MB;
}

/**
 * Calculate the size for a given segment (counted from 0).
 * @param seg_id the id of the segment. 
 * @return the number of bytes that can be stored in this segments.
 */
static off_seg_t segment_to_size(int seg_id)
{
	if (seg_id < 8)
		return (off_seg_t)SIZE_MB; 

	if (seg_id < 15)
		return (off_seg_t)8*SIZE_MB; 

	if (seg_id < 22)
		return (off_seg_t)64*SIZE_MB; 

	if (seg_id < 29)
		return (off_seg_t)512*SIZE_MB; 

	return (off_seg_t)2*SIZE_GB; 
}

/**
 * Verify the integrity of a meta object.
 * @param meta the meta object.
 * @param m_size the size of the meta object.
 * @return verification result (true: yes).
 */
static bool verify_meta_object(SorMetaData *meta, int m_size)
{
	if (meta->magic != META_MAGIC) {
		return false;
	}

	if (meta->size < 0)
		return false;

	if (meta->size <= MAX_ATTACHMENT_SIZE) { // data in payload.
		if (m_size < SOR_META_HEADER+meta->size)
			return false;
		return true;
	}

	unsigned int n_seg=size_to_segments(meta->size);

	if (m_size < (int)(SOR_META_HEADER+n_seg*sizeof(ObjID)))
		return false;

	return true;
}

/** 
 * construct from a meta object.
 * @param meta the meta object.
 * @para m_size the size of the meta-object.
 * Note: meta cannot be NULL, the content of meta
 * must be verified already.
 */
DataSegments::DataSegments(SorMetaData *meta, int m_size, int f) :
	obj_size(meta->size),
	dirty(false),
	flags(f)
{
	unsigned int n_guids=size_to_segments(meta->size);
	for (unsigned int i=0; i<n_guids; i++) {
		off_seg_t sz=segment_to_size(i);
		Segment *seg=new Segment(meta->payload.data_segs[i], sz, flags, true);
		this->push_back(seg);
	}
}

/**
 * Truncate the data segments.
 * @param cli the Storage access point.
 * @param hmap the host map.
 * @param size the new size, must be larger than MAX_ATTACHMENT_SIZE.
 * @return status (0: success; otherwise, error reasons).
 */
int DataSegments::truncate(Client *cli, HostMap *hmap, off64_t size)
{
	assert(size > MAX_ATTACHMENT_SIZE);

	if (size != obj_size)
		dirty = true;
	else
		return FILE_RES_OKAY;

	unsigned int n_segs=size_to_segments(size);
	unsigned const int old_n_segs=size_to_segments(obj_size);
	if (n_segs < old_n_segs) {
		// Need to get rid of some segments.
		for (unsigned int i=n_segs; i<old_n_segs; i++) {
			(*this)[i]->remove(cli, hmap);
			delete (*this)[i];
			(*this)[i]=NULL;
		}
		this->resize(n_segs);
	}
	else if (n_segs > old_n_segs) {
		// first expand the current last segment.
		if (old_n_segs > 0) {
			off_seg_t actual_size;
			off_seg_t seg_size = segment_to_size(old_n_segs-1);
			int ret=(*this)[old_n_segs-1]->truncate(cli, hmap, seg_size, actual_size);
			if (ret == FILE_RES_OKAY) {
				obj_size = segments_to_max_size(old_n_segs-1)+actual_size;
				if (seg_size != actual_size) {
					return FILE_RES_ESPACE;
				}
			}
			else {
				return ret;
			}
		}

		// expand some more segments.
		this->reserve(n_segs);

		// expand full size segments.
		for (unsigned int i=old_n_segs; i<n_segs-1; i++) {
			off_seg_t sz=segment_to_size(old_n_segs);
			Segment *seg=new Segment(sz);
			off_seg_t actual_size;
			off_seg_t seg_size = segment_to_size(i);
			int ret=FILE_RES_OKAY;
			ret=seg->truncate(cli, hmap, seg_size, actual_size);
			if (ret==FILE_RES_OKAY) {
				obj_size += actual_size;
				this->push_back(seg);
				if (seg_size != actual_size) {
					return FILE_RES_ESPACE;
				}
			}
			else {
				return ret;
			}
		}

		// expand the last segment.
		off_seg_t sz=segment_to_size(n_segs-1);
		Segment *seg=new Segment(sz);
		this->push_back(seg);
	}
	else {
		// only need to expand the last segment
	}

	// expanding the last segment.
	if (n_segs > 0) {
		off_seg_t in_seg_offset=(off_seg_t)(size-segments_to_max_size(n_segs-1));
		off_seg_t actual_size;
		int ret=(*this)[n_segs-1]->truncate(cli, hmap, in_seg_offset, actual_size);
		if (ret == FILE_RES_OKAY) {
			obj_size=segments_to_max_size(n_segs-1)+actual_size;
			if (in_seg_offset != actual_size) {
				return FILE_RES_ESPACE;
			}
		}
		else {
			obj_size=segments_to_max_size(n_segs-1);
			return ret;
		}
	}

	return FILE_RES_OKAY;
}

/**
 * Read a certain number of bytes.
 * @param cli the Storage access point.
 * @param hmap the host map.
 * @param offset the offset of the starting byte.
 * @param len the number of bytes to be read.
 * @param read_bytes the actual number of bytes being read.
 * @return status (0: success; otherwise, error reasons).
 */
int DataSegments::read(Client *cli, HostMap *hmap, off64_t offset, char *buf, off_seg_t len, off_seg_t &read_bytes)
{
	if ( (len ==0) || (offset >= obj_size)) {
		read_bytes = 0;
		return FILE_RES_OKAY;
	}

	if (offset+len > obj_size)
		len=obj_size-offset;

	unsigned int n_segs=size_to_segments(offset+1);
	unsigned int seg_id = n_segs-1;
	off_seg_t off_in_seg = (off_seg_t)(offset-segments_to_max_size(seg_id));
	off_seg_t size_in_seg=segment_to_size(seg_id)-off_in_seg;
	read_bytes = 0;
	while (read_bytes < len) {
		off_seg_t x_bytes = (read_bytes + size_in_seg > len)? len-read_bytes: size_in_seg;
		off_seg_t y_bytes;
		int ret=(*this)[seg_id]->read(cli, hmap, off_in_seg, buf, x_bytes, y_bytes);
		if (ret == FILE_RES_OKAY) {
			read_bytes += y_bytes;
			if (x_bytes != y_bytes) {
				return FILE_RES_EREAD;
			}
			seg_id++;
			buf+=y_bytes;
			off_in_seg = 0;
			size_in_seg=segment_to_size(seg_id);
		}
		else {
			return ret;
		}
	}

	return FILE_RES_OKAY;
}

/**
 * Write a certain number of bytes.
 * @param cli the Storage access point.
 * @param hmap the host map.
 * @param offset the offset of the starting byte.
 * @param len the number of bytes to be written.
 * @param written_bytes the actual number of bytes being written.
 * @return status (0: success; otherwise, error reasons).
 */
int DataSegments::write(Client *cli, HostMap *hmap, off64_t offset, char *buf, off_seg_t len, off_seg_t &written_bytes)
{
	if (flags == SOR_RD_ONLY)
		return FILE_RES_RD_ONLY;

	if (len == 0) {
		written_bytes=0;
		return FILE_RES_OKAY;
	}

	if (offset+len >= obj_size) {
		int ret=truncate(cli, hmap, offset+len);
		if (ret != FILE_RES_OKAY)
			return ret;
	}

	unsigned int n_segs=size_to_segments(offset+1);
	unsigned int seg_id = n_segs-1;
	off_seg_t off_in_seg=(off_seg_t)(offset-segments_to_max_size(seg_id));
	off_seg_t size_in_seg=segment_to_size(seg_id)-off_in_seg;
	written_bytes = 0;
	while (written_bytes < len) {
		off_seg_t x_bytes = (written_bytes + size_in_seg > len)? len-written_bytes: size_in_seg;
		off_seg_t y_bytes;
		int ret=(*this)[seg_id]->write(cli, hmap, off_in_seg, buf, x_bytes, y_bytes);
		if (ret == FILE_RES_OKAY) {
			written_bytes += y_bytes;
			if (x_bytes != y_bytes) {
				return FILE_RES_EWRITE;
			}
			seg_id++;
			buf+=y_bytes;
			off_in_seg = 0;
			size_in_seg=segment_to_size(seg_id);
		}
		else {
			return ret;
		}
	}

	return FILE_RES_OKAY;
}

/**
 * Close the data segments, not going to write any more.
 * Note: although in the current implementation, read/write
 * after close is still possible, it is not guaranteed to
 * be so in the future.
 * @param cli the storage access point.
 * @return status code (0: success; otherwise, error reason).
 */
int DataSegments::close(Client *cli)
{
	unsigned int n_seg = size_to_segments(obj_size);
	int final_ret=FILE_RES_OKAY;
	for (unsigned int i=0; i<n_seg; i++) {
		int ret=(*this)[i]->close(cli);
		if (ret != FILE_RES_OKAY) {
			final_ret = ret;
		}
	}

	return final_ret;
}

/**
 * Pack the data segments into a SorMetaData object.
 * @param p_meta the pointer to the meta object.
 * @param m_size the size of the buffer.
 * @return the actual size of the meta object.  Or 
 * -1 on errors. Error reason: (1) buffer too small.
 * (2) the size of DataSegments too small.
 */
int DataSegments::pack(SorMetaData *p_meta, int m_size)
{
	if (!p_meta)
		return -1;

	unsigned int n_segs=size_to_segments(obj_size);
	int real_size = (int)(SOR_META_HEADER + sizeof(ObjID)*n_segs);
	if (m_size < real_size)
		return -1;

	if (obj_size <= MAX_ATTACHMENT_SIZE)
		return -1;

	p_meta->magic = META_MAGIC;
	p_meta->size = obj_size;
	for (unsigned int i=0; i<n_segs; i++) {
		memcpy(&(p_meta->payload.data_segs[i]), &((*this)[i]->id), sizeof(ObjID));
	}

	return real_size;
}

/**
 * Remove all existing segments.
 * @param cli storage access point.
 * @param hmap the host map.
 * @return status (0: success; otherwise, error reasons).
 */
int DataSegments::remove_all(Client *cli, HostMap *hmap)
{
	int last_error=FILE_RES_OKAY;
	unsigned int n_segs = size_to_segments(obj_size);
	for (unsigned int i=0; i<n_segs; i++) {
		int ret=(*this)[i]->remove(cli, hmap);
		if (ret != FILE_RES_OKAY) 
			last_error = ret;
	}

	this->resize(0);
	this->obj_size = 0;
	this->dirty = true;

	return last_error;
}

/**
 * SorFile constructor.
 * @param cli storage access point.
 * @param hmap host map.
 * @param entry the directory server entry.
 * @param new_file whether it is a new file?
 * Note that we explicitly require the caller to specify
 * whether it is a new file or not. If a user wants to do
 * a creation on demand, he can first try to load the
 * object without setting the new_file and in the catch 
 * part of the program try again to create the object.
 */
SorFile::SorFile(Client *cli, HostMap *hmap, DIR_Cli_AP *ap, DIR_Entry *entry, int f, bool new_file) :
	root_segment(NULL),
	attachment(NULL),
	data_segs(NULL),
	flags(f)
{
	// first get the up-to-date root_guid
	int ret=entry->get_guid(ap, root_guid);
	if (ret != DIR_RES_OKAY) {
		throw ret;
	}

	memcpy(&root, entry, sizeof(DIR_Entry));

	root_segment = new Segment(root_guid, MAX_META_SIZE, flags, !new_file);

	if (new_file) {
		SorMetaData meta;
		meta.magic=META_MAGIC;
		meta.size=0;
		off_seg_t bytes_written;
		ret=root_segment->write(cli, hmap, 0U, (char *)&meta, sizeof(meta), bytes_written);
		if (ret != FILE_RES_OKAY)
			throw ret;
		if ( bytes_written != sizeof(meta) ) {
			tmpi_error(DBG_VERBOSE, "bytes_written = %d, sizeof(meta) = %d.\n", bytes_written, sizeof(meta));
			throw FILE_RES_ESPACE;
		}
		attachment = new Attachment(meta.payload.data, meta.size, flags);
	}
	else {
		char buf[MAX_META_SIZE]; // 60KB.
		off_seg_t bytes_read;
		ret=root_segment->read(cli, hmap, 0U, buf, MAX_META_SIZE, bytes_read);
		if (ret != FILE_RES_OKAY)
			throw ret;
		SorMetaData *p_meta=(SorMetaData *)buf;
		if (!verify_meta_object(p_meta, (int)bytes_read))
			throw FILE_RES_EMETA;
		if (Attachment::is_legal_size(p_meta->size)) {
			attachment=new Attachment(p_meta->payload.data, p_meta->size, flags);
		}
		else {
			data_segs = new DataSegments(p_meta, (int)bytes_read, flags);
		}
	}
}

SorFile::~SorFile()
{
	if (data_segs)
		delete data_segs;

	if (attachment)
		delete attachment;

	if (root_segment)
		delete root_segment;
}

/**
 * Truncate the file.
 * @param cli the Storage access point.
 * @param hmap the host map.
 * @param size the new size.
 * @return status (0: success; otherwise, error reasons).
 */
int SorFile::truncate(Client *cli, HostMap *hmap, off64_t size)
{
	if (attachment && data_segs) {
		return FILE_RES_EFILE; // very bad error!
	}

	// may change from attachment to data_segs, or vice versa.
	if (data_segs && (size <= MAX_ATTACHMENT_SIZE)) {
		// change to attachment mode.
		char buf[(off_seg_t)size];
		off_seg_t read_bytes;
		int ret=data_segs->read(cli, hmap, 0, buf, (off_seg_t)size, read_bytes);
		if (ret == FILE_RES_OKAY) {
			// if ( (off_seg_t)size != read_bytes)
				// return FILE_RES_EDATA;

			try {
				attachment = new Attachment(buf, (off_seg_t)read_bytes, SOR_WR);
				data_segs->remove_all(cli, hmap); /**< I will live with whatever errors here */
				data_segs->close(cli);
				delete data_segs;
				data_segs = NULL;

				return FILE_RES_OKAY;
			}
			catch (...) {
				delete attachment;
				attachment = NULL;
				return FILE_RES_ATTACHMENT;
			}
		}

		return ret;
	}
	else if ( attachment && (size > MAX_ATTACHMENT_SIZE)) {
		// change to attachment mode.
		char buf[MAX_ATTACHMENT_SIZE];
		off_seg_t read_bytes;
		int ret=attachment->read(0, buf, MAX_ATTACHMENT_SIZE, read_bytes);
		if (ret != FILE_RES_OKAY)
			return ret;
		data_segs = new DataSegments();
		ret=data_segs->truncate(cli, hmap, size);
		off_seg_t write_bytes = 0;
		if (ret == FILE_RES_OKAY) {
			ret=data_segs->write(cli, hmap, 0, buf, read_bytes, write_bytes);
		}

		if ( (ret != FILE_RES_OKAY) || (write_bytes != read_bytes) ) {
			data_segs->remove_all(cli, hmap);
			delete data_segs;
			data_segs = NULL;

			if (write_bytes != read_bytes) {
				ret=FILE_RES_ESPACE;
			}
		}
		else {
			delete attachment;
			attachment = NULL;
		}

		return ret;
	}
	else if (attachment) {
		off_seg_t actual_size;
		try {
			int ret=attachment->truncate(size, actual_size);
			if (ret != FILE_RES_OKAY)
				return ret;
			if (size != actual_size)
				return FILE_RES_ESPACE;
			return ret;
		} catch (int errnum) {
			return errnum;
		}
	}
	else if (data_segs) {
		int ret=data_segs->truncate(cli, hmap, size);
		return ret;
	}
	else {
		return FILE_RES_EFILE; // very bad error!
	}
}

/**
 * Read a certain number of bytes.
 * @param cli the Storage access point.
 * @param hmap the host map.
 * @param offset the offset of the starting byte.
 * @param len the number of bytes to be read.
 * @param read_bytes the actual number of bytes being read.
 * @return status (0: success; otherwise, error reasons).
 */
int SorFile::stat(DIR_Param *param)
{
	if (attachment && data_segs) {
		return FILE_RES_EFILE; // very bad error!
	}

	memcpy(param, root.get_param(), sizeof(DIR_Param));
	if (attachment) {
		param->size=attachment->get_size();
	}
	else if (data_segs) {
		param->size=data_segs->get_size();
	}
	else {
		return FILE_RES_EFILE; // very bad error!
	}

	return FILE_RES_OKAY;
}

/**
 * Read a certain number of bytes.
 * @param cli the Storage access point.
 * @param hmap the host map.
 * @param offset the offset of the starting byte.
 * @param len the number of bytes to be read.
 * @param read_bytes the actual number of bytes being read.
 * @return status (0: success; otherwise, error reasons).
 */
int SorFile::read(Client *cli, HostMap *hmap, off64_t offset, char *buf, off_seg_t len, off_seg_t &read_bytes)
{
	if (attachment && data_segs) {
		return FILE_RES_EFILE; // very bad error!
	}

	if (attachment) {
		if (offset>MAX_ATTACHMENT_SIZE) {
			read_bytes=0;
			return FILE_RES_OKAY;
		}
		return attachment->read((off_seg_t)offset, buf, len, read_bytes);
	}
	else if (data_segs) {
		return data_segs->read(cli, hmap, offset, buf, len, read_bytes);
	}
	else {
		return FILE_RES_EFILE; // very bad error!
	}
}

/**
 * Get the size of the file.
 * @param size the size of the file (to be filled in).
 * @return status (0: success; otherwise, error reasons).
 */
int SorFile::get_size(off64_t & size)
{
	if (attachment && data_segs) {
		return FILE_RES_EFILE; // very bad error!
	}

	if (!attachment && !data_segs) {
		return FILE_RES_EFILE; // very bad error!
	}

	size = (attachment)?attachment->get_size(): data_segs->get_size();
	return FILE_RES_OKAY;
}

/**
 * Get the object ID of the file.
 * @param the object ID (to be filled in).
 * @return status (0: success; otherwise, error reasons).
 */
int SorFile::get_oid(ObjID &oid)
{
	oid=root_guid;
	return FILE_RES_OKAY;
}

/**
 * Write a certain number of bytes.
 * @param cli the Storage access point.
 * @param hmap the host map.
 * @param offset the offset of the starting byte.
 * @param len the number of bytes to be written.
 * @param written_bytes the actual number of bytes being written.
 * @return status (0: success; otherwise, error reasons).
 */
int SorFile::write(Client *cli, HostMap *hmap, off64_t offset, char *buf, off_seg_t len, off_seg_t &written_bytes)
{
	off64_t cur_size;
	int ret=get_size(cur_size);
	if (ret != FILE_RES_OKAY) {
		return ret;
	}

	if (offset+len > cur_size) {
		ret=this->truncate(cli, hmap, offset+len);
		if (ret != FILE_RES_OKAY)
			return ret;
	}

	try {
		if (attachment) {
			ret=attachment->write((off_seg_t)offset, buf, len, written_bytes);
		}
		else if (data_segs) {
			ret=data_segs->write(cli, hmap, offset, buf, len, written_bytes);
		}
		else {
			return FILE_RES_EFILE; // very bad error!
		}
	} catch (int errnum) {
		return errnum;
	}

	if (ret != FILE_RES_OKAY)
		return ret;

	return 0;
}

/**
 * Close a file.
 * @param ap the directory service access point.
 * @param cli the storage access point.
 * @param hmap the host map.
 * @return status (0: success; otherwise, error reasons).
 */
int SorFile::close(DIR_Cli_AP *ap, Client *cli, HostMap *hmap)
{
	if (attachment && data_segs) {
		return FILE_RES_EFILE; // very bad error!
	}

	int ret_final=FILE_RES_OKAY;
	off64_t actual_size;
	char buf[MAX_META_SIZE];
	SorMetaData *p_meta=(SorMetaData *)buf;
	int m_size=0;
	if (attachment) {
		actual_size=attachment->get_size();
		if (attachment->is_dirty()) {
			m_size=attachment->pack(p_meta, MAX_META_SIZE);
		}
	}
	else if (data_segs) {
		ret_final=data_segs->close(cli);
		actual_size=data_segs->get_size();
		if (data_segs->is_dirty()) {
			m_size=data_segs->pack(p_meta, MAX_META_SIZE);
		}
	}
	else {
		return FILE_RES_EFILE;
	}

	if (m_size>0) {
		// dirty meta-block, update it.
		off_seg_t bytes_written;
		int ret=FILE_RES_OKAY;
		ret=root_segment->write(cli, hmap, 0U, (char *)p_meta, (off_seg_t)m_size, bytes_written);
		int ret2=root_segment->close(cli);
		if (ret_final == FILE_RES_OKAY) {
			if (ret != FILE_RES_OKAY)
				ret_final = ret;
			else if (m_size != (int)bytes_written)
				ret_final = FILE_RES_ESPACE;
			else
				ret_final = ret2;
		}
	}
	else {
		ret_final=root_segment->close(cli);
	}
	
	// update dir_srv entry.
	off64_t cur_size;
	if (root.get_size(ap, cur_size) == DIR_RES_OKAY) {
		if (cur_size != actual_size) {
			DIR_Param param;

			memcpy(&param, root.get_param(), sizeof(DIR_Param));
			param.size=actual_size;

			root.setattr(ap, &param);
		}
	}
	else {
		// failed to update directory server data, but that is fine.
	}

	return ret_final;
}

/**
 * Unlink a file.
 * @param cli the storage access point.
 * @param hmap the host map.
 * @param guid the guid of the file to be unlinked.
 * @return status (0: success; otherwise, error reasons).
 * This is a static method and can be called without having
 * an active SorFile instance.
 */
int SorFile::unlink(Client *cli, HostMap *hmap, ObjID &guid)
{
	Segment *rseg = new Segment(guid, MAX_META_SIZE, SOR_WR, true);

	char buf[MAX_META_SIZE]; // 60KB.
	off_seg_t bytes_read;
	int ret=FILE_RES_OKAY;
	ret=rseg->read(cli, hmap, 0U, buf, MAX_META_SIZE, bytes_read);
	if (ret != FILE_RES_OKAY)
		return ret;
	SorMetaData *p_meta=(SorMetaData *)buf;
	if (!verify_meta_object(p_meta, (int)bytes_read))
		return FILE_RES_EMETA;

	int ret2=FILE_RES_OKAY;
	if (!(Attachment::is_legal_size(p_meta->size))) {
		DataSegments *dsegs = new DataSegments(p_meta, (int)bytes_read, SOR_WR);
		ret=rseg->remove(cli, hmap);
		delete rseg;
		ret2=dsegs->remove_all(cli, hmap);
		delete dsegs;
	}
	else {
		ret=rseg->remove(cli, hmap);
	}

	if (ret != FILE_RES_OKAY)
		return ret;

	if (ret2 != FILE_RES_OKAY)
		return ret2;

	return FILE_RES_OKAY;
}
