// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: srvcatalogupdater.cc,v $ - Thread that updates server
 *            catalog.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Aziz Gulbeden, gulbeden@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2004/04/29 02:17:14 $ by $Author: gulbeden $
 * Version :  $Revision: 1.15 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

# include <stdlib.h>
# include <unistd.h>
# include <pthread.h>
# include <fcntl.h>
# include <unistd.h>
# include <sys/types.h>
# include <set>

# include "host_map.h"
# include "protocol.h"
# include "replicator.h"
# include "sock_conn.h"			// SOCK_PORT_ANY
# include "srvcatalog.h"
# include "srvcatalogupdater.h"
# include "tmpi_debug.h"
# include "udp_cli.h"
# include "../../expr/fsd/heap.h"
# include "gsl/gsl_rng.h"

using namespace std;

class RNG {
 private:
	gsl_rng *gen;

 public:
	/** the constructor */
	RNG()
	{
		unsigned long seed; 

		int fd = open("/dev/urandom", O_RDONLY);		
		if (fd < 0) {
			time((time_t *)&seed);
			seed = seed ^ ((unsigned long)getpid())<<8;
		}
		else {
			read(fd, &seed, sizeof(seed));
			close(fd);
		}
		gen=gsl_rng_alloc(gsl_rng_default);
		gsl_rng_set(gen, seed);
	}

	/** the destructor */
	~RNG()
	{
		gsl_rng_free(gen);
	}

	/** getting a random double between [0, 1) */
	double get_uniform()
	{
		return gsl_rng_uniform(gen);
	}

	// return uint between [0, n-1].
	unsigned int get_uniform_int(unsigned int n)
	{
		return (unsigned int)gsl_rng_uniform_int(gen, (unsigned long)n);
	}
};

class TimeHeap
{
 private:
	pthread_mutex_t lock;
    HostID *hid_buffer;
    time_t *time_buffer;
	typedef HeapMin<time_t, HostID> time_heap_t;
	time_heap_t * heap;
	typedef set<HostID> hid_set_t;
	hid_set_t live_hosts;

	bool add_internal(time_t t, HostID &hid) 
	{
		hid_set_t::const_iterator iter = live_hosts.find(hid);
		if (iter != live_hosts.end()) {
			// found it.
			remove_internal(hid);
		}

		bool ret = heap->add(t, hid);
		ret = ret && (live_hosts.insert(hid).second);

		// attempt to roll back the effect.
		if (ret == false) {
			remove_internal(hid);
		}

		return ret;
	}

	bool remove_internal(HostID &hid) 
	{
		int cnt = heap->howMany();
		int index=-1;
		for (int i=0; i<cnt; i++) {
			if (memcmp(&(hid_buffer[i]), &hid, sizeof(HostID)) == 0) {
				index=i;
				break;
			}
		}
		if (index != -1) {
			heap->remove_by_index(index);
		}
		hid_set_t::iterator pi = live_hosts.find(hid);
		if (pi != live_hosts.end()) {
			live_hosts.erase(pi);
		}

		return true;
	}

	bool peek_internal(time_t now, HostID &hid) 
	{
		time_t t;

		if (heap->peekMin(t, hid) == false) {
			return false;
		}

		if (t > now) {
			return false;
		}

		heap->removeMin(t, hid);
		hid_set_t::iterator pi = live_hosts.find(hid);
		if (pi != live_hosts.end()) {
			live_hosts.erase(pi);
		}

		return true;
	}

 public:
 	TimeHeap(int size) :
	hid_buffer(NULL),
	time_buffer(NULL)
	{
		pthread_mutex_init(&lock, NULL);
		hid_buffer=(HostID *)calloc(size, sizeof(HostID));
		if (hid_buffer == NULL)
			throw string("TimeHeap::TimeHeap: out of memory");

		time_buffer=(time_t *)calloc(size, sizeof(time_t));
		if (time_buffer == NULL)
			throw string("TimeHeap::TimeHeap: out of memory");

		heap = new time_heap_t(size, time_buffer, hid_buffer);
	}

 	~TimeHeap() {
		if (heap)
			delete heap;

		if (time_buffer)
			free(time_buffer);

		if (hid_buffer)
			free(hid_buffer);

		pthread_mutex_destroy(&lock);
	}

	bool add(time_t t, HostID &hid) {
		if (pthread_mutex_lock(&lock)<0) {
			return false;
		}
		bool ret = add_internal(t, hid);
		pthread_mutex_unlock(&lock);
		return ret;
	}

	bool remove(HostID &hid) {
		if (pthread_mutex_lock(&lock)<0) {
			return false;
		}
		bool ret = remove_internal(hid);
		pthread_mutex_unlock(&lock);
		return ret;
	}

	// check if the minimum time is obsolete comparing
	// to now.
	bool peek(time_t now, HostID &hid) {
		if (pthread_mutex_lock(&lock)<0) {
			return false;
		}
		bool ret = peek_internal(now, hid);
		pthread_mutex_unlock(&lock);
		return ret;
	}
};

// time in between pull requests
// set it to two hours
# define CHECK_BACK_TIME 3600		
# define STG_SCU_UDP_TOUT 1000
# define HEAP_SIZE 1024

SrvCatalogUpdater::SrvCatalogUpdater(SrvCatalog *src, HostMap *hm, HostID hid,
									 Replicator *rep):
	sc(src),
	hmap(hm),
	myHID(hid),
	time_heap(NULL),
	rng(NULL),
	use_tcp(false),
	recv_buf(NULL),
	buf_size(0),
	replicator(rep)
{
    cli = new Udp_cli(NULL, SOCK_PORT_ANY, STG_SCU_UDP_TOUT);
	time_heap = new TimeHeap(HEAP_SIZE);
	rng = new RNG;
	buf_size = 64*1024; /*< set to 64K initially */
	recv_buf = (ObjID *)malloc(buf_size);
	if (!recv_buf) {
		throw string("SrvCatalogUpdater::SrvCatalogUpdater: out of memory");
	}
}

SrvCatalogUpdater::~SrvCatalogUpdater()
{
	if (recv_buf)
		free(recv_buf);

	delete rng;
	delete time_heap;
	delete cli;
}

/**
 * We start with UDP based pulling, but if we receive an
 * error from ANY server that the size does not fit in
 * UDP, then we permamently switch to TCP based pulling.
 * This design is based on the assumption that different
 * hosts should have similar number of objects.
 * 
 * Todo : change return values into bool or int.
 */
void SrvCatalogUpdater::pull_from_host(ServerAddr &saddr)
{
	while (1) {
		int hint=(use_tcp)?STG_USE_TCP:0;
		Stg_Request *req=new_stg_req_get_range(myHID);
		if (!req) {
			tmpi_error(DBG_INTERNAL, "Out of memory.\n");
			return;
		}
		
		Stg_Req_Handle reqh;
		int ret=stg_start_request_no_payload(cli, saddr, req, reqh, hint);
		stg_req_free(req);

		if (ret != STG_RES_OKAY) {
			tmpi_error(DBG_VERBOSE, "Failed to send request (err=%d).\n", ret);
			return;
		}

		Stg_Res_Handle resh;
		ret=stg_wait_response_with_payload(cli, reqh, resh, (char *)recv_buf, buf_size);
		if (ret == STG_RES_ESIZE) {
			if (use_tcp == false) {
				use_tcp = true;
				continue;
			}
			else {
				tmpi_error(DBG_VERBOSE, "Unexpected error code received (%d).\n", ret);
				return;
			}
		}

		// special handling for buffer-too-small error.
		if (ret==STG_RES_EBUFFER) {
			// resize the buffer
			buf_size=resh.response()->payload_size();
			recv_buf = (ObjID *)realloc(recv_buf, buf_size);
			if (!recv_buf) {
				tmpi_error(DBG_INTERNAL, "Out of memory.\n");
				return;
			}
			ret = stg_wait_response_with_payload_cont(cli, reqh, resh, (char *)recv_buf, buf_size);
		}

		if (ret != STG_RES_OKAY) {
			tmpi_error(DBG_VERBOSE, "Failed to get_range from remote server(%d).\n", ret);
			return;
		}

		Stg_Response *res=resh.response();

		if ( (res->header.tag != STG_RES_GET_RANGE_SMALL) && 
			 (res->header.tag != STG_RES_GET_RANGE_BIG) ) {
			tmpi_error(DBG_VERBOSE, "Unexpected response received (tag=%d).\n", res->header.tag);
			return;
		}

		int n_objs;
		HostID remoteHid;
		if (res->header.tag == STG_RES_GET_RANGE_SMALL) {
			n_objs=res->u.u_get_range_small.n_objs;
			remoteHid=res->u.u_get_range_small.hid;
		}
		else {
			n_objs=res->u.u_get_range_big.n_objs;
			remoteHid=res->u.u_get_range_big.hid;
		}

		ServerAddr saddr;
		if(hmap->get_address(remoteHid, saddr) == false){
			return;
		}
		
		for (int i=0; i<n_objs; i++) {
			sc->add(recv_buf[i], remoteHid);

			// Also replicator needs to know about these objects because it
			// keeps track of them. Maybe I (Aziz) should combine server 
			// catalog and replicator.

			// I assume that the remote version is 0, should be corrected
			// TODO
			
			// replicator has to keep track of this segment, since this node
			// is the home host.
			if(replicator)
				if(replicator->addSegment(recv_buf[i], saddr)==false){ 
					return /*STG_RES_ESERVER*/;
				}

			// This call is unnecessary after the first but we make it
			// over and over again, problem: there is no authority to
			// decide on replication degrees except for the repDegree
			// variable. problem2: I am assuming version is 0 because
			// that data is not sent.
					//	if(replicator)
		//if(replicator->setRepDegree(recv_buf[i], Replicator::repDegree, 0)== false){
   		//replicator->removeSegment(recv_buf[i]);
   		//return /*STG_RES_ESERVER*/;
		//   }
		}
		return;
	}

	return;
}

void *SrvCatalogUpdater::run() 
{
	while (!should_quit()) {
		time_t now;
		time(&now);

		HostID hid;

		if (time_heap->peek(now, hid) == false) {
			sleep(1);
			continue;
		}

		ServerAddr saddr;
		if (hmap->get_address(hid, saddr)) {
			pull_from_host(saddr);
			now = now + rng->get_uniform_int(100)+CHECK_BACK_TIME-50;
			time_heap->add(now, hid);
		}
		else {
			// cannot find the address of hid?
			// should seldom happen.
			time_heap->remove(hid);
		}
	}

	tmpi_error(DBG_VERBOSE, "Server catalog updater stopped\n");
	
    return NULL;
}

/**
 * adds the host hid to the heap, information is pulled from the server 
 * within a random amount of time.
 * @param hid HostID of the host to added.
 * @return status, true: success, false: failure.
 */
bool SrvCatalogUpdater::addHost(HostID &hid)
{
	tmpi_error(DBG_VERBOSE, "host join notify received\n");
	time_t now;
	time(&now);
	now = now + rng->get_uniform_int(10)+5;
	return time_heap->add(now, hid);
}

/**
 * adds the host hid to the leftHosts list so that 
 * when the tile comes, the 
 * host will be deleted from the heap.
 * @param hid HostID of the host to be deleted.
 * @return status, true: success, false: failure.
 */
bool SrvCatalogUpdater::removeHost(HostID &hid)
{
	tmpi_error(DBG_VERBOSE, "host remove notify received\n");
	return time_heap->remove(hid);
}
