// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: subscriber.cc,v $ - class implementation for load 
 *            subscriber.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/10/03 01:44:48 $
 * Version :  $Revision: 1.17 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

# include <netinet/in.h>

# include "tmpi_debug.h"
# include "sock_conn.h"
# include "../shared/daemon_common.h"
# include "subscriber.h"

using namespace std;

Subscriber::Subscriber(SorrentoConf &sconf) :
	ancee(NULL),
	expiration(1000) // default with a very long expiration time.
{
	char *mc_group=sconf.mc_group();
	int mc_port=sconf.mc_port();
	int interval=sconf.mc_interval();

	init(mc_group, mc_port, interval);
	free(mc_group);

	return;
}

Subscriber::Subscriber(const char *mc_group, int mc_port, int interval) :
	ancee(NULL),
	expiration(1000) // default with a very long expiration time.
{
	init(mc_group, mc_port, interval);
	return;
}

void Subscriber::init(const char *mc_group, int mc_port, int interval)
{
	struct sockaddr_in sin;

	if ((!mc_group)||(mc_port<=0)) {
		tmpi_error(DBG_CALLER, "Bad parameters: mc_group=%s, mc_port=%d.\n", mc_group, mc_port);
		throw string("Subscriber::init: Cannot get multicast channel for Subscriber().");
	}

	if (sock_filladdr(mc_group, mc_port, &sin)<0) {
		tmpi_error(DBG_CALLER, "Bad address: %s:%d.\n", mc_group, mc_port);
		throw string("Subscriber::init: Bad IP address for Subscriber().");
	}

	if (sock_is_mcast_addr(&sin, sizeof(sin))!=1) {
		tmpi_error(DBG_CALLER, "Bad multi-cast address: %s:%d.\n", mc_group, mc_port);
		throw string("Subscriber::init: Bad multicast address for Subscriber().");
	}

	ancee = new Mc_ancee(mc_group, mc_port);

	if (!ancee->good()) {
		tmpi_error(DBG_CALLER, "Mc_ancee() failed.\n");
		throw string("Subscriber::init: Cannot create multicast subscriber.");
	}

	expiration=(interval*EXPIRATION)/1000;

	// at least 5 seconds expiration time.
	if (expiration<5)
		expiration=5;

	return;
}

/**
 * note that the anc cound be from directory server or a normal server.
 * currently I use some heuristics to distinguish the two. maybe better
 * to add an explicit field for that.
 */
bool Subscriber::add_loads(Mc_anc *anc)
{
	Mc_rec **r;
	HostID hid;
	
	r=anc->lookup(KEY_SOR_IDENT);

	// one and only one
	if ((!r)||(!r[0])||(r[1])) {
		if (r)
			anc->free_result(r);
		return false;
	}

	string ident((char *)r[0]->val());
	anc->free_result(r);

	char ip[16];
	strcpy(ip, anc->addr());

	r=anc->lookup(KEY_SVR_IDENT);
	if ((!r)||(!r[0])) {
		// not a valid storage server announcement.
		if (r)
			anc->free_result(r);

		// check whether it is a good dir_srv announcement.
		r=anc->lookup(KEY_DIRSRV_PORT);
		if ((!r)||(!r[0])) {
			if (r)
				anc->free_result(r);
			return false;
		}
		int port;
		memcpy(&port, r[0]->val(), sizeof(int));
		anc->free_result(r);
		HostAddr dir_addr(ip, port);
		table.update_dir_addr(ident, dir_addr);
	}
	else {
		memcpy(&hid, (char *)r[0]->val(), sizeof(HostID));

		anc->free_result(r);

		HostLoad *load=new HostLoad(ident, hid);
		int tcp_port, udp_port;

		Mc_rec **records=anc->dump();

		if (!records) {
			delete load;
			return false;
		}

		struct InfoTemp {
			int cpu_total;
			int cpu_user;
			int cpu_kernel;
			int cpu_io;
			int mem_total;
			int mem_used;
			int swap_total;
			int swap_free;
			int n_disks;

			InfoTemp() :
				cpu_total(0),
				cpu_user(0),
				cpu_kernel(0),
				cpu_io(0),
				mem_total(0),
				mem_used(0),
				swap_total(0),
				swap_free(0),
				n_disks(0)
			{}

			bool valid(void)
			{
				if ((cpu_total==0)||(mem_total==0)||(swap_total==0)) {
					return false;
				}
				return true;
			}
		} infotmp;

		for (Mc_rec **p=records; *p; p++) {
			Mc_rec *rec=*p;
			switch (rec->var) {
				case KEY_CPU_TOT: 
					memcpy(&(infotmp.cpu_total), rec->val(), sizeof(infotmp.cpu_total));
					break;
				case KEY_CPU_USR:
					memcpy(&(infotmp.cpu_user), rec->val(), sizeof(infotmp.cpu_user));
					break;
				case KEY_CPU_KER:
					memcpy(&(infotmp.cpu_kernel), rec->val(), sizeof(infotmp.cpu_kernel));
					break;
				case KEY_CPU_IOW:
					memcpy(&(infotmp.cpu_io), rec->val(), sizeof(infotmp.cpu_io));
					break;
				case KEY_CPU_IDL:
					break;
				case KEY_CPU_NUM:
					memcpy(&(load->n_cpu), rec->val(), sizeof(load->n_cpu));
					break;
				case KEY_CPU_CXW:
					break;
				case KEY_MEM_TOT:
					memcpy(&(infotmp.mem_total), rec->val(), sizeof(infotmp.mem_total));
					load->mem_total=infotmp.mem_total;
					break;
				case KEY_MEM_USE:
					memcpy(&(infotmp.mem_used), rec->val(), sizeof(infotmp.mem_used));
					break;
				case KEY_MEM_FRE:
				case KEY_MEM_SHR:
				case KEY_MEM_BUF:
				case KEY_MEM_CCH:
					break;
				case KEY_SWP_TOT:
					memcpy(&(infotmp.swap_total), rec->val(), sizeof(infotmp.swap_total));
					load->swap_total=infotmp.swap_total;
					break;
				case KEY_SWP_FRE:
					memcpy(&(infotmp.swap_free), rec->val(), sizeof(infotmp.swap_free));
					break;
				case KEY_DSK_FRE:
					if (load->ndisks>MAX_DISKS)
						break;
					memcpy(&(load->disks[load->ndisks]), rec->val(), sizeof(load->disks[0]));
					load->ndisks++;
					break;
				case KEY_DSK_NUM:
					memcpy(&(infotmp.n_disks), rec->val(), sizeof(infotmp.n_disks));
					break;
				case KEY_SVR_UPTIME:
					memcpy(&(load->uptime), rec->val(), sizeof(load->uptime));
					break;
				case KEY_SVR_PORT:
					// memcpy(udp_port, rec->val(), sizeof(int));
					break;
				case KEY_SVR_QLEN:
					memcpy(&(load->qlen), rec->val(), sizeof(load->qlen));
					break;
				case KEY_SVR_CMHZ:
					memcpy(&(load->cpu_speed), rec->val(), sizeof(load->cpu_speed));
					break;
				case KEY_SVR_VNODES:
					memcpy(&(load->vnodes), rec->val(), sizeof(int));
					break;
				case KEY_SVR_TCP_PORT:
					memcpy(&tcp_port, rec->val(), sizeof(int));
					break;
				case KEY_SVR_UDP_PORT:
					memcpy(&udp_port, rec->val(), sizeof(int));
					break;
				default:
					;
			}
		}

		load->addr.init(ip, tcp_port, udp_port);

		free(records);

		if (!infotmp.valid()) {
			delete load;
			return false;
		}

		if (infotmp.n_disks != load->ndisks) {
			tmpi_error(DBG_APP, "Number of disks unmatched.\n");
		}

		load->cpu_comp=(float)(infotmp.cpu_user+infotmp.cpu_kernel)/infotmp.cpu_total;
		load->cpu_io=(float)(infotmp.cpu_io)/infotmp.cpu_total;
		load->mem_used=(float)(infotmp.mem_used)/infotmp.mem_total;
		load->swap_used=(float)(infotmp.swap_total-infotmp.swap_free)/infotmp.swap_total;

		table.update(load);

		// pthread_mutex_lock(&lock);
		list.push_back(load);
		// pthread_mutex_unlock(&lock);
	}

	return true;
}

void Subscriber::clean_expired()
{
	time_t now;

	// pthread_mutex_lock(&lock);
	time(&now);

	HostLoad *p=(HostLoad *)list.get_head();
	while (p && (now>(p->t_recv+expiration))) {
		tmpi_error(DBG_VERBOSE,"now = %u, t_recv = %u, expiration = %u.\n", now, p->t_recv, expiration);
		p->unlink();
		if (p->never_expire) {
			// special load item.
			p->t_recv=now;
			list.push_back(p);
		}
		else {
			table.remove(p);
			delete p;
		}
		p=(HostLoad *)list.get_head();
	}

	// pthread_mutex_unlock(&lock);

	return;
}

void * Subscriber::run(void)
{
	// tmpi_error(DBG_VERBOSE, "Load Subscriber enters the run() loop.\n");
	while (!should_quit()) {
		Mc_anc *anc=ancee->listen(1000);
		clean_expired();

		if (!anc) {
			continue;
		}

		// tmpi_error(DBG_VERBOSE, "Load received.\n");
		add_loads(anc);
		delete anc;

	}

	// tmpi_error(DBG_VERBOSE, "Load Subscriber is quitting normally ...\n");

	return NULL;
}
