// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: subscriber.h,v $ - header file for load subscriber.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/10/03 01:44:48 $
 * Version :  $Revision: 1.12 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 
# ifndef _SUBSCRIBER_H_
# define _SUBSCRIBER_H_

# include <time.h>
# include <map>
# include <string>
# include <pthread.h>

# include "cbs_thread.h"
# include "mcast_annc.h"
# include "sorrento_types.h"
# include "sorrento_conf.h"
# include "sor_dll.h"
# include "host_map.h"

/**
 * Usage:
 *
 * Subscriber *sub=new Subscriber(config);
 * 
 * sub->start();
 *
 * ...
 *
 * sub->stop();
 *
 * ...
 *
 * sub->finalize();
 *
 */

/**
 * We maintain a two-level mapping table:
 * The first level maps from Sorrento ID to a group of hosts share the same ID:
 * Sorrento-ID -> HostMap
 * In side each HostMap, the second level maps from Host ID to a Load struct:
 * 		HostID -> HostLoad (maintained by the HostMap class)
 */

struct sz_cmp : public std::less<std::string> {
	bool operator()(const std::string &x, const std::string &y) const {
		return (x<y);
	}
};

class GlobalMap :  std::map<std::string, HostMap *, sz_cmp> {
 private:
	pthread_mutex_t lock; /**< the lock protecting the map */
 public:

	GlobalMap() {
		pthread_mutex_init(&lock, NULL);
	}

	// I create a new map on demand. The returned pointer
	// can be used later to retrieve the whole list of
	// hosts.
	HostMap *lookup(const std::string &ident)
	{
		HostMap *ret=NULL;

		pthread_mutex_lock(&lock);

		GlobalMap::const_iterator pi=this->find(ident);

		if (pi==this->end()) {
			ret=new HostMap(ident.c_str());
			this->insert(GlobalMap::value_type(ident, ret));
		}
		else {
			ret=pi->second;
		}

		pthread_mutex_unlock(&lock);

		return ret;
	}

	// returns true when a new host is added, false
	// when the host already exists.
	bool update(HostLoad *load)
	{
		HostMap *hmap=lookup(load->ident);
		return hmap->update(load);
	}

	bool update_dir_addr(std::string &ident, HostAddr &ha)
	{
		HostMap *hmap=lookup(ident);
		return hmap->set_dir_addr(ha);
	}

	// return true when load is found and removed.
	// false when load is not found.
	bool remove(HostLoad *load)
	{
		HostMap *hmap=NULL;
		hmap=lookup(load->ident);
		return hmap->remove(load);

		/**
		pthread_mutex_lock(&lock);
		GlobalMap::iterator pi=this->find(load->ident);
		if (pi == this->end()) {
			pthread_mutex_unlock(&lock);
			return false;
		}

		pthread_mutex_unlock(&lock);

		return pi->second->remove(load);
		**/
	}

	~GlobalMap() 
	{
		// release all HostMap structures.
		// release all HostLoad structures.
		for (GlobalMap::iterator pi=this->begin(); pi!=this->end(); pi++) {
			if (pi->second) {
				delete pi->second;
			}
			pi->second=NULL;
		}
		pthread_mutex_destroy(&lock);
	}
};

/**
 * When we have not received any heart beat packets
 * for EXPIRATION intervals, we assume the host
 * is dead.
 */
# define EXPIRATION 5

/**
 * Subscriber also needs to manage an vector of 
 * virtual node IDs for the purpose of consistent 
 * hashing. The array is updated in two places:
 * (1) when table.update() returns true, we need
 * to insert new virtual nodes corresponds to the
 * new host ID.
 * (2) in check_expired(), when we found an expired
 * host, we need to unlink it from the list, then
 * remove it from the corresponding HostMap, and
 * finally remove then from the array of virtual
 * nodes.
 */
// I assume that the danger of not guarding
// the top level mapping table is low, because
// it should be updated very infrequently. 
// Plan to change the top level GlobalMap to an array
// instead of a <map> struct and make sure that 
// each new HostMap is never deallocated once it
// is stored into the array. To find host structs
// based on ident, we just need to scan the whole
// array.
class Subscriber : public CBS_Thread {
 private:
	Mc_ancee *ancee;
	// HostLoad structures sorted by t_recv.
	DoubleLinkList list; 
	GlobalMap table;
	int expiration;
	// RWLock lock;
	// pthread_mutex_t lock; /**< the lock protecting list */

	bool add_loads(Mc_anc *anc);
	void clean_expired();
	void init(const char *mc_group, int mc_port, int interval);

 public:
	Subscriber(SorrentoConf &sconf);
	Subscriber(const char *mc_group, int mc_port, int interval);

	virtual ~Subscriber(void)
	{
		if (ancee)
			delete ancee;
	}

	virtual void * run(void);

	HostMap *get_host_map_by_ident(const char *ident)
	{
		std::string str_ident(ident);
		return table.lookup(str_ident);
	}
};

typedef Subscriber Proxy;

# endif
