// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: xput_tracker.cc,v $ - Module that is responsible from
 *            replication. 
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Aziz Gulbeden, gulbeden@cs.ucsb.edu
 *            William Strathearn, bill@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2004/04/10 03:12:16 $ by $Author: bill $
 * Version :  $Revision: 1.2 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

# include "client.h"			// to connect to other servers
# include "host_map.h"
# include "protocol.h"
# include "xput_tracker.h"
# include "tmpi_debug.h"
# include "udp_cli.h"
# include "verfs.h"

# define STG_SCU_UDP_TOUT 1000

Xput_Tracker::Xput_Tracker(HostMap *hostMap):
        hm(hostMap)
{

	cli = new Client();        
	if (pthread_mutex_init(&blockMapLock, NULL) != 0)
                throw -1;
        udpCli = new Udp_cli(NULL, SOCK_PORT_ANY, STG_SCU_UDP_TOUT);
      

}


Xput_Tracker::~Xput_Tracker(){
        delete cli;
	//TODO delete the blockMap
        pthread_mutex_destroy(&blockMapLock);
}

bool Xput_Tracker::addXput(int desc, ClientID &cid, int bytes){

	//lookup desc entry in blockMap
	blockMapType::iterator block; 
    if(pthread_mutex_lock(&blockMapLock) != 0)
	    return false;	
	//lookup cid,desc pair 
	block = blockMap.find(desc);
	if(block == blockMap.end()){
		tmpi_error(DBG_VERBOSE, "We did not find desc: %d, why not?\n",desc);
        dumpBlockMap();
        return false; 
	}else{
		blockAccessType access = (*block).second;
        blockAccessType::iterator clients_access;
        clients_access = access.find(cid);
        if(block == blockMap.end()){
            //insert new entry for cid,desc
            tmpi_error(DBG_VERBOSE, "We did not find cid: %d for desc: %d, why not?\n",cid.id,desc);
            dumpBlockMap();
            return false;
        }else{
		    (*clients_access).second = (*clients_access).second + bytes; 
            tmpi_error(DBG_VERBOSE, "Updating client:%d access bytes for desc: %d to %d..\n",cid.id,desc,(*clients_access).second);
            (*block).second = access;
		    tmpi_error(DBG_VERBOSE, "Block has been updated\n");
        }
	}
    pthread_mutex_unlock(&blockMapLock);    
    return true;
}


void Xput_Tracker::dumpBlockMap(){

  blockMapType::iterator i; 
  tmpi_error(DBG_VERBOSE, "BLOCK MAP DUMP =======\n");
  tmpi_error(DBG_VERBOSE, "SIZE: %d\n", blockMap.size());

  for(i=blockMap.begin(); i != blockMap.end(); i++){
     tmpi_error(DBG_VERBOSE, "\tDESC: %03d -----\n", (*i).first);
     dumpAccessType((*i).second); 
  }
  tmpi_error(DBG_VERBOSE, "======================\n", blockMap.size());
  return;
}

void Xput_Tracker::dumpAccessType(blockAccessType access){

  blockAccessType::iterator i; 

  tmpi_error(DBG_VERBOSE, "\tBLOCK ACCESS DUMP =======\n");
  tmpi_error(DBG_VERBOSE, "\tSIZE: %d\n", access.size());

  for(i=access.begin(); i != access.end(); i++){
     tmpi_error(DBG_VERBOSE, "\t\t  CID: %03d -----\n", (*i).first.id);
     tmpi_error(DBG_VERBOSE, "\t\tBYTES: %d -----\n", (*i).second); 
  }
  
  tmpi_error(DBG_VERBOSE, "\t:======================\n", blockMap.size());
  return;

}

  



//assuming caller does not induce transfer traffic at this time...
//assuming that saddr is the same for every creation...
bool Xput_Tracker::addSegment(int desc, const ClientID &cid){
   
    blockAccessType blockAccess;   
  
    //blockListType::iterator same;
    if(pthread_mutex_lock(&blockMapLock) != 0)
		return false;
	
	//check to make sure desc is not in map
    if(blockMap.find(desc) == blockMap.end()){
		tmpi_error(DBG_VERBOSE, "%d, %d is not already in the Block Map\n",desc,cid.id);
		//create new map to hold server ID's and xput figures
		pair<const ClientID, long> blockAccessEntry(cid,0);
        blockAccess.insert(blockAccessEntry);
		pair<int, blockAccessType> blockMapEntry(desc,blockAccess);
		//insert new map
		blockMap.insert(blockMapEntry);
		tmpi_error(DBG_VERBOSE, "Block has been inserted \n");
	}else{
        tmpi_error(DBG_VERBOSE, "%d, %d is already in the Block Map\n",desc,cid.id);
	}
        
	pthread_mutex_unlock(&blockMapLock);
	return true; 
}


bool Xput_Tracker::removeSegment(int desc){
	return true;
}

void *Xput_Tracker::run(){

    while (!should_quit()) {
		sleep(30);
        dumpBlockMap();
	}
    tmpi_error(DBG_VERBOSE, "Xput_Tracker stopped\n");

    return NULL;
}

bool Xput_Tracker::start(){

        // first call parent class
        CBS_Thread::start();

        // now learn this server's address
        return true;
}
