// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*- 

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: const_data_adaptor.h,v $ - header file for the simplest
 *            data collector/extractors - constants
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2002/10/14 08:21:12 $
 * Version :  $Revision: 1.1 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

# ifndef _CONST_DATA_ADAPTOR_H_
# define _CONST_DATA_ADAPTOR_H_

# include "mcast_adaptor.h"

/** MDC_Const: multicast data collectors for constants */
class MDC_Const : public MDC {
 private:
	/** @name private variables **/
	//@{
	// the type of the data
	unsigned char type;
	// the address of the data.
	void *data;
	// the length of the data.
	unsigned short len;
	// is the data private or a pointer passed in?
	bool is_private;
	// is it the first multicast?
	bool first_time;
	// last reference time.
	int last_time;
	// mcast interval
	int interval;
	//@}
 public:
	/** @name public methods **/
	//@{
	/**
	 * Constructor
 	 *  @param   t - type of the data
 	 *  @param   d - data
 	 *  @param   l - length of the data
 	 *  @param   i - multicast interval
 	 *  @param   do_copy - should we copy data?
	 */
	MDC_Const(unsigned char t, void *d, unsigned short l, int i, bool dup=false) :
		type(t),
		data(d),
		len(l),
		is_private(false),
		first_time(true),
		last_time(0),
		interval(i)
	{
		if ((!d)||(i<=0))
			throw -1;

		if (dup) {
			void *tmp=malloc(l);
			if (!tmp) {
				throw -1;
			}
			memcpy(tmp, d, l);
			data=tmp;
			is_private=true;
		}
	}

	/**
 	 * read the local state data and put it into a
	 * multicast announcement packet. Override the
	 * pure abstract method in MDC class.
 	 *  @param   pkt - multicast announcement packet
 	 *  @param   ms_time - current reference timer expressed 
	 * 	         in milli-seconds, intend to be used to allow
	 * 	         data collectors implement their own update
	 * 	         frequencies. E.g. one adaptor may wish to
	 * 	         only insert an update field every 1 second,
	 * 	         it may achieve that by remembering the last
	 * 	         time it was called can check if the current
	 * 	         time is 1 second after the previous update
	 * 	         time.
 	 *  @return  true: a field is successfully added; 
	 *           false: no field is added (not necessarily 
	 *             an error).
 	 */
	virtual bool put(Mc_anc_data &pkt, int ms_time)
	{
		if (ms_time>=(last_time+interval)) {
			// nothing
		}
		else if (first_time) {
			first_time=false;
		}
		else {
			return false;
		}

		if (pkt.add(type, data, len)==0) {
			last_time=ms_time;
			return true;
		}

		return false;
	}

	/** destructor */
	virtual ~MDC_Const() {
		if (is_private) {
			if (data)
				free(data);
		}
	}
	//@}
}; // end of class MDC_Const

/**
 * MDE_Const : multicast data extractors for constants
 * Not sure if it would be useful at all.
 * I just keep it here as a template to implement other
 * MDE adaptors.
 */

class MDE_Const : public MDE {
 private:
	/** @name private variables **/
	//@{
	// the type of the data interested
	unsigned char type;
	// should we discard the data or truncate it under
	// buffer overflow.
	bool discard_if_overflow;
	// the address of the buffer where data can be deposited.
	void *buf;
	// the length of the data buffer.
	int len;
	//@}

 public:
	/** @name public methods  **/
	//@{
	/**
	 * Constructor
 	 *  @param   t - type of the data
 	 *  @param   b - buffer
 	 *  @param   l - length of the buffer
 	 *  @param   discard - should we discard if buffer overflows?
	 */
	MDE_Const(unsigned char t, void *b, int l, bool discard=true) :
		type(t),
	    discard_if_overflow(discard),
		buf(b),
		len(l)
	{
		if ((!b)||(l<=0))
			throw -1;
	}

	/**
 	 * read multicast announcement packet and update
	 * the local state data
 	 *  @param   pkt - multicast announcement packet
 	 *  @return  true: a field is successfully extracted 
	 *             and local state updated; 
	 *
	 *           false: no field is extracted and no 
	 *             local state update (not signify an 
	 *             error).
 	 */
	virtual bool get(Mc_anc &pkt)
	{
		Mc_rec **recs=pkt.lookup(type);
		if ((!recs)||(recs[0]==NULL)) {
			return false;
		}

		int xlen=recs[0]->len;
		if (xlen>len) {
			if (discard_if_overflow) {
				return false;
			}
			else {
				xlen=len;
			}
		}

		memcpy(buf, recs[0]->val(), xlen);

		return true;
	}

	/** destructor */
	virtual ~MDE_Const() {}
	//@}
}; // end of class MDE_Const

# endif // _CONST_DATA_ADAPTOR_H_
