// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*- 

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: mcast_publisher.h,v $ - header file for multi-cast publisher.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/07/16 06:40:23 $
 * Version :  $Revision: 1.2 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

# ifndef _MCAST_PUBLISHER_H_
# define _MCAST_PUBLISHER_H_

# include <netinet/in.h>
# include <unistd.h>
# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <pthread.h>

# include <vector>

# include "tmpi_debug.h"
# include "usc_time.h"
# include "sock_conn.h"
# include "cbs_thread.h"
# include "mcast_adaptor.h"

/** 
 * A generic multicast publisher.
 * Usage:
 *  // first create an publisher object which
 *  // publishes info to svc_addr:svc_port.
 *  MC_Publisher publisher(svc_addr, svc_port, 100);
 *  // add a few data-collector adaptors.
 *  publisher.add(load_data_collector);
 *  publisher.add(...);
 *  publisher.add(...);
 *  // start the publisher thread
 *  publisher.start();
 * 
 *  ...
 *
 *  // stop the publisher thread
 *  publisher.stop(...);
 *  // wait till the publisher quits
 *  publisher.finalize(...);
 *  // release the data-collector adaptors
 *  delete load_data_collector;
 *  delete ...
 *
 *  // note that it is okay to add adaptors
 *  // after the publisher is started. However
 *  // I am not sure if that would be 
 *  // a common usage.
 */

class MC_Publisher : public CBS_Thread {
 private:
 	/** @name private type definition */
	//@{
	// a vector of multicast data collector adaptors.
	typedef vector<MDC *> VP_MDC;

	// private enum constants.
	// MAX_ERRORS maximum number of errors tolerated.
	enum {MAX_ERRORS=1000};
	//@}

 	/** @name private member variables */
	//@{
	// the set of data collector adaptors
	VP_MDC vp_mdc;
	// lock protecting vp_mdc
	pthread_mutex_t lock;
	// multicast announcer
	Mc_ancer *ancer;
	// check interval in milli-seconds.
	int ms_int; 
	// the descriptive name of the publisher
	char *name; 
	//@}

 public:
 	/** @name public member methods */
	//@{
	/**
	 * Constructor:
	 *  @param  id - descriptive name of the publisher 
	 *            (e.g. ISC-publisher).
	 *  @param  mgrp - multicast IP/host address (as a string).
	 *  @param  mg_port - multicast port.
	 *  @param  interval - (optional), collect data and attempt
	 *          to publish data every interval milli-seconds.
	 */
	MC_Publisher(char *id, char *mgrp, int mg_port, int interval=100) :
		ancer(NULL),
		ms_int(interval),
		name(strdup(id))
	{
		struct sockaddr_in sin;

		if ((!mgrp)||(mg_port<=0)||(!name)) {
			throw -1;
		}

		if (sock_filladdr(mgrp, mg_port, &sin)<0) {
			throw -1;
		}

		if (sock_is_mcast_addr(&sin, sizeof(sin))!=1) {
			throw -1;
		}

		if (pthread_mutex_init(&lock, NULL)<0) {
			throw -1;
		}

		ancer=new Mc_ancer(mgrp, mg_port);

		if (!ancer->good()) {
			pthread_mutex_destroy(&lock);
			delete ancer;
			throw -1;
		}
	}

	// Virtual destructor
	virtual ~MC_Publisher() {
		pthread_mutex_destroy(&lock);
		if (ancer)
			delete ancer;

		if (name)
			free(name);
	}

	// thread loop  routine.
	virtual void *run()
	{
		unsigned int ref_time_ms=0;
		int errors=0;
		while (!should_quit()) {
			if (errors>MAX_ERRORS) {
				tmpi_error(DBG_APP, "%s quitting due to too many errors ...\n", name);
				return NULL;
			}

			Mc_anc_data pkt;
			int vp_mdc_size=vp_mdc.size();
			int n_items=0;
			for (int i=0; i<vp_mdc_size; i++) {
				if (vp_mdc[i]->put(pkt, ref_time_ms))
					n_items++;
			}

			if (n_items>0) {
				if (ancer->announce(&pkt)) {
					tmpi_error(DBG_INTERNAL, "%s info publishing failed.\n", name);
					errors++;
				}
			}

			ref_time_ms+=random_usleep(ms_int*1000)/1000;
		}

		tmpi_error(DBG_APP, "%s quitting normally ...\n", name);

		return NULL;
	}

	/**
	 * add a data collector adaptor
	 *  @param   mdc - the adaptor to be added.
	 *  @return  true: mdc added successfully, false: faled.
	 */
	bool add_adaptor(MDC *mdc)
	{
		if (!ancer)
			return false;

		if (pthread_mutex_lock(&lock)<0) {
			return false;
		}

		vp_mdc.push_back(mdc);

		pthread_mutex_unlock(&lock);

		return true;
	}

	//@}

 private:
 	/** @name private member methods */
	//@{
	/** 
	 * do usleep with a random interval between 0.5*usec and
	 * 1.5*usec. 
	 *  @param   usec - suggested sleep value.
	 *  @return  the actual # of usec slept.
	 */
	unsigned int random_usleep(unsigned int usec)
	{
		usc_time_t t1, t2;
		t1=usc_gettime();
		int least=usec/2;
		// I expect srandom is set at the application level.
		usleep(least+random()%usec);
		t2=usc_gettime();
		return (unsigned)(int)(usc_difftime(t2, t1));
	}
	//@}
}; // end of class PC_Publisher

# endif // _MCAST_PUBLISHER_H_
