// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: scsd_format.cc,v $ - tool to format a ISC partition.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/06/03 00:30:50 $
 * Version :  $Revision: 1.1 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 
# include <sys/types.h>
# include <sys/stat.h>
# include <fcntl.h>
# include <dirent.h>
# include <stdio.h>
# include <stdlib.h>
# include <time.h>
# include <unistd.h>
# include <getopt.h>
# include <string.h>

# include "resmon.h"
# include "sock_conn.h"

char *prog=NULL;

static void do_help()
{
	printf("\
Usage: %s [options] [mount].\n\
   -f         Format even if the partition is not empty. \n\
   -i         Do not perform mount point check. \n\
   -g         Debug mode (data directory level reduced from 4 to 2).\n\
   -h         Help, this screen.\n\
", prog);
	exit(0);
}

/**
 * Hexical to char conversion
 */
static char h2c(int h)
{
	if ((h<0)||(h>15)) {
		return (char)0;
	}

	if (h<10) {
		return '0'+h;
	}

	return 'A'+(h-10);
}

static int time_to_str(time_t tt, char *s, int l)
{
	if (!s)
		return -1;

	struct tm ttm;
	memcpy(&ttm, gmtime(&tt), sizeof(struct tm));
	if (strftime(s, l, "%b %e %T %Y", &ttm)==0)
		return -1;
	return 0;
}

int main(int argc, char *argv[])
{
	bool force=false;
	bool check=true;
	bool debug=false;

	prog=argv[0];
	
	int c;
	
	while ((c=getopt(argc, argv, "figh"))!=EOF) 
	{
		switch (c) {
			case 'f':
				force=true;
				break;
			case 'i':
				check=false;
				break;
			case 'g':
				debug=true;
				break;
			case 'h':
				do_help();
				break;
			default:
				do_help();
				break;
		}
	}

	if ((argc-optind)<=0) {
		fprintf(stderr, "No mount point specified.\n");
		do_help();
	}

# define MAX_MP_LEN 1024
	char mp[MAX_MP_LEN];

	if (strlen(argv[optind])>=MAX_MP_LEN) {
		fprintf(stderr, "Mount point path too long.\n");
		exit(-1);
	}
	
	if (check) {
		if (ResMon::find_mount_point(argv[optind], mp, MAX_MP_LEN)) {
			fprintf(stderr, "Cannot determine mount point.\n");
			exit(-1);
		}
	}
	else {
		strcpy(mp, argv[optind]);
	}

	char pwd[MAX_MP_LEN];
	if (getcwd(pwd, MAX_MP_LEN)==NULL) {
		fprintf(stderr, "Current directory path too long.\n");
		exit(-1);
	}

	if (chdir(mp)<0) {
		fprintf(stderr, "Cannot access directory %s.\n", mp);
		exit(-1);
	}
	
	int ret=0;

	DIR *mp_dir=opendir(".");
	if (mp_dir==NULL) {
		fprintf(stderr, "Cannot open directory %s.\n", mp);
		exit(-1);
	}

	struct dirent *dent;
	bool success=true;

	while ((dent=readdir(mp_dir))) {
		if (strcmp(dent->d_name, "lost+found")) {
			if (
					(strcmp(dent->d_name, "..")==0) ||
					(strcmp(dent->d_name, ".")==0)) {
				continue;
			}
			if (force==false) {
				fprintf(stderr, "Directory not empty, format aborted.\n");
				success=false;
				break;
			}
			else {
				/* attempt to remove it */
				char cmd[1024];
				snprintf(cmd, 1024, "/bin/rm -rf %s", dent->d_name);
				printf("Removing files/directories: %s/%s.\n", mp, dent->d_name);
				if (system(cmd)<0) {
					fprintf(stderr, "Cannot remove directories.\n");
					success=false;
					break;
				}
			}
		}
	}
	
	closedir(mp_dir);

	if (!success) {
		ret=-1;
		goto _exit;
	}

	{
		int err=0;

# define MODE 0700
# define DIR_DATA ".data"
# define DIR_ETC ".etc"
# define DIR_TMP ".tmp"
# define DIR_DB ".db"
# define FILE_SIG ".SCS_ROOT"

		err+=mkdir(DIR_DATA, MODE);
		err+=mkdir(DIR_ETC, MODE);
		err+=mkdir(DIR_TMP, MODE);
		err+=mkdir(DIR_DB, MODE);

		char d_tmp[1024];
		int total=0x10*0x10;
		printf("0%%");
		fflush(stdout);

		for (int i=0; i<0x10; i++) {
			sprintf(d_tmp, "%s/%c", DIR_DATA, h2c(i));
			err+=mkdir(d_tmp, MODE);
			for (int j=0; j<0x10; j++) {
				sprintf(d_tmp, "%s/%c/%c", DIR_DATA, h2c(i), h2c(j));
				err+=mkdir(d_tmp, MODE);
				if (!debug) {
					for (int k=0; k<0x10; k++) {
						sprintf(d_tmp, "%s/%c/%c/%c", DIR_DATA, h2c(i), h2c(j), h2c(k));
						err+=mkdir(d_tmp, MODE);
						for (int l=0; l<0x10; l++) {
							sprintf(d_tmp, "%s/%c/%c/%c/%c", 
									DIR_DATA, h2c(i), h2c(j), h2c(k), h2c(l));
							err+=mkdir(d_tmp, MODE);
						}
					}
				}
				printf("\r%d%%", (i*0x10+j+1)*100/total);
				fflush(stdout);
			}
		}

		printf("\n");

		if (err) {
			fprintf(stderr, "Cannot create directories.\n");
			ret=-1;
			goto _exit;
		}
	}

	{
		FILE *fp=fopen(FILE_SIG, "w");
		if (!fp) {
			fprintf(stderr, "Cannot create files.\n");
			ret=-1;
			goto _exit;
		}

		char ts[256]="";
		time_to_str(time(NULL), ts, 256);

		fprintf(fp, "Date: %s\n", ts);
		fprintf(fp, "Host: %s\n", sock_hostname());
		fprintf(fp, "Prime IP: %s\n", sock_prime_ip());
		fclose(fp);

	}

	printf("Format complated.\n");

_exit:
	chdir(pwd);
	return ret;
}
