/* -*- Mode: C; c-basic-offset:4 ; -*- */
/* 
 * Debug functions for tracing PVFS API calls.
 */

#ifdef DEBUG_TRACE

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#include <pvfs.h>
#include <sys/uio.h>
#include "usc_time.h"


#define DEFAULT_TRACE_FILE  "/tmp/pvfs.trace.log"

FILE *log = NULL;


void
open_log()
{
	char *file = NULL;
	if ( (file = getenv("PVFS_TRACE_FILE")) == NULL ) {
		file = DEFAULT_TRACE_FILE;
	}

	log = fopen(file, "w+");
	if ( log == NULL ) {
		fprintf(stderr, "Error opening log file: %s\n", file);
		exit(0);
	}

	t_usc_init();
	fprintf(log, "wall_clock %u, usc_time %llx, MHz rate %.3f\n",
			time(NULL),
			usc_gettime(),
			usc_mhzrate());
}


void
debug_log(const char *fmt, ...)
{
	va_list vl;

	if ( log == NULL )
		open_log();

	va_start(vl, fmt);
	vfprintf(log, fmt, vl);
	va_end(vl);

	fflush(log);
}


/* ----------------------------------------------------------------------
 * PVFS wrapper functions
 * ---------------------------------------------------------------------- */
int
debug_pvfs_open64(const char* pathname, int flag, int mode, void *pstat, void *last)
{
	int ret;
	usc_time_t start_t, end_t;

	start_t = usc_gettime();
	ret = pvfs_open64(pathname, flag, mode, pstat, last);
	end_t = usc_gettime();

	debug_log("start = %llx, (%d)pvfs_open64(%s, %d, %d), end = %llx, duration = %.3f\n",
			start_t,
			ret,
			pathname,
			flag,
			mode,
			end_t,
			usc_difftime(end_t, start_t));
	return ret;
}

int
debug_pvfs_close(int fd)
{
	int ret;
	usc_time_t start_t, end_t;

	start_t = usc_gettime();
	ret = pvfs_close(fd);
	end_t = usc_gettime();

	debug_log("start = %llx, (%d)pvfs_close(%d), end = %llx, duration = %.3f\n",
			start_t,
			ret,
			fd,
			end_t,
			usc_difftime(end_t, start_t));
	return ret;
}

int
debug_pvfs_unlink(const char *pathname)
{
	int ret;
	usc_time_t start_t, end_t;

	start_t = usc_gettime();
	ret = pvfs_unlink(pathname);
	end_t = usc_gettime();

	debug_log("start = %llx, (%d)pvfs_unlink(%s), end = %llx, duration = %.3f\n",
			start_t,
			ret,
			pathname,
			end_t,
			usc_difftime(end_t, start_t));
	return ret;
}

int64_t
debug_pvfs_lseek64(int fd, int64_t off, int whence)
{
	int64_t ret;
	usc_time_t start_t, end_t;

	start_t = usc_gettime();
	ret = pvfs_lseek64(fd, off, whence);
	end_t = usc_gettime();

	debug_log("start = %llx, (%lld)pvfs_lseek64(%d, %lld, %d), end = %llx, duration = %.3f\n",
			start_t,
			ret,
			fd,
			off,
			whence,
			end_t,
			usc_difftime(end_t, start_t));
	return ret;
}


int
debug_pvfs_ioctl(int fd, int cmd, void *data)
{
	int ret;
	usc_time_t start_t, end_t;

	start_t = usc_gettime();
	ret = pvfs_ioctl(fd, cmd, data);
	end_t = usc_gettime();

	debug_log("start = %llx, (%d)pvfs_ioctl(%d, %d, %p), end = %llx, duration = %.3f\n",
			start_t,
			ret,
			fd,
			cmd,
			data,
			end_t,
			usc_difftime(end_t, start_t));
	return ret;
}

int
debug_pvfs_ftruncate64(int fd, int64_t length)
{
	int ret;
	usc_time_t start_t, end_t;

	start_t = usc_gettime();
	ret = pvfs_ftruncate64(fd, length);
	end_t = usc_gettime();

	debug_log("start = %llx, (%d)pvfs_ftruncate64(%d, %lld), end = %llx, duration = %.3f\n",
			start_t,
			ret,
			fd,
			length,
			end_t,
			usc_difftime(end_t, start_t));
	return ret;
}

int
debug_pvfs_read(int fd, char *buf, size_t count)
{
	int ret;
	usc_time_t start_t, end_t;

	start_t = usc_gettime();
	ret = pvfs_read(fd, buf, count);
	end_t = usc_gettime();

	debug_log("start = %llx, (%d)pvfs_read(%d, %p, %u), end = %llx, duration = %.3f\n",
			start_t,
			ret,
			fd,
			buf,
			count,
			end_t,
			usc_difftime(end_t, start_t));
	return ret;
}

int 
debug_pvfs_read_list(int     fd,
		            int     mem_list_count,
					char   *mem_offsets[],
					int mem_lengths[],
					int file_list_count,
					int64_t file_offsets[],
					int32_t file_lengths[])
{
	int ret, i;
	usc_time_t start_t, end_t;

	start_t = usc_gettime();
	ret = pvfs_read_list(fd, mem_list_count, mem_offsets, mem_lengths, file_list_count,
			file_offsets, file_lengths);
	end_t = usc_gettime();

	debug_log("start = %llx, (%d)pvfs_read_list(%d, %d), end = %llx, duration = %.3f\n",
			start_t,
			ret,
			mem_list_count,
			file_list_count,
			end_t,
			usc_difftime(end_t, start_t));
	for ( i = 0; i < file_list_count; i++ ) {
		debug_log("\t%d %d\n", file_offsets[i], file_lengths[i]);
	}
	return ret;
}

int
debug_pvfs_write(int fd, char *buf, size_t count)
{
	int ret;
	usc_time_t start_t, end_t;

	start_t = usc_gettime();
	ret = pvfs_write(fd, buf, count);
	end_t = usc_gettime();

	debug_log("start = %llx, (%d)pvfs_write(%d, %p, %u), end = %llx, duration = %.3f\n",
			start_t,
			ret,
			fd,
			buf,
			count,
			end_t,
			usc_difftime(end_t, start_t));
	return ret;
}

int
debug_pvfs_writev(int fd, const struct iovec *vector, size_t count)
{
	int ret;
	usc_time_t start_t, end_t;
	int i, total = 0;

	start_t = usc_gettime();
	ret = pvfs_writev(fd, vector, count);
	end_t = usc_gettime();

	for ( i = 0; i < count; i++ )
		total += vector[i].iov_len;

	debug_log("start = %llx, (%d)pvfs_writev(%d, %d), end = %llx, duration = %.3f\n",
			start_t,
			ret,
			fd,
			total,
			end_t,
			usc_difftime(end_t, start_t));
	return ret;
}


int 
debug_pvfs_write_list(int     fd,
		            int     mem_list_count,
					char   *mem_offsets[],
					int *mem_lengths,
					int file_list_count,
					int64_t file_offsets[],
					int32_t file_lengths[])
{
	int ret, i;
	usc_time_t start_t, end_t;

	start_t = usc_gettime();
	ret = pvfs_write_list(fd, mem_list_count, mem_offsets, mem_lengths, file_list_count,
			file_offsets, file_lengths);
	end_t = usc_gettime();

	debug_log("start = %llx, (%d)pvfs_write_list(%d, %d), end = %llx, duration = %.3f\n",
			start_t,
			ret,
			mem_list_count,
			file_list_count,
			end_t,
			usc_difftime(end_t, start_t));
	for ( i = 0; i < file_list_count; i++ ) {
		debug_log("\t%d %d\n", file_offsets[i], file_lengths[i]);
	}
	return ret;
}




#endif
