// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

# include <fcntl.h>
# include <pthread.h>

//# include "sor_client.h"
# include "SClientStub.h"
# include "client.h"
# include "sor_fs.h"
# include "sor_file.h"
# include "../dir_srv/dir_cli_ap_udp.h"
# include "tmpi_debug.h"
# include <string>
# include "debug.h"

# define CID_NULL 		0x0
# define CID_CHDIR 		0x1
# define CID_MKDIR 		0x2
# define CID_RMDIR 		0x3
# define CID_DELTREE 	0x4
# define CID_READDIR 	0x5
# define CID_READDIR_CONT 	0x6
# define CID_UNLINK 	0x7
# define CID_CREATE 	0x8
# define CID_OPEN 		0x9
# define CID_CLOSE 		0xa
# define CID_READ 		0xb
# define CID_WRITE 		0xc
# define CID_FTRUNCATE 	0xd
# define CID_STAT 		0xf

/** Variable controls lock debugging */
#define DEBUG_LOCK 0

#if (!DEBUG_LOCK)
#define LOCK(i) pthread_mutex_lock(&(files[i].lock))
#define UNLOCK(i) pthread_mutex_unlock(&(files[i].lock))
#define TRYLOCK(i) pthread_mutex_trylock(&(files[i].lock))
#else
#define LOCK(i) \
tmpi_error(DBG_INFO, "   Before T(%d) lock %d %f\n", pthread_self(), i, getTime()); \
pthread_mutex_lock(&(files[i].lock)); \
tmpi_error(DBG_INFO, "   After  T(%d) lock %d %f\n", pthread_self(), i, getTime())

#define UNLOCK(i) \
tmpi_error(DBG_INFO, "   Before T(%d) unlock %d %f\n", pthread_self(), i, getTime()); \
pthread_mutex_unlock(&(files[i].lock)); \
tmpi_error(DBG_INFO, "   After  T(%d) unlock %d %f\n", pthread_self(), i, getTime())

#endif

/**
 * Translate a Sorrento error code into a standard Unix error code
 * @param errno Sorrento error code
 * @return Unix error code
 */
int 
sore2errno(int err)
{
	static int table[][2] = {
		{DIR_RES_ECREATE, EPERM},      /* no corresponding one */
		{DIR_RES_EEXIST, EEXIST},
		{DIR_RES_ENONEXIST, ENOENT},
		{DIR_RES_ESERVER, EIO},        /* no corresponding one */
		{DIR_RES_ENETWORK, ENETDOWN},  /* no corresponding one */
		{DIR_RES_ETOUT, ETIMEDOUT},
		{DIR_RES_EDATA, ENODATA},      /* no corresponding one */
		{DIR_RES_ECLIENT, EIO},        /* no corresponding one */
		{DIR_RES_EINVALID, EIO},       /* no corresponding one */
		{DIR_RES_EBUFFER, EIO},        /* no corresponding one */
		{DIR_RES_EMEM, EIO},           /* no corresponding one */
		{DIR_RES_EAGAIN, EAGAIN},
		{DIR_RES_EEND, EIO},           /* no corresponding one */
		{DIR_RES_ESELF, EIO},          /* no corresponding one */
		{DIR_RES_EISDIR, EISDIR},
		{DIR_RES_ENOTDIR, ENOTDIR},
		{DIR_RES_MAX_ERROR, EIO},      /* no corresponding one */
		{FILE_RES_EMEM, EIO},          /* no corresponding one */
		{FILE_RES_ESCLIENT, EIO},      /* no corresponding one */
		{FILE_RES_ESPACE, EIO},        /* no corresponding one */
		{FILE_RES_EREAD, EIO},         /* no corresponding one */
		{FILE_RES_EWRITE, EIO},        /* no corresponding one */
		{FILE_RES_NOT_FOUND, ENOENT},
		{FILE_RES_EMETA, EIO},         /* no corresponding one */
		{FILE_RES_EFILE, EIO},         /* no corresponding one */
		{FILE_RES_EDATA, EIO},         /* no corresponding one */
		{FILE_RES_ATTACHMENT, EIO},    /* no corresponding one */
		{FILE_RES_ESERVER, EIO},       /* no corresponding one */
		{FILE_RES_ENETWORK, EIO},      /* no corresponding one */
		{FILE_RES_EIO, EIO},
		{FILE_RES_EEXIST, EEXIST},
		{FILE_RES_ECLIENT, EIO},       /* no corresponding one */
		{0, EPERM}                     /* default value */
	};

	for (int i =0; ; i++) {
		if (table[i][0] == 0) 
			return EPERM;

		if (table[i][0] == err){
			fprintf(stderr,"sore2errno: For Error: %d, Returning table[%d][1], which is: %d\n",err,i,table[i][1]);
			return table[i][1];
		}
	}
	return EPERM;
}



FileTable::FileTable(HostMap *h, APTable *a, ClientPool *c) 
	: hmap(h), apt(a), cp(c)
{
	for (int i = 0; i < FILE_TABLE_SIZE; i++) {
		files[i].file = NULL;
		pthread_mutex_init(&(files[i].lock), NULL);
		files[i].ino = INVALID_INO;
		files[i].valid = false;
	}
}

FileTable::~FileTable()
{
	Client *cli = cp->get_client();

	for (int i = 0; i < FILE_TABLE_SIZE; i++) {
		if (files[i].valid) {
			LOCK(i);
			if (files[i].file != NULL) { 
				/* close the meta object, data segments, 
				 * and update dir_srv entry. 
				 */
				files[i].file->close(apt->get_ap(), cli, hmap);
				apt->put_ap();
				cp->put_client();
				delete files[i].file;
				files[i].file = NULL;
			}
			UNLOCK(i);
		}
		pthread_mutex_destroy(&(files[i].lock));
	}

	cp->put_client();

	/* Just a reference */
	hmap = NULL;
}

/**
 * Add a new SorFile to the table. Multi-thread safe.
 * @param fp SorFile object
 * @param ino inode number of the file
 * @return file descriptor (index in the table) for future access
 */
int
FileTable::put_file(SorFile *fp, long ino)
{
	int ret;

	if (fp == NULL || ino == INVALID_INO)
		return -1;

	for (int i = 0; i < FILE_TABLE_SIZE; i++) {
#if (!DEBUG_LOCK)
		ret = TRYLOCK(i);
#else
		tmpi_error(DBG_INFO, "   Before T(%d) trylock %d\n", pthread_self(), i); 
		ret = pthread_mutex_trylock(&(files[i].lock));
		tmpi_error(DBG_INFO, "   After  T(%d) trylock %d [%d]\n", pthread_self(), i, ret);
#endif

		if (ret == 0 && (!files[i].valid) ) {
			files[i].file = fp;
			files[i].ino = ino;
			files[i].valid = true;
			UNLOCK(i);

			tmpi_error(DBG_INFO, "put_file fd %d, ino %ld\n", i, ino);
			return i;
		}

		if (ret == 0) {
			UNLOCK(i);
		}
    }
	
	/* no available slot left */
    return -1;
}


/**
 * Get a SorFile object by index
 * @param fd file descriptor
 * @return SorFile object, NULL if error
 */
SorFile*
FileTable::get_file(int fd)
{
	if (fd <0 || fd >= FILE_TABLE_SIZE)
		return NULL;

	return files[fd].file;
}

/**
 * Find the file descriptor for a given inode
 * @param ino inode number
 * @return file descriptor, -1 if not found
 */
int
FileTable::get_fd(long ino)
{
	if (ino == INVALID_INO)
		return -1;

	for (int i = 0; i < FILE_TABLE_SIZE; i++) {
		LOCK(i);
		if (files[i].valid && files[i].ino == ino) {
			UNLOCK(i);
			return i;
		}
		UNLOCK(i);
	}
	return -1;
}

/**
 * Close a SorFile object by index
 * @param fd file descriptor
 * @return 0 for success, or error number
 */
int
FileTable::close_file(int fd)
{
	int ret;

	if (fd <0 || fd >= FILE_TABLE_SIZE)
		return EBADF;

	tmpi_error(DBG_INFO, "close_file fd %d, ino %ld\n", fd, files[fd].ino);
	LOCK(fd);
	if (files[fd].file != NULL) {
		ret = files[fd].file->close(apt->get_ap(), cp->get_client(), hmap);
		apt->put_ap();
		cp->put_client();
		delete files[fd].file;
	}
	files[fd].file = NULL;
	files[fd].ino = INVALID_INO;
	files[fd].valid = false;
	UNLOCK(fd);

	if (ret != FILE_RES_OKAY) {
		return sore2errno(ret);
	}
	return 0;
}

/** 
 * Read a number of bytes from a certain offset. Multi-thread safe.
 * @param fd file descriptor
 * @param offset starting offset
 * @param buf buffer to be filled in
 * @param len buffer length
 * @param read_bytes number of bytes read
 * @return status (0 for sucess, or errno)
 */
int 
FileTable::read(int fd, 
				off64_t offset, 
				char *buf, 
				off_seg_t len, 
				off_seg_t & read_bytes)
{
	int ret = 0;

	if (fd <0 || fd >= FILE_TABLE_SIZE || !files[fd].valid)
		return EBADF;

	LOCK(fd);
	try {
		ret = files[fd].file->read(cp->get_client(), hmap, offset, 
								   buf, len, read_bytes);
		cp->put_client();
	}
	catch (int err) { // error in read file
		tmpi_error(DBG_INTERNAL, "read internal error(%d) fd[%d], offset %lld, len %d\n", 
				err, fd, offset, len);
		UNLOCK(fd);
		ret = err;
		close_file(fd);
		goto OUT;
	}
	UNLOCK(fd);

OUT:
	if (ret != FILE_RES_OKAY) {
		ret = sore2errno(ret);
	} else {
		ret = 0;
	}

	return ret;
}

/** 
 * Write a number of bytes from a certain offset. Multi-thread safe.
 * @param fd file descriptor
 * @param offset starting offset
 * @param buf buffer to be filled in
 * @param len buffer length
 * @param read_bytes number of bytes read
 * @return status (0 for sucess, or errno)
 */
int 
FileTable::write(int fd, 
				 off64_t offset, 
				 char *buf, 
				 off_seg_t len, 
				 off_seg_t & written_bytes)
{	
	int ret = 0;
    
    LOCK(fd);
	if (fd <0 || fd >= FILE_TABLE_SIZE || !files[fd].valid){
        UNLOCK(fd);
		return EBADF;
    }

	tmpi_error(DBG_INFO, "write fd[%d], offset %lld, len %d\n", fd, offset, len);

	ret = files[fd].file->write(cp->get_client(), hmap, offset, 
								buf, len, written_bytes);
	cp->put_client();
	if (ret != FILE_RES_OKAY) {
		ret = sore2errno(ret);
	} else {
		ret = 0;
	}
	UNLOCK(fd);

	return ret;
}




/** Allocat a struct CP entry */
struct ClientPool::CP*
ClientPool::allocate_CP()
{
	struct ClientPool::CP *p = new (struct ClientPool::CP);

	if (p == NULL)
		throw std::string("ClientPool: not enough memory.");

	pthread_mutex_init( &(p->lock), NULL);
	p->thread = THREAD_UNUSED;
	p->inuse = false;
	p->cli = new Client();

	return p;
}

/** Constructor 
 * @throw std::string
 */
ClientPool::ClientPool()
{
	struct CP *p;

	for (int i = 0; i < CLIENT_POOL_SIZE; i++) {
		p = allocate_CP();
		cp.push_back(p);
	}
}

/** Destructor */
ClientPool::~ClientPool()
{
	for (unsigned i = 0; i < cp.size(); i++) {
		struct CP *p = cp[i];
		
		delete p->cli;
		delete p;
	}
}

/**
 * Get an access point and holding the lock. Must call put_client() after
 * using the AP. 
 * @return access point (holding lock)
 * @throw std::string if allocation failed
 */
Client*
ClientPool::get_client()
{
	int ret;
	struct CP *p;

	/* search in available entries if there is not in use */
	for (unsigned i = 0; i < cp.size(); i++) {
		p = cp[i];
		if (p->inuse)
			continue;

		ret = pthread_mutex_trylock( &(p->lock) );
		if (p->inuse && ret == 0 ) {
			pthread_mutex_unlock( &(p->lock) );
		} else if ( !p->inuse && ret == 0 ) {
			p->inuse = true;
			p->thread = pthread_self();

			return p->cli;
		}
	}

	p = allocate_CP();
	pthread_mutex_lock( &(p->lock) );
	p->inuse = true;
	p->thread = pthread_self();
	cp.push_back(p);

	return p->cli;
}


/** Release an AP by unlocking */
void
ClientPool::put_client()
{
	pthread_t self = pthread_self();

	for (unsigned i = 0; i < cp.size(); i++) {
		struct CP *p = cp[i];
		if (p->inuse && p->thread == self) {
			p->inuse = false;
			p->thread = THREAD_UNUSED;
			pthread_mutex_unlock( &(p->lock) );

			return;
		}
	}

	/* TODO: remove entries from pool if the size becomes big */

	tmpi_error(DBG_WARN1, "%s: T(%d) not using any AP\n", __FUNCTION__);
}







/**
 * Constructor
 * @param sp SorFS object
 */
APTable::APTable(SorFS *sp)
{
	if (!sp) 
		throw std::string("APTable::APTable: NULL argument.");

	for (int i = 0; i < AP_TABLE_SIZE; i++) {
		aps[i].ap = sp->new_dir_cli_ap();
		if (aps[i].ap == NULL)
			throw std::string("APTable::APTable: can't get a new access point.");

		pthread_mutex_init( &(aps[i].lock), NULL );
		aps[i].thread = THREAD_UNUSED;
		aps[i].inuse = false;
	}
}

/** Destructor */
APTable::~APTable()
{
	for (int i = 0; i < AP_TABLE_SIZE; i++) {
		SorFS::free_dir_cli_ap(aps[i].ap);
		aps[i].ap = NULL;
	}
}

/**
 * Get an access point and holding the lock. Must call put_ap() after
 * using the AP. 
 *
 * TODO: adding a blocking call that never fails.
 *
 * @return access point (holding lock), or NULL if no 
 *         available (not holding any lock)
 */
DIR_Cli_AP*
APTable::get_ap()
{
	int ret;
	//pthread_t self = pthread_self();

	for (int i = 0; i < AP_TABLE_SIZE; i++) {
		if (aps[i].inuse) {
			//if (aps[i].thread == self)
			//return aps[i].ap; /* same thread tries again */

			continue;
		}

		ret = pthread_mutex_trylock( &(aps[i].lock));
		if (aps[i].inuse && ret == 0 ) {
			pthread_mutex_unlock( &(aps[i].lock) );
		} else if ( !aps[i].inuse && ret == 0 ) {
			aps[i].inuse = true;
			aps[i].thread = pthread_self();

			//tmpi_error(DBG_INFO, "T[%d] get AP[%d]\n", self, i);
			return aps[i].ap;
		}
	}

	tmpi_error(DBG_WARN1, "%s: no available access point\n", __FUNCTION__);
	return NULL;
}

/** Release an AP by unlocking */
void
APTable::put_ap()
{
	pthread_t self = pthread_self();

	for (int i = 0; i < AP_TABLE_SIZE; i++) {
		if (aps[i].inuse && aps[i].thread == self) {
			aps[i].inuse = false;
			aps[i].thread = THREAD_UNUSED;
			pthread_mutex_unlock( &(aps[i].lock) );

			//tmpi_error(DBG_INFO, "T[%d] put AP[%d]\n", self, i);
			return;
		}
	}

	tmpi_error(DBG_WARN1, "%s: T(%d) not using any AP\n", __FUNCTION__);
}




/**
 * Constructor.
 * @param ap Directory Service Access Point.
 */
SClientStub::SClientStub(SorFS *sp) :
	hmap(NULL),
	//cli(NULL),
	errnum(DIR_RES_OKAY),
	callid(CID_NULL)
{
	if (!sp) 
		throw std::string("SClientStub::SClientStub: NULL superblock argument.");
	hmap=sp->get_hmap();
	cp = new ClientPool();
	apt = new APTable(sp);
	ft = new FileTable(hmap, apt, cp);
}

/** Destructor. */
SClientStub::~SClientStub()
{
	delete ft; /* has to be called before ap and cli */
	//delete cli;
	delete cp;
}


/**
 * Make a new directory.  Namespace operation.
 * @param path the directory to be created.
 * @return status (0: success, or errno).
 */
int
SClientStub::mkdir(const char *path)
{
	DIR_Entry entry;
	int ret = cwd.mkdir(apt->get_ap(), path, &entry);
	apt->put_ap();

	if (ret == DIR_RES_OKAY) {
		return 0;
	}

	return sore2errno(ret);
}

/**
 * Remove an existing directory.  Namespace operation.
 * @param path the directory to be removed.
 * @return status (0: success, or errno).
 */
int 
SClientStub::rmdir(const char *path)
{
	int ret=cwd.rmdir(apt->get_ap(), path);
	apt->put_ap();

	if (ret == DIR_RES_OKAY) {
		return 0;
	}

	return sore2errno(ret);
}

/**
 * Read the content of a directory. Namespace operation.
 * @param path the directory to be read.
 * @param list the list of entries in the directory.
 * @param buf_len the memory consumption constraint. 
 * @param cookie the cookie used for subsequent readdir_cont() calls.
 * @return status (true: success).
 */
int 
SClientStub::readdir(const char *path, 
					 DIR_List &list, 
					 int buf_len, 
					 DIR_Entry *cookie)
{
	DIR_Cli_AP *cli_ap = apt->get_ap();
	int ret = cwd.lookup(cli_ap, path, cookie);

	if (ret != DIR_RES_OKAY) {
		goto _exit;
	}

	if (!cookie->is_dir(cli_ap)) {
		ret = DIR_RES_ENOTDIR;
		goto _exit;
	}

	ret=cookie->readdir(cli_ap, list, buf_len);
	if ( (ret != DIR_RES_OKAY) ) {
		goto _exit;
	}

	apt->put_ap();
	return 0;

_exit:
	apt->put_ap();
	return sore2errno(ret);
}

/**
 * Check whether we need to make another readdir_cont() call. 
 * @param cookie the cookie from previous call of readdir or readdir_cont.
 * @return status (true: yes).
 */
bool SClientStub::is_eod(const DIR_Entry *cookie) const
{
	return cookie->is_eod();
}

/** 
 * Continue the previous reading.
 * @param cookie the cookie used for subsequent readdir_cont() calls.
 * @param list the list of entries in the directory.
 * @param buf_len the memory consumption constraint. 
 * @return status (0: success, or errno).
 */
int 
SClientStub::readdir_cont(DIR_Entry *cookie, 
						  DIR_List &list, 
						  int buf_len)
{
	int ret = cookie->readdir_cont(apt->get_ap(), list, buf_len);
	apt->put_ap();

	if ( (ret != DIR_RES_OKAY) && (ret != DIR_RES_EEND) ) {
		return sore2errno(ret);
	}

	return 0;
}

/**
 * Remove an existing file.  Namespace and storage operation.
 * @param path to the file to be removed.
 * @return status (0: success, or errno).
 */
int
SClientStub::unlink(const char *path)
{
	DIR_Cli_AP *cli_ap = apt->get_ap();

	// retrieve the entry first.
	DIR_Entry entry;
	int ret=cwd.lookup(cli_ap, path, &entry);
	if (ret != DIR_RES_OKAY) {
		goto _exit;
	}

	// store away the guid
	// CHECK: should I check return value of get_guid??
	guid_t guid;
	entry.get_guid(cli_ap, guid);
	// ObjID oid=guid;

	// then remove the directory entry (possibly with the guard 
	// of entry later).
	ret = cwd.remove(cli_ap, path);
	if (ret != DIR_RES_OKAY) {
		goto _exit;
	}

	// finally, remove the meta object and the data segments.
	try {
		ret = SorFile::unlink(cp->get_client(), hmap, guid);
	}
	catch (int err) {
		ret = err;
	}
	cp->put_client();
	if (ret != FILE_RES_OKAY) {
		goto _exit;
	}

	apt->put_ap();
	return 0;

_exit:
	apt->put_ap();
	return sore2errno(ret);
}

/**
 * Create a new file.
 * @param path to of the file to be created.
 * @return file handle (NULL if failed.)
 */
SorFile *
SClientStub::create(const char *path)
{
	SorFile *sf=NULL;
	DIR_Entry entry;
	DIR_Cli_AP *cli_ap = apt->get_ap();

	DIR_Param param;
	param.reset();
	param.options = DIR_OPTION_FILE;

	int ret = cwd.create(cli_ap, path, &param, &entry);
	if (ret != DIR_RES_OKAY) {
		goto _exit;
	}

	try {
		/** create a new file, create the meta object */
		entry.refresh_info(cli_ap);
		sf = new SorFile(cp->get_client(), hmap, cli_ap, &entry, 0, true);
		cp->put_client();

		apt->put_ap();
		return sf;
	}
	catch (int error) {
		ret=error;
		goto _exit;
	}

_exit:
	errnum=ret;
	callid=CID_CREATE;

	apt->put_ap();
	return NULL;
}

/**
 * Open an existing file. The file is put in internal table so that next
 * time we can use its inode number for access.
 * @param path to the file to be opened.
 * @param flags (eg, read/write modes, ignored).
 * @param ino this file's inode number
 * @return status (0: success, or errno).
 */
int
SClientStub::open(const char *path, int flags, long ino)
{
	DIR_Entry entry;
	SorFile *sf=NULL;
	int fd;
	DIR_Cli_AP *cli_ap = apt->get_ap();

	int ret = cwd.lookup(cli_ap, path, &entry);
	if (ret != DIR_RES_OKAY) {
		fprintf(stderr,"Could not lookup %s\n",path);
		if(!this->create(path)){
			fprintf(stderr,"Could not create %s\n",path);
			goto _exit;
		}else{
			ret = cwd.lookup(cli_ap, path, &entry);
			if (ret != DIR_RES_OKAY) {
				fprintf(stderr,"Could not open %s\n",path);
				goto _exit;				
			}	
		}

	} else {
		if (entry.is_dir(cli_ap)) {
			ret=DIR_RES_EISDIR;
			goto _exit;
		}
	}

	/* check if aleady opened */
	fd = ft->get_fd(ino);
	if (fd >= 0)
		return 0;

	/* try to open the file */
	try {
		entry.refresh_info(cli_ap);
		sf = new SorFile(cp->get_client(), hmap, cli_ap, &entry, 0); 
		cp->put_client();

		fd = ft->put_file(sf, ino);
		if (fd < 0) {
			tmpi_error(DBG_INTERNAL, "File table is full\n");
			sf->close(cli_ap, cp->get_client(), hmap);
			cp->put_client();
			delete sf;

			apt->put_ap();
			return EMFILE;
		}
		apt->put_ap();
		return 0;
	} 
	catch (int error) {
		ret=error;
		goto _exit;
	}

_exit:
	apt->put_ap();
	return sore2errno(ret);
}

/**
 * Close an opened file. 
 * @param sf Sorrento file handle.
 * @return status (0: success, or errno.)
 */
int 
SClientStub::close(SorFile *sf)
{
	/** 
	 * close the meta object, data segments, and update dir_srv entry. 
	 */
	int ret=sf->close(apt->get_ap(), cp->get_client(), hmap);
	apt->put_ap();
	cp->put_client();

	if (ret != FILE_RES_OKAY) {
		return sore2errno(ret);
	}

	delete sf;
	return 0;
}

/**
 * Close an opened file after kernel reference count becomes 0. 
 * @param ino Sorrento file inode number.
 * @return status (0: success, or errno.)
 */
int 
SClientStub::release(long ino)
{
	int fd;

	fd = ft->get_fd(ino);
	if (fd < 0)
		return EBADF;

	return ft->close_file(fd);
}
/** 
 * Read a number of bytes from a certain offset. 
 * @param sf Sorrento file handle.
 * @param offset the offset of the starting range.
 * @param buf the data buffer to be filled.
 * @param len the length of the buffer.
 * @param read_bytes the number of bytes read.
 * @return status (0: success; or errno). 
 */
int 
SClientStub::read(long ino, 
				  off64_t offset, 
				  char *buf, 
				  off_seg_t len, 
				  off_seg_t & read_bytes)
{
	int fd, ret;

	fd = ft->get_fd(ino);
	if (fd < 0)
		return EBADF;

	ret = ft->read(fd, offset, buf, len, read_bytes);
	if (ret != FILE_RES_OKAY) {
		return sore2errno(ret);
	}

	return 0;
}

/** 
 * Write a number of bytes to a certain offset.
 * @param ino Sorrento file's inode number
 * @param offset the offset of the starting range.
 * @param buf the data buffer to be filled.
 * @param len the length of the buffer.
 * @param written_bytes the number of bytes being written.
 * @return status (0: success; or errno).
 */
int 
SClientStub::write(long ino, 
				   off64_t offset, 
				   char *buf, 
				   off_seg_t len, 
				   off_seg_t & written_bytes)
{
	int fd, ret;

	fd = ft->get_fd(ino);
	if (fd < 0)
		return EBADF;

	ret = ft->write(fd, offset, buf, len, written_bytes);
	if (ret != FILE_RES_OKAY) {
		return sore2errno(ret);
	}

	return 0;
}

/** 
 * Change the size of an opened file.
 * @param sf Sorrento file handle.
 * @param new_size the new size of the file.
 * @return status (0: success, or errno).
 */
int 
SClientStub::ftruncate(SorFile *sf, off64_t new_size)
{
	int ret = sf->truncate(cp->get_client(), hmap, new_size);
	cp->put_client();

	if (ret != FILE_RES_OKAY) {
		return sore2errno(ret);
	}

	return 0;
}

/**
 * Change the size of a file.
 * @param path Sorrento file name.
 * @param new_size the new size of the file.
 * @return status (0: success, or errno).
 */
int 
SClientStub::truncate(const char *path, off64_t new_size)
{
	DIR_Entry entry;
	SorFile *sf = NULL;
	DIR_Cli_AP *cli_ap = apt->get_ap();

	int ret = cwd.lookup(cli_ap, path, &entry);
	if (ret != DIR_RES_OKAY) {
		goto _exit;
	} else {
		if (entry.is_dir(cli_ap)) {
			ret=DIR_RES_EISDIR;
			goto _exit;
		}
	}

	/* try to open the file */
	try {
		entry.refresh_info(cli_ap);
		sf = new SorFile(cp->get_client(), hmap, cli_ap, &entry, 0);
		cp->put_client();
	} 
	catch (int error) {
		ret = error;
		goto _exit;
	}

	if ( (ret = this->ftruncate(sf, new_size)) != 0) {
		this->close(sf);
		goto _exit;
	}

	if ( (ret = this->close(sf)) != 0)
		goto _exit;

	apt->put_ap();
	return 0;

 _exit:
	apt->put_ap();
	return sore2errno(ret);
}

/** 
 * Get attribute of a certain file or directory based on name. This is 
 * in fact a namespace operation.
 * @param path the path to the file.
 * @param param the attributes.
 * @return status (0: success, errno otherwise).
 */
int 
SClientStub::stat(const char *path, DIR_Param *param)
{
	DIR_Entry entry;
	DIR_Cli_AP *cli_ap = apt->get_ap();

	int ret=cwd.lookup(cli_ap, path, &entry);
	if (ret != DIR_RES_OKAY) {
		goto _exit;
	}

	ret=entry.getattr(cli_ap, param);
	if (ret != DIR_RES_OKAY) {
		goto _exit;
	}

	apt->put_ap();
	return 0;

_exit:
	apt->put_ap();
	return sore2errno(ret);
}

std::string SClientStub::str_error()
{
	if (callid == CID_NULL) {
		return std::string("No error.");
	}

	std::string cmd, reason;

	switch (this->callid) {
		case CID_CHDIR: 		cmd="chdir"; 	break;
		case CID_MKDIR: 		cmd="mkdir"; 	break;
		case CID_RMDIR: 		cmd="rmdir"; 	break;
		case CID_DELTREE: 		cmd="deltree"; 	break;
		case CID_READDIR: 		cmd="readdir"; 	break;
		case CID_READDIR_CONT: 	cmd="readdir_cont"; break;
		case CID_UNLINK: 		cmd="unlink"; 	break;
		case CID_CREATE: 		cmd="create"; 	break;
		case CID_OPEN: 			cmd="open"; 	break;
		case CID_CLOSE: 		cmd="close"; 	break;
		case CID_READ: 			cmd="read"; 	break;
		case CID_WRITE: 		cmd="write"; 	break;
		case CID_FTRUNCATE: 	cmd="ftruncate"; break;
		case CID_STAT: 			cmd="stat"; 	break;
		default: 				cmd="<unknown>"; break;
	}

	switch (this->errnum) {
		case DIR_RES_OKAY:
			reason="Okay"; break;
		case DIR_RES_ENONEXIST: 
			if ((callid==CID_MKDIR)||(callid==CID_CREATE))
				reason="Unable to create file or directory now, try again."; 
			else
				reason="File or directory does not exist."; 
			break;
		case DIR_RES_EEXIST: 
			if (callid==CID_RMDIR)
				reason="Directory not empty."; 
			else
				reason="File or directory already exists."; 
			break;
		case DIR_RES_ESELF: 
			reason="Current directory no longer exists."; break;
		case DIR_RES_EINVALID: 
			reason="Invalid argument."; break;
		case DIR_RES_EISDIR: 
			reason="Target is a directory."; break;
		case DIR_RES_ENOTDIR: 
			reason="Target is a not a directory."; break;
		case DIR_RES_EBUFFER: 
		case DIR_RES_EMEM: 
		case FILE_RES_EMEM :
			reason="Out of memory."; break;
		case DIR_RES_ESERVER: 
		case FILE_RES_ESERVER :
			reason="Server error."; break;
		case DIR_RES_ETOUT: 
			reason="Request timed out. Maybe network error."; break;
		case DIR_RES_ENETWORK: 
		case FILE_RES_ENETWORK :
			reason="Network error."; break;
		case DIR_RES_ECLIENT: 
		case FILE_RES_ECLIENT :
			reason="Client error."; break;
		case FILE_RES_ESPACE :
			reason="Server out of space."; break;
		case FILE_RES_EREAD :
			reason="Failed to read."; break;
		case FILE_RES_EWRITE :
			reason="Failed to write."; break;
		case FILE_RES_NOT_FOUND :
			reason="File not found."; break;
		case FILE_RES_EMETA :
			reason="Metadata corrupted."; break;
		case FILE_RES_EFILE :
			reason="File object corrupted."; break;
		case FILE_RES_EDATA :
			reason="Object data info corrupted."; break;
		case FILE_RES_ATTACHMENT :
			reason="Attachment corrupted."; break;
		case FILE_RES_EIO :
			reason="Server I/O error."; break;
		case FILE_RES_EEXIST :
			reason="File or object already exists."; break;
		case FILE_RES_ESCLIENT :
			reason="Subscriber error."; break;
		default:
			{
				char buf[32];
				sprintf(buf, "%d", errnum);
				reason="Internal error. Code = ";
				reason= reason+buf;
			}

			break;
	}

	return cmd+": "+reason;
}

void SClientStub::stop_service(void)
{
	const VHostLoad *hostloads=(hmap)?hmap->get_host_loads():NULL;
	if (hostloads) {
		Client *cli = cp->get_client();
		for (unsigned int i=0; i<hostloads->size(); i++) {
			cli->stop_server((*hostloads)[i].addr);
		}
		cp->put_client();
	}
	if (hmap)
		hmap->release_host_loads(hostloads); // Important: release_host_loads even if it is NULL.

	HostAddr dir_addr;
	if (hmap->get_dir_addr(dir_addr)) {
		Udp_cli udp_cli(NULL, SOCK_PORT_ANY, 1000);
		udp_cli.stop_srv(dir_addr.ip, dir_addr.port, "", 1, 10);
	}
}

void SClientStub::dump_view(void)
{
	HostAddr dir_addr;
	if (hmap->get_dir_addr(dir_addr)) {
		printf("Directory server: %s.\n", dir_addr.to_string().c_str());
	}

	const VHostLoad *hostloads=(hmap)?hmap->get_host_loads():NULL;
	if (hostloads) {
		for (unsigned int i=0; i<hostloads->size(); i++) {
			printf(" %-16s %-4d %04.1f   %04.1f   %04.1f   %04.1f   %04.1f   %04.1f  <", 
					(*hostloads)[i].addr.ip,
					(*hostloads)[i].vnodes,
					(*hostloads)[i].cpu_comp*100,
					(*hostloads)[i].cpu_ewma_comp*100,
					(*hostloads)[i].cpu_io*100,
					(*hostloads)[i].cpu_ewma_io*100,
					(*hostloads)[i].mem_used*100,
					(*hostloads)[i].swap_used*100
			);

			for (int j=0; j<(*hostloads)[i].ndisks; j++) {
				if (j>0)
					printf(",");
				printf("%d", (*hostloads)[i].disks[j].bfree/1024); 
			}
			printf(">\n");
		}
	}
	if (hmap)
		hmap->release_host_loads(hostloads); // Important: release_host_loads even if it is NULL.
}
