// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: fsd.cc,v $ - file size distribution analyzer.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2002/10/08 07:32:51 $
 * Version :  $Revision: 1.6 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 
// -*- mode: c++; -*-

# include <sys/types.h>
# include <sys/stat.h>
# include <unistd.h>
# include <stdio.h>
# include <stdlib.h>
# ifdef __linux__
# include <getopt.h>
# endif
# include <dirent.h>

# include <queue>
# include <vector>

using namespace std;

# include "mystatfs.h"
# include "heap.h"

/**
 * A program that traverse a set of
 * file systems specified in the command 
 * line and calculates the aggregated
 * file size distribution.
 * Syntax: fsd [options] <dir-list>
 */

static char *prog=NULL;

static void do_help()
{
	printf("\
File size distribution calculation.\n\
Usage: %s [options] <dir-list>.\n\
   -o <output>     Output file (default to stdout).\n\
   -x <num>        Min # of data points for object percentage.\n\
   -y <num>        Min # of data points for space percentage.\n\
   -m              Output Matlab array (default).\n\
   -M              No Matlab output.\n\
   -t              Output in tabular format (default).\n\
   -T              No tabular output.\n\
   -i              Show inaccessible directories (default).\n\
   -I              Don't show inaccessible directories.\n\
   -s              Show symbolic links (default).\n\
   -S              Don't show symbolic links.\n\
   -h              Help, this screen.\n\
\n\
This program check the aggregated file size distribution\n\
on a list of file systems. Unaccessible directories are\n\
ignored.\n\
", prog);
	exit(0);
}

struct FID {
	dev_t id_dev;
	ino_t id_ino;

	FID(dev_t d, ino_t i) : id_dev(d), id_ino(i) { }
	FID() { }

	int cmp(FID &fid) {
		if (id_dev!=fid.id_dev)
			return (id_dev>fid.id_dev)?1:-1;
		if (id_ino!=fid.id_ino)
			return (id_ino>fid.id_ino)?1:-1;

		return 0;
	}
};

struct SHARED {
	unsigned long long total_bytes;
	unsigned int total_dir, total_sym, total_iacc, total_other;
	double total_obj;
	off_t *heap_keys;
	FID *heap_vals;
	HeapMax<off_t, FID> *heap;
	vector<char *>v_iacc;
	vector<char *>v_sym;

	SHARED(unsigned int heap_size) : total_bytes(0), 
		total_dir(0), total_sym(0), total_iacc(0), 
		total_other(0), total_obj(0.0), 
		heap_keys(NULL), heap_vals(NULL), heap(NULL) 
	{
		heap_keys=(off_t *)malloc(sizeof(off_t)*heap_size);
		heap_vals=(FID *)malloc(sizeof(FID)*heap_size);
		if ((heap_keys)&&(heap_vals)) {
			heap=new HeapMax<off_t, FID>(heap_size, heap_keys, heap_vals);
		}
		else 
			throw -1;
	}

	~SHARED() {
		if (heap_keys)
			free(heap_keys);
		if (heap_vals)
			free(heap_vals);
		if (heap)
			delete heap;
		for (unsigned int i=0; i<v_iacc.size(); i++) {
			free(v_iacc[i]);
		}
		for (unsigned int i=0; i<v_sym.size(); i++) {
			free(v_sym[i]);
		}
	}

	void add_file(struct stat &st) {
		//total_bytes+=(st.st_size/st.st_nlink);
		//total_obj+=1.0/st.st_nlink;
		FID fid(st.st_dev, st.st_ino);
		if (heap->add(st.st_size, fid)) {
			total_bytes+=st.st_size;
			total_obj+=1;
		}
	}
};

static void traverse(char *f, SHARED &state, unsigned int total_node, dev_t device)
{
	static unsigned int cnt=0;

	cnt++;
	if (cnt%100==0) {
		double progress=(state.total_obj+
					state.total_dir+state.total_sym+
					state.total_iacc+state.total_other)/total_node;
		if (progress>1)
			progress=1;
		printf("\r%.3f%% done.       ", progress*100);
		fflush(stdout);
	}

	struct stat st;
	if (lstat(f, &st)<0) {
		fprintf(stderr, "lstat failed: %s.\n", f);
		state.total_iacc++;
		state.v_iacc.push_back(strdup(f));
		return;
	}

	mode_t m=st.st_mode;

	if (S_ISREG(m)) {
		state.add_file(st);
	}
	else if ((S_ISDIR(m))&&(st.st_dev==device)) {
		DIR *dir=opendir(f);
		if (!dir) {
			state.total_iacc++;
			state.v_iacc.push_back(strdup(f));
			return;
		}
		else {
			state.total_dir++;
		}

		struct dirent *de=NULL;
		queue<char *> dirlist;
		while ((de=readdir(dir))) {
			if ((de->d_name[0]=='.')&&(de->d_name[1]=='\0'))
				continue;

			if ((de->d_name[0]=='.')&&(de->d_name[1]=='.')&&(de->d_name[2]=='\0'))
				continue;

			char *d_name=strdup(de->d_name);
			if (d_name) {
				dirlist.push(d_name);
			}
		}
		closedir(dir);

		char ff[4096];
		while (!dirlist.empty()) {
			char *d_name=dirlist.front();
			dirlist.pop();
			snprintf(ff, 4096, "%s/%s", f, d_name);
			free(d_name);
			traverse(ff, state, total_node, device);
		}
	}
	else {
		if (S_ISDIR(m)) {
			state.total_dir++;
		}
		else if (S_ISLNK(m)) {
			state.v_sym.push_back(strdup(f));
			char real_f[4096];
			if (realpath(f, real_f)==NULL) {
				strcpy(real_f, "<unresolved_link>");
				state.total_iacc++;
			}
			else {
				state.total_sym++;
			}
			state.v_sym.push_back(strdup(real_f));
		}
		else {
			state.total_other++;
		}
	}

	return;
}

struct Point {
	double pcent_obj, pcent_spc;
	off_t threshold;
	Point(double x, double y, off_t t) : pcent_obj(x), pcent_spc(y), threshold(t) {}
};

int main(int argc, char *argv[])
{
	prog=argv[0];

	char *output=NULL;
	unsigned int gran_x=1000, gran_y=1000;
	bool tabular_display=true;
	bool matlab_display=true;
	bool show_sym=true;
	bool show_iacc=true;

	int c;
	while ((c=getopt(argc, argv, "o:x:y:mMtTiIsSh"))!=EOF) {
		switch (c) {
			case 'o':
				output=optarg;
				break;
			case 'x':
				gran_x=(unsigned)atoi(optarg);
				break;
			case 'y':
				gran_y=(unsigned)atoi(optarg);
				break;
			case 'm':
				matlab_display=true;
				break;
			case 'M':
				matlab_display=false;
				break;
			case 't':
				tabular_display=true;
				break;
			case 'T':
				tabular_display=false;
				break;
			case 'i':
				show_iacc=true;
				break;
			case 'I':
				show_iacc=false;
				break;
			case 's':
				show_sym=true;
				break;
			case 'S':
				show_sym=false;
				break;
			case 'h':
				do_help();
				break;
			default:
				do_help();
				break;
		}
	}

	if ((!matlab_display)&&(!tabular_display)) {
		fprintf(stderr, "No output format specified.\n");
		do_help();
	}

	if ((argc-optind)<=0) {
		fprintf(stderr, "No file system specified.\n");
		do_help();
	}

	unsigned int used_inodes=0;
	for (int i=optind; i<argc; i++) {
		char real_p[4096];
		if (realpath(argv[i], real_p)==NULL) {
			fprintf(stderr, "realpath failed: %s (ignored).\n", argv[i]);
			continue;
		}
		struct fs_info mst;
		if (mystatfs(real_p, &mst)<0) {
			fprintf(stderr, "statfs failed: %s (ignored).\n", real_p);
			continue;
		}
		used_inodes+=(unsigned int)(mst.f_files-mst.f_ffree);
	}

	if (used_inodes==0) {
		fprintf(stderr, "No files found.\n");
		return -1;
	}

	SHARED state(used_inodes*2);

	for (int i=optind; i<argc; i++) {
		char real_p[4096];
		if (realpath(argv[i], real_p)==NULL) {
			continue;
		}
		struct stat st;
		if (stat(real_p, &st)<0) {
			continue;;
		}
	
		traverse(real_p, state, used_inodes, st.st_dev);
	}

	printf("\r%.3f%% done.      \n", 100.0);

	if (state.total_obj>=(unsigned int)(used_inodes*1.9)) {
		fprintf(stderr, "Potential file system corruption: total files(%u)>used inode(%u).\n", 
				(unsigned int)state.total_obj, used_inodes);
	}

	vector<Point> data;

	unsigned int frag_obj=1, frag_size=1;
	off_t key;
	FID val;
	unsigned long long agg_size=0;
	unsigned int agg_obj=0;
	// FID last_fid(0, 0);

	while (state.heap->removeMax(key, val)) {
		// if (val.cmp(last_fid)==0)
			// continue;
		// last_fid=val;
		agg_size+=key;
		agg_obj++;
		bool add_point=false;
		if (agg_size>=(frag_size*state.total_bytes/gran_y)) {
			add_point=true;
			frag_size=(unsigned int)(agg_size*gran_y/state.total_bytes)+1;
		}
		if (agg_obj>=(frag_obj*state.total_obj/gran_x)) {
			add_point=true;
			frag_obj=(agg_obj*gran_x/(int)(state.total_obj))+1;
		}
		if (add_point) {
			data.push_back(Point(
				(double)agg_obj/state.total_obj,
				(double)agg_size/state.total_bytes,
				key));
		}

		if ((frag_obj>=gran_x)||(frag_size>=gran_y))
			break;
	}

	FILE *fp_out=NULL;
	if ((output==NULL)||(strcmp(output, "-")==0)) {
		fp_out=stdout;
	}
	else {
		fp_out=fopen(output, "w");
		if (!fp_out) {
			fprintf(stderr, "Cannot write to %s, use stdout instead.\n", output);
			fp_out=stdout;
		}
	}

	fprintf(fp_out, "total_bytes=%lld;\ntotal_obj=%u;\ntotal_dir=%u;\n"
			"total_sym=%u;\ntotal_iacc=%u;\ntotal_others=%u;\n", 
			state.total_bytes, (unsigned int)(state.total_obj), state.total_dir, 
			state.total_sym, state.total_iacc, state.total_other);
	if (show_iacc) {
		for (unsigned int i=0; i<state.v_iacc.size(); i++) {
			fprintf(fp_out, "%% [%d] %s\n", i, state.v_iacc[i]);
		}
	}
	if (show_sym) {
		for (unsigned int i=0; i<state.v_sym.size()/2; i++) {
			fprintf(fp_out, "%% [%d] %s -> %s \n", i, state.v_sym[i*2], state.v_sym[i*2+1]);
		}
	}
	if (data.size()>0) {
		if (tabular_display) {
			fprintf(fp_out, "%%\t%%obj\t%%space\tsize(bytes)\n");
			for (unsigned int i=0; i<data.size(); i++) {
				fprintf(fp_out, "%%\t%.3f\t%.3f\t%lu\n", data[i].pcent_obj*100, data[i].pcent_spc*100, data[i].threshold);
			}
		}
		if (matlab_display) {
			fprintf(fp_out, "pcent_obj=[%.3f", data[0].pcent_obj*100);
			for (unsigned int i=1; i<data.size(); i++) {
				fprintf(fp_out, ", %.3f", data[i].pcent_obj*100);
				if (i%10==0) {
					fprintf(fp_out, " ...\n\t");
				}
			}
			// fprintf(fp_out, ", 100.0];\n");
			fprintf(fp_out, "];\n");

			fprintf(fp_out, "pcent_spc=[%.3f", data[0].pcent_spc*100);
			for (unsigned int i=1; i<data.size(); i++) {
				fprintf(fp_out, ", %.3f", data[i].pcent_spc*100);
				if (i%10==0) {
					fprintf(fp_out, " ...\n\t");
				}
			}
			// fprintf(fp_out, ", 100.0];\n");
			fprintf(fp_out, "];\n");

			fprintf(fp_out, "threshold=[%lu", data[0].threshold);
			for (unsigned int i=1; i<data.size(); i++) {
				fprintf(fp_out, ", %lu", data[i].threshold);
				if (i%10==0) {
					fprintf(fp_out, " ...\n\t");
				}
			}
			// fprintf(fp_out, ", 0];\n");
			fprintf(fp_out, "];\n");
		}
	}
	else {
		fprintf(stderr, "No data points collected.\n");
	}


	if (fp_out!=stdout)
		fclose(fp_out);

	return 0;
}
