// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2000 Teoma Technologies, Inc.
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: heap.h,v $ - Heap template.
 * Authors :  Wei Wang (ww@teoma.com), Teoma Technologies, Inc.
 *            Hong Tang, htang@cs.ucsb.edu,
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/08/14 23:50:22 $
 * Version :  $Revision: 1.4 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

#ifndef H_HEAP
#define H_HEAP

/*[]-----------------------------------------[]
  |  System Includes                          |
  []-----------------------------------------[]*/



/*[]-----------------------------------------[]
  |  Project Includes                         |
  []-----------------------------------------[]*/



/*[]-----------------------------------------[]
  |  Local Includes                           |
  []-----------------------------------------[]*/



/*[]-----------------------------------------[]
  |  Forward References                       |
  []-----------------------------------------[]*/

/** swap the values of two parameters.
    @param x 1st variable
    @param y 2nd variable */
template<typename T>
void Swap(T& x, T& y)
{
    T z = x;   x = y;   y = z;
}

/** HeapMin is a wrap for Heap algorithm. It uses external buffers
    to keep data. The root of the Heap will always be the smallest
    value. The HeapMin means a Heap that has the following property:
    each entity within Heap has key bigger than that of its parent
    node. */
template<typename Tkey, typename Tval>
class HeapMin {

    public:
        /**@name public methods */
        //@{
        /** the constructor
            @param size the size of buffer keys/values
            @param keys buffer for keys
            @param values buffer for values */
        HeapMin(int size, Tkey* keys, Tval* values)
        {
            size_ = size;
            keys_ = keys;
            values_ = values;
            cnt_ = 0;
        }
        /// the destructor
        ~HeapMin() {}
        /** check if the heap is full.
            @return true if the heap is full, false otherwise. */
        bool isFull() { return (cnt_ >= size_); }
        /** check if the heap is empty.
            @return true if the heap is empty, false otherwise. */
        bool isEmpty() { return ( cnt_ <= 0); }
        /** check how many entities are in Heap.
            @return the \# of entities in Heap. */
        int  howMany() { return cnt_; }
        /** clean all entities within heap */
        void clean() { cnt_ = 0; }
        /** add one entity into heap.
            @param key key
            @param value value corresponding to this key
            @return true if heap is not full and key/value has been added
            into heap, false otherwise and the smallest entity to be
            dropped will be assigned to parameters key and value when
            return. */
        bool add(Tkey& key, Tval& value)
        {
            if (cnt_ >= size_)   // heap is full
            {
                if (key > keys_[0])
                {
                    Swap(key, keys_[0]);
                    Swap(value, values_[0]);
                    topdownHeapify(0);
                }
                return false;
            }
            keys_[cnt_] = key;
            values_[cnt_] = value;
            bottomupHeapify(cnt_);
            cnt_++;
            return true;
        }
        /** remove the minimum entity out of heap.
            @param key key
            @param value value corresponding to this key
            @return true if heap is not empty and key/value has been
            fetched, false otherwise. */
        bool removeMin(Tkey& key, Tval& value)
        {
            if (cnt_ <= 0)
                return false;
            key = keys_[0];
            value = values_[0];
            cnt_--;
            keys_[0] = keys_[cnt_];
            values_[0] = values_[cnt_];
            topdownHeapify(0);
            return true;
        }
        /** remove an entity by index.
			Note: since the buffers are supplied externally.
			a program can first get the count of elements in
			the heap, then look through the buffer to find
			the element it wants to remove, then call
			remove_by_index for that index.
			*IMPORTANT*, after a successful removal, the data
			in the heap buffres may be modified.
            @param i the index
            @return true if i is valid.
        */
        bool remove_by_index(int index)
        {
			// check the index
			if ((index<0)||(index>=cnt_))
                return false;

            cnt_--;
            keys_[index] = keys_[cnt_];
            values_[index] = values_[cnt_];
            Heapify(index);
            return true;
        }
        /** just take a look at the minimum entity in the heap and not
            remove it from the heap.
            @param key key
            @param value value corresponding to this key
            @return true if heap is not empty and key/value has been
            fetched, false otherwise. */
        bool peekMin(Tkey& key, Tval& value)
        {
            if (cnt_ <= 0)
                return false;
            key = keys_[0];
            value = values_[0];
            return true;
        }
        //@}


    protected:

    private:
        /**@name private member variables */
        //@{
        /// how many entities in heap
        int   cnt_;
        /// maximum \# of entities to be hold in heap
        int   size_;
        /// array of keys
        Tkey* keys_;
        /// array of values
        Tval* values_;
        //@}

        /**@name private methods */
        //@{
        /** operate top-down to make the array back into status in which
            the Heap property holds.
            @param idx the index to array where the operation starts */
        void topdownHeapify(int idx)
        {
            while (idx+1 < cnt_)
            {
                int left = idx * 2 + 1;
                int right = left + 1;
                if (right < cnt_)
                {
                    if ((keys_[idx] > keys_[left]) ||
                        (keys_[idx] > keys_[right]))
                    {
                        if (keys_[left] > keys_[right])   // move right up
                        {
                            Swap(keys_[idx], keys_[right]);
                            Swap(values_[idx], values_[right]);
                            idx = right;
                        }
                        else   // move left up
                        {
                            Swap(keys_[idx], keys_[left]);
                            Swap(values_[idx], values_[left]);
                            idx = left;
                        }
                    }
                    else
                    {
                        break;   // in position
                    }
                }
                else if (left < cnt_)
                {
                    if (keys_[idx] > keys_[left])
                    {
                        Swap(keys_[idx], keys_[left]);
                        Swap(values_[idx], values_[left]);
                        idx = left;
                    }
                    else
                    {
                        break;   // in position
                    }
                }
                else   // leaf node
                {
                    break;   // in position
                }
            }
        }
        /** operate bottom-up to make the array back into status in which
            the Heap property holds.
            @param idx the index to array where the operation starts */
        void bottomupHeapify(int idx)
        {
            while (idx > 0)
            {
                int parent = (idx-1) / 2;
                if (keys_[idx] > keys_[parent])
                    break;   // in position
                Swap(keys_[idx], keys_[parent]);
                Swap(values_[idx], values_[parent]);
                idx = parent;
            }
        }

        /** operate either top-down or bottom-up to make the array back into status in which
            the Heap property holds.
            @param idx the index to array where the operation starts */
        void Heapify(int idx)
        {
			if (idx == 0) {
				topdownHeapify(0);
				return;
			}

			int parent = (idx-1)/2;
			if (keys_[parent] > keys_[idx]) {
				bottomupHeapify(idx);
				return;
			}

			topdownHeapify(idx);
		}
        //@}
        
        // the copy constructor
        HeapMin(const HeapMin& rhs);
        // the assignment operator
        HeapMin& operator=(const HeapMin& rhs);

};	// end of class HeapMin


/** HeapMax is a wrap for Heap algorithm. It uses external buffers
    to keep data. The root of the Heap will always be the smallest
    value. The HeapMax means a Heap that has the following property:
    each entity within Heap has key bigger than that of its parent
    node. */
template<typename Tkey, typename Tval>
class HeapMax {

    public:
        /**@name public methods */
        //@{
        /** the constructor
            @param size the size of buffer keys/values
            @param keys buffer for keys
            @param values buffer for values */
        HeapMax(int size, Tkey* keys, Tval* values)
        {
            size_ = size;
            keys_ = keys;
            values_ = values;
            cnt_ = 0;
        }
        /// the destructor
        ~HeapMax() {}
        /** check if the heap is full.
            @return true if the heap is full, false otherwise. */
        bool isFull() { return (cnt_ >= size_); }
        /** check if the heap is empty.
            @return true if the heap is empty, false otherwise. */
        bool isEmpty() { return ( cnt_ <= 0); }
        /** check how many entities are in Heap.
            @return the \# of entities in Heap. */
        int  howMany() { return cnt_; }
        /** clean all entities within heap */
        void clean() { cnt_ = 0; }
        /** add one entity into heap.
            @param key key
            @param value value corresponding to this key
            @return true if heap is not full and key/value has been added
            into heap, false otherwise and the smallest entity to be
            dropped will be assigned to parameters key and value when
            return. */
        bool add(Tkey& key, Tval& value)
        {
            if (cnt_ >= size_)   // heap is full
            {
                if (key < keys_[0])
                {
                    Swap(key, keys_[0]);
                    Swap(value, values_[0]);
                    topdownHeapify(0);
                }
                return false;
            }
            keys_[cnt_] = key;
            values_[cnt_] = value;
            bottomupHeapify(cnt_);
            cnt_++;
            return true;
        }
        /** remove the maximum entity out of heap.
            @param key key
            @param value value corresponding to this key
            @return true if heap is not empty and key/value has been
            fetched, false otherwise. */
        bool removeMax(Tkey& key, Tval& value)
        {
            if (cnt_ <= 0)
                return false;
            key = keys_[0];
            value = values_[0];
            cnt_--;
            keys_[0] = keys_[cnt_];
            values_[0] = values_[cnt_];
            topdownHeapify(0);
            return true;
        }
        /** just take a look at the maximum entity in the heap and not
            remove it from the heap.
            @param key key
            @param value value corresponding to this key
            @return true if heap is not empty and key/value has been
            fetched, false otherwise. */
        bool peekMax(Tkey& key, Tval& value)
        {
            if (cnt_ <= 0)
                return false;
            key = keys_[0];
            value = values_[0];
            return true;
        }
        //@}


    protected:

    private:
        /**@name private member variables */
        //@{
        /// how many entities in heap
        int   cnt_;
        /// maximum \# of entities to be hold in heap
        int   size_;
        /// array of keys
        Tkey* keys_;
        /// array of values
        Tval* values_;
        //@}

        /**@name private methods */
        //@{
        /** operate top-down to make the array back into status in which
            the Heap property holds.
            @param idx the index to array where the operation starts */
        void topdownHeapify(int idx)
        {
            while (idx+1 < cnt_)
            {
                int left = idx * 2 + 1;
                int right = left + 1;
                if (right < cnt_)
                {
                    if ((keys_[idx] < keys_[left]) ||
                        (keys_[idx] < keys_[right]))
                    {
                        if (keys_[left] < keys_[right])   // move right up
                        {
                            Swap(keys_[idx], keys_[right]);
                            Swap(values_[idx], values_[right]);
                            idx = right;
                        }
                        else   // move left up
                        {
                            Swap(keys_[idx], keys_[left]);
                            Swap(values_[idx], values_[left]);
                            idx = left;
                        }
                    }
                    else
                    {
                        break;   // in position
                    }
                }
                else if (left < cnt_)
                {
                    if (keys_[idx] < keys_[left])
                    {
                        Swap(keys_[idx], keys_[left]);
                        Swap(values_[idx], values_[left]);
                        idx = left;
                    }
                    else
                    {
                        break;   // in position
                    }
                }
                else   // leaf node
                {
                    break;   // in position
                }
            }
        }
        /** operate bottom-up to make the array back into status in which
            the Heap property holds.
            @param idx the index to array where the operation starts */
        void bottomupHeapify(int idx)
        {
            while (idx > 0)
            {
                int parent = (idx-1) / 2;
                if (keys_[idx] < keys_[parent])
                    break;   // in position
                Swap(keys_[idx], keys_[parent]);
                Swap(values_[idx], values_[parent]);
                idx = parent;
            }
        }
        //@}
        
        // the copy constructor
        HeapMax(const HeapMax& rhs);
        // the assignment operator
        HeapMax& operator=(const HeapMax& rhs);

};	// end of class HeapMax


/** HeapMin2Key is a variation for HeapMin. It uses two key fields
    each time to decide ordering and uses external buffers to keep
    data. The 1st key field will determain the primary order and
    the 2nd field for the secondary order. The root of the Heap will
    always be the smallest value. The HeapMin means a Heap that has
    the following property: each entity within Heap has key bigger
    than that of its parent node. */
template<typename Tkey1, typename Tkey2, typename Tval>
class HeapMin2Key {

    public:
        /**@name public methods */
        //@{
        /** the constructor
            @param size the size of buffer keys/values
            @param keys buffer for keys
            @param values buffer for values */
        HeapMin2Key(int size, Tkey1* keys1, Tkey2* keys2, Tval* values)
        {
            size_ = size;
            keys1_ = keys1;
            keys2_ = keys2;
            values_ = values;
            cnt_ = 0;
        }
        /// the destructor
        ~HeapMin2Key() {}
        /** check if the heap is full.
            @return true if the heap is full, false otherwise. */
        bool isFull() { return (cnt_ >= size_); }
        /** check if the heap is empty.
            @return true if the heap is empty, false otherwise. */
        bool isEmpty() { return ( cnt_ <= 0); }
        /** check how many entities are in Heap.
            @return the \# of entities in Heap. */
        int  howMany() { return cnt_; }
        /** clean all entities within heap */
        void clean() { cnt_ = 0; }
        /** add one entity into heap.
            @param key key
            @param value value corresponding to this key
            @return true if heap is not full and key/value has been added
            into heap, false otherwise and the smallest entity to be
            dropped will be assigned to parameters key and value when
            return. */
        bool add(Tkey1& key1, Tkey2& key2, Tval& value)
        {
            if (cnt_ >= size_)   // heap is full
            {
                if (bigger(key1, key2, keys1_[0], keys2_[0]))
                {
                    Swap(key1, keys1_[0]);
                    Swap(key2, keys2_[0]);
                    Swap(value, values_[0]);
                    topdownHeapify(0);
                }
                return false;
            }
            keys1_[cnt_] = key1;
            keys2_[cnt_] = key2;
            values_[cnt_] = value;
            bottomupHeapify(cnt_);
            cnt_++;
            return true;
        }
        /** remove the minimum entity out of heap.
            @param key1 primary key
            @param key2 secondary key
            @param value value corresponding to this key
            @return true if heap is not empty and key/value has been
            fetched, false otherwise. */
        bool removeMin(Tkey1& key1, Tkey2& key2, Tval& value)
        {
            if (cnt_ <= 0)
                return false;
            key1 = keys1_[0];
            key2 = keys2_[0];
            value = values_[0];
            cnt_--;
            keys1_[0] = keys1_[cnt_];
            keys2_[0] = keys2_[cnt_];
            values_[0] = values_[cnt_];
            topdownHeapify(0);
            return true;
        }
        /** just take a look at the minimum entity in the heap and not
            remove it from the heap.
            @param key1 primary key
            @param key2 secondary key
            @param value value corresponding to this key
            @return true if heap is not empty and key/value has been
            fetched, false otherwise. */
        bool peekMin(Tkey1& key1, Tkey2& key2, Tval& value)
        {
            if (cnt_ <= 0)
                return false;
            key1 = keys1_[0];
            key2 = keys2_[0];
            value = values_[0];
            return true;
        }
        //@}


    protected:

    private:
        /**@name private member variables */
        //@{
        /// how many entities in heap
        int    cnt_;
        /// maximum \# of entities to be hold in heap
        int    size_;
        /// array of 1st key fields
        Tkey1* keys1_;
        /// array of 2nd key fields
        Tkey2* keys2_;
        /// array of values
        Tval*  values_;
        //@}

        /**@name private methods */
        //@{
        /** decide ordering for two key fields
         */
        bool bigger(Tkey1 key1_1, Tkey2 key2_1, Tkey1 key1_2, Tkey2 key2_2)
        {
            return ((key1_1 > key1_2) ||
                    ((key1_1 == key1_2) && (key2_1 > key2_2)));
        }
        /** operate top-down to make the array back into status in which
            the Heap property holds.
            @param idx the index to array where the operation starts */
        void topdownHeapify(int idx)
        {
            while (idx+1 < cnt_)
            {
                int left = idx * 2 + 1;
                int right = left + 1;
                if (right < cnt_)
                {
                    if (bigger(keys1_[idx], keys2_[idx],
                               keys1_[left], keys2_[left]) ||
                        bigger(keys1_[idx], keys2_[idx],
                               keys1_[right], keys2_[right]))
                    {
                        if (bigger(keys1_[left], keys2_[left],
                                   keys1_[right], keys2_[right]))
                        {
                            Swap(keys1_[idx], keys1_[right]);
                            Swap(keys2_[idx], keys2_[right]);
                            Swap(values_[idx], values_[right]);
                            idx = right;
                        }
                        else   // move left up
                        {
                            Swap(keys1_[idx], keys1_[left]);
                            Swap(keys2_[idx], keys2_[left]);
                            Swap(values_[idx], values_[left]);
                            idx = left;
                        }
                    }
                    else
                    {
                        break;   // in position
                    }
                }
                else if (left < cnt_)
                {
                    if (bigger(keys1_[idx], keys2_[idx],
                               keys1_[left], keys2_[left]))
                    {
                        Swap(keys1_[idx], keys1_[left]);
                        Swap(keys2_[idx], keys2_[left]);
                        Swap(values_[idx], values_[left]);
                        idx = left;
                    }
                    else
                    {
                        break;   // in position
                    }
                }
                else   // leaf node
                {
                    break;   // in position
                }
            }
        }
        /** operate bottom-up to make the array back into status in which
            the Heap property holds.
            @param idx the index to array where the operation starts */
        void bottomupHeapify(int idx)
        {
            while (idx > 0)
            {
                int parent = (idx-1) / 2;
                if (bigger(keys1_[idx], keys2_[idx],
                           keys1_[parent], keys2_[parent]))
                    break;   // in position
                Swap(keys1_[idx], keys1_[parent]);
                Swap(keys2_[idx], keys2_[parent]);
                Swap(values_[idx], values_[parent]);
                idx = parent;
            }
        }
        //@}
        
        // the copy constructor
        HeapMin2Key(const HeapMin2Key& rhs);
        // the assignment operator
        HeapMin2Key& operator=(const HeapMin2Key& rhs);

};	// end of class HeapMin2Key



/*[]-----------------------------------------[]
  |  Inline Methods                           |
  []-----------------------------------------[]*/



/*[]-----------------------------------------[]
  |  External References                      |
  []-----------------------------------------[]*/



#endif // H_HEAP
