// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: loc_entities.cc,v $ - Defining all entities in BF location simulation.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/07/22 02:25:30 $
 * Version :  $Revision: 1.11 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

# include "gsldistrib.h"
# include "usc_time.h"
# include "../../daemon/load_monitor/cbs_bf.h"
# include "loc_events.h"
# include "loc_sim.h"
# include "loc_entities.h"

/* maximum # of objects per server */
# define MAX_MAX_OBJECTS 1000000

using namespace std;

Server::Server(int sid, int init_objs, double update_int) :
	CBS_BF((init_objs>MAX_MAX_OBJECTS)?MAX_MAX_OBJECTS:init_objs),
	id(sid),
	bcast_interval(update_int),
	min_obj(init_objs)
{
	// nothing
}

Server::~Server()
{
	// nothing
}

bool Server::good()
{
	if ((id>=0)&&(bcast_interval>0))
		return true;

	return false;
}

bool Server::add(const guid_t &guid)
{
	GUID_KEY key(guid);

	// check if it is already existent.
	GUID_SET::iterator pi=local_store.find(key);
	if (pi!=local_store.end()) {
		return false;
	}

	if (CBS_BF::add(guid)==false)
		return false;

	// hopefully the following should always return true.
	pair<GUID_SET::iterator, bool> ret=local_store.insert(key);

	if (zero_pcent()<0.45) {
		expand();
	}

	return ret.second;
}

bool Server::del(const guid_t &guid)
{
	GUID_KEY key(guid);

	GUID_SET::iterator pi=local_store.find(key);
	if (pi==local_store.end()) 
		return false;

	local_store.erase(pi);

	if (CBS_BF::del(guid)) {
		// if ((zero_pcent()>0.80)&&(n_obj>min_obj)) {
			// shrink();
		// }
		return true;
	}
	else {
		return false;
	}
}

bool Server::find(const guid_t &guid)
{
	GUID_KEY key(guid);
	GUID_SET::iterator pi=local_store.find(key);
	if (pi==local_store.end()) 
		return false;

	return true;
}

bool Server::check_consistency()
{
	GUID_SET::iterator pi=local_store.begin();
	for (; pi!=local_store.end(); pi++) {
		if (check((*pi).key)==false)
			return false;
	}

	return true;
}

bool Server::shrink()
{
	// if (pthread_mutex_lock(&lock)<0) {
		// return false;
	// }

	bool ret=adjust(msize/2);
	// pthread_mutex_unlock(&lock);

	return ret;
}

bool Server::expand()
{
	// if (pthread_mutex_lock(&lock)<0) {
		// return false;
	// }
	bool ret=adjust(msize*2);
	// pthread_mutex_unlock(&lock);

# if defined(__DEBUG__)
	tmpi_error(DBG_VERBOSE, "Size expansion for Server %d.\n", id);
# endif
	return ret;
}

bool Server::adjust(int new_msize)
{
	msize=new_msize;
	fsize=prime_within(msize*BF_PAGE_SIZE*8);
	if (mask) {
		delete []mask;
	}
	mask=new unsigned char [msize*BF_PAGE_SIZE];

	if (bits_cnt) {
		delete []bits_cnt;
	}
	bits_cnt=new unsigned char[fsize];

	if (page_clock) {
		delete []page_clock;
	}

	page_clock=new unsigned char[msize];

	reset();

	GUID_SET::iterator pi=local_store.begin();
	for (; pi!=local_store.end(); pi++) {
		CBS_BF::add(pi->key);
	}

	update_reset(); // clear update info

# if defined(__DEBUG__)
	if (check_consistency()==false) {
		tmpi_error(DBG_INTERNAL, "server %d failed consistency check after size adjustment.\n", id);
	}
# endif
	return true;
}

Client::Client(int cid, int svr_num, int init_obj) :
	id(cid),
	bfs(NULL),
	n_bfs(svr_num),
	n_add(0),
	n_add_fail(0),
	n_lookup(0),
	n_lookup_fail(0),
	n_false_pos(0),
	n_true_pos(0),
	n_false_neg(0),
	n_true_neg(0),
	n_follow_bcast(0),
	n_del(0),
	n_del_fail(0),
	n_bcast_rcvd(0),
	n_bcast_bytes_rcvd(0),
	n_unicast_rcvd(0),
	n_unicast_bytes_rcvd(0),
	n_bcast_drop(0)
{
	if ((svr_num<=0)&&(init_obj<=0))
		return;

	if (init_obj>MAX_MAX_OBJECTS)
		init_obj=MAX_MAX_OBJECTS;

	bfs=(CBS_BaseBF **)calloc(svr_num, sizeof(CBS_BaseBF));
	if (bfs) {
		for (int i=0; i<svr_num; i++) {
			bfs[i]=new CBS_BaseBF(init_obj);
		}
	}
}

Client::~Client()
{
	if (bfs) {
		for (int i=0; i<n_bfs; i++) {
			if (bfs[i]) {
				delete bfs[i];
			}
		}
		free(bfs);
	}
}

bool Client::good()
{
	if (bfs)
		return true;

	return false;
}

bool Client::handle_request(Simulator *sim, RequestEvent *e)
{
	if ((!sim)||(!e))
		return false;

	bool do_delete=false;
	switch (e->type()) {
	case RequestEvent::ADD:
		{
			n_add++;
			if (sim->add_object()==false) {
				n_add_fail++;
			}

			// e->dump();
			break;
		}
	case RequestEvent::DEL:
		do_delete=true;
		n_del++;
		// fall through to do lookup.
	case RequestEvent::LOOKUP:
		{
			n_lookup++;
			// always generate a valid GUID.
			guid_t guid;
			if (sim->global->get_random(guid, do_delete)==false) {
				n_lookup_fail++;
				break;
			}

# if defined(__DEBUG__)
			{
				char guid_str[100];
				Guid tmp_guid((void *)&guid, GUID_LEN);
				tmp_guid.to_string(guid_str, 100);
				tmpi_error(DBG_VERBOSE, "Lookup object %s.\n", guid_str);
			}
# endif
			// local lookup
			bool found=false;
			bool did_udp=false;
			BaseBF::check_handle_t ch;
			CBS_BaseBF::prepare_check(guid, ch); // prepare check
			for (int i=0; i<n_bfs; i++) {
				if (bfs[i]->post_check(ch)==true) {
# if defined(__DEBUG__)
					tmpi_error(DBG_VERBOSE, "Found object in BF[%d].\n", i);
# endif
					// server verification
					did_udp=true;
					if (sim->servers[i]->find(guid)==true) {
# if defined(__DEBUG__)
						tmpi_error(DBG_VERBOSE, "Object existence in Server %d verified.\n", i);
# endif
						n_true_pos++;
						found=true;
						if (do_delete) {
							sim->servers[i]->del(guid);
						}
					}
					else {
# if defined(__DEBUG__)
						tmpi_error(DBG_VERBOSE, "Object does not exist in Server %d.\n", i);
# endif
						n_false_pos++;
					}
				}
			}

			if (!found) {
				if (did_udp)
					n_follow_bcast++;
				for (int i=0; i<sim->get_svr_num(); i++) {
					if (sim->servers[i]->find(guid)==true) {
						found=true;
						if (do_delete) {
							sim->servers[i]->del(guid);
						}
						break;
					}
				}
				if (!found)
					n_true_neg++;
				else
					n_false_neg++;
			}
			break;
		}
	default:
		tmpi_error(DBG_INTERNAL, "Unknown event type:%d.\n", e->type());
		break;
	}
	
	return true;
}	

bool Client::handle_bcast(Simulator *sim, BcastEvent *e)
{
	if ((!sim)||(!e)) {
		return false;
	}

	int sid=e->get_sid();
	const unsigned short pkt_size=1450;
	unsigned short pkt_len=pkt_size/sizeof(unsigned int)*sizeof(unsigned int);
	unsigned int pkt_buf[pkt_len];
	sim->servers[sid]->make_bf_update_packet
		((CBS_BF_PKT *)pkt_buf, pkt_len);

	// start with a packet loss filter
	if (sim->pkt_drop>0) {
		double should_drop=gsldd_flat(0,1);
		if (should_drop<sim->pkt_drop) {
			n_bcast_drop++;
			// if (sim->servers[sid]->cmp(bfs[sid])==true) {
			// tmpi_error(DBG_INTERNAL, "Filters still consistent after packet drop (server=%d).\n", sid);
			// }
			return true;
		}
	}

	// ((CBS_BF_PKT *)pkt_buf)->dump();
	n_bcast_rcvd++;
	n_bcast_bytes_rcvd+=pkt_len;
	if (bfs[sid]->get_fsize() == ((CBS_BF_PKT *)pkt_buf)->n_bits) {
		bfs[sid]->update_from_packet((CBS_BF_PKT *)pkt_buf);
		// # if defined(__DEBUG__)
		// if (sim->servers[sid]->cmp(bfs[sid])==false) {
			// tmpi_error(DBG_INTERNAL, "Inconsistent after partial update (server=%d) lab=%d, pg=%d.\n", sid, 
					   // ((CBS_BF_PKT *)pkt_buf)->n_lab_entries, ((CBS_BF_PKT *)pkt_buf)->n_bf_pages);
		// }
		// # endif

	}
	else {
		bfs[sid]->clone(*(sim->servers[sid]));
		n_unicast_rcvd++;
		n_unicast_bytes_rcvd+=bfs[sid]->get_size();
		// # if defined(__DEBUG__)
		tmpi_error(DBG_VERBOSE, "Server %d expansion detected, resync.\n", sid);
		// # endif

		// # if defined(__DEBUG__)
		// if (sim->servers[sid]->check_consistency()==false) {
			// tmpi_error(DBG_INTERNAL, "Inconsistent after complete update (server=%d).\n", sid);
		// }
		// # endif
	}

	return true;
}

bool Client::handle_icheck(Simulator *sim, InstantCheckEvent *e)
{
	int bad_udp=0, good_bcast=0, bad_bcast=0, n_udpbcast=0;
	const int N_REQUEST=10000;
	int n_req=0;
	for (; n_req<N_REQUEST; n_req++) {
		guid_t guid;
		if (sim->global->get_random(guid, false)==false) {
			break;
		}

		// local lookup
		bool found=false;
		bool did_udp=false;
		BaseBF::check_handle_t ch;
		CBS_BaseBF::prepare_check(guid, ch); // prepare check
		for (int i=0; i<n_bfs; i++) {
			if (bfs[i]->post_check(ch)==true) {
				// server verification
				did_udp=true;
				if (sim->servers[i]->find(guid)==true) {
					found=true;
				}
				else {
					bad_udp++;
				}
			}
		}

		if (!found) {
			if (did_udp)
				n_udpbcast++;
			for (int i=0; i<sim->get_svr_num(); i++) {
				if (sim->servers[i]->find(guid)==true) {
					found=true;
					break;
				}
			}
			if (!found)
				bad_bcast++;
			else
				good_bcast++;
		}
	}

	if (n_req<=0) {
		tmpi_error(DBG_INTERNAL, "Instant check @ %.3f failed: cannot get object ID.\n", 
				   e->time());
	}
	else if (bad_bcast>0) {
		tmpi_error(DBG_INTERNAL, "Instant check @ %.3f failed: should be no true negative.\n", 
				   e->time());
	}
	else {
		double fp_udp=(double)bad_udp/n_req;
		double fn_bcast=(good_bcast+n_udpbcast)*100.0/n_req;
		double fpfn_bcast=(n_udpbcast)*100.0/n_req;
		tmpi_error(DBG_APP, "Instant check @ %.3f: FP_UDP=%.4f, BCAST=%.3f%%, FP_BCAST=%.3f%%, FPFN_BCAST=%.3f%%.\n",
				   e->time(), fp_udp, fn_bcast+fpfn_bcast, fn_bcast, fpfn_bcast);
	}

	return true;
}

double Client::handle_kcheck(Simulator *sim, double &fp)
{
	const int N_REQUEST=10000;
	int n_req=0;
	int ifound=0;
	usc_time_t t_start=usc_gettime();
	for (; n_req<N_REQUEST; n_req++) {
		guid_t guid;
		if (sim->global->get_random(guid, false)==false) {
			break;
		}

		// local lookup
		BaseBF::check_handle_t ch;
		CBS_BaseBF::prepare_check(guid, ch); // prepare check
		for (int i=0; i<n_bfs; i++) {
			if (bfs[i]->post_check(ch)==true) {
				ifound++;
			}
		}
	}
	usc_time_t t_end=usc_gettime();
	double t1=usc_difftime(t_end, t_start);

	t_start=usc_gettime();
	for (int j=0; j<n_req; j++) {
		guid_t guid;
		if (sim->global->get_random(guid, false)==false) {
			break;
		}
	}
	t_end=usc_gettime();
	double t2=usc_difftime(t_end, t_start);
	fp=ifound*1.0/n_req-1;

	return (t1-t2)/n_req;
}

void Client::report()
{
	printf("ADD=%d, DEL=%d, LOOKUP(including pre-DEL LOOKUP)=%d\n"\
		   "MCAST=%d, MDROP=%d, MC_BYTES=%d, UNICAST=%d, UNI_BYTES=%d.\n", 
		   n_add-n_add_fail,
		   n_del-n_del_fail,
		   n_lookup-n_lookup_fail,
		   n_bcast_rcvd,
		   n_bcast_drop,
		   n_bcast_bytes_rcvd,
		   n_unicast_rcvd,
		   n_unicast_bytes_rcvd);
	printf("Lookup statistics:\n");
	printf("\t%.4f%% UDP, %.4f%% BCAST, %.4f%% UDP/BCAST\n",
		   (double)n_udp_pos_request()*100.0/(n_lookup-n_lookup_fail),
		   (double)(n_bcast()-n_follow_bcast)*100.0/(n_lookup-n_lookup_fail),
		   (double)n_follow_bcast*100.0/(n_lookup-n_lookup_fail)
		   );
	printf("\tavg_udp=%.4f, avg_fp_udp=%.4f.\n",
		   avg_udp(), avg_false_pos_udp());
	return;
}

God::God(int max_obj) :
	gtab(NULL),
	size(max_obj),
	idx(0)
{
	if (max_obj>0) {
		gtab=(guid_t *)calloc(max_obj, GUID_LEN);
	}
}

God::~God()
{
	if (gtab)
		free(gtab);
}

bool God::good()
{
	return (gtab!=NULL);
}

bool God::add(guid_t &guid)
{
	if (idx==size)
		return false;
	memcpy(&(gtab[idx]), &guid, GUID_LEN);
	idx++;
	return true;
}

bool God::get_random(guid_t &guid, bool del_it)
{
	if (idx==0)
		return false;

	unsigned int ui=Guid::random_ui();
	ui=ui%(unsigned int)idx;
	memcpy(&guid, &(gtab[ui]), GUID_LEN);
	if (del_it) {
		idx--;
		if ((unsigned int)idx!=ui) {
			memcpy(&(gtab[ui]), &(gtab[idx]), GUID_LEN);
		}
		memset(&(gtab[idx]), 0, GUID_LEN);
	}

	return true;
}

void Client::read_init_bfs(Server **servers, int nsvr)
{
	if ((!good())||(nsvr!=n_bfs))
		return;
	
	for (int i=0; i<nsvr; i++) {
		bfs[i]->clone(*(servers[i]));
		n_unicast_rcvd++;
		n_unicast_bytes_rcvd+=bfs[i]->get_size();
		servers[i]->update_reset();
		/**
		   char page_buf[4096];
		   int n_pgs_in_buf=4096/BF_PAGE_SIZE;
		   int n_bf_pgs=servers[i]->get_num_page();
		   int next_page=0;
		   while (n_bf_pgs>0) {
		   int n=(n_bf_pgs>n_pgs_in_buf)?n_pgs_in_buf:n_bf_pgs;
		   n=servers[i]->get_pages(page_buf, next_page, n);
		   bfs[i]->update_pages(page_buf, next_page, n);
		   next_page+=n;
		   n_bf_pgs-=n;
		   }
		**/
# if defined(__DEBUG__)
		if (servers[i]->cmp(bfs[i])==false) {
			tmpi_error(DBG_INTERNAL, "Initial update failed.\n");
		}
		if (servers[i]->check_consistency()==false) {
			tmpi_error(DBG_INTERNAL, "server %d failed consistency check.\n", i);
		}
# endif
	}
}
