// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: loc_entities.h,v $ - header file for loc_entities.cc - defining all 
 *            entities in a BF location simulator.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/07/22 02:25:30 $
 * Version :  $Revision: 1.8 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

#ifndef _LOC_ENTITIES_H_
#define _LOC_ENTITIES_H_

# include <string.h>
# include <set>
# include "guid.h"
# include "../../daemon/load_monitor/cbs_bf.h"

class Simulator;
class RequestEvent;
class BcastEvent;
class InstantCheckEvent;


class Server : public CBS_BF {
 private:
	struct GUID_KEY {
		guid_t key;

		GUID_KEY()
		{
			memcpy((char *)&key, (char *)&Guid::GUID_NIL, GUID_LEN);
		}

		GUID_KEY(const guid_t &id)
		{
			memcpy((char *)&key, (char *)&id, GUID_LEN);
		}
	};

	struct GUID_cmp : public std::less<GUID_KEY> {
		bool operator()(const GUID_KEY &x, const GUID_KEY &y) const {
			return(memcmp((const char *)&(x.key), (const char *)&(y.key), GUID_LEN)<0);
		}
	};

	bool shrink();
	bool expand();
	bool adjust(int new_msize);

	typedef std::set<GUID_KEY, GUID_cmp> GUID_SET;
	int id;
	GUID_SET local_store;
	double bcast_interval;
	int min_obj;

 public:
	Server(int sid, int max_objs, double update_int);
	~Server();
	bool good();
	bool add(const guid_t &id);
	bool del(const guid_t &id);
	bool find(const guid_t &id);
	double get_bcast_int() { return bcast_interval; }
	bool check_consistency();
	unsigned int get_obj_cnt2() { return local_store.size(); }
	// static char *update_policy_string() {
		// return CBS_BF::update_policy_string();
	// }
};

class Client {
 private:
	int id;
	CBS_BaseBF **bfs;
	int n_bfs;

	// statistics
	// addition statistics
	int n_add; // # of object additions
	// addition failures: 
	// global table full, server table full.
	// as long as some servers successfully adds the object
	// we do not consider it as a failure.
	int n_add_fail; 

	// the following are lookup statistics
	int n_lookup; // # of object lookup (including the lookup in DEL op)
	int n_lookup_fail; // # of object lookup failed internally.
	int n_false_pos; // # of occurances of false positive 
	int n_true_pos; // # of occurances of true positive
	int n_false_neg; // # of occurances of false negative
	int n_true_neg; // # of occurances of true negative
	int n_follow_bcast; // # of bcasts that follow failed UDP lookup

	// the following are del statistics
	int n_del; // # of object deletions
	int n_del_fail; // # of object deletion failures.

	// the following are update statistics
	int n_bcast_rcvd;
	int n_bcast_bytes_rcvd;
	int n_unicast_rcvd;
	int n_unicast_bytes_rcvd;
	int n_bcast_drop;

 public:
	Client(int cid, int svr_num, int max_obj);
	~Client();
	bool good();
	void read_init_bfs(Server **, int nsvr);
	bool handle_request(Simulator *sim, RequestEvent *e);
	bool handle_bcast(Simulator *sim, BcastEvent *e);
	bool handle_icheck(Simulator *sim, InstantCheckEvent *e);
	double handle_kcheck(Simulator *sim, double &fp);
	void report();
	// # of broadcast (because no hit found through UDP).
	int n_bcast() { return n_true_neg+n_false_neg; }
	// # of UDP verification sent
	int n_udp() { return n_true_pos+n_false_pos; }
	// # of requests ended in a successful UDP verification
	int n_udp_pos_request() { return n_lookup-n_lookup_fail-n_bcast(); }
	// # of requests turned into UDP verifications
	int n_udp_request() { return n_udp_pos_request()+n_follow_bcast; }
	// average # of UDP verifications per lookup request
	double avg_udp() { return (double)n_udp()/(n_udp_request()); }
	// average # of failed UDP verifications per lookup request
	double avg_false_pos_udp() { return (double)n_false_pos/(n_udp_request()); }
	// average # of successful UDP verifications per lookup request
	double avg_true_pos_udp() { return (double)n_true_pos/(n_udp_request()); }
};

class God {
 private:
	guid_t *gtab;
	int size;
	int idx;

 public:
	God(int max_obj);
	~God();
	bool good();
	bool add(guid_t &guid);
	bool get_random(guid_t &guid, bool del_it=false);
	int get_size() { return idx; }
};

# endif
