// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: loc_sim.cc,v $ - CBS Bloom Filter Object Locationing Simulator
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/07/22 02:25:30 $
 * Version :  $Revision: 1.9 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

# include <stdio.h>
# include <stdlib.h>
# include <unistd.h>

# include "gsldistrib.h"
# include "guid.h"
# include "usc_time.h"
# include "calevent.h"
# include "progress.h"
# include "loc_requestgen.h"
# include "loc_entities.h"
# include "loc_events.h"
# include "loc_sim.h"

static double random_bcast_delta(double itvl)
{
	double sigma=itvl/3;
	double tt=gsldd_gaussian(sigma);

	if (tt>0)
		tt=tt-(int)(tt/sigma)*sigma;
	else 
		tt=tt+(int)(-tt/sigma)*sigma;
	tt+=itvl;

	return tt;
}

bool Simulator::add_object()
{
	guid_t oid;
	Guid guid;
	guid.gen();
	guid.pack(oid);
	if (global->add(oid)==false) {
		return false;
	}

	// prepare for a list of servers
	// that are going to host the object.
	int dup_svr_ids[dup_degree];
	int dsi_idx=0;

	while (dsi_idx<dup_degree) {
		double pp=gsldd_flat(0, 1);
		int cand=(int)(pp*svr_num);
		if (cand>=svr_num)
			cand=svr_num-1;
		bool exist=false;
		for (int i=0; i<dsi_idx; i++) {
			if (dup_svr_ids[i]==(int)cand) {
				exist=true;
				break;
			}
		}
		if (!exist) {
			dup_svr_ids[dsi_idx]=cand;
			dsi_idx++;
		}
	}

	bool obj_added=false;

# if defined(__DEBUG__)
	char guid_str[100];
	guid.to_string(guid_str,100);
# endif

	for (int i=0; i<dup_degree; i++) {
		if (servers[dup_svr_ids[i]]->add(oid)==true) {
			obj_added=true;
# if defined(__DEBUG__)
			tmpi_error(DBG_VERBOSE, "Object %s added to Server %d.\n", guid_str, dup_svr_ids[i]);
# endif
		}
# if defined(__DEBUG__)
		else {
			tmpi_error(DBG_VERBOSE, "Failed to add object %s to Server %d.\n", guid_str, dup_svr_ids[i]);
		}
# endif
	}	

	return obj_added;
}

void Simulator::init(double update_int, int stb_obj, int max_obj, int init_obj)
{
	servers=(Server **)calloc(svr_num, sizeof(Server *));
		
	if (servers) {
		for (int i=0; i<svr_num; i++) {
			servers[i]=new Server(i, (init_obj/svr_num), update_int);
		}
	}
	client=new Client(0, svr_num, init_obj/svr_num);
	global=new God(max_obj);
	sim_core=new EventChain();

	if (good()) {
		int show_prog=init_obj/100;
		for (int i=0; i<init_obj; i++) {
			add_object();
			if ((show_prog>0)&&(i%100==0)) {
				printf("\rInitializing: %.4f%% completed...", i*100.0/init_obj);
				fflush(stdout);
			}
		}
		if (show_prog>0) {
			printf("\rInitialization done!                   \n");
			fflush(stdout);
		}
		client->read_init_bfs(servers, svr_num);
	}
}

bool Simulator::start()
{
	if ((isatty(fileno(stdout)))&&(sim_time>0)) {
		set_progress_implementation("bar");
		prog=progress_create((long)0, (long)sim_time*1000);
		t_start=usc_gettime();
	}

	RequestEvent *re=req_gen->generate_next(global->get_size());
	if (!re) {
		tmpi_error(DBG_FATAL, "Cannot put seed request.\n");
		return false;
	}
	sim_core->schedule(re);

	if (snapshot>0) {
		InstantCheckEvent *ie=new InstantCheckEvent((double)snapshot);
		if (!ie) {
			tmpi_error(DBG_FATAL, "Cannot put seed request.\n");
			return false;
		}
		sim_core->schedule(ie);
	}

	for (int i=0; i<svr_num; i++) {
		double delta=random_bcast_delta(servers[i]->get_bcast_int());
		BcastEvent *be=new BcastEvent(i, delta);
		if (!be) {
			tmpi_error(DBG_FATAL, "Cannot put seed bcast event.\n");
			return false;
		}

		sim_core->schedule(be);
	}

	// init progress parameters
	if (prog) {
		sim_t_previous=0;
		t_start=t_previous=usc_gettime();
	}

	return true;
}

void Simulator::release()
{
	if (prog) {
		prog=NULL;
	}

	if (sim_core) {
		delete sim_core;
		sim_core=NULL;
	}
	
	if (global) {
		delete global;
		global=NULL;
	}

	if (client) {
		delete client;
		client=NULL;
	}

	if (servers) {
		for (int i=0; i<svr_num; i++) {
			if (servers[i]) {
				delete servers[i];
				servers[i]=NULL;
			}
		}
		free(servers);
		servers=NULL;
	}

	if (req_gen) {
		delete req_gen;
		req_gen=NULL;
	}
}

Simulator::Simulator(double t, double update_int, double r_add_del, double r_lookup, 
					 int ns, int degree, int stb_obj, int max_obj, int init_obj, 
					 double pkt_loss, int instant_dump) :
	prog(NULL),
	sim_time(t),
	dup_degree(degree),
	svr_num(ns),
	sim_core(NULL),
	req_gen(NULL),
	global(NULL),
	client(NULL),
	servers(NULL),
	pkt_drop(pkt_loss),
	snapshot(instant_dump)
{
	if ((t<=0)||(update_int<=0)||(r_add_del<=0)||(r_lookup<=0)||
		(ns<=0)||(degree<=0)||(stb_obj<=0)||(max_obj<=0)||
		(init_obj<=0))
		return;

	req_gen=new RequestGen(stb_obj, max_obj, r_add_del, r_lookup);
	if ((!req_gen)||(!req_gen->good()))
		return;
	init(update_int, stb_obj, max_obj, init_obj);
}

Simulator::Simulator(char *tc_path, double scale, double update_int, int ns, int degree, 
					  int stb_obj, int max_obj, int init_obj, double pkt_loss, int instant_dump) :
	prog(NULL),
	sim_time(0),
	dup_degree(degree),
	svr_num(ns),
	sim_core(NULL),
	req_gen(NULL),
	global(NULL),
	client(NULL),
	servers(NULL),
	pkt_drop(pkt_loss),
	snapshot(instant_dump)
{
	if ((!tc_path)||(update_int<=0)|| (ns<=0)||(degree<=0)||
		(stb_obj<=0)||(max_obj<=0)||(init_obj<=0))
		return;

	req_gen=new RequestGen(tc_path, scale);
	if (!req_gen)
		return;

	if (req_gen->good()) {
		sim_time=req_gen->get_trace_duration();
	}
	else {
		return;
	}

	init(update_int, stb_obj, max_obj, init_obj);
}

bool Simulator::good()
{
	if (sim_time<=0)
		return false;
	if (dup_degree<=0)
		return false;
	if (svr_num<=0)
		return false;
	if ((!global)||(!global->good()))
		return false;
	if ((!client)||(!client->good()))
		return false;
	if (!sim_core)
		return false;
	if ((!req_gen)||(!req_gen->good()))
		return false;
	if (!servers)
		return false;
	for (int i=0; i<svr_num; i++) {
		if (!servers[i])
			return false;
		if (!servers[i]->good())
			return false;
	}

	return true;
}

Simulator::~Simulator()
{
	release();
}

void Simulator::update_prog(double sim_t_now)
{
	if (prog) {
		double sim_delta_ms=(sim_t_now-sim_t_previous)*1000;
		if (sim_delta_ms>10) { // 10ms progress granularity
			usc_time_t now=usc_gettime();
			if (usc_difftime(now, t_previous)>10000) { // 10ms update frequency
				progress_update(prog, (long)sim_delta_ms, (long)usc_difftime(now, t_start));
				t_previous=now;
				sim_t_previous=sim_t_now;
			}
		}
	}
}

void Simulator::finalize_prog()
{
	if (prog) {
		usc_time_t now=usc_gettime();
		progress_finish(prog, (long)usc_difftime(now, t_start));
		prog=NULL;
	}
}
		
bool Simulator::run()
{
	///////////////////////////////////
	// building dispatching table
	struct ETAB {
		int etype;
		bool (Simulator::*f)(Event *);
	} etab0[] = {
		{ EVENT_REQUEST_ADD, &(Simulator::handle_request)}
		, { EVENT_REQUEST_DEL, &(Simulator::handle_request)}
		, { EVENT_REQUEST_LOOKUP, &(Simulator::handle_request)}
		, { EVENT_BCAST, &(Simulator::handle_bcast)}
		, { EVENT_INSTANT_CHECK, &(Simulator::handle_icheck)}
		, { -1, NULL}
	};		

	bool (Simulator::*etab[1000])(Event *);
	memset(etab, 0, sizeof(struct ETAB *)*1000);
	struct ETAB *p=etab0;

	while ((p->etype)!=-1) {
		etab[p->etype]=p->f;
		p++;
	}

	double sim_t_now;
	bool success=true;
	while ((sim_t_now=sim_core->time())<=sim_time) {
		update_prog(sim_t_now);

		// handle events;
		Event *e=sim_core->next_event();
		if (!e)
			break;

		if (!e->good()) {
			tmpi_error(DBG_INTERNAL, "Bad events.\n");
			success=false;
			e->dump();
			break;
		}

# if defined(__DEBUG__)
		e->dump();
# endif

		int etype=e->type();
		if ((etype<0)||(etype>=1000)) {
			tmpi_error(DBG_INTERNAL, "Bad event type: %d.\n", etype);
			success=false;
			break;
		}

		if ((this->*etab[etype])(e)==false) {
			success=false;
			break;
		}
	}

	finalize_prog();

	if (success)
		client->report();

	report();

	return success;
}

void Simulator::report()
{
	unsigned int total_bf_bytes=0;
	unsigned int total_objs=0;
	// unsigned int total_objs2=0;
	double zero_pcent=0;
	// double zero_pcent2=0;
	for (int i=0; i<svr_num; i++) {
		total_bf_bytes+=servers[i]->get_size();
		zero_pcent+=servers[i]->zero_pcent();
		// zero_pcent2+=servers[i]->zero_pcent_debug();
		total_objs+=servers[i]->get_obj_cnt();
		// total_objs2+=servers[i]->get_obj_cnt2();
		// tmpi_error(DBG_VERBOSE, "SERVER [%d] size=%d, fsize=%d, zero_pcent=%.4f%%, total_objs=%d.\n",
				   // i, servers[i]->get_size(), servers[i]->get_fsize(), servers[i]->zero_pcent()*100, servers[i]->get_obj_cnt());
	}
	printf("\tbf_mem=%d bytes, avg n_obj=%.2f, avg zero_pcent=%.4f%%.\n", 
		   total_bf_bytes, (float)total_objs/svr_num, zero_pcent/svr_num*100);
}

bool Simulator::handle_request(Event *e)
{
	RequestEvent *re=(RequestEvent *)e;
	if (client->handle_request(this, re)==false)
		return false;

	delete re;

	// generate next request
	re=req_gen->generate_next(global->get_size());
	if (re)
		sim_core->schedule(re);

	return true;
}

bool Simulator::handle_bcast(Event *e)
{
	BcastEvent *be=(BcastEvent *)e;
	if (client->handle_bcast(this, be)==false)
		return false;

	// generate next bcast
	double delta=random_bcast_delta(servers[be->get_sid()]->get_bcast_int());
	be->set_time(be->time()+delta);
	sim_core->schedule(be);

	return true;
}

bool Simulator::handle_icheck(Event *e)
{
	InstantCheckEvent *ie=(InstantCheckEvent *)e;
	if (client->handle_icheck(this, ie)==false)
		return false;

	// generate next instant check
	ie->set_time(ie->time()+snapshot);
	sim_core->schedule(ie);

	return true;
}
