// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: main.cc,v $ - The main LOC_SIM simulation control module.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/07/22 02:25:30 $
 * Version :  $Revision: 1.7 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

# include <sys/types.h>
# include <sys/stat.h>
# include <fcntl.h>
# include <stdio.h>
# include <stdlib.h>
# include <unistd.h>
# include <getopt.h>
# include <libgen.h>

# include "guid.h"
# include "tmpi_debug.h"
# include "loc_sim.h"
# include "loc_entities.h"
# include "../../daemon/load_monitor/cbs_bf.h"

// maximum 2000 hosts
# define MAX_NUM_HOSTS 2000

// maximum 20M objects (the real maximum
// would be MAX_NUM_OBJS/degree)
# define MAX_NUM_OBJS 20*1024*1024

static char *prog_name=NULL;

static void do_help()
{
	printf("\
Usage: %s [options]\n\
  -d <degree>        Replication degree (interger).\n\
  -e <pkt_loss>      Packet loss rate.\n\
  -f <file>          Trace file (cannot be combined with -t).\n\
  -h                 Help, this screen.\n\
  -i <init_obj>      Initial object count (interger).\n\
  -I <sec>           Dump instantaneous fp/fn rate (every sec seconds).\n\
  -k                 Checking bloom filter lookup performance.\n\
  -n <#servers>      Number of server nodes (interger).\n\
  -o <stb_obj>       Stable object count (interger).\n\
  -O <max_obj>       Maximum object count (interger).\n\
  -r <add/del rate>  Add/delete request rate (double, imply -t).\n\
  -R <lookup rate>   Lookup request rate (double, imply -t).\n\
  -s <t_scale>       Scaling factor in converting trace events (double, imply -f).\n\
  -t <duration>      Synthetic simulation duration (in seconds, cannot be with -f).\n\
  -u <interval>      Broadcast interval (in seconds).\n\
  -X <random_file>   Read random numbers from a file.\n\
", prog_name);

	exit(0);
}

int main(int argc, char *argv[])
{
	double sim_duration=0;
	double r_add_del=10; // 10 requests/second
	double r_lookup=100; // 100 requests/second
	char *trace_file=NULL;
	double trace_scale=0.1;

	double update_interval=1; // once per second.
	int ns=100; 
	int degree=2;
	int stb_obj=10000000;
	int max_obj=11000000;
	int init_obj=100000;
	char *random_file=NULL;
	double pkt_loss=0;
	int instant_dump=0;
	bool bloom_performance=false;

	prog_name=argv[0];

	int c;
	while ((c=getopt(argc, argv, "f:s:t:r:R:u:n:d:i:o:O:X:e:I:hk"))!=EOF) {
		switch (c) {
		case 'f' :
			if (trace_file) {
				tmpi_error(DBG_CALLER, "Trace already specified.\n");
				do_help();
				break;
			}
			trace_file=optarg;
			break;
		case 's' :
			trace_scale=atof(optarg);
			break;
		case 't' :
			sim_duration=atof(optarg);
			break;
		case 'r' :
			r_add_del=atof(optarg);
			break;
		case 'R' :
			r_lookup=atof(optarg);
			break;
		case 'u' :
			update_interval=atof(optarg);
			break;
		case 'n' :
			ns=atoi(optarg);
			break;
		case 'd' :
			degree=atoi(optarg);
			break;
		case 'i' :
			init_obj=atoi(optarg);
			break;
		case 'o' :
			stb_obj=atoi(optarg);
			break;
		case 'O' :
			max_obj=atoi(optarg);
			break;
		case 'X' :
			random_file=optarg;
			break;
		case 'e' :
			pkt_loss=atof(optarg);
			break;
		case 'I' :
			instant_dump=atoi(optarg);
			break;
		case 'k' :
			bloom_performance=true;
			break;
		case 'h' :
			do_help();
			break;
		default:
			do_help();
		}
	}

	if (random_file) {
		int random_fd=open(random_file, O_RDONLY);
		if (random_fd>=0) {
			Guid::set_random_fd(random_fd);
		}
		else {
			tmpi_error(DBG_CALLER, "Unable to open random file: %s. Use default instead.\n", random_file);
		}
	}

	if ((pkt_loss<0)||(pkt_loss>=1)) {
		tmpi_error(DBG_CALLER, "Packet loss rate (0-1) out of range: %.2f.\n", pkt_loss);
		do_help();
	}

	if ((!trace_file)&&(sim_duration<=0)) {
		tmpi_error(DBG_CALLER, "At least one of the -f or -t options must be specified.\n");
		do_help();
	}

	if ((trace_file)&&(sim_duration>0)) {
		tmpi_error(DBG_CALLER, "Only one of the -f or -t options can be specified.\n");
		do_help();
	}

	Simulator *sim=NULL;
	char log_name[1024];
	snprintf(log_name, 1024, "%s.%d.log", basename(strdup(argv[0])), getpid());
	FILE *fp=fopen(log_name, "a+");
	if (fp) {
		tmpi_set_output(fp);
	}
	else {
		fprintf(stderr, "Cannot open log file: %s.\n", log_name);
		exit(-1);
	}

	if (trace_file) {
		sim=new Simulator(trace_file, trace_scale, update_interval, ns, degree, stb_obj, max_obj, init_obj, pkt_loss, instant_dump);
		printf("Trace driven simulation:\n");
		printf("  trace_scale=%.4f.\n", trace_scale);
	}
	else {
		sim=new Simulator(sim_duration, update_interval, r_add_del, r_lookup, ns, degree, stb_obj, max_obj, init_obj, pkt_loss, instant_dump);
		printf("Synthetic simulation:\n");
		printf("  r_add_del=%.4f (req/sec).\n", r_add_del);
		printf("  r_lookup=%.4f (req/sec).\n", r_lookup);
	}

	if ((!sim)||(!sim->good())) {
		tmpi_error(DBG_INTERNAL, "Cannot initialize the Simulator object.\n");
		exit(-1);
	}

	if (bloom_performance) {
		double dummy;
		double t=sim->client->handle_kcheck(sim, dummy);
		printf("svr_num=%d, per_lookup=%.3fus.\n", ns, t);
		return 0;
	}

	printf("  simulation duration: %.2f sec.\n", sim->get_sim_time());
	printf("  update policy: %s.\n", CBS_BF::update_policy_string());
	printf("  update_interval=%.4f (sec).\n", update_interval);
	printf("  svr_num=%d.\n", ns);
	printf("  degree=%d.\n", degree);
	printf("  stb_obj=%d.\n", stb_obj);
	printf("  max_obj=%d.\n", max_obj);
	printf("  init_obj=%d.\n", init_obj);
	printf("  pkt_loss=%.4f.\n", pkt_loss);
	printf("  dump instant info=");
	if (instant_dump<=0) {
		printf("false");
	}
	else {
		printf("every %d seconds", instant_dump);
	}
	printf(".\n");

	{
		char cmd_line[1024];
		char *p=cmd_line;
		int blen=1024;
		for (int i=0; i<argc; i++) {
			int l=strlen(argv[i]);
			if (blen<l+5) {
				strcpy(p, "...");
				break;
			}
			snprintf(p, blen, "%s ", argv[i]);
			blen-=(l+1);
			p+=(l+1);
		}
		tmpi_error(DBG_APP, "command: %s\n", cmd_line);
	}

	sim->report();
	if (sim->start()==false) {
		tmpi_error(DBG_INTERNAL, "Cannot start simulator.\n");
		exit(-1);
	}

	sim->run();

	if (fp) {
		fclose(fp);
	}

	return 0;
}
