// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: micro_ben2.cc,v $ - micro benchmark.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Aziz Gulbeden, gulbeden@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/10/01 00:47:44 $ by $Author: gulbeden $
 * Version :  $Revision: 1.8 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

# include <getopt.h>
# include <stdio.h>
# include <signal.h>

# include "usc_time.h"
# include "sor_client.h"
# include "sor_fs.h"
# include "sor_file.h"
# include "subscriber.h"
# include "tmpi_debug.h"

using namespace std;

static char *program_name=NULL;
static Subscriber *sub=NULL;
static SorClient *scli=NULL;

/* range of files that will be generated */
int lower=0;
int upper=1;


static void usage(int stat){

	printf("\nMicrobenchmark 2:\nThis benchmark creates files. Parameters:\n  f: Configuration file.\n  l: Lower bound on the file names to be created.\n  u: Upper bound on the file names to be created (not inclusive).\nLast argument needs to the the deployment ID.\n");
	exit(stat);
}


static void parse_cmd_args_and_init(int argc, char *argv[])
{
	static struct option const long_options[] = {
		{"config", required_argument, 0, 'f'},
		{"mc-addr", required_argument, 0, 'm'},
		{"mc-port", required_argument, 0, 'p'},
		{"lower", required_argument, 0, 'l'},
		{"upper", required_argument, 0, 'u'},
		{"silent", no_argument, 0, 's'},
		{"debug", no_argument, 0, 'd'},
		{"help", no_argument, 0, 'h'},
		{"version", no_argument, 0, 'V'},
		{NULL, 0, 0, 0}
	};

	int optc;
	char *config=NULL;
	char *mc_addr=NULL;
	int mc_port=-1;
	char *ident=NULL;

	program_name=argv[0];

	while ((optc=getopt_long(argc, argv, \
					"f:m:p:l:u:sdhV", long_options, NULL)) !=-1) {
		switch (optc) {
		case 0:
			break;
		case 'f':
			if (config) {
				printf("Sorrento configuration file already specified.\n");
				usage(1);
			}
			config=optarg;
			break;
		case 'm':
			if (mc_addr) {
				printf("Sorrento multicast address already specified.\n");
				usage(1);
			}
			mc_addr=optarg;
			break;
		case 'p':
			if (mc_port>0) {
				printf("Sorrento multicast port already specified.\n");
				usage(1);
			}
			mc_port=atoi(optarg);
			if (mc_port<=0) {
				printf("invalid port.\n");
				usage(1);
			}
			break;
			
		case 'l':
			lower = atoi(optarg);
			break;
			
		case 'u':
			upper = atoi(optarg);
			break;
			
		case 's':
			tmpi_set_dbg_level(DBG_FATAL-1);
			tmpi_hide_location();
			break;
		case 'd':
			tmpi_set_dbg_level(DBG_ALL);
			tmpi_show_location();
			break;
		case 'V':
			//do_version();
			exit(0);
			break;
		case 'h':
			usage(0);
			break;
		default:
			//usage(1);
			usage(0);
			;
		}
	}

	if (optind != (argc-1)) {
		printf("No demployment ID specified.\n");
		usage(1);
	}

	ident=argv[optind];

	if ((config)&&(mc_addr)) {
		printf("Both configuration file and multicast address specified.\n");
		usage(0);
	}

	if ((!config)&&(!mc_addr)) {
		printf("Neither configuration file nor multicast address specified.\n");
		usage(0);
	}

	if (mc_addr && (mc_port<=0)) {
		printf("Multicast port not specified.\n");
	}

	try {
		if (config) {
			SorrentoConf sconf(config);
			sub = new Subscriber(sconf);
		}
		else {
			sub = new Subscriber(mc_addr, mc_port, 1000);
		}
	
		sub->start();
		printf("Starting the proxy, please wait ...\n");
		sleep(3); // wait for 3 seconds.
		SorFS sorfs(sub, string(ident));
		scli = new SorClient(&sorfs);
	}
	catch (const string &errmsg) {
		tmpi_error(DBG_INTERNAL, "sor_shell: %s\n", errmsg.c_str());
		goto _exit;
	}
	catch (int errnum) {
		tmpi_error(DBG_INTERNAL,"sor_shell: error caught (code=%d).\n",errnum);
		goto _exit;
	}
	catch (...) {
		tmpi_error(DBG_INTERNAL, "Unknown error, initialization failed.\n");
		goto _exit;
	}

	return;

_exit:
	if (sub)
		delete sub;
	exit(-1);
}

static void output_error(const char *cmd, bool do_output)
{
	if (do_output)
		printf("%s failed: %s.\n", cmd, scli->str_error().c_str());
}



/**
 *
 */
int main(int argc, char *argv[])
{
	signal(SIGPIPE, SIG_IGN);
	const int blockSize = 4*1024*1024;
	const int numBlocks = 128;
	char *buf;

	tmpi_set_dbg_level(DBG_ALL);
	tmpi_show_location();

	bool do_output=true;
	bool do_timing=true;

	char objName[16];

	int i;

	// might be no return.
	parse_cmd_args_and_init(argc, argv);

	usc_time_t t_start=0, t_end=0;
	int cmd_cnt=0;
	int err_cnt=0;

	SorFile *psf;

	off_seg_t written_len;

	buf = new char[blockSize];

	printf("Lower : %d upper : %d\n", lower, upper);

	/* Start measurement */
	if (do_timing) {
		t_start=usc_gettime();
	}

	/* create , write , close */

	for(i=lower; i<upper; i++){

		sprintf(objName, "%d.2", i);

		psf=scli->create(objName);

		if (psf == NULL) {
			printf("ERROR: Create (%s) : ", objName);
			output_error("Create", do_output);
			err_cnt++;
			continue;
		}

		for(int j=0; j<numBlocks; j++){
			if(scli->write(psf,j*blockSize,buf,blockSize,written_len)==false)
			{
				printf("ERROR: Write (%s) : ", objName);
				output_error("Write", do_output);
				err_cnt++;
			}
			cmd_cnt++;
		}

		/* close */
		if (scli->close(psf) == false) {
			printf("ERROR: close\n");
			output_error("Close", do_output);
				err_cnt++;
		}
		
	}

	/* end measurement */

	if (do_timing) {
		t_end=usc_gettime();
		printf("Total %d commands executed, execution time %.3fms.(with %d errors)\n", 
			   cmd_cnt, usc_difftime(t_end, t_start)/1000, err_cnt);
		fprintf(stderr, "%.3f\n", usc_difftime(t_end, t_start)/1000);
	}

	sub->stop();
	sub->finalize();
	delete sub;
	delete scli;
	return 0;
}
