// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: sprite.cc,v $ - converted Sprite File System trace parser.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2002/10/08 07:32:51 $
 * Version :  $Revision: 1.2 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

# include <netinet/in.h>
# include <zlib.h>
# include <time.h>
# include "tmpi_debug.h"
# include "sprite.h"
# include "sprite_struct.h"

SpriteRecord::SpriteRecord(void *buf, enum SpriteRecord::OP t) :
	type(t)
{
	if (buf) {
		BaseType *tbuf=(BaseType *)buf;
		tv.tv_sec=ntohl(tbuf->sec);
		tv.tv_usec=ntohl(tbuf->usec);
	}
}

// bad ID will have all the chars to be '\0';
bool SpriteRecord::is_good_id(IDType &id)
{
	bool good=false;

	for (unsigned int i=0; i<4; i++) {
		if (id.p[i]!=-1) {
			good=true;
			break;
		}
	}

	return good;
}

static void ntohIDType(IDType &src, IDType &dst)
{
	for (int i=0; i<4; i++)
		dst.p[i] = ntohl(src.p[i]);
}

static char *time_str(struct timeval &tv, char *buf, int sz)
{
	if ((!buf)||(sz<=0))
		return "<error>";

	time_t tt=tv.tv_sec;
	struct tm ttm;
	memcpy(&ttm, gmtime(&tt), sizeof(struct tm));

	buf[0]='\0';
	strftime(buf, sz, "%b %d %Y, %H:%M:%S", &ttm);

	int l=strlen(buf);
	if (l+5>sz) {
		return buf;
	}

	sprintf(buf+l, ".%03d", (int)((tv.tv_usec/1000)%1000));

	return buf;
}

static char *fid_str(IDType &id, char *buf, int sz)
{
	if ((!buf)||(sz<=0))
		return "<error>";

	/* snprintf(buf, 1024, "%08X%08X%08X%08X", id.p[0], id.p[1], id.p[2], id.p[3]); */
	snprintf(buf, sz, "[%X-%X-%X-%X]", id.p[0], id.p[1], id.p[2], id.p[3]);
	return buf;
}

class SpriteRecordOpen : public SpriteRecord {
 private:
	IDType fid;
	IDType sid;
	int mode; // ignored
	int size;

 public:
	SpriteRecordOpen(void *buf) :
		SpriteRecord(buf, SpriteRecord::OPEN)
	{
		OpenType *tbuf=(OpenType *)buf;

		ntohIDType(tbuf->fid, fid);
		ntohIDType(tbuf->sid, sid);
		mode=ntohl(tbuf->mode);
		size=ntohl(tbuf->size);
	}

	virtual bool get_fid(IDType &fid) {
		memcpy((char *)&fid, (char *)&(this->fid), sizeof(IDType)); 
		return true;
	}

	virtual bool get_sid(IDType &sid) {
		memcpy((char *)&sid, (char *)&(this->sid), sizeof(IDType)); 
		return true;
	}

	virtual bool get_size(int &size) {
		size=this->size;
		return true;
	}

	virtual void dump(FILE *fp=NULL) {
		char szTime[1024], szFid[1024], szSid[1024];
		if (!fp)
			fp=stdout;
		fprintf(fp, "<OPEN:%s> fid=%s, sid=%s, size=%d.\n", time_str(tv, szTime, 1024), 
				fid_str(fid, szFid, 1024), fid_str(sid, szSid, 1024), size); 
	}
};

class SpriteRecordUnlink : public SpriteRecord {
 private:
	IDType fid;
	int size; // should be ignored

 public:
	SpriteRecordUnlink(void *buf) :
		SpriteRecord(buf, SpriteRecord::UNLINK)
	{
		UnlinkType *tbuf=(UnlinkType *)buf;

		ntohIDType(tbuf->fid, fid);
		size=ntohl(tbuf->size);
	}

	virtual bool get_fid(IDType &fid) {
		memcpy((char *)&fid, (char *)&(this->fid), sizeof(IDType)); 
		return true;
	}

	virtual void dump(FILE *fp=NULL) {
		char szTime[1024], szFid[1024];
		if (!fp)
			fp=stdout;
		fprintf(fp, "<UNLINK:%s> fid=%s.\n", time_str(tv, szTime, 1024), fid_str(fid, szFid, 1024)); 
	}
};

class SpriteRecordCreate : public SpriteRecord {
 private:
	IDType fid;
	int size; // should be ignored

 public:
	SpriteRecordCreate(void *buf) :
		SpriteRecord(buf, SpriteRecord::CREATE)
	{
		CreateType *tbuf=(CreateType *)buf;

		ntohIDType(tbuf->fid, fid);
	}

	virtual bool get_fid(IDType &fid) {
		memcpy((char *)&fid, (char *)&(this->fid), sizeof(IDType)); 
		return true;
	}

	virtual void dump(FILE *fp=NULL) {
		char szTime[1024], szFid[1024];
		if (!fp)
			fp=stdout;
		fprintf(fp, "<CREATE:%s> fid=%s.\n", time_str(tv, szTime, 1024), fid_str(fid, szFid, 1024)); 
	}
};

// Note that a seek in fact could be read/write/seek

class SpriteRecordSeek : public SpriteRecord {
 private:
	IDType sid;
	int start;
	int end;
	int rw;

 public:
	SpriteRecordSeek(void *buf) :
		SpriteRecord(buf)
	{
		SeekType *tbuf=(SeekType *)buf;

		ntohIDType(tbuf->sid, sid);
		start=ntohl(tbuf->start);
		end=ntohl(tbuf->end);
		rw=ntohl(tbuf->rw);

		if ((rw&SEEK_READ)&&(rw&SEEK_WRITE)) {
			tmpi_error(DBG_CALLER, "Both SEEK_READ and SEEK_WRITE set.\n");
			type=READ;
			return; // conflict
		}

		type=SEEK;
		if (rw&SEEK_READ)
			type=READ;
		else if (rw&SEEK_WRITE)
			type=WRITE;
	}

	virtual bool get_sid(IDType &sid) {
		if (!is_good_op(type))
			return false;

		memcpy((char *)&sid, (char *)&(this->sid), sizeof(IDType)); 
		return true;
	}

	virtual bool get_offset(int &offset) {
		if (!is_good_op(type))
			return false;

		if ((type==READ)||(type==WRITE)) {
			if (start<end)
				offset=start;
			else
				offset=end;
		}
		else {
			// seek
			offset=end;
		}

		return true;
	}

	virtual bool get_size(int &size) {
		if (!is_good_op(type))
			return false;

		if ((type==READ)||(type==WRITE)) {
			if (end<start)
				size=(start-end);
			else
				size=(end-start);
			return true;
		}

		return false;
	}

	virtual void dump(FILE *fp=NULL) {
		if (!fp)
			fp=stdout;

		if (!is_good_op(type))
			return;

		switch (type) {
			case SEEK : 
				{
					int offset;
					get_offset(offset);
					char szTime[1024], szSid[1024];
					fprintf(fp, "<SEEK:%s> sid=%s, offset=%d.\n", 
							time_str(tv, szTime, 1024), fid_str(sid, szSid, 1024), offset); 
					break;
				}
			case READ :
				{
					int offset;
					int size;
					get_offset(offset);
					get_size(size);
					char szTime[1024], szSid[1024];
					fprintf(fp, "<READ:%s> sid=%s, offset=%d, size=%d.\n", 
							time_str(tv, szTime, 1024), fid_str(sid, szSid, 1024), offset, size); 
					break;
				}
			case WRITE :
				{
					int offset;
					int size;
					get_offset(offset);
					get_size(size);
					char szTime[1024], szSid[1024];
					fprintf(fp, "<WRITE:%s> sid=%s, offset=%d, size=%d.\n", 
							time_str(tv, szTime, 1024), fid_str(sid, szSid, 1024), offset, size); 
					break;
				}
			default:
				tmpi_error(DBG_INTERNAL, "Unknown type in Seek.dump().\n");
				break;
		}
	}
};

class SpriteRecordClose : public SpriteRecord {
 private:
	IDType sid;
	int size; // should be ignored

 public:
	SpriteRecordClose(void *buf) :
		SpriteRecord(buf, SpriteRecord::CLOSE)
	{
		CloseType *tbuf=(CloseType *)buf;

		ntohIDType(tbuf->sid, sid);
		size=ntohl(tbuf->size);
	}

	virtual bool get_sid(IDType &sid) {
		memcpy((char *)&sid, (char *)&(this->sid), sizeof(IDType)); 
		return true;
	}

	virtual void dump(FILE *fp=NULL) {
		if (!fp)
			fp=stdout;
		char szTime[1024], szSid[1024];
		fprintf(fp, "<CLOSE:%s> sid=%s.\n", time_str(tv, szTime, 1024), fid_str(sid, szSid, 1024)); 
	}
};

class SpriteRecordTruncate : public SpriteRecord {
 private:
	IDType sid;
	int oldLen; // ignored
	int newLen;

 public:
	SpriteRecordTruncate(void *buf) :
		SpriteRecord(buf, SpriteRecord::TRUNCATE)
	{
		TruncateType *tbuf=(TruncateType *)buf;

		ntohIDType(tbuf->sid, sid);
		oldLen=ntohl(tbuf->oldLen);
		newLen=ntohl(tbuf->newLen);
	}

	virtual bool get_sid(IDType &sid) {
		memcpy((char *)&sid, (char *)&(this->sid), sizeof(IDType)); 
		return true;
	}

	virtual bool get_size(int &size) {
		size=newLen;
		return true;
	}

	virtual void dump(FILE *fp=NULL) {
		if (!fp)
			fp=stdout;
		int size;
		get_size(size);
		char szTime[1024], szSid[1024];
		fprintf(fp, "<TRUNCATE:%s> sid=%s, size=%d.\n", 
				time_str(tv, szTime, 1024), fid_str(sid, szSid, 1024), size); 
	}
};

class SpriteRecordReadWrite : public SpriteRecord {
 private:
	IDType sid;
	int offset;
	int bytes; // should be ignored

 public:
	SpriteRecordReadWrite(void *buf, bool isRead) :
		SpriteRecord(buf)
	{
		ReadWriteType *tbuf=(ReadWriteType *)buf;

		ntohIDType(tbuf->sid, sid);
		offset=ntohl(tbuf->offset);
		bytes=ntohl(tbuf->bytes);

		if (isRead) {
			type=READ;
		}
		else {
			type=WRITE;
		}
	}

	virtual bool get_sid(IDType &sid) {
		memcpy((char *)&sid, (char *)&(this->sid), sizeof(IDType)); 
		return true;
	}

	virtual bool get_offset(int &offset) {
		offset=this->offset;
		return true;
	}

	virtual bool get_size(int &size) {
		size=bytes;
		return true;
	}

	virtual void dump(FILE *fp=NULL) {
		if (!fp)
			fp=stdout;
		int size;
		get_size(size);
		char szTime[1024], szSid[1024];
		char *op_str=(char *)((type==READ)?"READ":"WRITE");
		fprintf(fp, "<%s:%s> sid=%s, offset=%d, size=%d.\n", 
				op_str, time_str(tv, szTime, 1024), fid_str(sid, szSid, 1024), offset, size); 
	}
};

void SpriteParser::free_record(SpriteRecord *&r)
{
	if (r) {
		delete r;
		r=NULL;
	}
}

SpriteRecord *SpriteParser::read_record()
{
	if (!fp)
		return NULL;

	int type;
	union {
		OpenType i_open;
		UnlinkType i_delete;
		CreateType i_create;
		SeekType i_seek;
		CloseType i_close;
		TruncateType i_truncate;
		ReadWriteType i_read;
	} item;

	if (gzread(fp, &type, sizeof(int))!=sizeof(int)) {
		return NULL;
	}

	type=htonl(type);

	SpriteRecord *ret=NULL;

	switch (type) {
		case SPT_OPEN:
			if (gzread(fp, &(item.i_open), sizeof(item.i_open))==sizeof(item.i_open)) {
				ret=new SpriteRecordOpen(&(item.i_open));
			}
			break;
		case SPT_DELETE:
			if (gzread(fp, &(item.i_delete), sizeof(item.i_delete))==sizeof(item.i_delete)) {
				ret=new SpriteRecordUnlink(&(item.i_delete));
			}
			break;
		case SPT_CREATE:
			if (gzread(fp, &(item.i_create), sizeof(item.i_create))==sizeof(item.i_create)) {
				ret=new SpriteRecordCreate(&(item.i_create));
			}
			break;
		case SPT_SEEK:
			if (gzread(fp, &(item.i_seek), sizeof(item.i_seek))==sizeof(item.i_seek)) {
				ret=new SpriteRecordSeek(&(item.i_seek));
			}
			break;
		case SPT_CLOSE:
			if (gzread(fp, &(item.i_close), sizeof(item.i_close))==sizeof(item.i_close)) {
				ret=new SpriteRecordClose(&(item.i_close));
			}
			break;
		case SPT_TRUNCATE:
			if (gzread(fp, &(item.i_truncate), sizeof(item.i_truncate))==sizeof(item.i_truncate)) {
				ret=new SpriteRecordTruncate(&(item.i_truncate));
			}
			break;
		case SPT_READ:
			if (gzread(fp, &(item.i_read), sizeof(item.i_read))==sizeof(item.i_read)) {
				ret=new SpriteRecordReadWrite(&(item.i_read), true);
			}
			break;
		case SPT_WRITE:
			if (gzread(fp, &(item.i_read), sizeof(item.i_read))==sizeof(item.i_read)) {
				ret=new SpriteRecordReadWrite(&(item.i_read), false);
			}
			break;
		default:
			break;
	}

	if (ret) {
		if (first_record) {
			if (ret->get_time(start_tv))
				first_record=false;
		}
		ret->get_time(last_tv);
		counters[ret->get_op()]++;
		nrec++;
	}

	return ret;
}

void SpriteParser::report(FILE *fp)
{
	if (fp==NULL)
		fp=stdout;

	char szStart[1024], szEnd[1024];
	if (first_record)
		return;

	time_str(start_tv, szStart, 1024);
	time_str(last_tv, szEnd, 1024);

	fprintf(fp, "Trace start @ %s, end @ %s, total %d records.\n", szStart, szEnd, nrec);
	fprintf(fp, "Record type break down: ");
	for (int i=SpriteRecord::OP_START+1; i<SpriteRecord::OP_END; i++) {
		fprintf(fp, "<%s, %d> ", SpriteRecord::op_str(i), counters[i]);
	}
	fprintf(fp, "\n");
}
