// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: trace_loc.cc,v $ - This program extracts all the object location 
 *            related events and write to the disk in the following 
 *            format: <time> <event_type>.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/05/06 09:19:22 $
 * Version :  $Revision: 1.3 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 
# include <stdio.h>
# include <stdlib.h>
# include <getopt.h>
# include <queue>
# include <set>

# include "sprite.h"
# include "tmpi_debug.h"

/**
 */

using namespace std;

static char *prog=NULL;

static void do_help()
{
	printf("\
Sprite Trace Object Location Event Extractor.\n\
Usage: %s [options] [trace list ...].\n\
   -f <trace>             Trace file.\n\
   -o <output>            Output file.\n\
   -h                     Help, this screen.\n\
\n\
Multiple trace files can be specified by the command \n\
line, and the analyzer will generate output trace for\n\
all of them (traces will be separated by a line of\n\
statistics;\n\
#NAME=%%s#ADD=%%d#DEL=%%d#LOOKUP=%%d#TOTAL_TIME=%%.3fsec#\n\
", prog);
	exit(0);
}

struct IDType_cmp : public less<IDType> {
	bool operator()(const IDType &x, const IDType &y) const {
		return(memcmp(x.p+1, y.p+1, sizeof(int)*3)<0);
	}
};

typedef set<IDType, IDType_cmp> ID_SET;

struct TraceStat {
	FILE *ofp;
	unsigned int n_add;
	unsigned int n_del;
	unsigned int n_lookup;
	double total_time;
	ID_SET sid_set;
	ID_SET fid_set;

public:
	TraceStat(FILE *output) :
		ofp(output),
		n_add(0),
		n_del(0),
		n_lookup(0),
		total_time(0)
	{ }

	void report(char *caption, FILE *fp=NULL)
	{
		if (!fp)
			fp=ofp;

		fprintf(fp, "#NAME=%s#ADD=%d#DEL=%d#LOOKUP=%d#TOTAL_TIME=%.3fsec#\n", 
				caption, n_add, n_del, n_lookup, total_time); 
	}

	void handle_record(double &start_time, SpriteRecord *r);

	double tv_to_double(struct timeval &tv) {
		return (double)tv.tv_sec+(double)tv.tv_usec/1000000.0;
	}
};

void TraceStat::handle_record(double &first_rec_time, SpriteRecord *r)
{
	int op=r->get_op();

	if (first_rec_time==0) {
		struct timeval tv;
		if (r->get_time(tv))
			first_rec_time=tv_to_double(tv);
	}

	switch (op) {
		case SpriteRecord::OPEN:
			{
				IDType fid;
				IDType sid;
				struct timeval tv;
				bool ret=true;
				ret = ret && (r->get_fid(fid));
				ret = ret && (r->get_sid(sid));
				ret = ret && (r->get_time(tv));

				if (!ret) {
					tmpi_error(DBG_APP, "Bad OPEN record passed from parser.\n");
					break;
				}

				ID_SET::iterator pi=fid_set.find(fid);
				if (pi==fid_set.end()) {
					fid_set.insert(fid);
					double tt=tv_to_double(tv);
					fprintf(ofp, "%.3f\t%.3f\tADD\n", tt, tt-first_rec_time);
					n_add++;
				}
				else {
					double tt=tv_to_double(tv);
					fprintf(ofp, "%.3f\t%.3f\tLOOKUP\n", tt, tt-first_rec_time);
					n_lookup++;
				}
				
				pi=sid_set.find(sid);
				if (pi==sid_set.end()) {
					sid_set.insert(sid);
				}
			}
			break;
		case SpriteRecord::CLOSE:
			{
				IDType sid;
				if (!r->get_sid(sid)) {
					tmpi_error(DBG_APP, "Bad CLOSE record passed from parser.\n");
					break;
				}

				ID_SET::iterator pi=sid_set.find(sid);
				if (pi!=sid_set.end()) {
					sid_set.erase(pi);
				}
			}
			break;
		case SpriteRecord::CREATE:
			{
				IDType fid;
				struct timeval tv;
				bool ret=true;

				ret = ret && r->get_fid(fid);
				ret = ret && r->get_time(tv);

				if (!ret) {
					tmpi_error(DBG_APP, "Bad CREATE record passed from parser.\n");
					break;
				}

				ID_SET::iterator pi=fid_set.find(fid);
				if (pi==fid_set.end()) {
					fid_set.insert(fid);
					double tt=tv_to_double(tv);
					fprintf(ofp, "%.3f\t%.3f\tADD\n", tt, tt-first_rec_time);
					n_add++;
				}
			}
			break;
		case SpriteRecord::UNLINK:
			{
				IDType fid;
				struct timeval tv;
				bool ret=true;

				ret = ret && r->get_fid(fid);
				ret = ret && r->get_time(tv);
				if (!ret) {
					tmpi_error(DBG_APP, "Bad UNLINK record passed from parser.\n");
					break;
				}

				ID_SET::iterator pi=fid_set.find(fid);
				if (pi!=fid_set.end()) {
					fid_set.erase(pi);
					double tt=tv_to_double(tv);
					fprintf(ofp, "%.3f\t%.3f\tDEL\n", tt, tt-first_rec_time);
					n_del++;
				}
			}
			break;
		case SpriteRecord::READ:
		case SpriteRecord::WRITE:
		case SpriteRecord::SEEK:
		case SpriteRecord::TRUNCATE:
			{
				bool ret=true;
				IDType sid;
				ret = ret && (r->get_sid(sid));

				if (!ret) {
					tmpi_error(DBG_APP, "Bad READ/WRITE record passed from parser.\n");
					break;
				}

				ID_SET::iterator pi=sid_set.find(sid);
				if (pi==sid_set.end()) {
					sid_set.insert(sid);
				}
			}
			break;
		default:
			tmpi_error(DBG_APP, "Un-handled record: op=%d.\n", op);
			r->dump();
			break;
	}
}

int main(int argc, char *argv[])
{
	prog=argv[0];
	char *ofp_name=NULL;
	FILE *default_ofp=stdout;

	queue<char *>trace_queue;

	int c;
	while ((c=getopt(argc, argv, "f:o:h"))!=EOF) {
		switch (c) {
			case 'f':
				trace_queue.push(optarg);
				break;
			case 'o':
				if (ofp_name) {
					tmpi_error(DBG_CALLER, "Output file name already specified.\n");
					do_help();
				}
				ofp_name=optarg;
				break;
			case 'h':
				do_help();
				break;
			default:
				do_help();
				break;
		}
	}

	for (; optind<argc; optind++) {
		trace_queue.push(argv[optind]);
	}

	if (trace_queue.empty()) {
		tmpi_error(DBG_VERBOSE, "No trace file specified in the command.\n");
		do_help();
	}

	FILE *ofp=default_ofp;
	if (ofp_name) {
		ofp=fopen(ofp_name, "w");
		if (!ofp) {
			tmpi_error(DBG_VERBOSE, "Cannot open output file.\n");
			do_help();
		}
	}

	double first_rec_time=0;
	while (!trace_queue.empty()) {
		char *trace=trace_queue.front();
		TraceStat per_trace(ofp);
		trace_queue.pop();
		SpriteParser parser(trace);
		SpriteRecord *r;
		first_rec_time=0;
		while ((r=parser.read_record())) {
			per_trace.handle_record(first_rec_time, r);
			parser.free_record(r);
		}
		double tt=0;
		if (!parser.total_time(tt)) {
			tmpi_error(DBG_INTERNAL, "Cannot get trace interval time.\n");
		}
		else {
			per_trace.total_time=tt;
		}
		char *p=strrchr(trace, '/');
		if (p)
			p++;
		else
			p=trace;
		per_trace.report(p);
	}

	if (ofp!=default_ofp) {
		fclose(ofp);
	}

	return 0;
}
