/**********************************************************************************
Copyright (c) 2003 The Regents of the University of California. 
All rights reserved.

Sorrento project contact: Hong Tang (hongt@cs.ucsb.edu)
			  Lingkun Chu (lkchu@cs.ucsb.edu),
                          Tao Yang (tyang@cs.ucsb.edu),
Web site: http://www.cs.ucsb.edu/projects/sorrento
Department of Computer Science, UC Santa Barbara.

This software is distributed to individuals for personal non-commercial 
use and to non-profit entities for non-commercial purposes only. Permission 
is also granted to commercial enterprises to use this software for evaluation 
purposes only. It is licensed on a non-exclusive basis, free of charge for 
these uses. The license is subject to the following conditions:

1. Redistribution is not permitted. UC Santa Barbara will maintain a copy 
   of this software as a directly downloadable file, available under the terms 
   of license specified in this agreement.
2. All advertising materials mentioning features or use of this software must 
   display the following acknowledgment: "This product includes software 
   developed by the University of California, Santa Barbara." 
3. Neither the name of the University nor the names of its contributors may 
   be used to endorse or promote products derived from this software without 
   specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*********************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <string.h>
#include <errno.h>
#include <getopt.h>
#include <pthread.h>
#include "config.h"
#include "argdef.h"
#include "indextable.h"
#include "freespacelist.h"

extern "C"
{
#include "socket_opt.h"
}

#include "sor_client.h"
#include "sor_fs.h"
#include "subscriber.h"
#include "sor_file.h"

typedef struct
{
	char url[URLLEN];
	int seqnum;
	int datalen;
	char data[BLOCKSIZE];
}One_Block;

typedef struct
{
	char filename[MAXFILENAMLEN];
	char feature[MAXFEATURELEN];
	int valid;
}Single_File;

class Webcache_Service;
typedef struct
{
	int sktid;
	Webcache_Service* ws;
	char cmd[10];
}Arg_List;

class Webcache_Service
{
	public:
	Single_File file_list[MAXFILENUMBER];
	int file_num;

	Subscriber *sub;
	SorClient *sclt;
	SorFS *sorfs;
	
	Index_Table index_table;
	char indextable_filename[MAXFILENAMLEN];
	Freespace_List free_list[MAXFILENUMBER];
	char freelist_filename[MAXFILENAMLEN];

	pthread_mutex_t tlock;
	pthread_mutex_t llock;

	public:

	Webcache_Service(){};
	~Webcache_Service(){};
	
	/* return value
 	* -2: fileindex error, the data file which saves the webpages not found;
 	* -1: argument error, e.g., url=NULL, buflen=0;
 	* 0: file already exist, or none is written;
 	* >=1: succeed, the number of blocks written;
 	*/
	int Save_toFile(int fileindex, const char* url, const char* filebuf, int buflen);

	/* return value
	 * -2: fileindex error, the data file which saves the webpages not found;
	 * -1: argument error, e.g. url=NULL;
	 * 0: file not found, or none is read;
	 * 1: succeed;
	 */
	int Get_fromFile(char* url, char** filebuf, int* buflen);

	/* return value
	 * -1: argument error, e.g. url = NULL;
	 * 0: file not found;
	 * 1: succeed;
	 */
	int Delete_fromFile(char* url);
	
	/* return value
	 * -1: error;
	 * 1: succeed;
	 */
	int Init_Webcache(char* configfile);
	
	void Final_Webcache();
	
	/* Transform an url to the corresponding index of the data file
	 * in which the web page should be saved
	 */
	int Get_Fileindex(char* url);

	/* the following methods are implementations of the protocol between
	 * cache server and clients
	 */
	void Webcache_Put(int sktid);
	void Webcache_Get(int sktid);
	void Webcache_Del(int sktid);

	/* save the index table and the free lists
	 */
	void Save_Structure();
};


void Webcache_Service::Save_Structure()
{
	if (USELOCK)
		pthread_mutex_lock(&tlock);
	if (index_table.Ifmodified())
		index_table.Write_toFile(indextable_filename);
	if (USELOCK)
		pthread_mutex_unlock(&tlock);

	if (USELOCK)
		pthread_mutex_lock(&llock);
	for (int i=0; i<file_num; i++)
	{
		if (free_list[i].Ifmodified())
		{
			char freelistfilename[MAXFILENAMLEN];
			sprintf(freelistfilename, "%s%d", freelist_filename, i);
			free_list[i].Write_toFile(freelistfilename);
		}
	}
	if (USELOCK)
		pthread_mutex_unlock(&llock);
}

int Webcache_Service::Get_Fileindex(char* url)
{
	char ft[MAXFEATURELEN];
	strncpy(ft, url, 1);
	ft[1] = '\0';
	for(int i=0; i<file_num; i++)
	{
		if (strcmp(ft, file_list[i].feature) == 0)
			return i;
	}
	return -1;
}

int Webcache_Service::Delete_fromFile(char* url)
{
	if (url == NULL)
		return -1;

	if (USELOCK)
		pthread_mutex_lock(&tlock);
	//locate the file in the webpage data file
	File_Position fpos;
	if (index_table.Retrieve_File(url, &fpos) == 0)
	{
		if (USELOCK)
			pthread_mutex_unlock(&tlock);
		return 0;
	}
	//delete the file from the index table
	index_table.Delete_File(url);
	if (USELOCK)
		pthread_mutex_unlock(&tlock);
	
	if (USELOCK)
		pthread_mutex_lock(&llock);
	//return back the space;
	free_list[fpos.fileno].Return_Freespace(fpos.beginblock, fpos.endblock-fpos.beginblock+1);
	if (USELOCK)
		pthread_mutex_unlock(&llock);
	return 1;	
}

int Webcache_Service::Get_fromFile(char* url, char** filebuf, int* buflen)
{
	if (url == NULL)
		return -1;

	if (Get_Fileindex(url) < 0)
		return -2;	
	//locate the file in the webpage data file
	File_Position fpos;
	
	if (USELOCK)
		pthread_mutex_lock(&tlock);
	if (index_table.Retrieve_File(url, &fpos) == 0)
	{
		if (USELOCK)
			pthread_mutex_unlock(&tlock);
		return 0;
	}
	if (USELOCK)
		pthread_mutex_unlock(&tlock);
	
	if ((fpos.fileno >= file_num) || (!file_list[fpos.fileno].valid))
	{
		return -2;
	}
	int blocknum = fpos.endblock - fpos.beginblock + 1;
	
	SorFile* sfilehandle;
	FILE* filehandle;
	if (USESORRENTO)
	{
		//open the data file
		sfilehandle = sclt->open(file_list[fpos.fileno].filename, SOR_RD_ONLY);
		if (sfilehandle == NULL)
		{
			return -2;
		}
	}
	else
	{
		filehandle = fopen(file_list[fpos.fileno].filename, "r");
		if (filehandle == NULL)
		{
			return -2;
		}
	}
	//read the webpage from the data file
	(*filebuf) = (char *) malloc(BLOCKSIZE*blocknum);
	(*buflen) = 0;
	One_Block block;
	char* index = (*filebuf);
	for (int i=0; i<blocknum; i++)
	{
		if (USESORRENTO)
		{
			off_seg_t readbyte;
			bool rresult = sclt->read(sfilehandle, (off64_t)((fpos.beginblock+i)*sizeof(One_Block)), (char *)&block, (off_seg_t)sizeof(One_Block), readbyte);
			if ((rresult != true) || (readbyte != sizeof(One_Block)))
			{
				fprintf(stderr, "Read file error!\n");
				break;
			}
		}
		else
		{
			if (fseek(filehandle, (fpos.beginblock+i)*sizeof(One_Block), SEEK_SET) != 0)
			{
				fprintf(stderr, "Seek file error!\n");
				break;	
			}
			if (fread(&block, sizeof(One_Block), 1, filehandle) != 1)
			{
				fprintf(stderr, "Read file error!\n");
				break;
			}
			
		}
		if ((block.datalen > BLOCKSIZE) || (block.datalen <= 0))
		{
			if (i == (blocknum - 1))
				block.datalen = 0;
			else
				block.datalen = BLOCKSIZE;
		}
		strncpy(index, block.data, block.datalen);
		index += block.datalen;
		(*buflen) += block.datalen;
	}
	//close the data file
	if (USESORRENTO)
		sclt->close(sfilehandle);
	else
		fclose(filehandle);
	if (*buflen <= 0)
	{
		free(*filebuf);
		return 0;
	}
	else
	{
		return 1;
	}
}

int Webcache_Service::Save_toFile(int fileindex, const char* url, const char* filebuf, int buflen)
{
	if ((fileindex >= file_num) || (!file_list[fileindex].valid))
		return	-2;
	if ((url == NULL) || (buflen <= 0))
		return -1;
	if (USELOCK)
		pthread_mutex_lock(&tlock);
	if (index_table.Locate_File((char *)url) != NULL)
	{
		if (USELOCK)
			pthread_mutex_unlock(&tlock);
		return 0;
	}
	if (USELOCK)
		pthread_mutex_unlock(&tlock);
	
	//open the data file
	SorFile* sfilehandle;
	FILE* filehandle;
	if (USESORRENTO)
	{
		sfilehandle = sclt->open(file_list[fileindex].filename, 0);
		if (sfilehandle == NULL)
		{
			return -2;
		}
	}
	else
	{
		filehandle = fopen(file_list[fileindex].filename, "r+");
		if (filehandle == NULL)
		{
			return -2;
		}
	}
	//get the space needed
	int blocknum;
	if (buflen % BLOCKSIZE != 0)
		blocknum = (buflen / BLOCKSIZE) + 1;
	else
		blocknum = buflen / BLOCKSIZE;
	if (USELOCK)
		pthread_mutex_lock(&llock);
	size_t beginblock = free_list[fileindex].Get_Freespace(blocknum);
	if (USELOCK)
		pthread_mutex_unlock(&llock);
	
	//write to the data file
	One_Block block;
	int i;
	strncpy(block.url, url, URLLEN);
	
	for (i=0; i<blocknum; i++)
	{
		if (  i < (blocknum - 1))
		{
			block.seqnum = i;
			block.datalen = BLOCKSIZE;
			strncpy(block.data, filebuf+i*BLOCKSIZE, BLOCKSIZE);
		}
		else
		{
			block.seqnum = -1;
			block.datalen = buflen - i*BLOCKSIZE;
			strncpy(block.data, filebuf+i*BLOCKSIZE, block.datalen);
		}
		if (USESORRENTO)
		{
			off_seg_t writebyte;
			bool wresult = sclt->write(sfilehandle, (off64_t)((beginblock+i)*sizeof(One_Block)), (char*)&block, (off_seg_t)sizeof(One_Block), writebyte);
			if ((wresult != true) || (writebyte != sizeof(One_Block)))
			{
				fprintf(stderr, "write file error! exit...\n");
				break;
			}
		}
		else
		{
			if (fseek(filehandle, (long)(beginblock+i)*sizeof(One_Block), SEEK_SET)!=0)
				break;
			if (fwrite(&block, sizeof(One_Block), 1, filehandle) != 1)
				break;
		}
	}
	//close the data file;
	if (USESORRENTO)
		sclt->close(sfilehandle);
	else
	{
		fflush(filehandle);			
		fclose(filehandle);
	}

	//return back the unused free space
	if (i != blocknum)
	{
		if (USELOCK)
			pthread_mutex_lock(&llock);
		free_list[fileindex].Return_Freespace(beginblock+i, blocknum-i);
		if (USELOCK)
			pthread_mutex_unlock(&llock);
	}
	
	//update the index table
	if (i > 0)
	{
		if (USELOCK)
			pthread_mutex_lock(&tlock);
		index_table.Add_File((char *)url, fileindex, beginblock, beginblock+i-1);
		if (USELOCK)
			pthread_mutex_unlock(&tlock);
	}
	
	return i;
}

int Webcache_Service::Init_Webcache(char* configfile)
{
    // Initialize the service data

	sub = NULL;
	sclt = NULL;
	sorfs = NULL;
	if (USESORRENTO)
	{
		try
		{
			SorrentoConf sconf(configfile);
			sub = new Subscriber(sconf);
			sub->start();
			fprintf(stderr, "Sorrento client starting up ");
			fflush(stderr);
			bool done = false;
			int max_retry = 10;

			while ((!done) && (max_retry -- ))
			{
				try
				{
					sleep(3);
					fprintf(stderr, ".");
					fflush(stderr);
					sorfs = new SorFS(sub, std::string(sconf.ident()));
					fprintf(stderr, ".");
					fflush(stderr);
					sclt = new SorClient(sorfs);
					fprintf(stderr, ".");
					fflush(stderr);
					done = true;
				}
				catch (const std::string &errmsg)
				{
					if (sorfs)
					{
						delete sorfs;
						sorfs = NULL;
					}
					if (sclt)
					{
						delete sclt;
						sclt = NULL;
					}
				}
			}
			fprintf(stderr, "\n");
			if (sclt == NULL)
			{
				fprintf(stderr, "Cannot locate the specified service (%s).\n", sconf.ident());
				return -1;
			}
		}
		catch(const std::string &errmsg)
		{
			fprintf(stderr, "%s\n", errmsg.c_str()); 
			return -1;
		}
		catch (int errnum)
		{	
			fprintf(stderr, "error caught (code=%d).\n", errnum); 
			return -1;
		}
		catch(...)
		{
			fprintf(stderr, "Unknown error!\n"); 
			return -1;
		}
	}

	if (USELOCK)
	{
		pthread_mutex_init(&tlock, NULL);
		pthread_mutex_init(&llock, NULL);
	}

	char conf_file[MAXFILENAMLEN];
	sprintf(conf_file, "%scache_config", CONFIGPATH);
	char hostname[MAXHOSTNAMELEN];
	if (gethostname(hostname, MAXHOSTNAMELEN) != 0)
	{
		fprintf(stderr, "Cannot get the host name!\n");
		return -1;
	}
	Config conf(conf_file, hostname);
	
	if (conf.GetFilenum() <= 0)
	{
		fprintf(stderr, "Cache File not specified in Conf\n"); 
		return -1;
	}
	file_num = conf.GetFilenum();
	strncpy(indextable_filename, conf.GetIndexname(), MAXFILENAMLEN);
	strncpy(freelist_filename, conf.GetFreename(), MAXFILENAMLEN);
	
	char * fn;
	fn = conf.GetFilename();
	for(int i=0; i<file_num ; i++)
	{
		strncpy(file_list[i].filename, fn+i*MAXFILENAMLEN, MAXFILENAMLEN);
		char* end;
		char* begin;
		end = file_list[i].filename + strlen(file_list[i].filename);
		begin = end - 1;
		while ((begin != file_list[i].filename) && (begin[0] != '/'))
		{
			if (begin[0] == '.')
				end = begin;
			begin--;
		}
		if (begin[0] == '/')
		{
			strncpy(file_list[i].feature, begin+1, MAXFEATURELEN);
			file_list[i].feature[end-begin-1] = '\0';
		}
		else
		{
			strncpy(file_list[i].feature, begin, MAXFEATURELEN);
			file_list[i].feature[end-begin] = '\0';
		}
		
		SorFile* sfilehandle;
		FILE* filehandle;
		if (USESORRENTO)
		{
		  sfilehandle = sclt->open(file_list[i].filename, SOR_RD_ONLY);
		  if (sfilehandle == NULL)
			{
				fprintf(stderr, "Data File %s not exist, create it...\n", file_list[i].filename);
				sfilehandle = sclt->create(file_list[i].filename);
				if (sfilehandle == NULL)
				{
					fprintf(stderr, "Create data file %s failed, exit...\n", file_list[i].filename);
					return -1;
				}
				else
					sclt->close(sfilehandle);
			}
			else
			{
				sclt->close(sfilehandle);
				sfilehandle = NULL;
			}
		}
		else
		{
			filehandle = fopen(file_list[i].filename, "r");
			if (filehandle == NULL)
			{
				fprintf(stderr, "Data File %s not exist, create it...\n", file_list[i].filename);
				fclose(fopen(file_list[i].filename, "w"));
			}
			else
			{
				fclose(filehandle);
				filehandle = NULL;
			}
		}
		file_list[i].valid = 1;
	}
	
	index_table.Initialize(indextable_filename);
	for(int i=0; i<file_num; i++)
	{
		char freelistfilename[MAXFILENAMLEN];
		sprintf(freelistfilename, "%s%d", freelist_filename, i);
		free_list[i].Initialize(freelistfilename);
	}

	return 1;
}

void Webcache_Service::Final_Webcache()
{
	if (USELOCK)
		pthread_mutex_lock(&tlock);
	index_table.Finalize(indextable_filename);
	if (USELOCK)
		pthread_mutex_unlock(&tlock);
	if (USELOCK)
		pthread_mutex_lock(&llock);
	for(int i=0; i<file_num; i++)
	{
		char freelistfilename[MAXFILENAMLEN];
		sprintf(freelistfilename, "%s%d", freelist_filename, i);
		free_list[i].Finalize(freelistfilename);
	}
	if (USELOCK)
		pthread_mutex_unlock(&llock);

	if (USESORRENTO)
	{
		if (sub) 
		{
	    	sub->stop();
	    	sub->finalize();
			delete sub;
    	}
    	if (sorfs) 
        	delete sorfs;
    	if (sclt) 
        	delete sclt;
	}
}

void Webcache_Service::Webcache_Put(int sktid)
{
	char* url;
	char* filebuf;
	int buflen;
	int status;
	int ret;
	char* putnumstr;
	int putnum;
	int fileindex;

	ret = read_msg(sktid, &putnumstr);
	if (ret <= 0)
	{
		fprintf(stderr, "Socket error! Webpage Put aborted...\n");
		return;
	}
	putnum = atoi(putnumstr);
	free(putnumstr);

	for(int i=0; i<putnum; i++)
	{
		ret = read_msg(sktid, &url);
		if (ret <= 0)
		{
			fprintf(stderr, "Socket error! Webpage Put aborted...\n");
			break;
		}
		buflen = read_msg(sktid, &filebuf);
		if (buflen <= 0)
		{
			fprintf(stderr, "Socket error! Webpage Put aborted...\n");
			free(url);
			break;
		}
		
		fileindex = Get_Fileindex(url);
		if (fileindex < 0)
			status = -1;
		else
			status = Save_toFile(fileindex, url, filebuf, buflen);
	
		if (status < 0)
		{
			ret = write_msg(sktid, "ERROR", 5);
			fprintf(stderr, "Put url=%s ERROR...\n", url);
		}
		if (status == 0)
		{
			ret = write_msg(sktid, "EXIST", 5);
			fprintf(stderr, "Put url=%s EXIST...\n", url);
		}
		if (status > 0)
		{
			ret = write_msg(sktid, "OK", 2);
			fprintf(stderr, "Put url=%s Succeeds...\n", url);
		}
		free(url);
		free(filebuf);
		if (ret <= 0)
		{
			fprintf(stderr, "Socket error! Webpage Put aborted...\n");
			break;
		}
	}
	Save_Structure();
	return;
}

void Webcache_Service::Webcache_Get(int sktid)
{
	char* filebuf;
	int buflen;
	char* url;
	int status;
	int ret;
	char* getnumstr;
	int getnum;

	ret = read_msg(sktid, &getnumstr);
	if (ret <= 0)
	{
		fprintf(stderr, "Socket error! Webpage Get aborted...\n");
		return;
	}
	getnum = atoi(getnumstr);
	free(getnumstr);

	for(int i=0; i<getnum; i++)
	{
		ret = read_msg(sktid, &url);
		if (ret <= 0)
		{
			fprintf(stderr, "Socket error! Webpage Get aborted...\n");
			break;
		}
	
		status = Get_fromFile(url, &filebuf, &buflen);
		if (status < 0)
		{
			ret = write_msg(sktid, "ERROR", 5);
			fprintf(stderr, "Get url=%s ERROR...\n", url);
		}
		if (status == 0)
		{
			ret = write_msg(sktid, "NOTFOUND", 8);
			fprintf(stderr, "Get url=%s Not Found...\n", url);
		}
		if (status > 0)
		{
			ret = write_msg(sktid, "OK", 2);
			fprintf(stderr, "Get url=%s Succeeds...\n", url);
		}
		free(url);
		if (ret <= 0)
		{
			fprintf(stderr, "Socket error! Webpage Get aborted...\n");
			if (status > 0)
				free(filebuf);
			break;
		}
		if (status > 0)
		{
			ret = write_msg(sktid, filebuf, buflen);
			if (ret <= 0)
			{
				fprintf(stderr, "Socket error! Webpage Get aborted...\n");
				free(filebuf);
				break;
			}
			else
				free(filebuf);
		}
	}
	return;	
}

void Webcache_Service::Webcache_Del(int sktid)
{
	char* url;
	int status;
	int ret;
	char* delnumstr;
	int delnum;

	ret = read_msg(sktid, &delnumstr);
	if (ret <= 0)
	{
		fprintf(stderr, "Socket error! Webpage Delete aborted...\n");
		return;
	}
	delnum = atoi(delnumstr);
	free(delnumstr);

	for(int i=0; i<delnum; i++)
	{
		ret = read_msg(sktid, &url);
		if (ret <= 0)
		{
			fprintf(stderr, "Socket error! Webpage Delete aborted...\n");
			break;
		}
	
		status = Delete_fromFile(url);
		if (status < 0)
		{
			ret = write_msg(sktid, "ERROR", 5);
			fprintf(stderr, "Delete url=%s ERROR...\n", url);
		}
		if (status == 0)
		{
			ret = write_msg(sktid, "NOTFOUND", 8);
			fprintf(stderr, "Delete url=%s Not Found...\n", url);
		}
		if (status > 0)
		{
			ret = write_msg(sktid, "OK", 2);
			fprintf(stderr, "Delete url=%s Succeeds...\n", url);
		}
		free(url);
		if (ret <= 0)
		{
			fprintf(stderr, "Socket error! Webpage Delete aborted...\n");
			break;
		}
	}
	Save_Structure();
	return;
}


void* One_Connection(void* arg)
{
	Arg_List* arglist;
	Webcache_Service* ws;
	int sktid;

	arglist = (Arg_List *) arg;
	sktid = arglist->sktid;
	ws = arglist->ws;
	
	if (strcmp(arglist->cmd, "PUT") == 0)
		ws->Webcache_Put(sktid);
	if (strcmp(arglist->cmd, "GET") == 0)
		ws->Webcache_Get(sktid);
	if (strcmp(arglist->cmd, "DEL") == 0)
		ws->Webcache_Del(sktid);
		
	close(sktid);
	arglist->ws = NULL;
	free(arglist);
	pthread_exit(NULL);
}

int main(int argc, char* argv[])
{
	int ret;
	Webcache_Service ws;
	char* cmd;
	
	int listenfd, connfd;
	struct sockaddr_in servaddr;
	struct sockaddr_in clientaddr;
	int clientaddr_len;
	int on;

	Arg_List* arglist;
	pthread_t tid;
	pthread_attr_t tattr;

	if (argc != 2)
	{
		fprintf(stderr, "Command Error!\nUsage:cachesvr configfile\n");
		exit(1);
	}

	ret = ws.Init_Webcache(argv[1]);
	if (ret != 1)
	{
		fprintf(stderr, "Initialize the web cache Service Failed, Exit...\n");
		if (ws.sub)
		{
			ws.sub->stop();
			ws.sub->finalize();
			delete ws.sub;
		}
    	if (ws.sorfs) 
        	delete ws.sorfs;
    	if (ws.sclt) 
        	delete ws.sclt;
		return -1;
	}
	
	signal(SIGPIPE, SIG_IGN);

	listenfd = socket(AF_INET, SOCK_STREAM, 0);
	if (listenfd < 0)
	{
		fprintf(stderr, "Cache Server creats socket error, exit!\n");
		ws.Final_Webcache();
		exit(1);
	}
	bzero(&servaddr, sizeof(servaddr));
	servaddr.sin_family = AF_INET;
	servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
	servaddr.sin_port = htons(SERVER_PORT);
	if (bind(listenfd, (struct sockaddr *)&servaddr, sizeof(servaddr)) < 0)
	{
		fprintf(stderr, "Server bind error, exit!\n");
		ws.Final_Webcache();
		exit(1);
	}
	on = 1;
	setsockopt(listenfd, SOL_SOCKET, SO_REUSEADDR, (char*)on, sizeof(on));
	if (listen(listenfd, BACKLOG) < 0)
	{
		fprintf(stderr, "Server listen error, exit!\n");
		ws.Final_Webcache();
		exit(1);
	}
	
	for(;;)
	{
		clientaddr_len = sizeof(clientaddr);
		connfd = accept(listenfd, (struct sockaddr*)&clientaddr, (socklen_t *)&clientaddr_len);
		if (connfd < 0)
		{
			fprintf(stderr, "Server accept error!\n");
			continue;
		}
		fprintf(stderr, "New client %s coming...\n", inet_ntoa(clientaddr.sin_addr));
		fflush(stderr);
	
		//if the EXIT client
		ret = read_msg(connfd, &cmd);
		if (ret <= 0)
		{
			fprintf(stderr, "Client socket error!\n");
			fflush(stderr);
			close(connfd);
			continue;
		}
		else
		{
			if(strcmp(cmd, "EXIT") == 0)
			{
				close(connfd);
				close(listenfd);
				ws.Final_Webcache();
				free(cmd);
				fprintf(stderr, "Server Exits!\n");
				pthread_exit(NULL);
			}
		}
		
		if ((arglist=(Arg_List *) malloc(sizeof(Arg_List))) == NULL)
		{
			fprintf(stderr, "Not enough memory, socket closed!\n");
			fflush(stderr);
			close(connfd);
			continue;
		}
		arglist->sktid = connfd;
		arglist->ws = &ws;
		strncpy(arglist->cmd, cmd, 10);
		free(cmd);
		pthread_attr_init(&tattr);
		/* set detach state for each child thread */
		if (pthread_attr_setdetachstate(&tattr, PTHREAD_CREATE_DETACHED) != 0)
		{
			fprintf(stderr, "Set detach state failed!\n");
			fflush(stderr);
			close(connfd);
			continue;
		}
		pthread_attr_setscope(&tattr, PTHREAD_SCOPE_SYSTEM);
		if (pthread_create(&tid, &tattr, One_Connection, arglist) != 0)
		{
			fprintf(stderr, "Server create thread failed, the connection for this client is closed!\n");
			fflush(stderr);
			close(connfd);
			free(arglist);
			arglist = NULL;
			continue;
		}
		arglist = NULL;
	}
}
