// -*- mode:C; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: asmalloc.c,v $ - a capable memory checker for C.
 * Authors :  Max Ibel, ibel@cs.ucsb.edu,
 *            Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2002/10/08 07:32:52 $
 * Version :  $Revision: 1.2 $
 *
 * Originally from RENS active service framework.
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include "asmalloc_api.h"

static int do_trace=0;

/* single process version */
/* array size that stores pointer and size for all mallocs in program
that are generated with AS_Malloc or AS_Strdup (don't use AS_free on
any pointer that was not allocated with AS_Malloc or AS_Strdup) -
increase this value as necessary for larger programs */

#define MALLARRSZ 1000000

/* 'blowup' value: if the malloc bug is not due to freeing twice or
   freeing sth. that hasn't been malloced than it must be due to
   overwriting malloc boundaries. we can figure this out by allocating
   inside this wrapper more than actually requested (size*BLOWUP) and
   fill the excess memory ([size*1 - size*BLOWUP-1]) with
   TESTCHARs. During AS_free()s we can test this excess memory if the
   application overwrote the TESTCHARs. The trick is setting BLOWUP to a

   'high enough' value that the code runs to completion (if the malloc
   bug is caused by a boundary overwriting there always exists such a
   BLOWUP value that the code won't break). Since it completes we'll see

   all the AS_free()'s corresponding to the AS_Mallocs()s...and
   therefore can see all possible boundary overwritings ... well, that
   is, only if YOU are a consentious programmer that frees the stuff
   that you malloc and don't rely on exit() to take care of that...

   an alternative to increasing BLOWUP until the code won't break is
   using alarm() and periodically check for boundary overwritings in
   here...but that was too tedious for me at the time ;) (and anyways
   you have to calibrate the alarm timer and all...) */

/* # define PADDING_SIZE 64 */
# define PADDING_SIZE 16
# define TESTCHAR ((char)0x15)

/* the array with the info about all mallocs (AS_Mallocs and AS_Strdups)
*/
typedef struct {
	void *ptr;
	int size;
	int line;
	char *file;
} memblockDesc;

static memblockDesc *m;

/* the position until which the array is filled */
static int m_pos = 0;

void AS_InitMem()
{
	m=(memblockDesc *)calloc(MALLARRSZ, sizeof(memblockDesc));
	if (m)
		do_trace=1;
}

/*
modified malloc routine that is used instead of standard malloc

  blowup every malloc by a factor and initialize with magic value;
  either the @see checkmem routine or @see free, also called by every
  AS_Malloc1, will first check that the magic has not been
  overwritten; If it does, then we know that some one has written
  randomly into memory; Essentially, a poor man's electric fence.

  we also record line# and filename for every malloc, so to tackle the
  offended malloc as close as possible

  @param size size of requested block
  @param line line number of malloc (__LINE__)
  @param file filename (__FILE__)

  @returns memory or NULL if unsuccessful
  */

static int SAMPLING=100;

void enable_malloc_check(int k)
{
	if (k<=0)
		k=1;
	SAMPLING=k;
}

void disable_malloc_check()
{
	SAMPLING=0;
}

static int sample=0;

char * AS_StrDup1(char *str, const int line, const char *file)
{
	if (do_trace) {
		size_t size=strlen(str)+1;
		char *ptr=(char *)AS_Malloc1(size, line, file);
	
		if (ptr)
			strcpy(ptr, str);
	
		return ptr;
	}
	else 
		return strdup(str);
}

void * AS_Calloc1(const size_t nelem, const size_t elmsize, const int line, char  *file)
{
	if (do_trace) {
		size_t size=nelem*elmsize;
		char *ptr=(char *)AS_Malloc1(size, line, file);
	
		if (ptr)
			memset(ptr, 0, size);
	
		return (void *)ptr;
	}
	else
		return calloc(nelem, elmsize);
}

void * AS_Malloc1(const size_t size, const int line, const char *file)
{
	if (do_trace) {
		void *ptr;
		char *cptr;
		int i=0;
	
		if ( (SAMPLING != 0) && (((++sample)%SAMPLING)==0) )
			AS_CheckMem();
	
		if(size==0) {
			fprintf(stderr, "allocating 0 bytes at (%s:%d)? Strange ...\n", file, line);
			fflush(stderr);
		}
	
		if ((ptr = malloc(size+2*PADDING_SIZE)) == NULL) {
			fprintf(stderr, "malloc failed (size=%d)\n", size+2*PADDING_SIZE);
			fflush(stderr);
			return NULL;
		}

		cptr = (char*)ptr;

		/* write TESTCHAR in excess memory */
		for (i=0; i<PADDING_SIZE; i++)
			cptr[i] = TESTCHAR;

		for (i=size+PADDING_SIZE; i<size+2*PADDING_SIZE; i++)
			cptr[i] = TESTCHAR;

		/* insert in array: if there's a NULL entry between 0 and m_pos than
		     	there was a AS_free for this spot and we can reuse it */
		/**
		if (m_pos<MALLARRSZ) {
			i=m_pos++;
		}
		else {
			for (i=0; i<m_pos; i++)
				if (m[i].ptr == NULL) {
					break;
				}
		}
		**/

		for (i=0; i<m_pos; i++)
			if (m[i].ptr == NULL) {
				break;
			}
	
		if ((i==m_pos)&&(m_pos<MALLARRSZ)) {
			m_pos++;
		}
	
		if ((i>=0)&&(i<MALLARRSZ)) {
			m[i].ptr = cptr+PADDING_SIZE;
			m[i].size = size;
			m[i].line=line;
			m[i].file=(char *)file;
			return cptr + PADDING_SIZE;
		}
		else {
			fprintf(stderr, "AS_Malloc: ran out of array space\n");
			fflush(stderr);
			return NULL;
		}
	}
	else
		return malloc(size);
}
	

/**
  checking version of free().

  we check that no previous Malloc has been overwritten and that we
  are freeing something that has been malloced before. Since we are
  using GC this is now somewhat superfluous :-)

  @param ptr pointer to be freed
  @returns nothing
 */

static void report_error(int block_id, int index)
{
	char *cptr=(char *)m[block_id].ptr-PADDING_SIZE;

	fprintf(stderr,"memwrite at index %d, size %d, (value 0x%x), " \
		"allocated at %s:%d", 
		index-PADDING_SIZE, m[block_id].size, cptr[index], m[block_id].file, \
		m[block_id].line);
	if (index<PADDING_SIZE) {
		fprintf(stderr, "value (as string): %s (0x%x,0x%x,0x%x) before that\n", \
			(char *)m[block_id].ptr, cptr[index-3], cptr[index-2], cptr[index-1]);
	}
	else {
		fprintf(stderr, "value (as string): %s (0x%x,0x%x,0x%x) after that\n", \
			(char *)m[block_id].ptr, cptr[index+1], cptr[index+2], cptr[index+3]);
	}
	fprintf(stderr, "out-of-bounds error\n");
	fflush(stderr);
}

static int safe_mode=1;

void enable_safe_mode()
{
	safe_mode=1;
}

void disable_safe_mode()
{
	safe_mode=1;
}

void AS_Free1(void *ptr, const int line, const char *file)
{
	if (do_trace) {
		char *cptr;
		int i,j,error=0;
	
		if (!ptr) {
			fprintf(stderr,"Trying to free NULL pointer! (%s,%d)\n",file,line);
			fflush(stderr);
			return;
		}
	
		/* find ptr in array */
		for (i=0; i<m_pos; i++) {
			if (m[i].ptr == ptr) {
				cptr = (char*)ptr-PADDING_SIZE;
	
				/* detection of 1st possible malloc bug: check for malloc boundary 
						         	violation in excess memory */
				
				for (j=0; j<PADDING_SIZE; j++) {
					if (cptr[j] != TESTCHAR) {
						report_error(i, j);
						error=1;
						break;
					}
				}
	
				for (j=m[i].size+PADDING_SIZE; j<m[i].size+PADDING_SIZE*2; j++) {
					if (cptr[j] != TESTCHAR) {
						report_error(i, j);
						error=1;
						break;
					}
				}
				if(error==1) {
					return;
				}
				bzero(cptr,m[i].size+PADDING_SIZE*2);
				m[i].ptr = NULL;
				m[i].size=-1;
				free(cptr);
				if (i==m_pos) {
					do {
						m_pos--;
					} while (m[m_pos-1].ptr==NULL);
				}
				return;
			}
		}
		/* detection of 2nd possible malloc bug: ptr wasn't found in
		     	array...what are you freeing, man? */
		fprintf(stderr, "AS_free: freeing something that hasn't been malloc'd"
	    	" or freed before! (%s,%d)\n",file,line);
		fflush(stderr);
		if (!safe_mode) /* maybe the memory is allocated from other modules */
			free(ptr);
	}
	else
		free(ptr);
}

/**
   give a dump of the memory allocations
*/
void AS_DumpMem()
{
	if (do_trace) {
		int i;

		for (i=0; i<m_pos; i++) {
			/* detection of 1st possible malloc bug: check for malloc boundary
		       	violation in excess memory */
			if(m[i].ptr) {
				fprintf(stderr, "m[%d]=0x%p,size = %d, line = %u, file = %s\n", 
			    	i, m[i].ptr, m[i].size, m[i].line, m[i].file?m[i].file:"UNDEF");
				fflush(stderr);
			}
		}
	}
	else
		return;
}


/**
  check that no write in previously allocated blocks has happened
  useful to find out-of-bounds array accesses and such.
  @returns nothing, but will abort if error is diagnosed
  */
void AS_CheckMem()
{
	if (do_trace) 
	{
		int i,j;
		char *cptr;

		for (i=0; i<m_pos; i++) {
			cptr=(char *) m[i].ptr;
			if(cptr==NULL)
				continue;
			cptr-=PADDING_SIZE;
			for (j=0; j<PADDING_SIZE; j++) {
				if (cptr[j] != TESTCHAR) {
						report_error(i, j);
					abort();
				}
			}
			for (j=m[i].size+PADDING_SIZE; j<m[i].size+PADDING_SIZE*2; j++) {
				if (cptr[j] != TESTCHAR) {
					report_error(i, j);
					abort();
				}
			}
		}
	}
	else
		return;
}
