// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: bloom.cc,v $ - base Bloom filter, MT-safe.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2002/10/26 02:21:40 $
 * Version :  $Revision: 1.19 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 
# include <stdlib.h>
# include <pthread.h>
# include <math.h>
# include <string.h>
# include <zlib.h>
# include "bloom.h"
# include "../md5/md5.h"

static const unsigned int sha_consts[]={0x5a827999, 0x6ed9eba1, 0x8f1bbcdc, 0xca62c1d6};

# define BLOOM_HASH MD5
// # define BLOOM_HASH CRC

# ifndef USE_SIX
# define NBITS 4
# else
# define NBITS 6
# endif

static void crc2ints(unsigned char *key, int len, unsigned int *ints)
{
	ints[0]=(unsigned int)crc32(0, key, len);
	for (int i=1; i<NBITS; i++) {
		ints[i]=crc32(ints[i-1], key, len);
	}

	return;
}

static void md52ints(unsigned char *key, int len, unsigned int *ints)
{
	Md5::md5(key, len, (unsigned char *)ints);
# ifdef USE_SIX
	ints[4]=ints[0]^ints[1]^sha_consts[0];
	ints[5]=ints[2]^ints[3]^sha_consts[1];
# endif

	return;
}

// primality tester, wheel factorization
bool BaseBF::is_prime(unsigned int p)
{
	if (p<6) {
		if ((p==3)||(p==5)||(p==2)) {
			return true;
		}
		else {
			return false;
		}
	}

	if ( (!(p%3)) || (!(p%5)) || (!(p%2)) ) {
		return false;
	}

	unsigned int f=(unsigned int)sqrt(p)+1;
	unsigned int wheel[]={7, 11, 13, 17, 19, 23, 29, 31};

	for (unsigned int i=0; i<=f; i+=30) {
		for (unsigned int j=0; j<sizeof(wheel)/sizeof(unsigned int); j++) {
			unsigned factor=i+wheel[j];
			if (factor>f)
				break;
			if (!(p%factor))
				return false;
		}
	}

	return true;
}

// Find the closest but smaller prime number to upper
unsigned int BaseBF::prime_within(unsigned int upper)
{
	if (upper==2)
		return upper;

	if (!(upper&0x1)) {
		upper--;
	}
	while ((upper>1)&&(!is_prime(upper))) {
		upper-=2;
	}

	return (upper<=1)?0: upper;
}

# define BF_BITS_PER_BYTE (8)
# define BF_BITS_PER_PAGE (BF_PAGE_SIZE*BF_BITS_PER_BYTE)

BaseBF::BaseBF(int max_objs) :
	mask(NULL),
	msize(0),
	fsize(0)
{
	if (max_objs<=0) {
		return;
	}

	fsize=(unsigned int)(4*max_objs*SCALE_FACTOR);
	msize=(fsize+BF_BITS_PER_PAGE-1)/BF_BITS_PER_PAGE;
	fsize=prime_within(msize*BF_BITS_PER_PAGE);
	mask=new unsigned char[msize*BF_PAGE_SIZE];

	if (!mask) {
		return;
	}

	if (pthread_mutex_init(&lock, NULL)<0) {
		delete []mask;
		mask=NULL;
		return;
	}

	memset(mask, 0, msize*BF_PAGE_SIZE);

	return;
}

BaseBF::~BaseBF()
{
	if (mask) {
		delete []mask;
		mask=NULL;
		pthread_mutex_destroy(&lock);
	}
}

bool BaseBF::reset()
{
	if (!mask)
		return false;

	memset(mask, 0, msize*BF_PAGE_SIZE);

	return true;
}

bool BaseBF::update_block(void *pmask, int offset, int len)
{
	if (!mask)
		return false;

	if ((offset<0)||(((unsigned int)offset+(unsigned int)len)>msize*BF_PAGE_SIZE)) {
		return false;
	}

	if (pthread_mutex_lock(&lock)<0)
		return false;
	memcpy((unsigned char *)mask+offset, pmask, len);
	pthread_mutex_unlock(&lock);

	return true;
}

bool BaseBF::update_blocks(void *blocks, unsigned int *block_idx, int size)
{
	if ((!blocks)||(!block_idx)||(size<=0))
		return false;

	bool ret=true;
	for (int i=0; i<size; i++) {
		ret = ret && update_block
			((unsigned char *)blocks+BF_PAGE_SIZE*i, 
			 block_idx[i]*BF_PAGE_SIZE, BF_PAGE_SIZE);
	}

	return ret;
}

static bool is_bit_set(unsigned char *bit_array, unsigned int idx)
{
	unsigned int byte_idx=idx/BF_BITS_PER_BYTE;
	unsigned int inbyte_idx=idx%BF_BITS_PER_BYTE;
	return (bit_array[byte_idx]&((unsigned char)1<<inbyte_idx)) ?
		true: false;
}

static void bit_set(unsigned char *bit_array, unsigned int idx)
{
	unsigned int byte_idx=idx/BF_BITS_PER_BYTE;
	unsigned int inbyte_idx=idx%BF_BITS_PER_BYTE;

	bit_array[byte_idx]|=(unsigned char)1<<inbyte_idx;

	return;
}

static void bit_clr(unsigned char *bit_array, unsigned int idx)
{
	unsigned int byte_idx=idx/BF_BITS_PER_BYTE;
	unsigned int inbyte_idx=idx%BF_BITS_PER_BYTE;

	bit_array[byte_idx]&=~((unsigned char)1<<inbyte_idx);

	return;
}

bool BaseBF::update_bits(unsigned int *bit_indices, int size)
{
	static void (*f[])(unsigned char *, unsigned int) = 
		{ bit_clr, bit_set};

	if ((!bit_indices)||(size<=0))
		return false;

	if (pthread_mutex_lock(&lock)<0)
		return false;
	for (int i=0; i<size; i++) {
		unsigned int tmp_v=bit_indices[i];
		if ((tmp_v>>1)>=fsize) {
			// ignore bad indices
			continue;
		}
		(*(f[tmp_v&0x1]))(mask, tmp_v>>1);
	}
	pthread_mutex_unlock(&lock);

	return true;
}

// assuming digest is of length 16 and ints of length 4
// we do not perform direct memory casting to avoid endian problems.
// essentially we do the big-endian conversion for the MD5 digest.
static void conv_digest_ints(void *digest, unsigned int *ints)
{
	for (int i=0; i<4; i++) {
		unsigned char *pc=(unsigned char *)digest+i*4;
		ints[i]=0;
		for (int j=0; j<4; j++) {
			ints[i]=(ints[i]<<8)+(*(pc+j));
		}
	}

	return;
}

bool BaseBF::post_check(check_handle_t &ch)
{
	if (!mask)
		throw -1;

	unsigned int indice[NBITS];

	for (int i=0; i<NBITS; i++) {
		indice[i]=ch.ui[i]%fsize;
	}

# if defined(USE_SIX)
	if (is_bit_set(mask, indice[0]) &&
		is_bit_set(mask, indice[1]) &&
		is_bit_set(mask, indice[2]) &&
		is_bit_set(mask, indice[3]) &&
		is_bit_set(mask, indice[4]) &&
		is_bit_set(mask, indice[5]) 
		// && is_bit_set(mask, indice[6])    
		) 
	{
		return true;
	}
# else
	if (is_bit_set(mask, indice[0]) &&
		is_bit_set(mask, indice[1]) &&
		is_bit_set(mask, indice[2]) &&
		is_bit_set(mask, indice[3])    ) 
	{
		return true;
	}

# endif
	return false;
}

void BaseBF::prepare_check(const void *key, int len, check_handle_t &ch)
{
	if ((!key)||(len<=0)) {
		throw -1;
	}

# if (BLOOM_HASH==CRC)
	crc2ints((unsigned char *)key, len, ch.ui);
# else
	md52ints((unsigned char *)key, len, ch.ui);
# endif

	return;
}


bool BaseBF::clone(BaseBF &bf_from) 
{
	if (pthread_mutex_lock(&lock)<0) {
		return false;
	}

	if ((mask)&&(bf_from.msize!=msize)) {
		delete []mask;
		mask=NULL;
	}

	if (!mask) {
		mask=new unsigned char[bf_from.msize*BF_PAGE_SIZE];
	}
	msize=bf_from.msize;
	fsize=bf_from.fsize;

	memcpy(mask, bf_from.mask, msize*BF_PAGE_SIZE);

	pthread_mutex_unlock(&lock);

	return true;
}

void BF::release()
{
	if (mask) {
		delete []mask;
		mask=NULL;
	}

	if (bits_cnt) {
		delete []bits_cnt;
		bits_cnt=NULL;
	}

	if (page_clock) {
		delete []page_clock;
		page_clock=NULL;
	}

	if (lab_dat) {
		delete []lab_dat;
		lab_dat=NULL;
	}
}

bool BaseBF::cmp(BaseBF *bf)
{
	if (msize!=bf->msize)
		return false;
	if (fsize!=bf->fsize)
		return false;
	return (memcmp(mask, bf->mask, BF_PAGE_SIZE*msize)==0);
}

BF::BF(int max_objs, int n_lab_entries) : 
	BaseBF(max_objs),
	bits_cnt(NULL),
	page_clock(NULL),
	n_setbit(0),
	prev_page(-1), 
	obj_cnt(0),
	bcast_gen(0),
	prev_base(0),
	cur_base(0),
	lab_dat(NULL),
	lab_idx(0),
	lab_last(0),
	lab_size(n_lab_entries)
{
	if (!mask)
		return;

	bits_cnt=new unsigned char[fsize];
	page_clock=new unsigned char[msize];
	lab_dat=new unsigned int[lab_size];

	if ((!bits_cnt)||(!page_clock)||(!lab_dat)) {
		goto _exit;
	}

	memset(bits_cnt, 0, fsize);
	memset(page_clock, 0, msize);
	memset(lab_dat, 0xff, sizeof(unsigned int)*lab_size);

	return;

_exit: 
	release();

	return;
}

BF::~BF()
{
	release();
}

bool BF::add(const void *key, int len)
{
	if (!mask)
		return false;

	unsigned int indice[NBITS];
# if (BLOOM_HASH==CRC)
	crc2ints((unsigned char *)key, len, indice);
# else
	md52ints((unsigned char *)key, len, indice);
# endif

	for (int i=0; i<NBITS; i++) {
		indice[i]%=fsize;
	}

	if (pthread_mutex_lock(&lock)<0)
		return false;

	bool overflow=false;

	for (int i=0; i<NBITS; i++) {
		if (bits_cnt[indice[i]]==255)  {
			overflow=true;
			break;
		}
	}

	if (!overflow) {
		for (int i=0; i<NBITS; i++) {
			unsigned int tmp_idx=indice[i];
			if (!bits_cnt[tmp_idx]) { // first time, set bit
				bit_set(mask, tmp_idx);
				n_setbit++;
				page_clock[tmp_idx/BF_BITS_PER_PAGE]=1;
				lab_dat[lab_idx]=(tmp_idx<<1)|0x1;
				lab_idx=(lab_idx+1)%lab_size;
			}
			bits_cnt[tmp_idx]++;
		}
		obj_cnt++;
	}

	pthread_mutex_unlock(&lock);
	
	if (!overflow)
		return true;

	return false;
}

bool BF::del(const void *key, int len)
{
	if (!mask)
		return false;

	unsigned int indice[NBITS];
# if (BLOOM_HASH==CRC)
	crc2ints((unsigned char *)key, len, indice);
# else
	md52ints((unsigned char *)key, len, indice);
# endif

	for (int i=0; i<NBITS; i++) {
		indice[i]%=fsize;
	}

	if (pthread_mutex_lock(&lock)<0)
		return false;
	bool underflow=false;
	for (int i=0; i<NBITS; i++) {
		if (bits_cnt[indice[i]]==0)  {
			underflow=true;
			break;
		}
	}

	if (!underflow) {
		for (int i=0; i<NBITS; i++) {
			unsigned int tmp_idx=indice[i];
			bits_cnt[tmp_idx]--;
			if (!bits_cnt[tmp_idx]) { // needs to clear the bit.
				n_setbit--;
				bit_clr(mask, tmp_idx);
				page_clock[tmp_idx/BF_BITS_PER_PAGE]=1;
				lab_dat[lab_idx]=(tmp_idx<<1);
				lab_idx=(lab_idx+1)%lab_size;
			}
		}
		obj_cnt--;
	}
	pthread_mutex_unlock(&lock);
	
	if (!underflow)
		return true;

	return false; // very bad error!!!
}

bool BF::first_mod_page(void *page, int *offset, int *len)
{
	if (!mask || (*len<BF_PAGE_SIZE))
		return false;

	prev_page=-1;

	return next_mod_page(page, offset, len);
}

// return a random integer in range from low to up (inclusive).
inline static unsigned int rand_in_range(unsigned int low, unsigned int up)
{
	return rand()%(up+1-low)+low;
}

// to avoid flocking effect, we initiate the clock
// for an unmodified page to be a random number.
# define RAND_CLK_START 100
# define RAND_CLK_END 255

bool BF::next_mod_page(void *page, int *offset, int *len)
{
	if (!mask || (*len<BF_PAGE_SIZE))
		return false;

	int mod_page=-1;

	// No matter whether we have found a page or not, we will
	// decrement all the >1 counters. 
	for (unsigned int i=prev_page+1; i<prev_page+msize; i++) {
		unsigned page_idx=i%msize;
		switch (page_clock[page_idx]) {
		case 0: // nothing
			break;
		case 1:
			if (mod_page>=0)
				break;

			// found "modified page"
			// It could well be an unmodified page that hasn't
			// been broadcast for a long time.
			{
				bool false_alarm=false;
				if (pthread_mutex_lock(&lock)<0)
					break;
				if (page_clock[page_idx]==1) { // nobody has changed it yet
					page_clock[page_idx]=(unsigned char)
						rand_in_range(RAND_CLK_START, RAND_CLK_END);
				}
				else {
					false_alarm=1;
				}
				pthread_mutex_unlock(&lock);
				if (!false_alarm) {
					*offset=page_idx*BF_PAGE_SIZE;
					*len=BF_PAGE_SIZE;
					mod_page=page_idx;
					memcpy(page, mask+*offset, *len);
				}
			}
			break;
		default : // decrement the clock
			page_clock[page_idx]--;
		}
	}

	if (mod_page>=0) {
		prev_page=mod_page;
		return true;
	}
	else {
		return false;
	}
}

void BF::add_to_mod_page_buf(void *pg_buf, unsigned int *pg_idx, 
		 unsigned int &mpage_idx, unsigned int page_idx, bool touch_unmod)
{
	if (touch_unmod)
		page_clock[page_idx]=(unsigned char)
			rand_in_range(RAND_CLK_START, RAND_CLK_END);
	else 
		page_clock[page_idx]=0;

	memcpy((unsigned char *)pg_buf+BF_PAGE_SIZE*mpage_idx, 
		   mask+BF_PAGE_SIZE*page_idx, BF_PAGE_SIZE);
	pg_idx[mpage_idx]=page_idx;
	mpage_idx++;
	prev_page=page_idx;
}

int BF::get_mod_pages(void *pg_buf, unsigned int *pg_idx, int max_pgs, bool touch_unmod)
{
	if (!mask || (max_pgs<=0))
		return 0;

	if (pthread_mutex_lock(&lock)<0)
		return 0;

	unsigned int mpage_idx=0;
# define MAX_PCENT_PER_SCAN 0.1
	unsigned int delta=cur_base-prev_base;
	int upper=prev_page+(int)ceil(msize*MAX_PCENT_PER_SCAN);
	for (int i=prev_page+1; i<=upper; i++) {
		unsigned int page_idx=(unsigned int)i%msize;
		if (page_idx == 0) {
			prev_base=cur_base;
			cur_base=bcast_gen;
			delta=cur_base-prev_base;
		}

		if (page_clock[page_idx]==0)
			continue;

		if (page_clock[page_idx]<=bcast_gen-prev_base) {
			page_clock[page_idx]=1;
		}
		else {
			page_clock[page_idx]-=delta;
		}

		if (page_clock[page_idx]==1) {
			// found "modified page"
			// It could well be an unmodified page that hasn't
			// been broadcast for a long time.
			add_to_mod_page_buf(pg_buf, pg_idx, mpage_idx, page_idx, touch_unmod);
			if (mpage_idx==(unsigned)max_pgs)
				break;
		}

	}

	bcast_gen++;
	pthread_mutex_unlock(&lock);

	return mpage_idx;
}

int BF::get_pages(void *pg_buf, int first_page, int max_pgs)
{
	if ((!pg_buf)||(first_page<0)||(max_pgs<=0))
		return 0;

	int n_pages=((msize-(unsigned)first_page)>(unsigned)max_pgs)?
		(unsigned)max_pgs:(msize-(unsigned)first_page);
	memcpy(pg_buf, mask+BF_PAGE_SIZE*first_page, BF_PAGE_SIZE*n_pages);
	return n_pages;
}

int BF::get_lab_entries(unsigned int *lab_entries, int max_entries)
{
	if (lab_dat[lab_idx]!=(unsigned int)-1) {
		lab_last=lab_idx;
	}

	int i=0;
	for (; (i<max_entries); i++, lab_last=(lab_last+1)%lab_size) {
		unsigned int tmp_v=lab_dat[lab_last];
		if (tmp_v==(unsigned int)-1)
			break;
		lab_entries[i]=tmp_v;
		lab_dat[lab_last]=(unsigned int)-1;
	}

	return i;
}

int BF::get_usable_lab_size()
{
	if (lab_dat[lab_idx]!=(unsigned int)-1) {
		return lab_size;
	}
	return (int)((lab_idx+lab_size-lab_last)%lab_size);
}

bool BF::first_page(void *page, int *offset, int *len)
{
	if (!mask || (*len<BF_PAGE_SIZE))
		return false;

	prev_page=-1;
	return next_page(page, offset, len);
}

bool BF::next_page(void *page, int *offset, int *len)
{
	if (!mask || (*len<BF_PAGE_SIZE))
		return false;

	if ((unsigned)prev_page==(msize-1))
		return false;

	prev_page++;

	if (pthread_mutex_lock(&lock)<0)
		return false;
	page_clock[prev_page]=(unsigned char)
		rand_in_range(RAND_CLK_START, RAND_CLK_END);
	pthread_mutex_unlock(&lock);
	*offset=prev_page*BF_PAGE_SIZE;
	*len=BF_PAGE_SIZE;
	memcpy(page, mask+*offset, *len);
	return true;
}

bool BF::reset()
{
	if (!BaseBF::reset())
		return false;

	if (!bits_cnt)
		return false;
	memset(bits_cnt, 0, fsize);

	if (!page_clock)
		return false;
	memset(page_clock, 0, msize);

	if (!lab_dat)
		return false;
	memset(lab_dat, 0xff, lab_size*sizeof(unsigned int));

	prev_page=-1;
	n_setbit=0;
	obj_cnt=0;
	bcast_gen=0;
	prev_base=0;
	cur_base=0;
	lab_idx=0;
	lab_last=0;

	return true;
}

bool BF::update_reset()
{
	if (!page_clock)
		return false;
	memset(page_clock, 0, msize);

	if (!lab_dat)
		return false;
	memset(lab_dat, 0xff, lab_size*sizeof(unsigned int));
	prev_page=-1;
	bcast_gen=0;
	prev_base=0;
	cur_base=0;
	lab_idx=0;
	lab_last=0;

	return true;
}

double BaseBF::zero_pcent()
{
	if ((!mask)||(!fsize))
		return (double)-1;

	unsigned int zb_cnt=0;
	for (unsigned int i=0; i<fsize; i++) {
		if (!is_bit_set(mask, i)) {
			zb_cnt++;
		}
	}

	return (double)zb_cnt/fsize;
}

double BF::zero_pcent_debug()
{
	if ((!mask)||(!fsize))
		return (double)-1;

	unsigned int zb_cnt=0;
	for (unsigned int i=0; i<fsize; i++) {
		if (!is_bit_set(mask, i)) {
			zb_cnt++;
		}
	}

	return (double)zb_cnt/fsize;
}

double BF::zero_pcent()
{
	if ((!mask)||(!fsize))
		return (double)-1;

	return 1.0-((double)n_setbit/fsize);
}
