// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*- 

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: bloom.h,v $ - header file for Bloom filter base classes.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2002/10/26 02:21:40 $
 * Version :  $Revision: 1.16 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

# ifndef _BLOOM_H_
# define _BLOOM_H_

# include <stdlib.h>
# include <pthread.h>

// The implementation uses MD5 for hashing,
// and then each hashed value is broken into 4 integers
// each sets one bit. If -DUSE_SIX is set during compilation,
// then we will xor the first two and the second two integers
// and get a total of 6 integers.
// Every node has a set of base BFs and a full BF
// The full BF tracks at large granularity of modified bits
// and allows for incremental replication to remote nodes.
 
// Restrictions: currently we do not allow for dynamic expansion
// of bloom filters.
// Idea/TO-DO: allow for dynamically size-adjustable filters.
// Need to use a generation number to track size change and
// keep both the old and new filter. Keep using the old filter
// for an amount of time (because getting the new filter filled
// up would take quite some time).
//
// Idea: We should broadcast un-changed chunks at a lower
// frequency so that newly joined nodes will be able to catch up.
// Use a counter for each page. The initial counter for any
// page would be a random number between 10 and 255. Each time, 
// a page counter would be decremented by 1. When it reaches 1,
// it will be broadcast and the counter re-set to a random number.
// If a page consists of all 0s, the counter would be 0, in that
// case, subsequent scans will not decrement the counter at all. If
// a page is modified, the counter will be reset to 1.
//
// The size of the mask is calculated by the following criterion:
// 1. Suppose we will store N objects, then each object needs to
// have one bit unique for itself. A scale factor f great than 1
// is used so that the number of 1's in the mask is expected to
// be fN when the number of maximum objects is reached.

# ifndef SCALE_FACTOR
# define SCALE_FACTOR 3
# endif

// 128 Bytes per page so that the space overhead is <1% for
// page_clock;
// We will try to cram a combination of 63KB's of data per
// announcement: 32KB (maximum) for dirty pages; 31KB (maximum)
// for individual bit indices. The final data will then be 
// compressed with zlib.

# define BF_PAGE_SIZE 128

// 2. We want to have the mask half full, so the total number of
// bits would be around 2fN.
// 3. We want to page-align the memory taken by the mask. Since 
// we choose a page size of 128B. We want to round UP 2fN to be 
// divisable by 1024.
// 4. We want to choose a mask size to be a prime number.

class BaseBF {
protected:
	pthread_mutex_t lock; // the lock
	unsigned char *mask;	// bit mask
	unsigned int msize;	// size of the mask (in the unit of BF_PAGEs)
	unsigned int fsize;	// size of the filter (in bits)
public:
	// constructors/destructors
	BaseBF(int max_objs);
	BaseBF(BaseBF &bf_from) :
		mask(NULL),
		msize(0),
		fsize(0)
	{ 
		pthread_mutex_init(&lock, NULL); 
		clone(bf_from);
	}

	virtual ~BaseBF();

	bool reset(); 		// clear the mask
	// update arbitrary size block (most likely for TCP stream)
	bool update_block(void *pmask, int offset, int len);
	bool update_pages(void *pgs, int first_page, int n_pgs) {
		return update_block(pgs, first_page*BF_PAGE_SIZE, BF_PAGE_SIZE*n_pgs);
	}
	// update fixed-size blocks (BF_PAGE_SIZE)
	bool update_blocks(void *blocks, unsigned int *block_idx, int size);
	bool update_bits(unsigned int *bit_indices, int size);

	typedef struct {
		// we pessimistically assume 6 bits rather than 4.
		unsigned int ui[6]; 
	} check_handle_t;

	static void prepare_check(const void *key, int len, check_handle_t &ch);
	bool post_check(check_handle_t &ch);

	bool check(const void *key, int len) {
		check_handle_t ch;
		prepare_check(key, len, ch);
		return post_check(ch);
	}

	virtual double zero_pcent();
	bool cmp(BaseBF *bf);
	unsigned int get_size() { return msize*BF_PAGE_SIZE; }
	unsigned int get_fsize() { return fsize; }
	virtual bool clone(BaseBF &bf_from); 

	// __ALL_PUBLIC__ is defined to test internal methods.
# ifndef __ALL_PUBLIC__
protected: 
# endif
	BaseBF(); // illegal constructor
	// Find the closest but smaller prime number to upper
	static unsigned int prime_within(unsigned int upper);
	static bool is_prime(unsigned int p);
};  

class BF : public BaseBF {
 protected:
	unsigned char *bits_cnt;	// a count for each bit <256
	unsigned char *page_clock;	// a byte clock counter for each page of bits
 private:
	unsigned int n_setbit; // # of set bits.
	int prev_page;

 protected:
	unsigned int obj_cnt;

 private:
	void release();
	// each time we call get_mod_pages(), bcast_gen will be incremented.
	// each non-zero page will be set a future generation when it is 
	// expired and needs to be re-broadcast. To save space, we only use
	// one byte to set this expiration time. 
	unsigned int bcast_gen;
	// prev_base and cur_base are the base generation of the first broadcast
	// for the whole page_array. While we scan the page_clock array, those
	// pages smaller than the current page is offset by cur_base and those
	// larger than the current page is offset by prev_base.
	unsigned int prev_base;
	unsigned int cur_base;
	void add_to_mod_page_buf
		( void *pg_buf, unsigned int *pg_idx, 
		  unsigned int &mpage_idx, unsigned int page_idx, bool touch_unmod);
	unsigned int *lab_dat;
	unsigned int lab_idx;
	unsigned int lab_last;
	unsigned int lab_size;
	BF(); 		// illegal constructor

 public:
	// constructors/destructors
	// max_objs is the suggested #. to optimize the flase rate and yet
	// reduce memory usage.
	BF(int max_objs, int n_lab_entry=65536); 
	virtual ~BF();

	// operations
	bool add(const void *key, int len);
	bool del(const void *key, int len);

	////////////////////////////////////////////////////////////
	// obsoleted. use get_mod_pages() instead.
	// the following functions will clear the page modification bits
	// page must points to pre-allocated memory
	bool first_mod_page(void *page, int *offset, int *len);
	bool next_mod_page(void *page, int *offset, int *len);

	// end of obsoleted functions
	//////////////////////////////////////////////////////////////

	//////////////////////////////////////////////////////////////
	// replacement function
	int get_mod_pages(void *pg_buf, unsigned int *pg_idx, int max_pgs, bool touch_unmod=true);
	int get_pages(void *pg_buf, int first_page, int max_pgs);
	int get_num_page() { return msize; }

	int get_lab_entries(unsigned int *lab_entries, int max_entries);
	int get_lab_size() { return lab_size; }
	// available lab data that can be used.
	int get_usable_lab_size();

	// the following functions will be used for complete update
	bool first_page(void *page, int *offset, int *len);
	bool next_page(void *page, int *offset, int *len);
	bool reset(); 	
	bool update_reset(); 	
	virtual double zero_pcent();
	double zero_pcent_debug();
	unsigned int get_obj_cnt() { return obj_cnt; }
};

# endif
