// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: simple_bloom.cc,v $ - simple Bloom filter, serializable, loadable,
 *            restrictive access: readonly or sequential.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2002/11/14 06:05:06 $
 * Version :  $Revision: 1.5 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

# include <stdlib.h>
# include <pthread.h>
# include <math.h>
# include <string.h>
# include <zlib.h>
# include "simple_bloom.h"
# include "zlib.h"
# include "../md5/md5.h"

static const unsigned int sha_consts[]={0x5a827999, 0x6ed9eba1, 0x8f1bbcdc, 0xca62c1d6};

# ifndef USE_SIX
# define NBITS 4
# else
# define NBITS 6
# endif

static void crc2ints(unsigned char *key, int len, unsigned int *ints)
{
	ints[0]=(unsigned int)crc32(0, key, len);
	for (int i=1; i<NBITS; i++) {
		ints[i]=crc32(ints[i-1], key, len);
	}

	return;
}

static void md52ints(unsigned char *key, int len, unsigned int *ints)
{
	Md5::md5(key, len, (unsigned char *)ints);
# ifdef USE_SIX
	ints[4]=ints[0]^ints[1]^sha_consts[0];
	ints[5]=ints[2]^ints[3]^sha_consts[1];
# endif

	return;
}

// primality tester, wheel factorization
bool SimpleBF::is_prime(unsigned int p)
{
	if (p<6) {
		if ((p==3)||(p==5)||(p==2)) {
			return true;
		}
		else {
			return false;
		}
	}

	if ( (!(p%3)) || (!(p%5)) || (!(p%2)) ) {
		return false;
	}

	unsigned int f=(unsigned int)sqrt(p)+1;
	unsigned int wheel[]={7, 11, 13, 17, 19, 23, 29, 31};

	for (unsigned int i=0; i<=f; i+=30) {
		for (unsigned int j=0; j<sizeof(wheel)/sizeof(unsigned int); j++) {
			unsigned factor=i+wheel[j];
			if (factor>f)
				break;
			if (!(p%factor))
				return false;
		}
	}

	return true;
}

// Find the closest but smaller prime number to upper
unsigned int SimpleBF::prime_within(unsigned int upper)
{
	if (upper==2)
		return upper;

	if (!(upper&0x1)) {
		upper--;
	}
	while ((upper>1)&&(!is_prime(upper))) {
		upper-=2;
	}

	return (upper<=1)?0: upper;
}

# ifndef BF_PAGE_SIZE
# define BF_PAGE_SIZE 128
# endif

# ifndef SCALE_FACTOR
# define SCALE_FACTOR 3
# endif

# define BF_BITS_PER_BYTE (8)
# define BF_BITS_PER_PAGE (BF_PAGE_SIZE*BF_BITS_PER_BYTE)


SimpleBF::SimpleBF(int max_objs) :
	mask(NULL),
	msize(0),
	fsize(0),
	n_bit_set(0),
	n_object(0)
{
	if (max_objs<=0) {
		return;
	}

	fsize=(unsigned int)(4*max_objs*SCALE_FACTOR);
	msize=(fsize+BF_BITS_PER_PAGE-1)/BF_BITS_PER_PAGE;
	fsize=prime_within(msize*BF_BITS_PER_PAGE);
	mask=new unsigned char[msize*BF_PAGE_SIZE];

	if (!mask) {
		throw -1;
	}

	memset(mask, 0, msize*BF_PAGE_SIZE);

	return;
}

SimpleBF::SimpleBF(FILE *fp) :
	mask(NULL),
	msize(0),
	fsize(0),
	n_bit_set(0),
	n_object(0)
{
	if (!load(fp)) {
		throw -1;
	}
}

# define SIMPLE_BF_MAGIC 0x13568acd

bool SimpleBF::serialize(FILE *fp) 
{
	unsigned int magic=SIMPLE_BF_MAGIC;
	if (fwrite(&magic, sizeof(unsigned int), 1, fp)<=0) {
		return false;
	}
	if (fwrite(this, sizeof(SimpleBF), 1, fp) <= 0) {
		return false;
	}

	if (fwrite(mask, BF_PAGE_SIZE, msize, fp) < msize) {
		return false;
	}

	return true;
}

bool SimpleBF::load(FILE *fp) 
{
	mask=NULL;
	unsigned int magic=0;
	if (fread(&magic, sizeof(unsigned int), 1, fp)<=0) {
		return false;
	}

	if (magic != SIMPLE_BF_MAGIC)
		return false;

	if (fread(this, sizeof(SimpleBF), 1, fp)<=0) {
		return false;
	}

	mask=(unsigned char *)calloc(msize, BF_PAGE_SIZE);
	if (!mask)
		return false;

	if (fread(mask, BF_PAGE_SIZE, msize, fp) != msize) {
		free(mask);
		mask=NULL;
		return false;
	}

	return true;
}

SimpleBF::~SimpleBF()
{
	if (mask) {
		delete []mask;
		mask=NULL;
	}
}

bool SimpleBF::reset()
{
	if (!mask)
		return false;

	memset(mask, 0, msize*BF_PAGE_SIZE);

	return true;
}

inline static bool is_bit_set(unsigned char *bit_array, unsigned int idx)
{
# if defined(_Solaris_)
	unsigned int byte_idx=idx/BF_BITS_PER_BYTE;
	unsigned int inbyte_idx=idx%BF_BITS_PER_BYTE;
	return (bit_array[byte_idx]&((unsigned char)1<<inbyte_idx)) ?
		true: false;
# else
	unsigned int word_idx=idx/(BF_BITS_PER_BYTE*sizeof(unsigned int));
	unsigned int inword_idx=idx%(BF_BITS_PER_BYTE*sizeof(unsigned int));

	return (asm_is_bit_set((unsigned int *)bit_array+word_idx, inword_idx)!=0);
# endif
}

inline static void bit_set(unsigned char *bit_array, unsigned int idx)
{
	unsigned int byte_idx=idx/BF_BITS_PER_BYTE;
	unsigned int inbyte_idx=idx%BF_BITS_PER_BYTE;

	bit_array[byte_idx]|=(unsigned char)1<<inbyte_idx;

	return;
}

inline static void bit_clr(unsigned char *bit_array, unsigned int idx)
{
	unsigned int byte_idx=idx/BF_BITS_PER_BYTE;
	unsigned int inbyte_idx=idx%BF_BITS_PER_BYTE;

	bit_array[byte_idx]&=~((unsigned char)1<<inbyte_idx);

	return;
}

// assuming digest is of length 16 and ints of length 4
// we do not perform direct memory casting to avoid endian problems.
// essentially we do the big-endian conversion for the MD5 digest.
inline static void conv_digest_ints(void *digest, unsigned int *ints)
{
	for (int i=0; i<4; i++) {
		unsigned char *pc=(unsigned char *)digest+i*4;
		ints[i]=0;
		for (int j=0; j<4; j++) {
			ints[i]=(ints[i]<<8)+(*(pc+j));
		}
	}

	return;
}

bool SimpleBF::check(const void *key, int len)
{
	unsigned int indice[NBITS];

	// if (!mask)
		// return false;

	// if ((!key)||(len<=0)) {
		// return false;
	// }

	// crc2ints((unsigned char *)key, len, indice);
	md52ints((unsigned char *)key, len, indice);

// # if defined(USE_SIX)
	// indice[4]=indice[0]^indice[1]^sha_consts[0];
	// indice[5]=indice[2]^indice[3]^sha_consts[1];
	// indice[6]=indice[4]^indice[5]^sha_consts[2];
// # endif

	for (int i=0; i<NBITS; i++) {
		indice[i]%=fsize;
	}

# if defined(USE_SIX)
	if (is_bit_set(mask, indice[0]) &&
		is_bit_set(mask, indice[1]) &&
		is_bit_set(mask, indice[2]) &&
		is_bit_set(mask, indice[3]) &&
		is_bit_set(mask, indice[4]) &&
		is_bit_set(mask, indice[5]) 
		// && is_bit_set(mask, indice[6])    
		) 
	{
		return true;
	}
# else
	if (is_bit_set(mask, indice[0]) &&
		is_bit_set(mask, indice[1]) &&
		is_bit_set(mask, indice[2]) &&
		is_bit_set(mask, indice[3])    ) 
	{
		return true;
	}

# endif
	return false;
}

bool SimpleBF::add(const void *key, int len)
{
	if (!mask)
		return false;

	unsigned int indice[NBITS];
	// crc2ints((unsigned char *)key, len, indice);
	md52ints((unsigned char *)key, len, indice);

	// conv_digest_ints(digest, indice);

// # if defined(USE_SIX)
	// indice[4]=indice[0]^indice[1]^sha_consts[0];
	// indice[5]=indice[2]^indice[3]^sha_consts[1];
	// indice[6]=indice[4]^indice[5]^sha_consts[2];
// # endif

	n_object++;

	for (int i=0; i<NBITS; i++) {
		unsigned int tmp_idx=indice[i]%fsize;
		if (!is_bit_set(mask, tmp_idx)) {
			bit_set(mask, tmp_idx);
			n_bit_set++;
		}
	}

	return true;
}

double SimpleBF::set_bit_pcent()
{
	if ((!mask)||(!fsize))
		return (double)-1;

	return ((double)n_bit_set/fsize);

	/**
	unsigned int zb_cnt=0;
	for (unsigned int i=0; i<fsize; i++) {
		if (!is_bit_set(mask, i)) {
			zb_cnt++;
		}
	}

	return (double)zb_cnt/fsize;
	**/
}

int SimpleBF::get_size()
{ 
	return msize*BF_PAGE_SIZE; 
}
