// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: fc_test2.cc,v $ - File Cache tester II.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/04/09 09:59:58 $
 * Version :  $Revision: 1.6 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 
# include <sys/types.h>
# include <sys/stat.h>
# include <sys/param.h>
# include <unistd.h>
# include <stdio.h>
# include <stdlib.h>
# include <getopt.h>
# include <queue>
# include <map>

using namespace std;

# include "../usc/usc_time.h"
# include "../tpool/cbs_thread.h"
# include "../../expr/sprite/sprite.h"
# include "../debug/tmpi_debug.h"
# include "file_cache.h"

/**
 * A dummy program that reads a trace and make calls
 * to File_Cache.
 * fc_test -n <cache_size> -f <trace list ...>
 */

static char *prog=NULL;

static int get_mode(char *f, mode_t &m)
{
	struct stat st;
	if (stat(f, &st)<0)
		return -1;

	m=st.st_mode;

	return 0;
}

# define MAX_TRACE_FILES 32

static void do_help()
{
	printf("\
File Cache Performance Test.\n\
Usage: %s [options] [trace list ...].\n\
   -n <cache_size>        Cache Size (0: disabled)\n\
   -f <trace>             Trace file.\n\
   -t <tmp>               Temp directory.\n\
   -l <log>               Log file.\n\
   -h                     Help, this screen.\n\
\n\
Multiple trace files can be specified by the command \n\
line, and each will be driven by a thread. There is\n\
a hard limit that no more than %d trace files can be\n\
specified.\n\
", prog, MAX_TRACE_FILES);
	exit(0);
}

struct IDType_cmp : public less<IDType> {
	bool operator()(const IDType &x, const IDType &y) const {
		return(memcmp(x.p+1, y.p+1, sizeof(int)*3)<0);
	}
};

// mapping sids to fids
class TC_StreamMap {
 private:
	pthread_mutex_t lock;
	typedef map<IDType, IDType, IDType_cmp> SID_MAP;
	SID_MAP sid_map;

 public:
	TC_StreamMap() { pthread_mutex_init(&lock, NULL); }
	~TC_StreamMap() { pthread_mutex_destroy(&lock); }
	bool get(IDType &key, IDType &val) {
		if (pthread_mutex_lock(&lock)<0) {
			return false;
		}

		key.p[0]=0;
		SID_MAP::iterator pi=sid_map.find(key);

		bool ret=false;

		if (pi!=sid_map.end()) {
			val=pi->second;
			ret=true;
		}

		pthread_mutex_unlock(&lock);

		return ret;
	}

	bool set(IDType key, IDType &val) {
		if (pthread_mutex_lock(&lock)<0) {
			return false;
		}

 		// ideally there should be no overriding.
		// but it is possible that the caller might
		// forget to call del after close the handle
		key.p[0]=0;
		SID_MAP::iterator pi=sid_map.find(key);
		if (pi!=sid_map.end()) {
			pi->second=val;
		}
		else {
			sid_map.insert(SID_MAP::value_type(key, val));
		}

		pthread_mutex_unlock(&lock);

		return true;
	}

	bool del(IDType key) {
		if (pthread_mutex_lock(&lock)<0) {
			return false;
		}

		key.p[0]=0;
		SID_MAP::iterator pi=sid_map.find(key);
		bool ret=false;
		if (pi!=sid_map.end()) {
			sid_map.erase(pi);
			ret=true;
		}

		pthread_mutex_unlock(&lock);

		return ret;
	}
};

# define ODS_SIZE 1000

class TC_Tester : public CBS_Thread {
 private:
	SpriteParser parser;
	char *tmp;
	File_Cache *fcache;
	TC_StreamMap *sid_map;
	void handle_record(SpriteRecord *r);
	bool get_file_handle(IDType &sid, IDType &fid);
	FILE *fp_log;
	IDType on_demand_sid[ODS_SIZE];
	int ods_idx;

 public:
	TC_Tester(
		char *f_trace, 
		char *temp, 
		File_Cache *file_cache, 
		TC_StreamMap *smap, 
		FILE *logf
	) : 
		parser(f_trace), 
		tmp(temp), 
		fcache(file_cache), 
		sid_map(smap), 
		fp_log(logf),
		ods_idx(0)
	{ 
		for (int i=0; i<ODS_SIZE; i++) {
			for (int j=0; j<4; j++) {
				on_demand_sid[i].p[j]=-1;
			}
		}
	}

	virtual void *run();
};

void *TC_Tester::run()
{
	SpriteRecord *r;
	while ((r=parser.read_record())) {
		// r->dump(fp_log);
		handle_record(r);
		parser.free_record(r);
	}

	return NULL;
}

static unsigned char hex_to_char(unsigned char hex)
{
	if (hex>=0x10)
		return 'X';
	if (hex<10)
		return '0'+hex;
	return 'A'+(hex-10);
}

static void id_to_name(char *prefix, IDType &id, char *buf, int len)
{
	bool add_slash=false;
	int prefix_len=0;

	if (prefix) {
		prefix_len=strlen(prefix);
	}
	else {
		prefix="";
	}

	if (prefix_len>0) {
		if (prefix[prefix_len-1] == '/') {
			prefix[prefix_len-1] = '\0';
		}
		add_slash=true;
	}

	unsigned char c=0;
	unsigned char *pc=(unsigned char *)(id.p+1);
	for (unsigned int i=0; i<sizeof(int)*3; i++) {
		c^=pc[i];
	}

	unsigned char hex1=c/0x10, hex2=c%0x10;
	unsigned char char1=hex_to_char(hex1);
	unsigned char char2=hex_to_char(hex2);
	char sub_dir[10];
	snprintf(sub_dir, 10, "%c/%c", char1, char2);

	if (add_slash) {
		snprintf(buf, len, "%s/%s/%08X%08X%08X", 
				prefix, sub_dir, id.p[1], id.p[2], id.p[3]);
	}
	else {
		snprintf(buf, len, "%s/%08X%08X%08X", 
				sub_dir, id.p[1], id.p[2], id.p[3]);
	}
}

bool TC_Tester::get_file_handle(IDType &sid, IDType &fid)
{
	if (!sid_map->get(sid, fid)) {
		IDType old_sid=on_demand_sid[ods_idx];
		if (SpriteRecord::is_good_id(old_sid)) {
			IDType fid0;
			if (sid_map->get(old_sid, fid0)) {
				// char fname[1024];
				// id_to_name(tmp, fid0, fname, 1024);
				// fcache->unlink(fname);
				sid_map->del(old_sid);
			}
		}
		on_demand_sid[ods_idx]=sid;
		ods_idx=(ods_idx+1)%ODS_SIZE;

		fid=sid;

		sid_map->set(sid, fid);
		char fname[1024];
		id_to_name(tmp, fid, fname, 1024);
		fcache->truncate(fname, (off_t)65536); // guess an initial size.
		tmpi_error(DBG_APP, "On-demand file handle added: %s.\n", fname);
		return true;
	}
	else {
		return true;
	}
}

void TC_Tester::handle_record(SpriteRecord *r)
{
	int op=r->get_op();
	switch (op) {
		case SpriteRecord::OPEN:
			{
				bool ret=true;
				IDType fid;
				IDType sid;
				int size;
				ret = ret && (r->get_fid(fid));
				ret = ret && (r->get_sid(sid));
				ret = ret && (r->get_size(size));

				if (!ret) {
					tmpi_error(DBG_APP, "Bad OPEN record passed from parser.\n");
					break;
				}

				char fname[1024];
				id_to_name(tmp, fid, fname, 1024);
				if (!fcache->truncate(fname, (off_t)size)) {
					tmpi_error(DBG_APP, "Cannot open file %s.\n", fname);
					break;
				}

				sid_map->set(sid, fid);
				// tmpi_error(DBG_VERBOSE, "OPEN DONE.\n");
			}
			break;
		case SpriteRecord::CLOSE:
			{
				IDType sid;
				if (!r->get_sid(sid)) {
					tmpi_error(DBG_APP, "Bad CLOSE record passed from parser.\n");
					break;
				}

				sid_map->del(sid);
				// tmpi_error(DBG_VERBOSE, "CLOSE DONE.\n");
			}
			break;
		case SpriteRecord::CREATE:
			{
				IDType fid;
				if (!r->get_fid(fid)) {
					tmpi_error(DBG_APP, "Bad CREATE record passed from parser.\n");
					break;
				}

				char fname[1024];
				id_to_name(tmp, fid, fname, 1024);
				if (fcache->touch(fname)) {
					// tmpi_error(DBG_VERBOSE, "CREATE DONE.\n");
				}
				else
					tmpi_error(DBG_APP, "CREATE failed.\n");
			}
			break;
		case SpriteRecord::UNLINK:
			{
				IDType fid;
				if (!r->get_fid(fid)) {
					tmpi_error(DBG_APP, "Bad UNLINK record passed from parser.\n");
					break;
				}

				char fname[1024];
				id_to_name(tmp, fid, fname, 1024);
				if (fcache->unlink(fname)) {
					// tmpi_error(DBG_VERBOSE, "UNLINK DONE.\n");
				}
				else
					tmpi_error(DBG_APP, "UNLINK failed.\n");
			}
			break;
		case SpriteRecord::READ:
			{
				bool ret=true;
				int offset;
				int size;
				IDType sid;
				ret = ret && (r->get_offset(offset));
				ret = ret && (r->get_size(size));
				ret = ret && (r->get_sid(sid));

				if (!ret) {
					tmpi_error(DBG_APP, "Bad READ record passed from parser.\n");
					break;
				}

				IDType fid;
				if (!get_file_handle(sid, fid)) {
					break;
				}

				char fname[1024];
				id_to_name(tmp, fid, fname, 1024);

				char buf[1024];
				int len=size;
				while (len>0) {
					int len1=(len>1024)?1024:len;
					len1=fcache->read(fname, (long)offset, buf, len1);
					if (len1<=0)
						break;
					len-=len1;
					offset+=len1;
				}

				if (size-len>=0) {
					// tmpi_error(DBG_VERBOSE, "%d byes READ.\n", size-len);
				}
				else
					tmpi_error(DBG_APP, "READ failed.\n");
			}
			break;
		case SpriteRecord::WRITE:
			{
				bool ret=true;
				int offset;
				int size;
				IDType sid;
				ret = ret && (r->get_offset(offset));
				ret = ret && (r->get_size(size));
				ret = ret && (r->get_sid(sid));

				if (!ret) {
					tmpi_error(DBG_APP, "Bad WRITE record passed from parser.\n");
					break;
				}

				IDType fid;
				if (!get_file_handle(sid, fid)) {
					break;
				}

				char fname[1024];
				id_to_name(tmp, fid, fname, 1024);

				char buf[1024];
				int len=size;
				while (len>0) {
					int len1=(len>1024)?1024:len;
					len1=fcache->write(fname, (long)offset, buf, len1);
					if (len1<=0)
						break;
					len-=len1;
					offset+=len1;
				}

				if (size-len>0) {
					// tmpi_error(DBG_VERBOSE, "%d byes Written.\n", size-len);
				}
				else
					tmpi_error(DBG_APP, "WRITE failed.\n");
			}
			break;
		case SpriteRecord::SEEK:
			{
				// don't do anything for seek.
				// tmpi_error(DBG_VERBOSE, "SEEK done.\n");
			}
			break;
		case SpriteRecord::TRUNCATE:
			{
				IDType sid;
				int size;
				bool ret=true;
				ret = ret && (r->get_sid(sid));
				ret = ret && (r->get_size(size));
				if (!ret)
					break;

				IDType fid;
				if (!get_file_handle(sid, fid)) {
					break;
				}

				char fname[1024];
				id_to_name(tmp, fid, fname, 1024);

				if (fcache->truncate(fname, (off_t)size)) {
					// tmpi_error(DBG_VERBOSE, "TRUNCATE DONE.\n");
				}
				else
					tmpi_error(DBG_APP, "TRUNCATE failed.\n");
			}
			break;
		default:
			tmpi_error(DBG_APP, "Un-handled record.\n");
			break;
	}
}

static bool md_tmp(char *tmp)
{
	char dir_buf[1024];
	for (unsigned char i=0; i<0x10; i++) {
		unsigned char ci=hex_to_char(i);
		snprintf(dir_buf, 1024, "%s/%c", tmp, ci);
		mode_t m;
		if (get_mode(dir_buf, m)<0) {
			if (mkdir(dir_buf, 0770)<0)
				return false;

			if (get_mode(dir_buf, m)<0) {
				return false;
			}
		}

		if (!S_ISDIR(m)) {
			return false;
		}

		for (unsigned char j=0; j<0x10; j++) {
			unsigned char cj=hex_to_char(j);
			snprintf(dir_buf, 1024, "%s/%c/%c", tmp, ci, cj);
			mode_t m;
			if (get_mode(dir_buf, m)<0) {
				if (mkdir(dir_buf, 0770)<0)
					return false;

				if (get_mode(dir_buf, m)<0) {
					return false;
				}
			}

			if (!S_ISDIR(m)) {
				return false;
			}
		}
	}

	return true;
}

int main(int argc, char *argv[])
{
	prog=argv[0];

	int cache_size=512;
	char tmp[MAXPATHLEN]="";
	queue<char *>trace_queue;
	char *log=NULL;

	int c;
	while ((c=getopt(argc, argv, "n:f:t:l:h"))!=EOF) {
		switch (c) {
			case 'n':
				cache_size=atoi(optarg);
				break;
			case 'f':
				trace_queue.push(optarg);
				break;
			case 'l':
				if (log) {
					tmpi_error(DBG_CALLER, "log file is already specified.\n");
					do_help();
				}
				log=optarg;
				break;
			case 't':
				{
					if (tmp[0]) {
						tmpi_error(DBG_CALLER, 
								"temp directory specified multiple "
								"times in the command.\n");
						do_help();
					}
					if (realpath(optarg, tmp)==NULL) {
						tmpi_error(DBG_CALLER, "Cannot resolve the absolute path"
								" of the temp directory: %s.\n", optarg);
					}
				}
				break;
			case 'h':
				do_help();
				break;
			default:
				do_help();
				break;
		}
	}

	if (!tmp[0]) {
		strcpy(tmp, "/tmp");
	}

	for (; optind<argc; optind++) {
		trace_queue.push(argv[optind]);
	}

	if (trace_queue.empty()) {
		tmpi_error(DBG_VERBOSE, "No trace file specified in the command.\n");
		do_help();
	}

	if (trace_queue.size()>MAX_TRACE_FILES) {
		tmpi_error(DBG_CALLER, "Too many trace files specified in the command.\n");
		do_help();
	}

	mode_t tmp_m=0;
	get_mode(tmp, tmp_m);
	if (!S_ISDIR(tmp_m)) {
		tmpi_error(DBG_CALLER, "'%s' is not a directory.\n", tmp);
		do_help();
	}

	if (!md_tmp(tmp)) {
		tmpi_error(DBG_CALLER, "Cannot read/write tmp directory; %s.\n", tmp);
		exit(-1);
	}

	int total_traces=trace_queue.size();
	TC_Tester **thr_queue=(TC_Tester **)calloc
		(total_traces, sizeof(TC_Tester *));

	FILE *fp_log=NULL;
	if (log) {
		fp_log=fopen(log, "w");
	}

	if (fp_log)
		tmpi_set_output(fp_log);

	File_Cache fcache(cache_size);
	if (!thr_queue) {
		tmpi_error(DBG_FATAL, "out of memory.\n");
		exit(-1);
	}

	TC_StreamMap sid_map;

	int k=0;
	while (!trace_queue.empty()) {
		char *trace=trace_queue.front();
		trace_queue.pop();
		struct stat st;

		if (stat(trace, &st)) {
			tmpi_error(DBG_VERBOSE, "Cannot access trace file: %s.\n", trace);
			continue;
		}

		thr_queue[k]=new TC_Tester(trace, tmp, &fcache, &sid_map, fp_log);
		if (thr_queue[k]) {
			k++;
		}
		else {
			tmpi_error(DBG_INTERNAL, "Cannot create thread for trace: %s.\n", trace);
		}
	}

	if (k==0) {
		tmpi_error(DBG_VERBOSE, "Nothing to do.\n");
		goto _exit;
	}

	{
		usc_time_t usc_start=usc_gettime();
		for (int i=0; i<k; i++) {
			thr_queue[i]->start();
		}

		for (int i=0; i<k; i++) {
			thr_queue[i]->join();
		}

		usc_time_t usc_end=usc_gettime();

		fcache.report();
		printf("Total execution time: %.3fms.\n", usc_difftime(usc_end, usc_start)/1000);

		for (int i=0; i<k; i++) {
			printf("Thread [%d]: execution time: %.3fms.\n", i, thr_queue[i]->exe_time());
		}
	}

_exit:
	if (fp_log)
		fclose(fp_log);
	if (thr_queue) {
		for (int i=0; i<total_traces; i++) {
			if (thr_queue[i]) {
				delete thr_queue[i];
			}
		}
		free(thr_queue);
	}

	return 0;
}
