// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: file_cache.cc,v $ - file cache.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2004/04/06 09:48:27 $
 * Version :  $Revision: 1.15 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

# define _XOPEN_SOURCE 500
# include <unistd.h>

# include <sys/stat.h>
# include <sys/types.h>
# include <sys/vfs.h>
# include <sys/time.h>
# include <sys/resource.h>
# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <fcntl.h>
# include <pthread.h>

# include "../md5/md5.h"
# include "../debug/tmpi_debug.h"
# include "file_cache.h"


using namespace std;

// increase the rlimt of file descriptors to maximum.
static int set_rlimit_nofile_to_max()
{
	// attempt to set the OPEN FILE limit to
	// the system allowable limit 
	// and make sure that csize
	// is smaller than that limit.
	// In the worst case, when we are not able to get the
	// actual limit, the program will proceed with the danger
	// that we are not able to open files even though
	// cache is not full yet.
	struct rlimit rlim;
	if (getrlimit(RLIMIT_NOFILE, &rlim)==0) {
		int cur_ofn_lim=rlim.rlim_cur;
		if (rlim.rlim_cur<rlim.rlim_max) {
			rlim.rlim_cur=rlim.rlim_max;
			if (setrlimit(RLIMIT_NOFILE, &rlim)==0) {
				cur_ofn_lim=rlim.rlim_cur;
			}
		}

		return cur_ofn_lim;
	}
	else {
		return -1;
	}
}


# define TURNAROUND_FD_NUM 10

File_Cache::File_Cache(short size) :
	cache(NULL),
	csize(size),
	n_md5(0),
	n_miss(0),
	n_replace(0), 
	n_empty(0), 
	n_second_hit(0), 
	n_first_hit(0), 
	n_unlink(0), 
	n_pack(0)
{
	if (csize>0) {
		int ret=set_rlimit_nofile_to_max();
		if (ret>TURNAROUND_FD_NUM) {
			// leave 10 descriptors for turn-around
			if (csize>(ret-TURNAROUND_FD_NUM)) { 
				// user specified limit is too high, reduce it.
				csize=ret-TURNAROUND_FD_NUM;
			}
		}
		else {
			tmpi_error(DBG_VERBOSE, "The NOFILE rlimit is dangerously low (%d)!\n", ret);
		}

		pthread_mutex_init(&lock, NULL);
		cache=(Entry *)calloc(csize, sizeof(Entry));
		if (cache) {
			for (int i=0; i<csize; i++) {
				cache[i].init();
			}
		}
	}
}

File_Cache::~File_Cache()
{
	if (csize) {
		if (cache) {
			for (int i=0; i<csize; i++) {
				cache[i].finalize();
			}

			free(cache);
		}
		pthread_mutex_destroy(&lock);
	}
}

# define HIGH_BITS (~((unsigned int)(~0)>>4))
void File_Cache::pjw(char *path, unsigned int &sum)
{
	unsigned int h=0x5a827998;
	unsigned int g;

	for (; *path; ++path) {
		if ( (g = h & HIGH_BITS) )
			h = (h^(g>>24)) & ~HIGH_BITS;
		h = ( h << 4) + *path;
	}

	sum=h;
}

void File_Cache::md5(char *path, unsigned char *digest)
{
	Md5 md5;
	md5.stuff(path, strlen(path));
	md5.close();
	md5.pack(digest);
}

bool File_Cache::good_handle0(FC_Handle &handle, short idx)
{
	if (cache[idx].empty())
		return false;

	if (cache[idx].generation!=handle.gen())
		return false;

	return true;
}

bool File_Cache::good_handle(FC_Handle &handle)
{
	if (csize<=0) {
		return (handle.fd()>=0);
	}

	short idx=handle.idx();

	if ((idx<0)||(idx>=csize))
		return false;

	return good_handle0(handle, idx);
}

// If cur is a better replacement than old, return true.
// Otherwise, return false.
// replacable object = empty() or !in_use.
bool File_Cache::better_replacement(short cur, short old)
{
	// cur is an in-use handle.
	if ((!cache[cur].empty())&&(cache[cur].in_use>0)) {
		return false;
	}

	// cur is a replacable object,
	// and old is uninitialized (first call), choose cur.
	if (old<0)
		return true;

	// cur is a replacable object,
	// and old is an empty object, choose old.
	if (cache[old].empty())
		return false;

	// cur is an empty object,
	// and old is a replacable object, choose cur.
	if (cache[cur].empty())
		return true;

	// neither cur or old is an empty object.
	// both cur and old are replacable objects.
	// if (cache[cur].nhits<cache[old].nhits) {
		// return true;
	// }
	if (cache[cur].laccess<cache[old].laccess) {
		// cur's last access is older than old's, choose cur.
		return true;
	}

	// choose old
	return false;
}

# ifndef NUM_LOOK_AHEAD
# define NUM_LOOK_AHEAD 5
/* # define NUM_LOOK_AHEAD csize */
# endif

// searching without holding the lock.
short File_Cache::try_search(char *path, unsigned int &sum, bool &valid_digest, unsigned char *digest)
{
	short idx_start, idx=-1;
	idx_start=(short)(sum%(unsigned int)csize);
	bool found=false;

	for (unsigned int i=(unsigned int)idx_start; 
			i<(unsigned int)idx_start+NUM_LOOK_AHEAD; i++) {
		idx=(short)(i%(unsigned int)csize);
		if ((cache[idx].empty())||(cache[idx].sum!=sum)) {
			continue;
		}

		if (!valid_digest) {
			md5(path, digest);
			valid_digest=true;
			n_md5++;
		}

		if (memcmp(cache[idx].tag, digest, MD5_LEN)) {
			continue;
		}

		found=true;
		break;
	}

	return (found)?idx:-1;
}

# define FC_OPEN_FLAGS (O_RDWR|O_CREAT)
# define FC_OPEN_MODE (S_IRWXU|S_IRWXG)

FC_Handle File_Cache::fopen(char *path)
{
	if (csize<=0) {
		int fd=::open(path, FC_OPEN_FLAGS, FC_OPEN_MODE);
		return FC_Handle(fd);
	}

	unsigned int sum;
	pjw(path, sum);
	short idx_start;
	idx_start=(short)(sum%(unsigned int)csize);
	bool found=false;
	unsigned char digest[MD5_LEN];
	bool valid_digest=false;

	// a quick verification search, hopefully not much conflict will occur at this
	// level.
	short idx=try_search(path, sum, valid_digest, digest);
	if (idx>=0) {
		// potential hit.
		FC_Handle h_ret;
		bool good_to_return=false;
		if (pthread_mutex_lock(&(cache[idx].elock))<0) {
			return FC_Handle();
		}

		if ((sum==cache[idx].sum)&&(memcmp(digest, cache[idx].tag, MD5_LEN)==0)) {
			// verification check passed.
			if ((!cache[idx].deleted) && (cache[idx].in_use<0xffff)) {
				cache[idx].in_use++;
				time(&(cache[idx].laccess));
				// cache[idx].nhits++;
				// if (cache[idx].nhits==0) {
					// cache[idx].nhits--; // over-flow
					// the following op might cause
					// transient inconsistency in terms
					// of the LFU calculation, but
					// since it is so fast, the effect
					// shouldn't be too bad.
					// for (short i=0; i<csize; i++) {
						// cache[i].nhits >>= 1;
					// }
				// }
				h_ret=FC_Handle(cache[idx].generation, idx);
				good_to_return=true;
			}
			else {
				// in_use count too high.
				tmpi_error(DBG_INTERNAL, "Too many concurrent users.\n");
			}
		}
		else {
			// entry has been changed since try_search.
			// false alarm, continue with complete search.
		}

		pthread_mutex_unlock(&(cache[idx].elock));
		if (good_to_return) {
			n_first_hit++;
			return h_ret;
		}
	}

	if (pthread_mutex_lock(&lock)<0) {
		return FC_Handle();
	}

	short replacable=-1;
	// Now we failed the try_search phase. Do it through a complete pass.
	for (unsigned int i=(unsigned int)idx_start; 
			i<(unsigned int)idx_start+NUM_LOOK_AHEAD; i++) {
		idx=(short)(i%(unsigned int)csize);

		// We first need to check for potential cache hits.
		if ( (!cache[idx].empty()) && (cache[idx].sum == sum) ) {
			if (!valid_digest) {
				md5(path, digest);
				valid_digest=true;
				n_md5++;
			}
			if (memcmp(cache[idx].tag, digest, MD5_LEN) == 0) {
				found=true;
				break;
			}
		}

		if (better_replacement(idx, replacable)) {
			replacable=idx;
		}
	}

	FC_Handle h_ret; // an invalid handle

	if (found) {
		n_second_hit++;
		if (pthread_mutex_lock(&(cache[idx].elock))==0) {
			if ((!cache[idx].deleted) && (cache[idx].in_use<0xffff)) {
				cache[idx].in_use++;
				time(&(cache[idx].laccess));
				h_ret=FC_Handle(cache[idx].generation, idx);
			}
			else {
				tmpi_error(DBG_INTERNAL, "Too many concurrent users.\n");
			}
			pthread_mutex_unlock(&(cache[idx].elock));
		}
	}
	else {
		n_miss++;
		if (replacable>=0) {
			if (!valid_digest) {
				md5(path, digest);
				valid_digest=true;
				n_md5++;
			}

			if (pthread_mutex_lock(&(cache[replacable].elock))==0) {
				if (cache[replacable].empty()) {
					n_empty++;
				}
				else {
					n_replace++;
				}

				int fd=::open(path, FC_OPEN_FLAGS, FC_OPEN_MODE);
				if (fd>=0) {
					cache[replacable].set(digest, sum, fd);
					cache[replacable].in_use++;
					h_ret=FC_Handle(cache[replacable].generation, replacable);
				}
				else {
					tmpi_error(DBG_INTERNAL, "std::fopen failed: %s.\n", path);
				}
				pthread_mutex_unlock(&(cache[replacable].elock));
			}
		}
		else {
			tmpi_error(DBG_INTERNAL, "Cache full!.\n");
			// cache full
		}
	}

	pthread_mutex_unlock(&lock);

	return h_ret;
}

bool File_Cache::unlink(char *path, bool force)
{
	if (csize<=0) {
		return (::unlink(path)==0);
	}

	unsigned int sum;
	pjw(path, sum);
	short idx_start, idx=-1;
	idx_start=sum%csize;
	unsigned char digest[MD5_LEN];
	bool valid_digest=false;

	if (pthread_mutex_lock(&lock)<0)
		return false;

	n_unlink++;

	idx=try_search(path, sum, valid_digest, digest);
	if (idx<0) {
		bool ret=(::unlink(path)==0);
		pthread_mutex_unlock(&lock);
		return ret;
	}

	if (pthread_mutex_lock(&(cache[idx].elock))<0) {
		pthread_mutex_unlock(&lock);
		return false;
	}

	bool ret=false;

	if (!cache[idx].deleted)
		ret=(::unlink(path)==0);

	if (ret) {
		if ((!force)&&(cache[idx].in_use>0)) {
			cache[idx].deleted=true;
		}
		else {
			cache[idx].clear();
		}
	}

	pthread_mutex_unlock(&(cache[idx].elock));
	pthread_mutex_unlock(&lock);

	return ret;
}

bool File_Cache::fclose(FC_Handle &handle)
{
	if (csize<=0) {
		int fd=handle.fd();
		int ret=(fd>=0)?::close(fd):-1;
		handle=FC_Handle();
		return (ret==0);
	}

	short idx=handle.idx();

	if ((idx<0)||(idx>=csize))
		return false;

	if (pthread_mutex_lock(&(cache[idx].elock))<0) {
		return false;
	}

	if (cache[idx].in_use==0) {
		pthread_mutex_unlock(&cache[idx].elock);
		return false;
	}

	if (!good_handle0(handle, idx)) {
		pthread_mutex_unlock(&cache[idx].elock);
		return false;
	}

	cache[idx].in_use--;

	if ((cache[idx].deleted==true)&&(cache[idx].in_use==0)) {
		cache[idx].clear();
	}

	pthread_mutex_unlock(&cache[idx].elock);

	handle=FC_Handle();

	return true;
}

int File_Cache::read0(int fd, unsigned int offset, char *data, unsigned int size)
{
	if (size==0)
		return 0;

	if ((!data)||(fd<0)) 
		return -1;

	return pread(fd, data, (size_t)size, (off_t)offset);
}

int File_Cache::fread(FC_Handle &handle, unsigned int offset, char *data, unsigned int size)
{
	if (csize<=0) {
		int fd=handle.fd();
		return read0(fd, offset, data, size);
	}
	else {
		short idx=handle.idx();

		if ((idx<0)||(idx>=csize))
			return -1;

		if (pthread_mutex_lock(&(cache[idx].elock))<0)
			return -1;

		if (cache[idx].in_use==0) {
			pthread_mutex_unlock(&(cache[idx].elock));
			return -1;
		}

		if (!good_handle0(handle, idx)) {
			pthread_mutex_unlock(&(cache[idx].elock));
			return -1;
		}

		int ret=read0(cache[idx].fd, offset, data, size);
		pthread_mutex_unlock(&(cache[idx].elock));
		return ret;
	}
}

int File_Cache::write0(int fd, unsigned int offset, char *data, unsigned int size)
{
	if (size==0)
		return 0;

	if ((!data)||(fd<0))
		return -1;

	return pwrite(fd, data, (size_t)size, (off_t)offset);
}

int File_Cache::fwrite(FC_Handle &handle, unsigned int offset, char *data, unsigned int size)
{
	if (csize<=0) {
		int fd=handle.fd();
		return write0(fd, offset, data, size);
	}
	else {
		short idx=handle.idx();
		if ((idx<0)&&(idx>=csize))
			return -1;

		if (pthread_mutex_lock(&(cache[idx].elock))<0)
			return -1;

		if (cache[idx].in_use==0) {
			pthread_mutex_unlock(&(cache[idx].elock));
			return -1;
		}

		if (!good_handle0(handle, idx)) {
			pthread_mutex_unlock(&(cache[idx].elock));
			return -1;
		}

		int ret=write0(cache[idx].fd, offset, data, size);
		cache[idx].clean=false;
		pthread_mutex_unlock(&(cache[idx].elock));
		return ret;
	}
}

int File_Cache::append0(int fd, char *data, unsigned int size)
{
	if (size==0)
		return 0;

	if ((!data)||(fd<0))
		return -1;

	struct stat st;
	if (fstat(fd, &st) < 0)
		return -1;

	return pwrite(fd, data, (size_t)size, (off_t)st.st_size);
}

int File_Cache::fappend(FC_Handle &handle, char *data, size_t size)
{
	if (csize<=0) {
		int fd=handle.fd();
		return append0(fd, data, size);
	}
	else {
		short idx=handle.idx();
		if ((idx<0)&&(idx>=csize))
			return -1;

		if (pthread_mutex_lock(&(cache[idx].elock))<0)
			return -1;

		if (cache[idx].in_use==0) {
			pthread_mutex_unlock(&(cache[idx].elock));
			return -1;
		}

		if (!good_handle0(handle, idx)) {
			pthread_mutex_unlock(&(cache[idx].elock));
			return -1;
		}

		int ret=append0(cache[idx].fd, data, size);
		cache[idx].clean=false;
		pthread_mutex_unlock(&(cache[idx].elock));
		return ret;
	}
}

bool File_Cache::truncate0(int fd, unsigned int size, bool touch)
{
	if ((fd<0)||(size<0))
		return false;

	unsigned int old_size=0;
	unsigned int bsize=1024;

	if (touch) {
		struct stat st;
		if (fstat(fd, &st)<0)
			return false;
		old_size=(unsigned int)st.st_size;

		struct statfs stf;
		if (fstatfs(fd, &stf)<0)
			return false;
		bsize=(unsigned int)stf.f_bsize;
	}

	if (::ftruncate(fd, (off_t)size)<0)
		return false;

	if (touch) {
		for (unsigned int i=old_size; i<size; i+=bsize) {
			if (write0(fd, i, "", 1)<0)
				return false;
		}
	}

	return true;
}

bool File_Cache::ftruncate(FC_Handle &handle, off_t size, bool touch)
{
	if (csize<=0) {
		int fd=handle.fd();

		return truncate0(fd, size, touch);
	}
	else {
		short idx=handle.idx();
		if ((idx<0)&&(idx>=csize))
			return false;

		if (pthread_mutex_lock(&(cache[idx].elock))<0)
			return false;

		if (cache[idx].in_use==0) {
			pthread_mutex_unlock(&(cache[idx].elock));
			return false;
		}

		if (!good_handle0(handle, idx)) {
			pthread_mutex_unlock(&(cache[idx].elock));
			return false;
		}

		bool ret=truncate0(cache[idx].fd, size, touch);
		cache[idx].clean=false;
		pthread_mutex_unlock(&(cache[idx].elock));

		return ret;
	}
}

bool File_Cache::flush0(int fd)
{
	return (fdatasync(fd)==0);
}

bool File_Cache::fflush(FC_Handle &handle)
{
	if (csize<=0) {
		int fd=handle.fd();

		return flush0(fd);
	}
	else {
		short idx=handle.idx();
		if ((idx<0)&&(idx>=csize))
			return false;

		if (pthread_mutex_lock(&(cache[idx].elock))<0)
			return false;

		if (cache[idx].in_use==0) {
			pthread_mutex_unlock(&(cache[idx].elock));
			return false;
		}

		if (!good_handle0(handle, idx)) {
			pthread_mutex_unlock(&(cache[idx].elock));
			return false;
		}

		if (!cache[idx].clean) {
			cache[idx].clean=flush0(cache[idx].fd);
		}

		bool ret=cache[idx].clean;

		pthread_mutex_unlock(&(cache[idx].elock));

		return ret;
	}
}

int File_Cache::read(char *path, unsigned int offset, char *data, unsigned int size)
{
	FC_Handle h=fopen(path);
	int ret=fread(h, offset, data, size);
	fclose(h);

	return ret;
}

int File_Cache::write(char *path, unsigned int offset, char *data, unsigned int size)
{
	FC_Handle h=fopen(path);
	int ret=fwrite(h, offset, data, size);
	fclose(h);

	return ret;
}

int File_Cache::append(char *path, char *data, unsigned int size)
{
	FC_Handle h=fopen(path);
	int ret=fappend(h, data, size);
	fclose(h);

	return ret;
}

bool File_Cache::truncate(char *path, unsigned int size, bool touch)
{
	FC_Handle h=fopen(path);
	bool ret=ftruncate(h, size, touch);
	fclose(h);

	return ret;
}

bool File_Cache::flush(char *path)
{
	FC_Handle h=fopen(path);
	int ret=fflush(h);
	fclose(h);

	return ret;
}

bool File_Cache::touch(char *path)
{
	FC_Handle h=fopen(path);
	return fclose(h);
}

int File_Cache::pack()
{
	if (csize<=0) {
		return 0;
	}

	if (pthread_mutex_lock(&lock)<0)
		return -1;

	n_pack++;

	int k=0;
	for (short i=0; i<csize; i++) {
		if (pthread_mutex_lock(&(cache[i].elock))<0) {
			continue;
		}

		if ((!cache[i].empty())&&(cache[i].in_use==0)) {
			cache[i].clear();
			k++;
		}

		pthread_mutex_unlock(&(cache[i].elock));
	}

	pthread_mutex_unlock(&lock);

	return k;
}

void File_Cache::sync()
{
	if (csize<=0) {
		return;
	}

	for (short i=0; i<csize; i++) {
		if (pthread_mutex_lock(&(cache[i].elock))<0) {
			continue;
		}

		if ((!cache[i].empty())&&(!cache[i].clean)) {
			cache[i].clean=(fdatasync(cache[i].fd)==0);
		}

		pthread_mutex_unlock(&(cache[i].elock));
	}

	return;
}

void File_Cache::report()
{
	unsigned int total_hits=n_first_hit+n_second_hit;
	if ((csize>0)&&((total_hits+n_miss)>0)) {
		printf("CACHE_SIZE=%d, UNLINK=%d, PACK=%d, MD5=%d\n"
			"1HIT=%d, 2HIT=%d, MISS=%d (EMISS=%d, RMISS=%d), "
			"hit rate=%.2f%%\n", 
			csize, n_unlink, n_pack, n_md5, 
			n_first_hit, n_second_hit, n_miss, n_empty, n_replace, 
			(double)(total_hits)*100.0/(total_hits+n_miss));
	}
}
