// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: file_cache.h,v $ - header file for file cache. 
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2004/04/06 09:48:27 $
 * Version :  $Revision: 1.14 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

# ifndef _FILE_CACHE_H_
# define _FILE_CACHE_H_

# include <unistd.h>
# include <sys/types.h>
# include <pthread.h>
# include <string.h>
# include <time.h>
# include "../md5/md5.h"

class FC_Handle {
 private:
	union {
		int fd;
		struct {
			unsigned short gen;
			short idx;
		} s;
	} u;

 public:
	FC_Handle(int fd) { u.fd=fd; return;}
	FC_Handle(unsigned short gen, short idx) 
	{ u.s.gen=gen; u.s.idx=idx; return;}
	FC_Handle() { u.s.idx=-1; return;}

	int & fd() { return u.fd; }
	unsigned short & gen() { return u.s.gen; }
	short & idx() { return u.s.idx; }
};

class File_Cache {
private:
	struct Entry {
		unsigned int sum; /* sum */
		char tag[MD5_LEN];
		int fd;
		// unsigned int nhits; // hits
		time_t laccess; // last access time, I think LRU makes more sense than LFU.
		unsigned short in_use; // share count
		unsigned short generation; // used in forced clear.
		pthread_mutex_t elock;
		bool clean;
		bool deleted; // deferred delete

		void init() {
			memset(this, 0, sizeof(Entry));
			pthread_mutex_init(&elock, NULL);
			fd = -1;
		}

		void finalize() {
			clear();
			pthread_mutex_destroy(&elock);
		}

		bool empty() { return (fd<0); }

		// it must be called when the lock is acquired.
		void set(unsigned char *digest, unsigned int sum, int fd) {
			this->sum=sum;
			memcpy(this->tag, digest, MD5_LEN);
			// nhits=0;
			time(&laccess);
			if (this->fd >= 0)
				::close(this->fd);
			this->fd=fd;
			in_use=0;
			// nhits=0;
			generation++;
			clean=true;
			deleted=false;
		}

		// it must be called when the lock is acquired.
		void clear() {
			sum=0;
			memset(tag, 0, MD5_LEN);
			if (fd) {
				::close(fd);
				fd=-1;
			}
			// in_use=0;
			// clean=true;
			// deleted=false;
			// last_op=0;
		}
	};

	Entry *cache; // all associative, LFU replacement.
	short csize;
	unsigned int n_md5; // md5 calculation
	unsigned int n_miss; // total misses
	unsigned int n_replace; // capacity miss/first seen miss
	unsigned int n_empty; // cold start miss
	unsigned int n_second_hit; // found after acquiring the global lock
	unsigned int n_first_hit; // found without acquiring the global lock
	unsigned int n_unlink; // number of unlink op (require global lock)
	unsigned int n_pack; // number of pack op (require global lock)
	pthread_mutex_t lock;

	static void md5(char *path, unsigned char *digest);
	static void pjw(char *path, unsigned int &sum);
	bool better_replacement(short cur, short old);
	short try_search(char *path, unsigned int &sum, bool &valid_digest, unsigned char *digest);

	int read0(int fd, unsigned int offset, char *data, unsigned int size);
	int write0(int fd, unsigned int offset, char *data, unsigned int size);
	int append0(int fd, char *data, unsigned int size);
	bool truncate0(int fd, unsigned int size, bool touch);
	bool flush0(int fd);
	bool good_handle0(FC_Handle &handle, short idx); // check validity of a handle.

 public:
	// constructor/destructors
	File_Cache(short size=512); // size<=0, disable cache
	~File_Cache();

	// operations on path
	int read(char *path, unsigned int offset, char *data, unsigned int size);
	int write(char *path, unsigned int offset, char *data, unsigned int size);
	int append(char *path, char *data, unsigned int size);
	bool truncate(char *path, unsigned int size, bool touch=false);
	bool flush(char *path);
	bool touch(char *path);

	bool unlink(char *path, bool force=false); // remove entries in the cache.

	// operations on handles
	FC_Handle fopen(char *path); // return a handle that can be used subsequently
	bool fclose(FC_Handle &handle); // declare that the handle will not be used.
	bool good_handle(FC_Handle &handle); // check validity of a handle.

	int fread(FC_Handle &handle, unsigned int offset, char *data, unsigned int size);
	int fwrite(FC_Handle &handle, unsigned int offset, char *data, unsigned int size);
	int fappend(FC_Handle &handle, char *data, unsigned int size);
	bool ftruncate(FC_Handle &handle, off_t size, bool touch=false);
	bool fflush(FC_Handle &handle);

	// if other applications receive errors due to insufficient file descriptors
	// we can free some through pack() call.
	int pack();
	void sync();

	// cache performance report
	void report();
};

# endif
