// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: trace_analyzer.cc,v $ - Sprite trace analyzer.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/04/09 09:59:58 $
 * Version :  $Revision: 1.7 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

# include <sys/types.h>
# include <sys/stat.h>
# include <sys/param.h>
# include <unistd.h>
# include <stdio.h>
# include <stdlib.h>
# include <getopt.h>
# include <queue>
# include <set>

# include "../usc/usc_time.h"
# include "../../expr/sprite/sprite.h"
# include "../debug/tmpi_debug.h"

using namespace std;

/**
 * A dummy program that reads a trace and make calls
 * to File_Cache.
 * fc_test -n <cache_size> -f <trace list ...>
 */

static char *prog=NULL;

static void do_help()
{
	printf("\
Sprite Trace Analyzer.\n\
Usage: %s [options] [trace list ...].\n\
   -f <trace>             Trace file.\n\
   -h                     Help, this screen.\n\
\n\
Multiple trace files can be specified by the command \n\
line, and the analyzer will generate a report for the\n\
aggregated traces only. \n\
", prog);
	exit(0);
}

struct IDType_cmp : public less<IDType> {
	bool operator()(const IDType &x, const IDType &y) const {
		return(memcmp(x.p+1, y.p+1, sizeof(int)*3)<0);
	}
};

typedef set<IDType, IDType_cmp> ID_SET;

struct TraceStat {
	unsigned int n_rec;
	unsigned int n_open;
	unsigned int n_close;
	unsigned int n_unlink;
	unsigned int n_create;
	unsigned int n_read;
	unsigned int n_write;
	unsigned int n_seek;
	unsigned int n_truncate;
	unsigned int n_files; // files
	unsigned int n_ufiles; // failed unlink
	unsigned int n_umstreams;// un-mapped stream.
	unsigned int n_mstreams; // mapped stream.
	double total_time;
	ID_SET sid_set;
	ID_SET fid_set;

public:
	TraceStat() :
	n_open(0),
	n_close(0),
	n_unlink(0),
	n_create(0),
	n_read(0),
	n_write(0),
	n_seek(0),
	n_truncate(0),
	n_files(0),
	n_ufiles(0), // bad unlink
	n_umstreams(0),
	n_mstreams(0),
	total_time(0)
	{}

	void report(char *caption)
	{
		int n_rec=n_open+n_close+n_unlink+n_create+n_read+n_write+n_seek+n_truncate;
		if (n_rec>0) {
			printf("%20s: OPEN=%.3f%%(%d) CLOSE=%.3f%%(%d) UNLINK=%.3f%%(%d) CREATE=%.3f%%(%d)\n",
					caption, (double)n_open*100/n_rec, n_open, (double)n_close*100/n_rec, n_close,
					(double)n_unlink*100/n_rec, n_unlink, (double)n_create*100/n_rec, n_create);
			printf("%20s: READ=%.3f%%(%d) WRITE=%.3f%%(%d) SEEK=%.3f%%(%d) TRUNCATE=%.3f%%(%d)\n",
					caption, (double)n_read*100/n_rec, n_read, (double)n_write*100/n_rec, n_write,
					(double)n_seek*100/n_rec, n_seek, (double)n_truncate*100/n_rec, n_truncate);
			printf("%20s: N_FILE=%d N_STREAM=%d B_UNLINK=%d B_STREAM=%d MIN_MISS=%d, TIME=%.3fsec\n\n",
					caption, n_files, n_mstreams, n_ufiles, n_umstreams, (n_files+n_umstreams), total_time);
		}
	}
};

void handle_record(TraceStat *ts, SpriteRecord *r)
{
	int op=r->get_op();
	switch (op) {
		case SpriteRecord::OPEN:
			{
				ts->n_open++;
				IDType fid;
				IDType sid;
				bool ret=true;
				ret = ret && (r->get_fid(fid));
				ret = ret && (r->get_sid(sid));

				if (!ret) {
					tmpi_error(DBG_APP, "Bad OPEN record passed from parser.\n");
					break;
				}

				ID_SET::iterator pi=ts->fid_set.find(fid);
				bool fid_found=true;
				if (pi==ts->fid_set.end()) {
					fid_found=false;
					ts->fid_set.insert(fid);
					ts->n_files++;
					ts->n_mstreams++;
				}
				
				pi=ts->sid_set.find(sid);
				if (pi==ts->sid_set.end()) {
					ts->sid_set.insert(sid);
					if (fid_found) {
						ts->n_mstreams++;
					}
				}
			}
			break;
		case SpriteRecord::CLOSE:
			{
				ts->n_close++;

				IDType sid;
				if (!r->get_sid(sid)) {
					tmpi_error(DBG_APP, "Bad CLOSE record passed from parser.\n");
					break;
				}

				ID_SET::iterator pi=ts->sid_set.find(sid);
				if (pi==ts->sid_set.end()) {
					ts->n_umstreams++;
				}
				else {
					ts->sid_set.erase(pi);
				}
			}
			break;
		case SpriteRecord::CREATE:
			{
				ts->n_create++;
				IDType fid;
				if (!r->get_fid(fid)) {
					tmpi_error(DBG_APP, "Bad CREATE record passed from parser.\n");
					break;
				}

				ID_SET::iterator pi=ts->fid_set.find(fid);
				if (pi==ts->fid_set.end()) {
					ts->fid_set.insert(fid);
					ts->n_files++;
				}
			}
			break;
		case SpriteRecord::UNLINK:
			{
				ts->n_unlink++;
				IDType fid;
				if (!r->get_fid(fid)) {
					tmpi_error(DBG_APP, "Bad UNLINK record passed from parser.\n");
					break;
				}

				ID_SET::iterator pi=ts->fid_set.find(fid);
				if (pi==ts->fid_set.end()) {
					ts->n_ufiles++;
				}
				else {
					ts->fid_set.erase(pi);
				}
			}
			break;
		case SpriteRecord::READ:
		case SpriteRecord::WRITE:
		case SpriteRecord::SEEK:
		case SpriteRecord::TRUNCATE:
			{
				if (op==SpriteRecord::READ)
					ts->n_read++;
				else if (op==SpriteRecord::WRITE)
					ts->n_write++;
				else if (op==SpriteRecord::SEEK)
					ts->n_seek++;
				else
					ts->n_truncate++;

				bool ret=true;
				IDType sid;
				ret = ret && (r->get_sid(sid));

				if (!ret) {
					tmpi_error(DBG_APP, "Bad READ/WRITE record passed from parser.\n");
					break;
				}

				ID_SET::iterator pi=ts->sid_set.find(sid);
				if (pi==ts->sid_set.end()) {
					ts->sid_set.insert(sid);
					ts->n_umstreams++;
				}
			}
			break;
		default:
			tmpi_error(DBG_APP, "Un-handled record: op=%d.\n", op);
			r->dump();
			break;
	}
}

int main(int argc, char *argv[])
{
	prog=argv[0];

	queue<char *>trace_queue;

	int c;
	while ((c=getopt(argc, argv, "f:h"))!=EOF) {
		switch (c) {
			case 'f':
				trace_queue.push(optarg);
				break;
			case 'h':
				do_help();
				break;
			default:
				do_help();
				break;
		}
	}

	for (; optind<argc; optind++) {
		trace_queue.push(argv[optind]);
	}

	if (trace_queue.empty()) {
		tmpi_error(DBG_VERBOSE, "No trace file specified in the command.\n");
		do_help();
	}

	TraceStat aggregate;
	int n_traces=0;
	while (!trace_queue.empty()) {
		char *trace=trace_queue.front();
		TraceStat per_trace;
		trace_queue.pop();
		SpriteParser parser(trace);
		SpriteRecord *r;
		while ((r=parser.read_record())) {
			handle_record(&aggregate, r);
			handle_record(&per_trace, r);
			parser.free_record(r);
		}
		double tt=0;
		if (!parser.total_time(tt)) {
			tmpi_error(DBG_INTERNAL, "Cannot get trace interval time.\n");
		}
		else {
			aggregate.total_time+=tt;
			per_trace.total_time=tt;
			char *p=strrchr(trace, '/');
			if (p)
				p++;
			else
				p=trace;
			per_trace.report(p);
			n_traces++;
		}
	}

	if (n_traces>1)
		aggregate.report("Total");

	return 0;
}
