// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: guid.cc,v $ - Global unique ID (GUID).
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/05/06 01:53:25 $
 * Version :  $Revision: 1.5 $
 *
 * Based on GUID implementation in ext2 file system code.
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

# include <sys/types.h>
# include <sys/socket.h>
// # include <sys/sockio.h>
# include <sys/stat.h>
# include <sys/file.h>
# include <sys/time.h>
# include <sys/ioctl.h>
# include <net/if.h>
# include <netinet/in.h>
# include <fcntl.h>
# include <unistd.h>
# include <stdlib.h>
# include <string.h>
# include <errno.h>
# include <ctype.h>
# include <stdio.h>

# include "guid.h"

const guid_t Guid::GUID_NIL=
{
{ 	0, 0, 0, 0,
	0, 0, 0, 0,
	0, 0, 0, 0,
	0, 0, 0, 0 	}
};

int	Guid::rfd=-2;

int Guid::get_random_fd(void)
{
	struct timeval	tv;
	int		i;

	if (rfd == -2) {
		gettimeofday(&tv, 0);
		rfd = open("/dev/urandom", O_RDONLY);
		if (rfd == -1)
			rfd = open("/dev/random", O_RDONLY | O_NONBLOCK);
		srand((getpid() << 16) ^ getuid() ^ tv.tv_sec ^ tv.tv_usec);
		/* Crank the random number generator a few times */
		gettimeofday(&tv, 0);
		for (i = (tv.tv_sec ^ tv.tv_usec) & 0x1F; i > 0; i--)
			rand();
	}
	return rfd;
}


/*
 * Generate a series of random bytes.  Use /dev/urandom if possible,
 * and if not, use srandom/random.
 */
void Guid::get_random_bytes(void *buf, int nbytes)
{
	int i, fd = get_random_fd();
	int lose_counter = 0;
	char *cp = (char *) buf;

	if (fd >= 0) {
		while (nbytes > 0) {
			i = read(fd, cp, nbytes);
			if ((i < 0) &&
			    ((errno == EINTR) || (errno == EAGAIN)))
				continue;
			if (i <= 0) {
				if (lose_counter++ == 8)
					break;
				continue;
			}
			nbytes -= i;
			cp += i;
			lose_counter = 0;
		}
	}

	/* XXX put something better here if no /dev/random! */
	for (i = 0; i < nbytes; i++)
		*cp++ = rand() & 0xFF;
	return;
}

static inline int max(int a, int b)
{
	return  ((a) > (b) ? (a) : (b));
}

/*
 * Get the ethernet hardware address, if we can find it...
 */
static int get_node_id(unsigned char *node_id)
{
	int	sd;
	struct ifreq ifr, *ifrp;
	struct ifconf ifc;
	char buf[1024];
	int	n, i;
	unsigned char 	*a;
	
#define ifreq_size(i) sizeof(struct ifreq)

	sd = socket(AF_INET, SOCK_DGRAM, IPPROTO_IP);
	if (sd < 0) {
		return -1;
	}
	memset(buf, 0, sizeof(buf));
	ifc.ifc_len = sizeof(buf);
	ifc.ifc_buf = buf;
	if (ioctl (sd, SIOCGIFCONF, (char *)&ifc) < 0) {
		close(sd);
		return -1;
	}
	n = ifc.ifc_len;
	for (i = 0; i < n; i+= ifreq_size(ifr) ) {
		ifrp = (struct ifreq *)((char *) ifc.ifc_buf+i);
		strncpy(ifr.ifr_name, ifrp->ifr_name, IFNAMSIZ);
#ifdef SIOCGIFHWADDR
		if (ioctl(sd, SIOCGIFHWADDR, &ifr) < 0)
			continue;
		a = (unsigned char *) &ifr.ifr_hwaddr.sa_data;
#else
#ifdef SIOCGENADDR
		if (ioctl(sd, SIOCGENADDR, &ifr) < 0)
			continue;
		a = (unsigned char *) ifr.ifr_enaddr;
#else
		/*
		 * XXX we don't have a way of getting the hardware
		 * address
		 */
		close(sd);
		return 0;
#endif /* SIOCGENADDR */
#endif /* SIOCGIFHWADDR */
		if (!a[0] && !a[1] && !a[2] && !a[3] && !a[4] && !a[5])
			continue;
		if (node_id) {
			memcpy(node_id, a, 6);
			close(sd);
			return 1;
		}
	}
	close(sd);
	return 0;
}

/* Assume that the gettimeofday() has microsecond granularity */
#define MAX_ADJUSTMENT 10

static int get_clock(unsigned int *clock_high, unsigned int *clock_low, unsigned short *ret_clock_seq)
{
	static int			adjustment = 0;
	static struct timeval		last = {0, 0};
	static unsigned short			clock_seq;
	struct timeval 			tv;
	unsigned long long		clock_reg;
	
try_again:
	gettimeofday(&tv, 0);
	if ((last.tv_sec == 0) && (last.tv_usec == 0)) {
		Guid::get_random_bytes(&clock_seq, sizeof(clock_seq));
		clock_seq &= 0x1FFF;
		last = tv;
		last.tv_sec--;
	}
	if ((tv.tv_sec < last.tv_sec) ||
	    ((tv.tv_sec == last.tv_sec) &&
	     (tv.tv_usec < last.tv_usec))) {
		clock_seq = (clock_seq+1) & 0x1FFF;
		adjustment = 0;
		last = tv;
	} else if ((tv.tv_sec == last.tv_sec) &&
	    (tv.tv_usec == last.tv_usec)) {
		if (adjustment >= MAX_ADJUSTMENT)
			goto try_again;
		adjustment++;
	} else {
		adjustment = 0;
		last = tv;
	}
		
	clock_reg = tv.tv_usec*10 + adjustment;
	clock_reg += ((unsigned long long) tv.tv_sec)*10000000;
	clock_reg += (((unsigned long long) 0x01B21DD2) << 32) + 0x13814000;

	*clock_high = clock_reg >> 32;
	*clock_low = clock_reg;
	*ret_clock_seq = clock_seq;
	return 0;
}

/**
 * Member functions:
 */

void Guid::gen_time()
{
	static unsigned char node_id[6];
	static int has_init = 0;
	unsigned int	clock_mid;

	if (!has_init) {
		if (get_node_id(node_id) <= 0) {
			get_random_bytes(node_id, 6);
			/*
			 * Set multicast bit, to prevent conflicts
			 * with IEEE 802 addresses obtained from
			 * network cards
			 */
			node_id[0] |= 0x80;
		}
		has_init = 1;
	}
	get_clock(&clock_mid, &time_low, &clock_seq);
	clock_seq |= 0x8000;
	time_mid = (unsigned short) clock_mid;
	time_hi_and_version = (clock_mid >> 16) | 0x1000;
	memcpy(node, node_id, 6);
	status=IN_FIELD;
}

void Guid::gen_random()
{
	get_random_bytes(u.packed, sizeof(u.packed));
	unpack();

	clock_seq = (clock_seq & 0x3FFF) | 0x8000;
	time_hi_and_version = (time_hi_and_version & 0x0FFF) | 0x4000;
	status=IN_FIELD;
}

/*
 * This is the generic front-end to gen_random and
 * gen_time.  It uses gen_random only if
 * /dev/urandom is available, since otherwise we won't have
 * high-quality randomness.
 */
void Guid::gen()
{
# if defined(__FORCE_TIME_GEN__)
	gen_time();
	return;
# endif

	if (get_random_fd() >= 0)
		gen_random();
	else
		gen_time();
}

time_t Guid::get_time(struct timeval *ret_tv)
{
	unsigned int			high;
	struct timeval		tv;
	unsigned long long	clock_reg;

	if (status==IN_NONE)
		return (time_t)-1;

	if (status==IN_PACKED) {
		unpack();
	}
	
	high = time_mid | ((time_hi_and_version & 0xFFF) << 16);
	clock_reg = time_low | ((unsigned long long) high << 32);

	clock_reg -= (((unsigned long long) 0x01B21DD2) << 32) + 0x13814000;
	tv.tv_sec = clock_reg / 10000000;
	tv.tv_usec = (clock_reg % 10000000) / 10;

	if (ret_tv)
		*ret_tv = tv;

	return tv.tv_sec;
}

int Guid::pack(void *ptr, int len)
{
	if ((len<GUID_LEN)||(status==IN_NONE))
		return -1;

	if (status==IN_FIELD) {
		pack();
	}

	memcpy(ptr, u.packed, GUID_LEN);
	return 0;
}

void Guid::pack()
{
	unsigned int	tmp;

	tmp = time_low;
	u.packed[3] = (unsigned char) tmp;
	tmp >>= 8;
	u.packed[2] = (unsigned char) tmp;
	tmp >>= 8;
	u.packed[1] = (unsigned char) tmp;
	tmp >>= 8;
	u.packed[0] = (unsigned char) tmp;
	
	tmp = time_mid;
	u.packed[5] = (unsigned char) tmp;
	tmp >>= 8;
	u.packed[4] = (unsigned char) tmp;

	tmp = time_hi_and_version;
	u.packed[7] = (unsigned char) tmp;
	tmp >>= 8;
	u.packed[6] = (unsigned char) tmp;

	tmp = clock_seq;
	u.packed[9] = (unsigned char) tmp;
	tmp >>= 8;
	u.packed[8] = (unsigned char) tmp;

	memcpy(u.packed+10, node, 6);
	status=IN_BOTH;
}

int Guid::unpack(const void *guid, int len)
{
	if ((len!=GUID_LEN))
		return -1;

	memcpy(u.packed, guid, len);
	
	// this is added to facilitate lazy unpacking
	status=IN_PACKED;
	// unpack(); // this is commented for lazy unpacking.
	
# if defined(__DEBUG__)
	// The following two statements are for testing purposese only
	// to force us to discard the cached packed copy
	unpack();
	status=IN_FIELD;
# endif
	
	return 0;
}

void Guid::unpack()
{
	const unsigned char	*ptr = u.packed;
	unsigned int		tmp;

	tmp = *ptr++;
	tmp = (tmp << 8) | *ptr++;
	tmp = (tmp << 8) | *ptr++;
	tmp = (tmp << 8) | *ptr++;
	time_low = tmp;

	tmp = *ptr++;
	tmp = (tmp << 8) | *ptr++;
	time_mid = tmp;
	
	tmp = *ptr++;
	tmp = (tmp << 8) | *ptr++;
	time_hi_and_version = tmp;

	tmp = *ptr++;
	tmp = (tmp << 8) | *ptr++;
	clock_seq = tmp;

	memcpy(node, ptr, 6);

	status=IN_BOTH;
}

int Guid::parse(const char *szIn)
{
	int 		i;
	const char	*cp;
	char		buf[3];

	if (strlen(szIn) != 36)
		return -1;

	// Check the format.
	for (i=0, cp = szIn; i <= 36; i++,cp++) {
		if ((i == 8) || (i == 13) || (i == 18) || (i == 23))
			if (*cp == '-')
				continue;
		if (i== 36)
			if (*cp == 0)
				continue;
		if (!isxdigit(*cp))
			return -1;
	}

	// convert it
	time_low = strtoul(szIn, NULL, 16);
	time_mid = strtoul(szIn+9, NULL, 16);
	time_hi_and_version = strtoul(szIn+14, NULL, 16);
	clock_seq = strtoul(szIn+19, NULL, 16);
	cp = szIn+24;
	buf[2] = 0;
	for (i=0; i < 6; i++) {
		buf[0] = *cp++;
		buf[1] = *cp++;
		node[i] = strtoul(buf, NULL, 16);
	}
	
	status=IN_FIELD;

	return 0;
}

int Guid::to_string(char *szOut, int len)
{
	if ((status==IN_NONE) || (len<37))
		return -1;

	if (status==IN_PACKED)
		unpack();

	sprintf(szOut,
		"%08x-%04x-%04x-%02x%02x-%02x%02x%02x%02x%02x%02x",
		time_low, time_mid, time_hi_and_version,
		clock_seq >> 8, clock_seq & 0xFF,
		node[0], node[1], node[2],
		node[3], node[4], node[5]);

	return 0;
}


# define UUCMP(u1,u2) \
do { \
	if (u1 != u2) \
		return((u1 < u2) ? -1 : 1); \
} while (0)

int Guid::compare(Guid *oGuid)
{
	if (status==IN_NONE) {
		if (oGuid->status==IN_NONE)
			return 0;
		return -1;
	}
	if (oGuid->status==IN_NONE)
		return 1;

	// Now neither of them has status IN_NONE
	if (status==IN_PACKED)
		unpack();

	if (oGuid->status==IN_PACKED)
		oGuid->unpack();

	UUCMP(time_low, oGuid->time_low);
	UUCMP(time_mid, oGuid->time_mid);
	UUCMP(time_hi_and_version, oGuid->time_hi_and_version);
	UUCMP(clock_seq, oGuid->clock_seq);
	return memcmp(node, oGuid->node, 6);
}
