// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*- 

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: guid.h,v $ - header file for GUID.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/07/26 06:40:36 $
 * Version :  $Revision: 1.10 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

# ifndef _GUID_H_
# define _GUID_H_

// # include <sys/types.h>
# include <sys/time.h>
# include <time.h>
# include <string.h>
# include <set>

# ifndef GUID_LEN
# define GUID_LEN 16
# endif

typedef struct {
	unsigned char data[GUID_LEN];
} guid_t;

# define TIME_OFFSET_HIGH 0x01B21DD2
# define TIME_OFFSET_LOW  0x13814000

class Guid {
public:
	static const guid_t  GUID_NIL;
    Guid() // start an empty GUID
	{
		status=IN_NONE;
	} 

    Guid(const void *guid, int len) // start with a packed GUID
	{
		if (len!=GUID_LEN)
			status=IN_NONE;
		memcpy(u.packed, guid, len);
		status=IN_PACKED;
	}

    Guid(const guid_t &guid) // start with a packed GUID
	{
		memcpy(u.packed, guid.data, GUID_LEN);
		status=IN_PACKED;
	}

    ~Guid() {}

    void clear(void) // reset it to an empty GUID
	{
		status=IN_NONE;
	}

    int is_null(void) // is it an empty GUID?
	{
		switch (status) {
			case IN_NONE: return 1;
			case IN_FIELD: pack();
			case IN_PACKED: 
			case IN_BOTH: 
				for (int i=0; i<int(GUID_LEN/sizeof(int)); i++) {
					if (u.align[i])
						return 0;
				}
				return 1;
			default: 
				// internal error
				return 0;
		}
	}

    void gen(); // create a new GUID

    int pack(void *guid, int len);
    int pack(guid_t &guid) { return pack(guid.data, GUID_LEN); }
    int unpack(const void *guid, int len);
    int unpack(const guid_t &guid) { return unpack(guid.data, GUID_LEN); }

	int parse(const char *szIn); 
	int to_string(char *szOut, int len);

	int compare(Guid *oGuid);
	time_t get_time(struct timeval *ret_tv);

	Guid *clone()
	{
		if (status==IN_NONE)
			return new Guid();

		if (status!=IN_PACKED)
			pack();

		return new Guid(u.packed, GUID_LEN);
	}

	static unsigned int random_ui() 
	{
	    Guid guid;
	    unsigned int aui[4];

	    guid.gen();
	    guid.pack((unsigned char *)aui, sizeof(unsigned int)*4);
	    return aui[0]^aui[1]^aui[2]^aui[3];
	}

	static void get_random_bytes(void *buf, int nbytes);
	static int get_random_fd();
	static void set_random_fd(int fd)
	{
		if (fd>=0) {
			rfd=fd;
		}
	}

private:
	static int rfd;
	enum {IN_NONE=0, IN_PACKED=1, IN_FIELD=2, IN_BOTH=3};
	int status; // IN_PACKED, IN_FIELD, IN_BOTH.
	
	// the follow are the fields storing the unpacked version
	unsigned int	time_low;
	unsigned short	time_mid;
	unsigned short	time_hi_and_version;
	unsigned short	clock_seq;
	unsigned char	node[6];

	// the follow is the field storing the packed version
	union {
		int align[GUID_LEN/sizeof(int)];
		unsigned char	packed[GUID_LEN];
	} u;

	void gen_time(void);
	void gen_random(void);
	void pack(void);
	void unpack(void);
};

class guid_cmp : public std::less<guid_t> {
 public:
	bool operator()(const guid_t &x, const guid_t &y) const {
		return (memcmp((char *)&(x), (char *)&(y), sizeof(guid_t))<0);
	}
};

bool operator<(const guid_t &guid1, const guid_t &guid2);

typedef std::set<guid_t, guid_cmp> guid_set_t;

#endif /* _GUID_H_ */
