// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: hdch.cc,v $ - high-dimensional consistent hashing.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/04/09 09:59:58 $
 * Version :  $Revision: 1.7 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 
# include <iostream>
# include "hdch.h"
# include "../rscoder/galois.h"

using namespace std;

/**
 * Permutation table.
 * 0 - 012
 * 1 - 021
 * 2 - 102
 * 3 - 120
 * 4 - 201
 * 5 - 210
 */
const char HDCH::perm_rev[]={0,1,2,4,3,5};

int HDCH::permute(char pid, unsigned char *c0, unsigned char *c1, unsigned char *c2)
{
	unsigned char tmp;

	switch (pid) {
		case 0:
			; break;
		case 1:
			tmp=*c2; *c2=*c1; *c1=tmp;
			; break;
		case 2:
			tmp=*c0; *c0=*c1; *c1=tmp;
			; break;
		case 3:
			tmp=*c0; *c0=*c1; *c1=*c2; *c2=tmp;
			; break;
		case 4:
			tmp=*c0; *c0=*c2; *c2=*c1; *c1=tmp;
			; break;
		case 5:
			tmp=*c0; *c0=*c2; *c2=tmp;
			; break;
		default:
			return -1;
	}

	return 0;
}

# ifndef __NO_ENCODE__
/* encoding: last two digits: 0->1, 1->3, 2->2, 3->0 */
/* decoding last two digits: 0->3, 1->0, 2->2, 3->1 */
static unsigned int	sha_encode[]=
	{0x5a827998|0x1, 0xca62c1d4|0x2, 0x8f1bbcdc|0x0, 0x6ed9eba0|0x3};

static unsigned int	sha_decode[]=
	{0x6ed9eba0|0x3, 0x5a827998|0x1, 0x8f1bbcdc|0x0, 0xca62c1d4|0x2};
# endif

unsigned short gal_lllog(unsigned short s)
{
	unsigned short dummy;
	return gal_log_one(gal_log_one(gal_log_one(s, &dummy), &dummy), &dummy);
}

unsigned short gal_iiilog(unsigned short s)
{
	unsigned short dummy;
	return gal_ilog_one(gal_ilog_one(gal_ilog_one(s, &dummy), &dummy), &dummy);
}

unsigned int HDCH::encode_bucket(unsigned int buck_id)
{
# ifndef __NO_ENCODE__
	buck_id=buck_id^(sha_encode[buck_id&0x3]);
	if (gal_field_size()==16) {
		unsigned short s1=(unsigned short)((buck_id&0xffff0000)>>16);
		unsigned short s0=(unsigned short)(buck_id&0xffff);
		s1=gal_iiilog(s1);
		s0=gal_iiilog(s0);
		buck_id=(((unsigned int)s1)<<16)|s0;
	}
	buck_id=(buck_id>>24)|(buck_id<<8);
	unsigned int fixed=buck_id&0xff;
	unsigned int shift=(fixed&0x1f)%24;
	unsigned int perm=(fixed>>5)%6;
	unsigned int r_mask=(((unsigned int)1<<shift)-1)<<8;
	unsigned int l_mask=(((unsigned int)1<<(24-shift))-1)<<(8+shift);
	unsigned int result= ((buck_id&r_mask)<<(24-shift)) |
				((buck_id&l_mask)>>shift);
	unsigned char c1=(unsigned char)(result>>24);
	unsigned char c2=(unsigned char)((result>>16)&0xff);
	unsigned char c3=(unsigned char)((result>>8)&0xff);

	permute(perm, &c1, &c2, &c3);
	result=((unsigned int)c1<<24) | 
		((unsigned int)c2<<16) |
		((unsigned int)c3<<8) |
		fixed;

	// cout << "Key: "; print_key(buck_id); cout << endl;
	// cout << "Encoded: "; print_key(result); cout << endl;
	return result;
# else
	return buck_id;
# endif
}

unsigned int HDCH::decode_bucket(unsigned int buck_id)
{	
# ifndef __NO_ENCODE__
	unsigned int fixed=buck_id&0xff;
	unsigned int shift=(fixed&0x1f)%24;
	unsigned int perm=HDCH::perm_rev[(fixed>>5)%6];
	unsigned int r_mask=(((unsigned int)1<<(24-shift))-1)<<8;
	unsigned int l_mask=(((unsigned int)1<<shift)-1)<<(32-shift);
	unsigned char c1=(unsigned char)(buck_id>>24);
	unsigned char c2=(unsigned char)((buck_id>>16)&0xff);
	unsigned char c3=(unsigned char)((buck_id>>8)&0xff);

	permute(perm, &c1, &c2, &c3);

	unsigned int result=((unsigned int)c1<<24) | 
		((unsigned int)c2<<16) |
		((unsigned int)c3<<8);

	result= ((result&r_mask)<<shift) |
		((result&l_mask)>>(24-shift)) |
		fixed;
	// cout << "Key: "; HDCH::print_key(buck_id); cout << endl;
	// cout << "Decoded: "; HDCH::print_key(result); cout << endl;
	result=((result&0xff)<<24)|(result>>8);

	if (gal_field_size()==16) {
		unsigned short s1=(unsigned short)((result&0xffff0000)>>16);
		unsigned short s0=(unsigned short)(result&0xffff);
		s1=gal_lllog(s1);
		s0=gal_lllog(s0);
		result=(((unsigned int)s1)<<16)|s0;
	}
	result=result^(sha_decode[result&0x3]);

	return result;
# else
	return buck_id;
# endif
}

int HDCH::add_bucket(unsigned int buck_id)
{
	if (buck_cnt>=max_buck) {
		return -1;
	}

	if (buck_id==0) {
		return -1;
	}

	buck_id=encode_bucket(buck_id);

	int i=buck_id%max_buck;
	int j=i;
	for (; j<i+max_buck; j++) {
		if (!bucks[j%max_buck])
			break;
		if (bucks[j%max_buck]==buck_id) // already added
			return 0;
	}
	if (bucks[j%max_buck]) // no empty slot
		return -1;
	j=j%max_buck;
	bucks[j]=buck_id;
	unsigned char digest[MD5_LEN];
	Md5::md5((unsigned char *)&buck_id, sizeof(unsigned int), digest);
	unsigned int *uis=(unsigned int *)&digest;
	for (int k=0; k<4; k++) {
		bucks_md5[k+j*4]=uis[k];
	}
	buck_cnt++;
	return 0;
}
	
int HDCH::del_bucket(unsigned int buck_id)
{
	if (buck_id==0) {
		return -1;
	}

	buck_id=encode_bucket(buck_id);

	int i=buck_id%max_buck;
	int j=i;
	for (; j<i+max_buck; j++) {
		if (!bucks[j%max_buck]) // not found
			return -1;
		if (bucks[j%max_buck]==buck_id) {
			// found
			j=j%max_buck;
			bucks[j]=0;
			for (int k=0; k<4; k++) {
				bucks_md5[k+j*4]=0;
			}
			buck_cnt--;
			return 0;
		}
	}

	// not found
	return -1;
}

int HDCH::distance(unsigned int a, unsigned int b)
{
	static int vdist[256]={
		0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4,
		1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
		1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
		2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
		1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
		2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
		2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
		3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
		1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
		2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
		2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
		3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
		2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
		3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
		3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
		4, 5, 5, 6, 5, 6, 6, 7, 5, 6, 6, 7, 6, 7, 7, 8
	};
	unsigned int c=(a^b);
	unsigned char *pc=(unsigned char *)&c;
	return (vdist[pc[0]]+vdist[pc[1]]+vdist[pc[2]]+vdist[pc[3]]);
}

# if (0)
int HDCH::distance(unsigned int a, unsigned int b)
{
	unsigned int c=(a^b);
	int i=0;

	// Now count the 1's in c
	while (c) {
		c=c&(c-1);
		i++;
	}

	return i;
}
# endif

// using MD5 first
int HDCH::hash(unsigned int item)
{
	int min_dist=64;
	int min_dist_md5=64;
	unsigned int res_buck=0xffffffff;
	unsigned int min_sec_dist=0;

	if (buck_cnt==0)
		return -1;

	for (int i=0; i<max_buck; i++) {
		int dist;
		int dist_md5;
		unsigned int sec_dist;
		
		if (!bucks[i])
			continue;

		unsigned int buck_md5_int=0;
		for (int j=0; j<4; j++) {
			buck_md5_int=bucks_md5[i*4+j]^buck_md5_int;
		}
		dist_md5=distance(buck_md5_int, item);
		dist=distance(bucks[i], item);
		sec_dist=bucks[i]^item;

		if (dist_md5<min_dist_md5) {
			min_dist_md5=dist_md5;
			min_dist=dist;
			min_sec_dist=sec_dist;
			res_buck=bucks[i];
			continue;
		}
		if (dist_md5==min_dist_md5) {
			if (dist<min_dist) {
				min_dist=dist;
				min_sec_dist=sec_dist;
				res_buck=bucks[i];
				continue;
			}
			if (dist==min_dist) {
				if (sec_dist<min_sec_dist) {
					min_sec_dist=sec_dist;
					res_buck=bucks[i];
					continue;
				}
			}
		}
	}

	if (min_dist_md5!=64)
		return decode_bucket(res_buck);

	return 0;
}

# if (0)
// enhenced, using a vector
int HDCH::hash(unsigned int item)
{
	int min_dist[5]={64, 64, 64, 64, 64};;
	unsigned int res_buck=0xffffffff;
	unsigned int min_sec_dist=0;

	if (buck_cnt==0)
		return -1;

	for (int i=0; i<max_buck; i++) {
		int dist[5];
		unsigned int sec_dist;
		
		if (!bucks[i])
			continue;

		int j=0;
		for (; j<5; j++) {
			if (j==0) {
				dist[j]=distance(bucks[i], item);
			}
			else {
				dist[j]=distance(bucks_md5[i*4+j-1], item);
			}

			if (dist[j]==min_dist[j]) {
				continue;
			}
			break;
		}

		if (j<5) {
			if (dist[j]<min_dist[j]) {
				for (j++; j<5; j++) {
					dist[j]=distance(bucks_md5[i*4+j-1], item);
				}
				sec_dist=bucks[i]^item;
				memcpy(min_dist, dist, 5*sizeof(int));
				min_sec_dist=sec_dist;
				res_buck=bucks[i];
			}
			continue;
		}

		sec_dist=bucks[i]^item;
		if (min_sec_dist>sec_dist) {
			memcpy(min_dist, dist, 5*sizeof(int));
			min_sec_dist=sec_dist;
			res_buck=bucks[i];
			continue;
		}
	}

	if (min_dist[0]!=64)
		return decode_bucket(res_buck);

	return 0;
}
# endif

# if (0)
// original
int HDCH::hash(unsigned int item)
{
	int min_dist=65536;
	unsigned int res_buck=0xffffffff;
	unsigned int min_sec_dist=0;

	if (buck_cnt==0)
		return -1;

	for (int i=0; i<max_buck; i++) {
		if (!bucks[i])
			continue;

		int dist=0;
		unsigned int sec_dist=0;
		
		dist=distance(bucks[i], item);
		// for (int j=0; j<4; j++) {
			// dist+=distance(bucks_md5[i*4+j], item);
		// }

		sec_dist=bucks[i]^item;

		if (dist<min_dist) {
			min_dist=dist;
			min_sec_dist=sec_dist;
			res_buck=bucks[i];
		}
		else if (dist==min_dist) {
			if (sec_dist<min_sec_dist) {
				min_sec_dist=sec_dist;
				res_buck=bucks[i];
			}
		}
	}

	if (min_dist!=65536)
		return decode_bucket(res_buck);

	return 0;
}
# endif

void HDCH::print_key(unsigned int key)
{
	unsigned int i=1<<31;
	for (int j=0; j<32; j++, i>>=1) {
		if (i&key) {
			cout << "1";
		}
		else {
			cout << "0";
		}
	}
}

int HDCH::get_bucket_cnt(void)
{
	return buck_cnt;
}
