// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*- 

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: hdch.h,v $ - header file for High-Dimensional Consistent
 *            Hashing.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2002/10/08 07:32:53 $
 * Version :  $Revision: 1.5 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

# ifndef _HDCH_H_
# define _HDCH_H_

# include <string.h>
# include <stdlib.h>
# include "../md5/md5.h"

/**
 * High-Dimensional-Consistent-Hashing.
 * It operates for items and buckets in the space of (0,1)^n.
 * This class only works for n=32 or all the unsigned intergers.
 * Particularly it targets to the case where buckets are 
 * identified as IP addresses. For IP addresses in a LAN,
 * they share lots of bits in common (the network part), so 
 * we use the lower 5 bits (0-4 in network order) as a shift counter
 * to shift the upper 24 bits. And then use bits 5-7 to do a permutation
 * for the upper three bytes., Finally we apply the distance
 * calculation algorithm.
 *
 * It is very important that either the bucket IDs have quite random
 * bit patterns or they only share higher bits in common.
 */ 

class HDCH {
# ifndef __ALL_PUBLIC__
private:
# else
public:
# endif
	int max_buck;
	int buck_cnt;
	unsigned int *bucks;
	unsigned int *bucks_md5; // each bucket consists of 4 unsigned ints.

	/**
	 * Permutation table.
	 */
	static const char perm_rev[];
	
	static int permute(char pid, unsigned char *c0, 
			unsigned char *c1, unsigned char *c2);
	static unsigned int encode_bucket(unsigned int buck_id);
	static unsigned int decode_bucket(unsigned int buck_id);
	static int distance(unsigned int a, unsigned int b);
public:
	HDCH() :
		max_buck(-1),
		buck_cnt(0),
		bucks(NULL),
		bucks_md5(NULL)
	{ }

	HDCH(int max_buckets) :
		max_buck(max_buckets),
		buck_cnt(0),
		bucks(NULL),
		bucks_md5(NULL)
	{
		if (max_buck>0) {
			bucks=(unsigned int *)calloc(max_buck, sizeof(unsigned int));
			bucks_md5=(unsigned int *)calloc(max_buck*(MD5_LEN/sizeof(unsigned int)), sizeof(unsigned int));
			if (!bucks || !bucks_md5) {
				throw -1;
			}
		}
	}

	~HDCH()
	{
		if (bucks) {
			free(bucks);
		}

		if (bucks_md5) {
			free(bucks_md5);
		}
	}

	int set_max_buckets(int max_buckets)
	{
		if (bucks) // already in use.
			return -1;

		if (max_buckets>0) {
			bucks=(unsigned int *)calloc(max_buckets, sizeof(unsigned int));
			bucks_md5=(unsigned int *)calloc(max_buckets*(MD5_LEN/sizeof(unsigned int)), sizeof(unsigned int));
			if (!bucks || !bucks_md5) {
				return -1;
			}
			max_buck=max_buckets;
			return 0;
		}
		return -1;
	}

	void reset()
	{
		if (bucks) {
			free(bucks);
			bucks=NULL;
		}

		if (bucks_md5) {
			free(bucks_md5);
			bucks_md5=NULL;
		}
		max_buck=-1;
		buck_cnt=0;
	}

	int add_bucket(unsigned int buck_id);
	int del_bucket(unsigned int buck_id);
	int hash(unsigned int item);

	static void print_key(unsigned int key);
	int get_bucket_cnt(void);

};

# endif
