// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: t9.cc,v $ - HDCH tester IX.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2002/10/08 07:32:53 $
 * Version :  $Revision: 1.2 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 
# include <sys/types.h>
# include <sys/stat.h>
# include <fcntl.h>
# include <stdio.h>
# include <unistd.h>
# include <stdlib.h>
# include <math.h>
# include <libgen.h>
# include <getopt.h>
# include <errno.h>
# include <string.h>

# ifndef _USE_CH_
# include "hdch.h"
# else
# include "ch.h"
# define HDCH CH
# endif

static int random_fd=-1;
static unsigned int get_random_ui()
{
	unsigned int x=0;
	unsigned char *p=(unsigned char *)&x;
	unsigned int n=0;
	while (n<sizeof(int)) {
		int m=read(random_fd, p, (sizeof(int)-n));
		if (m<0) {
			if ((errno==EINTR)||(errno==EAGAIN))
				continue;
			else {
				fprintf(stderr, "Read failed.\n");
				throw -1;
			}
		}
		n+=m;
		p+=m;
	}

	return x;
}

static char *prog_name=NULL;

static void do_help()
{
	printf("\
Usage: %s [options] <#buckets>\n\
   -h                 Help, this screen.\n\
   -n <buckets>       Number of buckets.\n\
   -N <items>         Total number of items.\n\
   -i <items>         Avg no of items per bucket.\n\
   -d <delta>         Delta of the # of buckets (1).\n\
   -f <rfile>         Getting random numbers from a file.\n\
", prog_name);

	exit(0);
}

int main(int argc, char *argv[])
{
	int n_buckets=0;
	int n_items=0;
	int n_avg_items=0;
	int n_delta=0;
	int c;
	char *random_file="/dev/urandom";

	prog_name=argv[0];
	while ((c=getopt(argc, argv, "n:N:i:d:f:h"))!=EOF) {
		switch (c) {
			case 'h': do_help(); break;
			case 'n': n_buckets=atoi(optarg); break;
			case 'N': n_items=atoi(optarg); break;
			case 'i': n_avg_items=atoi(optarg); break;
			case 'd': n_delta=atoi(optarg); break;
			case 'f': random_file=optarg; break;
			default: do_help();
		}
	}

	int n_missing_arg=0;
	if (!n_buckets) {
		n_missing_arg++;
	}

	if (n_missing_arg != (argc-optind)) {
		fprintf(stderr, "Syntax error.\n");
		do_help();
	}

	if (!n_buckets) {
		n_buckets=atoi(argv[optind++]);
	}

	if (!n_items && !n_avg_items) {
		n_avg_items=1000; // default
	}

	if (n_items && n_avg_items) {
		fprintf(stderr, "Cannot specify both total # of items and # of items/bucket.\n");
		do_help();
	}

	random_fd=open(random_file, O_RDONLY);
	if (random_fd<0) {
		fprintf(stderr, "Cannot open file: %s.\n", random_file);
		exit(-1);
	}

	if (n_avg_items) {
		n_items = n_buckets*n_avg_items;
	}

	if (n_items > 2500000)
		n_items = 2500000;

	if (n_delta==0) {
		n_delta=n_buckets/100;
		if (n_delta*100<n_buckets)
			n_delta++;
	}

	unsigned int *items=(unsigned int *)calloc(n_items, sizeof(unsigned int));
	char buck_ids1[3*n_buckets];
	char buck_ids2[3*n_buckets];
	int n_buck1, n_buck2;
	if (n_delta>0) {
		n_buck1=n_buckets;
		n_buck2=n_buckets+n_delta;
	}
	else {
		n_buck1=n_buckets+n_delta;
		n_buck2=n_buckets;
	}

	unsigned int buck_id_base=(128<<24)|(111<<16)|(40<<8)|2;

	printf("Testing item distribution with large number of bins.\n");

	for (int i=0; i<n_items; i++) {
		items[i]=get_random_ui();
		if (i%10000==0) {
			printf("\rStep 1 of 2: %.3f%% done ...    ", i*100.0/n_items);
			fflush(stdout);
		}
	}
	printf("\n");

	memset(buck_ids1, 0, 3*n_buckets);
	memset(buck_ids2, 0, 3*n_buckets);

	for (int i=0; i<n_buck1; i++) {
		int bid=0;
		do {
			bid=rand()%(3*n_buckets);
		} while (buck_ids1[bid]);
		buck_ids1[bid]=1;
		buck_ids2[bid]=1;
	}

	for (int i=0; i<(n_buck2-n_buck1); i++) {
		int bid=0;
		do {
			bid=rand()%(3*n_buckets);
		} while (buck_ids2[bid]);
		buck_ids2[bid]=1;
	}

	HDCH h1(n_buck1), h2(n_buck2);
	for (int i=0; i<3*n_buckets; i++) {
		if (buck_ids1[i]) {
			unsigned int bid=buck_id_base+i;
			bid=(bid<<24)|((bid&0xff00)<<8)|(bid>>16);
			h1.add_bucket(bid);
		}
	}
	for (int i=0; i<3*n_buckets; i++) {
		if (buck_ids2[i]) {
			unsigned int bid=buck_id_base+i;
			bid=(bid<<24)|((bid&0xff00)<<8)|(bid>>16);
			h2.add_bucket(bid);
		}
	}

	int diff_cnt=0;
	int *buck_no1=(int *)calloc(n_items, sizeof(int));
	int *buck_no2=(int *)calloc(n_items, sizeof(int));
	int buck_cnt1[3*n_buckets];
	int buck_cnt2[3*n_buckets];
	memset(buck_cnt1, 0, sizeof(int)*3*n_buckets);
	memset(buck_cnt2, 0, sizeof(int)*3*n_buckets);

	for (int i=0; i<n_items; i++) {
		unsigned int idx1=h1.hash(items[i]);
		unsigned int idx2=h2.hash(items[i]);
		if (idx1!=idx2)
			diff_cnt++;
		idx1=((idx1>>24)|((idx1>>8)&0xff00)|(idx1<<16))-buck_id_base;
		idx2=((idx2>>24)|((idx2>>8)&0xff00)|(idx2<<16))-buck_id_base;
		buck_no1[i]=idx1;
		buck_no2[i]=idx2;
		buck_cnt1[idx1]++;
		buck_cnt2[idx2]++;
		if (i%1000==0) {
			printf("\rStep 2 of 2: %.3f%% done ...    ", i*100.0/n_items);
			fflush(stdout);
		}
	}
	printf("\n");

	double avg1=(double)n_items/n_buck1;
	double sdv1=0, sdv2=0;
	int max_item1=0, max_item2=0;

	int sum=0;
	for (int i=0; i<3*n_buckets; i++) {
		if (buck_ids1[i]) {
			sum+=buck_cnt1[i];
			if (buck_cnt1[i]>max_item1)
				max_item1=buck_cnt1[i];
			sdv1+=(double)(buck_cnt1[i]-avg1)*(buck_cnt1[i]-avg1);
		}
		else {
			if (buck_cnt1[i]) {
				fprintf(stderr, "Error1!\n");
				return -1;
			}
		}
	}

	sdv1/=(n_buck1-1);
	sdv1=sqrt(sdv1);

	if (sum!=n_items) {
		fprintf(stderr, "sum1 wrong! Test failed!!!!!\n");
		return -1;
	}

	double avg2=(double)n_items/(n_buck2);
	sdv2=0;
	sum=0;
	int mov_buck=0;
	for (int i=0; i<3*n_buckets; i++) {
		if (buck_ids2[i]) {
			sum+=buck_cnt2[i];
			if (buck_cnt2[i]>max_item2)
				max_item2=buck_cnt2[i];
			sdv2+=(double)(buck_cnt2[i]-avg2)*(buck_cnt2[i]-avg2);
			if (buck_ids1[i]) {
				if (buck_cnt1[i]!=buck_cnt2[i]) {
					mov_buck++;
				}
			}
		}
		else {
			if (buck_cnt2[i]) {
				fprintf(stderr,  "Error2!\n");
				return -1;
			}
		}
	}

	sdv2/=(n_buck2-1);
	sdv2=sqrt(sdv2);

	if (sum!=n_items) {
		fprintf(stderr, "sum2 wrong! Test failed!!!!!\n");
		return -1;
	}

	printf("#bucket: %d==>%d, mig_obj: %d/%.3f%%, std_dev: %.3f==>%.3f,\n"
		   "#buck_change: %d, avg: %.3f==>%.3f, avg/max: %.3f%%==>%.3f%%\n", 
			n_buckets, n_buckets+n_delta, diff_cnt, diff_cnt*100.0/n_items, sdv1, sdv2, 
			mov_buck, avg1, avg2, avg1*100.0/max_item1, avg2*100.0/max_item2);

	return 0;
}
