// -*- mode:C; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: scon.c,v $ - Sequential consistence checking
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2002/10/08 07:32:53 $
 * Version :  $Revision: 1.2 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

/**
#include <sys/types.h>
**/
# include <stdlib.h>
# include <stdio.h>
# include <pthread.h>

# include "../atomic.h"

# define REPEAT ((unsigned long)100000)
# define MSG_SIZE 64

volatile int x=0, y=0, lx=0, ly=0, b1=0, b2=0;
int error=0;

void *sconx(void *);
void *scony(void *);

int n=REPEAT;

int main(int argc, char *argv[]) 
{
	int i;

	pthread_t tid[2];

	if (argc>=2)
		n=atoi(argv[1]);

	pthread_create(&(tid[0]), NULL, sconx, NULL);
	pthread_create(&(tid[1]), NULL, scony, NULL);

	for (i=0; i<2; i++) {
		pthread_join(tid[0], NULL);
		pthread_join(tid[1], NULL);
	}

	if (error==0) {
		printf("Possible sequential consistent architecture.\n");
	}
	return 0;
}

/**
 * Repeat the following operation in two threads (init: i=j=0):
 * Thread 1: <barrier> <x=i> <ly=y> <barrier> <i=i+1>
 * Thread 2: <barrier> <y=j> <lx=x> <barrier> <j=j+1>
 *
 * In side the two barriers, i==j. So at least one of the lx or ly
 * must be equal to the current value of i (or j) on a sequential
 * consistent system.
 */

void *sconx(void *arg)
{
	int i;
	register int t=0;

	fprintf(stderr, "start\n");
	fflush(stderr);
	for (i=1; i<=n; i++) {
		/* entry barrier */
		anf32(&b1, 1);
		while (b1!=2)
			;

		anf32(&b2, 1);
		while (b2!=2)
			;

		/**
		 * x=i;
		 * ly=y;
		 */
		/**
		asm __volatile__ ( 
			"movl %4, %0\n\t"
			"movl %3, %2\n\t"
			"movl %2, %1\n\t"
			: "=m"(x), "=m"(ly), "=r"(t)
			: "m" (y), "r"(i), "2"(t)
			);
		**/
		asm __volatile__ ( 
			"movl %4, %0\n\t"
			LOCK_PREFIX "addl $0, %0\n\t"
			"movl %3, %2\n\t"
			"movl %2, %1\n\t"
			: "=m"(x), "=m"(ly), "=r"(t)
			: "m" (y), "r"(i), "2"(t)
			);
		/* exit barrier */
		snf32(&b1, 1);
		while (b1!=0)
			;

		snf32(&b2, 1);
		while (b2!=0)
			;

		if ( (ly!=i) && (lx!=i) ) {
			fprintf(stderr, "Sequential consistency "
					"violated! lx=%d, ly=%d, x=%d, y=%d\n", lx, ly, x, y);
			error++;
			return NULL;
		}
 	}

	return NULL;
}

void *scony(void *arg)
{
	int i;
	register int t=0;

	fprintf(stderr, "start\n");
	fflush(stderr);
	for (i=1; i<=n; i++) {
	/* entry barrier */
	anf32(&b1, 1);
	while (b1!=2)
		;

	anf32(&b2, 1);
	while (b2!=2)
		;

	/**
	 * y=i;
	 * lx=x;
	 */
	/**
	asm __volatile__ (
		"movl %4, %0\n\t"
		"movl %3, %2\n\t"
		"movl %2, %1\n\t"
		: "=m"(y), "=m"(lx), "=r"(t)
		: "m" (x), "r"(i), "2"(t)
		);
	**/
	asm __volatile__ (
		"movl %4, %0\n\t"
		LOCK_PREFIX "addl $0, %0\n\t"
		"movl %3, %2\n\t"
		"movl %2, %1\n\t"
		: "=m"(y), "=m"(lx), "=r"(t)
		: "m" (x), "r"(i), "2"(t)
		);
	
	/* exit barrier */
	snf32(&b1, 1);
	while (b1!=0)
		;

	snf32(&b2, 1);
	while (b2!=0)
		;

	if ( (ly!=i) && (lx!=i) ) {
		fprintf(stderr, "Sequential consistency violated! lx=%d, ly=%d, x=%d, y=%d\n", lx, ly,
		x,y);
		error++;
		return NULL;
	}
	}

   return NULL;
}
