// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: mmap_pool.cc,v $ - mmapped, readonly, paged file cache.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/04/09 09:59:58 $
 * Version :  $Revision: 1.5 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 
# include <sys/time.h>
# include <sys/resource.h>
# include <sys/types.h>
# include <sys/stat.h>
# include <sys/vfs.h>
# include <sys/mman.h>
# include <unistd.h>
# include <fcntl.h>
# ifndef __SINGLE_PROC__
# include <pthread.h>
# endif
# include <errno.h>
# include <string.h>
# include <math.h>
# include <set>
# include <map>

# include "../debug/tmpi_debug.h"
# include "mmap_pool.h"

using namespace std;

static void *mmap_wrap(size_t length, int flags, int fd, off_t offset)
{
	void *buf;
	
	void *try_addrs[]={(void *)0x0L, (void *)0x0d000000};
	for (unsigned int i=0; i<sizeof(try_addrs)/sizeof(void *); i++) {
		buf=mmap(try_addrs[i], length, PROT_READ, flags|MAP_SHARED, fd, offset);
		if (buf!=MAP_FAILED) 
			return buf;
	}

	tmpi_error(DBG_VERBOSE, "mmap failed: %s (len=%u, offset=%u)\n", 
			strerror(errno), (unsigned)length, (unsigned)offset);

	return NULL;
}

// primality tester, wheel factorization
static bool is_prime(unsigned int p)
{
	if (p<6) {
		if ((p==3)||(p==5)||(p==2)) {
			return true;
		}
		else {
			return false;
		}
	}

	if ( (!(p%3)) || (!(p%5)) || (!(p%2)) ) {
		return false;
	}

	unsigned int f=(unsigned int)sqrt(p)+1;
	unsigned int wheel[]={7, 11, 13, 17, 19, 23, 29, 31};

	for (unsigned int i=0; i<=f; i+=30) {
		for (unsigned int j=0; j<sizeof(wheel)/sizeof(unsigned int); j++) {
			unsigned factor=i+wheel[j];
			if (factor>f)
				break;
			if (!(p%factor))
				return false;
		}
	}

	return true;
}

// Find the closest but smaller prime number to upper
static unsigned int prime_within(unsigned int upper)
{
	if (upper==2)
		return upper;

	if (!(upper&0x1)) {
		upper--;
	}
	while ((upper>1)&&(!is_prime(upper))) {
		upper-=2;
	}

	return (upper<=1)?0: upper;
}

short MFile::next_idx=0;
short MFile::page_sz=4096;
# ifndef __SINGLE_PROC__
pthread_mutex_t	MFile::glck=PTHREAD_MUTEX_INITIALIZER;
# endif
MFile::ENT * MFile::cache=NULL;
unsigned int MFile::cache_size=0;
MFile::FID_SET MFile::fid_set;
# ifndef __SINGLE_PROC__
pthread_mutex_t MFile::gclck=PTHREAD_MUTEX_INITIALIZER;
# endif
unsigned int MFile::gcc_idx=0;
unsigned int MFile::gcc_ridx=0;
unsigned int MFile::gcc[MFILE_GCC_SIZE];
int MFile::mem_used=0;
int MFile::mem_allowed=0;
int MFile::touch_mem_allowed=0;
unsigned int MFile::n_hit=0;
unsigned int MFile::n_capacity=0;
unsigned int MFile::n_conflict=0;
unsigned int MFile::n_miss=0;
unsigned int MFile::n_mmap=0;
unsigned int MFile::n_munmap=0;
unsigned int MFile::n_malloc=0;
unsigned int MFile::n_free=0;
unsigned int MFile::n_egc=0;
unsigned int MFile::n_gc=0;
unsigned long long MFile::n_tot_bytes=0;

// 256K cache
// changed to 512K cache entries
# define CACHE_BITS (unsigned char)19
# define CACHE_SIZE MFILE_EXP2(CACHE_BITS)
//
// 256MB minimum. 2.5GB default
# define MFILE_DEFAULT_MEM (2500*1024)
# define MFILE_MIN_MEM (256*1024)
# define MFILE_KILO 1024

bool MFile::init(int mem_in_kb, int touch_in_kb)
{
	if (cache)
		return true; // already initialized.

	// Sanity check, in case we made a wrong
	// definition of CACHE_SIZE.
 	// Set the cache size minimum to be the number
	// of regular chunks needed to fill the whole
	// address space, then multipled by 2 (for gapping
	// chunks.
	// With current parameters, csize will be in the
	// range of 32K and 32M.

	unsigned char min_csize=MFILE_CHUNK_ADDR;
	unsigned char max_csize=MFILE_TAG_LEN;

	if ((CACHE_BITS<min_csize)||(CACHE_BITS>max_csize))
		return false;

	if (!glock())
		return false;

	if (MFile::cache) {
		// omebody did it for me. :)
		gunlock();
		return true;
	}

	int mem_page_sz=getpagesize();

	if (mem_page_sz>=0)
		page_sz=mem_page_sz;

	/* set # of files to maximum */
	struct rlimit rlim;
	if (getrlimit(RLIMIT_NOFILE, &rlim)==0) {
		rlim.rlim_cur=rlim.rlim_max;
		setrlimit(RLIMIT_NOFILE, &rlim);
	}

	cache_size=prime_within(CACHE_SIZE);
	// get rid of object overhead.
	mem_in_kb-=((sizeof(ENT)*cache_size)+
				 sizeof(MFile)*MFILE_EXP2(MFILE_MAX_ID)+
				 sizeof(unsigned int)*MFILE_GCC_SIZE
			)/MFILE_KILO;

	int min_mem_in_kb=MFILE_MIN_MEM; 

	if (mem_in_kb<min_mem_in_kb)
		mem_in_kb=min_mem_in_kb;

	mem_allowed=mem_in_kb;
	touch_mem_allowed=touch_in_kb;
	if (touch_mem_allowed>mem_allowed) {
		touch_mem_allowed=mem_allowed;
	}

	cache=new ENT[cache_size];

	gunlock();

	return (cache!=NULL);
}

int MFile::fd_read(unsigned int offset, void *buf, unsigned int len)
{
	if (!flock()) {
		return -1;
	}
	try {
		if (lseek(fd, (off_t)offset, SEEK_SET)<0)
			throw -1;
		throw (int)(::read(fd, buf, len));
	}
	catch (int len) {
		funlock();
		return len;
	}

	tmpi_error(DBG_INTERNAL, "Impossible things happened.\n");
	funlock(); // should never happen.

	return -1;
}

MFile::MFile(char *map_file, unsigned int hint_flags) :
	idx(-1),
	fd(-1),
	fsz(0),
	hint(hint_flags),
	dat(NULL)
{
# ifndef __SINGLE_PROC__
	pthread_mutex_init(&flck, NULL);
# endif
	if (init(MFILE_DEFAULT_MEM)==false)
		return;

	fd=open(map_file, O_RDONLY);

	if (fd<0)
		return;

	try {
		struct stat st;
		if (fstat(fd, &st)!=0) {
			tmpi_error(DBG_INTERNAL, "Cannot access file: %s.\n", map_file);
			throw -1;
		}

		fsz=st.st_size;
		glock();
		n_tot_bytes+=fsz;
		gunlock();

		struct statfs stfs;
		if (fstatfs(fd, &stfs)!=0) {
			tmpi_error(DBG_INTERNAL, "Cannot access file: %s.\n", map_file);
			throw -1;
		}

		unsigned int chunk_size=MFILE_EXP2(MFILE_CHUNK);
		if ( (chunk_size/stfs.f_bsize*stfs.f_bsize) != chunk_size ) {
			tmpi_error(DBG_VERBOSE, "Warning: mmap chunk size not a multiple of I/O block.\n");
		}

# ifndef __NO_MAP_OVERLAP__
		if ( (page_sz/stfs.f_bsize*stfs.f_bsize) != page_sz ) {
			tmpi_error(DBG_VERBOSE, "Warning: gap chunk mmap enabled while its size is not a multiple of I/O block.\n");
		}
# endif

		// handle MFILE_MALLOC_ONCE
		if (hint&MFILE_MALLOC_ONCE) {
			inc32(&n_malloc);
			dat=calloc(fsz, 1);
			if (!dat)
				throw -1;
			if (((unsigned)fd_read(0, dat, fsz))!=fsz)
				throw -1;

			aadd32(&mem_used, fsz/MFILE_KILO);
		}
		else if (hint&MFILE_MMAP_ONCE) {
			dat=mmap_wrap(ROUND_UP(fsz, page_sz), (hint&MFILE_LOCKED)?MAP_LOCKED:0, fd, 0);
			inc32(&n_mmap);
			if (!dat)
				throw -1;

			aadd32(&mem_used, fsz/MFILE_KILO);
		}
		else {
			idx=fna16s(&next_idx, 1);
			if (idx>=(short)MFILE_EXP2(MFILE_MAX_ID)) {
				// too many Files.
				tmpi_error(DBG_INTERNAL, "Too many MFile objects.\n");
				dec16(&next_idx);
				throw -1;
			}

			add_to_fid_set(idx);

			if (hint&MFILE_TOUCH) {
				// unsigned chunk_sz=MFILE_EXP2(MFILE_CHUNK);
				for (unsigned int offset=0; (mem_used<touch_mem_allowed)&&(offset<fsz); offset+=page_sz) {
					MData data;
					request(offset, 1, data);
				}
			}
		}
	}
	catch (...) {
		if (fd>=0) {
			close(fd);
			fd=-1;
		}
		if (dat) {
			free(dat);
			dat=NULL;
		}
		return;
	}
}

MFile::~MFile()
{
	// I do not clean garbages in the cache
	// because I assume that MFiles will
	// never be destroyed. the fid_set is 
	// used to track active fids so that
	// so that in the worst case when we
	// are out of memory, we can still do gc().
	if (fd>=0) {
		close(fd);
	}
	
	if (hint&MFILE_MALLOC_ONCE) {
		if (dat)
			free(dat);
	}
	else if (hint&MFILE_MMAP_ONCE) {
		munmap(dat, ROUND_UP(fsz, page_sz));
	}
	else {
		remove_from_fid_set(idx);
	}

# ifndef __SINGLE_PROC__
	pthread_mutex_destroy(&flck);
# endif
}

bool MFile::request(unsigned int offset, unsigned int len, MData &data)
{
	// if (offset>=fsz) {
		// return false;
	// }

	if (idx<0) {
		assert((hint&MFILE_MALLOC_ONCE) || 
			   (hint&MFILE_MMAP_ONCE) );
		assert(dat);
		data.init((char *)dat+offset, len, fsz-offset, -3);
		return true;
	}

	unsigned int tag;
	unsigned int buf_offset;
	unsigned int max_len=length_suggest0(offset, tag, buf_offset);

	tag=tag|(((unsigned int)idx)<<(MFILE_CHUNK_ADDR));
	unsigned int cidx=tag%cache_size;
	ENT &ent=cache[cidx];

	if (!ent.lock())
		return false;

	try {
		bool cconflict=false;
		if ((tag!=ent.tag)||(ent.buf==NULL)) {
			unsigned int block_len=max_len+(offset-buf_offset);
			unsigned int n_kbs=block_len/MFILE_KILO;

			if (ent.buf!=NULL) {
				if (ent.ref_cnt == 0) {
					inc32(&n_miss);
					// replace cache entry
					ent.release();
				}
				else {
					inc32(&n_conflict);
					cconflict=true;
				}
			}
			else {
				inc32(&n_capacity);
				// create new entry
				if (anf32s(&mem_used, n_kbs) > mem_allowed) {
					if (gc(cidx) > mem_allowed) {
						// out of memory :(
						tmpi_error(DBG_INTERNAL, "Out of memory.\n");
						asub32(&mem_used, n_kbs);
						throw -1;
					}
				}
			}

			if (!cconflict) {
				// fill cache entry.
				bool mmap_success=false;
				if ( (!(hint&MFILE_MALLOC)) && ((block_len+buf_offset)<fsz) 
# if defined(__NO_MAP_OVERLAP__)
				// no overlapped address space
					&& (!(tag&MFILE_EXP2(MFILE_CHUNK_ADDR-1)))
# endif
				) {
					// try mmap first.
					ent.buf=mmap_wrap(block_len, (hint&MFILE_LOCKED)?MAP_LOCKED:0, fd, buf_offset);
					inc32(&n_mmap);
					if (ent.buf!=NULL) {
						mmap_success=true;
						ent.is_malloced=false;
					}
				}

				if (mmap_success==false) {
					ent.buf=calloc(block_len, 1);
					inc32(&n_malloc);
					if (!ent.buf) {
						asub32(&mem_used, n_kbs);
						throw -1;
					}
					// tmpi_error(DBG_VERBOSE, "Memory allocated for %d bytes.\n", block_len);
					int buf_len=fd_read(buf_offset, ent.buf, block_len);
					if ( (buf_len<0) || 
					     ( ((unsigned)buf_len<block_len) &&
						   (((unsigned)buf_len+buf_offset)<fsz) ) )
					{
						free(ent.buf);
						ent.buf=NULL;
						asub32(&mem_used, n_kbs);
						throw -1;
					}
					ent.is_malloced=true;
				}

				ent.tag=tag;
			}
			else {
				// solve cache conflict with a private
				// MData object.
				tmpi_error(DBG_INTERNAL, "Cache conflict.\n");
				void *buf=NULL;
# ifndef __NO_MAP_OVERLAP__
				// try mmap first. (however, we do not perform mmap if
				// it is required not to overlap mmapped regions.
				if ( (!hint&MFILE_MALLOC) &&
					 (max_len+offset<fsz) ) {
					unsigned int moffset= ROUND_DOWN(offset, page_sz);
					size_t mlen=ROUND_UP(max_len, page_sz);
					buf=mmap_wrap(mlen, (hint&MFILE_LOCKED)?MAP_LOCKED:0, fd, moffset);
					inc32(&n_mmap);
					if (buf) {
						data.init((char *)buf+(offset-moffset), 
								  len, max_len, -2);
					}
				}
				if (!buf) {
# endif
					if (max_len+offset>fsz) {
						max_len=fsz-offset;
					}
					buf=calloc(max_len, 1);
					inc32(&n_malloc);
					if (!buf) {
						throw -1;
					}
					// tmpi_error(DBG_INTERNAL, "Memory allocated for %d bytes.\n", max_len);
					int buf_len=fd_read(offset, buf, max_len);
					if ( (buf_len<0) || ((unsigned)buf_len<max_len) ) {
						free(buf);
						throw -1;
					}
					data.init(buf, len, max_len, -1);
# ifndef __NO_MAP_OVERLAP__
				}
# endif
			}
		}
		else {
			inc32(&n_hit);
		}

		if (!cconflict) {
			// now ent is good.
# if defined(__NO_PADDING__)
			// if we are not allowed to pad 0's at the end of the file.
			if (max_len+offset>fsz) {
				max_len=fsz-offset;
			}
# endif
	
			data.init((char *)ent.buf+(offset-buf_offset), 
					  len, max_len, cidx);
		}
	}
	catch (...) {
		ent.unlock();
		return false;
	}

	ent.unlock();
	return true;
}

// len will keep the actual bytes read into buffer
bool MFile::read(unsigned int offset, void *vbuf, unsigned int &len)
{
	unsigned int rem_len=len;
	char *buf=(char *)vbuf;

	while (rem_len>0) {
		unsigned int rlen=MFILE_EXP2(MFILE_CHUNK);
		if (rlen>rem_len) {
			rlen=rem_len;
		}

		MData data;
		if (!request(offset, rlen, data)) {
			break;
		}

		unsigned int dlen=data.length();
		if (rlen>dlen) {
			rlen=dlen;
		}
		memcpy(buf, data.addr(), rlen);
		offset+=rlen;
		rem_len-=rlen;
		buf+=rlen;
	}
	
	// got 0 bytes.
	if (len==rem_len)
		return false;

	len-=rem_len;
	return true;
}

// return the mem recycled by checking one cache entry.
int MFile::gc_one(unsigned int i, unsigned int me)
{
	if (i==me)
		return 0;

	ENT &ent=cache[i];
	if (!ent.trylock()) {
		return 0;
	}

	if (ent.ref_cnt==0) {
		int buf_len=ent.release();
		ent.unlock();
		return buf_len;
	}

	ent.unlock();
	return 0;
}

// performing garbage collection till mem_used is below
// mem_allowed, return the mem_allowed after gc().
// Will skip collect things in cache[me]. To avoid trouble,
// gc() is done sequentially.
int MFile::gc(unsigned int me)
{
	int lmem_used=mem_used;

	if (!gclock()) {
		return lmem_used;
	}

	if (lmem_used>mem_allowed) {
		// search for cache first
		// If there is 0.5% free space, we are done.
		// tmpi_error(DBG_VERBOSE, "Garbage collection started.\n");
		inc32(&n_gc);
		int mem_safe=mem_allowed-mem_allowed/200; 

		if (gcc[gcc_idx%MFILE_GCC_SIZE]!=0) {
			gcc_ridx=gcc_idx;
		}
		unsigned int i=gcc_ridx;
		for (unsigned int j=0; j<MFILE_GCC_SIZE; i++, j++) {
			unsigned int idx=i%MFILE_GCC_SIZE;
			unsigned int cidx=gcc[idx];
			if (cidx==0)
				continue;
			int rl=gc_one(cidx-1, me);
			if (rl>0) {
				lmem_used=snf32s(&mem_used, rl/MFILE_KILO);
			}
			gcc[idx]=0;
			if (lmem_used<=mem_safe)
				break;
		}
		gcc_ridx=i;

		if (lmem_used>mem_allowed) {
			// tmpi_error(DBG_VERBOSE, "Garbage collection exhaustive search started.\n");
			inc32(&n_egc);
			for (int i=0; i<(int)cache_size; i++) {
				int rl=gc_one(i, me);
				if (rl>0) {
					lmem_used=snf32s(&mem_used, rl/MFILE_KILO);
				}

				if (lmem_used<=mem_allowed)
					break;
			}
		}
	}
	else {
		// looks like I don't need to do gc() at all
	}

	gcunlock();

	return lmem_used;
}

void MFile::report()
{
	unsigned int tmiss=n_capacity+n_conflict+n_miss;
	unsigned int total=n_hit+tmiss;

	printf("total=%u, hit=%.3f%%, miss=%.3f%% (capacity=%.3f%%, conflict=%.3f%%, compulsary=%.3f%%).\n", 
		total,
		(double)n_hit*100.0/total,
		(double)tmiss*100.0/total,
		(double)n_capacity*100.0/total,
		(double)n_miss*100.0/total,
		(double)n_conflict*100.0/total
		  );
	printf("mmap=%u, munmap=%u, malloc=%u, free=%u.\n",
			n_mmap, n_munmap, n_malloc, n_free);
	printf("garbage collection=%u, extensive gc=%u.\n",
			n_gc, n_egc);
	printf("memory: used=%uKB, allowed=%uKB, files=%d, tot_bytes=%llu.\n",
			mem_used, mem_allowed, (int)next_idx, n_tot_bytes);
	int n_empty=0;
	for (unsigned int i=0; i<cache_size; i++) {
		if (cache[i].buf==NULL) {
			n_empty++;
		}
	}
	printf("%.3f%% cache entries empty.\n", (double)n_empty*100.0/cache_size);
	return;
}
