// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*- 

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: mmap_pool.h,v $ - header file for mmap_pool.cc, a read-only 
 *            mmapped and paged file cache.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/07/16 05:22:49 $
 * Version :  $Revision: 1.7 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 
# ifndef _MMAP_POOL_H_
# define _MMAP_POOL_H_

/**
 * Paged mmap pool for 32bits architecture. Pages 
 * are mapped for read-only!
 *
 * Interface: 
 * 	Each MFile object corresponds to a file
 * backed by a number of mmapped chunks (invisible 
 * to end users). One can request data at offsets 
 * of a multiple of memory pages in the file using
 * MFile::request(). The requested data length is
 * a hint and an actual data block with less data
 * could be returned instead. However, it is guaranteed
 * that a request with length <= MEM_PAGE_SIZE is 
 * guaranteed to satisfy. A data length exceeding
 * 256KB will NEVER be satisfied. (See details below.)
 * A MData object will be returned as a result. 
 * You may get the address with MData::addr(), and 
 * size with MData::length() (which could be longer
 * or short than requested length). MData 
 * objects must be released through MFile::release().
 *
 * Design:
 * 	All MFile objects share a common cache. The
 * cache is directly mapped, whose tag is as follows:
 * |FID|G|CHUNKADDR|
 * FID       : 10 bits, file handle ID (unsigned).
 * G         : 1 bit, 0: normal 256KB chunk, 1: gap 
 *             2*PG chunk.
 * CHUNKADDR : 14 bits, (unsigned) for normal chunk 
 *             (G=0), the in-file offset is 
 *             CHUNKADDR*256K. For gap chunk (G=1), 
 *             the in-file offset is CHUNKADDR*256K-PG.
 *
 * Each entry also has the following attributes:
 * malloced    : bool, false: mmaped, true: malloced (last 
 *             page, mmap failed but fread okay)
 * ref_cnt     : unsigned short, reference count.
 *
 * Cache size is defined in mmap_pool.cc through
 * macro CACHE_SIZE.
 *
 * Minimum # of cache entries: (2^32)/(256K)*2=32K.
 * Suggested: 256K (so that if # of files <=8, then
 * no cache entry conflict.)
 *
 * A page replacement buffer (STL set<int>). 
 * Change to the buffer happens:
 * 1) page counter increase and page in buffer
 * 	  the entry will be removed.
 * 2) page counter decrease to 0 and buffer size
 *    less than 1024, add into buffer.
 * 3) Cache entry value==NULL and MAX_MEM_PAGE 
 *    reached, get one from buffer, unmap it (or free 
 *    it), set entry to NULL, remap the new thing 
 *    (or malloc, read).
 *
 * Replacable buffer is dropped because it is not sure
 * how often we will get into a situation when we need
 * to replace cache entries (very rare due to cache
 * entry conflict, most likely running out of memory.)
 *
 * Optimization:
 * For the best performance, callers of this module
 * must follow the following procedure.
 * 1. Request for small-size data block: use read() 
 *  (internally small-size block requests have also
 *  been specificially optimized, but still with 
 *  some overhead).
 * 2. Request for multiple consecutive small-size
 *  data block: use request() then decode the buffer.
 * 3. Request for large data blocks:
 * 	a) chunk the data into blocks whose lengthes are
 * 	obtained by suggest_length().
 * 	b) If needs to be completely buffered, use read().
 */

# include <sys/mman.h>
# include <unistd.h>
# ifndef __SINGLE_PROC__
# include <pthread.h>
# endif
# include <set>
# include <map>

# ifndef __SINGLE_PROC__
# include "../machine/atomic.h"
# else
# include "../machine/atomic.debug.h"
# endif

// 0000100000 (n 0's)
# define MFILE_EXP2(n) (0x1U<<(n))

// 0000011111 (n 1's)
# define MFILE_MASK(n) ((0x1U<<(n))-1)

# define ROUND_UP(x, y) (((x)+(y)-1)/(y)*(y))
# define ROUND_DOWN(x, y) ((x)/(y)*(y))

class MData;

// Flags:
/* Equivalence to MAP_LOCKED, ignored if MFILE_MALLOC is set */
# define MFILE_LOCKED 1
/* Pages are malloc-ed rather than mmapped */
# define MFILE_MALLOC 2
/* Copy to main memory at once */
# define MFILE_MALLOC_ONCE 4
/* Mmap at once */
# define MFILE_MMAP_ONCE 8
/* eager touch, only useful if we use CHUNKED mmap/malloc, ignored when MFILE_xxx_ONCE is set */
# define MFILE_TOUCH 16

// 1024 mapped files
# define MFILE_MAX_ID 10
// 256KB chunks
// changed to 64KB
// changed to 512KB
# define MFILE_CHUNK 19
// 14bit CHUNK address. (Max file size: 2GB)
// Actually we can lift the restriction of 2GB.
// Instead, we use a 1GB (or 30bits) comman because most of
// the files are less than 1GB. And we can have higher bits
// stored beyond the FID.
// TO-DO: change tag format to be:
// |CHUNK_ID_HIGH|FID|G|CHUNK_ID_LOW|
// CHUNK_ID_HIGH: 8 bits
// FID: file id: 10 bits
// G: gap: 1 bit
// CHUNK_ID_LOW: 12 bits
// This we, we can use up-to 256GB files.
# define MFILE_CHUNK_ADDR (31-(MFILE_CHUNK)+1)

/**
# if (MFILE_CHUNK_ADDR > 16)
# error "Regular Chunk Size Too Small!\n"
# endif
**/

# define MFILE_TAG_LEN (MFILE_MAX_ID+MFILE_CHUNK_ADDR)

// Make sure the tag fits in a 32 bits interger.
# if (MFILE_TAG_LEN > 32)
# error "Tag size too long!\n"
# endif

class MFile {
	struct ENT;

	friend class MData;
	friend struct MFile::ENT;

 private:
 	struct ENT {
		friend class MFile;
# ifndef __SINGLE_PROC__
		pthread_mutex_t lck;
# endif
		/* |FID|CADDR| where CADDR is |G|CHUNK_ID| */
		unsigned int tag; 
		unsigned short ref_cnt;
		bool is_malloced;
	 	void *buf;
	
		ENT() : tag(0), ref_cnt(0), is_malloced(false), buf(NULL)
		{ 
# ifndef __SINGLE_PROC__
			pthread_mutex_init(&lck, NULL); 
# endif
		}

		~ENT() { 
# ifndef __SINGLE_PROC__
			pthread_mutex_destroy(&lck); 
# endif
		}

		int release() {
			if (!buf)
				return 0;

			int block_len=MFILE_EXP2(MFILE_CHUNK);
			if (is_gap_chunk()) {
				block_len=MFile::page_sz*2;
			}

			if (is_malloced) {
				free(buf);
				inc32(&(MFile::n_free));
			}
			else {
				munmap(buf, block_len);
				inc32(&(MFile::n_munmap));
			}
			reset();
			return block_len;
		}

		bool is_in_buffer(void *addr, unsigned int len) {
			int block_len=MFILE_EXP2(MFILE_CHUNK);
			if (is_gap_chunk()) {
				block_len=MFile::page_sz*2;
			}
			if ( (addr>=buf) && 
				 ((char *)addr+len<=(char *)buf+block_len) ) {
				return true;
			}

			return false;
		}

		void reset() { tag=0; ref_cnt=0; buf=NULL; }
		bool lock()   {
# ifndef __SINGLE_PROC__
			return (pthread_mutex_lock(&lck)==0);   
# else
			return true;
# endif
		}

		bool trylock(){
# ifndef __SINGLE_PROC__
			return (pthread_mutex_trylock(&lck)==0);
# else
			return true;
# endif
		}

		bool unlock() {
# ifndef __SINGLE_PROC__
			return (pthread_mutex_unlock(&lck)==0); 
# else
			return true;
# endif
		}

		/**
	 	unsigned short file_id() { 
			return (unsigned short)(tag>>MFILE_CHUNK_ADDR); 
		}
		**/
	 	bool is_gap_chunk() { 
			return (tag&MFILE_EXP2(MFILE_CHUNK_ADDR-1)); 
		}
		/**
	 	unsigned short chunk_addr() { 
			return (unsigned short)(tag&MFILE_MASK(MFILE_CHUNK_ADDR-1));
		}
		**/
		void inc_ref_cnt() { inc16(&ref_cnt); }
		void dec_ref_cnt() { dec16(&ref_cnt); }

		/**
		static unsigned int make_tag(unsigned short fid, unsigned int cid)
		{
			return ( (((unsigned int)fid & 
					   MFILE_MASK(MFILE_MAX_ID)) << MFILE_CHUNK_ADDR) | 
					 (cid&MFILE_MASK(MFILE_CHUNK_ADDR)) );
		}
		**/
 	};

 	static short next_idx; // keep increasing ...
	static short page_sz; // memory page size

 	// single lock protecting fid_set
	// atomic instructions used for:
	// next_id, mem_used
# ifndef __SINGLE_PROC__
	static pthread_mutex_t glck;
# endif
	static bool glock() { 
# ifndef __SINGLE_PROC__
		return (pthread_mutex_lock(&glck)==0); 
# else
		return true;
# endif
	}
	static bool gunlock() { 
# ifndef __SINGLE_PROC__
		return (pthread_mutex_unlock(&glck)==0); 
# else
		return true;
# endif
	}

	static ENT *cache; // the cache
	static unsigned int cache_size;

	typedef std::set<short> FID_SET;
	static FID_SET fid_set; // existing fid_set
	static void add_to_fid_set(short idx) {
		if (glock()) {
			fid_set.insert(idx);
			gunlock();
		}
	}
	static void remove_from_fid_set(short idx) {
		if (glock()) {
			FID_SET::iterator pi=fid_set.find(idx);
			if (pi!=fid_set.end()) {
				fid_set.erase(pi);
			}
			gunlock();
		}
	}

	// Garbage-Collection-Cache (GCC) size
// # define MFILE_GCC_SIZE 1024
# define MFILE_GCC_SIZE 4096

# ifndef __SINGLE_PROC__
	static pthread_mutex_t gclck;
# endif
	static bool gclock() { 
# ifndef __SINGLE_PROC__
		return (pthread_mutex_lock(&gclck)==0); 
# else
		return true;
# endif
	}
	static bool gcunlock(){ 
# ifndef __SINGLE_PROC__
		return (pthread_mutex_unlock(&gclck)==0);
# else
		return true;
# endif
	}

	static unsigned int gcc_idx;
	static unsigned int gcc_ridx;
	static unsigned int gcc[];
	static int gc_one(unsigned int i, unsigned int me);
	static int gc(unsigned int me);

	static void dec_ref_cnt(int i) {
		cache[i].dec_ref_cnt();
		if (cache[i].ref_cnt==0) {
			unsigned int gi=fna32(&gcc_idx, 1);
			gi=gi%MFILE_GCC_SIZE;
			gcc[gi]=i+1; // 0 means illegal
		}
	}

	static int mem_used; // in KB
	static int mem_allowed; // in KB
	static int touch_mem_allowed; // in KB

	// statistics
	// cache performance
	static unsigned int n_hit, n_capacity, n_conflict, n_miss; 
	// memory operations
	static unsigned int n_mmap, n_munmap, n_malloc, n_free;
	// (extensive) garbage collection 
	static unsigned int n_egc, n_gc; 
	// total mapped bytes
	static unsigned long long n_tot_bytes;

	// per MFile object data:
# ifndef __SINGLE_PROC__
	pthread_mutex_t flck; // lock for fd access
# endif
	short idx; // file idx.
	short fd; // file descriptor.
	unsigned int fsz; // file size.
	unsigned int hint; // mapping hints.
	void *dat; // only for completely in-mem object.

	bool flock() { 
# ifndef __SINGLE_PROC__
		return (pthread_mutex_lock(&flck)==0); 
# else
		return true;
# endif
	}
	bool funlock() { 
# ifndef __SINGLE_PROC__
		return (pthread_mutex_unlock(&flck)==0); 
# else
		return true;
# endif
	}
	int fd_read(unsigned int offset, void *buf, unsigned int len); 
	MFile(); // illegal constructor

 public:
	// Global initialization.
	// @Param: mem_in_kb: memory usage for the all MFile cache memory.
	// @Param: touch_in_kb: the amount of memory used in eager touch.
	static bool init(int mem_in_kb, int touch_mem_in_kb=0);

	// when we want to read large amount of data,
	// we will query about the size that can be returned
	// at a certain offset.
	static int length_suggest0(unsigned int offset, 
							  unsigned int &chunk_addr, 
							  unsigned int &buf_offset)
	{
		chunk_addr=(offset>>MFILE_CHUNK);
		unsigned int chunk_start_offset=chunk_addr<<MFILE_CHUNK;

		unsigned int chunk_end_offset=
			chunk_start_offset+MFILE_EXP2(MFILE_CHUNK);

		unsigned int gap_start_offset=chunk_end_offset-page_sz;

		if (offset>=gap_start_offset) {
				chunk_addr=chunk_addr|MFILE_EXP2(MFILE_CHUNK_ADDR-1);

				buf_offset=gap_start_offset;

			return (chunk_end_offset-offset+page_sz);
		}

		buf_offset=chunk_start_offset;

		return chunk_end_offset-offset;
	}

	static int length_suggest(unsigned int offset) {
		unsigned int dummy_ca, dummy_bo;
		return length_suggest0(offset, dummy_ca, dummy_bo);
	}

	MFile(char *fname, unsigned int hint_flags);
	~MFile();
	// Even if we cannot meet the len requirement,
	// we still try to return a valid MData object
	bool request(unsigned int offset, unsigned int len, MData &data);

	bool read(unsigned int offset, void *buf, unsigned int &len);
	bool good() const { return (fd>=0); }
	unsigned int get_size() { return fsz; }
	void * get_data() { return dat; }
	static void report();
};

// Data in MData optimization is only useful for random access
// of small datablocks.
class MData {
	friend class MFile;

# define MDATA_INIT (MFILE_EXP2(31)|MFILE_EXP2(30))
 public:
	MData(MData &d) // copy constructor
	{
		tag = d.tag;

		// uninitialized
		if (tag == MDATA_INIT) {
			return;
		}

		// data in object
		if (tag & MFILE_EXP2(31)) {
			memcpy(u.data.buf, d.u.data.buf, sizeof(u.ref));
			return;
		}
		
		// data in reference
		//
		// data in cache
		if (d.u.ref.idx>=0) {
			u.ref.idx=d.u.ref.idx;
			u.ref.ptr=d.u.ref.ptr;
			MFile::cache[u.ref.idx].inc_ref_cnt(); // increase reference counter
			return;
		}

		switch (-d.u.ref.idx) {
			case 1:
			case 2:
			{
				// data malloced or mmapped
				// because we don't keep fd, we are left
				// with no choice but to copy 
				u.ref.idx=-1;
				unsigned int l=length();
				u.ref.ptr=malloc(l);
				if (!u.ref.ptr) {
					tag = MDATA_INIT;
				}
				else {
					memcpy(u.ref.ptr, d.u.ref.ptr, l);
				}
			}
			break;
			case 3:
			{
				// point to completely in-mem (or mmaped) object.
				u.ref.idx=-3;
				u.ref.ptr=d.u.ref.ptr;
			}
			break;
			default:
				tag=MDATA_INIT;
		}
	}

	MData() : tag(MDATA_INIT) { 
		/* just set object to invalid state. */ 
	}

	~MData() { release(); }

	const void * const addr() const {
		switch (tag>>30) {
		case 1:  // invalid
		case 3:  return NULL; // invalid
		case 0:  return (const void *)u.ref.ptr;
		case 2:  return (const void *)(u.data.buf);
		default: return NULL;
		}
	}

	const unsigned int off_from_beginning() const {
		switch (tag>>30) {
		case 1:  return 0; // invalid
		case 3:  return 0; // invalid
		case 2:  return 0; // data in buffer
		case 0: // data in u.ref.ptr
			{
				int idx=u.ref.idx;
				if (idx>=0) {
					return ((char *)u.ref.ptr-(char *)MFile::cache[idx].buf);
				}
				else {
					switch (-idx) {
					case 1:
						// malloced
						return 0;
					case 2:
						// mmapped
						return (char *)u.ref.ptr-(char *)ROUND_DOWN
							((unsigned long)(u.ref.ptr), MFile::page_sz);
					case 3:
						// completely mapped, since I don't have a back
						// reference to the MFile object, 
						// I don't know for completely mapped objects
						// But this should not be a common case.
						return 0;
					default:
						// what is wrong?
						return 0;
					}
				}
			}
		default: return 0;
		}
	}

	const unsigned int length() const { return tag & (~(0x3U<<30)); }

 private:
	union {
		struct A {
			// if idx >= 0, ptr is within buffers in cache
			// if idx == -1, ptr points to private mem.
			// if idx == -2, ptr points to mmap() mem.
			// if idx == -3, ptr points to completely in-mem file.
			// the actual mmapped address and size
			// are round to the closest page boundary.
			int idx; 
			void *ptr; 
		} ref;
		struct B { 
			unsigned char buf[sizeof(A)];
		} data;
	} u;

	// The tag is in the following format:
	// |P|B|LENGTH|
	// P : 1 bit, 0: pointer in object; 1: data in object
	// B : 1 bit, bad (uninitialized object).
	// LENGTH : 30 bits, length of the buffer. maximum 1GB.
	unsigned int tag; 
	//
	// p is the pointer to the usable data
	// max_len bytes can be read from p
	// i is a reference to the cache entry 
	// if p is indeed pointing into buffers
	// in some cache entry.
	void init(void *p, unsigned int req_len, unsigned int max_len, int i)
	{
		// if (tag!=MDATA_INIT) {
			// oops, did you forget to release something?
			release();
		// }

		if (i<0) { // never dup private entries.
			tag=max_len;
			u.ref.ptr=p;
			u.ref.idx=i;
		}
		else if (req_len<=sizeof(u.ref)) {
# if defined(__NO_PADDING__)
			// if padding is allowed, max_len always greater than or equal to
			// a memory page, so must be greater than req_len here.
			if (req_len>max_len)
				req_len=max_len;
# endif
			tag=(MFILE_EXP2(31))|req_len;
			memcpy(u.data.buf, p, req_len);
		}
		else {
			tag=max_len;
			u.ref.ptr=p;
			u.ref.idx=i;
			MFile::cache[i].inc_ref_cnt(); // increase reference counter
		}
	}

	void release() // Never explicitly call release().
	{
		if (tag&MFILE_EXP2(31)) {
			// data in object, do nothing
		}
		else if (u.ref.idx>=0) {
			// might need to put it into gcc[].
			MFile::dec_ref_cnt(u.ref.idx);
		}
		else {
			switch (-u.ref.idx) {
				case 1:
					free(u.ref.ptr);
					break;
				case 2:
					{
						void *rbuf=(void *)ROUND_DOWN
							((unsigned long)(u.ref.ptr), MFile::page_sz);
						size_t rsize=ROUND_UP(length(), MFile::page_sz);
						munmap(rbuf, rsize);
						inc32(&(MFile::n_munmap));
					}
					break;
				case 3:
				default:
					// do nothing
					break;
			}
		}

		// make sure calling release()
		// twice will not do harm.
		tag=MDATA_INIT;
	}
};

# endif
