// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: mmap_test.cc,v $ - mmap_pool tester I.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2002/10/08 07:32:53 $
 * Version :  $Revision: 1.3 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 
# include <stdio.h>
# include <stdlib.h>

# include "../usc/usc_time.h"

# define NTHREAD 10
// # define NTHREAD 1
# define NITER 50000

# if (NTHREAD!=1)
# include <pthread.h>
# endif

# include "../sim/gsldistrib.h"
# include "../guid/guid.h"
# include "../md5/md5.h"
# include "mmap_pool.h"

struct XFile {
	int id;
	MFile f;
	double pcent; // access frequency.
	// XFile (char *fname, int fid, double p) : id(fid), f(fname, MFILE_MALLOC), pcent(p) {}
	XFile (char *fname, int fid, double p) : id(fid), f(fname, 0), pcent(p) {}
	bool good() { return f.good(); }
};

XFile **files=NULL;
int xf_sz=0;


static void md5tostr(unsigned char *digest, char *md5_str)
{
	unsigned int uints[4];
	for (int i=0; i<4; i++, digest+=4) {
		memcpy(&(uints[i]), digest, 4);
	}

	sprintf(md5_str, "%08X%08X%08X%08X", uints[0], uints[1], uints[2], uints[3]);

	return;
}

static void *mmap_test(void *arg)
{
	int id=(int)arg;
	char trace_file[30];
	snprintf(trace_file, 30, "trace.%d", id);

	FILE *trace=fopen(trace_file, "w");
	if (!trace) {
		return NULL;
	}
	for (int i=0; i<NITER; i++) {
		double p=gsldd_flat(0, 1);
		int j=0;
		for (; j<xf_sz; j++) {
			if (p<=files[j]->pcent)
				break;
		}
		if (j>=xf_sz) // weird
			j=xf_sz-1;

		MFile &f=files[j]->f;

		unsigned int offset=Guid::random_ui();
		offset=offset%(f.get_size()-16);
		unsigned int len=Guid::random_ui()%300000+16;
		MData data;
		if (!f.request(offset, len, data)) {
			fprintf(trace, "FID=%d OFF=%u LEN=%u <failed>\n", files[j]->id, offset, len);
		}
		else {
			unsigned char digest[16];
			char md5_str[128];
			// Md5::md5((unsigned char *)data.addr(), (int)data.length(), digest);
			memcpy(digest, data.addr(), 16);
			md5tostr(digest, md5_str);
			fprintf(trace, "FID=%d OFF=%u LEN=%u RLEN=%u MD5=%s\n", files[j]->id, offset, len, data.length(), md5_str);
		}
	}

	fclose(trace);

	return NULL;
}

int main(int argc, char *argv[])
{
	MFile::init(2000*1024); // 2.5GB
	// MFile::init(1500*1024); // 1.5GB
	if (argc==1) {
		printf("Usage %s [[<big_file>]...]\n", argv[0]);
		return -1;
	}

	files=new (XFile*)[argc-1];
	memset(files, 0, sizeof(XFile *)*(argc-1));

	FILE *fp=fopen("trace.files", "w");
	if (!fp) {
		printf("Failed to open file: trace.files.\n");
		return -1;
	}

	double prob=1.0;
	double decay=1.7;
	double prob_acc=0;
	for (int i=1; i<argc; i++) {
		XFile *pxf=new XFile(argv[i], xf_sz, prob);
		if (pxf->good()) {
			files[xf_sz]=pxf;
			fprintf(fp, "FILE=%s, ID=%d.\n", argv[i], xf_sz);
			prob_acc+=prob;
			prob/=decay;
			xf_sz++;
		}
		else {
			printf("Failed to open MFile: %s.\n", argv[i]);
			delete pxf;
		}
	}

	fclose(fp);

	if (xf_sz==0) {
		printf("No good MFiles.\n");
		return -1;
	}

	unsigned long long total_sz=0;

	// scaling the probability.
	double acc=0.0;
	for (int i=0; i<xf_sz; i++) {
		files[i]->pcent/=prob_acc;
		files[i]->pcent+=acc;
		acc=files[i]->pcent;
		total_sz+=files[i]->f.get_size();
		printf("[%d], %.3f%%\n", files[i]->id, files[i]->pcent*100);
	}

	files[xf_sz-1]->pcent=1.0;

	printf("Loading %d files with %llu bytes.\n", xf_sz, total_sz);

	usc_time_t t_start=usc_gettime();
# if (NTHREAD>1)
	pthread_t thrs[NTHREAD];
	for (int i=0; i<NTHREAD; i++) {
		pthread_create(&thrs[i], NULL, &mmap_test, (void *)i);
	}

	for (int i=0; i<NTHREAD; i++) {
		pthread_join(thrs[i], NULL);
	}
# else
	mmap_test((void *)0);
# endif	

	usc_time_t t_end=usc_gettime();

	printf("Total execution time: %.3fms.\n", usc_difftime(t_end, t_start)/1000);
	MFile::report();

	return 0;
}
