// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: mmap_test2.cc,v $ - mmap_pool tester II.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2002/10/08 07:32:53 $
 * Version :  $Revision: 1.2 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 
# include <sys/mman.h>
# include <sys/types.h>
# include <sys/stat.h>
# include <fcntl.h>
# include <unistd.h>
# include <stdio.h>
# include <stdlib.h>
# include <errno.h>

# include "../guid/guid.h"
# include "../md5/md5.h"
# include "../debug/tmpi_debug.h"

// # define PAGE_SZ_BITS 16
# define PAGE_SZ_BITS 18
# define PAGE_SZ (0x1U<<(PAGE_SZ_BITS))
# define NUM_ENTRY (0x1U<<(32-PAGE_SZ_BITS))

static void *mmap_wrap(size_t length, int flags, int fd, off_t offset)
{
	void *buf;
	
	void *try_addrs[]={(void *)0x0L, (void *)0x0d000000};
	for (unsigned int i=0; i<sizeof(try_addrs)/sizeof(void *); i++) {
		buf=mmap(try_addrs[i], length, PROT_READ, flags|MAP_SHARED, fd, offset);
		if (buf!=MAP_FAILED) 
			return buf;
	}

	tmpi_error(DBG_VERBOSE, "mmap failed: %s (len=%u, offset=%u)\n", 
			   strerror(errno), (unsigned)length, (unsigned)offset);

	return NULL;
}

class MData {
private:
	void *ptr;
	unsigned int len;
public:
	MData(void *p=NULL, unsigned int l=0) : ptr(p), len(l) {}
	unsigned int length() { return len; }
	void *addr() { return ptr; }
	void init(void *p, unsigned int l) { ptr=p; len=l; }
};

class XFile {
private:
	int fd;
	unsigned int fsz;
	void **pages;
public:
	XFile (char *fname) : fd(-1), fsz(0), pages(NULL) {
		try {
			pages=(void **)calloc(NUM_ENTRY, sizeof(void *));
			if (!pages) {
				throw -1;
			}
			fd=open(fname, O_RDONLY);
			if (fd<0) {
				throw -1;
			}
			struct stat st;
			if (fstat(fd, &st)<0) {
				throw -1;
			}
			fsz=st.st_size;
		}
		catch (...) {
			if (pages) {
				free(pages);
				pages=NULL;
			}
			if (fd>=0) {
				close(fd);
				fd=-1;
			}
		}
	}

	~XFile() {
		if (!good())
			return;

		for (unsigned int i=0; i<NUM_ENTRY; i++) {
			if (pages[i]) {
				munmap(pages[i], PAGE_SZ);
			}
		}
		close(fd);
	}

	bool request(unsigned int offset, int len, MData &data)
	{
		if (!good())
			return false;

		unsigned int idx=offset>>PAGE_SZ_BITS;
		unsigned int buf_offset=idx<<PAGE_SZ_BITS;
		if (pages[idx]==NULL) {
			if ((buf_offset+PAGE_SZ)>=fsz) {
				pages[idx]=calloc(PAGE_SZ, 1);
				if (!pages[idx]) {
					return false;
				}
				lseek(fd, buf_offset, SEEK_SET);
				read(fd, pages[idx], fsz-buf_offset);
		    }
		    else {
				pages[idx]=mmap_wrap(PAGE_SZ, 0, fd, buf_offset);
		    }
		}

		if (pages[idx]==NULL) {
			return false;
		}
		
		unsigned int delta=(offset-buf_offset);
		data.init((char *)pages[idx]+delta, PAGE_SZ-delta);

		return true;
	}

	bool good() { return (fd>=0)&&(pages); }

	unsigned int get_size() { return fsz; }
};

# define NITER 1000000

static void md5tostr(unsigned char *digest, char *md5_str)
{
	unsigned int uints[4];
	for (int i=0; i<4; i++, digest+=4) {
		memcpy(&(uints[i]), digest, 4);
	}

	sprintf(md5_str, "%08X%08X%08X%08X", uints[0], uints[1], uints[2], uints[3]);

	return;
}

static bool mmap_test(XFile *f)
{
	FILE *trace=fopen("trace.result", "w");
	if (!trace) {
		return false;
	}

	for (int i=0; i<NITER; i++) {
		unsigned int offset=Guid::random_ui();
		offset=offset%(f->get_size()-1);
		unsigned int len=Guid::random_ui()%300000+1;
		MData data;
		if (!f->request(offset, len, data)) {
			fprintf(trace, "OFF=%u LEN=%u <failed>.\n", offset, len);
		}
		else {
			unsigned char digest[16];
			char md5_str[128];
			Md5::md5((unsigned char *)data.addr(), (int)data.length(), digest);
			md5tostr(digest, md5_str);
			fprintf(trace, "OFF=%u LEN=%u MD5=%s.\n", offset, len, md5_str);
		}
	}

	fclose(trace);

	return true;
}

int main(int argc, char *argv[])
{
	if (argc!=2) {
		printf("Usage %s [[<big_file>]...]\n", argv[0]);
		return -1;
	}

	XFile xfile(argv[1]);
	if (!xfile.good()) {
		printf("Cannot create XFile object.\n");
		return -1;
	}

	mmap_test(&xfile);

	return 0;
}
