// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: pshash.cc,v $ - persistent hashing.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/05/16 21:01:32 $
 * Version :  $Revision: 1.11 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 
# include <string.h>
# include <db3/db_cxx.h>
# include <sys/vfs.h>
# include "../debug/tmpi_debug.h"
# include "pshash.h"

// TO-DO
// Add a LogFile stream to all the constructors.
// Default to stderr. And all errors will be directed
// to the error stream.

PsHash::PsHash(char *backstore, unsigned int keylen, unsigned int datalen, unsigned int nelem, unsigned int cache_sz) // create a new hash table
{
	int ret=0;

	db=NULL;
	dbc=NULL;
	// on error, leave it in an invalid state.
	if ((ret=create(backstore, keylen, datalen, nelem, cache_sz))!=0) {
		// would be nice to connect to the high-level logging facilities
		tmpi_error(DBG_APP, "Cannot create persistent hash table: file=%s!\n", backstore);
	}
}

PsHash::PsHash(char *backstore, unsigned int cache_sz)
{
	int ret=0;

	db=NULL;
	dbc=NULL;
	// on error, leave it in an invalid state.
	if ((ret=load(backstore, cache_sz))!=0) {
		// would be nice to connect to the high-level logging facilities
		tmpi_error(DBG_APP, "Cannot load  persistent hash table from file \"%s\"!\n", 
				backstore);
	}
}

PsHash::~PsHash()
{
	release();
}

static unsigned int get_io_page_size(char *file)
{
	struct statfs s;
	unsigned int io_page_size=0;

	if (statfs(file, &s)==0) {
		io_page_size=s.f_bsize;
	}
	else {
		io_page_size=IO_PAGE_SIZE;
	}

	return io_page_size;
}

bool PsHash::sync()
{
	if (!db)
		return false;
	if (db->sync(0))
		return false;
	return true;
}

int PsHash::load(char *backstore, unsigned int cache_sz)
{
	int ret=0;

	if (db) // db already opened.
		return -1;

	if (strlen(backstore)>=MAX_FILE_NAME) // name for the backstore too long
		return -1;

	strcpy(bsfile, backstore);

	db=new Db(NULL, DB_CXX_NO_EXCEPTIONS);
	if (!db)
		return -1;

	unsigned int io_page_size=get_io_page_size(backstore);
	if (set_db_cache(cache_sz, io_page_size)<0) {
		goto _exit1;
	}

	if ((ret=db->open(bsfile, NULL, DB_HASH, DB_THREAD, 0660))!=0) {
		goto _exit1;
	}

	return 0;

_exit1:
	if (db) {
		db->close(0);
		delete db;
		db=NULL;
	}
	return -1;
}

////////////////
// Also set the cache page size
int PsHash::set_db_cache(unsigned int cache_sz, unsigned int io_page_size)
{
	int ret=0;	
	const unsigned int kilo=1024;
	const unsigned int mega=kilo*kilo;

	ret|=db->set_pagesize(io_page_size);
	ret|=db->set_cachesize(cache_sz/mega, cache_sz%mega*kilo, 0);
	
	if (ret)
		return -1;
	return 0;
}

int PsHash::set_db_hash(unsigned int keylen, unsigned int datalen, 
		unsigned int nelem, unsigned int io_page_size)
{
	int ret=0;
	int hf=(io_page_size-32)/(keylen+datalen+8);

	ret|=db->set_h_ffactor(hf);
	ret|=db->set_h_nelem(nelem);

	if (ret)
		return -1;
	return 0;
}

///////////////////////////////////////////////
// The cache_sz is in the unit of kilo-bytes
int PsHash::create(char *backstore, unsigned int keylen, unsigned int datalen, unsigned int nelem, unsigned int cache_sz)
{
	if (db) // db already opened.
		return -1;

	if (strlen(backstore)>=MAX_FILE_NAME) // name for the backstore too long
		return -1;

	strcpy(bsfile, backstore);

	db=new Db(NULL, DB_CXX_NO_EXCEPTIONS);

	if (!db)
		return -1;

	int ret=0;
	unsigned int io_page_size=get_io_page_size(backstore);

	if (set_db_cache(cache_sz, io_page_size)<0) {
		goto _exit0;
	}

	if (set_db_hash(keylen, datalen, nelem, io_page_size)<0) {
		goto _exit0;
	}

	if ((ret=db->open(bsfile, NULL, DB_HASH, DB_CREATE|DB_EXCL|DB_THREAD, 0660))!=0) {
		goto _exit0;
	}

	return 0;

_exit0:
	if (db) {
		db->close(0);
		delete db;
		db=NULL;
	}
	return -1;
}

int PsHash::release()
{
	int ret1=0, ret2=0;

	if (dbc) {
		ret1=dbc->close();
		dbc=NULL;
	}

	if (db) {
		ret2=db->close(0);
		delete (db);
		db=NULL;
	}
	else
		ret2=-1;


	return (ret1||ret2)?-1:0;
}
	
int PsHash::get_backstore(char *backstore, unsigned int len)
{
	if (!db)
		return -1;

	if (strlen(bsfile)>=(unsigned)len)
		return -1;

	strcpy(backstore, bsfile);

	return 0;
}

static Dbt *new_dbt(void *buf, unsigned int len)
{
	Dbt *dbt=new Dbt(buf, len);
	if (!dbt)
		return NULL;

	dbt->set_ulen(len);
	dbt->set_flags(DB_DBT_USERMEM);
	return dbt;
}

int PsHash::set(void *key, unsigned int klen, void *data, unsigned int dlen)
{
	if (!db)
		return -1;

	Dbt okey(key, klen), odat(data, dlen);
	int ret=0;

	if ((ret = db->put(NULL, &okey, &odat, 0)) != 0) {
		// db->err(ret, "pshash\n");
		return -1;
	}

	return 0;
}

int PsHash::get(void *key, unsigned int klen, void *data, unsigned int *dlen)
{
	if (!db)
		return -1;

	Dbt okey(key, klen);
	Dbt *odat=new_dbt(data, *dlen);
	int ret=0;

	if ((ret = db->get(NULL, &okey, odat, 0)) != 0) {
		// db->err(ret, "pshash\n");
		ret=-1;
	}
	else {
		*dlen=odat->get_size();
	}

	delete odat;

	return ret;
}

int PsHash::del(void *key, unsigned int len)
{
	if (!db)
		return -1;

	Dbt okey(key, len);
	int ret=0;

	if ((ret = db->del(NULL, &okey, 0)) != 0) {
		return -1;
	}

	return 0;
}

int PsHash::cursor_get(void *key, unsigned int *klen, void *data, unsigned int *dlen, int flags)
{
	if (!db)
		return -1;

	int ret;

	if (!dbc) {
		ret=db->cursor(NULL, &dbc, 0); 
		if (ret)
			dbc=NULL;
	}

	if (!dbc)
		return -1;

	Dbt *okey=new_dbt(key, *klen); 
	Dbt *odat=new_dbt(data, *dlen);

	ret=dbc->get(okey, odat, flags);

	if (ret) {
		ret= -1;
	}
	else {
		*klen=okey->get_size();
		*dlen=odat->get_size();
	}

	delete  odat;
	delete  okey;

	return ret;
}

int PsHash::get_first(void *key, unsigned int *klen, void *data, unsigned int *dlen)
{
	return cursor_get(key, klen, data, dlen, DB_FIRST);
}

int PsHash::get_last(void *key, unsigned int *klen, void *data, unsigned int *dlen)
{
	return cursor_get(key, klen, data, dlen, DB_LAST);
}

int PsHash::get_next(void *key, unsigned int *klen, void *data, unsigned int *dlen)
{
	return cursor_get(key, klen, data, dlen, DB_NEXT);
}

int PsHash::get_current(void *key, unsigned int *klen, void *data, unsigned int *dlen)
{
	return cursor_get(key, klen, data, dlen, DB_CURRENT);
}

int PsHash::get_prev(void *key, unsigned int *klen, void *data, unsigned int *dlen)
{
	return cursor_get(key, klen, data, dlen, DB_PREV);
}

int PsHash::is_loaded()
{
	if (db)
		return 1;

	return 0;
}
