// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: pstree.cc,v $ - persistent tree.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/09/24 01:20:32 $
 * Version :  $Revision: 1.18 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 
# include <string.h>
# include <db3/db_cxx.h>
# include <sys/vfs.h>
# include <stdlib.h>

# include "../debug/tmpi_debug.h"
# include "pstree.h"

# if defined(__CHECK_MEM__)
# include "../asmalloc/asmalloc.h"
# endif

const PsTree::node_t PsTree::NIL=(PsTree::node_t)0;
const PsTree::node_t PsTree::ROOT=(PsTree::node_t)1;
const PsTree::node_t PsTree::FREE=(PsTree::node_t)2;

const int PsTree::REC_HEADER_SZ=((int)(&(((struct PsTree::Rec *)(0L))->user)));
const int PsTree::FREEREC_HEADER_SZ=((int)(&(((struct PsTree::FreeRec *)(0L))->free_id)));

static Dbt *new_dbt(void *buf, unsigned int len)
{
	Dbt *dbt=new Dbt(buf, len);

	if (!dbt)
		return NULL;

	dbt->set_ulen(len);
	dbt->set_flags(DB_DBT_USERMEM);

	return dbt;
}

static unsigned int get_io_page_size(char *file)
{
	struct statfs s;
	unsigned int io_page_size=0;

	if (statfs(file, &s)==0) {
		io_page_size=s.f_bsize;
	}
	else {
		io_page_size=IO_PAGE_SIZE;
	}

	return io_page_size;
}

PsTree::PsTree(char *backstore, void *root, int len, unsigned int cache_sz, \
		unsigned int max_val_len) :
	db(NULL),
	MAX_VAL_LEN(max_val_len)
{
	int ret=0;

	// on error, leave it in an invalid state.
	if ((ret=create(backstore, root, len, cache_sz))!=0) {
		tmpi_error(DBG_APP, "Cannot create persistent tree: file=\"%s\"!\n", backstore);
	}
}

PsTree::PsTree(char *backstore, unsigned int cache_sz, unsigned int max_val_len) :
	db(NULL),
	MAX_VAL_LEN(max_val_len)
{
	int ret=0;

	// on error, leave it in an invalid state.
	if ((ret=load(backstore, cache_sz))!=0) {
		tmpi_error(DBG_APP, "Cannot load  persistent tree from \"%s\"!\n", backstore);
	}
}

PsTree::~PsTree()
{
	release();
}

int PsTree::put_first_free_id()
{
	if (put(FREE, &first_free_id, sizeof(node_t))!=0) {
		return -1;
	}

	return 0;
}

int PsTree::get_first_free_id()
{
	int len=sizeof(node_t);

	if (get(FREE, &first_free_id, &len)!=0) {
		return -1;
	}

	return 0;
}

int PsTree::load(char *backstore, unsigned int cache_sz)
{
	int ret=0;

	if (db) // db already opened.
		return -1;

	if (strlen(backstore)>=MAX_FILE_NAME) // name for the backstore too long
		return -1;

	strcpy(bsfile, backstore);

	db=new Db(NULL, DB_CXX_NO_EXCEPTIONS);

	if (!db)
		return -1;

	unsigned int io_page_size=get_io_page_size(backstore);

	if (set_db_cache(cache_sz, io_page_size)<0) {
		goto _exit1;
	}

	if ((ret=db->open(bsfile, NULL, DB_RECNO, DB_THREAD, 0660))!=0) {
		tmpi_error(DBG_APP, "DB error: %s.\n", DbEnv::strerror(ret));
		goto _exit1;
	}

	if (get_first_free_id()) {
		goto _exit1;
	}

	return 0;

_exit1:
	if (db) {
		db->close(0);
		delete db;
		db=NULL;
	}

	return -1;
}

bool PsTree::sync()
{	
	if (!db)
		return false;

	if (db->sync(0))
		return false;

	return true;
}

////////////////
// Also set the cache page size
int PsTree::set_db_cache(unsigned int cache_sz, unsigned int io_page_size)
{
	int ret=0;	
	const unsigned int kilo=1024;
	const unsigned int mega=kilo*kilo;

	ret|=db->set_pagesize(io_page_size);
	ret|=db->set_cachesize(cache_sz/mega, cache_sz%mega*kilo, 0);
	
	if (ret)
		return -1;

	return 0;
}

int PsTree::create_root_rec(void *root, int len)
{
	Rec *root_rec=(Rec *)malloc(REC_HEADER_SZ+len);
	if (!root_rec) {
		return -1;
	}

	root_rec->parent=ROOT;
	root_rec->child=NIL;
	root_rec->sibling=NIL;
	memcpy(root_rec->user, root, len);

	if (put(ROOT, root_rec, REC_HEADER_SZ+len) != 0) {
		free(root_rec);
		return -1;
	}

	free(root_rec);

	return 0;
}

///////////////////////////////////////////////
// The cache_sz is in the unit of kilo-bytes
int PsTree::create(char *backstore, void *root, int len, unsigned int cache_sz)
{
	if (db) // db already opened.
		return -1;

	if ((!root)||(len<=0)) {
		return -1;
	}

	if (strlen(backstore)>=MAX_FILE_NAME) // name for the backstore too long
		return -1;

	strcpy(bsfile, backstore);

	db=new Db(NULL, DB_CXX_NO_EXCEPTIONS);

	if (!db)
		return -1;

	int ret=0;
	unsigned int io_page_size=get_io_page_size(backstore);

	if (set_db_cache(cache_sz, io_page_size)<0) {
		goto _exit0;
	}

	if ((ret=db->open(bsfile, NULL, DB_RECNO, DB_CREATE|DB_EXCL|DB_THREAD, 0660))!=0) {
		tmpi_error(DBG_APP, "DB error: %s.\n", DbEnv::strerror(ret));
		goto _exit0;
	}

	first_free_id=NIL;

	if (put_first_free_id()) {
		goto _exit0;
	}

	if (create_root_rec(root, len)) {
		goto _exit0;
	}

	return 0;

_exit0:
	if (db) {
		db->close(0);
		delete db;
		db=NULL;
	}

	return -1;
}

/**
 * the validity of node is required
 */
int PsTree::get(node_t node, void *val, int *len)
{
# if defined(__DEBUG__)
	if (node==NIL) {
		tmpi_error(DBG_CALLER, "Bad node ID in PsTree::get().\n");
		return -1;
	}

	if (!db) {
		tmpi_error(DBG_APP, "DB uninitialized.\n");
		return -1;
	}
# endif

	Dbt okey(&node, sizeof(node_t));
	Dbt *odat=new_dbt(val, *len);

	if (!odat)
		return -1;
	
	if (db->get(NULL, &okey, odat, 0) != 0) {
		delete odat;
		return -1;
	}

	*len=odat->get_size();
	delete odat;

	return 0;
}

int PsTree::put(node_t node, void *val, int len)
{
# if defined(__DEBUG__)
	if (node==NIL) {
		tmpi_error(DBG_CALLER, "Bad node ID in PsTree::put().\n");
		return -1;
	}

	if (!db) {
		tmpi_error(DBG_APP, "DB uninitialized.\n");
		return -1;
	}
# endif

	Dbt okey(&node, sizeof(node_t));
	Dbt odat(val, len);

	if (db->put(NULL, &okey, &odat, 0) != 0) {
		return -1;
	}

	return 0;
}

int PsTree::del(node_t node)
{
# if defined(__DEBUG__)
	if ((node==NIL)||(node==FREE)) {
		tmpi_error(DBG_CALLER, "Bad node ID in PsTree::del().\n");
		return -1;
	}

	if (!db) {
		tmpi_error(DBG_APP, "DB uninitialized.\n");
		return -1;
	}
# endif

	Dbt okey(&node, sizeof(node_t));
	if (db->del(NULL, &okey, 0)) {
		return -1;
	}

	return 0;
}

int PsTree::free_node(node_t node)
{
# if defined(__DEBUG__)
	if ((node==NIL)||(node==FREE)) {
		tmpi_error(DBG_CALLER, "Bad node ID in PsTree::del().\n");
		return -1;
	}

	if (!db) {
		tmpi_error(DBG_APP, "DB uninitialized.\n");
		return -1;
	}
# endif

	if (first_free_id==FREE) {
		tmpi_error(DBG_INTERNAL, "PSTREE structure corrupted.\n");
		return -1;
	}

	if (first_free_id!=NIL) { /* add to the root record if room allowed */
		FreeRec *first_free_rec=(FreeRec *)malloc(REC_HEADER_SZ+MAX_VAL_LEN);
		int len=REC_HEADER_SZ+MAX_VAL_LEN;

		if (!first_free_rec) {
			tmpi_error(DBG_INTERNAL, "Out of memory.\n");
			return -1;
		}

		if (get(first_free_id, first_free_rec, &len)) {
			free(first_free_rec);
			return -1;
		}

		// verification
		if (first_free_rec->nil!=NIL) {
			tmpi_error(DBG_INTERNAL, "PSTREE structure corrupted.\n");
			return -1;
		}

		if (first_free_rec->idx<first_free_rec->size) {
			first_free_rec->free_id[first_free_rec->idx]=node;
			first_free_rec->idx++;

			if (del(node)) {
				tmpi_error(DBG_APP, "Deletion of node %u failed but ignored.\n", node);
			}

			if (put(first_free_id, first_free_rec, len)) {
				free(first_free_rec);
				return -1;
			}

			free(first_free_rec);
			return 0;
		}

		free(first_free_rec);
	}

	// first_free_id is NIL or already full.
	
	FreeRec *to_be_freed=(FreeRec *)malloc(REC_HEADER_SZ+MAX_VAL_LEN);

	if (!to_be_freed) {
		tmpi_error(DBG_INTERNAL, "Out of Memory.\n");
		return -1;
	}

	int len=REC_HEADER_SZ+MAX_VAL_LEN;

	if (get(node, to_be_freed, &len)) {
		free(to_be_freed);
		return -1;
	}

	to_be_freed->nil=NIL; // free record mark.
	to_be_freed->size=(len-FREEREC_HEADER_SZ)/sizeof(node_t);
	to_be_freed->idx=0;
	to_be_freed->next=first_free_id;
	first_free_id=node;

	// overwriting the old record.
	if (put(node, to_be_freed, FREEREC_HEADER_SZ+to_be_freed->size*sizeof(node_t))) {
		first_free_id=to_be_freed->next; // roll back...
		free(to_be_freed);
		return -1;
	}

	if (put_first_free_id()) {
		first_free_id=to_be_freed->next; // roll back...
		// put_first_free_id(); // another attempt?
		free(to_be_freed);
		return -1;
	}

	free(to_be_freed);

	return 0;
}

int PsTree::alloc_node(node_t *node)
{
	*node=NIL; // invalidate the result

# if defined(__DEBUG__)
	if (!db) {
		tmpi_error(DBG_APP, "DB uninitialized.\n");
		return -1;
	}
# endif

	if (first_free_id==NIL) { /* no more free node from free list */
		return -1;
	}

	if (first_free_id==FREE) {
		tmpi_error(DBG_INTERNAL, "PSTREE data structure corrupted.\n");
		return -1;
	}

	FreeRec *first_free_rec=(FreeRec *)malloc(REC_HEADER_SZ+MAX_VAL_LEN);
	int len=REC_HEADER_SZ+MAX_VAL_LEN;

	if (!first_free_rec) {
		tmpi_error(DBG_INTERNAL, "Out of memory.\n");
		return -1;
	}

	if (get(first_free_id, first_free_rec, &len)) {
		free(first_free_rec);
		return -1;
	}

	if (first_free_rec->nil!=NIL) {
		tmpi_error(DBG_INTERNAL, "PSTREE structure corrupted.\n");
		return -1;
	}

	if (first_free_rec->idx>0) {
		*node=first_free_rec->free_id[--(first_free_rec->idx)];

		if ((*node==NIL)||(*node==FREE)) {
			tmpi_error(DBG_INTERNAL, "PSTREE structure corrupted.\n");
			free(first_free_rec);
			return -1;
		}

		if (put(first_free_id, first_free_rec, len)) {
			free(first_free_rec);
			return -1;
		}

		free(first_free_rec);
		return 0;
	}

	*node=first_free_id;
	first_free_id=first_free_rec->next;
	free(first_free_rec);

	if (put_first_free_id()) {
		tmpi_error(DBG_INTERNAL, "Cannot write back first_free_id record.\n");
		// no way to rollback now.
		return -1;
	}

	return 0;
}

int PsTree::release()
{
	int ret=0;

	if (db) {
		ret=db->close(0);
		delete (db);
		db=NULL;
	}

	return ret;
}
	
int PsTree::get_backstore(char *backstore, unsigned int len)
{
	if (!db)
		return -1;

	if (strlen(bsfile)>=(unsigned)len)
		return -1;

	strcpy(backstore, bsfile);

	return 0;
}

bool PsTree::is_loaded()
{
	if (db)
		return true;

	return false;
}

int PsTree::set_node(node_t node, Rec *rec, unsigned int user_len)
{
	if (!db)
		return -1;

	if ((node==FREE)||(node==NIL)) {
		return -1;
	}

	return put(node, rec, REC_HEADER_SZ+user_len);
}

PsTree::Rec *PsTree::get_node(node_t node, unsigned int *user_len)
{
	if (!db)
		return NULL;

	if ((node==FREE)||(node==NIL)) {
		return NULL;
	}

	Rec *rec=(Rec *)malloc(REC_HEADER_SZ+MAX_VAL_LEN);

	if (!rec) {
		return NULL;
	}

	int len=REC_HEADER_SZ+MAX_VAL_LEN;

	if (get(node, rec, &len)) {
		free(rec);
		return NULL;
	}

	/*
	 * sanity check.
	 */
	if ((rec->parent==NIL)||(rec->parent==FREE)||(rec->child==FREE)||(rec->sibling==FREE)) {
		free(rec);
		return NULL;
	}

	*user_len=len-REC_HEADER_SZ;
	
	return rec;
}

void PsTree::free_node(Rec *rec)
{
	if (rec)
		free(rec);
}

int PsTree::put_node(node_t *node, Rec *rec, unsigned int user_len)
{
	if (!db)
		return -1;

	{
		Dbt okey(node, sizeof(node_t));
		okey.set_ulen(sizeof(node_t));
		okey.set_flags(DB_DBT_USERMEM);
		Dbt odat(rec, REC_HEADER_SZ+user_len);

		if (db->put(NULL, &okey, &odat, DB_APPEND)==0) {
			return 0;
		}
	}
		
	// try to get a node-id from internal free list
	if ((alloc_node(node))<0) { /* node should be valid now */
		return -1;
	}

	return put(*node, rec, REC_HEADER_SZ+user_len);
}

/**
 * prev_node must be a child of node.
 */
PsTree::node_t PsTree::make_child(node_t node, void *data, unsigned int dlen, node_t prev_node=NIL)
{
	if ((!data)||(dlen<=0)) {
		return NIL;
	}

	if ((prev_node==FREE)||(prev_node==ROOT)) { /* root has no peer */
		return NIL;
	}

# if defined(__DEBUG__)
	{
		static bool first_time=true;
		if (first_time) {
			tmpi_error(DBG_INTERNAL, "Check make_child() param consistency in debug mode.\n");
			first_time=false;
		}
		if (prev_node!=NIL) {
			// tmpi_error(DBG_INTERNAL, "BEGIN Check make_child() param consistency.\n");
			unsigned int node_len;
			Rec *node_rec=get_node(node, &node_len);
			if (!node_rec) {
				tmpi_error(DBG_INTERNAL, "checking failed due to failure in get_node()\n");
			}
			else {
				node_t next_list=node_rec->child;
				free(node_rec);
				bool found_match=false;
				while (next_list!=NIL) {
					if (next_list==prev_node) {
						found_match=true;
						break;
					}
					node_rec=get_node(next_list, &node_len);
					if (!node_rec) {
						tmpi_error(DBG_INTERNAL, "checking failed due to failure in get_node()\n");
						break;
					}
					next_list=node_rec->sibling;
					free(node_rec);
				}
				if (!found_match) {
					tmpi_error(DBG_INTERNAL, "Bad prev_node in make_child().\n");
				}
			}

			// tmpi_error(DBG_INTERNAL, " END  Check make_child() param consistency.\n");
		}
	}
# endif
	unsigned int cur_len;
	Rec *cur=NULL;

	if (prev_node==NIL) {
		cur=get_node(node, &cur_len); /* node and db check will be here */
	}
	else {
		cur=get_node(prev_node, &cur_len); /* node and db check will be here */
	}

	if (!cur) {
		return NIL;
	}

	Rec *child=(Rec *)malloc(REC_HEADER_SZ+dlen);

	if (!child) {
		free(cur);
		return NIL;
	}

	child->parent=node;
	if (prev_node==NIL)
		child->sibling=cur->child;
	else
		child->sibling=cur->sibling;
	child->child=NIL;
	memcpy(child->user, data, dlen);

	node_t child_id=NIL;
	if (put_node(&child_id, child, dlen)) {
		free(cur);
		free(child);
		return NIL;
	}

	free(child);

	if (prev_node==NIL)
		cur->child=child_id;
	else 
		cur->sibling=child_id;

	int ret=0;
	if (prev_node==NIL)
		ret=put(node, cur, REC_HEADER_SZ+cur_len);
	else
		ret=put(prev_node, cur, REC_HEADER_SZ+cur_len);

	free(cur);

	if (ret) {
		return NIL;
	}

	return child_id;
}

PsTree::node_t PsTree::first_child(node_t node)
{
	unsigned int cur_len;

	Rec *cur=get_node(node, &cur_len);
	if (!cur) {
		return NIL;
	}

	node_t child_id=cur->child;
	free(cur);

	return child_id;
}

PsTree::node_t PsTree::next_peer(node_t node)
{
	unsigned int cur_len;

	Rec *cur=get_node(node, &cur_len);
	if (!cur) {
		return NIL;
	}

	node_t sibling_id=cur->sibling;
	free(cur);

	return sibling_id;
}

PsTree::node_t PsTree::parent(node_t node)
{
	unsigned int cur_len;

	Rec *cur=get_node(node, &cur_len);
	if (!cur) {
		return NIL;
	}

	node_t parent_id=cur->parent;
	free(cur);

	return parent_id;
}

/**
 * prev_id is a hint value, if presented, it will be checked.
 * otherwise, it will be ignored.
 */
int PsTree::remove_node(node_t prev_id, node_t node)
{
	if (prev_id==FREE) {
		return -1;
	}

	if (node==ROOT) {
		return -1;
	}

	unsigned int cur_len;
	Rec *cur=get_node(node, &cur_len);

	if (!cur)
		return -1;
	
	if (cur->child) {
		free(cur);
		return -1;
	}

	node_t cur_sib=cur->sibling;
	node_t cur_par=cur->parent;
	free(cur);

	Rec *prev_node=NULL;
	bool use_child=true;
	unsigned int prev_len;

	if (prev_id!=NIL) {
		prev_node=get_node(prev_id, &prev_len);
		if (!prev_node) 
			prev_id=NIL;
		else if (prev_node->child==node)
			use_child=true;
		else if (prev_node->sibling==node)
			use_child=false;
		else {
			/* Bad hint value, discard it */
			free(prev_node);
			prev_node=NULL;
			prev_id=NIL;
		}
	}

	if (prev_id==NIL) {
		prev_id=cur_par;
		do {
			prev_node=get_node(prev_id, &prev_len);
			if (!prev_node) {
				return -1;
			}

			if (use_child) {
				if (prev_node->child==node)
					break;
				use_child=false;
				prev_id=prev_node->child;
			}
			else {
				if (prev_node->sibling==node)
					break;
				prev_id=prev_node->sibling;
			}

			free(prev_node);
		} while (prev_id!=NIL);

		if (prev_id==NIL) { 
			/* no previous found! */
			/* very weird! */
			tmpi_error(DBG_INTERNAL, 
					   "Could not find ancestorfor a node.\n");
			return -1;
		}
	}

	if (use_child) {
		prev_node->child=cur_sib;
	}
	else {
		prev_node->sibling=cur_sib;
	}

	if (put(prev_id, prev_node, REC_HEADER_SZ+prev_len)<0) {
		return -1;
	}

	free(prev_node);

	return free_node(node);
}
