// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: pstree.h,v $ - header file for persistent tree structure
 *            using Berkeley DB.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2002/10/08 07:32:53 $
 * Version :  $Revision: 1.12 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 
# ifndef _PSTREE_H_
# define _PSTREE_H_

/* # include <db_cxx.h> */

# if defined(__CHECK_MEM__)
# include "../asmalloc/asmalloc.h"
# endif

# ifndef MAX_FILE_NAME
# define MAX_FILE_NAME 256
# endif

/**
 * default I/O Block size, will be replaced with the
 * real block size if we are able to get it through
 * the statfs() call.
 */
# ifndef IO_PAGE_SIZE
# define IO_PAGE_SIZE 1024
# endif

class Db;

/**
 * Persistent tree using DB::RECNO
 */
class PsTree {
public:
	typedef unsigned int node_t;
	
	struct Rec {
		node_t parent;
		node_t child;
		node_t sibling;
		char user[1];
	};

private:
	struct FreeRec {
		node_t nil; /* corresponds to Rec::parent. */
		node_t next;
		unsigned short size; // total items that can be stored 
		unsigned short idx; // how many have been stored so far?
		node_t free_id[1];
	};

protected:
	static const int REC_HEADER_SZ;

private:
	static const int FREEREC_HEADER_SZ;

public:
	static const node_t ROOT;
	static const node_t NIL;
	static const node_t FREE;
	PsTree(unsigned int max_val_len) : db(NULL), MAX_VAL_LEN(max_val_len) { return; }
	PsTree(char *backstore, void *root, int len, \
			unsigned int cache_sz, unsigned int max_val_len); // create a new tree
	PsTree(char *backstore, unsigned int cache_sz, unsigned int max_val_len); // load an existing tree
	~PsTree();
	int load(char *backstore, unsigned int cache_sz);
	int create(char *backstore, void *root, int len, \
			unsigned int cache_sz);
	int release();
	int get_backstore(char *backstore, unsigned int len);

	node_t make_child(node_t node, void *data, unsigned int dlen, node_t prev_node);
	node_t first_child(node_t node);
	node_t next_peer(node_t node);
	node_t parent(node_t node);
	Rec *get_node(node_t node, unsigned int *len);
	void free_node(Rec *rec);
	int set_node(node_t node, Rec *rec, unsigned int len);
	int remove_node(node_t prev_id, node_t node);
	bool sync();
	
	// do we have a hash table loaded yet?
	bool is_loaded();

private :
	node_t first_free_id; // in-core image at recno=1 (for easy update).
	Db *db;
	int MAX_VAL_LEN;
	char bsfile[MAX_FILE_NAME];

	int set_db_cache(unsigned int cache_sz, unsigned int io_page_size);
	int free_node(node_t node);
	int alloc_node(node_t *node);
	int put(node_t node, void *buf, int len);
	int get(node_t node, void *buf, int *len);
	int del(node_t node);
	int put_first_free_id();
	int get_first_free_id();
	int create_root_rec(void *root, int len);
	int put_node(node_t *node, Rec *rec, unsigned int len);
};

# endif
