// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: pstree_core.cc,v $ - in-core tree structure for debugging.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/09/18 01:08:38 $
 * Version :  $Revision: 1.3 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 
# include <string.h>
# include <stdlib.h>

# include "../debug/tmpi_debug.h"
# include "pstree_core.h"

# if defined(__CHECK_MEM__)
# include "../asmalloc/asmalloc.h"
# endif

# define MAX_REC_NUM 0x200000

const PsTree_Core::node_t PsTree_Core::NIL=(PsTree_Core::node_t)0;
const PsTree_Core::node_t PsTree_Core::ROOT=(PsTree_Core::node_t)1;
const PsTree_Core::node_t PsTree_Core::FREE=(PsTree_Core::node_t)2;

const int PsTree_Core::REC_HEADER_SZ=((int)(&(((struct PsTree_Core::Rec *)(0L))->user)));

PsTree_Core::PsTree_Core(char *backstore, void *root, int len, unsigned int cache_sz, \
		unsigned int max_val_len) :
	table(NULL),
	tab_idx(NIL),
	MAX_VAL_LEN(max_val_len)
{
	int ret=0;

	// on error, leave it in an invalid state.
	if ((ret=create(backstore, root, len, cache_sz))!=0) {
		tmpi_error(DBG_APP, "Cannot create persistent tree: file=\"%s\"!\n", backstore);
	}
}

PsTree_Core::PsTree_Core(char *backstore, unsigned int cache_sz, unsigned int max_val_len) :
	table(NULL),
	tab_idx(NIL),
	MAX_VAL_LEN(max_val_len)
{
	tmpi_error(DBG_APP, "Cannot load PSTREE in debug mode.\n");
}

PsTree_Core::~PsTree_Core()
{
	release();
}

/**
 * load always fails
 */
int PsTree_Core::load(char *backstore, unsigned int cache_sz)
{
	tmpi_error(DBG_APP, "Cannot load PSTREE in debug mode.\n");

	return -1;
}

bool PsTree_Core::sync()
{	
	return true;
}

int PsTree_Core::create_root_rec(void *root, int len)
{
	Rec *root_rec=(Rec *)malloc(REC_HEADER_SZ+len);

	if (!root_rec) {
		return -1;
	}

	root_rec->parent=ROOT;
	root_rec->child=NIL;
	root_rec->sibling=NIL;
	memcpy(root_rec->user, root, len);

	if (put(ROOT, root_rec, REC_HEADER_SZ+len) != 0) {
		free(root_rec);
		return -1;
	}

	return 0;
}

///////////////////////////////////////////////
// The cache_sz is in the unit of kilo-bytes
int PsTree_Core::create(char *backstore, void *root, int len, unsigned int cache_sz)
{
	if (table) 
		return -1;

	if ((!root)||(len<=0)) {
		return -1;
	}

	if (strlen(backstore)>=MAX_FILE_NAME) // name for the backstore too long
		return -1;

	strcpy(bsfile, backstore);

	table=(Rec **)calloc(MAX_REC_NUM, sizeof(Rec *));

	if (!table)
		return -1;

	if (create_root_rec(root, len)) {
		free(table);
		table=NULL;
		return -1;
	}

	tab_idx=FREE+1;

	return 0;
}

int PsTree_Core::put(node_t node, void *val, int len)
{
# if defined(__DEBUG__)
	if ((node==NIL) || (node>=MAX_REC_NUM)) {
		tmpi_error(DBG_CALLER, "Bad node ID in PsTree_Core::put().\n");
		return -1;
	}

	if (!table) {
		tmpi_error(DBG_APP, "Object uninitialized.\n");
		return -1;
	}
# endif

	if (table[node]) {
		free(table[node]);
	}

	table[node]=(Rec *)val;
	table[node]->r_size=len;

	return 0;
}

int PsTree_Core::free_node(node_t node)
{
# if defined(__DEBUG__)
	if ((node==NIL)||(node==FREE)||(node>=MAX_REC_NUM)) {
		tmpi_error(DBG_CALLER, "Bad node ID in PsTree_Core::del().\n");
		return -1;
	}

	if (!table) {
		tmpi_error(DBG_APP, "Object uninitialized.\n");
		return -1;
	}
# endif

	if (!table[node]) {
		return -1;
	}

	free(table[node]);
	table[node]=NULL;

	freeq.push(node);

	return 0;
}

int PsTree_Core::alloc_node(node_t *node)
{
	*node=NIL; // invalidate the result

	if (freeq.empty())
		return -1;

	*node=freeq.front();
	freeq.pop();

	return 0;
}

int PsTree_Core::release()
{
	if (table) {
		for (node_t i=0; i<MAX_REC_NUM; i++) {
			if (table[i])
				free(table[i]);
		}
		free(table);
		table=NULL;
	}

	return 0;
}
	
int PsTree_Core::get_backstore(char *backstore, unsigned int len)
{
	if (!table)
		return -1;

	if (strlen(bsfile)>=(unsigned)len)
		return -1;

	strcpy(backstore, bsfile);

	return 0;
}

bool PsTree_Core::is_loaded()
{
	if (table)
		return true;

	return false;
}

int PsTree_Core::set_node(node_t node, Rec *rec, unsigned int user_len)
{
	if (!table)
		return -1;

	int len=REC_HEADER_SZ+user_len;
	Rec *i_rec=(Rec *)malloc(len);
	if (!i_rec)
		return -1;

	memcpy(i_rec, rec, len);

	return put(node, i_rec, len);
}

PsTree_Core::Rec *PsTree_Core::get_node(node_t node, unsigned int *user_len)
{
	if (!table)
		return NULL;

	if ((node==FREE)||(node==NIL)||(node>=MAX_REC_NUM)) {
		return NULL;
	}

	if (!table[node])
		return NULL;


	int len=table[node]->r_size;
	Rec *rec=(Rec *)malloc(len);

	if (!rec) {
		return NULL;
	}

	memcpy(rec, table[node], len);

	/*
	 * sanity check.
	 */
	if ((rec->parent==NIL)||(rec->parent==FREE)||(rec->child==FREE)||(rec->sibling==FREE)) {
		free(rec);
		return NULL;
	}

	*user_len=len-REC_HEADER_SZ;
	
	return rec;
}

void PsTree_Core::free_node(Rec *rec)
{
	if (rec)
		free(rec);
}

int PsTree_Core::put_node(node_t *node, Rec *rec, unsigned int user_len)
{
	if (!table)
		return -1;

	int len=REC_HEADER_SZ+user_len;

	{
		if ((tab_idx<MAX_REC_NUM)) {
			if (put(tab_idx, rec, len)) {
				return -1;
			}
			*node=tab_idx;
			tab_idx++;
			return 0;
		}
	}
	
	if (alloc_node(node)<0) {
		return -1;
	}

	if (put(*node, rec, len)<0) {
		return -1;
	}

	return 0;
}

/**
 * prev_node must be a child of node.
 * if prev_node==NIL, add as the first child of "node";
 * otherwise, add it as a sibling to "prev_node".
 */
PsTree_Core::node_t PsTree_Core::make_child(node_t node, void *data, unsigned int dlen, node_t prev_node=NIL)
{
	if ((!data)||(dlen<=0)) {
		return NIL;
	}

	if ((prev_node==FREE)||(prev_node==ROOT)) { /* root has no peer */
		return NIL;
	}

# if defined(__DEBUG__)
	{
		static bool first_time=true;
		if (first_time) {
			tmpi_error(DBG_INTERNAL, "Check make_child() param consistency in debug mode.\n");
			first_time=false;
		}
		if (prev_node!=NIL) {
			// tmpi_error(DBG_INTERNAL, "BEGIN Check make_child() param consistency.\n");
			unsigned int node_len;
			Rec *node_rec=get_node(node, &node_len);
			if (!node_rec) {
				tmpi_error(DBG_INTERNAL, "checking failed due to failure in get_node()\n");
			}
			else {
				node_t next_list=node_rec->child;
				free(node_rec);
				bool found_match=false;
				while (next_list!=NIL) {
					if (next_list==prev_node) {
						found_match=true;
						break;
					}
					node_rec=get_node(next_list, &node_len);
					if (!node_rec) {
						tmpi_error(DBG_INTERNAL, "checking failed due to failure in get_node()\n");
						break;
					}
					next_list=node_rec->sibling;
					free(node_rec);
				}
				if (!found_match) {
					tmpi_error(DBG_INTERNAL, "Bad prev_node in make_child().\n");
				}
			}

			// tmpi_error(DBG_INTERNAL, " END  Check make_child() param consistency.\n");
		}
	}
# endif
	unsigned int cur_len;
	Rec *cur=NULL;

	if (prev_node==NIL) {
		cur=get_node(node, &cur_len); /* node and db check will be here */
	}
	else {
		cur=get_node(prev_node, &cur_len); /* node and db check will be here */
	}

	if (!cur) {
		return NIL;
	}

	Rec *child=(Rec *)malloc(REC_HEADER_SZ+dlen);

	if (!child) {
		free(cur);
		return NIL;
	}

	child->parent=node;
	if (prev_node==NIL)
		child->sibling=cur->child;
	else
		child->sibling=cur->sibling;
	child->child=NIL;
	memcpy(child->user, data, dlen);

	node_t child_id=NIL;
	if (put_node(&child_id, child, dlen)) {
		free(cur);
		free(child);
		return NIL;
	}

	if (prev_node==NIL)
		cur->child=child_id;
	else 
		cur->sibling=child_id;

	int ret=0;
	if (prev_node==NIL)
		ret=put(node, cur, REC_HEADER_SZ+cur_len);
	else
		ret=put(prev_node, cur, REC_HEADER_SZ+cur_len);

	if (ret) {
		return NIL;
	}

	return child_id;
}

PsTree_Core::node_t PsTree_Core::first_child(node_t node)
{
	unsigned int cur_len;

	Rec *cur=get_node(node, &cur_len);
	if (!cur) {
		return NIL;
	}

	node_t child_id=cur->child;
	free(cur);

	return child_id;
}

PsTree_Core::node_t PsTree_Core::next_peer(node_t node)
{
	unsigned int cur_len;

	Rec *cur=get_node(node, &cur_len);
	if (!cur) {
		return NIL;
	}

	node_t sibling_id=cur->sibling;
	free(cur);

	return sibling_id;
}

PsTree_Core::node_t PsTree_Core::parent(node_t node)
{
	unsigned int cur_len;

	Rec *cur=get_node(node, &cur_len);
	if (!cur) {
		return NIL;
	}

	node_t parent_id=cur->parent;
	free(cur);

	return parent_id;
}

/**
 * prev_id is a hint value, if presented, it will be checked.
 * otherwise, it will be ignored.
 */
int PsTree_Core::remove_node(node_t prev_id, node_t node)
{
	if (prev_id==FREE) {
		return -1;
	}

	if (node==ROOT) {
		return -1;
	}

	unsigned int cur_len;
	Rec *cur=get_node(node, &cur_len);

	if (!cur)
		return -1;
	
	if (cur->child) {
		free(cur);
		return -1;
	}

	node_t cur_sib=cur->sibling;
	node_t cur_par=cur->parent;
	free(cur);

	Rec *prev_node=NULL;
	bool use_child=true;
	unsigned int prev_len;

	if (prev_id!=NIL) {
		prev_node=get_node(prev_id, &prev_len);
		if (!prev_node)
			/* Bad hint value, discard it */
			prev_id=NIL;
		else if (prev_node->child==node)
			use_child=true;
		else if (prev_node->sibling==node)
			use_child=false;
		else {
			/* Bad hint value, discard it */
			free(prev_node);
			prev_node=NULL;
			prev_id=NIL;
		}
	}

	if (prev_id==NIL) {
		prev_id=cur_par;
		do {
			prev_node=get_node(prev_id, &prev_len);
			if (!prev_node) {
				return -1;
			}

			if (use_child) {
				if (prev_node->child==node)
					break;
				use_child=false;
				prev_id=prev_node->child;
			}
			else {
				if (prev_node->sibling==node)
					break;
				prev_id=prev_node->sibling;
			}

			free(prev_node);
		} while (prev_id!=NIL);

		if (prev_id==NIL) { 
			/* no previous found! */
			/* very weird! */
			tmpi_error(DBG_INTERNAL, 
					   "Could not find ancestorfor a node.\n");
			return -1;
		}
	}

	if (use_child) {
		prev_node->child=cur_sib;
	}
	else {
		prev_node->sibling=cur_sib;
	}

	if (put(prev_id, prev_node, REC_HEADER_SZ+prev_len)<0) {
		return -1;
	}

	return free_node(node);
}
