// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*- 

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: unit_ring.cc,v $ - implementation of the unit ring.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2004/03/18 17:20:43 $
 * Version :  $Revision: 1.7 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

# include "unit_ring.h"
# include "gsl/gsl_rng.h"

# ifndef ENABLE_INF_LOOP_GUARD
# define ENABLE_INF_LOOP_GUARD 1
# endif

/**
 * RandomSequence implements a random sequence generator
 * based on GSL RNG module.
 */
class RandomSequence {
 private:
	gsl_rng *generators[4];
	unsigned int index;
	unsigned long seeds[4];

 public:
	/** the constructor */
	RandomSequence(guid_t & seed) :
		index(0)
	{
		memcpy((char *)seeds, (char *)&seed, sizeof(guid_t));
	}

	/** the destructor */
	~RandomSequence()
	{
		int n_gen=(index<4)?index:4;
		for (int i=0; i<n_gen; i++) {
			gsl_rng_free(generators[i]);
		}
	}

	/** getting a random number */
	double get_one()
	{
		if (index<4) {
			generators[index]=gsl_rng_alloc(gsl_rng_default);
			gsl_rng_set(generators[index], seeds[index]);
		}
		double ret=gsl_rng_uniform(generators[index%4]);
		index++;
		return ret;
	}
};

/** The UnitRing constructor */
UnitRing::UnitRing(bool vnode_enabled) :
	vnoded(vnode_enabled),
	key_cnt(0)
{ 
	pthread_mutex_init(&lock, NULL);
}

/** The UnitRing destructor */
UnitRing::~UnitRing()
{ 
	pthread_mutex_destroy(&lock);
}

/** 
 * add a key to the ring.
 * @param key the key.
 * @param vnode the number of virtual nodes we desire for the key.
 * @return status (true: success ; false: failure).
 * If the key to be added is already in the ring, add() returns
 * false. The vnode parameter for the same key must stay unchanged
 * for all add() calls.
 */
bool UnitRing::add(guid_t &key, int vnode)
{
	if (!vnoded && (vnode != 1))
		return false;

	bool ret;

	pthread_mutex_lock(&lock);

	RandomSequence rs(key);
	double pos=rs.get_one();
	Ring_t::iterator pi=ring.find(UR_Item(pos, key));
	if (pi == ring.end()) {
		key_cnt++;
		ret=true;

		for (int i=0; i<vnode; i++, pos=rs.get_one()) {
			ring.insert(UR_Item(pos, key));
		}
	}
	else {
		ret=false;
	}

	pthread_mutex_unlock(&lock);

	return ret;
}

/** 
 * remove a key from the ring.
 * @param key the key.
 * @return status (true: success; false: key not found).
 * Note the number of virtual nodes will be automatically
 * derived.
 */
bool UnitRing::remove(guid_t &key)
{
	pthread_mutex_lock(&lock);
	RandomSequence rs(key);
	double pos=rs.get_one();
	Ring_t::iterator pi=ring.find(UR_Item(pos, key));
	int n_erased=0;

	while (pi != ring.end()) {
		ring.erase(pi);
		n_erased++;
		if (vnoded) {
			pos=rs.get_one();
			pi=ring.find(UR_Item(pos, key));
		}
		else {
			break;
		}
	}

	if (n_erased>0)
		key_cnt--;

	pthread_mutex_unlock(&lock);

	return (n_erased>0);
}

/** 
 * Get the control ranges for a certain key, the key
 * must be inserted before.
 * @param key the key.
 * @param ranges the ranges controlled by the key.
 * @return status (true: success; false: key not found).
 * In general, if replication is supported (or replication
 * of home sites, we need to traverse back r hosts, where
 * r is the replication degree).
 */
bool UnitRing::get_ranges(guid_t & key, VecURRange & ranges)
{
	pthread_mutex_lock(&lock);
	RandomSequence rs(key);
	double pos=rs.get_one();
	Ring_t::iterator pi=ring.find(UR_Item(pos, key));
	int n_keys=0;

	while (pi != ring.end()) {
		Ring_t::iterator qi=pi;
		if (qi == ring.begin()) {
			qi = ring.end();
		}
		qi--;
		n_keys++;

		double p1=pi->position, p2=qi->position;
		if (p1 != p2) {
			ranges.push_back(URRange(p2, p1));
		}
		else if (pi == qi) {
			ranges.push_back(URRange(p1, p1));
		}
		else {
			// 0 range, ignored.
		}

		if (vnoded) {
			pos=rs.get_one();
			pi=ring.find(UR_Item(pos, key));
		}
		else {
			break;
		}
	}

	pthread_mutex_unlock(&lock);

	return (n_keys>0);
}

/**
 * estimate how many keys in the range. 
 * @param range the range.
 * @return the number of keys in range.
 * since I do not have a good way to calculate the
 * range, I simply return the worst situation.
 */
int UnitRing::n_keys_in_range(URRange & range) const
{
	return ring.size();
}

/**
 * Help functions to get keys within a certain range.
 * @param begin the begin item.
 * @param end the end item (not inclusive).
 * @param keys the guid_t buffer
 * @param len the buffer size.
 * @return the number of keys stored in the buffer.
 * Note: begin will never be ring.end(), and *begin will
 * always be inserted into keys. Len is always > 0.
 */
int UnitRing::get_keys_in_range_no_dup(
		const Ring_t::const_iterator & begin, 
		const Ring_t::const_iterator & end, 
		guid_t * keys, int len)
{
	int n_keys=0;
	// first round.
	Ring_t::const_iterator pi = begin;
	memcpy((char *)&(keys[n_keys]), (char *)&(pi->key), sizeof(guid_t));
	pi++;
	n_keys++;

	while ((pi != end)&&(n_keys<len)) {
		if (pi == ring.end()) {
			pi = ring.begin();
			continue;
		}
		memcpy((char *)&(keys[n_keys]), (char *)&(pi->key), sizeof(guid_t));
		pi++;
		n_keys++;
	}

	return n_keys;
}

/**
 * Help functions to get keys within a certain range.
 * Same as get_keys_in_range_no_dup except that we
 * first store the keys in a set so that duplicates
 * will be elimiated and then we copy them back to
 * the guid_t buffer.
 * @param begin the begin item.
 * @param end the end item (not inclusive).
 * @param keys the guid_t buffer
 * @param len the buffer size.
 * @return the number of keys stored in the buffer.
 */
int UnitRing::get_keys_in_range_check_dup(
		const Ring_t::const_iterator & begin, 
		const Ring_t::const_iterator & end, 
		guid_t * keys, int len)
{
	guid_set_t key_set;
	Ring_t::const_iterator pi = begin;
	key_set.insert(pi->key);
	pi++;

	while ((pi != end)&&((int)key_set.size()<len)) {
		if (pi == ring.end()) {
			pi = ring.begin();
			continue;
		}
		key_set.insert(pi->key);
		pi++;
	}

	int n_keys=0;
	for (guid_set_t::const_iterator pi=key_set.begin();  pi != key_set.end(); pi++, n_keys++) {
		memcpy((char *)&(keys[n_keys]), (char *)&(*pi), sizeof(guid_t));
	}

	return n_keys;
}

/**
 * Get keys that fall in a certain range.
 * @param range the range.
 * @param keys the buffer to be used for fill-in.
 * @param len the size of the buffer.
 * @return the number of keys in range.
 */
int UnitRing::get_keys_in_range(URRange & range, guid_t *keys, int len)
{
	if ((keys == NULL) || (len <= 0))
		return 0;

	int n_keys=0;

	// if (vnoded)
	//	throw "get_keys_in_range() not implemented for vnoded rings.";

	pthread_mutex_lock(&lock);

	if (key_cnt == 0) {
		pthread_mutex_unlock(&lock);
		return 0;
	}

	if (range.from == range.to) {
		if (vnoded)
			n_keys=get_keys_in_range_check_dup(ring.begin(), ring.end(), keys, len);
		else
			n_keys=get_keys_in_range_no_dup(ring.begin(), ring.end(), keys, len);
	}
	else {
		Ring_t::const_iterator start=ring.lower_bound(UR_Item(range.from, Guid::GUID_NIL));
		if (start==ring.end()) {
			start=ring.begin();
		}

		bool is_empty=true;
		if (range.from < range.to) {
			if ((start->position>=range.from)&&(start->position<range.to))
				is_empty=false;
		}
		else {
			if ((start->position>=range.from)||(start->position<range.to))
				is_empty=false;
		}

		if (is_empty) {
			pthread_mutex_unlock(&lock);
			return 0;
		}

		Ring_t::const_iterator end=ring.lower_bound(UR_Item(range.to, Guid::GUID_NIL));
		if (end == ring.end()) {
			end=ring.begin();
		}

		if (vnoded)
			n_keys=get_keys_in_range_check_dup(start, end, keys, len);
		else
			n_keys=get_keys_in_range_no_dup(start, end, keys, len);
	}

	pthread_mutex_unlock(&lock);

	return n_keys;
}

/**
 * Get the neighbors of a certain key. The key may or may not be vnoded.
 * @param neighbors fill-in buffer.
 * @param n_neighbor the size of the buffer.
 * @param key the key.
 * @param n_vnodes the vnodes for the key.
 * @return the number of neighbors found.
 */
int UnitRing::get_neighbors(guid_t * neighbor, int n_neighbor, guid_t & key, int n_vnodes) 
{
	if (n_vnodes<1)
		return -1;

	if ((!neighbor)||(n_neighbor==0))
		return 0;

	// to get neighbors on for a vnoded key, we 
	// can use a heap to keep the nearest neighbors,
	// and for each position of the vnodes, we try n_neighbor
	// neighbors.
	if (n_vnodes > 1)
		throw "get_neighbors() not implemented for n_vodes>1.";

	pthread_mutex_lock(&lock);
	if (n_neighbor > key_cnt)
		n_neighbor = key_cnt;

	RandomSequence rs(key);
	double pos=rs.get_one();
	Ring_t::const_iterator start=ring.lower_bound(UR_Item(pos, key));

	if (!vnoded) {
		int index=0;
# if (ENABLE_INF_LOOP_GUARD)
		bool end_encountered=false; // debugging purpose.
# endif
		while (index < n_neighbor) {
			if (start == ring.end()) {
# if (ENABLE_INF_LOOP_GUARD)
				if (end_encountered) {
					n_neighbor = index; // bad things happened, correct it.
					break;
				}
				end_encountered = true;
# endif
				start=ring.begin();
				continue;
			}
			memcpy((char *)&(neighbor[index]), (char *)&(start->key),
					sizeof(guid_t));
			index++;
			start++;
		}
	}
	else {
		guid_set_t neighbor_set;
		int i=0;
# if (ENABLE_INF_LOOP_GUARD)
		bool end_encountered=false; // debugging purpose.
# endif
		while ((int)neighbor_set.size()<n_neighbor) {
			if (start == ring.end()) {
# if (ENABLE_INF_LOOP_GUARD)
				if (end_encountered) {
					n_neighbor = neighbor_set.size(); // bad things happened, correct it.
					break;
				}
				end_encountered = true;
# endif
				start=ring.begin();
				continue;
			}
			guid_set_t::iterator pi=neighbor_set.find(start->key);
			if (pi == neighbor_set.end()) {
				neighbor_set.insert(start->key);
				memcpy((char *)&(neighbor[i]), (char *)&(start->key), sizeof(guid_t));
				i++;
			}
			start++;
		}
	}

	pthread_mutex_unlock(&lock);

	return n_neighbor;
}

/**
 * Dump the content of the unit ring to stdout.
 */
void UnitRing::dump()
{
	pthread_mutex_lock(&lock);
	for (Ring_t::const_iterator pi=ring.begin(); pi != ring.end(); pi++) {
		char guid_str[128];
		Guid guid(pi->key);
		guid.to_string(guid_str, 128);
		printf("%.5f\t%s\n", pi->position, guid_str);
	}
	pthread_mutex_unlock(&lock);
}

/**
 * Simple integrity check of the unit ring.
 */
bool UnitRing::self_check()
{
	double prev_pos=0.0;
	guid_set_t key_set;
	int sc_key_cnt=0;
	bool ret=true;

	pthread_mutex_lock(&lock);

	for (Ring_t::const_iterator pi=ring.begin(); pi != ring.end(); pi++) {
		if (pi->position<prev_pos) {
			ret = false;
			goto _exit;
		}
		guid_set_t::const_iterator pi2=key_set.find(pi->key);
		if (pi2==key_set.end()) {
			sc_key_cnt++;
			key_set.insert(pi->key);
		}
		else {
			if (!vnoded) {
				ret = false;
				goto _exit;
			}
		}
	}
	if (sc_key_cnt != key_cnt) {
		ret = false;
		goto _exit;
	}

_exit:
	pthread_mutex_unlock(&lock);

	return ret;
}

/**
 * Determine the tuple <pos, key> is on the ring or not.
 * @param pos: the position of the key.
 * @param key: the key.
 * @return: true of the tuple exists, false otherwise.
 */
bool UnitRing::exists(double pos, guid_t key)
{
	pthread_mutex_lock(&lock);
	Ring_t::const_iterator pi=ring.find(UR_Item(pos, key));

	bool ret = (pi != ring.end());
	pthread_mutex_unlock(&lock);

	return ret;
}
