// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*- 

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: unit_ring.h,v $ - header file for a circular unit ring.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2004/03/18 00:02:55 $
 * Version :  $Revision: 1.6 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

# ifndef _UNIT_RING_H_
# define _UNIT_RING_H_

# include "../guid/guid.h"
# include <set>
# include <vector>
# include <pthread.h>

typedef struct URRange {
	double from;
	double to;
	URRange(double a, double b) : from(a), to(b) {}
};

typedef std::vector<URRange> VecURRange;

class UnitRing {
 private:
	bool vnoded;
	struct UR_Item {
		double position;
		guid_t key;
		UR_Item(double p, const guid_t &g) : position(p), key(g) 
		{}
	};

	class UR_Item_cmp : public std::less<UR_Item> {
	 public:
		bool operator()(const UR_Item &x, const UR_Item &y) const {
			if (x.position!=y.position)
				return (x.position<y.position);

			return (memcmp((char *)&(x.key), (char *)&(y.key), 
						sizeof(guid_t))<0);
		}
	};

	typedef std::set<UR_Item, UR_Item_cmp> Ring_t;
	Ring_t ring;
	pthread_mutex_t lock;
	int key_cnt;
	int get_keys_in_range_no_dup(const Ring_t::const_iterator & begin, 
		const Ring_t::const_iterator & end, guid_t * keys, int len);
	int get_keys_in_range_check_dup(const Ring_t::const_iterator & begin, 
		const Ring_t::const_iterator & end, guid_t * keys, int len);

 public:
	UnitRing(bool vnode_enabled=false); /**< the constructor */
	~UnitRing(); /**< the destructor */
	bool add(guid_t & key, int vnode=1); /**< add a key to the ring */
	bool remove(guid_t & key); /**< remove a key from the ring */
	/** get the ranges covered by a certain key, the key must be inserted before */
	bool get_ranges(guid_t & key, VecURRange & ranges);
	int n_keys_in_range(URRange & range) const; /**< estimate how many keys in the range. */
	/** get keys in range (no duplicates) */
	int get_keys_in_range(URRange & range, guid_t *keys, int n_keys); 
	/** get the closest neighbours (no duplicates) */
	int get_neighbors(guid_t * neighbour, int n_neighbour, guid_t & key, int n_vnodes=1); 
	int get_size() const { return ring.size(); }
	int get_key_cnt() const { return key_cnt; }
	void dump(); /**< dump the content of the ring */
	bool self_check(); /**< Checking the data consistency of the ring */
	bool exists(double pos, guid_t key); /**< Checking the existence of a <pos, key> 
										   tuple in the of the ring */
};

# endif
