// -*- mode:C; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: galois.c,v $ - Galois Field (2^16) implementation.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2002/10/08 07:32:53 $
 * Version :  $Revision: 1.3 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 
# include <stdlib.h>
# include "galois_config.h"
# include "galois.h"
# include "rs_types.h"

/**
 * The generation of logarithm tables for GF is
 * based on "A Tutorial on Reed-Solomon Coding for Fault-Tolerance in
 * RAID-like Systems.
 */
static const UINT q_x=GF_GEN;
static const UINT leading_x=1<<(GF_LEN);
static const UINT leading_x_minus_one=(1<<(GF_LEN))-1;
static USHORT *gflog=NULL;
static USHORT *gfilog=NULL;

/**
 * An implementation of Galois GF(2^16) with the generation polynomial
 * x^16+x^12+x^3+x+1 (can be redefined by GF_GEN in galois_config.h).
 */

int gal_init()
{
	UINT b, log;

	if (gflog && gfilog)
		return 0;

	gflog=(USHORT *)malloc(sizeof(USHORT)*leading_x);
	gfilog=(USHORT *)malloc(sizeof(USHORT)*leading_x);

	if ((!gflog)||(!gfilog))
		goto _exit_gal_init;

	b=1;
	gflog[0]=leading_x_minus_one;
	gfilog[leading_x_minus_one]=0;
	for (log = 0; log < leading_x_minus_one; log++) {
		gflog[b]=(USHORT)log;
		gfilog[log]=(USHORT)b;
		b=b<<1;
		if (b & leading_x)
			b=b^q_x;
	}

	return 0;

_exit_gal_init:
	gal_finalize();
	return -1;
}

int gal_finalize()
{
	if (gflog) {
		free(gflog);
		gflog=NULL;
	}

	if (gfilog) {
		free(gfilog);
		gfilog=NULL;
	}

	return 0;
}

int gal_field_size()
{
	return GF_LEN;
}

int gal_poly_gen()
{
	return GF_GEN;
}

/**
 * Galois addition for two vectors of n unsigned shorts
 */
int gal_plus_one(USHORT a, USHORT b, USHORT *c)
{
	*c=a^b;

	return *c;
}

int gal_plus(USHORT *a, USHORT *b, USHORT *c, int n)
{
	int i=0;

	for (; i<n; i++) {
		c[i]=a[i]^b[i];
	}

	return 0;
}

/**
 * Galois multiplication for two vectors of n unsigned shorts
 */
static int gal_mult_one_(USHORT a, USHORT b, USHORT *c)
{
	if ((!a)||(!b)) {
		*c=0;
	}
	else {
		int sum_log;
		sum_log=gflog[a]+gflog[b];
		if (sum_log >=leading_x_minus_one)
			sum_log -= leading_x_minus_one;
		*c=gfilog[sum_log];
	}

	return *c;
}

int gal_mult_one(USHORT a, USHORT b, USHORT *c)
{
	if (gal_init()<0) {
		return -1;
	}

	return gal_mult_one_(a, b, c);
}

int gal_mult(USHORT *a, USHORT *b, USHORT *c, int n)
{
	int i=0;

	if (gal_init()<0) {
		return -1;
	}

	for (; i<n; i++) {
		gal_mult_one_(a[i], b[i], &c[i]);
	}

	return 0;
}

/**
 * Galois division for two vectors of n unsigned shorts
 * Returns the number of elements that failed the division.
 */
static int gal_div_one_(USHORT a, USHORT b, USHORT *c)
{
	if (!b)
		return -1;

	if (!a)
		*c=0;
	else {
		int diff_log;
		diff_log=gflog[a]-gflog[b];
		if (diff_log < 0)
			diff_log += leading_x_minus_one;
		*c=gfilog[diff_log];
	}

	return *c;
}

int gal_div_one(USHORT a, USHORT b, USHORT *c)
{
	if (gal_init()<0) {
		return -1;
	}

	return gal_div_one_(a, b, c);
}

int gal_div(USHORT *a, USHORT *b, USHORT *c, int n)
{
	int i=0;
	int has_error=0;

	if (gal_init()<0) {
		return -1;
	}

	for (; i<n; i++) {
		if (gal_div_one_(a[i], b[i], &c[i])<0)
			has_error++;
	}

	return has_error;
}

static int gal_log_one_(USHORT a, USHORT *c)
{
	/**
	if (!a)
		return -1;
	**/
	*c=gflog[a];

	return *c;
}

int gal_log_one(USHORT a, USHORT *c)
{
	if (gal_init()<0) {
		return -1;
	}

	return gal_log_one_(a, c);
}

int gal_log(USHORT *a, USHORT *c, int n)
{
	int i=0;
	int has_error=0;

	if (gal_init()<0) {
		return -1;
	}

	for (; i<n; i++) {
		if (gal_log_one_(a[i], &c[i])<0)
			has_error++;
	}

	return has_error;
}

static int gal_ilog_one_(USHORT a, USHORT *c)
{
	*c=gfilog[a];

	return *c;
}

int gal_ilog_one(USHORT a, USHORT *c)
{
	if (gal_init()<0) {
		return -1;
	}

	return gal_ilog_one_(a, c);
}

int gal_ilog(USHORT *a, USHORT *c, int n)
{
	int i=0;

	if (gal_init()<0) {
		return -1;
	}

	for (; i<n; i++) {
		gal_ilog_one(a[i], &c[i]);
	}

	return 0;
}
