// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.
 
  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
 
   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.
 
   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
 
   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.
 
   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.
 
  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: rwlock.cc,v $ - read-write lock implementation.
 *            This implementation is based on Shlomi Fish's implementation
 *            at http://vipe.technion.ac.il/~shlomif/rwlock/. We assume
 *            that threads blocked on the same condition variable
 *            (pthread_cond_t) are waken up in the same order as they call
 *            pthread_cond_wait(), and globally, threads are granted 
 *            the lock in the FCFS order. One extension from the original
 *            implementation, this version supports lock promotion. Aka,
 *            a thread holding a read lock can prompt the lock it holds 
 *            to a write lock.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/06/11 08:44:06 $
 * Version :  $Revision: 1.8 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */

# include <sys/time.h>
# include <time.h>
# include <pthread.h>
# include <errno.h>

# include "rwlock.h"


RWLockItem::RWLockItem(bool writer, int rid) :
		is_writer(writer),
		in_progress(false),
		n_threads(0),
		round(rid)
{
	pthread_cond_init(&cond, NULL);
}

RWLockItem::~RWLockItem()
{
	pthread_cond_destroy(&cond);
}

RWLock::RWLock() :
		n_readers(0),
		is_writer(false),
		current_round(0),
		next_round(1)
{
	pthread_mutex_init(&lock, NULL);
}

RWLock::~RWLock()
{
	pthread_mutex_destroy(&lock);
}

// The correctness of the original program relies on the fact that
// a cond_signal or cond_bcast will never get lost.
// A.k.a: in the following example:
// 	t1: interrupt to process P.
// 	t2: cond_signal to process P.
// 	t3: process P wake up because of the interrupt.
// 	t4: P goes back to cond_wait()
// *t5* P receives the cond_signal sent out on t2!
// This may not be true, so we modified the program such that the wait
// stops only on any of the two conditions: 
// (1) timed-out; 
// (2) current_round == item->round.
// This makes the program more reliable. For instance, if two 
// threads are waken up, one is due to the cond_signal, the other is due to
// interrupt. The second process may continue to enter the 
// critical section and change item->round. The first thread is
// waken up, and detects that current_round != item->round, it 
// will try again.
bool RWLock::pending_on_item(RWLockItem *item, int tout)
{
	item->n_threads++;

	int ret;

	if (tout<0) {
		do {
			ret=pthread_cond_wait(&(item->cond), &(lock));
		} while (item->round!=current_round);
	} else {
		struct timespec abstime;
		struct timeval now;

		gettimeofday(&now, NULL);

		abstime.tv_sec = now.tv_sec + tout/1000;
		abstime.tv_nsec = now.tv_usec * 1000 + (tout%1000)*1000000;

		/*
		 * We try to get a timed wait on cond. If it times-out, we ask
		 * the user if he wishes to continue.
		 */
		do {
			ret = pthread_cond_timedwait(&(item->cond), &(lock), &abstime);
		} while ((ret!=ETIMEDOUT)&&(item->round!=current_round));
	}

	assert(item->n_threads > 0);

	item->n_threads--;

	/**
	 * We failed for some reason (probably a time out - let's disable
	 * this item, so it won't be processed.
	 */
	if (current_round != item->round) {
		// The item may become a garbage,
		// but since it is not at the queue head,
		// I cannot clear it yet.

		return false;
	}

	return true;
}

// I am exiting the item now, check whether
// the item may need to be released.
void RWLock::exit_from_item(RWLockItem *item)
{
	/**
	 * Check if there are more threads pending. If not - we can extract
	 * and destroy the item 
	 */
	if (item->n_threads == 0) {
		// I am the last one exiting the pack.
		// We don't need the item anymore so let's deallocate it.

		RWLockItem *item2=(RWLockItem *)queue.pop_front();
		assert(item == item2);
		delete item;
	}

	return;
}

bool RWLock::write_lock_generic_internal(int tout)
{
	remove_junk_from_head();

	if (current_round != 0) {
		// the lock is currently held by somebody.
		assert( (is_writer) || (n_readers>0) || !queue.is_empty() );

		bool wait_for_access=(tout>0)||(tout==-1);

		if (! wait_for_access) {
			return false;
		}

		RWLockItem * item=(RWLockItem *)queue.peek_tail();

		if (
		    /* The queue is not empty */
		    (item != NULL) &&
		    /* It is a writers' pack */
		    item->is_writer &&
		    /* The first thread was not yet accepted from it */
		    (! item->in_progress)
		) {
			// don't have to do anything.
		} else {
			/* Someone is using the RWLock. Let's enqueue myself */
			item = new RWLockItem(true, inc_round()); // create a new writer item.
			queue.push_back(item);
		}

		/* Wait upon the condition variable to gain access to the rwlock */
		if (pending_on_item(item, tout) == false) {
			return false;
		}

		assert( (n_readers == 0) && (is_writer == false) );

		/* Now we have a write access */
		is_writer = true;
		item->round=inc_round(); // change the item's round id.

		exit_from_item(item);

		return true;
	} else {
		// assert( current_round == 0);
		assert( (is_writer==false) && (n_readers==0) && queue.is_empty() );

		current_round = inc_round();

		/**
		 * There aren't any readers or writers in the queue we
		 * can gain access immidiately.
		 */
		/* Mark the lock as such that contains a writer */
		is_writer = true;

		return true;
	}
}

void RWLock::remove_junk_from_head(void)
{
	/**
	 * This loop removes junk items (those whith n_threads == 0) from the queue until it 
	 * encounters a good pack.
	 */
	RWLockItem *item = (RWLockItem *)queue.peek_head();

	/**
	 * As long as we did not reach the end of the queue and as long as the item
	 * is junk - destroy the item, dequeue it and peek the next item
	 */
	while ((item != NULL) && (item->n_threads == 0)) {
		item=(RWLockItem *)queue.pop_front();
		delete item;
		item = (RWLockItem *)queue.peek_head();
	}
}

bool RWLock::read_lock_generic_internal(int tout)
{
	/*
	 * It is possible that there are some disabled items clogging the
	 * queue, which will prevent this thread from being accepted immidiately.
	 *
	 * Therefore, try to remove them
	 * */
	remove_junk_from_head();

	if (current_round == 0) {
		// an idle lock.
		assert( (is_writer==false) && (n_readers==0) && queue.is_empty() );
		current_round = inc_round();
		/* Increment the number of readers. */
		n_readers++;

		/* Now we have read access */
		return true;
	}
	else if ((is_writer == false) && queue.is_empty()) {
		assert(n_readers>0);
		// the lock is currently being held by readers.
		// and nobody is waiting to enter the queue.

		/* Increment the number of readers. */
		n_readers++;

		/* Now we have read access */
		return true;
	} else {
		bool wait_for_access=(tout>0)||(tout==-1);

		if (! wait_for_access) {
			return false;
		}

		RWLockItem *item = (RWLockItem *)queue.peek_tail();
		if (
		    /* The queue is not empty */
		    (item != NULL) &&
		    /* It is a readers' pack */
		    (! item->is_writer) &&
		    /* The first thread was not yet accepted from it */
			(! item->in_progress)
		) {
			/* Use this item */
		} else {
			/* Let's enqueue myself */
			item = new RWLockItem(false, inc_round());
			queue.push_back(item);
		}

		/* Wait upon the condition variable to gain access to the rwlock */
		if (pending_on_item(item, tout) == false) {
			return false;
		}

		assert(is_writer==false);

		/* Now we have a read access */
		n_readers++;

		exit_from_item(item);

		return true;
	}
}

bool RWLock::accept_pending_items(void)
{
	RWLockItem * item;

	item=(RWLockItem *)queue.peek_head();

	/* The queue is empty, so we don't need to release anything */
	if (item == NULL) {
		current_round = 0;
		return true;
	}

	/* Prohibit other readers/writers from joining this pack from now on */
	item->in_progress = true;
	current_round = item->round;

	int ret;
	/* Check if it's a writer or a reader */
	if (item->is_writer) {
		/* Wake up one writer. */
		ret=pthread_cond_signal( &(item->cond) );
	} else {
		/* Wake up all readers. */
		ret=pthread_cond_broadcast( &(item->cond) );
	}

	return (ret==0);
}

bool RWLock::release_internal(void)
{
	remove_junk_from_head();

	/*
	 * If there is a writer locking the lock, than obviously he is the one
	 * that releases it.
	 * */
	if (is_writer) {
		assert(n_readers == 0);

		is_writer = false;

		return accept_pending_items();
	} else {
		/*
		 * Croak if there are also no readers - someone is trying to release
		 * a lock which he does not have...
		 * */
		assert(n_readers > 0);

		n_readers--;

		/*
		 * Check if the lock is now free of readers. If so, we can accept
		 * pending writers.
		 * */
		if (n_readers == 0) {
			return accept_pending_items();
		}
	}

	return true;
}

# if (0)
bool RWLock::promote_generic_internal(bool wait_for_access)
{
    if (!wait_for_access)
        return false;

    if (release_internal() == false) {
        assert(false);
    }

    if (write_lock_generic_internal(-1) == false) {
        assert(false);
    }

    return true;
}
# endif

bool RWLock::promote_generic_internal(bool wait_for_access)
{
	remove_junk_from_head();

	// check the impossible. The current thrad
	// must hold a read lock.
	assert((is_writer == false) && (n_readers>0) && (current_round != 0));

	if (n_readers == 1) {
		// I am the only reader. Acquire the lock immediately.
		n_readers = 0;
		is_writer = true;
		current_round=inc_round();

		return true;
	}

	if (!wait_for_access) {
		return false;
	}

	n_readers --;
	RWLockItem *item = (RWLockItem *)queue.peek_head();

	if (item) {
        item->in_progress=false; // stop the progress of the first item.
		item->round=inc_round(); // prevent the next pack from waking up.
    }

	item = new RWLockItem(true, inc_round());
	queue.push_front(item);

	/* Wait upon the condition variable to gain access to the rwlock */
	if (pending_on_item(item, -1) == false) {
		// terrible things have happened.
		// maybe I should re-acquire the readlock?
		assert(false); 
	}

	assert( (is_writer == false) && (n_readers == 0));
	/* Now we have a write access */
	is_writer = true;

	item->round=inc_round(); // change the item's round id.

	exit_from_item(item);

	return true;
}

bool RWLock::demote_internal()
{
	remove_junk_from_head();

	// check the impossible. The current thrad
	// must hold a write lock.
	assert((is_writer == true) && (n_readers == 0) && (current_round != 0));

	is_writer = false;
	n_readers = 1;

	return true;
}
