// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.
 
  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
 
   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.
 
   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
 
   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.
 
   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.
 
  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: rwlock.h,v $ - header file for read-write locks.
 *            This implementation is based on Shlomi Fish's implementation
 *            at http://vipe.technion.ac.il/~shlomif/rwlock/. We assume
 *            that threads blocked on the same condition variable
 *            (pthread_cond_t) are waken up in the same order as they call
 *            pthread_cond_wait(), and globally, threads are granted 
 *            the lock in the FCFS order. One extension from the original
 *            implementation, this version supports lock promotion. Aka,
 *            a thread holding a read lock can prompt the lock it holds 
 *            to a write lock. Some bugs in the original code have also
 *            been removed.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/06/11 08:44:06 $
 * Version :  $Revision: 1.7 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */

# ifndef _RWLOCK_H_
# define _RWLOCK_H_

# include <pthread.h>
# include "queue.h"

class RWLockItem : public Listable
{
public:
	pthread_cond_t cond;
	bool is_writer;
	bool in_progress;
	int n_threads;
	unsigned int round;

public:
	RWLockItem::RWLockItem(bool writer, int rid);
	~RWLockItem();
};

class RWLock
{
private:
	pthread_mutex_t lock;
	Queue queue;
	int n_readers;
	bool is_writer;
	unsigned int current_round;
	unsigned int next_round;

private:
	// (current_round == 0) means that nobody holds the lock.
	unsigned int inc_round() 
	{ 
		unsigned int ret=0;
		do {
			ret=next_round++;
		} while (ret == 0);

		return ret;
	}

	bool read_lock_generic_internal(int tout);
	bool write_lock_generic_internal(int tout);
	bool promote_generic_internal(bool wait_for_access);
	bool release_internal();
	bool demote_internal();

	bool read_lock_generic(int tout)
	{
		if (pthread_mutex_lock(&lock)<0)
			return false;
		bool ret=read_lock_generic_internal(tout);
		pthread_mutex_unlock(&lock)
		;

		return ret;
	}

	bool write_lock_generic(int tout)
	{
		if (pthread_mutex_lock(&lock)<0)
			return false;
		bool ret=write_lock_generic_internal(tout);
		pthread_mutex_unlock(&lock)
		;

		return ret;
	}

	bool promote_generic(bool wait_for_access)
	{
		if (pthread_mutex_lock(&lock)<0)
			return false;
		bool ret=promote_generic_internal(wait_for_access);
		pthread_mutex_unlock(&lock)
		;

		return ret;
	}

	void remove_junk_from_head(void);
	bool accept_pending_items(void);
	bool pending_on_item(RWLockItem *item, int tout);
	void exit_from_item(RWLockItem *item);

public:
	RWLock();
	~RWLock();

	// All timeout values are in mili-seconds.

	bool read_lock(void)
	{
		return read_lock_generic(-1);
	}
	bool read_trylock(void)
	{
		return read_lock_generic(0);
	}
	bool read_timedlock(int tout)
	{
		if (tout<=0)
			return false;
		return read_lock_generic(tout);
	}

	bool write_lock(void)
	{
		return write_lock_generic(-1);
	}
	bool write_trylock(void)
	{
		return write_lock_generic(0);
	}
	bool write_timedlock(int tout)
	{
		if (tout<=0)
			return false;
		return write_lock_generic(tout);
	}

	bool promote_lock(void)
	{
		return promote_generic(true);
        // return (release() && write_lock());
	}

	bool demote_lock(void)
	{
		if (pthread_mutex_lock(&lock)<0)
			return false;
		bool ret=demote_internal();
		pthread_mutex_unlock(&lock);

		return ret;
	}

	bool promote_trylock(void)
	{
        // return false;
		return promote_generic(false);
	}

	bool release(void)
	{
		if (pthread_mutex_lock(&lock)<0)
			return false;
		bool ret=release_internal();
		pthread_mutex_unlock(&lock)
		;

		return ret;
	}
};

# endif
