// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: calevent.cc,v $ - calendar queue.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2002/10/08 07:32:53 $
 * Version :  $Revision: 1.7 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 
/*
 * ID:            $Id: calevent.cc,v 1.7 2002/10/08 07:32:53 htang Exp $
 * DESCRIPTION:   Calendar Events Implementation
 * AUTHOR:        Hong Tang (htang@cs.ucsb.edu)
 * Last Modified: $Date: 2002/10/08 07:32:53 $
 * Version:       $Revision: 1.7 $
 */

/*************************************************************************/
/*                                                                       */
/*  Calendar Queueing Method for Future Event List Manipulation:         */
/*      As presented in the article by Randy Brown in Communications of  */
/*  the ACM, Oct. 1988 Vol. 30 Num. 10.                                  */
/*	    Code based on a C implementation by Eric Callman 3/4/92          */
/*	    Object-oriented design. Each event needs to derive from the      */
/*	Event object and must contains three virtual functions:              */
/*  bool good(), void dump(), and the destructor.                        */
/*	    The estimation of bucket width is based on history instead of    */
/*  sampling, which is presumably better and faster.                     */
/*      Maximum number of buckets has been lifted to 262144.             */
/*                                                                       */
/*************************************************************************/

# include <stdlib.h>
# include <math.h>
# include "calevent.h"

EventChain::EventChain() :
	calendar(NULL),
	resize_enable(true),
	sim_time(0),
	// latest_time(0),
	qsize(0),
	e_samples_idx(0),
	t_samples_idx(0),
	t_samples_sum(0),
	laste(NULL),
	lastt(-1),
	lastdqt(-1),
	n_enqueue(0),
	n_dequeue(0),
	n_xdq(0),
	qsize_op(0),
	n_traverse(0),
	resize_cnt(0) 
	// , qsize_integral(0)
{
	for (int i=0; i<CAL_SAMPLE_SIZE; i++)
		e_samples[i]=0;
	calq=(Event **)calloc(CAL_QSPACE, sizeof(Event *));
	init();
}

EventChain::~EventChain() 
{
	if (calq)
		free(calq);
}

# define TOO_MANY_TRAVERSE 16

bool EventChain::schedule(Event *e)
{
	// if (e == NULL)
		// return false;

	double t=e->time();
	if (sim_time>t)
		return false;


	// if (latest_time<t)
		// latest_time=t;

	qsize_op+=qsize;

	/**
	 * Shortcut for same time consecutive events
	 */
	int k=0;
	if (lastt==t) {
		e->e_cnt=0;
		e->next()=laste->next();
		laste->next()=e;
		qsize++;
	}
	else {
		k=insert(e, t);
	}

	n_traverse+=k;
	t_samples_sum+=(k-t_samples[t_samples_idx]);
	t_samples[t_samples_idx]=k;
	t_samples_idx=(t_samples_idx+1)%CAL_SAMPLE_SIZE;

	laste=e;
	lastt=t;

	n_enqueue++;

	if ((qsize > top_threshold) && ((2*nbuckets) <= CAL_MAXNBUCKETS) && resize_enable)
		resize(2 * nbuckets); /* double the size of the calendar */
	else {
		int too_many_traverse=TOO_MANY_TRAVERSE*((qsize+nbuckets-1)/nbuckets);
		int avg_trv=t_samples_sum/CAL_SAMPLE_SIZE;
		if (avg_trv>(too_many_traverse)) {
			// adjust t_sampel_sum and see if that makes it unnecessary.
			int adj_avg_trv=0;
			int j=0;
			for (int i=0; i<CAL_SAMPLE_SIZE; i++) {
				if (t_samples[i]<=(avg_trv*3)) {
					adj_avg_trv+=t_samples[i];
					j++;
				}
			}
			adj_avg_trv/=j;
			if (adj_avg_trv>too_many_traverse) {
				resize(nbuckets); /* adjust width */
			}
		}
	}

	return true;
}

/*************************************************************************/
/*                                                                       */
/* insert:                                                               */
/*     This routine adds one entry to the future event queue.            */
/*                                                                       */
/*************************************************************************/

// width: time period per bucket.
int EventChain::insert(Event *e, double t)
{
	/* calculate the number of the bucket to place new entry in */
	// double t=e->time();
	int i = (int)(t / (double)width); /* find virtual bucket */
	i=i%nbuckets;

	int k=0;
	Event **ppcur = &(calendar[i]);

	e->e_cnt=1;
	while ((*ppcur!=NULL) && ((*ppcur)->time()<t)) {
		k+=(*ppcur)->e_cnt;
		ppcur = &((*ppcur)->next());
	}

	if ((*ppcur!=NULL) && ((*ppcur)->time()<=t)) {
		k++;
		e->e_cnt=0;
		do {
			ppcur = &((*ppcur)->next());
		} while ((*ppcur!=NULL) && ((*ppcur)->time()<=t));
	}

	e->next() = *ppcur;
	*ppcur = e;
	qsize++; /* update the size of the queue */

	return k;
}

/*************************************************************************/
/*                                                                       */
/* remove:                                                               */
/*     This function removes the next item from the calendar queue.      */
/*                                                                       */
/*************************************************************************/

Event *EventChain::remove()
{
	if (qsize==0)
		return NULL;

	int i = lastbucket; /* search for next entry */
	bool exhaust=false;
	while (1) {
		if (calendar[i]!=NULL) {
			exhaust=true; // might need to do exhaustive search.
		}
		if ((calendar[i] != NULL) && (calendar[i]->time() < buckettop)) {
			Event *e = calendar[i];
			calendar[i] = e->next();
			// lastprio = sim_time = e->time();
			sim_time = e->time();
			lastbucket = i; /* update position in calendar and set found flag */
			qsize--; /* update the size of the queue */
			return e;
		}
		else {
			i++;
			if (i == nbuckets)
				i = 0; /* go on to next bucket */
			buckettop += width;
			if (i == lastbucket)
				break;
		}
	}

	// hopefully this never happens!
	if (exhaust) {
		double earliest=-1.0;

		for (int j=0; j<nbuckets; j++) {
			if (calendar[j]==NULL)
				continue;
			double t=calendar[j]->time();
			if ((earliest<0)||(t<earliest)) {
				lastbucket=j;
				earliest=t;
			}
		}
		if (earliest<0)
			return NULL;
		Event *e = calendar[lastbucket];
		calendar[lastbucket] = e->next();
		// lastprio=sim_time = e->time();
		sim_time = e->time();
		buckettop=(double)((int) (sim_time/width)+1)*width+(0.5*width);
		qsize--; /* update the size of the queue */
		return e;
	}
	else {
		return NULL;
	}
}

Event *EventChain::next_event()
{
	Event *e=remove();
	if (e) {
		// int i=(e_samples_idx+CAL_SAMPLE_SIZE-1)%CAL_SAMPLE_SIZE;
		double t=e->time();
		// if (e_samples[i]!=t) { // no dup rec of events occurred at the same time
		if (lastdqt!=t) { // no dup rec of events occurred at the same time
			e_samples[e_samples_idx]=t;
			e_samples_idx=(e_samples_idx+1)%CAL_SAMPLE_SIZE;
			n_xdq++;
			lastdqt=t;
		}
		if ((qsize < bot_threshold)&&e&&(resize_enable)) /* cut the size of the queue in */
			resize((int)((double)nbuckets / 2)); /* half if needed */
		n_dequeue++;
		qsize_op+=(qsize+1);
		if (t==lastt) {
			laste=NULL;
			lastt=-1;
		}
	}

	return e;
}

/*************************************************************************/
/*                                                                       */
/* localinit:                                                            */
/*     This routine initializes a bucket array within the array calq[].  */
/* Calendar[0] is made equal to calq[qbase], the number of buckets is    */
/* nbuck.                                                                */
/*                                                                       */
/*************************************************************************/

void EventChain::localinit(int qbase, int nbuck, double bwidth, double startprio)
{
	int i;
	long n;

	firstsub = qbase; /* set position and size of new queue */
	calendar = calq+qbase; /* initialize the new queue */
	width = bwidth;
	nbuckets = nbuck;

	qsize = 0; /* initialize as empty */
	for (i = 0; i < nbuckets; i++)
		calendar[i] = NULL;
	// lastprio = startprio; /* set up initial position */
	n = (long)((double)startprio / width); /* virtual bucket */
	lastbucket = n % nbuckets;
	buckettop = (double)(n + 1) * width + (0.5 * width);

	/* set up queue size change thresholds */
	bot_threshold = (int)((double)nbuckets / 2) - 2;
	top_threshold = 2 * nbuckets;
}

# define MIN_NUM_BUCKETS 16
# define MIN_NUM_ITEMS 16
# define DEFAULT_WIDTH 0.1
# define DEFAULT_EPSILON_WIDTH 0.2
# define MIN_WIDTH 0.001

/*************************************************************************/
/*                                                                       */
/* init:                                                                 */
/*     This routine uses localinit() to start the calendar with          */
/* nothing in it for initializing.                                       */
/*                                                                       */
/*************************************************************************/

void EventChain::init() /* this initializes an empty queue */
{
	localinit(0, MIN_NUM_BUCKETS, DEFAULT_WIDTH, sim_time);
}

/*************************************************************************/
/*                                                                       */
/* resize:                                                               */
/*     This routine copies the queue into a calendar with the number of  */
/* buckets specified in the parameter newsize.  The new queue is in the  */
/* opposit end of calq from the original.                                */
/*                                                                       */
/*************************************************************************/

void EventChain::resize(int newsize)
{
	if (qsize<=MIN_NUM_ITEMS)
		return;

	if ((newsize<nbuckets)&&(newsize<MIN_NUM_BUCKETS))
		return;

	double bwidth = new_width(); /* find new bucket width */
	if (bwidth==0) // wait off till I can find a better width
		return;

	// if no change of the number of buckets, we will try 
	if ((newsize==nbuckets)&&(fabs(bwidth-width)<(width*DEFAULT_EPSILON_WIDTH)))
		return;

	resize_cnt++;

	Event **oldcalendar = calendar; /* save location and size of old calendar */
	int oldnbuckets = nbuckets;/* for use when copying calendar */

	// debug
	// int oldqsize=qsize; // debug
	// int ic=0; // debug

	if (firstsub == 0) /* initialize new calendar */
		// localinit(CAL_QSPACE-newsize, newsize, bwidth, lastprio);
		localinit(CAL_QSPACE-newsize, newsize, bwidth, sim_time);
	else
		// localinit(0, newsize, bwidth, lastprio);
		localinit(0, newsize, bwidth, sim_time);

	for (int i = oldnbuckets-1; i >= 0; i--) { 
		/* copy elements to new calendar */
		Event *temp = oldcalendar[i];
		while (temp != NULL) {
			/* traverse linked list of elements */
			Event *temp_next=temp->next();
			insert(temp, temp->time()); /* use insert routine for transfer */
			temp = temp_next;
			// ic++; // debug
		}
	}

	// if (ic!=oldqsize) { // debug
		// printf("\nInconsistent qsize found during resize() %d!=%d.\n", ic, oldqsize); // debug
		// abort(); // debug
	// } // debug
}

/*************************************************************************/
/*                                                                       */
/* new_width:                                                            */
/*     This function returns the width that the buckets should have      */
/* based on the time of the past 32 dequeue events so that there will    */
/* be about 3 items in each bucket.                                      */
/*                                                                       */
/*************************************************************************/

double EventChain::new_width() /* this calculates the width to use for buckets */
{
	int begin=e_samples_idx; // assuming event simulation starts at 0 
	double t_begin=e_samples[begin];
	int size=(t_begin>0)?CAL_SAMPLE_SIZE:begin;

	if (size<MIN_NUM_ITEMS) {
		return 0; // no change
	}

	int end=(begin+CAL_SAMPLE_SIZE-1)%CAL_SAMPLE_SIZE;
	double t_end=e_samples[end];
	double average = (t_end-t_begin) / size; /* get the average */

	double newaverage=0;
	int k=0;
	for (int i=end-size+2; i<=end; i++) {
		int j=(i+CAL_SAMPLE_SIZE)%CAL_SAMPLE_SIZE;
		int j_1=(j+CAL_SAMPLE_SIZE-1)%CAL_SAMPLE_SIZE;
		double delta=e_samples[j]-e_samples[j_1];
		if (delta<0) {
			printf("Weird! Delta width < 0.\n");
			abort();
		}

		if (delta<=(average*3)) {
			newaverage+=delta;
			k++;
		}
	}

	if (k<MIN_NUM_ITEMS/2) { 
		// should never happen except during the beginning! 
		// too small a sample size
		return 0;
	}

	newaverage/=k;

	// if (newaverage < MIN_WIDTH)
		// newaverage=MIN_WIDTH;

	return (newaverage*2);
}


void EventChain::statistics()
{
	printf("%.5fms/%.5fms width, %d resizes, %.2f trav, %lld enq, %lld deq, %.2f qlen.\n", 
			get_width()*1000, opt_width()*1000, get_resize_cnt(), get_num_traverse(), num_enqueue(), num_dequeue(),
			queue_length());
}
