// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: calevent.h,v $ - Header File for Calendar Events Implementation
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2002/10/08 07:32:53 $
 * Version :  $Revision: 1.7 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

# ifndef _CAL_EVENTS_H_
# define _CAL_EVENTS_H_
# include <stdio.h>
# include "../debug/tmpi_debug.h"

/*************************************************************************/
/*                                                                       */
/*  Calendar Queueing Method for Future Event List Manipulation:         */
/*      As presented in the article by Randy Brown in Communications of  */
/*  the ACM, Oct. 1988 Vol. 30 Num. 10.                                  */
/*	    Code based on a C implementation by Eric Callman 3/4/92          */
/*	    Object-oriented design. Each event needs to derive from the      */
/*	Event object and must contains three virtual functions:              */
/*  bool good(), void dump(), and the destructor.                        */
/*	    The estimation of bucket width is based on history instead of    */
/*  sampling, which is presumably better and faster.                     */
/*      Maximum number of buckets has been lifted to 262144.             */
/*                                                                       */
/*      Performance in the Service-Parallel Invocation Simulator         */
/* (on 400MHz PIII, RedHat Linux 6.0, including simulator overhead):     */
/* 22.8M enq, 22.5M deq, 0.28M qlen, total time 409 secs.                */
/* Average 9.04us per operation.                                         */
/*      Using gprof sampling, schedule() and next_event() accounts for   */
/* about 10.6% of the time. So the pure cost would be 0.96us per op.     */
/* In fact, schedule() is 1.87 times more expensive than next_event().   */
/* And resize() accounts for 12.3% of the queue op time which is mainly  */
/* included in schedule() because bucket shrinking down seldom happens   */
/* except when the queue is to be destroyed. Also, shrinking seems       */
/* cheaper than inflating. Each insert() will in average call 6.2 time() */
/* and 6.8 next(). Each remove() will in average call 2.3 time() and     */
/* 1 next(). Total actual resize(): 27 (half occurred during shutdown).  */
/*                                                                       */
/*************************************************************************/

// simulation event base class
class Event
{
 private:
	double sched_time;	// scheduled occuring time
	int event_type;		// application-specific event type
	Event *ne;	// next event in the chain

 public:
	char e_cnt; // hidden field only used by EventChain.
	Event(int typ=-1, double t=-1) :
		sched_time(t), event_type(typ), ne(NULL), e_cnt(1) {}
	virtual ~Event() {}
	Event * &next() { return ne; }
	int type() { return event_type; }
	double time() { return sched_time; }
	void set_time(double t) { sched_time=t; }
	virtual bool good() { return false; }
	virtual void dump() { tmpi_error(DBG_VERBOSE, "Generic Event @ %f.\n", time()); }
};

# define CAL_QSPACE 393217
# define CAL_MAXNBUCKETS 262144
# define CAL_SAMPLE_SIZE 32

class EventChain {
private:
	Event **calendar;                   /* pointer to use as a sub-array to calq */
	Event **calq;                   /* The real space */
	bool resize_enable;
	double sim_time;
	// double latest_time;
	int qsize;
	int nbuckets;
	double width;
	int lastbucket;
	int firstsub;
	// double lastprio;
	int top_threshold, bot_threshold;
	double buckettop;
	double e_samples[CAL_SAMPLE_SIZE];	// event dequeue time samples
	int e_samples_idx;
	int t_samples[CAL_SAMPLE_SIZE];	// event enqueue traversal number samples
	int t_samples_idx;
	int t_samples_sum;
	Event *laste;
	double lastt;
	double lastdqt;
	long long n_enqueue;
	long long n_dequeue;
	long long n_xdq; // dequeue couner for different time
	long long qsize_op;
	long long n_traverse;
	int resize_cnt;
	// double qsize_integral;
	
	void init();
	void localinit(int qbase, int nbuck, double bwidth, double startprio);
	void resize(int newsize);
	double new_width();
	int insert(Event *e, double t);
	Event *remove();

public:
	EventChain();
	~EventChain();
	bool schedule(Event *e);
	Event *next_event();
	double time() { return sim_time; }
	long long num_enqueue() { return n_enqueue; }
	long long num_dequeue() { return n_dequeue; }
	double get_width() { return width; }
	int get_resize_cnt() { return resize_cnt; }
	double get_num_traverse() { return (double)n_traverse/n_enqueue; }
	double queue_length() { if (n_enqueue>0) return (double)qsize_op/(n_enqueue+n_dequeue); return 0.0;}
	double opt_width() { return sim_time/n_xdq;}
	void statistics();
};

# endif
