// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: gsldistrib.h,v $ - Wrapper for GNU Scientific Library's distribution 
 *            generators.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/07/28 07:20:18 $
 * Version :  $Revision: 1.8 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

/**
 * Export the following functions and classes:
 *
 * double gsldd_<distribution>(parameters);
 *
 * class Distrib {
 *	virtual double get_one() = 0;
 * };
 *
 * class Distrib_<distribution> : Distrib {
 *	Distrib_<distribution>::Distrib_<distribution>(parameters);
 *  virtual double get_one();
 * };
 *
 * <distribution> could be one of:
 *	gaussian
 *	gaussian_tail
 *	exponential
 *	laplace
 *	exppow
 *	cauchy
 *	rayleigh
 *	rayleigh_tail
 *	// landau, (not available in GSL 0.9)
 *	levy
 *	levy_skew
 *	gamma
 *	flat
 *	lognormal
 *	chisq
 *	fdist
 *	tdist
 *	beta
 *	logistic
 *	pareto
 *	weibull
 *	gumbel1
 *	gumbel2
 */ 

#ifndef _GSLDISTRIB_H_
#define _GSLDISTRIB_H_

# include "gsl/gsl_rng.h"
# include "gsl/gsl_randist.h"

/**
 * Distribution for contiguous PDFs
 * GSL Double Distributions
 */

# define GSLDD_FUNC0(name) \
static double gsldd_##name() { \
    gsldd_init(); \
    return gsl_ran_##name(gsldd_r); \
} \
\
class Distrib_##name : Distrib { \
 public: \
	Distrib_##name() {} \
    virtual double get_one() { return GSLDD::gsldd_##name(); } \
    virtual char * get_name() { return #name; } \
} 

# define GSLDD_FUNC1(name, a) \
static double gsldd_##name(double a) { \
    gsldd_init(); \
    return gsl_ran_##name(gsldd_r, a); \
} \
\
class Distrib_##name : Distrib { \
 private: \
    double a; \
 public: \
	Distrib_##name(double a) {this->a=a;} \
    virtual double get_one() { return GSLDD::gsldd_##name(a); } \
    virtual char * get_name() { return #name; } \
} 

# define GSLDD_FUNC2(name, a, b) \
static double gsldd_##name(double a, double b) { \
    gsldd_init(); \
    return gsl_ran_##name(gsldd_r, a, b); \
} \
\
class Distrib_##name : Distrib { \
 private: \
    double a; \
    double b; \
 public: \
	Distrib_##name(double a, double b) {this->a=a; this->b=b;} \
    virtual double get_one() { return GSLDD::gsldd_##name(a, b); } \
    virtual char * get_name() { return #name; } \
} 

# define GSLDD_FUNC3(name, a, b, c) \
static double gsldd_##name(double a, double b, double c) { \
    gsldd_init(); \
    return gsl_ran_##name(gsldd_r, a, b, c); \
} \
\
class Distrib_##name : Distrib { \
 private: \
    double a; \
    double b; \
    double c; \
 public: \
	Distrib_##name(double a, double b, double c) {this->a=a; this->b=b; this->c=c;} \
    virtual double get_one() { return GSLDD::gsldd_##name(a, b, c); } \
    virtual char * get_name() { return #name; } \
} 

namespace GSLDD 
{
	extern gsl_rng *gsldd_r;
	void gsldd_init(void);

	// general distribution class
	class Distrib {
	 public:
		virtual double get_one() = 0;
		virtual char * get_name() = 0;
	};

	GSLDD_FUNC1(gaussian, sigma);
	GSLDD_FUNC2(gaussian_tail, a, sigma);
	GSLDD_FUNC1(exponential, mu);
	GSLDD_FUNC1(laplace, a);
	GSLDD_FUNC2(exppow, a, b);
	GSLDD_FUNC1(cauchy, a);
	GSLDD_FUNC1(rayleigh, sigma);
	GSLDD_FUNC2(rayleigh_tail, a, sigma);
	// not available in GSL 0.9 GSLDD_FUNC0(landau);
	GSLDD_FUNC2(levy, c, alpha);
	GSLDD_FUNC3(levy_skew, c, alpha, beta);
	GSLDD_FUNC2(gamma, a, b);
	GSLDD_FUNC2(flat, a, b);
	GSLDD_FUNC2(lognormal, zeta, sigma);
	GSLDD_FUNC1(chisq, nu);
	GSLDD_FUNC2(fdist, nu1, nu2);
	GSLDD_FUNC1(tdist, nu);
	GSLDD_FUNC2(beta, a, b);
	GSLDD_FUNC1(logistic, a);
	GSLDD_FUNC2(pareto, a, b);
	GSLDD_FUNC2(weibull, a, b);
	GSLDD_FUNC2(gumbel1, a, b);
	GSLDD_FUNC2(gumbel2, a, b);
};

using namespace GSLDD; /** export the namespace */

# endif
