// -*- mode:C; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: progress.c,v $ - Simulation Progress Bar.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2002/10/08 07:32:53 $
 * Version :  $Revision: 1.4 $
 *
 * This module is based on the Download progress module from Wget.
 * Modified by Hong Tang.
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

/**
 * WGET:
 * Copyright (C) 2001 Free Software Foundation, Inc.
 * GNU Wget is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * GNU Wget is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Wget; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

# include <sys/ioctl.h>
# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <assert.h>
# include <unistd.h>
# include <signal.h>
# include <termios.h>

/* # include "utils.h" */
# include "progress.h"

struct progress_implementation {
	char *name;
	void *(*create) (long, long);
	void (*update) (void *, long, long);
	void (*finish) (void *, long);
	void (*set_params) (const char *);
};

//# define KILO 1024
# define KILO 1000

/* Necessary forward declarations. */

static char *legible PARAMS ((long));
/* static char *legible_very_long PARAMS ((long long)); */
static int numdigit PARAMS ((long));
static char *number_to_string PARAMS ((char *, long));
static int determine_screen_width PARAMS ((void));

static void *dot_create PARAMS ((long, long));
static void dot_update PARAMS ((void *, long, long));
static void dot_finish PARAMS ((void *, long));
static void dot_set_params PARAMS ((const char *));

static void *bar_create PARAMS ((long, long));
static void bar_update PARAMS ((void *, long, long));
static void bar_finish PARAMS ((void *, long));
static void bar_set_params PARAMS ((const char *));

static struct progress_implementation implementations[] = {
  { "dot", dot_create, dot_update, dot_finish, dot_set_params },
  { "bar", bar_create, bar_update, bar_finish, bar_set_params },
  { NULL, NULL, NULL, NULL, NULL}
};

static struct progress_implementation *current_impl;
static int current_impl_locked;

/* Progress implementation used by default.  Can be overriden in
   wgetrc or by the fallback one.  */

#define DEFAULT_PROGRESS_IMPLEMENTATION "bar"

/* Fallnback progress implementation should be something that works
   under all display types.  If you put something other than "dot"
   here, remember that bar_set_params tries to switch to this if we're
   not running on a TTY.  So changing this to "bar" could cause
   infloop.  */

#define FALLBACK_PROGRESS_IMPLEMENTATION "dot"

/* Return non-zero if NAME names a valid progress bar implementation.
   The characters after the first : will be ignored.  */

int
valid_progress_implementation_p (const char *name)
{
	int i = 0;
	struct progress_implementation *pi = implementations;
	char *colon = strchr (name, ':');
	int namelen = colon ? colon - name : strlen (name);

	for (i = 0; ; i++, pi++) {
		if (pi->name==NULL)
			break;
		if (!strncmp (pi->name, name, namelen))
			return 1;
	}
	return 0;
}

/* Set the progress implementation to NAME.  */
void
set_progress_implementation (const char *name)
{
	int i, namelen;
	struct progress_implementation *pi = implementations;
	char *colon;

	if (!name)
		name = DEFAULT_PROGRESS_IMPLEMENTATION;

	colon = strchr (name, ':');
	namelen = colon ? colon - name : strlen (name);

	for (i = 0; ; i++, pi++) {
		if (pi->name==NULL)
			break;
		if (!strncmp (pi->name, name, namelen))
			{
				current_impl = pi;
				current_impl_locked = 0;

				if (colon)
					/* We call pi->set_params even if colon is NULL because we
					   want to give the implementation a chance to set up some
					   things it needs to run.  */
					++colon;

				if (pi->set_params)
					pi->set_params (colon);
				return;
			}
	}
	abort ();
}

static int output_redirected;

void
progress_schedule_redirect (void)
{
	output_redirected = 1;
}

/* Create a progress gauge.  INITIAL is the number of bytes the
   download starts from (zero if the download starts from scratch).
   TOTAL is the expected total number of bytes in this download.  If
   TOTAL is zero, it means that the download size is not known in
   advance.  */

void *
progress_create (long initial, long total)
{
	/* Check if the log status has changed under our feet. */
	if (output_redirected)
		{
			if (!current_impl_locked)
				set_progress_implementation (FALLBACK_PROGRESS_IMPLEMENTATION);
			output_redirected = 0;
		}

	return current_impl->create (initial, total);
}

/* Inform the progress gauge of newly received bytes.  DLTIME is the
   time in milliseconds since the beginning of the download.  */

void
progress_update (void *progress, long howmuch, long dltime)
{
	current_impl->update (progress, howmuch, dltime);
}

/* Tell the progress gauge to clean up.  Calling this will free the
   PROGRESS object, the further use of which is not allowed.  */

void
progress_finish (void *progress, long dltime)
{
	current_impl->finish (progress, dltime);
}

/* Dot-printing. */

struct dot_progress {
	long initial_length;		/* how many bytes have been downloaded
								   previously. */
	long total_length;		/* expected total byte count when the
							   download finishes */

	int accumulated;

	int rows;			/* number of rows printed so far */
	int dots;			/* number of dots printed in this row */
	long last_timer_value;
};

static void myfputs(char *s, FILE *dummy)
{
	fputs(s, dummy);
	fflush(dummy);
	// puts(s);
}

/* Dot-progress backend for progress_create. */

static void *
dot_create (long initial, long total)
{
	struct dot_progress *dp = malloc (sizeof (struct dot_progress));

	memset (dp, 0, sizeof (*dp));

	dp->initial_length = initial;
	dp->total_length   = total;

	if (dp->initial_length) 
		{
			int dot_bytes = 5; // 1 dot is 5 simulation seconds
			long row_bytes = dot_bytes * 40; // one line seconds.

			int remainder = (int) (dp->initial_length % row_bytes);
			long skipped = dp->initial_length - remainder;

			if (skipped)
				{
					int skipped_k = (int) (skipped / KILO); /* skipped amount in K */
					int skipped_k_len = numdigit (skipped_k);
					if (skipped_k_len < 5)
						skipped_k_len = 5;

					/* Align the [ skipping ... ] line with the dots.  To do
					   that, insert the number of spaces equal to the number of
					   digits in the skipped amount in K.  */
					printf ("\n%*s[ skipping %dK ]",
							2 + skipped_k_len, "", skipped_k);
				}

			printf ("\n%5ldK", skipped / KILO);
			for (; remainder >= dot_bytes; remainder -= dot_bytes)
				{
					if (dp->dots % 10 == 0)
						myfputs(" ", stdout);
					myfputs (",", stdout);
					++dp->dots;
				}
			assert (dp->dots < 40);

			dp->accumulated = remainder;
			dp->rows = skipped / row_bytes;
		}

	return dp;
}

static void
print_percentage (long bytes, long expected)
{
	int percentage = (int)(100.0 * bytes / expected);
	printf ("%3d%%", percentage);
}

static double
calc_rate (long bytes, long msecs, int *units)
{
	double dlrate;

	if (msecs == 0)
		msecs = 1;

	dlrate = (double)KILO * KILO * bytes / msecs;
	if (dlrate < KILO)
		*units = 0;
	else if (dlrate < KILO * KILO)
		*units = 1, dlrate /= KILO;
	else if (dlrate < KILO * KILO * KILO)
		*units = 2, dlrate /= (KILO * KILO);
	else
		/* Maybe someone will need this one day.  More realistically, it
		 *        will get tickled by buggy timers. */
		*units = 3, dlrate /= (KILO * KILO * KILO);

	return dlrate;
}

static char *
retr_rate (long bytes, long msecs, int pad)
{
	static char res[20];
	static char *rate_names[] = {"ms/s", "s/s", "Ks/s", "Gs/s" };
	int units = 0;

	double dlrate = calc_rate (bytes, msecs, &units);
	sprintf (res, pad ? "%7.2f %s" : "%.2f %s", dlrate, rate_names[units]);

	return res;
}

static void
print_download_speed (struct dot_progress *dp, long bytes, long dltime)
{
	printf (" %s",
			retr_rate (bytes, dltime - dp->last_timer_value, 1));
	dp->last_timer_value = dltime;
}

/* Dot-progress backend for progress_update. */

static void
dot_update (void *progress, long howmuch, long dltime)
{
	struct dot_progress *dp = progress;
	int dot_bytes = 5;
	long row_bytes = dot_bytes * 40;

	dp->accumulated += howmuch;
	for (; dp->accumulated >= dot_bytes; dp->accumulated -= dot_bytes)
		{
			if (dp->dots == 0)
				printf ("\n%5ldK", dp->rows * row_bytes / KILO);

			if (dp->dots % 10 == 0)
				myfputs (" ", stdout);
			myfputs (".", stdout);

			++dp->dots;
			if (dp->dots >= 40)
				{
					long row_qty = row_bytes;
					if (dp->rows == dp->initial_length / row_bytes)
						row_qty -= dp->initial_length % row_bytes;

					++dp->rows;
					dp->dots = 0;

					if (dp->total_length)
						print_percentage (dp->rows * row_bytes, dp->total_length);
					print_download_speed (dp, row_qty, dltime);
				}
		}

	fflush(stdout);
}

/* Dot-progress backend for progress_finish. */

static void
dot_finish (void *progress, long dltime)
{
	struct dot_progress *dp = progress;
	int dot_bytes = 5;
	long row_bytes = dot_bytes * 40;
	int i;

	if (dp->dots == 0)
		printf ("\n%5ldK", dp->rows * row_bytes / KILO);
	for (i = dp->dots; i < 40; i++)
		{
			if (i % 10 == 0)
				myfputs (" ", stdout);
			myfputs (" ", stdout);
		}
	if (dp->total_length)
		{
			print_percentage (dp->rows * row_bytes
							  + dp->dots * dot_bytes
							  + dp->accumulated,
							  dp->total_length);
		}

	{
		long row_qty = dp->dots * dot_bytes + dp->accumulated;
		if (dp->rows == dp->initial_length / row_bytes)
			row_qty -= dp->initial_length % row_bytes;
		print_download_speed (dp, row_qty, dltime);
	}

	myfputs ("\n\n", stdout);
	fflush(stdout);

	free (dp);
}

/* This function interprets the progress "parameters".  For example,
   if Wget is invoked with --progress=dot:mega, it will set the
   "dot-style" to "mega".  Valid styles are default, binary, mega, and
   giga.  */

static void
dot_set_params (const char *params)
{
	return;
}

/* "Thermometer" (bar) progress. */

/* Assumed screen width if we can't find the real value.  */
#define DEFAULT_SCREEN_WIDTH 80

/* Minimum screen width we'll try to work with.  If this is too small,
   create_image will overflow the buffer.  */
#define MINIMUM_SCREEN_WIDTH 45

static int screen_width = DEFAULT_SCREEN_WIDTH;

struct bar_progress {
	long initial_length;		/* how many bytes have been downloaded
								   previously. */
	long total_length;		/* expected total byte count when the
							   download finishes */
	long count;			/* bytes downloaded so far */

	long last_update;		/* time of the last screen update. */

	int width;			/* screen width we're using at the
						   time the progress gauge was
						   created.  this is different from
						   the screen_width global variable in
						   that the latter can be changed by a
						   signal. */
	char *buffer;			/* buffer where the bar "image" is
							   stored. */
	int tick;
};

static void create_image PARAMS ((struct bar_progress *, long));
static void display_image PARAMS ((char *));

static void *
bar_create (long initial, long total)
{
	struct bar_progress *bp = malloc (sizeof (struct bar_progress));

	memset (bp, 0, sizeof (*bp));

	bp->initial_length = initial;
	bp->total_length   = total;

	/* - 1 because we don't want to use the last screen column. */
	bp->width = screen_width - 1;
	/* + 1 for the terminating zero. */
	bp->buffer = malloc (bp->width + 1);

	myfputs ("\n", stdout);

	create_image (bp, 0);
	display_image (bp->buffer);

	return bp;
}

static void
bar_update (void *progress, long howmuch, long dltime)
{
	struct bar_progress *bp = progress;
	int force_update = 0;

	bp->count += howmuch;
	if (bp->total_length > 0
		&& bp->count + bp->initial_length > bp->total_length)
		/* We could be downloading more than total_length, e.g. when the
		   server sends an incorrect Content-Length header.  In that case,
		   adjust bp->total_length to the new reality, so that the code in
		   create_image() that depends on total size being smaller or
		   equal to the expected size doesn't abort.  */
		bp->total_length = bp->count + bp->initial_length;

	if (screen_width - 1 != bp->width)
		{
			bp->width = screen_width - 1;
			bp->buffer = realloc (bp->buffer, bp->width + 1);
			force_update = 1;
		}

// 	if (dltime - bp->last_update < 200 && !force_update)
// 		/* Don't update more often than five times per second. */
// 		return;

	bp->last_update = dltime;

	create_image (bp, dltime);
	display_image (bp->buffer);
}

static void
bar_finish (void *progress, long dltime)
{
	struct bar_progress *bp = progress;

	if (dltime == 0)
		/* If the download was faster than the granularity of the timer,
		   fake some output so that we don't get the ugly "----.--" rate
		   at the download finish.  */
		dltime = 1;

	create_image (bp, dltime);
	display_image (bp->buffer);

	myfputs ("\n\n", stdout);

	free (bp->buffer);
	free (bp);
}

#define APPEND_LITERAL(s) do {			\
  memcpy (p, s, sizeof (s) - 1);		\
  p += sizeof (s) - 1;				\
} while (0)

#ifndef MAX
# define MAX(a, b) ((a) >= (b) ? (a) : (b))
#endif

static void
create_image (struct bar_progress *bp, long dltime)
{
	char *p = bp->buffer;
	long size = bp->initial_length + bp->count;

	char *size_legible = legible (size);
	int size_legible_len = strlen (size_legible);

	/* The progress bar should look like this:
	   xx% [=======>             ] nn,nnn 12.34K/s ETA 00:00

	   Calculate the geometry.  The idea is to assign as much room as
	   possible to the progress bar.  The other idea is to never let
	   things "jitter", i.e. pad elements that vary in size so that
	   their variance does not affect the placement of other elements.
	   It would be especially bad for the progress bar to be resized
	   randomly.

	   "xx% " or "100%"  - percentage               - 4 chars
	   "[]"              - progress bar decorations - 2 chars
	   " nnn,nnn,nnn"    - downloaded bytes         - 12 chars or very rarely more
	   " 1012.56K/s"     - dl rate                  - 11 chars
	   " ETA xx:xx:xx"   - ETA                      - 13 chars

	   "=====>..."       - progress bar             - the rest
	*/
	int dlbytes_size = 1 + MAX (size_legible_len, 11);
	int progress_size = bp->width - (4 + 2 + dlbytes_size + 11 + 13);

	if (progress_size < 5)
		progress_size = 0;

	/* "xx% " */
	if (bp->total_length > 0)
		{
			int percentage = (int)(100.0 * size / bp->total_length);

			assert (percentage <= 100);

			if (percentage < 100)
				sprintf (p, "%2d%% ", percentage);
			else
				strcpy (p, "100%");
			p += 4;
		}
	else
		APPEND_LITERAL ("    ");

	/* The progress bar: "[====>      ]" */
	if (progress_size && bp->total_length > 0)
		{
			double fraction = (double)size / bp->total_length;
			int dlsz = (int)(fraction * progress_size);
			char *begin;

			assert (dlsz <= progress_size);

			*p++ = '[';
			begin = p;

			if (dlsz > 0)
				{
					/* Draw dlsz-1 '=' chars and one arrow char.  */
					while (dlsz-- > 1)
						*p++ = '=';
					*p++ = '>';
				}

			while (p - begin < progress_size)
				*p++ = ' ';

			*p++ = ']';
		}
	else if (progress_size)
		{
			/* If we can't draw a real progress bar, then at least show
			 *something* to the user.  */
			int ind = bp->tick % (progress_size * 2 - 6);
			int i, pos;

			/* Make the star move in two directions. */
			if (ind < progress_size - 2)
				pos = ind + 1;
			else
				pos = progress_size - (ind - progress_size + 5);

			*p++ = '[';
			for (i = 0; i < progress_size; i++)
				{
					if      (i == pos - 1) *p++ = '<';
					else if (i == pos    ) *p++ = '=';
					else if (i == pos + 1) *p++ = '>';
					else
						*p++ = ' ';
				}
			*p++ = ']';

			++bp->tick;
		}

	/* " 234,567,890" */
	sprintf (p, " %-11s", legible (size));
	p += strlen (p);

	/* " 1012.45K/s" */
	if (dltime && bp->count)
		{
			static char *short_units[] = { "m/s", "s/s", "K/s", "M/s" };
			int units = 0;
			double dlrate = calc_rate (bp->count, dltime, &units);
			sprintf (p, " %7.2f%s", dlrate, short_units[units]);
			p += strlen (p);
		}
	else
		APPEND_LITERAL ("   --.--K/s");

	/* " ETA xx:xx:xx" */
	if (bp->total_length > 0 && bp->count > 0)
		{
			int eta, eta_hrs, eta_min, eta_sec;
			double tm_sofar = (double)dltime / 1000;
			long bytes_remaining = bp->total_length - size;

			eta = (int) (((tm_sofar * bytes_remaining / bp->count)+(KILO/2))/KILO);

			eta_hrs = eta / 3600, eta %= 3600;
			eta_min = eta / 60,   eta %= 60;
			eta_sec = eta;

			/* Pad until the end of screen.  The padding is dependent on the
			   hour value.  */
			if (eta_hrs == 0 || eta_hrs > 99)
				/* Hours not printed: pad with three spaces (two digits and
				   colon). */
				APPEND_LITERAL ("   ");
			else if (eta_hrs < 10)
				/* Hours printed with one digit: pad with one space. */
				*p++ = ' ';
			else
				/* Hours printed with two digits: we're using maximum width,
				   don't pad. */
				;

			APPEND_LITERAL (" ETA ");

			if (eta_hrs > 99)
				/* Bogus value, probably due to a calculation overflow.  Print
				   something safe to avoid overstepping the buffer bounds. */
				sprintf (p, "--:--");
			else if (eta_hrs > 0)
				sprintf (p, "%d:%02d:%02d", eta_hrs, eta_min, eta_sec);
			else
				sprintf (p, "%02d:%02d", eta_min, eta_sec);
			p += strlen (p);
		}
	else if (bp->total_length > 0)
		APPEND_LITERAL ("    ETA --:--");

	assert (p - bp->buffer <= bp->width);

	while (p < bp->buffer + bp->width)
		*p++ = ' ';
	*p = '\0';
}

/* Print the contents of the buffer as a one-line ASCII "image" so
   that it can be overwritten next time.  */

static void
display_image (char *buf)
{
	myfputs("\r", stdout);
	myfputs(buf, stdout);
}

static void progress_handle_sigwinch (int sig)
{
	int sw = determine_screen_width ();
	if (sw && sw >= MINIMUM_SCREEN_WIDTH)
		screen_width = sw;
	signal (SIGWINCH, progress_handle_sigwinch);
}

static void
bar_set_params (const char *params)
{
	int sw;

	if (params
		&& 0 == strcmp (params, "force"))
		current_impl_locked = 1;

	if ( !isatty (fileno (stdout))
		 && !current_impl_locked)
		{
			/* We're not printing to a TTY, so revert to the fallback
			   display.  #### We're recursively calling
			   set_progress_implementation here, which is slightly kludgy.
			   It would be nicer if we provided that function a return value
			   indicating a failure of some sort.  */
			set_progress_implementation (FALLBACK_PROGRESS_IMPLEMENTATION);
			return;
		}

	sw = determine_screen_width();
	if (sw && sw >= MINIMUM_SCREEN_WIDTH)
		screen_width = sw;

	signal(SIGWINCH, progress_handle_sigwinch);
}

/* Engine for legible and legible_very_long; this function works on
   strings.  */

static char *
legible_1 (const char *repr)
{
  static char outbuf[128];
  int i, i1, mod;
  char *outptr;
  const char *inptr;

  /* Reset the pointers.  */
  outptr = outbuf;
  inptr = repr;
  /* If the number is negative, shift the pointers.  */
  if (*inptr == '-')
    {
      *outptr++ = '-';
      ++inptr;
    }
  /* How many digits before the first separator?  */
  mod = strlen (inptr) % 3;
  /* Insert them.  */
  for (i = 0; i < mod; i++)
    *outptr++ = inptr[i];
  /* Now insert the rest of them, putting separator before every
     third digit.  */
  for (i1 = i, i = 0; inptr[i1]; i++, i1++)
    {
      if (i % 3 == 0 && i1 != 0)
	*outptr++ = ',';
      *outptr++ = inptr[i1];
    }
  /* Zero-terminate the string.  */
  *outptr = '\0';
  return outbuf;
}

/* Legible -- return a static pointer to the legibly printed long.  */
static char *
legible (long l)
{
  char inbuf[24];
  /* Print the number into the buffer.  */
  number_to_string (inbuf, l);
  return legible_1 (inbuf);
}

/* Write a string representation of NUMBER into the provided buffer.
   We cannot use sprintf() because we cannot be sure whether the
   platform supports printing of what we chose for VERY_LONG_TYPE.

   Example: Gcc supports `long long' under many platforms, but on many
   of those the native libc knows nothing of it and therefore cannot
   print it.

   How long BUFFER needs to be depends on the platform and the content
   of NUMBER.  For 64-bit VERY_LONG_TYPE (the most common case), 24
   bytes are sufficient.  Using more might be a good idea.

   This function does not go through the hoops that long_to_string
   goes to because it doesn't aspire to be fast.  (It's called perhaps
   once in a Wget run.)  */

# if (0)
static void
very_long_to_string (char *buffer, long long number)
{
  int i = 0;
  int j;

  /* Print the number backwards... */
  do
    {
      buffer[i++] = '0' + number % 10;
      number /= 10;
    }
  while (number);

  /* ...and reverse the order of the digits. */
  for (j = 0; j < i / 2; j++)
    {
      char c = buffer[j];
      buffer[j] = buffer[i - 1 - j];
      buffer[i - 1 - j] = c;
    }
  buffer[i] = '\0';
}
# endif

/* The same as legible(), but works on VERY_LONG_TYPE.  See sysdep.h.  */
# if (0)
static char *
legible_very_long (long long l)
{
  char inbuf[128];
  /* Print the number into the buffer.  */
  very_long_to_string (inbuf, l);
  return legible_1 (inbuf);
}
# endif

/* Count the digits in a (long) integer.  */
static int
numdigit (long number)
{
  int cnt = 1;
  if (number < 0)
    {
      number = -number;
      ++cnt;
    }
  while ((number /= 10) > 0)
    ++cnt;
  return cnt;
}

#define ONE_DIGIT(figure) *p++ = n / (figure) + '0'
#define ONE_DIGIT_ADVANCE(figure) (ONE_DIGIT (figure), n %= (figure))

#define DIGITS_1(figure) ONE_DIGIT (figure)
#define DIGITS_2(figure) ONE_DIGIT_ADVANCE (figure); DIGITS_1 ((figure) / 10)
#define DIGITS_3(figure) ONE_DIGIT_ADVANCE (figure); DIGITS_2 ((figure) / 10)
#define DIGITS_4(figure) ONE_DIGIT_ADVANCE (figure); DIGITS_3 ((figure) / 10)
#define DIGITS_5(figure) ONE_DIGIT_ADVANCE (figure); DIGITS_4 ((figure) / 10)
#define DIGITS_6(figure) ONE_DIGIT_ADVANCE (figure); DIGITS_5 ((figure) / 10)
#define DIGITS_7(figure) ONE_DIGIT_ADVANCE (figure); DIGITS_6 ((figure) / 10)
#define DIGITS_8(figure) ONE_DIGIT_ADVANCE (figure); DIGITS_7 ((figure) / 10)
#define DIGITS_9(figure) ONE_DIGIT_ADVANCE (figure); DIGITS_8 ((figure) / 10)
#define DIGITS_10(figure) ONE_DIGIT_ADVANCE (figure); DIGITS_9 ((figure) / 10)

/* DIGITS_<11-20> are only used on machines with 64-bit longs. */

#define DIGITS_11(figure) ONE_DIGIT_ADVANCE (figure); DIGITS_10 ((figure) / 10)
#define DIGITS_12(figure) ONE_DIGIT_ADVANCE (figure); DIGITS_11 ((figure) / 10)
#define DIGITS_13(figure) ONE_DIGIT_ADVANCE (figure); DIGITS_12 ((figure) / 10)
#define DIGITS_14(figure) ONE_DIGIT_ADVANCE (figure); DIGITS_13 ((figure) / 10)
#define DIGITS_15(figure) ONE_DIGIT_ADVANCE (figure); DIGITS_14 ((figure) / 10)
#define DIGITS_16(figure) ONE_DIGIT_ADVANCE (figure); DIGITS_15 ((figure) / 10)
#define DIGITS_17(figure) ONE_DIGIT_ADVANCE (figure); DIGITS_16 ((figure) / 10)
#define DIGITS_18(figure) ONE_DIGIT_ADVANCE (figure); DIGITS_17 ((figure) / 10)
#define DIGITS_19(figure) ONE_DIGIT_ADVANCE (figure); DIGITS_18 ((figure) / 10)

/* Print NUMBER to BUFFER in base 10.  This should be completely
   equivalent to `sprintf(buffer, "%ld", number)', only much faster.

   The speedup may make a difference in programs that frequently
   convert numbers to strings.  Some implementations of sprintf,
   particularly the one in GNU libc, have been known to be extremely
   slow compared to this function.

   Return the pointer to the location where the terminating zero was
   printed.  (Equivalent to calling buffer+strlen(buffer) after the
   function is done.)

   BUFFER should be big enough to accept as many bytes as you expect
   the number to take up.  On machines with 64-bit longs the maximum
   needed size is 24 bytes.  That includes the digits needed for the
   largest 64-bit number, the `-' sign in case it's negative, and the
   terminating '\0'.  */

static char *
number_to_string (char *buffer, long number)
{
  char *p = buffer;
  long n = number;

  if (n < 0)
    {
      *p++ = '-';
      n = -n;
    }

  if      (n < 10)                   { DIGITS_1 (1); }
  else if (n < 100)                  { DIGITS_2 (10); }
  else if (n < 1000)                 { DIGITS_3 (100); }
  else if (n < 10000)                { DIGITS_4 (1000); }
  else if (n < 100000)               { DIGITS_5 (10000); }
  else if (n < 1000000)              { DIGITS_6 (100000); }
  else if (n < 10000000)             { DIGITS_7 (1000000); }
  else if (n < 100000000)            { DIGITS_8 (10000000); }
  else if (n < 1000000000)           { DIGITS_9 (100000000); }
  /* ``if (1)'' serves only to preserve editor indentation. */
  else if (1)                        { DIGITS_10 (1000000000); }

  *p = '\0';

  return p;
}

#undef ONE_DIGIT
#undef ONE_DIGIT_ADVANCE

#undef DIGITS_1
#undef DIGITS_2
#undef DIGITS_3
#undef DIGITS_4
#undef DIGITS_5
#undef DIGITS_6
#undef DIGITS_7
#undef DIGITS_8
#undef DIGITS_9
#undef DIGITS_10
#undef DIGITS_11
#undef DIGITS_12
#undef DIGITS_13
#undef DIGITS_14
#undef DIGITS_15
#undef DIGITS_16
#undef DIGITS_17
#undef DIGITS_18
#undef DIGITS_19

static int
determine_screen_width (void)
{
	/* If there's a way to get the terminal size using 
	 * POSIX tcgetattr(), somebody please tell me.  */
	/* TIOCGWINSZ */
	int fd;
	struct winsize wsz;

	fd = fileno (stderr);
	if (ioctl (fd, TIOCGWINSZ, &wsz) < 0)
		return 70;  /* most likely ENOTTY */

	return wsz.ws_col;
}
