// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: async_conn.cc,v $ - asynchronous TCP connection.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2002/10/08 07:32:53 $
 * Version :  $Revision: 1.2 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>

# define MAX_RETRY 10

/** @name   Function */
//@{
/** 
 * TCP asynchronous connection management.
 *  @param   sockfd - client socket
 *  @param   serv_addr - server address
 *  @param   addrlen - server address length
 *  @param   tout - timeout in ms (<0: blocked, =0: unblocked, >0: timeout)
 *  @return  0: success; -1: error. Actual error is indicated by errno.
 *  
 *  For errors not known, we set errno=ENETUNREACH.
 *  We set errno=EINTR when the routine fails too many times due to EINTR.
 */
//@}
extern "C" int async_connect(int sockfd, const struct sockaddr *serv_addr, socklen_t addrlen, int tout)
{
	if (tout<0) // blocked connect
		return connect(sockfd, serv_addr, addrlen);

	int flags;
	int n;
	flags = fcntl(sockfd, F_GETFL, 0);
	fcntl(sockfd, F_SETFL, flags | O_NONBLOCK);
	if ( (n=connect(sockfd, (struct sockaddr *) serv_addr, addrlen)) < 0)
	{
		// check whether the socket is already connected?
		if ((errno == EISCONN)) {
			fcntl(sockfd, F_SETFL, flags);
			return 0;
		}

		// check if we indeed initiated a SYN request
		if ((errno != EINPROGRESS) && (errno != EALREADY))
		{
			fcntl(sockfd, F_SETFL, flags);
			return -1;
		}
	}
	else {
		fcntl(sockfd, F_SETFL, flags);
		return 0;
	}

	struct timeval tv;
	int retry=0;
	tv.tv_sec=tout/1000;
	tv.tv_usec=(tout%1000)*1000;

	fd_set rset, wset;
	FD_ZERO(&rset);
	FD_ZERO(&wset);

	while (1) {
		/**
		 * blocked receive w/ timeout
		 */

		int n;
		int error = 0;
		socklen_t len;

		FD_SET(sockfd, &rset);
		FD_SET(sockfd, &wset);

		n=select(sockfd+1, &rset, &wset, NULL, &tv);
		if (n==0) { // timeout
			errno=ETIMEDOUT;
			fcntl(sockfd, F_SETFL, flags);
			return -1;
		}

		if (n<0) { // error
			if ((errno!=EINTR)||(retry>=MAX_RETRY)) { 
				// system error
				fcntl(sockfd, F_SETFL, flags);
				return -1;
			}

			// EINTR, try again please
			retry++;
			continue;
		}
			
		// n>0 might be good.
		/**
		 * a socket is connected when it is readable or 
		 * writeable, and there is no error
		 */

		if ((FD_ISSET(sockfd, &rset)) || (FD_ISSET(sockfd, &wset)))
		{
			len = sizeof(error);
			if (getsockopt(sockfd, SOL_SOCKET, SO_ERROR, &error, &len)<0) {
				fcntl(sockfd, F_SETFL, flags);
				errno=ENETUNREACH;
				return -1;
			}

			if (error) {
				fcntl(sockfd, F_SETFL, flags);
				errno=error;
				return -1;
			}

			fcntl(sockfd, F_SETFL, flags);
			return 0;
		} 
		else
		{
			fcntl(sockfd, F_SETFL, flags);
			errno=ENETUNREACH;
			return -1;
		}
	}
}
