// -*- mode:C; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: sock_conn.c,v $ - Socket library wrapper, TCP connection and
 *            name/addr management.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/09/30 01:14:12 $
 * Version :  $Revision: 1.11 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

# include <sys/types.h>
# include <sys/socket.h>
# include <netinet/in.h>
# include <arpa/inet.h>
# include <unistd.h>
# include <string.h>
# include "sock_conn.h"

# define TCP_NODELAY 1

/**
 * Get a list of hostent structures for a given host name
 */
struct hostent *sock_gethostbyname(const char *host)
{
	const char *xhost;
	char lhost[1024];

	if ((!host)||(host[0]=='\0'))
		return NULL;

	if (strcmp(host, ".") == 0) {
		strncpy(lhost, sock_hostname(), 1023);
		lhost[1023]='\0';
		xhost=lhost;
	}
	else {
		xhost=host;
	}
		
	return gethostbyname(xhost);
}

char  *sock_prime_ip()
{
	struct hostent *hp=sock_gethostbyname(".");
	if (!hp)
		return NULL;

	if (hp->h_addr_list[0]) {
		return inet_ntoa(*(struct in_addr *)(hp->h_addr_list[0]));
	}
	else {
		return NULL;
	}
}

/**
 * Connect to a specific server at specific port
 */
int sock_conn(const char *host, int port)
{
	struct hostent *hp;
	struct sockaddr_in serv;
	const int max_retry=10; /* is 10 retrial enough? */
	int connected=0;
	int fd=-1;
	int n_addr;
	int i;
	/* struct linger ll; */

	if (port<=0)
		return -1;

	if (port==SOCK_PORT_ANY)
		return -1;

	hp=sock_gethostbyname(host);
	if (!hp)
		return -1;

	n_addr=0;
	while (hp->h_addr_list[n_addr])
		n_addr++;

	memset((char *)&serv, 0, sizeof(serv));
	serv.sin_family=hp->h_addrtype;
	serv.sin_port=htons(port);

	for (i=0; i<n_addr; i++) {
		int retry=max_retry;

		memcpy((char *)&(serv.sin_addr), (char *)(hp->h_addr_list[i]), hp->h_length);

		while (retry--) {
			int flag=-1;

			fd=socket(AF_INET, SOCK_STREAM, 0);
			setsockopt(fd, IPPROTO_TCP, TCP_NODELAY, (void *)&flag, sizeof(int));
			sock_setbufsize(fd, SOCK_SNDBUF_SZ, SOCK_RCVBUF_SZ); /* 100KB buffer size */

			/**
			ll.l_onoff=1;
			ll.l_linger=0;

			setsockopt(fd, SOL_SOCKET, SO_LINGER, &ll, sizeof(ll));
			**/

			if (fd<0) {
				sleep(1);
				continue;
			}

			if (connect(fd, (struct sockaddr *)&serv, sizeof(serv))<0) {
				close(fd);
				fd=-1;
				sleep(1);
				continue;
			}
			else {
				connected=1;
				break;
			}
		}

		if (connected)
			break;
	}

	if (!connected) {
		if (fd>=0) {
			close(fd);
			fd=-1;
		}
		return -1;
	}

	return fd;
}

/* defined async_conn.cc */
int async_connect(int sockfd, const struct sockaddr *serv_addr, socklen_t addrlen, int tout);

int sock_async_conn(const char *host, int port, int tout)
{
	struct hostent *hp;
	struct sockaddr_in serv;
	int connected=0;
	int fd=-1;
	int n_addr;
	int i;

	if (port<=0)
		return -1;

	if (port==SOCK_PORT_ANY)
		return -1;

	hp=sock_gethostbyname(host);
	if (!hp) {
		return -1;
	}

	n_addr=0;
	while (hp->h_addr_list[n_addr])
		n_addr++;

	memset((char *)&serv, 0, sizeof(serv));
	serv.sin_family=hp->h_addrtype;
	serv.sin_port=htons(port);

	for (i=0; i<n_addr; i++) {
		int flag=-1;
		memcpy((char *)&(serv.sin_addr), (char *)(hp->h_addr_list[i]), hp->h_length);

		fd=socket(AF_INET, SOCK_STREAM, 0);
		if (fd<0) {
			continue;
		}

		setsockopt(fd, IPPROTO_TCP, TCP_NODELAY, (void *)&flag, sizeof(int));
		sock_setbufsize(fd, SOCK_SNDBUF_SZ, SOCK_RCVBUF_SZ); /* 100KB buffer size */

		if (async_connect(fd, (struct sockaddr *)&serv, sizeof(serv), tout)<0) {
			close(fd);
			fd=-1;
			continue;
		}
		else {
			connected=1;
			break;
		}
	}

	if (!connected) {
		if (fd>=0) {
			close(fd);
			fd=-1;
		}
		return -1;
	}

	return fd;
}

/**
 * connect to the server with the client binding to a
 * specific address.
 */
int sock_conn2(const char *host, int port, const char *bind_host)
{
	struct hostent *hp;
	struct sockaddr_in serv;
	const int max_retry=10; /* is 5 retrial enough? */
	int connected=0;
	int fd=-1;
	int n_addr;
	int i;
	struct sockaddr_in sin;
	/* struct linger ll; */

	if (port<=0)
		return -1;

	if (port==SOCK_PORT_ANY)
		return -1;

	if (sock_filladdr(bind_host, SOCK_PORT_ANY, &sin)<0) {
		return -1;
	}

	hp=sock_gethostbyname(host);
	if (!hp)
		return -1;

	n_addr=0;
	while (hp->h_addr_list[n_addr])
		n_addr++;

	memset((char *)&serv, 0, sizeof(serv));
	serv.sin_family=hp->h_addrtype;
	serv.sin_port=htons(port);

	for (i=0; i<n_addr; i++) {
		int retry=max_retry;

		memcpy((char *)&(serv.sin_addr), (char *)(hp->h_addr_list[i]), hp->h_length);

		while (retry--) {
			int flag=-1;

			fd=socket(AF_INET, SOCK_STREAM, 0);

			if (fd<0) {
				sleep(1);
				continue;
			}

			if (bind(fd, (struct sockaddr *)&sin, sizeof(sin))<0) {
				if (fd>=0) {
					close(fd);
					fd=-1;
				}
				return -1;
			}

			setsockopt(fd, IPPROTO_TCP, TCP_NODELAY, (void *)&flag, sizeof(int));
			sock_setbufsize(fd, SOCK_SNDBUF_SZ, SOCK_RCVBUF_SZ); /* 100KB buffer size */

			/**
			ll.l_onoff=1;
			ll.l_linger=0;

			setsockopt(fd, SOL_SOCKET, SO_LINGER, &ll, sizeof(ll));
			**/

			if (connect(fd, (struct sockaddr *)&serv, sizeof(serv))<0) {
				close(fd);
				fd=-1;
				sleep(1);
				continue;
			}
			else {
				connected=1;
				break;
			}
		}

		if (connected)
			break;
	}

	if (!connected) {
		if (fd>=0) {
			close(fd);
			fd=-1;
		}
		return -1;
	}

	return fd;
}

/**
 * Set the server to listen to a specific port.
 */
int sock_listen(int port, int backlog)
{
	int fd=-1;
	struct sockaddr_in sin;
	int flag=-1;
	/** struct linger ll; */

	fd=socket(AF_INET, SOCK_STREAM, 0);
	if (fd==-1)
		return -1;

	setsockopt(fd, IPPROTO_TCP, TCP_NODELAY, (void *)&flag, sizeof(int));
	sock_setbufsize(fd, SOCK_SNDBUF_SZ, SOCK_RCVBUF_SZ); /* 100KB buffer size */

	/**
	ll.l_onoff=1;
	ll.l_linger=0;

	setsockopt(fd, SOL_SOCKET, SO_LINGER, &ll, sizeof(ll));
	**/

	if (sock_filladdr(NULL, port, &sin)<0) {
		return -1;
	}

	if (bind(fd, (struct sockaddr *)&sin, sizeof(sin))<0) {
		close(fd);
		return -1;
	}

	if (listen(fd, backlog)<0) {
		close(fd);
		return -1;
	}

	return fd;
}

/**
 * Set the server to listen to a specific port with the address
 * bound to a specific host.
 */
int sock_listen2(int port, int backlog, const char *bind_host)
{
	int fd=-1;
	struct sockaddr_in sin;
	int flag=-1;
	/* struct linger ll; */

	if (sock_filladdr(bind_host, port, &sin)<0) {
		return -1;
	}

	fd=socket(AF_INET, SOCK_STREAM, 0);

	if (fd==-1)
		return -1;

	setsockopt(fd, IPPROTO_TCP, TCP_NODELAY, (void *)&flag, sizeof(int));
	sock_setbufsize(fd, SOCK_SNDBUF_SZ, SOCK_RCVBUF_SZ); /* 100KB buffer size */

	/**
	ll.l_onoff=1;
	ll.l_linger=0;

	setsockopt(fd, SOL_SOCKET, SO_LINGER, &ll, sizeof(ll));
	**/

	sin.sin_family = AF_INET;

	if (bind(fd, (struct sockaddr *)&sin, sizeof(sin))<0) {
		close(fd);
		return -1;
	}

	if (listen(fd, backlog)<0) {
		close(fd);
		return -1;
	}

	return fd;
}

int sock_accept(int fd)
{
	struct sockaddr_in from;
	int sinlen=sizeof(from);
	int newfd;

	newfd=accept(fd, (struct sockaddr *)&from, &sinlen);

	if (newfd<0)
		return -1;

	return newfd;
}

char *sock_hostname()
{
	static char hostname[1024];

	if (gethostname(hostname, 1024)<0)
		return NULL;

	return hostname;
}

int sock_myport(int fd)
{
	struct sockaddr_in sin;
	int sinlen=sizeof(sin);

	if (getsockname(fd, (struct sockaddr *)&sin, &sinlen)<0)
		return -1;

	return ntohs(sin.sin_port);
}

int sock_peerport(int fd)
{
	struct sockaddr_in sin;
	int sinlen=sizeof(sin);

	if (getpeername(fd, (struct sockaddr *)&sin, &sinlen)<0)
		return -1;

	return ntohs(sin.sin_port);
}

char *sock_myip(int fd)
{
	struct sockaddr_in sin;
	int sinlen=sizeof(sin);

	if (getsockname(fd, (struct sockaddr *)&sin, &sinlen)<0)
		return NULL;

	return inet_ntoa(sin.sin_addr);
}

char *sock_peerip(int fd)
{
	struct sockaddr_in sin;
	int sinlen=sizeof(sin);

	if (getpeername(fd, (struct sockaddr *)&sin, &sinlen)<0)
		return NULL;

	return inet_ntoa(sin.sin_addr);
}

char *sock_peerhost(int fd)
{
	struct sockaddr_in sin;
	int sinlen=sizeof(sin);
	struct hostent *hent;

	if (getpeername(fd, (struct sockaddr *)&sin, &sinlen)<0)
		return NULL;

	hent=gethostbyaddr((char *)&(sin.sin_addr), sizeof(sin.sin_addr), AF_INET);
	if (!hent)
		return inet_ntoa(sin.sin_addr);

	return hent->h_name;
}

/**
 * show binary address into string IP format.
 */
int sock_dotaddr(char *szAddr, int len, const struct in_addr *addr)
{
	char *cp=inet_ntoa(*addr);

	if (len<=strlen(cp))
		return -1;

	strcpy(szAddr, cp);

	return 0;
}

int sock_filladdr(const char *host, int port, struct sockaddr_in *addr)
{
	addr->sin_family=AF_INET;

	if (!host||(host[0]=='\0')) {
		addr->sin_addr.s_addr=INADDR_ANY;
	}
	else {
		struct hostent *hp;
		hp=sock_gethostbyname(host);

		if (!hp) {
			return -1;
		}

		memcpy(&(addr->sin_addr), (hp->h_addr_list[0]), hp->h_length);
	}

	if ((port==SOCK_PORT_ANY)||(port<0)) {
		addr->sin_port=htons(0);
	}
	else {
		addr->sin_port=htons(port);
	}

	return 0;
}

int sock_setbufsize(int s, int snd, int rcv)
{
	int ret=0;
	if (snd>0) {
		int buf_size=snd;
		socklen_t len=sizeof(int);
		if (setsockopt(s, SOL_SOCKET, SO_SNDBUF, &buf_size, len)<0) {
			ret++;
		}
	}
	else {
		ret++;
	}

	if (rcv>0) {
		int buf_size=rcv;
		socklen_t len=sizeof(int);
		if (setsockopt(s, SOL_SOCKET, SO_RCVBUF, &buf_size, len)<0) {
			ret++;
		}
	}
	else {
		ret++;
	}

	return (ret)? -1:0;
}
