// -*- mode:C; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: sock_dg.c,v $ - Socket library wrapper, datagram/multicast
 *            socket creation and send/mcast/receive.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/09/29 00:15:14 $
 * Version :  $Revision: 1.17 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

# include <sys/types.h>
# include <sys/time.h>
# include <sys/ioctl.h>
# include <sys/socket.h>
# include <net/if.h>
# include <netinet/in.h>
# include <arpa/inet.h>
# include <netdb.h>
# include <unistd.h>
# include <string.h>
# include <errno.h>
# include "sock_conn.h"
# include "sock_dg.h"
# include "sock_sos.h"

/**
 * Create a datagram socket and bind to the specified address.
 * host can be NULL, port=-1 means give me a random port.
 */
int sock_udp(const char *host, int port)
{
	int fd=-1;
	struct sockaddr_in sin;
	
	fd=socket(AF_INET, SOCK_DGRAM, 0);
	if (fd<0) {
		return -1;
	}

	/* no need to do bind? */
	/**
	 * The following is commented out because otherwise
	 * a subsequent getsockname on an un-binded socket will
	 * lead to errors.
	 */
	/**
	if ((!host)&&((port<0)||(port==SOCK_PORT_ANY))) {
		return fd; 
	}
	**/

	if (sock_filladdr(host, port, &sin)<0) {
		close(fd);
		return -1;
	}
	
	if (bind(fd, (struct sockaddr *)&sin, sizeof(sin))<0) {
		close(fd);
		return -1;
	}

	sock_setbufsize(fd, SOCK_SNDBUF_SZ, SOCK_RCVBUF_SZ); /* 100KB buffer size */

	return fd;
}

int sock_is_mcast_addr(const struct sockaddr_in *sa, int salen)
{
	int oct1;

	if (sa->sin_family!=AF_INET) {
		/** only handles IPv4 currently. **/
		return -1;
	}

	oct1=OCTET_1(sa->sin_addr.s_addr);
	if ((oct1>=224)&&(oct1<240))
		return 1;

	return 0;
}

/**
 * Create a multicast socket.
 * mgrp must be a mcast address, port is necessary. 
 * port reuse option is always specified. bind_host
 * is added for filtering purposes. If no filtering
 * is needed, choose NULL. Or else, one can choose
 * to bind to a local IP address (that will allow you
 * to receive packets sent to both the specific IP or
 * mcast group). The most restrictive case would be 
 * to set bind_host the same as mgrp so that it only
 * receives packets destined to the multicast group. 
 * in_if and out_if are not required and might only
 * be needed for multi-homed node.
 *
 * no_loop!=0 will disable local loopback, disable it
 * only when there is no processes the local node in 
 * the same mcast group.
 */
static int sock_join_(int fd, const struct in_addr *sin_mcast, 
		int salen, const char *in_if, int no_loop);

int sock_mcast(const char *mgrp, int port, const char *bind_host, 
		const char *in_if, const char *out_if, int no_loop)
{
	int fd=-1;
	struct sockaddr_in sin_mcast, sin_bind;
	int on=1;
	struct in_addr inaddr;

	// if (!mgrp) {
		// return -1; 
	// }

	if (mgrp) {
		/**
		 * Check the validity of the multicast address
		 */
		if (sock_filladdr(mgrp, port, &sin_mcast)<0) {
			return -1;
		}

		if (sock_is_mcast_addr(&sin_mcast, sizeof(sin_mcast))!=1) {
			return -1;
		}
	}
	else {
		// the socket being created will not be in any multicast group
		// initially.
	}

	/**
	 * make the binding address
	 */
	if (sock_filladdr(bind_host, port, &sin_bind)<0) {
		return -1;
	}

	/**
	 * create a UDP socket
	 */
	fd=socket(AF_INET, SOCK_DGRAM, 0);

	if (fd<0) {
		return -1;
	}

	if (mgrp) {
		/**
		 * set Reuse-port/address option. Some OS might not support
		 * port reuse, in that case, we try to set address reuse.
		 */
		if (
# if defined(SO_REUSEPORT)
				(setsockopt(fd, SOL_SOCKET, SO_REUSEPORT, &on, sizeof(on))<0) &&
# endif
			(setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on))<0) 	) {
			goto _exit;
		}

		sock_setbufsize(fd, SOCK_SNDBUF_SZ, SOCK_RCVBUF_SZ); /* 100KB buffer size */
	}

	/**
	 * bind the socket.
	 */
	if (bind(fd, (struct sockaddr *)&sin_bind, sizeof(sin_bind))<0) {
		goto _exit;
	}

	if (mgrp) {
		/**
		 * join the multicast group and set input physical interface
		 */
		if (sock_join_(fd, &sin_mcast.sin_addr, 
					sizeof(struct in_addr), in_if, no_loop)<0) {
			goto _exit;
		}
	}
	else {
		/**
		 * Set TTL to the default value
		 */
		unsigned char ttl=SOCK_MCAST_TTL;
		if (setsockopt(fd, IPPROTO_IP, IP_MULTICAST_TTL, &ttl, sizeof(ttl))<0) {
			goto _exit;
		}
	}

	/**
	 * Set the default output physical interface
	 */
	if (out_if != NULL) {
		if (sock_if2addr(fd, out_if, &inaddr)<0)
			goto _exit;
	} else
		inaddr.s_addr = htonl(INADDR_ANY);	/* remove prev. set default */

	if (setsockopt(fd, IPPROTO_IP, IP_MULTICAST_IF, 
				&inaddr, sizeof(struct in_addr))<0) {
		goto _exit;
	}
	
	return fd;

_exit:
	if (fd>=0) {
		close(fd);
	}

	return -1;
}

/**
 * Additional multicast memberships.
 * This is only allowed on sockets created by sock_mcast.
 * Unfortunately there is no API allows us to check the multicast
 * groups a socket has joined already. If one attempts to join the same
 * group with the same in_if, it will fail. We do check that either
 * SO_REUSEPORT or SO_REUSEADDR is set to 1. 
 */

int sock_join(int fd, const char *mcaddr, const char *in_if, int no_loop)
{
	struct hostent *hp;
	struct in_addr sin_mcast;

	if (!mcaddr)
		return -1;

	hp=sock_gethostbyname(mcaddr);
	if (!hp) {
		return -1;
	}
	memcpy(&sin_mcast, (hp->h_addr_list[0]), hp->h_length);

	return sock_join_(fd, &sin_mcast, sizeof(struct in_addr), in_if, no_loop);
}

/**
 * Attempts to first translate if as a hostname or IP address,
 * if not possible, interpret it as an interface name.
 * if no fd available, we will attempt to open one just
 * for this call.
 */
int sock_if2addr(int fd, const char *ifn, struct in_addr *addr)
{
	struct hostent *hp;
	struct ifreq ifreq;
	int closefd=0;
	int ret=0;

	if ((!ifn)||(!addr))
		return -1;

	if (fd<0) {
		fd=socket(AF_INET, SOCK_DGRAM, 0);
		if (fd<0)
			return -1;
		closefd=1;
	}

	hp=sock_gethostbyname(ifn);
	if (hp) {
		memcpy(addr, (hp->h_addr_list[0]), hp->h_length);
		ret=0;
		goto _end;
	}

	strncpy(ifreq.ifr_name, ifn, IFNAMSIZ);
	if (ioctl(fd, SIOCGIFADDR, &ifreq)<0) { /* I believe the first arg is ignored */
		ret=-1;
		goto _end;
	}

	memcpy(addr, &((struct sockaddr_in *)&ifreq.ifr_addr)->sin_addr, 
			sizeof(struct in_addr));

_end:
	if (closefd)
		close(fd);

	return ret;
}

static int sock_join_(int fd, const struct in_addr *sin_mcast, 
		int salen, const char *in_if, int no_loop)
{
# if defined(SO_REUSEPORT)
	int port_reuse=0;
	int plen=sizeof(int);
# endif
	int addr_reuse=0;
	int alen=sizeof(int);
	struct ip_mreq mreq;
	unsigned char loop;
	unsigned char ttl;

	if ( 
# if defined(SO_REUSEPORT)
			(getsockopt(fd, SOL_SOCKET, SO_REUSEPORT, &port_reuse, &plen)<0) &&
# endif

		 (getsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &addr_reuse, &alen)<0) )
		return -1;

	if (
# if defined(SO_REUSEPORT)
			(port_reuse==0)&&
# endif
			(addr_reuse==0)) {
		return -1;
	}

	memcpy(&mreq.imr_multiaddr, sin_mcast, sizeof(struct in_addr));
	if (in_if) {
		if (sock_if2addr(fd, in_if, &(mreq.imr_interface))<0)
			return -1;
	}
	else {
		mreq.imr_interface.s_addr=htonl(INADDR_ANY);
	}

	if (setsockopt(fd, IPPROTO_IP, IP_ADD_MEMBERSHIP, &mreq, sizeof(mreq))<0) {
		return -1;
	}

	/**
	 * enable/disable loopback
	 */
	if (no_loop) {
		loop=0;
	}
	else {
		loop=1;
	}
	
	if (setsockopt(fd, IPPROTO_IP, IP_MULTICAST_LOOP, &loop, sizeof(loop))<0) {
		return -1;
	}
	
	/**
	 * Set TTL to the default value
	 */
	ttl=SOCK_MCAST_TTL;
	if (setsockopt(fd, IPPROTO_IP, IP_MULTICAST_TTL, &ttl, sizeof(ttl))<0) {
		return -1;
	}

	return 0;
}

/**
 * Again, we don't know whether the socket has
 * joined the group or not. Also, if the user just wishes to
 * drop all the groups it current joins (possibly because
 * it only joins into only one group), we cannot do that.
 * We do not allow for dropping a membership on a specific 
 * interface. A drop always drops the membership on all
 * interfaces.
 */
int sock_leave(int fd, const char *mgrp)
{
	struct ip_mreq mreq;
	struct sockaddr_in sin_mcast;

	if (sock_filladdr(mgrp, SOCK_PORT_ANY, &sin_mcast)<0) {
		return -1;
	}

	if (sock_is_mcast_addr(&sin_mcast, sizeof(sin_mcast))!=1) {
		return -1;
	}

	memcpy(&mreq.imr_multiaddr, &(sin_mcast.sin_addr), sizeof(struct in_addr));
	mreq.imr_interface.s_addr = htonl(INADDR_ANY);

	return (setsockopt(fd, IPPROTO_IP, IP_DROP_MEMBERSHIP, &mreq, sizeof(mreq)));
}

# define SOCK_MAX_RETRY 10

/**
 * send datagrams
 */
int sock_writeto(int fd, const char *host, int port, void *data, int *len)
{
	struct sockaddr_in to;
	int slen=-1;
	int retry=0;

	if (sock_get_sos_sig()) {
		return -1;
	}

	if (sock_filladdr(host, port, &to)<0) {
		return -1;
	}

	while ((slen=sendto(fd, data, *len, 0, (struct sockaddr *)&to, sizeof(to)))<0) {
		if (sock_get_sos_sig()) {
			break;
		}
		if (((errno!=EINTR)||(retry>=SOCK_MAX_RETRY)))
			break;
		retry++;
	}

	if (slen>=0) {
		*len=slen;
		return 0;
	}
	else {
		return -1;
	}
}

/**
 * receive datagrams
 * host(in/out): if present, it is used to restrict the source of a message
 * 	if NULL, we will receive from any source
 * 	if empty string, we will receive from any source and the real source IP
 * 		will be filled into that buffer.
 * port(in/out): if present, we will only receive from the specific source port
 *  if NULL, we will not restrict the source port
 *  if *port==SOCK_PORT_ANY, we will not restrict the source port and fill
 *  back in the source port.
 * data(in): the actual data buffer
 * len(in/out): passed in the size of the data buffer;
 *  will be filled with the actual length of the received data, 
 * tout(in): the timeout value in the unit of ms.
 * tout=0: non-block.
 * tout<0: block till return.
 * tout>0: actual timeout value is subject to host OS's timer 
 * resolution, which is 10ms on the current Linux OS.
 * Important: when a timeout occurs, it will return 0 but
 * a returned negative lenght indicates that no valid data is received.
 */
int sock_readfrom(int fd, char *host, int szhlen, int *port, 
		void *data, int *len, int tout)
{
	struct sockaddr_in from;
	int slen=sizeof(from);
	int rlen=-1;
	int retry=0;
	struct timeval tv;
	struct timeval *ptv;

	if (sock_get_sos_sig()) {
		return -1;
	}

	if (!len||(*len<=0)||(host&&(szhlen<16))) {
		return -1;
	}

	if (port) {
		if (sock_filladdr(host, *port, &from)<0) {
			return -1;
		}
	}
	else {
		if (sock_filladdr(host, SOCK_PORT_ANY, &from)<0) {
			return -1;
		}
	}

	if (tout<0) {
		ptv=NULL;
	}
	else {
		ptv=&tv;
	}

	while (1) {
		fd_set rset;
		int nfd=-1;

		if (sock_get_sos_sig()) {
			*len=0;
			return -1;
		}

		FD_ZERO(&rset);
		FD_SET(fd, &rset);

		if (tout>0) {
			/**
			 * blocked receive w/ timeout
			 */
			tv.tv_sec=tout/1000;
			tv.tv_usec=(tout%1000)*1000;
		}
		else if (tout==0) {
			/**
		 	* non-blocked receive
		 	*/
			tv.tv_sec=tv.tv_usec=0;
		}

		nfd=select(fd+1, &rset, NULL, NULL, ptv);

		if (nfd>0)
			break;

		if (nfd==0) {
			*len=-1;
			return 0;
		}

		if ((errno!=EINTR)||(retry>=SOCK_MAX_RETRY)) {
			*len=0;
			return -1;
		}

		retry++;
	}

	while (((rlen=recvfrom(fd, data, *len, 0, (struct sockaddr *)&from, &slen))<0)) {
		if (sock_get_sos_sig()) {
			break;
		}

		if ((errno!=EINTR)||(retry>=SOCK_MAX_RETRY))
			break;
		retry++;
	}

	if (rlen>=0) {
		*len=rlen;
		if (host)
			sock_dotaddr(host, szhlen, &(from.sin_addr));
		if (port)
			*port=ntohs(from.sin_port);
		return 0;
	}

	return -1;
}
