// -*- mode:C; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: sock_dg.h,v $ - header file for sock_dg.c. Socket library
 *            wrapper for datagram and multicast applications.
 *            Note that there is no guarantee that we be able to
 *            deliver all the data. It is required that the writer
 *            doesn't write more than a packets that allows you to 
 *            write. It is suggested to embed the message length as
 *            the first 4 bytes of the datagram so that the receiver
 *            will be able to determine whether the received packet
 *            is complete.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/07/26 06:40:36 $
 * Version :  $Revision: 1.8 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 
# ifndef _SOCK_DG_H_
# define _SOCK_DG_H_

/* # include "sock_sos.h" */

# if defined(__cplusplus)
extern "C" {
# endif

# define MAX_SAFE_PKT_LEN 1200

# define SOCK_MCAST_TTL 4

/**
 * Datagram socket creations.
 */

/**
 * Create a datagram socket and bind to the specified address.
 * host can be NULL, port=-1 means give me a random port.
 */
int sock_udp(const char *host, int port);

/**
 * Create a multicast socket.
 * mgrp must be a mcast address, port is necessary. 
 * port reuse option is always specified. bind_host
 * is added for filtering purposes. If no filtering
 * is needed, choose NULL. Or else, one can choose
 * to bind to a local IP address (that will allow you
 * to receive packets sent to both the specific IP or
 * mcast group). The most restrictive case would be 
 * to set bind_host the same as mgrp so that it only
 * receives packets destined to the multicast group. 
 * in_if and out_if are not required and might only
 * be needed for multi-homed node. They can be expressed
 * as interface names (eth0, etc) or IP addresses.
 */
int sock_mcast(const char *mgrp, int port, const char *bind_host, 
		const char *in_if, const char *out_if, int no_loop);

int sock_is_mcast_addr(const struct sockaddr_in *sa, int salen);

/**
 * Additional multicast memberships.
 * This is only allowed on sockets created by sock_mcast.
 * They may never needed unless we want to join to multiple
 * multicast groups. This is not particularly useful unless
 * multiple services use the same port on different multicast
 * addresses.
 */
int sock_join(int fd, const char *mcaddr, const char *in_if, int no_loop);
int sock_leave(int fd, const char *mcaddr);

/**
 * datagram send/receive
 */
int sock_writeto(int fd, const char *host, int port, void *data, int *len);

/**
 * receive datagrams
 * host(in/out): if present, it is used to restrict the source of a message
 * 	if NULL, we will receive from any source
 * 	if empty string, we will receive from any source and the real source IP
 * 		will be filled into that buffer.
 * port(in/out): if present, we will only receive from the specific source port
 *  if NULL, we will not restrict the source port
 *  if *port==SOCK_PORT_ANY, we will not restrict the source port and fill
 *  back in the source port.
 * data(in): the actual data buffer
 * len(in/out): passed in the size of the data buffer;
 *  will be filled with the actual length of the received data, 
 * tout(in): the timeout value in the unit of ms.
 * tout=0: non-block.
 * tout<0: block till return.
 * tout>0: actual timeout value is subject to host OS's timer 
 * resolution, which is 10ms on the current Linux OS.
 */
int sock_readfrom(int fd, char *host, int szhlen, int *port, 
		void *data, int *len, int tout);

int sock_if2addr(int fd, const char *ifn, struct in_addr *addr);

# if defined(__cplusplus)
}
# endif

# endif
