// -*- mode:C; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2000, 2001, 2002 The Regents of the University of 
  California, All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: sock_sr.c,v $ - Socket library wrapper, TCP send/recv with
 *            timeout support.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/10/01 22:46:25 $
 * Version :  $Revision: 1.7 $
 *
 * This is originally part of the TMPI release.
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

# include <sys/types.h>
# include <sys/socket.h>
# include <sys/uio.h>
# include <sys/time.h>
# include <unistd.h>
# include <sys/poll.h>

# ifndef POLLRDNORM
# define POLLRDNORM 0
# endif

# ifndef POLLWRNORM
# define POLLWRNORM 0
# endif

# include <errno.h>
# include "sock_sr.h"
# include "sock_sos.h"
# include "../usc/usc_time.h"

# define MAX_RETRY 3

/**
 * No need to control rate after using TCP_NODELAY
# define MAX_READ (10240000)
# define MAX_WRITE (40960000*2)
**/

# define MIN(a,b) (((a)>(b))?(b):(a))
/**
 * one can keep alive a connection by sending
 * packets with payload size 0 or advatising
 * a 0 sized receive window. Here, I am
 * only going to retry after a 0 sized packet
 * for MAX_RETRY times.
 */
int sock_write(int fd, char *data, int len)
{
	int n=0;
	int m;
	int retry=0;

	while (n<len) {
		if (sock_get_sos_sig()) {
			return -1;
		}
		m=write(fd, data+n, (len-n));
		if (m==-1) { /* is it because of user signal? */
			if (errno!=EINTR) {
				return -1;
			}

			retry++;
			if (retry>=MAX_RETRY)
				return -1;

			continue;
		}

		if (m==0) {
			retry++;
			if (retry>=MAX_RETRY)
				return -1;
		}
		else {
			n+=m;
		}
	}


	return 0;
}

/* might change iov! */
int	sock_writev(int fd, struct iovec *iov, int count)
{
	int n=0;
	int m;
	int retry=0;
	int len=0, i=0;

	for (; i<count; i++)
		len+=iov[i].iov_len;

	while (n<len) {
		if (sock_get_sos_sig()) {
			return -1;
		}
		m=writev(fd, iov, count);
		if (m==-1) { /* is it because of user signal? */
			if (errno!=EINTR) {
				return -1;
			}
			retry++;
			if (retry>=MAX_RETRY)
				return -1;

			continue;
		}

		if (m==0) {
			retry++;
			if (retry>=MAX_RETRY)
				return -1;
		}
		else {
			n+=m;

			/* adjusting iov */
			for (i=0; i<count; i++) {
				int t;

				if (m>iov[i].iov_len) {
					t=iov[i].iov_len;
				}
				else {
					t=m;
				}
				
				iov[i].iov_len-=t;
				iov[i].iov_base+=t;
				m-=t;

				if (m==0)
					break;
			}

			if (m!=0) {
				/* weired */
				return -1;
			}

			/* adjust iov */
			for (i=0; i<count; i++)
				if (iov[i].iov_len>0)
					break;

			iov+=i;
			count-=i;
		}
	}

	return 0;
}

int	sock_read(int fd, char *data, int len)
{
	int n=0;
	int m;
	int retry=0;

	while (n<len) {
		if (sock_get_sos_sig()) {
			return -1;
		}
	   /* m=recv(fd, data+n, (len-n), MSG_WAITALL);*/ 
		m=recv(fd, data+n, (len-n), 0);
		if (m==-1) { /* is it because of user signal? */
			if (errno!=EINTR) {
				return -1;
			}
			retry++;
			if (retry>=MAX_RETRY)
				return -1;

			continue;
		}

		if (m==0) {
			retry++;
			if (retry>=MAX_RETRY)
				return -1;
		}
		else {
			n+=m;
		}
	}


	return 0;
}

int	sock_read_tout(int fd, char *data, int *len, int tout)
{
	char *p=data;
	int  remain_len=*len;
    usc_time_t start, now;
	int ret=0;


	if (sock_get_sos_sig()) {
		*len = 0;
		return EL3RST; // system down.
	}

	if (tout<0) {
		/**
		 * blocked receive
		 */
		while (remain_len>0) {
			int read_bytes;
			if (sock_get_sos_sig()) {
				ret=EL3RST; // system down.
				break;
			}
			read_bytes=recv(fd, p, remain_len, MSG_WAITALL);
			if (read_bytes<0) {
				if (errno == EINTR)
					continue;
				else {
					ret=errno;
					break;
				}
			}
			p+=read_bytes;
			remain_len-=read_bytes;
		}

		*len = *len - remain_len;

		return ret;
	}

    for (now = start = usc_gettime(); remain_len>0; now = usc_gettime()) {
		double remain_time;
		struct pollfd pfd;
		int nfd=-1;

		if (sock_get_sos_sig()) {
			ret = EL3RST; // system down.
			break;
		}

		remain_time = (double)tout-(usc_difftime(now, start)/1000);
		if (remain_time<0) {
			ret = ETIMEDOUT;
			break;
		}

        pfd.fd = fd;
        pfd.events = POLLIN|POLLRDNORM;
        pfd.revents = 0;

		nfd=poll(&pfd, 1, (int)remain_time);
		if (nfd < 0) {
			if (errno == EINTR) {
				continue;
			}

			ret = errno;
			break;
		}

		if (nfd == 0) {
			continue;
		}

		if (pfd.revents & (POLLIN|POLLRDNORM)) {
			// now we try to read stuff.
			int read_bytes;

			if (sock_get_sos_sig()) {
				ret=EL3RST; // system down.
				break;
			}

			read_bytes=recv(fd, p, remain_len, 0);
			if (read_bytes<0) {
				if (errno == EINTR)
					continue;
				else {
					ret=errno;
					break;
				}
			}
			p+=read_bytes;
			remain_len-=read_bytes;
		}
	}

	*len = *len - remain_len;

	return ret;
}

int	sock_read_till_close(int fd, char *data, int len)
{
	int n=0;
	int m;
	int retry=0;

	while (n<len) {
		if (sock_get_sos_sig()) {
			return n;
		}
		m=recv(fd, data+n, (len-n), 0);
		if (m==-1) { /* is it because of user signal? */
			if (errno!=EINTR) {
				return n;
			}
			retry++;
			if (retry>=MAX_RETRY)
				return n;

			continue;
		}

		if (m==0) {
			retry++;
			if (retry>=MAX_RETRY)
				return n;

			break;
		}
		else {
			n+=m;
		}
	}

	return n;
}

/* readmax return the length of the actual data received, up to len */
int sock_readmax(int fd, char *data, int len)
{
	return read(fd, data, len);
}

int sock_canread(int fd)
{
	struct pollfd fdst;
	int ret;

	fdst.fd=fd;
	fdst.events=POLLIN|POLLRDNORM;
	fdst.revents=0;

	ret=poll(&fdst, 1, 0);

	if (ret==-1)
		return -1;

	if (ret==0)
		return 0;

	if (fdst.revents & (POLLIN|POLLRDNORM))
		return 1;
	return -1;
}

int sock_canwrite(int fd)
{
	struct pollfd fdst;
	int ret;

	fdst.fd=fd;
	fdst.events=POLLOUT|POLLWRNORM;
	fdst.revents=0;

	ret=poll(&fdst, 1, 0);

	if (ret==-1)
		return -1;

	if (ret==0)
		return 0;

	if (fdst.revents & (POLLOUT|POLLWRNORM))
		return 1;
	return -1;
}

char *sock_gets(char *buf, int buf_size, int istream)
{
	int  idx;
	int  status;
	int retry=0;

	if (!buf)
		return NULL;

	buf[0]='\0';
	buf[buf_size-1]='\0';

	for (idx=0; idx<buf_size-1; idx++) {
		/* it is inefficient to read one byte at a time */
		/* will change it to a buffer management layer to improve it */
		if (sock_get_sos_sig()) {
			*(buf+idx)='\0';
			break;
		}
		status=read(istream,buf+idx,1);
		if (status<0) {
			if (errno!=EINTR) {
				*(buf+idx)='\0';
				break;
			}

			retry++;
			if (retry>=MAX_RETRY) {
				*(buf+idx)='\0';
				break;
			}
			else
				continue;
		}

		if (status==0) { /* connection closed */
			*(buf+idx)='\0';
			break;
		}

		if (*(buf+idx)=='\n') {
			*(buf+idx+1)='\0';
			break;
		}
	}

	if (buf[0]=='\0')
		return NULL;

	return buf;
}
