// -*- mode:C; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*- 
/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: sp_client.cc,v $ - header file for socket pool client
 *            module. A socket pool client module maintains a cache of connected sockets.
 *            The cache is implemented as a multimap, and is indexed by the 
 *            destination address (<ip, port> pair). Two operations are
 *            permitted on the socket pool: checking in a connection that
 *            we no longer in use. checking out a connection to a certain
 *            destination.  A connection returned by the socket pool may be already 
 *            closed by the server side.
 * Authors :  Lingkun Chu, lkchu@cs.ucsb.edu
 *            Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/10/03 01:44:48 $
 * Version :  $Revision: 1.5 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

# include <sys/types.h>
# include <sys/socket.h>
# include <netinet/in.h>
# include <arpa/inet.h>

# include <assert.h>
# include <string.h>
# include <stdio.h>

# include "sock_conn.h"
# include "sp_client.h"
# include "../debug/tmpi_debug.h"

// using namespace std;

int SocketPoolClient::checkout(const char *ip, int port)
{
    uint32_t addr;

	if (inet_aton(ip,  (struct in_addr *)&addr) == 0) {
    	struct hostent *hp;
		if ( (hp=sock_gethostbyname(ip)) == NULL ) {
			return -1;
		}

		assert(hp->h_length == sizeof(uint32_t));
		memcpy((void *)&addr, hp->h_addr_list[0], hp->h_length);
	}

    return checkout(addr, port);
}

int SocketPoolClient::checkout(uint32_t addr, int port)
{
	SocketKey r(addr, port);
	if (pthread_mutex_lock(&lock)<0) {
		return -1;
	}

	int sock = -1;

	std::pair<iterator, iterator> s = equal_range(r);
	iterator next;
	for (iterator i=next=s.first; i != s.second; i = next) {
		next++;
# ifndef NO_SOCK_CHECK
		int error;
		socklen_t len=sizeof(int);
		if ((getsockopt(i->second, SOL_SOCKET, SO_ERROR, &error, &len) == 0) && (error == 0)) {
# endif
			sock = i->second;
# ifndef NO_SOCK_CHECK
		}
# endif
		erase(i);
		if (sock >= 0) {
			break;
		}
	}

	if (sock<0) {
		n_misses++;
	}
	else {
        // tmpi_error(DBG_VERBOSE, "checkout returns %d\n", sock);
		n_hits++;
	}

	pthread_mutex_unlock(&lock);

	return sock;
}

int SocketPoolClient::checkin(int sock, const char *ip, int port)
{
    uint32_t addr;

	if (ip == NULL) {
		// I need to get the destination by myself.
		struct sockaddr_in sin;
		socklen_t sinlen=sizeof(sin);

		if (getpeername(sock, (struct sockaddr *)&sin, &sinlen)<0)
			return -1;

		memcpy((void *)&addr, &(sin.sin_addr), sizeof(uint32_t));
		port=ntohs(sin.sin_port);
	}
	else {
		// attempt to resolve the ip as the dot notation.
		if (inet_aton(ip,  (struct in_addr *)&addr) == 0) {
			// atempt to resolve the ip as a generic name.
			struct hostent *hp;
			if ( (hp=sock_gethostbyname(ip)) == NULL ) {
				return -1;
			}

			assert(hp->h_length == sizeof(uint32_t));
			memcpy((void *)&addr, hp->h_addr_list[0], hp->h_length);
		}
	}

    return checkin(sock, addr, port);
}

int SocketPoolClient::checkin(int sock, uint32_t addr, int port)
{
	SocketKey key(addr, port);

	pthread_mutex_lock(&lock);
	insert(value_type(key, sock));
    // tmpi_error(DBG_VERBOSE, "Socket %d checked in\n", sock);
	pthread_mutex_unlock(&lock);

	return 0;
}

void SocketPoolClient::summary()
{
	printf("Hits=%d(%.3f%%), Misses=%d(%.3f%%).\n", n_hits, (double)n_hits*100/(n_hits+n_misses),
			n_misses, (double)n_misses*100/(n_hits+n_misses));
}

void SocketPoolClient::dump()
{
	for (const_iterator ci=begin(); ci != end(); ci++) {
		char buf[128];
		sock_dotaddr(buf, 128, (struct in_addr *)&(ci->first.addr));
		printf("%s:%d -> %d\n", buf, ci->first.port, ci->second);
	}
}
