// -*- mode:C; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: sp_client.h,v $ - header file for socket pool client
 *            module. A socket pool client module maintains a cache of connected sockets.
 *            The cache is implemented as a multimap, and is indexed by the 
 *            destination address (<ip, port> pair). Two operations are
 *            permitted on the socket pool: checking in a connection that
 *            we no longer in use. checking out a connection to a certain
 *            destination.  A connection returned by the socket pool may be already 
 *            closed by the server side.
 * Authors :  Lingkun Chu, lkchu@cs.ucsb.edu
 *            Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/10/01 10:53:47 $
 * Version :  $Revision: 1.3 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

#ifndef SP_CLIENT_H
#define SP_CLIENT_H

# include <map>
# include <inttypes.h>
# include <pthread.h>

/** The key to index a socket */
struct SocketKey {
    uint32_t addr; /**< ipv4 address of a remote server */
    int port; /**< port on a remote server*/

    SocketKey(uint32_t a, int p): addr(a), port(p) { };
    SocketKey(const SocketKey &k) : addr(k.addr), port(k.port) { };
};

class SocketKey_cmp : public std::less<SocketKey> {
 public:
	bool operator()(const SocketKey &x, const SocketKey &y) const {
        return ((x.addr < y.addr) || ((x.addr == y.addr) && (x.port < y.port)));
	}
};

/** The socket pool data structure for the client */
class SocketPoolClient : std::multimap<SocketKey, int, SocketKey_cmp> {
    int max_size; /**< The maximum pool size, currently not used */
    pthread_mutex_t lock; /**< The lock to access the socket pool */
	int n_hits, n_misses;

public:
    SocketPoolClient(int sz): max_size(sz), n_hits(0), n_misses(0)
    { 
        pthread_mutex_init(&lock, NULL);
    }

    ~SocketPoolClient()
    { 
        pthread_mutex_destroy(&lock);
    }

    /**< check out a socket if it is available, -1 is returned if not */
    int checkout(uint32_t addr, int port); 
    int checkout(const char *ip, int port); 
    /**< check in a socket to the socket pool*/
    int checkin(int sock, uint32_t addr, int port); 
    int checkin(int sock, const char *ip, int port);
	void summary();
	void dump();
};

#endif
