// -*- mode:C; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: sp_server.cc,v $ - socket pool server module implemenation. 
 * 			  A socket pool server module maintains a pool of open
 *            sockets, and perform select() on them. Two types of operations
 *            are supported: get_one() blocks until one socket is readable.
 *            put_one() returns a socket back to the pool (which is still
 *            connected with the client).
 * Authors :  Lingkun Chu, lkchu@cs.ucsb.edu
 *            Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/10/03 01:44:48 $
 * Version :  $Revision: 1.5 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 

# include <stdio.h>
# include <unistd.h>
# include <stdlib.h>
# include <sys/socket.h>
# include <arpa/inet.h>
# include <netinet/in.h>
# include <net/if.h>
# include <netdb.h>
# include <errno.h>
# include <sys/poll.h>
# include <sys/stat.h>
# include <fcntl.h>

# include "sp_server.h"
# include "../usc/usc_time.h"
# include "../debug/tmpi_debug.h"

// using namespace std;

SocketPoolServer::SocketPoolServer(int sz, int p): max_size(sz)
{
    fds = (struct pollfd *) calloc(max_size,  sizeof(struct pollfd));
    if (fds == NULL) 
        throw -1;

	accept_fd=sock_listen(p, 100);

	if (accept_fd < 0)
		throw -1;

    int flags = fcntl(accept_fd, F_GETFL, 0);
    if (!(flags & O_NONBLOCK))
        fcntl(accept_fd, F_SETFL, flags | O_NONBLOCK);

    if (pipe(pipefds) == -1)
        throw -1;
}

SocketPoolServer::~SocketPoolServer(void)
{
	close(accept_fd);
	close(pipefds[0]);
	close(pipefds[1]);

    for (const_iterator i=begin(); i!=end(); i++) {
		if (*i >= 0)
        	close(*i);
    }

	if (fds)
    	free(fds);
}

// Note that we do not provide extensive status code here.
// If the caller gives a bad socket ID, we may still return success.
// If the socket is actually a duplicate, we won't detect it either.
bool SocketPoolServer::put_one(int sock)
{
    if (write(pipefds[1], (char *) &sock, sizeof(int)) == sizeof(int)) {
		int reply;
		read(pipefds[1], (char *)&reply, sizeof(int));
		if (reply == sock)
			return true;
	}

    return false;
}

void SocketPoolServer::dump()
{
	printf("[ %d %d ", accept_fd, pipefds[0]);
	for (const_iterator ci=begin(); ci != end(); ci++) {
		printf("%d ", *ci);
	}
	printf("]\n");
}

int SocketPoolServer::get_one(int tout_in_ms)
{
    usc_time_t start, now;
     
    for (now = start = usc_gettime(); true; now = usc_gettime()) {
		// dump();
		double remain_time=(double)tout_in_ms-(usc_difftime(now, start)/1000); 

		if (remain_time<0) {
			errno=ETIMEDOUT;
			return -1;
		}

        fds[0].fd = accept_fd;
        fds[0].events = POLLIN;
        fds[0].revents = 0;

        fds[1].fd = pipefds[0];
        fds[1].events = POLLIN;
        fds[1].revents = 0;
        int cnt = 2;

		iterator next;
		for (iterator i=next=begin(); i != end(); i = next) {
			next++;
			int error;
    		socklen_t len=sizeof(int);
		    if ( (getsockopt(*i, SOL_SOCKET, SO_ERROR, &error, &len) < 0) ||
					(error != 0)) {
				tmpi_error(DBG_VERBOSE, "Bad socket (%d)\n", *i);
                close(*i);
                erase(i);
                continue;
            }

            fds[cnt].fd = *i;
            fds[cnt].events = POLLIN;
            fds[cnt].revents = 0;
            cnt++;
            if (cnt == max_size) {
                errno = ENOMEM;
                return -1;
            }
        }

        int n = poll(fds, cnt, int(remain_time));
		if (n < 0) {
			if (errno == EBADF) {
				continue;
			}

			if (errno == EINTR) {
				continue;
			}

			return -1;
		}

        if (n == 0) {
			continue;
        }

		// checking the special sockets.
        if (fds[0].revents & POLLIN) {
			// tmpi_error(DBG_VERBOSE, "connection request received.\n");
			int one=sock_accept(accept_fd);
			if (one >= 0) {
            	insert(one);
				// dump();
				// tmpi_error(DBG_VERBOSE, "connection accepted (%d).\n", one);
            }
			else {
				tmpi_error(DBG_VERBOSE, "connection acceptance failed.\n");
			}
        }

        if (fds[1].revents & POLLIN) {
			int one;
            if (read(pipefds[0], (char *) &one, sizeof(int)) == sizeof(int)) {
            	insert(one);
				// dump();
				// tmpi_error(DBG_VERBOSE, "put_one accepted (%d).\n", one);
				// reply with the same fd.
				write(pipefds[0], (char *)&one, sizeof(int));
            }
        }

    	int fd = -1;
        for (int k=2; k<cnt; k++) {
            int e = fds[k].revents;
            if (e & POLLIN) {
				int error;
				socklen_t len=sizeof(int);
		    	if ( (getsockopt(fds[k].fd, SOL_SOCKET, SO_ERROR, &error, &len) == 0) && (error == 0)) {
            		fd = fds[k].fd;
					break;
				}
				else {
					erase(fd);
					tmpi_error(DBG_VERBOSE, "Bad socket (%d).\n", fd);
				}
            }
        }

        if (fd >= 0) {
            erase(fd);
			// tmpi_error(DBG_VERBOSE, "get_one returns %d.\n", fd);
            return fd;
        }
    }

	errno = ETIMEDOUT;
    return -1;
}
