// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: udp_cli.cc,v $ - UDP client library. Auto-retrial, sequence
 *            numbering. Asynchronous request/response.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/09/12 06:40:26 $
 * Version :  $Revision: 1.18 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 
# include <sys/time.h>
# include <vector>
# include <unistd.h>
# include <stdio.h>
# include "sock_conn.h"
# include "sock_dg.h"
# include "udp_cli.h"
# include "../guid/guid.h"

using namespace std;

# define REQ_TRY_INTERNAL (-2)

struct Udp_req {
	int req_try; // remaining retries
	unsigned int tout; // expiration interval
	unsigned int expire; // abs time to expire
	Udp_msg *req_msg;
	char *host;
	int port;
	Udp_res **res;
	int max_res;
	int min_res;
	int i_res;

public :
	void reset()
	{
		req_try=0;
		tout=0;
		expire=0;
		req_msg=NULL;
		res=NULL;
		host=NULL;
		port=-1;
		max_res=0;
		min_res=0;
		i_res=0;
	}

	Udp_req()
	{
		reset();
	}

	Udp_req(int maxtry, const void *req, int l, const char *h, int p, seq_t sq, int to, unsigned short mjc, int mxres, int mnres);
	~Udp_req() { release(); }
	int request(int fd);
	int expired();
	int to_expire();

	seq_t seq() {
		if (!req_msg) {
			return UDP_SEQ_ERR;
		}

		return req_msg->seq;
	}

private:
	void release();
};

Udp_req::Udp_req(int maxtry, const void *req, int l, const char *h, int p, seq_t sq, int tot, unsigned short mjc, int mxres, int mnres)
{
	reset();

	if ( (l<=0) || ((l+UDP_MSG_HEADER_SZ)>UDP_MAX_DG_SZ) || (!req) || 
			(sq==UDP_SEQ_ANY) || (sq==UDP_SEQ_ERR) || (!h) || (p<0) || 
			(p==SOCK_PORT_ANY) || (tot<=0) || (mxres<=0)|| (mnres<=0)) {
		return;
	}
	else {
		host=new char[strlen(h)+1];
		if (!host) {
			release();
			return;
		}
		strcpy(host, h);

		req_try=(maxtry>0)?maxtry:-1;
		tout=tot;
		port=p;
		max_res=mxres;
		min_res=mnres;

		res=new Udp_res*[max_res+1];
		if (!res) {
			release();
			return;
		}

		memset(res, 0, sizeof(Udp_res *)*(max_res+1));

		res[0]=new Udp_res(sq);
		if (!res[0]) {
			release();
			return;
		}
		req_msg=(Udp_msg *)malloc(UDP_MSG_HEADER_SZ+l);
		if (!req_msg) {
			release();
			return;
		}
		req_msg->magic=mjc;
		req_msg->len=l+UDP_MSG_HEADER_SZ;
		req_msg->seq=sq;
		memcpy(req_msg->buf, req, l);
	}
}

void Udp_req::release()
{
	if (req_msg) {
		free(req_msg);
		req_msg=NULL;
	}

	if (res) {
		for (int i=0; (i<=i_res)&&(i<max_res); i++) {
			if (res[i]) {
				delete res[i];
			}
		}
		delete []res;
		res=NULL;
	}
	
	if (host) {
		delete []host;
		host=NULL;
	}

	reset();
}

static unsigned int now()
{
	struct timeval tv;
	gettimeofday(&tv, NULL);

	return (tv.tv_sec*1000+tv.tv_usec/1000);
}

int Udp_req::request(int s)
{
	if ((!req_msg)||(!res)||(i_res>=max_res)||(!res[i_res]))
		return -1;

	if (res[i_res]->status!=UDP_RES_INIT) // weird
		return 0;

	if (req_try==REQ_TRY_INTERNAL) {
		return 0;
	}
	if ((req_try>0)||(req_try==-1)) {
		int l=req_msg->len;
		if (sock_writeto(s, host, port, req_msg, &l)<0) {
			res[i_res]->status=UDP_RES_SYST;
			req_try=REQ_TRY_INTERNAL;
			res[++i_res]=new Udp_res(seq()); // maintain consistency
			return 0;
		}
		if (l!=req_msg->len)  {
			res[i_res]->status=UDP_RES_LENG;
			req_try=REQ_TRY_INTERNAL;
			res[++i_res]=new Udp_res(seq()); // maintain consistency
			return 0;
		}
		expire=now()+tout;
	}
	else {
		res[i_res]->status=UDP_RES_TOUT;
		res[++i_res]=new Udp_res(seq());
		req_try=REQ_TRY_INTERNAL;
		return 0;
	}

	req_try=(req_try==-1)?-1:(req_try-1);

	return 0;
}

int Udp_req::to_expire()
{
	return expire-now();
}

int Udp_req::expired()
{
	if (now()>expire)
		return 1;

	return 0;
}

Udp_res::~Udp_res() 
{
	if (msg)
		free(msg);
}

// on failure, the response will merely be discarded
int Udp_res::deposit(const char *ad, int p, void *buf, int l, unsigned short s)
{
	if (status!=UDP_RES_INIT) {
		return -1;
	}

	if (strlen(ad)>15) {
		return -1;
	}

	if ((msg=(Udp_msg *)malloc(l))==NULL) {
		return -1;
	}
	
	status=s;
	port=(short)p;
	strcpy(addr, ad);
	len = l;
	memcpy(msg, buf, l);

	return 0;
}

int Udp_res::replace(void *buf, int l, unsigned short s)
{
	if (msg) {
		free(msg);
	}

	if ((msg=(Udp_msg *)malloc(l))==NULL) {
		return -1;
	}
	
	status=s;
	len = l;
	memcpy(msg, buf, l);

	return 0;
}

void Udp_res::dump()
{
	int exam_data=1;

	seq_t sq=seq();
	if (sq==UDP_SEQ_ANY) {
		printf("seq#\t: SEQ_ANY\n");
	}
	else if (sq==UDP_SEQ_ERR) {
		printf("seq#\t: SEQ_ERR\n");
	}
	else {
		printf("seq#\t: 0x%08X\n", sq);
	}

	printf("status\t: ");
	if (status==UDP_RES_OKAY) {
		printf("OKAY");
	}
	else if (status==UDP_RES_INIT) {
		printf("INIT");
		exam_data=0;
	}
	else {
		if (status&UDP_RES_PORT) {
			printf("PORT ");
		}
		if (status&UDP_RES_ADDR) {
			printf("ADDR ");
		}
		if (status&UDP_RES_LENG) {
			printf("LENG ");
		}
		if (status&UDP_RES_SYST) {
			printf("SYST ");
			exam_data=0;
		}
		if (status&UDP_RES_TOUT) {
			printf("TOUT ");
			exam_data=0;
		}
	}
	printf("\n");
	if (exam_data) {
		printf("port\t: %d\n", port);
		printf("addr\t: %s\n", addr);
		printf("length\t: %d\n", length());
		printf("data\t: ");
		unsigned char *data=(unsigned char *)result();
		if (!data) {
			printf("<NUL>");
		}
		else {
			for (int i=0; i<length(); i++) {
				printf("%02X ", data[i]);
			}
		}
	
		printf("\n");
	}
}

static seq_t justify_seq(seq_t sq)
{
	while ((sq==UDP_SEQ_ANY)||(sq==UDP_SEQ_ERR)) {
		sq++;
	}

	return sq;
}

Udp_cli::Udp_cli(int ss, int tot): s(ss), roe(false), tout(tot)
{
	next_seq=Guid::random_ui();
	next_seq=justify_seq(next_seq);
}

Udp_cli::Udp_cli(const char *host, int port, int tot) : roe(true), tout(tot)
{
	s=sock_udp(host, port);
	next_seq=Guid::random_ui();
	next_seq=justify_seq(next_seq);
}

void Udp_cli::free_results(Udp_res **res)
{
	if (res) {
		Udp_res **p=res;
		for (; *p; p++) {
			delete *p;
		}
		
		delete []res;
	}
}

seq_t Udp_cli::async_req_(const char *host, int port, const void *buf, 
		int len, int max_retry, unsigned short mjc, 
		int max_res=1, int min_res=1)
{
	if ((s<0)||(tout<=0))
		return UDP_SEQ_ERR;

	seq_t seq=next_seq++;
	next_seq=justify_seq(next_seq);
	Udp_req *req=new Udp_req(max_retry, buf, 
			len, host, port, seq, tout, 
			mjc, max_res, min_res);

	if (!req)
		return UDP_SEQ_ERR;

	if (!(req->req_msg)) {
		delete(req);
		return UDP_SEQ_ERR;
	}
	
	if (!(req->request(s))) {
		requests.push_back(req);
		return seq;
	}
	else {
		delete req;
		return UDP_SEQ_ERR;
	}

}

seq_t Udp_cli::async_req(const char *host, int port, void *buf, int len, int max_retry, int max_res, int min_res)
{
	return async_req_(host, port, buf, len, max_retry, UDP_MSG_REQ_MAG, max_res, min_res);
}

Udp_res ** Udp_cli::sync_req(const char *host, int port, void *buf, int len, int max_retry, int max_res, int min_res)
{
	seq_t seq=async_req(host, port, buf, len, max_retry);

	if (seq==UDP_SEQ_ERR) {
		return NULL;
	}

	return wait_one(seq);
}

// keep a response for later retrieval
Udp_res ** Udp_cli::wait_one(seq_t seq)
{
	if ((s<0)||(tout<=0))
		return NULL;

	try_to_recv_one(0);

	while (1) {
		vector<Udp_req *>::iterator p=requests.begin();
		int next_expire=-1;
		int no_pending=1;
		int no_match=1;

		for (; p!=requests.end();p++) {
			Udp_req *rq=*p;

			if (!rq)
				continue;

			// should we restart the request?
			// also calculate the soonest time-out
			if ((rq->i_res<rq->min_res)&&(rq->req_try!=REQ_TRY_INTERNAL)) {
				// found an unfinished request
				no_pending=0;
	
				int to_exp=rq->to_expire();
				if (to_exp<=0) {
					rq->request(s);
					to_exp=rq->to_expire();
				}

				if ((next_expire<0)||(to_exp<next_expire)) {
					next_expire=to_exp;
				}
			}

			// is it a match?
			seq_t sq=rq->seq();

			if ((sq==UDP_SEQ_ERR)||((seq!=UDP_SEQ_ANY)&&(seq!=sq))) {
				// not a match, next request
				continue;
			}

			// found matching
			no_match=0;

			// is this matching request done?
			if ((rq->i_res>=rq->min_res)||(rq->req_try==REQ_TRY_INTERNAL)) {
				Udp_res **ret=new Udp_res *[rq->i_res+1];
				memset(ret, 0, sizeof(Udp_res *)*(rq->i_res+1));
				for (int i=0, j=0; i<rq->i_res; i++) {
					if (rq->res[i]) {
						ret[j++]=rq->res[i];
						rq->res[i]=NULL;
					}
				}
				delete rq;
				requests.erase(p);
				// *p=NULL;
				return ret;
			}

			// found match request but it hasn't completed yet.
			if (seq!=UDP_SEQ_ANY) {
				break;
			}
		}

		if (no_match) { // no matching sequence number.
			return NULL;
		}

		// actually no_match==0 and no-return implies
		// no_pending==0 (because otherwise, the request
		// would have been done and returned.
		// On the other hand, if no_pending is 0, no_match
		// could be 0 or 1. The the following test could
        // redundant.
		if (no_pending) {
			return NULL;
		}

		// nothing to return so far, so do select here
		if (try_to_recv_one(next_expire)<0) {
			// error to receive, return NULL.
			return NULL;
		}
	}
}

int Udp_cli::try_to_recv_one(int tout)
{
	if ((tout<0)||(s<0))
		return -1;

	// first do a wildcard readfrom with time out,
	// then do a number of non-blocking readfrom
	while (1) {
		int port=SOCK_PORT_ANY;
		char addr[16]="";
		char buf[UDP_MAX_DG_SZ];
		int len=UDP_MAX_DG_SZ;
	
		if (sock_readfrom(s, addr, 16, &port, buf, &len, tout)<0)
			return -1;

		tout=0;

		if (len<0) // normal timeout
			return 0;

		// sanity check
		Udp_msg *msg=(Udp_msg *)buf;
		if (msg->magic!=UDP_MSG_RES_MAG)
			continue;

		// sequence number matching
		seq_t res_seq=msg->seq;
		vector<Udp_req *>::const_iterator p=requests.begin();
	
		for (; p!=requests.end();p++) {
			Udp_req *rq=*p;
	
			if (!rq)
				continue;
	
			seq_t sq=rq->seq();

			if ((sq==UDP_SEQ_ERR)||(res_seq!=sq)) {
				// no matching
				continue;
			}
			
			// found matching 
			
			if (rq->i_res>=rq->max_res) {
				// excessive responses.
				break;
			}

			// prepare to deposit the message
			unsigned short status=UDP_RES_OKAY;
			if ((rq->port>=0)&&(rq->port!=SOCK_PORT_ANY)&&(rq->port!=port))
				status|=UDP_RES_PORT;
	
			// host comparison is too costly and thus UDP_RES_ADDR error
			// is never checked.
			if (len!=msg->len)
				status|=UDP_RES_LENG;

			// check duplicates
			int i=0;
			for (; i<rq->i_res; i++) {
				Udp_res *ss=rq->res[i];
				if (ss) {
					if ((strcmp(ss->addr, addr)==0)&&(ss->port==port)) {
						// found match.
						if ((ss->status!=UDP_RES_OKAY)&&(status==UDP_RES_OKAY)) {
							// upgrade old response
							ss->replace(buf, len, status);
						}

						break;
					}
				}
			}

			if (i>=rq->i_res) {
				// no match found
				if (rq->res[rq->i_res]->deposit(addr, port, buf, len, status)==0) {
					rq->res[++rq->i_res]=new Udp_res(rq->seq());
				}
			}

			break;
		}

		// no match, discard it.
	}
}

Udp_res ** Udp_cli::wait_all()
{
	if ((s<0)||(tout<=0))
		return NULL;

	int done=0;
	int res_num=0;

	while (!done) {
		vector<Udp_req *>::const_iterator p=requests.begin();
		int next_expire=-1;

		done=1;
		res_num=0;

		for (; p!=requests.end();p++) {
			Udp_req *rq=*p;

			if (!rq)
				continue;

			// should we restart the request?
			// also calculate the soonest time-out
			if ((rq->i_res<rq->min_res)&&(rq->req_try!=REQ_TRY_INTERNAL)) {
				done=0;
	
				int to_exp=rq->to_expire();
				if (to_exp<=0) {
					rq->request(s);
					to_exp=rq->to_expire();
				}

				if ((next_expire<0)||(to_exp<next_expire)) {
					next_expire=to_exp;
				}
				// for incompleted requests, we need to
				// return at least one response with status=SYST
				res_num+=(rq->i_res)?rq->i_res:1;  
			}
			else {
				res_num+=rq->i_res;
			}

		}

		if (!done) {
			// do select here
			if (try_to_recv_one(next_expire)<0) {
				break;
			}
		}
	}

	// set those in UDP_RES_INIT to UDP_RES_SYST and return
	vector<Udp_req *>::iterator p=requests.begin();
	Udp_res **ret=new Udp_res *[res_num+1];
	memset(ret, 0, sizeof(Udp_res *)*(res_num+1));

	int i=0;

	for (; p!=requests.end();p++) {
		Udp_req *rq=*p;

		if (!rq)
			continue;

		if (rq->i_res) {
			for (int j=0; j<rq->i_res; j++) {
				if (rq->res[j]) {
					ret[i++]=rq->res[j];
					rq->res[j]=NULL;
				}
			}
		}
		else {
			if (rq->res[0]->status==UDP_RES_INIT) {
				rq->res[0]->status=UDP_RES_SYST;
			}
			ret[i++]=rq->res[0];
			rq->res[0]=NULL;
		}
		delete rq;
		//requests.erase(p);
		*p=NULL;
	}

	requests.erase(requests.begin(), requests.end());

	return ret;
}
			
int Udp_cli::stop_srv(const char *host, int port, const void *auth, int len, int max_retry)
{
	seq_t seq=async_req_(host, port, auth, len, max_retry, UDP_MSG_STP_MAG);

	if (seq==UDP_SEQ_ERR) {
		return -1;
	}

	Udp_res **res=wait_one(seq);

	if ((!res)||(!res[0])||(res[0]->status!=UDP_RES_OKAY)) {
		free_results(res);
		return -1;
	}

	free_results(res);

	return 0;
}

Udp_cli::~Udp_cli()
{
	if (roe==true)
		close(s);

	vector<Udp_req *>::iterator p=requests.begin();
	for (; p!=requests.end();p++) {
		if (*p) {
			delete *p;
			*p=NULL;
		}
	}

	// requests.clear();
}
