// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: udp_cli.h,v $ - header file for udp_cli.cc. UDP clients,
 *            Implement synchronous/asynchronous request/response,
 *            single request/multiple response, time-out, automatic
 *            retransmission, sequence numbering. Udp_Cli is not MT-safe and
 *            is not meant to be. It is suggested to use per-thread Udp_Cli
 *            object.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/09/12 06:40:26 $
 * Version :  $Revision: 1.18 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 
# ifndef _UDP_CLI_H_
# define _UDP_CLI_H_
# include <stdlib.h>
# include <vector>
# include "udp_common.h"

# define UDP_RES_OKAY ((unsigned short)0)
// response content uninitialized
# define UDP_RES_INIT ((unsigned short)1)
// server port not match
# define UDP_RES_PORT ((unsigned short)2)
// server address not match
# define UDP_RES_ADDR ((unsigned short)4)
// datagram truncated 
# define UDP_RES_LENG ((unsigned short)8)
// system error, e.g. socket closed
# define UDP_RES_SYST ((unsigned short)16)
// failed for too many retrials
# define UDP_RES_TOUT ((unsigned short)32)

struct Udp_res;

# ifndef MIN
# define MIN(a,b) (((a)>(b))?(b):(a))
# endif

struct Udp_res {
	unsigned short status;
	short port;
	char addr[16];
 // private:
	int len;
	seq_t res_seq;
	Udp_msg *msg;

public:
	Udp_res(seq_t sq) {
		status=UDP_RES_INIT;
		res_seq=sq;
		msg=NULL;
		// uninitiated or no-content messages will have the same address.
		addr[0]='\0';
		port=-1;
	}

	~Udp_res(); 

	// retrieve the states of the response
	void *result() {
		if ((status==UDP_RES_INIT)||(!msg))
			return NULL;

		return (void *)(msg->buf);
	}

	int length() {
		if ((status==UDP_RES_INIT)||(!msg))
			return -1;

		return MIN(len, msg->len)-UDP_MSG_HEADER_SZ;
	}

	seq_t seq() {
		return res_seq;
		// return msg->seq;
	}

	// keep a response for later retrieval
	int deposit(const char *ad, int p, void *buf, int l, unsigned short s);
	int replace(void *buf, int l, unsigned short s);
	void dump();
};

// private internal class
struct Udp_req;

class Udp_cli {
private:
	int s; // the socket
	bool roe; // should we close s on exit?
	int tout; // in mili-seconds
	std::vector<Udp_req *> requests; // request and response
	seq_t next_seq; 
	int try_to_recv_one(int tout);
	seq_t async_req_(const char *host, int port, const void *buf, 
			int len, int max_retry, unsigned short mjc, int max_res, int min_res); 

public:
	Udp_cli(int s, int tout); // wrap around an existing socket
	Udp_cli(const char *host, int port, int tout); // create a socket

	seq_t async_req(const char *host, int port, 
			void *buf, int len, int max_retry, int max_res=1, int min_res=1);

	Udp_res **sync_req(const char *host, int port, 
			void *buf, int len, int max_retry, int max_res=1, int min_res=1); 

	Udp_res **wait_one(seq_t seq); // wait for one request
	
	int stop_srv(const char *host, int port, 
			const void *auth, int len, int max_retry);

	Udp_res **wait_all(); // an array of n+1 Udp_res*'s, 
						  // where n is the actual nmber of 
						  // responses the last one is is NULL.
	~Udp_cli();

	bool good() { if (s>=0) return true; return false; }

	static void free_results(Udp_res **res);
};

# endif
