// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: udp_srv.h,v $ - header file for udp_srv.cc. UDP server. Much
 *            of the packet-loss handling is done on the UDP client side,
 *            thus, the UDP server is relatively simple.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/06/07 06:59:06 $
 * Version :  $Revision: 1.13 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 
# ifndef _UDP_SRV_H_
# define _UDP_SRV_H_
# include <stdlib.h>
# include <string.h>
# include <unistd.h>
# include "../machine/atomic.h"
# include "sock_dg.h"
# include "udp_common.h"

# ifndef MIN
# define MIN(a,b) (((a)>(b))?(b):(a))
# endif

struct Udp_srv_req {
	char *host;
	int port;
	int len; // raw packet length as received
	Udp_msg *msg; // raw packet

	Udp_srv_req()
	{
		host=NULL;
		port=-1;
		len=0;
		msg=NULL;
	}

	Udp_srv_req(void *req, int l, char *h, int p);
	~Udp_srv_req() { release(); }

	// retrieve the states of the request
	seq_t seq() {
		if (!msg) {
			return UDP_SEQ_ERR;
		}

		return msg->seq;
	}

	unsigned short magic() {
		if (!msg) {
			return UDP_MSG_ERR_MAG;
		}

		return msg->magic;
	}
	
	void *request() {
		if (!msg)
			return NULL;

		return (void *)(msg->buf);
	}

	int length() {
		if (!msg)
			return -1;

		return MIN(len, msg->len)-UDP_MSG_HEADER_SZ;
	}

	bool valid()
	{
		if ((!host)||(!msg)) {
			return false;
		}

		if (len!=msg->len) {
			return false;
		}

		unsigned short m=magic();
		if ((m!=UDP_MSG_REQ_MAG)&&(m!=UDP_MSG_STP_MAG)) {
			return false;
		}

		return true;
	}

	bool is_stop_req()
	{
		unsigned short m=magic();
		if ((m==UDP_MSG_STP_MAG)) {
			return true;
		}
		return false;
	}

private	:
	void release();
};

// server is almost stateless
// The following is a sample sketch of a UDP server
/**********************************************
int run(char *host, int port)
{
	Udp_srv *srv=new Udp_srv(host, port);
	while (Udp_srv_req *q=srv.wait_for_request()) {
		if (q.is_stop_req()) {
			stop_the_server();
			break;
		}
		handle_the_request(srv, q);
	}

	delete srv;
}

int handle_the_request(Udp_srv *srv, Udp_srv_req *q)
{
	///////////////////////////////
	// A non-blocking front-end
	//
	// enqueue the request (including server
	// handle into a task queue and
	// return
	//
//---8<------------------------------------------
	////////////////////////////////
	// The following should be considered
	// as server back-end and might be
	// executed in the context of the
	// other thread.
	//
	// dequeue a request from the task queue
	//
	// handle q and fill res and len
	//
	char *res;
	int len;

	srv.respond(q, res, len);
	delete q;
}
**************************************************/ 		

class Udp_srv {
private:
	int s;
	bool roe;
public:
	Udp_srv(int ss) {
		s=ss;
		roe=false;
	}

	Udp_srv(char *host, int port)
	{
		s=sock_udp(host, port);
		if (s>=0) 
			roe=true;
	}

	~Udp_srv()
	{
		stop();
	}

	Udp_srv_req *wait_for_request();
	int respond(Udp_srv_req *q, void *res, int len);
	char *host();
	int port();
	void stop() {
		int ls=s;
		if ((ls>=0)&&(cas32s(&s, ls, -1)==1)) {
			if (roe)
				close(ls);
		}
	}

	bool good() { return (s>=0); }
};

# endif
