// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: par_disk.cc,v $ - parallel file reader, demonstration
 *            of the thread-pool module. The program launches a set of threads
 *            and read a set of files. The goal is to see if increasing 
 *            the concurrency would increase better disk read throughput.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/08/07 06:29:01 $
 * Version :  $Revision: 1.9 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 
#define _XOPEN_SOURCE 500
# include <sys/types.h>
# include <sys/stat.h>
# include <fcntl.h>
# include <stdio.h>
# include <stdlib.h>
# include <unistd.h>
# include <getopt.h>
# include <dirent.h>
# include <sched.h>
# include <time.h>
# include <queue>

# include "../machine/atomic.h"
# include "../debug/tmpi_debug.h"
# include "../usc/usc_time.h"
// # include "pthread.h"
# include "thread_pool.h"
# include "../res/resmon.h"

/**
 * A program that reads a bounch of files specified in the command line
 * in parallel.
 * par_disk -n <num> -l <log> -f <fcache purger dir> <list...>
 */

struct SHARED {
	// pthread_mutex_t lock;
	ino_t log_inode;
	dev_t log_dev;
	//ino_t purger_inode;
	//dev_t purger_dev;
	int n_bytes;
	int phy_mem;

//-->8---------------------------------------//
	// SHARED(ino_t li, dev_t ld, ino_t pi, dev_t pd) :
	SHARED(ino_t li, dev_t ld) :
		log_inode(li), 
		log_dev(ld), 
		// purger_inode(pi), 
		// purger_dev(pd), 
		n_bytes(0),
		phy_mem(0)
	{ 
		ResMon::MEM s;
		if (ResMon::get_mem(&s)<0) {
			fprintf(stderr, "Error get_mem resource samples.\n");
			exit(-1);
		}
		phy_mem=s.mtotal*1024;
	// 	pthread_mutex_init(&lock, NULL); 
	}

	~SHARED()
	{
	// 	pthread_mutex_destroy(&lock);
	}
};

static int get_inode(char *f, ino_t &i, dev_t &d)
{
	struct stat st;

	if (stat(f, &st)<0) {
		return -1;
	}

	i=st.st_ino;
	d=st.st_dev;
	return 0;
}

static char *handler(void *ss, void *req, int myid)
{
	struct SHARED *state=(struct SHARED *)ss;

	if (state->n_bytes>state->phy_mem*2) {
		free(req);
		return NULL;
	}

	char *file_in=(char *)req;
	ino_t in_inode;
	dev_t in_device;

	char *msg=(char *)malloc(1024);

	if (get_inode(file_in, in_inode, in_device)<0) {
		snprintf(msg, 1024, "Cannot get inode for file: %s.", file_in);
		free(req);
		return msg;
	}

	/**
	 * skip the output file and log file.
	 */
	if ( // ((in_inode==state->purger_inode)&&(in_device==state->purger_dev)) ||
	     ((in_inode==state->log_inode)&&(in_device==state->log_dev)) ) {
		free(msg);
		free(req);
		return NULL;
	}

	int input=open(file_in, O_RDONLY);
	if (input<0) {
		snprintf(msg, 1024, "Cannot open file: %s.", file_in);
		free(req);
		return msg;
	}

	char block[4096];
	int cnt;
	int size=0;

	while ((cnt=pread(input, block, 4096, size*32))>0) {
		size+=cnt;
	}

	close(input);
	aadd32(&(state->n_bytes), size);

	free(msg);
	free(req);

	return NULL;
}

static char *prog=NULL;

static void do_help()
{
	printf("\
Parallel disk reader.\n\
Usage: %s [options] [files...].\n\
   -n <# threads>         The # of consumer threads.\n\
   -f <file cache purger> A file used to purge OS file cache.\n\
   -l <logfile>           Log file.\n\
   -h                     Help, this screen.\n\
\n\
This program recursively (no symbolic link follow-up) searches through\n\
the directories and files specified in the command-line argument, and\n\
reads these files in parallel using a fixed thread pool. It reports the\n\
total disk read throughput in the end.\n\
", prog);
	exit(0);
}

static int get_mode(char *f, mode_t &m)
{
	struct stat st;
	if (lstat(f, &st)<0)
		return -1;

	m=st.st_mode;

	return 0;
}

static void traverse(Thread_Pool *tp, char *f)
{
	mode_t m;

	if (get_mode(f, m)<0)
		return;

	if (S_ISREG(m)) {
		char *req=strdup(f);

		if (req) {
			tp->enqueue(req);
		}
	}
	else if (S_ISDIR(m)) {
		DIR *dir=opendir(f);
		if (!dir)
			return;
		struct dirent *de=NULL;
		std::queue<char *> dirlist;
		while ((de=readdir(dir))) {
			if ((de->d_name[0]=='.')&&(de->d_name[1]=='\0'))
				continue;

			if ((de->d_name[0]=='.')&&(de->d_name[1]=='.')&&(de->d_name[2]=='\0'))
				continue;

			char *d_name=strdup(de->d_name);
			if (d_name) {
				dirlist.push(d_name);
			}
		}
		closedir(dir);

		char ff[1024];
		while (!dirlist.empty()) {
			char *d_name=dirlist.front();
			dirlist.pop();
			snprintf(ff, 1024, "%s/%s", f, d_name);
			free(d_name);
			traverse(tp, ff);
		}
	}

	return;
}

void purge_file_cache(char *fc_purger)
{
	int phy_mem=0;
	ResMon::MEM s;
	if (ResMon::get_mem(&s)<0) {
		fprintf(stderr, "Error get_mem resource samples.\n");
		exit(-1);
	}
	phy_mem=s.mtotal*1024;

	int size=0;
	int update=0;
	std::queue<char *> filelist;

	filelist.push(strdup(fc_purger));

	while (!filelist.empty()) {
		char *file=filelist.front();
		filelist.pop();

		mode_t m;
		if (get_mode(file, m)<0) {
			free(file);
			continue;
		}

		if (S_ISREG(m)) {
			FILE *f_purger=fopen(file, "r");
			free(file);

			if (!f_purger) {
				continue;
			}

			while ((!feof(f_purger)) && (size<phy_mem)) {
				if (update%1000==0) {
					fprintf(stderr, "\rPurging file cache...%.2f%%   ", (double)size/phy_mem*100);
					fflush(stderr);
				}
				char garbage[4096];
				int cnt=fread(garbage, 1, 4096, f_purger);
				size+=cnt;
				update++;
			}
			fclose(f_purger);
			if (size>=phy_mem)
				break;
		}
		else if (S_ISDIR(m)) {
			DIR *dir=opendir(file);

			if (!dir) {
				free(file);
				continue;
			}

			struct dirent *de=NULL;
			while ((de=readdir(dir))) {
				if ((de->d_name[0]=='.')&&(de->d_name[1]=='\0'))
					continue;

				if ((de->d_name[0]=='.')&&(de->d_name[1]=='.')&&(de->d_name[2]=='\0'))
					continue;

				char ff[1024];
				snprintf(ff, 1024, "%s/%s", file, de->d_name);
				filelist.push(strdup(ff));
			}
			closedir(dir);
			free(file);
		}
	}

	if (size<phy_mem) {
		fprintf(stderr, "\nWarning: File cache purger too small.\n");
	}
	else {
		fprintf(stderr, "\rPurging file cache...done    \n");
	}

	return;
}

int main(int argc, char *argv[])
{
	prog=argv[0];

	int n_thr=5;
	char *logfile=NULL;
	char *fc_purger=NULL;

	int c;
	while ((c=getopt(argc, argv, "n:l:f:h"))!=EOF) {
		switch (c) {
			case 'n':
				n_thr=atoi(optarg);
				break;
			case 'l':
				logfile=optarg;
				break;
			case 'f':
				fc_purger=optarg;
				break;
			case 'h':
				do_help();
				break;
			default:
				do_help();
				break;
		}
	}

	if (!fc_purger) {
		fprintf(stderr, "File cache purger directory not specified.\n");
		do_help();
	}

	purge_file_cache(fc_purger);

	if ((argc-optind)<=0) {
		fprintf(stderr, "Nothing specified in the command.\n");
		do_help();
	}

	FILE *f_log=stderr;
	ino_t i_log=(ino_t)0;
	dev_t d_log=(dev_t)0;

	if (logfile) {
		f_log=fopen(logfile, "w");

		if (!f_log) {
			fprintf(stderr, "Cannot open log file: %s.\n", logfile);
			exit(-1);
		}

		if (get_inode(logfile, i_log, d_log)<0) {
			fprintf(stderr, "Cannot get inode for file: %s.\n", logfile);
			exit(-1);
		}

		tmpi_set_output(f_log);
	}

	// struct SHARED state(i_log, d_log, i_purger, d_purger);
	struct SHARED state(i_log, d_log);
	Thread_Pool tp(NULL, n_thr, &state, &handler);
	// Thread_Pool tp(1, 1, n_thr, &state, &handler);

	for (; optind<argc; optind++) {
		traverse(&tp, argv[optind]);
	}

	tp.enqueue(NULL);

	system("cat /proc/stat");
	tp.start();

	usc_time_t usc_start=usc_gettime();

	tp.join();

	system("cat /proc/stat");

	if (f_log!=stderr) {
		fclose(f_log);
	}

	usc_time_t usc_end=usc_gettime();
	double total_ms=usc_difftime(usc_end, usc_start)/1000;
	double read_bw=state.n_bytes/total_ms/1000;

	printf("Time: %.3fms, Read-BW: %.3fMB/s.\n", total_ms, read_bw);

	return 0;
}
