// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: par_disk2.cc,v $ - parallel file reader, demonstration
 *            of the thread-pool module. The program launches a set of threads
 *            and read a set of files. The goal is to see if increasing 
 *            the concurrency would increase better disk read throughput.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/08/07 06:29:01 $
 * Version :  $Revision: 1.6 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 
# include <sys/types.h>
# include <sys/stat.h>
# include <sys/vfs.h>
# include <sys/mman.h>
# include <fcntl.h>
# include <stdio.h>
# include <stdlib.h>
# include <unistd.h>
# include <getopt.h>
# include <dirent.h>
# include <sched.h>
# include <time.h>
# include <queue>

# include "../machine/atomic.h"
# include "../debug/tmpi_debug.h"
# include "../usc/usc_time.h"
// # include "pthread.h"
# include "thread_pool.h"
# include "../res/resmon.h"

/**
 * A program that randomly accesses an mmapped file for reading in
 * parallel.
 * par_disk2 -n <num> -l <log> -f <fcache purger dir> <mmap_file>
 */

struct SHARED {
	unsigned int n_bytes;
	unsigned phy_mem;
	char *addr;
	unsigned fs_bsize;
	unsigned f_size;
	unsigned int sum;
	unsigned int io_lb, io_ub;

//-->8---------------------------------------//
	SHARED(char *p, unsigned io_size, unsigned max_size, unsigned min_io, unsigned max_io) :
		n_bytes(0),
		phy_mem(0),
		addr(p),
		fs_bsize(io_size),
		f_size(max_size),
		sum(0),
        io_lb(min_io),
        io_ub(max_io)
	{ 
		ResMon::MEM s;
		if (ResMon::get_mem(&s)<0) {
			fprintf(stderr, "Error get_mem resource samples.\n");
			exit(-1);
		}
		phy_mem=(unsigned)(s.mtotal*1024);
		if (phy_mem*2>f_size) {
			fprintf(stderr, "Warning: mapped file size too small.\n");
		}
		printf("phy_mem=%dMB, iosize=%dbytes, file_size=%dMB\n", 
			phy_mem/1024/1024, fs_bsize, f_size/1024/1024);
	}

	~SHARED()
	{
	}
};

static char *handler(void *ss, void *req, int myid)
{
    // int myid=(int)req-1;
	struct SHARED *state=(struct SHARED *)ss;
	unsigned iosize, phy_mem, page_max, min_io, max_io;
	char *start=state->addr;
	iosize=state->fs_bsize;
	phy_mem=state->phy_mem;
	page_max=state->f_size/iosize;
    min_io=state->io_lb;
    max_io=state->io_ub;

	char buf[iosize*64];
	unsigned sum=0;
	while (state->n_bytes<(2*phy_mem)) {
        if (myid==0) {
            double pcent=(double)state->n_bytes/(phy_mem*2)*100;
            if (pcent>100) pcent=100;
			fprintf(stderr, "\rTesting read...%.2f%%   ", pcent);
			fflush(stderr);
        }
		int page_cnt=0;
		while (page_cnt<1024) {
			unsigned int start_page=(unsigned)random()%page_max;
			unsigned int read_range=(unsigned)random()%(max_io-min_io+1)+min_io; /* 1 to 64 pages */
			if ((start_page+read_range)>page_max) {
				read_range=page_max-start_page;
			}
			char *start2=start+start_page*iosize;
            memcpy(buf, start2, iosize*read_range);
		    sum+=(unsigned)(buf[0]+buf[iosize-1]);
			// for (unsigned i=0; i<read_range; i++, start2+=iosize) {
				// memcpy(buf, start2, iosize);
				// sum+=(unsigned)(buf[0]+buf[iosize-1]);
			// }
			page_cnt+=read_range;
		}
		aadd32(&(state->n_bytes), page_cnt*iosize);
		aadd32(&(state->sum), sum);
	}

    if (myid==0) {
		fprintf(stderr, "\rTesting read...done   \n");
    }

	return NULL;
}

static char *prog=NULL;

static void do_help()
{
	printf("\
Parallel disk reader.\n\
Usage: %s [options] [mmap file].\n\
   -n <# threads>         The # of consumer threads.\n\
   -f <file cache purger> A directory tree used to purge OS file cache.\n\
   -l <logfile>           Log file.\n\
   -h                     Help, this screen.\n\
\n\
This program launches a set of threads that randomly access an mmapped\n\
file.  It reports the total disk read throughput in the end.\n\
", prog);
	exit(0);
}

static int get_mode(char *f, mode_t &m)
{
	struct stat st;
	if (lstat(f, &st)<0)
		return -1;

	m=st.st_mode;

	return 0;
}

void purge_file_cache(char *fc_purger)
{
	int phy_mem=0;
	ResMon::MEM s;
	if (ResMon::get_mem(&s)<0) {
		fprintf(stderr, "Error get_mem resource samples.\n");
		exit(-1);
	}
	phy_mem=s.mtotal*1024;

	int size=0;
	int update=0;
	std::queue<char *> filelist;

	filelist.push(strdup(fc_purger));

	while (!filelist.empty()) {
		char *file=filelist.front();
		filelist.pop();

		mode_t m;
		if (get_mode(file, m)<0) {
			free(file);
			continue;
		}

		if (S_ISREG(m)) {
			FILE *f_purger=fopen(file, "r");
			free(file);

			if (!f_purger) {
				continue;
			}

			while ((!feof(f_purger)) && (size<phy_mem)) {
				if (update%1000==0) {
					fprintf(stderr, "\rPurging file cache...%.2f%%   ", (double)size/phy_mem*100);
					fflush(stderr);
				}
				char garbage[4096];
				int cnt=fread(garbage, 1, 4096, f_purger);
				size+=cnt;
				update++;
			}
			fclose(f_purger);
			if (size>=phy_mem)
				break;
		}
		else if (S_ISDIR(m)) {
			DIR *dir=opendir(file);

			if (!dir) {
				free(file);
				continue;
			}

			struct dirent *de=NULL;
			while ((de=readdir(dir))) {
				if ((de->d_name[0]=='.')&&(de->d_name[1]=='\0'))
					continue;

				if ((de->d_name[0]=='.')&&(de->d_name[1]=='.')&&(de->d_name[2]=='\0'))
					continue;

				char ff[1024];
				snprintf(ff, 1024, "%s/%s", file, de->d_name);
				filelist.push(strdup(ff));
			}
			closedir(dir);
			free(file);
		}
	}

	if (size<phy_mem) {
		fprintf(stderr, "\nWarning: File cache purger too small.\n");
	}
	else {
		fprintf(stderr, "\rPurging file cache...done    \n");
	}

	return;
}

static char *mmap_file(char *path, unsigned *iosize, unsigned *fsize)
{
	struct stat st;
	if (stat(path, &st)<0) {
		return NULL;
	}
	int page_sz=getpagesize();
	*fsize=st.st_size/page_sz*page_sz;

	struct statfs stfs;
	if (statfs(path, &stfs)<0) {
		return NULL;
	}
	*iosize=stfs.f_bsize;

	char *ret=NULL;

	int fd=open(path, O_RDONLY);
	if (fd<0)
		return NULL;

	ret= (char *)mmap(NULL,  *fsize, PROT_READ, MAP_SHARED, fd, 0);

	close(fd);

	if (ret==MAP_FAILED) {
		perror("mmap");
		return NULL;
	}

    madvise(ret, *fsize, MADV_RANDOM);

	return ret;
}

int main(int argc, char *argv[])
{
	prog=argv[0];

	int n_thr=5;
	char *logfile=NULL;
	char *fc_purger=NULL;
    unsigned min_io=1, max_io=64;

	int c;
	while ((c=getopt(argc, argv, "n:l:f:b:B:h"))!=EOF) {
		switch (c) {
			case 'n':
				n_thr=atoi(optarg);
				break;
			case 'l':
				logfile=optarg;
				break;
			case 'f':
				fc_purger=optarg;
				break;
			case 'b':
				min_io=atoi(optarg);
				break;
			case 'B':
				max_io=atoi(optarg);
				break;
			case 'h':
				do_help();
				break;
			default:
				do_help();
				break;
		}
	}

	if (!fc_purger) {
		fprintf(stderr, "File cache purger directory not specified.\n");
		do_help();
	}

    if (min_io>max_io) {
		fprintf(stderr, "Minimum read size is greater than maximum read size.\n");
        do_help();
    }

	if ((argc-optind)<=0) {
		fprintf(stderr, "Nothing specified in the command.\n");
		do_help();
	}

	if ((argc-optind)>1) {
		fprintf(stderr, "More than one mmap file specified in the command.\n");
		do_help();
	}

	FILE *f_log=stderr;

	if (logfile) {
		f_log=fopen(logfile, "w");

		if (!f_log) {
			fprintf(stderr, "Cannot open log file: %s.\n", logfile);
			exit(-1);
		}

		tmpi_set_output(f_log);
	}

	unsigned int iosize, fsize;
	char *addr;

	if ((addr=mmap_file(argv[optind], &iosize, &fsize))==NULL) {
		fprintf(stderr, "Cannot mmap file:%s\n", argv[optind]);
		exit(-1);
	}

	purge_file_cache(fc_purger);

	struct SHARED state(addr, iosize, fsize, min_io, max_io);
	Thread_Pool tp(NULL, n_thr, &state, &handler);

	for (int i=0; i<n_thr; i++) {
		tp.enqueue((void *)(i+1));
	}

	tp.enqueue(NULL);

	usc_time_t usc_start=usc_gettime();
	srandom((unsigned)time(NULL));

	tp.start();

	tp.join();

	usc_time_t usc_end=usc_gettime();

	if (f_log!=stderr) {
		fclose(f_log);
	}

	double total_ms=usc_difftime(usc_end, usc_start)/1000;
	double read_bw=state.n_bytes/total_ms/1000;

	printf("Time: %.3fms, Read-BW: %.3fMB/s.\n", total_ms, read_bw);

	return state.sum;
}
