// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: par_md5.cc,v $ - parallel MD5 checksum calculator, demonstration
 *            of the thread-pool module.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/08/07 06:29:01 $
 * Version :  $Revision: 1.11 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 
# include <sys/types.h>
# include <sys/stat.h>
# include <stdio.h>
# include <stdlib.h>
# include <unistd.h>
# include <getopt.h>
# include <dirent.h>
# include <sched.h>
# include <time.h>
# include <queue>

# include "../md5/md5.h"
# include "../debug/tmpi_debug.h"
# include "../usc/usc_time.h"
# include "pthread.h"
# include "thread_pool.h"

/**
 * A dummy program that calculates MD5 of the files 
 * specified in the command line.
 * par_md5 -n <num> -o <output> <list...>
 */

struct SHARED {
	FILE *fp;
	pthread_mutex_t lock;
	ino_t out_inode;
	dev_t out_dev;
	ino_t log_inode;
	dev_t log_dev;

//-->8---------------------------------------//
	SHARED(FILE *of, ino_t oi, dev_t od, ino_t li, dev_t ld) :
		fp(of), out_inode(oi), out_dev(od), log_inode(li), log_dev(ld)
	{ 
		pthread_mutex_init(&lock, NULL); 
	}

	~SHARED()
	{
		pthread_mutex_destroy(&lock);
	}
};

static int get_inode(char *f, ino_t &i, dev_t &d)
{
	struct stat st;

	if (stat(f, &st)<0) {
		return -1;
	}

	i=st.st_ino;
	d=st.st_dev;
	return 0;
}

static char *handler(void *ss, void *req, int myid)
{
	struct SHARED *state=(struct SHARED *)ss;
	char *file_in=(char *)req;
	ino_t in_inode;
	dev_t in_device;

	char *msg=(char *)malloc(1024);

	if (get_inode(file_in, in_inode, in_device)<0) {
		snprintf(msg, 1024, "Cannot get inode for file: %s.", file_in);
		free(req);
		return msg;
	}

	/**
	 * skip the output file and log file.
	 */
	if ( ((in_inode==state->out_inode)&&(in_device==state->out_dev)) ||
	     ((in_inode==state->log_inode)&&(in_device==state->log_dev)) ) {
		free(msg);
		free(req);
		return NULL;
	}

	FILE *input=fopen(file_in, "r");
	if (!input) {
		snprintf(msg, 1024, "Cannot open file: %s.", file_in);
		free(req);
		return msg;
	}

	Md5 md5;
	char block[512];
	int cnt;

	while ((cnt=fread(block, 1, 512, input))>0) {
		md5.stuff(block, cnt);
	}

	fclose(input);

	md5.close();
	unsigned char digest[MD5_LEN];
	md5.pack(digest);
	char digest_str[100];
	unsigned int *ints=(unsigned int *)digest;
	sprintf(digest_str, "%08X%08X%08X%08X", ints[0], ints[1], ints[2], ints[3]);
	pthread_mutex_lock(&(state->lock));
	fprintf(state->fp, "%s -> %s\n", digest_str, file_in);
	pthread_mutex_unlock(&(state->lock));

	free(msg);
	free(req);
	return NULL;
}

static char *prog=NULL;

static void do_help()
{
	printf("\
Parallel MD5 file checksum calculation.\n\
Usage: %s [options] [list].\n\
   -n <# threads>         The # of consumer threads.\n\
   -o <output>            Output file.\n\
   -l <logfile>           Log file.\n\
   -t <time>              Running time-out (in seconds).\n\
   -h                     Help, this screen.\n\
\n\
This program recursively (no symbolic link follow-up) searches through\n\
the directories and files specified in the command-line argument, and\n\
calculates the MD5 checksum of the files and write the result in to an\n\
output file.\n\
", prog);
	exit(0);
}

static int get_mode(char *f, mode_t &m)
{
	struct stat st;
	if (lstat(f, &st)<0)
		return -1;

	m=st.st_mode;

	return 0;
}

static void traverse(Thread_Pool *tp, char *f, time_t start_t, int to, int &cnt)
{
	if (cnt<0)
		return;

	mode_t m;

	if (get_mode(f, m)<0)
		return;

	if (S_ISREG(m)) {
		char *req=strdup(f);

		if (req) {
			tp->enqueue(req);
			if (to>0) {
				cnt=(cnt+1)%100;
				if (cnt==0) {
					time_t now;
					time(&now);
					if (((now-start_t)>to)) {
						cnt=-1;
						return;
					}
				}
			}
			sched_yield();
		}
	}
	else if (S_ISDIR(m)) {
		DIR *dir=opendir(f);
		if (!dir)
			return;
		struct dirent *de=NULL;
		std::queue<char *> dirlist;
		while ((de=readdir(dir))) {
			if ((de->d_name[0]=='.')&&(de->d_name[1]=='\0'))
				continue;

			if ((de->d_name[0]=='.')&&(de->d_name[1]=='.')&&(de->d_name[2]=='\0'))
				continue;

			char *d_name=strdup(de->d_name);
			if (d_name) {
				dirlist.push(d_name);
			}
		}
		closedir(dir);

		char ff[1024];
		while (!dirlist.empty()) {
			char *d_name=dirlist.front();
			dirlist.pop();
			snprintf(ff, 1024, "%s/%s", f, d_name);
			free(d_name);
			if (to>0) {
				cnt=(cnt+1)%100;
				if (cnt==0) {
					time_t now;
					time(&now);
					if (((now-start_t)>to)) {
						cnt=-1;
						return;
					}
				}
			}
			traverse(tp, ff, start_t, to, cnt);
			if (cnt<0)
				return;
		}
	}

	return;
}

int main(int argc, char *argv[])
{
	prog=argv[0];

	int n_thr=5;
	char *output="md5.out";
	char *logfile=NULL;
	int to=10;

	int c;
	while ((c=getopt(argc, argv, "n:o:l:t:h"))!=EOF) {
		switch (c) {
			case 'n':
				n_thr=atoi(optarg);
				break;
			case 'o':
				output=optarg;
				break;
			case 'l':
				logfile=optarg;
				break;
			case 't':
				to=atoi(optarg);
				break;
			case 'h':
				do_help();
				break;
			default:
				do_help();
				break;
		}
	}

	if ((argc-optind)<=0) {
		fprintf(stderr, "Nothing specified in the command.\n");
		do_help();
	}

	FILE *f_output=fopen(output, "w");

	if (!f_output) {
		fprintf(stderr, "Cannot open output file: %s.\n", output);
		exit(-1);
	}

	ino_t i_output;
	dev_t d_output;
	if (get_inode(output, i_output, d_output)<0) {
		fprintf(stderr, "Cannot get inode for file: %s.\n", output);
		exit(-1);
	}

	FILE *f_log=stderr;
	ino_t i_log=(ino_t)0;
	dev_t d_log=(dev_t)0;

	if (logfile) {
		f_log=fopen(logfile, "w");

		if (!f_log) {
			fprintf(stderr, "Cannot open log file: %s.\n", logfile);
			exit(-1);
		}

		if (get_inode(logfile, i_log, d_log)<0) {
			fprintf(stderr, "Cannot get inode for file: %s.\n", logfile);
			exit(-1);
		}

		tmpi_set_output(f_log);
	}

	usc_time_t usc_start=usc_gettime();

	time_t start_t;
	time(&start_t);

	struct SHARED state(f_output, i_output, d_output, i_log, d_log);
	Thread_Pool tp(NULL, n_thr, &state, &handler);
	// Thread_Pool tp(1, 1, n_thr, &state, &handler);
	tp.start();

	int cnt=0;
	for (; optind<argc; optind++) {
		traverse(&tp, argv[optind], start_t, to, cnt);
	}

	tp.enqueue(NULL);

	while (tp.is_running()) {
		if (to>0) {
			time_t now;
			time(&now);
			if (((now-start_t)>to)) {
				tp.stop();
				to=now-start_t+3; // add three seconds graceful period.
				// break; // it would be the same without the break.
			}
		}
		usleep(1);
	}

	tp.join();

	/**
	if (cnt>=0) {
		while (!tp.empty()) {
			time_t now;
			usleep(1);
			time(&now);
			if ((to>0)&&((now-start_t)>to))
				break;
		}
	}

	if (to>0) {
		tp.stop();
		tp.finalize();
	}
	else {
		tp.join();
	}
	**/

	fclose(f_output);

	if (f_log!=stderr) {
		fclose(f_log);
	}

	usc_time_t usc_end=usc_gettime();

	printf("Total execution time: %.3fms.\n", usc_difftime(usc_end, usc_start)/1000);

	return 0;
}
