// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.

   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: task_queue.cc,v $ - mt-safe task queue.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2003/08/11 23:40:10 $
 * Version :  $Revision: 1.6 $
 *
 * ISC project contact: htang@cs.ucsb.edu,  tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */ 
# include <sys/time.h>
# include <unistd.h>
# include <pthread.h>
# include <math.h>
# include "../machine/atomic.h"
# include "../sim/gsldistrib.h"
# include "task_queue.h"

Task_Queue::Task_Queue(int idle_qlen, int work_batch) :
	a(idle_qlen),
	b(work_batch),
	stopped(0),
	adm_control(true)
{
	if (a<0)
		a=TQ_IDLE_QLEN;

	if (b<=0)
		b=TQ_WORK_BATCH;

	pthread_mutex_init(&lock, NULL);
	pthread_cond_init(&cond, NULL);
}

Task_Queue::Task_Queue(bool admission) :
	a(TQ_IDLE_QLEN),
	b(TQ_WORK_BATCH),
	stopped(0),
	adm_control(admission)
{
	pthread_mutex_init(&lock, NULL);
	pthread_cond_init(&cond, NULL);
}

Task_Queue::~Task_Queue()
{
	pthread_mutex_destroy(&lock);
	pthread_cond_destroy(&cond);
}

bool Task_Queue::enqueue(void *item)
{
	if (stopped)
		return false;

	if (adm_control) {
		int qlen=q.size();
		double adm=exp((double)(a-qlen)/b);
		double tt=gsldd_flat(0, 1);
		if (tt>adm) {
			/* failed the admission test */
			return false;
		}
	}

	if (pthread_mutex_lock(&lock)<0)
		return false;

	if (stopped) {
		pthread_mutex_unlock(&lock);
		return false;
	}

	q.push_back(item);

	if (q.size()==1) {
		/* used to be empty */
		/* might be somebody waiting */
		pthread_cond_signal(&cond);
	}

	pthread_mutex_unlock(&lock);

	return true;
}

bool Task_Queue::prio_enqueue(void *item)
{
	if (stopped)
		return false;

# if (0)
// We ignore admission control for priority items.
	if (adm_control) {
		int qlen=q.size();
		double adm=exp((double)(a-qlen)/b);
		double tt=gsldd_flat(0, 1);
		if (tt>adm) {
			/* failed the admission test */
			return false;
		}
	}
# endif

	if (pthread_mutex_lock(&lock)<0)
		return false;

	if (stopped) {
		pthread_mutex_unlock(&lock);
		return false;
	}

	q.push_front(item);

	if (q.size()==1) {
		/* used to be empty */
		/* might be somebody waiting */
		pthread_cond_signal(&cond);
	}

	pthread_mutex_unlock(&lock);

	return true;
}

bool Task_Queue::dequeue(void *&item)
{
	if (stopped)
		return false;

	if (pthread_mutex_lock(&lock)<0)
		return false;

	do {
		if (stopped) {
			pthread_mutex_unlock(&lock);
			return false;
		}

		if (q.empty()) {
			struct timespec ts;
			struct timeval tv;

			/**
			 * set time-out to be 3 seconds
			 */
			gettimeofday(&tv, NULL);
			ts.tv_sec=tv.tv_sec+3;
			ts.tv_nsec=tv.tv_usec*1000;

			// should be safe even if we do pthread_cond_wait.
			pthread_cond_timedwait(&cond, &lock, &ts);
		}
		else {
			item=q.front();
			q.pop_front();

			/* might be somebody behind me waiting */
			if (!q.empty())
				pthread_cond_signal(&cond);

			pthread_mutex_unlock(&lock);
			return true;
		}
	} while (1);
}

bool Task_Queue::stop()
{ 
	if (pthread_mutex_lock(&lock)<0)
		return false;

	stopped=1; 
	commit32(&stopped); 

	pthread_cond_broadcast(&cond);

	pthread_mutex_unlock(&lock);

	return true;
}
