// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.
 
  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
 
   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.
 
   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
 
   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.
 
   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.
 
  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: block_map.cc,v $ - Block map file shows which version
 *            should be looked at for accessing a version, also the first 4
 *            bytes specify length of the corresponding segment.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Aziz Gulbeden, gulbeden@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2004/04/10 22:44:34 $
 * Version :  $Revision: 1.9 $ by $Author: gulbeden $
 *
 * ISC project contact: htang@cs.ucsb.edu, tyang@cs.ucsb.edu, 
 *                      gulbeden@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */

#include "block_map.h"
#include "../fcache/file_cache.h"
#include "../debug/tmpi_debug.h"
#include <math.h>				// ceil

BlockMap::BlockMap(char* bmf, File_Cache* fc) : 
	bmFileName(bmf),
	fcache(fc),
	mapFile(NULL),
	lastBlock(0),
	segmentLen(0),
	version(0)
{
	// read last block if file exists
	// first integer in the file is the length of the corresponding object

	mapFile = fopen(bmFileName, "r");
	
	if (mapFile==NULL){
		// the file is newly created,
		int zero = 0;
		mapFile = fopen(bmFileName, "w");
		int n = fwrite((char *)(& zero), 1, 4, mapFile); //length
		n = fwrite((char *)(& zero), 1, 4, mapFile); //version
		if (n<0){
			throw -1;
			tmpi_error(DBG_INTERNAL, "Write to block-map failed\n");
		}
	}
	
	else {
		int n = fread((char *)(& segmentLen), 1, 4, mapFile); 
		fread((char *)(& version), 1, 4, mapFile); 
		if (n<0){
			throw -1;
			tmpi_error(DBG_INTERNAL, "Read from block-map failed\n");
		}
		lastBlock = (segmentLen+BLOCK_SIZE-1)/BLOCK_SIZE;
		//tmpi_error(DBG_INTERNAL, "%d\n", segmentLen);
	}

	fclose(mapFile);
	
}


BlockMap::~BlockMap(){ 
	if(bmFileName != 0) 
		delete bmFileName; 
}


int BlockMap::flush(){
	if(bmFileName ==NULL)
		return -1;
	return fcache->flush(bmFileName);
}


int BlockMap::initializeBlocks(unsigned int lower, unsigned int higher){

	for(unsigned int i=lower; i<higher; i+=sizeof(int)){
		unsigned int ui = 0;
		char* c = (char *)(&ui);
		fcache->write(bmFileName, i, c, sizeof(int));
	}
	
	return 0;
}


int BlockMap::readRange(unsigned int offset, unsigned int len, 
						unsigned int * ret){
	int begOffset = offset - offset % BLOCK_SIZE;
	int totLen = len + offset % BLOCK_SIZE;

	for(int i=begOffset / BLOCK_SIZE;
		i< (ceil( ((float) (begOffset + totLen)) / (float)BLOCK_SIZE));
		i ++) 
	{
		// go to the corresponding offset
		char* c = (char *)(& ret[i - begOffset/BLOCK_SIZE]);
		
		if(i>=lastBlock){
			// block has not been not modified, 
			// so information not kept, forward request to
			// version 0, but this will result in trying to read a range
			// outside file's range.
			int zero = 0;
			memcpy (c, &zero, sizeof(int));
		}
		
		else if(!fcache || fcache->read(bmFileName, blockLoc(i), c, \
										 sizeof(int)) < 0) {
			tmpi_error(DBG_INTERNAL, "Block-map failed.\n");
			return -1;
		}
   	}

	return 0;
}


int BlockMap::updateBlock(int blockNum, int version){
	unsigned int ui = version;
	int stat;
	char* c = (char *)(&ui);

	if(blockNum>=lastBlock){
		initializeBlocks(blockLoc(lastBlock), blockLoc(blockNum+1));
		lastBlock= blockLoc(blockNum+1);
	}

	stat = fcache->write(bmFileName, blockLoc(blockNum), c, sizeof(int));
	if (stat < 0) return -1;

	return 0;
}


int BlockMap::updateLength(int newLen){
	int stat;

	stat = fcache->write(bmFileName, 0, (char *)(&newLen), sizeof(int));

	if (stat < 0) return -1;

	segmentLen = newLen;

	return 0;
}

int BlockMap::updateVersion(int newVer){
	int stat;

	stat = fcache->write(bmFileName, 1*sizeof(int),(char *)(&newVer), sizeof(int));

	if (stat < 0) return -1;

	version = newVer;

	return 0;
}
