// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.
 
  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
 
   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.
 
   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
 
   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.
 
   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.
 
  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: verfs.cc,v $ -  A file system implementation that 
 *            supportd versioning through copy-on-write.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Aziz Gulbeden, gulbeden@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2004/04/19 22:09:13 $ by $Author: jzhou $
 * Version :  $Revision: 1.41 $
 *
 * ISC project contact: htang@cs.ucsb.edu, tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */

#include <stdio.h>
#include <stdlib.h> 
#include <string.h>				/* strncmp */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dirent.h>
#include <queue>

#include "verfs.h"
#include "verfs_desc.h"
#include "block_map.h"
#include "../pshash/pshash.h"
#include "../res/resmon.h"
#include "../fcache/file_cache.h"
#include "../debug/tmpi_debug.h"

using namespace std;

#define MAX_LENGTH 1024			/* maximum filename length */
#define MAX_PATH_LEN 1024

#define WD "./"

/* metadata pointed by the descriptor */
class VerfsFileInfo {
public:
	VerfsFileInfo(int p=0, int n=0, int v=0) :
		partition(p),
		numOpened(n),
		version(v)
	{}
	
	int partition;				// which partition stores the file
	int numOpened;
	int version;
};


/* < operator for guid_t - required by map */
bool operator<(const guid_t &guid1, const guid_t &guid2){
	return memcmp(&guid1, &guid2, sizeof(guid_t));
}


/**
 * Hexical to char conversion
 * used for giving names to files
 */
static char h2c(int h)
{
	if ((h<0)||(h>15)) {
		return (char)0;
	}
	if (h<10) {
		return '0'+h;
	}
	return 'A'+(h-10);
}


/**
 * Checks if the path is correctly formatted and the required directories 
 * exist.
 */
static bool check_format(char* path, bool f_mod_debug){

	bool result = true;
	//char mp[MAX_LENGTH];
	char pwd[MAX_LENGTH];

	if (strlen(path)>=MAX_PATH_LEN) {
		tmpi_error(DBG_CALLER, "Mount point path too long.\n");
		exit(-1);
	}

	if (getcwd(pwd, MAX_LENGTH)==NULL) {
		tmpi_error(DBG_CALLER, "Current directory path too long.\n");
		return false;
	}

	if (chdir(path)<0) {
		tmpi_error(DBG_CALLER, "Cannot access directory %s.\n", path);
		return false;
	}
	
	if (chdir(DIR_DATA)<0) {
		return false;
	}

	char cmd[1024];

	if (f_mod_debug){
		for(int a=0; a<16 && result; a++){
			for(int b=0; b<16 && result; b++){
				snprintf(cmd, 1024, "%c/%c", h2c(a), h2c(b));
				if (chdir(cmd)<0) {
					result = false;
					break;
				}

				snprintf(cmd, 1024, "../..");
				if (chdir(cmd)<0) {
					result = false;
					break;
				}
			}
		}
	}

	else{
		for(int a=0; a<16 && result; a++){
			for(int b=0; b<16 && result; b++){
				for(int c=0; c<16 && result; c++){
					for(int d=0; d<16 && result; d++){
						snprintf(cmd, 1024, "%c/%c/%c/%c", h2c(a), h2c(b), 
								 h2c(c), h2c(d));
						if (chdir(cmd)<0) {
							result = false;
							break;
						}
						
						snprintf(cmd, 1024, "../../../..");
						if (chdir(cmd)<0) {
							result = false;
							break;
						}
					}
				}
			}
		}
	}
	
	chdir(pwd);

	return result;
}


static int get_mode(char *f, mode_t &m){
	struct stat st;
	if (lstat(f, &st)<0)
		return -1;

	m=st.st_mode;

	return 0;
}


/**
 * appends data. FileEnd is stored by descriptor
 */
int Verfs::append(int desc, char *data, unsigned int len) {
	VerfsDescriptor * vd = check_desc(desc);
	if (vd == NULL) return -1;
	return vd->append(data, len);
}


/**
 * decides on which partition to use.
 * the one with maximum free space.
 */
int Verfs::calcPartition(guid_t &guid) {
	int maxFreeSp;
	int maxFreeInd = 0;
	time_t timeNow;
	bool status = false;
		
	time (&timeNow);

	if (timeNow - lastDIUpdate >= DI_UPDATE_INTERVAL)
		status = refreshDiskInfos();

	if (status == false || diskInfos == NULL)
		return descriptor % wDirCnt; // disk info not available, roundrobin

	maxFreeSp = diskInfos[0].bfree;

	for (int i=0; i<wDirCnt; i++){
		if (maxFreeSp > diskInfos[i].bfree){
			maxFreeSp = diskInfos[i].bfree;
			maxFreeInd = i;
		}
	}
	return maxFreeInd;
}


/**
 * Gets the corresponding VerfsDescriptor object
 * @return NULL if descriptor nor found.
 */
VerfsDescriptor * Verfs::check_desc(int desc) {
	std::map<int, VerfsDescriptor *>::iterator iter = desc2fi.find(desc);
	if (iter == desc2fi.end()) return NULL;
	return iter->second;
}


/**
 * closes the file referenced by the descriptor.
 * @param desc descriptor of the file to close
 * @param version of the segment after close operation (out parameter)
 */
int Verfs::close(int desc, int &version){

	VerfsDescriptor *vd = check_desc(desc);
	if (vd == NULL)
		return -1;

	vd->close();

	if(pthread_mutex_lock(&d2fLock) != 0)
		return -1;

	pthread_mutex_unlock(&d2fLock);

 	// store new file version in PSHash
 	if(pthread_mutex_lock(&g2pLock) != 0)
 		return -1;

 	VerfsFileInfo vfi(0,0,0);
 	unsigned int valSize = sizeof(VerfsFileInfo);

 	guid_t guid = vd->getGuid();

	if(guid2vfi->get(&guid, sizeof(guid_t), &vfi, &valSize) != 0 ){
		// file does not exist ??
		goto error;
	}

	version = vd->getVersion(); // out parameter
	vfi.version = version;

	if(guid2vfi->set(&guid, sizeof(guid_t), &vfi, valSize) != 0 ){
		// file does not exist ??
		goto error;
	}
	guid2vfi->sync(); /* sync to disk for fast recovery */

	pthread_mutex_unlock(&g2pLock);

	desc2fi.erase(desc);

	delete vd;
	return 0;

 error:;
	pthread_mutex_unlock(&g2pLock);
	tmpi_error(DBG_INTERNAL, "File does not exist. (This should never happen.)");
	return -1;	
}


/**
 * @return descriptor of the created file or -1 in case of failure.
 */
int Verfs::create(guid_t guid, int &version) {
	int pIndex, sessionId;
	VerfsFileInfo vfi(0,0,0);
	unsigned int valSize = sizeof(VerfsFileInfo);
	VerfsDescriptor *fi;

	// get locks
	if(pthread_mutex_lock(&d2fLock) != 0)
		return -1;

	if(pthread_mutex_lock(&g2pLock) != 0){
		pthread_mutex_unlock(&d2fLock);
		tmpi_error(DBG_INTERNAL, "Unable to get lock.\n");
		return -1;
	}

	// check if guid already exists
	if(guid2vfi->get(&guid, sizeof(guid_t), &vfi, &valSize) == 0 ){
		// file exists
		//tmpi_error(DBG_INTERNAL, "This GUID already exists.\n");
		goto error;
	}

	pIndex = calcPartition(guid);
	vfi.partition = pIndex;
	vfi.numOpened = 0;
	vfi.version = 0;

	if(guid2vfi->set(&guid, sizeof(guid_t), &vfi, sizeof(VerfsFileInfo))!=0){
		tmpi_error(DBG_INTERNAL, "Persistent hashtable failed.\n");
		goto error;
	}

	guid2vfi->sync();

	// file created, now open it
	version = 0;
	fi= new VerfsDescriptor(wDirs[pIndex], guid, fcache);

	pair<int, VerfsDescriptor*> *d2vfi;
	d2vfi = new pair<int, VerfsDescriptor*>(descriptor, fi);

	if(desc2fi.insert(*d2vfi).second == false){
		tmpi_error(DBG_INTERNAL, "Insert into map failed\n");
		goto error;
	}

	sessionId = 0;				// since this is the first open
	fi->openForWrite(sessionId);
	
	descriptor++;

	pthread_mutex_unlock(&g2pLock);
	pthread_mutex_unlock(&d2fLock);

	tmpi_error(DBG_VERBOSE, "created %d\n", descriptor-1);

	return descriptor-1;

 error:;
	pthread_mutex_unlock(&g2pLock);
	pthread_mutex_unlock(&d2fLock);
	return -1;
}


int Verfs::dumpObject(int desc) {
	tmpi_error(DBG_CALLER, "Dump object not implemented in C MODE");
	return 0;
}


int Verfs::flush(int desc) {
 	VerfsDescriptor * vd = check_desc(desc);
 	if (vd == NULL) return -1;
	return vd->flush();
}


/**
 * set iterator to the beginning and get first
 * @return true if successful, false if there is no first element or if 
 * failed.
 */
bool Verfs::getFirst(guid_t &guid) {
	VerfsFileInfo vfi;
	unsigned int valSize = sizeof(VerfsFileInfo);
	unsigned int keySize = sizeof(guid_t);
	
	if (guid2vfi->get_first(&guid, &keySize,
						    &vfi, &valSize) == 0) return true;
	return false;
}


/** 
 * Traverses the GUID list. 
 * @return true if operation is successful, false if operation fails or end
 * of the list is reached.
 */
bool Verfs::getNext(guid_t &guid){
	VerfsFileInfo vfi;
	unsigned int valSize = sizeof(VerfsFileInfo);
	unsigned int keySize = sizeof(guid_t);

	if (guid2vfi->get_next(&guid, &keySize, 
						   &vfi, &valSize) == 0) return true;
	return false;
}

/**
 * returns the corresponding GUID of a descriptor.
 */
guid_t Verfs::getGUID(int desc){
 	VerfsDescriptor * vd = check_desc(desc);
 	if (vd == NULL) return Guid::GUID_NIL;
	return vd->getGuid();
}


/**
 * returns -1 if file not found.
 */
int Verfs::getVersion(guid_t guid){

	unsigned int valSize = sizeof(VerfsFileInfo);
	VerfsFileInfo vfi;

	// shall I get the lock at this point?
	if(guid2vfi->get(&guid, sizeof(guid_t), &vfi, &valSize) != 0 ){
		// file does not exist ??
		return -1;
	}

	return vfi.version;
}

/**
 * common initializations among constructors.
 * @param clean Flag if last time we were cleanly showdown
 * @return true if successful
 */
bool Verfs::init(bool &clean)
{
	if (pthread_mutex_init(&d2fLock, NULL) != 0) 
		return false;
	if (pthread_mutex_init(&g2pLock, NULL) != 0)
		return false;

	char db_name[MAX_LENGTH];

	
	char flag_name[MAX_LENGTH]; /* flag if we were shutdown cleanly */

	if (homeDir == NULL) {
		sprintf(db_name, "guid2vfi.db");
		sprintf(flag_name, "startup.dirty");
	} else {
		sprintf(db_name, "%s/guid2vfi.db", homeDir);
		sprintf(flag_name, "%s/startup.dirty", homeDir);
	}

	/* check if we were cleanly shutdown */
	struct stat stat_buf;
	if ( stat(flag_name, &stat_buf) ) {
		clean = true; /* no dirty flag, good to go */

		/* if this is the first time we start the provider, set to dirty */
		if ( stat(db_name, &stat_buf) )
			clean = false;
	} else {
		clean = false;
		unlink(db_name);			// persistent hast is not reused
								// it is created again
	}

	srand(time(NULL));

	fcache = new File_Cache(800);

	if (time(&lastDIUpdate) <0)
		return false;

	if ( clean ) {
		guid2vfi = new PsHash(db_name, 100*(sizeof(guid_t)+ sizeof(VerfsFileInfo)));
		if ( !guid2vfi->is_loaded() ) {
			tmpi_error(DBG_FATAL, "Cannot load database.\n");
			return false;
		}
	} else {
		guid2vfi = 
			new PsHash(db_name, sizeof(guid_t),
					   sizeof(VerfsFileInfo),100,100*(sizeof(guid_t)+sizeof(int)));
		if (!guid2vfi->is_loaded()) {
			tmpi_error(DBG_FATAL, "guid2vfi already exists.\n");
			delete guid2vfi;
			guid2vfi = new PsHash(db_name,
								  100*(sizeof(guid_t)+ sizeof(VerfsFileInfo)));
			if (!guid2vfi->is_loaded()) {
				tmpi_error(DBG_FATAL, "Cannot load database.\n");
				return false;
			}
		}
	}

	return true;
}


/**
 * since fcache is being used, just assigns a new descriptor 
 * problem: runs out of descriptors
 * returns -1 if object not found.
 * flags : if 'w' is present in the flags file is opened in write mode, 
 * otherwise it is in read mode.
 */
int Verfs::open(guid_t guid, char *flags, int &version){
	int pIndex, sessionId;
	VerfsFileInfo vfi(0,0,0);
	unsigned int valSize = sizeof(VerfsFileInfo);
	VerfsDescriptor *fi;

	// get the locks
	if (pthread_mutex_lock(&d2fLock) != 0)
		return -1;
	if (pthread_mutex_lock(&g2pLock) != 0){
		pthread_mutex_unlock(&d2fLock);
		return -1;
	}

	if (guid2vfi->get(&guid, sizeof(guid_t), &vfi, &valSize) != 0){
		tmpi_error(DBG_VERBOSE, "GUID to open not found\n");
		goto error;
	}

	pIndex = vfi.partition;
	fi = new VerfsDescriptor(wDirs[pIndex], guid, fcache);

	pair<int, VerfsDescriptor*> *d2vfi;
	d2vfi = new pair<int, VerfsDescriptor*>(descriptor, fi);
	if (desc2fi.insert(*d2vfi).second == false){
		tmpi_error(DBG_INTERNAL, "insert into map failed\n");
		goto error;
	}

	version = vfi.version;

	if(strchr(flags, 'w') != NULL){
		sessionId = ++vfi.numOpened; // increment session number
		fi->openForWrite(sessionId);
	}

	descriptor++;

	guid2vfi->set(&guid, sizeof(guid_t), &vfi, sizeof(VerfsFileInfo));
	guid2vfi->sync();

	// release the locks
	pthread_mutex_unlock(&g2pLock);
	pthread_mutex_unlock(&d2fLock);

	return descriptor-1;

 error:;
	pthread_mutex_unlock(&g2pLock);
	pthread_mutex_unlock(&d2fLock);
	return -1; // not fnd
}


/**
 * @return ???
 */
int Verfs::read(int desc, unsigned int offset, char* buff, unsigned int len) {
	VerfsDescriptor *vd = check_desc(desc);
	
	if (vd == NULL){
		tmpi_error(DBG_VERBOSE, "Descriptor %d not found\n", desc);
		return -1;
	}
   	int ret =  vd->read(offset, buff, len);
	
	return ret;
}


bool Verfs::refreshDiskInfos(void) {
	bool success = true;
	int ndisks=0;

	if (wDirCnt>0) {
		for (int i=0; i<wDirCnt; i++) {
			ResMon::DISK *d=ResMon::get_disk(wDirs[i]);
			if ((d->good)&&(d->btotal>0)) {
				ndisks++;

				DiskInfo dsk;
				dsk.btotal=d->btotal;
				dsk.bfree=d->bavail;

				diskInfos[i] = dsk;
				delete d;
			}
			else {
				
				diskInfos[i].btotal = -1;
				diskInfos[i].bfree = -1;

				tmpi_error(DBG_VERBOSE, "Mount point not found: %s.\n", 
						   wDirs[i]);
				success = false;
			}
		}
	}
	return success;
}


/**
 * @return 0 if unlink successful, -1 otherwise
 */
int Verfs::remove(guid_t guid)
{
	int pIndex;
	int ret;
	unsigned int valSize = sizeof(VerfsFileInfo);
	VerfsFileInfo vfi;
	
	if(pthread_mutex_lock(&g2pLock)!=0)
		return -1;
	
	if( guid2vfi->get(&guid, sizeof(guid_t), &vfi, &valSize) != 0){
		tmpi_error(DBG_VERBOSE, "GUID to be deleted is not found\n");
		pthread_mutex_unlock(&g2pLock);
		return -1;				// not found
	}
	
	pIndex = vfi.partition;

	// Sending version as 0 because I don't care it will be deleted anyway
	VerfsDescriptor *fi = new VerfsDescriptor(wDirs[pIndex], guid, fcache);
	
	ret = fi->remove();
	
	// TODO : There is an unsolved issue here, what if somebody removes a
	// file while it is open?

	if (guid2vfi->del(&guid, sizeof(guid))!=0){
		tmpi_error(DBG_INTERNAL, "Delete from persistent hash failed\n");
		pthread_mutex_unlock(&g2pLock);
		return -1;
	}
	guid2vfi->sync();

	pthread_mutex_unlock(&g2pLock);
	
	if (ret != 0)
		tmpi_error(DBG_VERBOSE, "Remove failed, status: %d\n", ret);

	delete fi;

	return ret;
}


/**
 * Truncates the file
 * @return new length if successful, -1 otherwise.
 */
int Verfs::truncate(int desc, unsigned int length) {
 	VerfsDescriptor * vfi = check_desc(desc);
 	if (vfi == NULL){
		tmpi_error(DBG_INTERNAL, "Descriptor not found");
 		return -1;
	}
	
	return vfi->truncate(length);
}


/**
 * changes regarding vfi need to be done ! also about the extra files
 */
void Verfs::traverse(char *f, int partNum) {
	mode_t m;
	struct stat fStat;

	if (get_mode(f, m)<0)
		return;

	if(stat(f, &fStat) != 0){
		tmpi_error(DBG_INTERNAL, "Stat failed for %s", f);
		return;					// stat failed
	}

	//if (pthread_mutex_lock(&g2pLock) != 0)
	//	return;

	if (S_ISREG(m)) {
		char *req=strdup(f);
		char hex[5];
		if (req) {
			/* regular file : add to guid2vfi table */

			/* convert string to GUID */
			guid_t guid;

			int offset=0;		// last position of /
			for(int a=0; req[a]!='\0'; a++){
				if(req[a] == '/') offset = a;
			}

			long dec;

			for(unsigned int i=0; i<2*sizeof(guid_t); i+=2){
				sprintf(hex, "0x%c%c", req[i+offset+1], req[i+2+offset]);
				dec= strtol(hex, NULL, 16);	// convert hex to int
				guid.data[i/2] = (char)dec;
			}

			VerfsFileInfo vfi;
			unsigned int valSize = sizeof(VerfsFileInfo);
			if (guid2vfi->get(&guid, sizeof(guid_t), &vfi, &valSize) != 0){
				// first time
				vfi.numOpened = -1;// -1 considering blockmap
			}
			else{
				vfi.numOpened++;
			}

			vfi.partition = partNum;

			// get version information from the blockmap
			// THIS CODE WILL NOT WORK WITH 2 OR MORE PARTITIONS!!!!!!! FIX
			char* bmName=new char[128];
			initFName(bmName, wDirs[0], "b", guid);

			BlockMap bm (bmName, fcache);

			vfi.version = bm.getVersion();

			//int fileEnd = fStat.st_size;
			//tmpi_error(DBG_VERBOSE, "File %s\t%d\n", f, vfi.version);

			// This works fine, except it cannot determine the version number
			// therefore it is commented out which means guid to partition
			// persistent hash should not be lost.
			guid2vfi->set(&guid, sizeof(guid_t), &vfi, sizeof(VerfsFileInfo));
		}
	}

	else if (S_ISDIR(m)) {
		DIR *dir=opendir(f);
		if (!dir){
			pthread_mutex_unlock(&g2pLock);
			return;
		}
		struct dirent *de=NULL;
		queue<char *> dirlist;
		while ((de=readdir(dir))) {
			if ((de->d_name[0]=='.')&&(de->d_name[1]=='\0'))
				continue;

			if ((de->d_name[0]=='.')&&(de->d_name[1]=='.')
				&&(de->d_name[2]=='\0'))
				continue;

			char *d_name=strdup(de->d_name);
			if (d_name) {
				dirlist.push(d_name);
			}
		}
		closedir(dir);

		char ff[1024];
		while (!dirlist.empty()) {
			char *d_name=dirlist.front();
			dirlist.pop();
			snprintf(ff, 1024, "%s/%s", f, d_name);
			free(d_name);
			traverse(ff, partNum);
		}
	}

	guid2vfi->sync();

	/* remove dirty flag */
	char flag_name[MAX_LENGTH];
	if (homeDir  == NULL) {
		sprintf(flag_name, "startup.dirty");
	} else {
		sprintf(flag_name, "%s/startup.dirty", homeDir);
	}
	unlink(flag_name);

	//pthread_mutex_unlock(&g2pLock);
	return;
}


/**
 * Constructor
 */
Verfs::Verfs() :
	descriptor(0),
	fcache(NULL),
	guid2vfi(NULL),
	wDirCnt(1),
	wDirs(NULL),
	diskInfos(NULL),
	lastDIUpdate(0),
	homeDir(NULL)
{
	wDirs = new char *[1];
	wDirs[0] = ".";

	char dirName[1024];
	bool clean;

	if (!init(clean)) throw -1;

	if(check_format(wDirs[0], true) == false){
		tmpi_error(DBG_CALLER, "Directory %s not correctly formatted\n", 
				   wDirs[0]);
		throw (-1);
	}
	
	sprintf(dirName, "%s/%s", wDirs[0], DIR_DATA);

	if (pthread_mutex_lock(&g2pLock)) 
		throw -1;
	if (!clean)
		traverse(dirName, 0);
	pthread_mutex_unlock(&g2pLock);
	
	diskInfos = new DiskInfo[wDirCnt];
}


/**
 * Constructor 2
 */
Verfs::Verfs(char **wds, int cnt) : 
	descriptor(0),
	fcache(NULL),
	guid2vfi(NULL),
	wDirCnt(cnt),
	wDirs(wds),
	diskInfos(NULL),
	lastDIUpdate(0),
	homeDir(NULL)
{
	int i;
	char dirName[MAX_LENGTH];
	bool clean;

	if (!init(clean)) throw -1;

	for(i=0; i<wDirCnt; i++){
		if(check_format(wDirs[i], true) == false){
			tmpi_error(DBG_CALLER, 
					   "Directory %s not correctly formatted\n", wDirs[i]);
			throw(-1);
		}
		sprintf(dirName, "%s/%s", wDirs[i], DIR_DATA);
		if (pthread_mutex_lock(&g2pLock)) 
			throw -1;
		if (!clean)
			traverse(dirName, i);
		pthread_mutex_unlock(&g2pLock);
	}

	diskInfos = new DiskInfo[wDirCnt];
}


/**
 * Constructor 3
 * takes a NULL terminated array of working directories as parameter.
 */
Verfs::Verfs(char **wds, char* hd) :
	descriptor(0),
	fcache(NULL),
	guid2vfi(NULL),
	wDirCnt(0),
	wDirs(wds),
	diskInfos(NULL),
	lastDIUpdate(0),
	homeDir(hd)
{
	int i;
	char dirName[MAX_LENGTH];
	bool clean;

	if (!init(clean)) throw -1;

	for(i=0; wDirs[i]!=NULL; i++){
		if(check_format(wDirs[i], true) == false){
			tmpi_error(DBG_CALLER, 
					   "Directory %s not correctly formatted\n", wDirs[i]);
			throw(-1);
		}
		sprintf(dirName, "%s/%s", wDirs[i], DIR_DATA);
		if (pthread_mutex_lock(&g2pLock)) 
			throw -1;
		if (!clean)
			traverse(dirName, i);
		pthread_mutex_unlock(&g2pLock);
		wDirCnt++;
	}
	diskInfos = new DiskInfo[wDirCnt];
}


Verfs::~Verfs()
{
	guid2vfi->sync(); /* just be safe, every write should be committed already */

	delete guid2vfi;
	delete fcache;
	delete [] diskInfos;
	pthread_mutex_destroy(&d2fLock);
	pthread_mutex_destroy(&g2pLock);
}


int Verfs::write(int desc, unsigned int offset, char* data, unsigned int len)
{
	VerfsDescriptor * vd = check_desc(desc);
	if (vd == NULL) return -1;
	return vd->write(offset, data, len);
}
