// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:t -*-

/************************************************************************
  Copyright (c) 2001, 2002 The Regents of the University of California
  All rights reserved.
 
  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
 
   1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.
 
   2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
 
   3. All advertising materials mentioning features or use of this
   software must display the following acknowledgement: This product
   includes software developed by the University of California, Santa
   Barbara.
 
   4. Neither the name of the University nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.
 
  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND THE AUTHORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/**
 * Name    :  $RCSfile: verfs_desc.cc,v $ -  File desc class implementation 
              that stores information about an open session.
 * Authors :  Hong Tang, htang@cs.ucsb.edu
 *            Aziz Gulbeden, gulbeden@cs.ucsb.edu
 *            Dept. of Computer Science, UC Santa Barbara
 * Last-Mod:  $Date: 2004/03/26 22:44:31 $ by $Author: jzhou $
 * Version :  $Revision: 1.21 $
 *
 * ISC project contact: htang@cs.ucsb.edu, tyang@cs.ucsb.edu
 * Web site: http://www.cs.ucsb.edu/projects/isc
 */

#include <math.h>
#include <string>

#include "verfs_desc.h"
#include "verfs.h"
#include "block_map.h"
#include "../fcache/file_cache.h"
#include "../debug/tmpi_debug.h"

#define min(X, Y)  ((X) < (Y) ? (X) : (Y))

using namespace std;

/**
 * Hexical to char conversion
 */
char h2c(int h){
	if ((h<0)||(h>15)) {
		return (char)0;
	}
	if (h<10) {
		return '0'+h;
	}
	return 'A'+(h-10);
}

int VerfsDescriptor::append(char *buff, int len){
	return write(fileEnd, buff, len);
}


int VerfsDescriptor::close(){
	if(mode!=READ_MODE){

		version++;
		//tmpi_error(DBG_VERBOSE, "%d\n", version);
		bm->updateVersion(version);

		if(modified != NULL){ 
			// for each modified block update blockmap
			for (unsigned int i=0; i<modified->size();  i++){
				if((*modified)[i]){
					bm->updateBlock(i, sid);
				}
			}
			bm->updateLength(fileEnd);
		}
	}
	return 0;
}


int VerfsDescriptor::flush(){
	bm->flush();
	fcache->flush(copyName);
	return 0;
}


void initFName(char* c, char* wDir, char* suffix, guid_t &guid){
	if (wDir==NULL){
		tmpi_error(DBG_INTERNAL, "wDir is null.\n");
		return;
	}

	/* construct filename */
	strcpy(c, wDir);
	c[strlen(wDir)] = '/';
	
	char dirName[1024] = "";
	
	if(SCSD_DEBUG){
		for(int i=0; i<2; i++){
			sprintf(&(dirName[i*2]), "%c/", h2c(guid.data[i] % 16));
		}
	}
	
	else{						// not tested
		for(int i=0; i<4; i++){
			sprintf(&(dirName[i*2]), "%c/", h2c(guid.data[i] % 16));
		}
	}

	sprintf((&(c[strlen(wDir)+1])), "%s/%s%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X.%s", DIR_DATA, dirName,
			 guid.data[0], guid.data[1], guid.data[2], guid.data[3],
			 guid.data[4], guid.data[5], guid.data[6], guid.data[7], 
			 guid.data[8], guid.data[9], guid.data[10],guid.data[11], 
			 guid.data[12],guid.data[13],guid.data[14],guid.data[15], suffix);
}


/**
 * creates a shadow copy
 */
int VerfsDescriptor::openForWrite(int sessionID){
	sid = sessionID;
	copyName = new char[1024];
	char suffix[32];

	sprintf(suffix, "%d", sessionID);
	initFName(copyName, wd, suffix, guid);

	fcache->truncate(copyName, fileEnd);

	mode = WRITE_MODE;

	modified = new RubberVector<bool>();

	version = bm->getVersion();

	return 0;
}


/**
 * reads one or more blocks from the specified version
 * @return number of bytes read, -1 on error
 */
int 
VerfsDescriptor::readBlock(int offset, char *buff, int vers, int len)
{
	int ret;
	char verName[1024];
	char verStr[32];
	sprintf(verStr, "%d", vers);
	initFName(verName, wd, verStr, guid);
	
	if( (ret = fcache->read(verName, offset, buff, len)) < 0){
		tmpi_error(DBG_INTERNAL, "Read failed\n");	
		return -1;
	}

	return ret;
}


/**
 * reads block by block from the correct version
 * @return number of bytes read or -1 if error occured
 */
int 
VerfsDescriptor::read(int offset, char *buff, int len)
{
	int status = 0, ret;
	const unsigned base_block = offset/BLOCK_SIZE;     /* first block index */
	const unsigned last_block = (offset + len) / BLOCK_SIZE; /* last block index */
	unsigned int *versions = new unsigned int[last_block - base_block + 1];

	//tmpi_error(DBG_INTERNAL, "Read request offset %d, %d bytes\n", offset, len);
	if(bm->readRange(offset, len, versions) != 0){
		tmpi_error(DBG_INTERNAL, "Block-map failed.");
		delete versions;
		return -1;				// blockmap has to work
	}

	int skip = 0;
	int vidx;  /* index in the version array currently using */
	int bytes = (base_block == last_block) ? 
		len : (BLOCK_SIZE - (offset % BLOCK_SIZE)); /* bytes to be read from first block */
	for (unsigned i = base_block; i <= last_block; i++) {
		unsigned j = 1;
		vidx = i - base_block; /* base_block is mapped to index 0 */

		if ( mode == READ_MODE || !((*modified)[i]) ) {
			// determine how many continus blocks we can read at one time
			for ( ; i + j <= last_block; j++ ) {
				if ( versions[vidx] != versions[i+j-base_block] || /* version changed */
					   (mode != READ_MODE && (*modified)[i +j]) )
					break;
				bytes += (i+j == last_block) ? 
					((offset + len) % BLOCK_SIZE) : BLOCK_SIZE;
			}

			// actually read this number of bytes
			//tmpi_error(DBG_INTERNAL, "Read block[%d] %ld bytes offset %ld.\n", i, bytes, offset+skip);
			if ( (ret = readBlock(offset+skip, buff+skip, versions[vidx], bytes)) < 0) {
				tmpi_error(DBG_INTERNAL, "Read block[%d] failed.\n", i);
				status = -1;
				break;
			}
		} else if (copyName != NULL) {
			// determine how many continus blocks we can read at one time
			for ( ; i + j <= last_block; j++ ) {
				if ( versions[vidx] != versions[i+j-base_block] || /* version changed */
						mode == READ_MODE || !(*modified)[i+j] ) 
					break;
				bytes += (i+j == last_block) ? 
					((offset + len) % BLOCK_SIZE) : BLOCK_SIZE;
			}

			// actually read this number of bytes
			//tmpi_error(DBG_INTERNAL, "Read block[%d] %ld bytes offset %ld.\n", i, bytes, offset+skip);
			if( (ret = fcache->read(copyName, offset+skip, buff+skip, bytes)) <0){
				tmpi_error(DBG_INTERNAL, "Read block[%d] failed.\n", i);
				status = -1;
				break;
			}
		}

		skip += ret;
		if (ret < bytes) break; /* we've reached end of file */

		bytes = (i+j == last_block) ? ((offset + len) % BLOCK_SIZE) : BLOCK_SIZE;
		i += j -1; /* i will be incremented by i++ */
	}

	delete versions;

	//tmpi_error(DBG_INTERNAL, "Read return %d bytes\n", skip);
	return (status==-1) ? status : skip;
}


int VerfsDescriptor::remove(){
	char comm[1024];
	comm[0] = 'r'; comm[1] = 'm'; comm[2] = ' ';
	initFName(&(comm[3]), wd, "*", guid);
	return system(comm);
}


int VerfsDescriptor::truncate(int length){
	bool success;

	if (mode != WRITE_MODE){
		tmpi_error(DBG_INTERNAL, "Cannot truncate, file opened for read.");
		return -1;
	}

	// copy the block that will be the last after truncate since it is modified
	write(length/*-length%BLOCK_SIZE*/, NULL, 0);

	// now do truncate
	fileEnd = length;
	success = fcache->truncate(copyName, length);
	bm->updateLength(length);	// blockmap persistently keeps this information

	return (success) ? length : -1;
}


VerfsDescriptor::VerfsDescriptor(char *wDir, guid_t &gid, File_Cache *fc) :
	bm(NULL),
	fcache(fc),
	guid(gid),
	wd(wDir),
	sid(-1),
	modified(NULL),
	mode(READ_MODE),
	copyName(NULL),
	fileEnd(0),
	version(0)
{
	char *bmName = new char[1024];

	if (bmName == NULL){
		tmpi_error(DBG_INTERNAL, "out of memory\n");
	}

	initFName(bmName, wDir, BM_EXTENSION, guid);
	
	bm = new BlockMap(bmName, fcache);
	fileEnd = bm->getLength();
	version = bm->getVersion();
	if (pthread_mutex_init(&fiLock, NULL) != 0) 
		throw -1;				// do we really need this lock
}

/**
 * Destructor
 */
VerfsDescriptor::~VerfsDescriptor(){ 
	if(modified != NULL) delete modified; 
	if(copyName != NULL) delete copyName;
	if(bm) delete bm;			// bmName is deleted by blockmap
}


/**
 * reads old block if necessary
 * @return length of write in case of success, -1 in case of failure
 */
int VerfsDescriptor::write(int offset, char *buff, int len){
	int writeSize;
	int status;

	if (modified == NULL || mode != WRITE_MODE){
		tmpi_error(DBG_CALLER, "Not in write mode.\n");
		return -1;
	}

	// copy partial block at the beginning
	if (offset % BLOCK_SIZE != 0){
		char tmpBuf[BLOCK_SIZE];
		read(((int)(offset/BLOCK_SIZE)) * BLOCK_SIZE, tmpBuf, BLOCK_SIZE);
		writeSize = (BLOCK_SIZE>offset+len)? 
			        (offset + len)%BLOCK_SIZE : BLOCK_SIZE;
		if(fcache->write(copyName, ((int)(offset/BLOCK_SIZE)) * BLOCK_SIZE,
						 tmpBuf, writeSize) < 0){
			tmpi_error(DBG_INTERNAL, "Copy first block for write failed.");
			status = -1;
		}
	}

	// copy partial block at the end
	if ((offset + len) % BLOCK_SIZE != 0){
		char tmpBuf[BLOCK_SIZE];
		read(((int)((offset+len)/BLOCK_SIZE))*BLOCK_SIZE, tmpBuf, BLOCK_SIZE);
		//tmpi_error(DBG_INTERNAL, "Read. %s %s\n", tmpBuf, copyName);
		writeSize = BLOCK_SIZE;
		if (fcache->write(copyName,((int)((offset+len)/BLOCK_SIZE))*BLOCK_SIZE,
						  tmpBuf, writeSize) <0){
			tmpi_error(DBG_INTERNAL, "Copy last block for write failed.");
			status = -1;
		}
	}

	// perform write
	if(fcache->write(copyName, offset, buff, len) < 0){
		tmpi_error(DBG_INTERNAL, "Write failed.");
		status = -1;
	}
	
	// update modified blocks index
	int begOffset = offset - offset % BLOCK_SIZE;
	int totLen = len + offset % BLOCK_SIZE;
	for(int i=begOffset / BLOCK_SIZE;
		i< (ceil( ((float) (begOffset + totLen)) / (float)BLOCK_SIZE));
		i ++)
   	{
   	    modified->setElement(i, true);
   	}

	// update file length
	if (offset + len > fileEnd)
		fileEnd = offset + len;
	
	return (status==-1) ? status : len;
}
